/* 
 * secsplit.c
 *
 * Shamir secret sharing
 *
 * Based on "How to Share a Secret", by Adi Shamir, Communications of
 * the ACM, November, 1979, Volume 22, Number 11, page 612.
 *
 * Original copyright:
 * Copyright (C) 1993 Hal Finney, 74076.1041@compuserve.com.
 * This software is being placed in the public domain.
 *
 * This program divides a file into n pieces, such that any k of them
 * are sufficient to reconstruct the original file, but that k-1
 * pieces give NO information about the original file (except its
 * length).
 *
 * It has been written for and tested on DOS and Unix systems.
 *
 * To split up a file, use the command; secsplit n k file where n is
 * the number of pieces to split it up to, and k is the number of
 * pieces needed to reconstruct it. k should be <= n and > 1.  If you
 * get n and k backwards the program will swap them for you so don't
 * worry too much about remembering the order.
 *
 * The program will output to file.001, file.002, .... If the file has
 * an extension (e.g. "file.c") the extension will be stripped off
 * before the ".001", etc., are added (so "file.c" will also output to
 * "file.001", etc.).
 *
 * To reconstruct a file, use the command:
 *    secsplit k file.*
 * or
 *    secsplit k file file1 file2...
 *
 * k should be >= the k used when the file was split; the minimum
 * number of pieces needed to reconstruct the file.  If you have too
 * few pieces then the program won't give an error, but you'll get the
 * wrong answer.
 *
 * The first command form is for DOS or other systems which won't
 * expand the ".*" for you; the program scans for file.000, file.001,
 * etc., and uses the first k of them that it finds.  In the second
 * form, the number of files given should be at least k, and again the
 * first k of them will be used.
 *
 * The output in the first form will be file.out; in the second form
 * it will be the first file on the command line, stripped of its
 * extension, and with ".out" added, so generally it will be file.out
 * too.
 *
 * Shamir's algorithm relies on cryptographically strong, unguessable,
 * random numbers.  This version of the program uses the arc4random(3)
 * interface.
 *
 * Revision history:
 *    Version 1.0	October 23, 1993
 *
 *    Version 1.1	October 24, 1993
 *       Added IDEA-based random-number-generator, initialized by MD5
 *       of input file, plus the time of day.
 *
 *    Version 1.2	March 31, 2000 Damien Miller <djm@mindrot.org>
 *       Removed IDEA-based random-number generator in favour of 
 *       /dev/random. Removed MSDOSisms. Rearranged source code and 
 *       formatting according to OpenBSD style(9)
 *
 *
 * The file formats used for output are as follows.  Each file starts
 * with one byte which is the index, from 1 through n, of that file.
 * This is the x value used for the polynomial evaluation in Shamir's
 * algorithm.  The files then consist of a series of 16-bit values
 * (high byte first), which are the result of applying Shamir's
 * splitting algorithm to the input file taken in 16-bit chunks.  The
 * prime used is slightly less than 2^16, meaning that input data
 * values close to 2^16 get turned into a pair of values (see below
 * for more information on this expansion).
 *
 * If the input file is of even length, the output files will each by
 * of that length+1 (because of the 1 byte at the beginning).  If the
 * input file is of odd length, the input is padded with a random byte
 * and processed normally to get a pair of output bytes, then each
 * output file is padded with an extra random byte to indicate this
 * fact.  So output files which have an odd length correspond to input
 * files with an even length, and vice versa.
 *
 * The output files could have encoded k and n information, but this
 * could be helpful to an attacker (he would know when he was close to
 * having enough pieces to reconstruct the file if he knew k).  So the
 * user is required to remember this himself.
 *
 * The file formats are system-independent so files split on one kind
 * of machine should be able to be reassembled on another kind of
 * machine.
 *
 * One complication exists with this code.  The algorithm must work
 * mod a prime which is bigger than all the data.  We want the data to
 * be about 16 bits therefore, so that we can multiply two numbers and
 * have them fit in a 32-bit int.  This means that the prime has to be
 * less than 2^16 in order for them to fit.  The largest such prime is
 * 65521.  This means that any data value > 65521 can't be processed
 * directly; instead, I split it into two values, a 65520, followed by
 * value-65520.  Also, 65520 itself is split as 65520 followed by 0.
 * This complication is pretty well limited to the input routine on
 * splitting and the output routine on assembly.
 *
 * The purpose of the "magic" is to xor the incoming value with a
 * different number each time.  This is purely intended to deal with
 * the problem that values of d >= limit-1 must be stored as four
 * bytes rather than two.  I was worried that some file might have a
 * lot of 0xff's which would balloon in size.  So I wanted to xor with
 * a rather random number.  Then, I change the random number each time
 * in order to make it less likely that some other file would happen
 * to have a lot of ballooning values.  This way it would be very rare
 * that a file managed to track my magic number in such a way that
 * many d's were over the size threshold.
 *
 * (The "magic" is not intended to increase the security or the
 * cryptographic strength of the algorithm in any way; it is purely to
 * keep the size of the output files from being much bigger than the
 * input.)
 *
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/mman.h>

#include <err.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "ssss.h"

#define LARGEST_PRIME 65521

#define S4_SMALLEST_INPUT	5

#define S4_FILE_MAGIC	"s4"

enum {
	SSSS_MEM_IO		= 0,
	SSSS_MAPPED_IO		= 1,
	SSSS_FILE_IO		= 2,

	SSSS_MAGIC_LENGTH	= 4
};

#ifndef USE_ARG
#define USE_ARG(x)     /*LINTED*/(void)&(x)
#endif

#ifndef __arraycount
#define	__arraycount(__x)	(sizeof(__x) / sizeof(__x[0]))
#endif

#if LARGEST_PRIME > 65536
#define S4_IO_SPACE(x)		(((x) * 3) / 2)
#else
#define S4_IO_SPACE(x)		(x)
#endif

/* Multiplicative inverses mod LARGEST_PRIME */
static const unsigned invtab[] = {
	    1, 32761, 43681, 49141, 52417, 54601, 56161, 57331,	/*   8 */
	58241, 58969, 11913, 60061, 60481, 60841, 61153, 61426, /*  16 */
	42396, 61881,  6897, 62245, 62401, 38717, 11395, 62791, /*  24 */
	49796, 63001, 41254, 63181, 58743, 63337, 25363, 30713, /*  32 */
	 3971, 21198, 63649, 63701, 54896, 36209, 63841, 63883, /*  40 */
	43148, 63961,  6095, 52119, 64065, 38458, 43216, 64156, /*  48 */
	 8023, 24898, 14132, 64261,  4945, 20627, 28591, 64351, /*  56 */
	 2299, 62132, 21100, 64429, 27927, 45442, 64481, 48117, /*  64 */
	64513, 34736, 26404, 10599, 47479, 64585,  5537, 64611, /*  72 */
	27824, 27448, 38439, 50865, 11062, 64681, 41469, 64702, /*  80 */
	57432, 21574, 48154, 64741, 60896, 35808, 19581, 58820, /*  88 */
	50061, 64793, 64801, 19229, 52135, 21608, 40692, 32078, /*  96 */
	52687, 36772, 23164, 12449, 53844,  7066, 60432, 64891, /* 104 */
	64897, 35233, 15921, 43074,  5410, 47056, 40139, 64936, /* 112 */
	 3479, 33910,  2279, 31066, 64961, 10550, 34137, 64975, /* 120 */
	 1083, 46724, 36223, 22721, 62376, 65001, 53655, 56819, /* 128 */
	23872, 65017, 53017, 17373, 47786, 13202, 21355, 38060, /* 136 */
	43043, 56500,  3771, 65053, 58086, 35529, 41237, 65066, /* 144 */
	37957, 13912, 46355, 13724, 47052, 51980, 40354, 58193, /* 152 */
	26551,  5531, 31281, 65101,  1252, 53495, 45329, 32351, /* 160 */
	10988, 28716, 39393, 10787, 53211, 24077, 16086, 65131, /* 168 */
	44973, 30448, 44447, 17904, 29920, 42551, 25834, 29410, /* 176 */
	50714, 57791, 18668, 65157,   362, 65161,  9309, 42375, /* 184 */
	63396, 58828, 27680, 10804, 43334, 20346, 57288, 16039, /* 192 */
	13240, 59104, 65185, 18386, 54878, 11582, 64204, 38985, /* 200 */
	52482, 26922, 17752,  3533, 34838, 30216, 59507, 65206, /* 208 */
	  627, 65209,  5900, 50377, 23686, 40721,  1219, 21537, /* 216 */
	50424,  2705, 31115, 23528, 53662, 52830, 39959, 32468, /* 224 */
	12813, 34500, 10391, 16955, 60657, 33900, 47368, 15533, /* 232 */
	55960, 65241, 61060,  5275, 13823, 49829, 54281, 65248, /* 240 */
	19031, 33302, 19144, 23362, 27813, 50872, 30771, 44121, /* 248 */
	59732, 31188,  6526, 65261, 54644, 59588, 42139, 61170  /* 256 */
};

typedef struct s4_t {
	int		oddsizefile;	/* last return included a pad */
	int		peek[SSSS_MAX_SHARES]; /* for two byte lookahead */
	unsigned	sharesc;	/* # of shares */
	unsigned	init[SSSS_MAX_SHARES];	/* initialised header */
} s4_t;

/* threshold header written to output file when splitting */
typedef struct s4_head_t {
	char		magic[SSSS_MAGIC_LENGTH];	/* magic string */
	uint8_t		coeff;		/* coefficient of this share */
	uint8_t		pad[3];		/* padding character */
	uint64_t	size;		/* size of original file */
} s4_head_t;

/* io stuff */
#define BSWAP64(x)      ((((x) & 0xffULL) << 56) |			\
                         (((x) & 0xff00ULL) << 40) |			\
                         (((x) & 0xff0000ULL) << 24) |			\
                         (((x) & 0xff000000ULL) << 8) |			\
                         (((x) & 0xff00000000ULL) >> 8) |		\
                         (((x) & 0xff0000000000ULL) >> 24) |		\
                         (((x) & 0xff000000000000ULL) >> 40) |		\
                         (((x) & 0xff00000000000000ULL) >> 56))

/* do 64bit integers in little-endian order */
static void
swapheader(s4_head_t *head)
{
        int     indian = 1;

        head->size = (*(char *)(void *)&indian) ? head->size : (uint64_t)BSWAP64(head->size);
}

/* check we have a sane header */
static int
sane_header(s4_head_t *head, s4_head_t *refhead)
{
	return (head->size == refhead->size);
}

/* read a header from the share, and return any info desired */
static int
get_header(s4_head_t *head, const uint8_t *mem)
{
	if (mem == NULL) {
		return 0;
	}
	(void) memcpy(head, mem, sizeof(*head));
	swapheader(head);
	return 1;
}

/* format the output header for split files */
static int
fmt_header(s4_head_t *head, const char *type, uint64_t size)
{
	(void) memset(head, 0x0, sizeof(*head));
	(void) snprintf(head->magic, sizeof(head->magic), "%s", type);
	head->size = size;
	swapheader(head);
	return 1;
}

/* allocate space from heap */
static int
allocate(ssss_str_t *share, const void *data, size_t size)
{
	if ((share->io.base = calloc(1, size)) == NULL) {
		return 0;
	}
	if (data) {
		(void) memcpy(share->io.base, data, size);
	}
	share->io.size = size;
	return 1;
}

/******************* Code related to splitting *********************/

/* Return a random number from 0 to x-1. */
#define CRANDOM(x)	(uint32_t)(arc4random() % (x))

/*
 * Evaluate the given polynomial, n coefficients, at point x=i.
 * Do it mod the LARGEST_PRIME
 *
 * poly = Polynomial coefficients
 * n = # coefficients (order of polynomial + 1)
 * i = Point to evaluate it at
 */
static uint32_t
eval(uint32_t *poly, unsigned n, unsigned i)
{
	uint32_t	prod;		/* Accumulated product */
	int		j;		/* index */

	prod = poly[n - 1];
	for (j = (n - 1); --j >= 0; ) {
		prod *= i;
		prod += poly[j];
		prod %= LARGEST_PRIME;
	}
	return prod;
}

#define SSSS_EOF		-1
#define SSSS_ODD_SIZED_FILE	-2

/* return the next character in the stream/memory array */
static int
ssss_getch(ssss_str_t *str, uint64_t last)
{
	int	ch;

	switch(str->iotype) {
	case SSSS_MEM_IO:
	case SSSS_MAPPED_IO:
		return (str->c == last) ? SSSS_EOF : (int)(str->io.base[str->c++]);
	case SSSS_FILE_IO:
		return ((ch = fgetc(str->fp)) == EOF) ? SSSS_EOF : ch;
	default:
		warn("weird method '%d'", str->iotype);
		return -1;
	}
}

/*
 * Return a 16-bit value from input
 */
static int
get_limited_16(s4_t *s4, ssss_str_t *str, unsigned *d)
{
	uint8_t	c[2];
	int	ch;

	/* Check if last return included a pad */
	if (s4->oddsizefile) {
		return SSSS_ODD_SIZED_FILE;
	}
	/* Read data (bigendian), do the magic */
	if ((ch = ssss_getch(str, str->io.size)) == SSSS_EOF) {
		return SSSS_EOF;
	}
	c[0] = (uint8_t)ch;
	if ((ch = ssss_getch(str, str->io.size)) == SSSS_EOF) {
		ch = CRANDOM(0x100);
		s4->oddsizefile = 1;
	}
	c[1] = (uint8_t)ch;
	*d = (((uint8_t)c[0] & 0xff) << 8) | ((uint8_t)c[1] & 0xff);
	return 1;
}

/*
 * Given a 24-bit value d, split it into sharesc files such that any k
 * of them can reconstruct it.
 */
static void
split_out(uint32_t d, unsigned sharesc, unsigned k, ssss_str_t *shares)
{
	uint32_t	poly[SSSS_MAX_SHARES];
	uint32_t	di;
	unsigned	i;
	unsigned	j;

	poly[0] = d;
	for (j = 1; j < k; j++) {
		poly[j] = CRANDOM(LARGEST_PRIME);
	}
	for (i = 0; i < sharesc; i++) {
		di = eval(poly, k, i + 1);
		if (/*CONSTCOND*/LARGEST_PRIME > 65536) {
			di |= ((CRANDOM(0x100) & 0xfe) << 16);
			shares[i].io.base[shares[i].c++] = (di >> 16) & 0xff;
		}
		shares[i].io.base[shares[i].c++] = (di >> 8) & 0xff;
		shares[i].io.base[shares[i].c++] = (di & 0xff);
	}
}

/*
 * Split the specified input file into sharesc output files, such that
 * any k of them are sufficient to reconstruct the input.  This is the
 * main routine for the splitting case.
 *
 * k = Threshhold for re-assembly
 */
static void
split(s4_t *s4, ssss_str_t *str, unsigned sharesc, unsigned k, ssss_str_t *shares, unsigned doheader)
{
	s4_head_t	head;
	unsigned	i;
	unsigned	d;

	if (doheader) {
		fmt_header(&head, S4_FILE_MAGIC, str->io.size);
	}
	for (i = 0; i < sharesc; i++) {
		/* Prefix each file with "x" coordinate, 1 byte */
		head.coeff = i + 1;
		if (doheader) {
			/* add header */
			(void) memcpy(&shares[i].io.base[shares[i].c], &head, sizeof(head));
			shares[i].c = sizeof(head);
		}
	}
	while (get_limited_16(s4, str, &d) != SSSS_EOF) {
		split_out(d, sharesc, k, shares);
	}
}

/******************* Code related to assembly *********************/

/*
 * Return the multiplicative inverse of small positive or negative value
 * x modulo the current prime.  Do a table lookup for speed. 
 *
 * x = Small value to find inverse of
 */
static uint32_t
inverse(int x)
{
	int	negative;

	negative = 0;
	if (x < 0) {
		x = -x;
		negative = 1;
	}
	if (x < 1 || x > (int)__arraycount(invtab)) {
		warn("inverse out of range %d", x);
		return 0;
	}
	return (negative) ? LARGEST_PRIME - invtab[x-1] : invtab[x-1];
}

/*
 ( Interpolate the polynomial specified at x and y coordinates
 * in the array of size n, at x=i.  Do it mod the specified modulus.
 * This algorithm is from Knuth, The Art of Computer Programming, Vol. 2,
 * Seminumerical Algorithms, section 4.6.4, Evaluation of Polynomials,
 * equation 43 and 44, page 485 (bottom) in 1981 hardcover edition.
 * Knuth has his indices go from 0 to n; mine go from 0 to n-1, a slight
 * notational change.
 *
 * x coordinate of interpolated point is 0
 * x[] = x coordinates of known points
 * y[] = y coordinates of known points
 * n = size of x, y, and alpha arrays
 * mod = modulus for reducing results
 */
static uint32_t 
interpolate(int *x, uint32_t *y, unsigned n)
{
	uint32_t	alpha[SSSS_MAX_SHARES];
	uint32_t	product;
	unsigned	i;
	unsigned	j;

	(void) memcpy(alpha, y, n * sizeof(alpha[0]));
	for (i = 1; i < n; i++) {
		for (j = n - 1; j >= i; j--) {
			if (alpha[j] > alpha[j - 1]) {
				alpha[j] = alpha[j] - alpha[j - 1];
			} else {
				alpha[j] = alpha[j] - alpha[j - 1] + LARGEST_PRIME;
			}
			alpha[j] *= inverse(x[j] - x[j - i]);
			alpha[j] = alpha[j] % LARGEST_PRIME;
		}
	}
	product = alpha[n - 1];
	for (i = n - 2; (int)i >= 0; i--) {
		product *= -x[i] + ((x[i] > 0) ? LARGEST_PRIME : 0);
		product += alpha[i];
		product %= LARGEST_PRIME;
	}
	return product;
}

/*
 * Return a 16-bit-value from the file, with a flag set if only the
 * high 8 bits of the reconstructed value will be valid.  This is known
 * because odd-size files are padded with an extra byte.  We have to stay
 * two bytes ahead to know this.
 */
static int 
get_assemble_16(s4_t *s4, ssss_str_t *inputs, unsigned coeff)
{
	unsigned	j;
	int		charsize;
	int		ret;
	int		ch;

	charsize = (/*CONSTCOND*/ LARGEST_PRIME > 65536) ? 3 : 2;
	if (!s4->init[coeff]) {
		s4->init[coeff] = 1;
		s4->peek[coeff] = 0;
		for (j = 0 ; j < (unsigned)charsize ; j++) {
			ch = ssss_getch(&inputs[coeff], inputs[coeff].io.size + charsize);
			s4->peek[coeff] = (s4->peek[coeff] << 8) | (ch & 0xff);
		}
	}
	ret = s4->peek[coeff];
	s4->peek[coeff] = 0;
	for (j = 0 ; j < (unsigned)charsize ; j++) {
		if ((ch = ssss_getch(&inputs[coeff], inputs[coeff].io.size + charsize)) == SSSS_EOF) {
			s4->peek[coeff] = SSSS_EOF;
			if (j > 0) {
				s4->oddsizefile = 1;
			}
			break;
		}
		s4->peek[coeff] = (s4->peek[coeff] << 8) | ((unsigned)ch & 0xff);
	}
	return ret;
}

/*
 * Get 16-bit values from each file, and interpolate them to get the value
 * at x=0.  That is how we do the assembly.  Set the podd flag if
 * only the high 8 bits of this interpolated value are valid.
 */
static int
get_assemble(s4_t *s4, unsigned nin, ssss_str_t *inputs, int *coeffs, unsigned *ret)
{
	unsigned	i;
	int		y[SSSS_MAX_SHARES];

	for (i = 0; i < nin; i++) {
		if ((y[i] = get_assemble_16(s4, inputs, i)) == SSSS_EOF) {
			return 0;
		}
	}
	*ret = interpolate(coeffs, (unsigned *)y, nin);
	return 1;
}

/*
 * Given a set of nin input descriptors, assemble them to generate the
 * original file.  This is the main routine for the assembly case.
 */
static int
assemble(s4_t *s4, ssss_str_t *inputs, unsigned nin, ssss_str_t *output, unsigned doheader)
{
	s4_head_t	heads[SSSS_MAX_SHARES];
	unsigned	ch;
	unsigned	i;
	int		coeffs[SSSS_MAX_SHARES];

	for (i = 0; i < nin; i++) {
		if (doheader) {
			get_header(&heads[i], inputs[i].io.base);
			if (!sane_header(&heads[i], &heads[0])) {
				warn("insane ssss header %u", i);
				return 0;
			}
			inputs[i].c += sizeof(heads[i]);
			coeffs[i] = heads[i].coeff;
		} else {
			coeffs[i] = i;
		}
	}
	while (get_assemble(s4, nin, inputs, coeffs, &ch)) {
		output->io.base[output->c++] = (ch >> 8) & 0xff;
		if (s4->oddsizefile) {
			break;
		}
		output->io.base[output->c++] = (ch & 0xff);
	}
	return 1;
}

/* check shares counts and threshold values are sane */
static int
check_values(unsigned threshold, unsigned sharesc)
{
	if (sharesc <= 1 || sharesc > SSSS_MAX_SHARES) {
		warn("Invalid number of shares (%u).", sharesc);
		return 0;
	}
	if (threshold < 1 || threshold > SSSS_MAX_SHARES) {
		warn("Invalid threshold number %u", threshold);
		return 0;
	}
	if (threshold > sharesc) {
		warn("threshold (%u) is larger than sharesc (%u)", threshold, sharesc);
		return 0;
	}
	return 1;
}

/* split the secret file into shares */
static int 
split_memory(ssss_t *ssss, const void *secret, size_t memsize, unsigned threshold, unsigned sharesc, unsigned doheader)
{
	ssss_str_t	mem;
	unsigned	i;
	char		small[8];
	s4_t		*s4;

	s4 = ssss->handle;
	if (!check_values(threshold, sharesc)) {
		return 0;
	}
	(void) memset(&mem, 0x0, sizeof(mem));
	if ((mem.io.size = memsize) < S4_SMALLEST_INPUT) {
		(void) memset(small, 0x0, sizeof(small));
		(void) memcpy(small, secret, memsize = mem.io.size = S4_SMALLEST_INPUT);
		mem.io.base = (uint8_t *)small;
	} else {
		mem.io.base = (uint8_t *)__UNCONST(secret);
	}
	/* allocate space for outputs */
	ssss->sharesc = (unsigned)sharesc;
	for (i = 0 ; i < sharesc ; i++) {
		/* possibly need space to grow for padded chars */
		ssss->shares[i].io.size = S4_IO_SPACE(memsize + sizeof(s4_head_t) + 2);
		ssss->shares[i].c = 0;
		if (ssss->shares[i].io.base == NULL &&
		    (ssss->shares[i].io.base = calloc(1, ssss->shares[i].io.size)) == NULL) {
			warn("Bad alloc");
			return 0;
		}
	}
	/* Do the work */
	split(s4, &mem, sharesc, threshold, ssss->shares, doheader);
	return 1;
}

/* reconstruct the shares from memory */
static int 
join_memory(ssss_t *ssss, ssss_str_t *input, size_t bytes, unsigned filec, ssss_str_t *str, unsigned doheader)
{
	s4_t	*s4;

	s4 = ssss->handle;
	(void) memset(str, 0x0, sizeof(*str));
	if (str->io.base == NULL &&
	    (str->io.base = calloc(1, str->io.size = bytes - sizeof(s4_head_t))) == NULL) {
		warn("Bad alloc %zu bytes", bytes);
		return 0;
	}
	/* Do the work */
	return assemble(s4, input, filec, str, doheader);
}

/**************************************************************************/
/* external functions */
/**************************************************************************/

/* initialise */
int
ssss_init(ssss_t *ssss, unsigned threshold, unsigned sharesc)
{
	s4_t	*s4;

	ssss->sharesc = sharesc;
	ssss->threshold = threshold;
	if ((s4 = calloc(1, sizeof(*s4))) == NULL) {
		warn("Unable to allocate ssss structure");
		return 0;
	}
	s4->sharesc = SSSS_MAX_SHARES; /* guess for just now */
	ssss->handle = s4;
	return 1;
}

/* add a share */
int
ssss_add_share(ssss_t *ssss, unsigned subscr, const void *data, ssize_t size)
{
	struct stat	 st;
	ssss_str_t	*share;
	unsigned	 n;
	ssize_t		 rc;
	ssize_t		 cc;
	FILE		*fp;

	n = (subscr == SSSS_SPLIT_SOURCE) ? subscr : ssss->availc++;
	share = &ssss->shares[n];
	if (size < 0) {
		if ((fp = fopen(data, "r")) == NULL) {
			warn("can't open file '%s'", (const char *)data);
			return 0;
		}
		(void) fstat(fileno(fp), &st);
		share->io.size = (size_t)st.st_size;
		share->io.base = mmap(NULL, share->io.size, PROT_READ, MAP_PRIVATE, fileno(fp), 0);
		if (share->io.base == MAP_FAILED) {
			if (!allocate(share, NULL, (size_t)st.st_size)) {
				ssss->shares[n].iotype = SSSS_FILE_IO;
				ssss->shares[n].fp = fp;
				return 1;
			}
			for (cc = 0 ; cc < st.st_size ; cc += rc) {
				if ((rc = read(fileno(fp), &share->io.base[cc], (size_t)(st.st_size - cc))) < 0) {
					break;
				}
			}
			ssss->shares[n].iotype = SSSS_MEM_IO;
		} else {
			ssss->shares[n].iotype = SSSS_MAPPED_IO;
		}
		(void) fclose(fp);
	} else {
		if (!allocate(share, data, (size_t)size)) {
			warn("can't allocate memory for %zd bytes", size);
			return 0;
		}
		ssss->shares[n].iotype = SSSS_MEM_IO;
	}
	return 1;
}

/* split via different algorithms */
int
ssss_split(ssss_t *ssss)
{
	ssss_str_t	*mem;

	mem = &ssss->shares[SSSS_MAX_SHARES];
	return split_memory(ssss, mem->io.base, mem->io.size, ssss->threshold, ssss->sharesc, 1);
}

/* split vector via iovec interface */
int
ssss_split_vec(ssss_t *ssss, const char *buf, size_t size, struct iovec *iov, int iovc)
{
	int	ret;
	int	i;

	ssss_add_share(ssss, (unsigned)SSSS_SPLIT_SOURCE, buf, (long)size);
	ret = split_memory(ssss, buf, size, ssss->threshold, ssss->sharesc, 0);
	for (i = 0 ; ret && i < iovc ; i++) {
		ssss_get_share(ssss, (unsigned)i, &iov[i].iov_base, &iov[i].iov_len);
	}
	return ret;
}

/* combine/join/decode into one */
int
ssss_combine(ssss_t *ssss)
{
	s4_head_t	 head;
	s4_t		*s4;

	(void) memset(&head, 0x0, sizeof(head));
	if (!get_header(&head, ssss->shares[0].io.base)) {
		return -1;
	}
	if ((s4 = calloc(1, sizeof(*s4))) == NULL) {
		warn("Unable to allocate ssss structure");
		return 0;
	}
	s4->sharesc = SSSS_MAX_SHARES; /* guess for just now */
	ssss->handle = s4;
	if (!join_memory(ssss, ssss->shares, 
			ssss->shares[0].io.size - 2, ssss->availc,
			&ssss->shares[SSSS_JOIN_DEST], 1)) {
		return -1;
	}
	return (int)ssss->shares[SSSS_MAX_SHARES].c;
}

/* combine a split vector via iovec interface */
int
ssss_combine_vec(ssss_t *ssss, struct iovec *iov, int iovc, char *buf, size_t size)
{
	s4_t	*s4;
	void	*vp;
	int	 i;

	for (i = 0 ; i < iovc ; i++) {
		ssss_add_share(ssss, (unsigned)i, iov[i].iov_base, (long)iov[i].iov_len);
	}
	if ((s4 = calloc(1, sizeof(*s4))) == NULL) {
		warn("Unable to allocate ssss structure");
		return 0;
	}
	s4->sharesc = SSSS_MAX_SHARES; /* guess for just now */
	ssss->handle = s4;
	if (!join_memory(ssss, ssss->shares, 
			ssss->shares[0].io.size - 2, ssss->availc,
			&ssss->shares[SSSS_JOIN_DEST], 0)) {
		return -1;
	}
	vp = buf;
	ssss_get_share(ssss, SSSS_JOIN_DEST, &vp, &size);
	free(s4);
	return (int)ssss->shares[SSSS_MAX_SHARES].c;
}

/* write a share to a file */
int
ssss_write_share(ssss_t *ssss, unsigned n, const char *f)
{
	ssss_str_t	*out;
	ssize_t		 wc;
	size_t		 last;
	size_t		 cc;
	FILE		*fp;
	int		 ok;

	if (n >= ssss->sharesc && n != SSSS_SPLIT_SOURCE) {
		warn("share %u out of range (max %u)", n, ssss->sharesc);
		return 0;
	}
	out = &ssss->shares[n];
	if (f == NULL) {
		fp = stdout;
	} else {
		if ((fp = fopen(f, "w")) == NULL) {
			warn("can't write to '%s'", f);
			return 0;
		}
	}
	last = (n == SSSS_JOIN_DEST) ? out->c - 2 : out->c;
	for (ok = 1, cc = 0 ; ok && cc < last ; cc += wc) {
		if ((wc = write(fileno(fp), &out->io.base[cc], last - cc)) <= 0) {
			warn("Short write");
			ok = 0;
		}
	}
	if (f != NULL) {
		(void) fclose(fp);
	}
	return ok;
}

/* get a share */
int
ssss_get_share(ssss_t *ssss, unsigned n, void **outp, size_t *size)
{
	ssss_str_t	*share;

	if (n >= ssss->sharesc && n != SSSS_SPLIT_SOURCE) {
		warn("share %u out of range (max %u)", n, ssss->sharesc);
		return 0;
	}
	share = &ssss->shares[(n == SSSS_SPLIT_SOURCE) ? SSSS_MAX_SHARES : n];
	*outp = (char *)share->io.base;
	*size = share->c;
	return 1;
}

/* finish up */
int
ssss_end(ssss_t *ssss)
{
	unsigned	 i;
	s4_t		*s4;

	s4 = ssss->handle;
	free(s4);
	ssss->handle = NULL;
	for (i = 0 ; i < ssss->sharesc ; i++) {
		switch (ssss->shares[i].iotype) {
		case SSSS_MAPPED_IO:
			(void) munmap(ssss->shares[i].io.base, ssss->shares[i].io.size);
			break;
		case SSSS_MEM_IO:
			free(ssss->shares[i].io.base);
			break;
		case SSSS_FILE_IO:
			(void)fclose(ssss->shares[i].fp);
			break;
		default:
			break;
		}
	}
	return 1;
}
