/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/stat.h>

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "circa.h"
#include "mat.h"
#include "netpgp.h"
#include "bag.h"

static int
copyfile(const char *to, const char *from)
{
	ssize_t	 cc;
	FILE	*fpin;
	FILE	*fpout;
	char	 buf[BUFSIZ * 10];
	int	 ok;

	if ((fpin = fopen(from, "r")) == NULL) {
		(void) fprintf(stderr, "can't read '%s'\n", from);
		return 0;
	}
	if ((fpout = fopen(to, "w")) == NULL) {
		(void) fprintf(stderr, "can't write '%s'\n", to);
		(void) fclose(fpin);
		return 0;
	}
	for (ok = 1; ok && (cc = read(fileno(fpin), buf, sizeof(buf))) > 0 ; ) {
		if (write(fileno(fpout), buf, (size_t)cc) != cc) {
			(void) fprintf(stderr, "short write\n");
			ok = 0;
		}
	}
	(void) fclose(fpin);
	(void) fclose(fpout);
	return ok;
}

static int
cmpfile(const char *f1, const char *f2)
{
	ssize_t	 cc1;
	ssize_t	 cc2;
	FILE	*fp1;
	FILE	*fp2;
	char	 buf1[BUFSIZ * 10];
	char	 buf2[BUFSIZ * 10];
	int	 ok;

	if ((fp1 = fopen(f1, "r")) == NULL) {
		(void) fprintf(stderr, "can't read '%s'\n", f1);
		return 0;
	}
	if ((fp2 = fopen(f2, "r")) == NULL) {
		(void) fprintf(stderr, "can't write '%s'\n", f2);
		(void) fclose(fp1);
		return 0;
	}
	for (ok = 1; ok ; ) {
		cc1 = read(fileno(fp1), buf1, sizeof(buf1));
		cc2 = read(fileno(fp2), buf2, sizeof(buf2));
		if (cc1 != cc2 || memcmp(buf1, buf2, (size_t)cc1) != 0) {
			ok = 0;
		}
		if (cc1 == 0) {
			break;
		}
	}
	(void) fclose(fp1);
	(void) fclose(fp2);
	return ok;
}

/*****************************************************************************/

int
bag_pack(char *bag, char *matfile, char *signer)
{
	netpgp_t	 netpgp;
	circa_t		 circa;
	mat_t		 mat;
	FILE		*listing;
	char		 copy[1024];
	char		 fsig[1024];
	char		 cmd[1024];
	char		 b1[1024];
	char		 b2[1024];
	char		 d[1024];
	char		 f[1024];
	char		*fp;
	char		*fsigp;
	char		*copyp;
	int		 ret;

	ret = 0;
	(void) memset(copy, 0x0, sizeof(copy));
	(void) memset(fsig, 0x0, sizeof(fsig));
	(void) memset(b1, 0x0, sizeof(b1));
	(void) memset(b2, 0x0, sizeof(b2));
	(void) memset(f, 0x0, sizeof(f));
	(void) snprintf(d, sizeof(d), "/tmp/bag.XXXXXX");
	if (mkdtemp(d) == NULL) {
		(void) fprintf(stderr, "can't mkdir '%s'\n", d);
		goto done;
	}
	(void) memset(&mat, 0x0, sizeof(mat));
	mat.verbose = 2;
	if (!mat_init(&mat, matfile, "read")) {
		(void) fprintf(stderr, "can't open '%s'\n", matfile);
		goto done;
	}
	(void) snprintf(f, sizeof(f), "%s/inventory", d);
	if ((listing = fopen(f, "w")) == NULL) {
		(void) fprintf(stderr, "can't write '%s'\n", f);
		mat_end(&mat);
		goto done;
	}
	if (!mat_list(&mat, NULL, listing)) {
		(void) fprintf(stderr, "can't mat_list '%s'\n", matfile);
		(void) fclose(listing);
		mat_end(&mat);
		goto done;
	}
	mat_end(&mat);
	(void) fclose(listing);
	/* sign inventory */
	(void) memset(&netpgp, 0x0, sizeof(netpgp));
	(void) snprintf(fsig, sizeof(fsig), "%s.gpg", f);
	netpgp_setvar(&netpgp, "hash", "sha256");
	netpgp_setvar(&netpgp, "max mem alloc", "4194304");
	netpgp_setvar(&netpgp, "need seckey", "1");
	netpgp_setvar(&netpgp, "need userid", "1");
	netpgp_set_homedir(&netpgp, getenv("HOME"),
		netpgp_getvar(&netpgp, "ssh keys") ? "/.ssh" : "/.gnupg", 1);
	if (!netpgp_init(&netpgp)) {
		(void) fprintf(stderr, "can't initialise netpgp\n");
		goto done;
	}
	if (!netpgp_sign_file(&netpgp, signer, f, fsig, 0, 0, 0)) {
		(void) fprintf(stderr, "can't sign inventory\n");
		goto done;
	}
	netpgp_end(&netpgp);
	/* make into a mat archive */
	fp = strrchr(f, '/') + 1;
	fsigp = strrchr(fsig, '/') + 1;
	(void) snprintf(copy, sizeof(copy), "%s/archive.mat", d);
	copyfile(copy, matfile);
	copyp = strrchr(copy, '/') + 1;
	(void) snprintf(b1, sizeof(b1), "%s.b1", "archive");
	if (!mat_vtar(__UNCONST("mat"), "-C", d, "-cf", b1, fp, fsigp, copyp, NULL)) {
		(void) fprintf(stderr, "can't mat bag files\n");
		goto done;
	}
	/* compress */
	(void) snprintf(b2, sizeof(b2), "%s/%s.b2", d, "archive");
	(void) snprintf(cmd, sizeof(cmd), "xz < %s > %s", b1, b2); /* XXX */
	system(cmd);
	/* circa */
	(void) memset(&circa, 0x0, sizeof(circa));
	if (!circa_file(&circa, CIRCA_SECTOR_SIZE, 1, b2, bag)) {
		(void) fprintf(stderr, "can't circa bag file\n");
		goto done;
	}
	circa_end(&circa);
	ret = 1;
done:
	if (b1[0]) {
		(void) unlink(b1);
	}
	if (b2[0]) {
		(void) unlink(b2);
	}
	if (f[0]) {
		(void) unlink(f);
	}
	if (fsig[0]) {
		(void) unlink(fsig);
	}
	if (copy[0]) {
		(void) unlink(copy);
	}
	(void) rmdir(d);
	return ret;
}

int
bag_unpack(char *bag, char *matfile)
{
	netpgp_t	 netpgp;
	circa_t		 circa;
	mat_t		 mat;
	FILE		*listing;
	char		 matinv[1024];
	char		 fsig[1024];
	char		 copy[1024];
	char		 cmd[1024];
	char		 b1[1024];
	char		 b2[1024];
	char		 f[1024];
	char		 d[1024];
	int		 ret;

	ret = 0;
	(void) memset(fsig, 0x0, sizeof(fsig));
	(void) memset(f, 0x0, sizeof(f));
	(void) memset(b1, 0x0, sizeof(b1));
	(void) memset(b2, 0x0, sizeof(b2));
	(void) memset(matinv, 0x0, sizeof(matinv));
	(void) memset(copy, 0x0, sizeof(copy));
	(void) snprintf(d, sizeof(d), "/tmp/bag.XXXXXX");
	if (mkdtemp(d) == NULL) {
		(void) fprintf(stderr, "can't mkdir '%s'\n", d);
		goto done;
	}
	/* circa */
	(void) memset(&circa, 0x0, sizeof(circa));
	(void) snprintf(b1, sizeof(b1), "%s/%s.b1", d, "archive");
	if (!circa_file(&circa, CIRCA_SECTOR_SIZE, 0, bag, b1)) {
		(void) fprintf(stderr, "can't uncirca bag file '%s'\n", bag);
		goto done;
	}
	circa_end(&circa);
	/* compress */
	(void) snprintf(b2, sizeof(b2), "%s/%s.b2", d, "archive");
	(void) snprintf(cmd, sizeof(cmd), "xz -d < %s > %s", b1, b2); /* XXX */
	system(cmd);
	if (!mat_vtar(__UNCONST("mat"), "-C", d, "-xf", b2, NULL)) {
		(void) fprintf(stderr, "can't unmat bag file '%s'\n", b2);
		goto done;
	}
	/* verify signature */
	(void) memset(&netpgp, 0x0, sizeof(netpgp));
	netpgp_setvar(&netpgp, "hash", "sha256");
	netpgp_setvar(&netpgp, "max mem alloc", "4194304");
	netpgp_set_homedir(&netpgp, getenv("HOME"),
		netpgp_getvar(&netpgp, "ssh keys") ? "/.ssh" : "/.gnupg", 1);
	if (!netpgp_init(&netpgp)) {
		(void) fprintf(stderr, "can't initialise netpgp\n");
		goto done;
	}
	(void) snprintf(fsig, sizeof(fsig), "%s/inventory.gpg", d);
	if (!netpgp_verify_file(&netpgp, fsig, NULL, 0)) {
		(void) fprintf(stderr, "can't verify inventory '%s'\n", fsig);
		goto done;
	}
	netpgp_end(&netpgp);
	(void) snprintf(f, sizeof(f), "%.*s", (int)(strlen(fsig) - 4), fsig);
	/* check f is same as matinv */
	(void) memset(&mat, 0x0, sizeof(mat));
	(void) snprintf(copy, sizeof(copy), "%s/archive.mat", d);
	mat.verbose = 2;
	if (!mat_init(&mat, copy, "read")) {
		(void) fprintf(stderr, "can't open '%s'\n", copy);
		goto done;
	}
	(void) snprintf(matinv, sizeof(matinv), "%s/matinventory", d);
	if ((listing = fopen(matinv, "w")) == NULL) {
		(void) fprintf(stderr, "can't write '%s'\n", matinv);
		mat_end(&mat);
		goto done;
	}
	if (!mat_list(&mat, NULL, listing)) {
		(void) fprintf(stderr, "can't mat_list '%s'\n", matfile);
		(void) fclose(listing);
		mat_end(&mat);
		goto done;
	}
	mat_end(&mat);
	(void) fclose(listing);
	if (!cmpfile(f, matinv)) {
		(void) fprintf(stderr, "signature mismatch '%s'\n", bag);
		goto done;
	}
	/* copyfile archive.mat to mat */
	copyfile(matfile, copy);
	ret = 1;
done:
	if (b1[0]) {
		(void) unlink(b1);
	}
	if (b2[0]) {
		(void) unlink(b2);
	}
	if (fsig[0]) {
		(void) unlink(fsig);
	}
	if (f[0]) {
		(void) unlink(f);
	}
	if (copy[0]) {
		(void) unlink(copy);
	}
	if (matinv[0]) {
		(void) unlink(matinv);
	}
	(void) rmdir(d);
	return ret;
}
