/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <inttypes.h>
#include <rs.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "circa.h"

/* for a description of the transformations here, please refer to libcirca(3) */

/* some basic definitions */
enum {
	Q_PARITY_SIZE		= 4,
	P_PARITY_SIZE		= 4,

	C3_FRAME_SIZE		= 24,
	C2_FRAME_SIZE		= C3_FRAME_SIZE + Q_PARITY_SIZE,
	C1_FRAME_SIZE		= C2_FRAME_SIZE + P_PARITY_SIZE,

	Q_PARITY_OFF		= C3_FRAME_SIZE / 2
};

/* do a 5-way swap on bytes */
#define PERMUTE5(x, f, mac, i1, i2, i3, i4, i5) do {			\
	uint8_t		_tmp = (x)[mac(f, i1)];				\
	(x)[mac(f, i1)] = (x)[mac(f, i2)];				\
	(x)[mac(f, i2)] = (x)[mac(f, i3)];				\
	(x)[mac(f, i3)] = (x)[mac(f, i4)];				\
	(x)[mac(f, i4)] = (x)[mac(f, i5)];				\
	(x)[mac(f, i5)] = _tmp;						\
} while (/*CONSTCOND*/0)

/* 2-way byte swap */
#define SWAP(arr, i1, i2) do {						\
	uint8_t		_tmp = (arr)[i1];				\
	(arr)[i1] = (arr)[i2];						\
	(arr)[i2] = _tmp;						\
} while (/*CONSTCOND*/0)

/* offsets in work arrays for different level frames */
#define C3OFF(x, y)	((((x) % circa->maxframes) * C3_FRAME_SIZE) + (y))
#define C1OFF(x, y)	((((x) % circa->maxframes) * C1_FRAME_SIZE) + (y))

#ifndef USE_ARG
#define USE_ARG(x)	/*LINTED*/(void)&(x)
#endif

#define BSWAP32(x)	((((x) & 0xff) << 24) |				\
			 (((x) & 0xff00) << 8) |			\
			 (((x) & 0xff0000) >> 8) |			\
			 (((x) & 0xff000000) >> 24))

/* magic number - leet! */
#define CIRCA_MAGIC	0xc14ca000

/* byte-swap 32 bits, if necessary, to get little-endian */
static uint32_t
le32(uint32_t in)
{
	int	indian = 1;

	return (*(char *)(void *)&indian) ? in : BSWAP32(in);
}

/***************************************************************************/

/* perform initialisations */
int
circa_init(circa_t *circa, size_t sectorsize)
{
	(void) memset(&circa->c2, 0x0, sizeof(circa->c2));
	rs_init(&circa->c2, RS_DEFAULT_POLYNOMIAL, C3_FRAME_SIZE, Q_PARITY_SIZE);
	rs_init(&circa->c1, RS_DEFAULT_POLYNOMIAL, C2_FRAME_SIZE, P_PARITY_SIZE);
	if ((circa->header.sectorsize = (uint32_t)sectorsize) == 0) {
		circa->header.sectorsize = CIRCA_SECTOR_SIZE;
	}
	if (circa->header.sectorsize % C1_FRAME_SIZE != 0) {
		(void) fprintf(stderr, "Bad sector size %zu\n", sectorsize);
		return 0;
	}
	circa->maxframes = (uint32_t)(sectorsize / C1_FRAME_SIZE);
	circa->datasize = (uint32_t)((sectorsize * 3 / 4) - sizeof(uint32_t));
	if ((circa->c1data = calloc(1, sectorsize)) == NULL ||
	    (circa->c3data = calloc(1, circa->datasize + C3_FRAME_SIZE)) == NULL) {
		(void) fprintf(stderr, "can't allocate c1data/c3data\n");
		return 0;
	}
	return circa->init = 1;
}

/* encode the information given in the data buffer in 'in' */
ssize_t
circa_encode(circa_t *circa, const void *inp, size_t insize, uint8_t *out, size_t outsize)
{
	uint32_t	isize;
	uint8_t		qparity[C1_FRAME_SIZE];
	int		f;
	int		b;

	if (!circa->init) {
		circa_init(circa, outsize);
	}
	isize = le32((uint32_t)insize);
	(void) memcpy(circa->c3data, &isize, sizeof(isize));
	(void) memcpy(&circa->c3data[sizeof(isize)], inp, insize);
	(void) memset(circa->c1data, 0x0, sizeof(circa->c1data));
	/* C3 even 4 bytes delayed 2 frames */
	for (f = 0 ; f < (int)circa->maxframes ; f++) {
		for (b = 0 ; b < C3_FRAME_SIZE ; b++) {
			if (((b / 4) & 1) == 0) {
				circa->c1data[C1OFF(f + 2, b)] = circa->c3data[C3OFF(f, b)];
			} else {
				circa->c1data[C1OFF(f, b)] = circa->c3data[C3OFF(f, b)];
			}
		}
	}
	/* intra-frame scramble */
	for (f = 0 ; f < (int)circa->maxframes ; f++) {
		PERMUTE5(circa->c1data, f, C1OFF, 2, 8, 10, 18, 6);
		PERMUTE5(circa->c1data, f, C1OFF, 3, 9, 11, 19, 7);
		PERMUTE5(circa->c1data, f, C1OFF, 4, 16, 20, 14, 12);
		PERMUTE5(circa->c1data, f, C1OFF, 5, 17, 21, 15, 13);
	}
	/* C2 Q parity RS encoding */
	for (f = 0 ; f < (int)circa->maxframes ; f++) {
		rs_parity(&circa->c2, &circa->c1data[C1OFF(f, 0)], C3_FRAME_SIZE, qparity, Q_PARITY_SIZE);
		(void) memmove(&circa->c1data[C1OFF(f, Q_PARITY_OFF + Q_PARITY_SIZE)],
			&circa->c1data[C1OFF(f, Q_PARITY_OFF)], C3_FRAME_SIZE / 2);
		(void) memcpy(&circa->c1data[C1OFF(f, Q_PARITY_OFF)], qparity, Q_PARITY_SIZE); 
	}
	/* sector disperse */
	for (f = 0 ; f < (int)circa->maxframes ; f++) {
		for (b = 1 ; b < C2_FRAME_SIZE ; b++) {
			SWAP(circa->c1data, C1OFF(f, b), C1OFF(f + b, b));
		}
	}
	/* C1 P parity rs encoding */
	for (f = 0 ; f < (int)circa->maxframes ; f++) {
		rs_parity(&circa->c1, &circa->c1data[C1OFF(f, 0)], C2_FRAME_SIZE,
			&circa->c1data[C1OFF(f, C2_FRAME_SIZE)], P_PARITY_SIZE);
	}
	/* C1 even bytes delayed 1 frame */
	for (f = (int)circa->maxframes - 1 ; f >= 0 ; --f) {
		for (b = C1_FRAME_SIZE - 1 ; b >= 0 ; --b) {
			if ((b & 1) == 0) {
				out[C1OFF(f + 1, b)] = circa->c1data[C1OFF(f, b)];
			} else {
				out[C1OFF(f, b)] = circa->c1data[C1OFF(f, b)];
			}
		}
	}
	return (ssize_t)circa->header.sectorsize;
}

/* decode the information given in the encoded buffer in 'in' */
ssize_t
circa_decode(circa_t *circa, const void *inp, size_t insize, uint8_t *out, size_t outsize)
{
	const uint8_t	*in = (const uint8_t *)inp;
	unsigned	 b;
	uint32_t	 isize;
	uint8_t		 qparity[C2_FRAME_SIZE];
	uint8_t		 qcalc[C2_FRAME_SIZE];
	uint8_t		 f1[C1_FRAME_SIZE];
	ssize_t		 cc;
	int		 f;

	USE_ARG(insize);
	if (!circa->init) {
		(void) fprintf(stderr, "decode: no init\n");
		return 0;
	}
	/* C1 even bytes delayed 1 frame */
	for (f = 0 ; f < (int)circa->maxframes ; f++) {
		for (b = 0 ; b < C1_FRAME_SIZE ; b++) {
			circa->c1data[C1OFF(f, b)] = ((b & 1) == 0) ? in[C1OFF(f + 1, b)] : in[C1OFF(f, b)];
		}
	}
	/* P parity C1 rs decoding */
	for (f = 0 ; f < (int)circa->maxframes ; f++) {
		(void) memcpy(f1, &circa->c1data[C1OFF(f, 0)], C1_FRAME_SIZE);
		cc = rs_decode(&circa->c1, f1, C1_FRAME_SIZE, &circa->c1data[C1OFF(f, 0)], C1_FRAME_SIZE);
		if (cc != C2_FRAME_SIZE) {
			circa->c1errc += 1;
		}
	}
	/* sector disperse */
	for (f = circa->maxframes - 1 ; f >= 0 ; --f) {
		for (b = C2_FRAME_SIZE - 1 ; b > 0 ; --b) {
			SWAP(circa->c1data, C1OFF(f, b), C1OFF(f + b, b));
		}
	}
	/* Q parity C2 rs decoding */
	for (f = 0 ; f < (int)circa->maxframes ; f++) {
		(void) memcpy(qparity, &circa->c1data[C1OFF(f, 0)], C3_FRAME_SIZE / 2);
		(void) memcpy(&qparity[C3_FRAME_SIZE / 2], &circa->c1data[C1OFF(f, (C3_FRAME_SIZE / 2) + Q_PARITY_SIZE)],
				C3_FRAME_SIZE / 2);
		(void) memcpy(&qparity[C3_FRAME_SIZE], &circa->c1data[C1OFF(f, C3_FRAME_SIZE / 2)], Q_PARITY_SIZE);
		cc = rs_decode(&circa->c2, qparity, C2_FRAME_SIZE, qcalc, sizeof(qcalc));
		if (cc != C3_FRAME_SIZE) {
			circa->c2errc += 1;
		}
		(void) memcpy(&circa->c3data[C3OFF(f, 0)], qcalc, C3_FRAME_SIZE);
	}
	/* intra-frame scramble */
	for (f = 0 ; f < (int)circa->maxframes ; f++) {
		PERMUTE5(circa->c3data, f, C3OFF, 2, 6, 18, 10, 8);
		PERMUTE5(circa->c3data, f, C3OFF, 3, 7, 19, 11, 9);
		PERMUTE5(circa->c3data, f, C3OFF, 4, 12, 14, 20, 16);
		PERMUTE5(circa->c3data, f, C3OFF, 5, 13, 15, 21, 17);
	}
	/* C3 even 4 bytes delayed 2 frames */
	for (f = 0 ; f < (int)circa->maxframes ; f++) {
		for (b = 0 ; b < C3_FRAME_SIZE ; b++) {
			circa->c1data[C3OFF(f, b)] = (((b / 4) & 1) == 0) ?
				circa->c3data[C3OFF(f + 2, b)] : circa->c3data[C3OFF(f, b)];
		}
	}
	(void) memcpy(&isize, circa->c1data, sizeof(isize));
	isize = le32(isize);
	(void) memcpy(out, &circa->c1data[sizeof(isize)], outsize);
	return isize;
}

/* get a header from an area of memory */
ssize_t
circa_get_header(circa_t *circa, const void *in, size_t insize)
{
	res_header_t	head;

	USE_ARG(insize);
	(void) memcpy(&head, in, sizeof(head));
	head.magic = le32(head.magic);
	if (head.magic != CIRCA_MAGIC) {
		return 0;
	}
	head.sectorsize = le32(head.sectorsize);
	(void) memcpy(&circa->header, &head, sizeof(circa->header));
	return sizeof(circa->header);
}

/* put a header into an area of memory */
ssize_t
circa_put_header(circa_t *circa, void *out, size_t outsize)
{
	res_header_t	head;

	USE_ARG(outsize);
	(void) memcpy(&head, &circa->header, sizeof(head));
	head.magic = le32(CIRCA_MAGIC);
	head.sectorsize = le32(head.sectorsize);
	(void) memcpy(out, &head, sizeof(head));
	return sizeof(head);
}

/* finalise things */
int
circa_end(circa_t *circa)
{
	if (circa->c1data) {
		free(circa->c1data);
		circa->c1data = NULL;
	}
	if (circa->c3data) {
		free(circa->c3data);
		circa->c3data = NULL;
	}
	return 1;
}

/* process a file - encoding or decoding */
int
circa_file(circa_t *circa, size_t sectorsize, int encoding, int header, char *in, char *fout)
{
	uint8_t	*buf;
	uint8_t	*out;
	ssize_t	 outcc;
	ssize_t	 cc;
	size_t	 readsize;
	FILE	*fpout;
	FILE	*fpin;
	int	 ok;

	if (in == NULL) {
		fpin = stdin;
	} else if ((fpin = fopen(in, "r")) == NULL) {
		(void) fprintf(stderr, "can't read '%s'\n", in);
		return 0;
	}
	if (fout == NULL) {
		fpout = stdout;
	} else if ((fpout = fopen(fout, "w")) == NULL) {
		(void) fprintf(stderr, "can't write '%s'\n", fout);
		if (fpin != stdin) {
			(void) fclose(fpin);
		}
		return 0;
	}
	if (encoding) {
		/* copy the encoding paramters to output */
		if (!circa_init(circa, sectorsize)) {
			(void) fprintf(stderr, "bad sectorsize %zu\n", sectorsize);
			return 0;
		}
		readsize = circa->datasize;
		if ((buf = calloc(1, sectorsize)) == NULL || (out = calloc(1, sectorsize)) == NULL) {
			(void) fprintf(stderr, "can't allocate sectorsize buffers\n");
			return 0;
		}
		if (header) {
			outcc = circa_put_header(circa, out, sectorsize);
			if (write(fileno(fpout), out, (size_t)outcc) != outcc) {
				(void) fprintf(stderr, "short write\n");
				return 0;
			}
		}
	} else {
		/* read the encoding paramters from input */
		if (header) {
			if (read(fileno(fpin), &circa->header, sizeof(circa->header)) != sizeof(circa->header)) {
				(void) fprintf(stderr, "short read\n");
				return 0;
			}
			if (circa_get_header(circa, &circa->header, sizeof(circa->header)) != sizeof(circa->header) ||
			    !circa_init(circa, readsize = sectorsize = circa->header.sectorsize)) {
				(void) fprintf(stderr, "bad sectorsize %zu\n", sectorsize);
				return 0;
			}
		} else {
			if (!circa_init(circa, readsize = sectorsize)) {
				(void) fprintf(stderr, "bad sectorsize %zu\n", sectorsize);
				return 0;
			}
		}
		if ((buf = calloc(1, sectorsize)) == NULL ||
		    (out = calloc(1, sectorsize)) == NULL) {
			(void) fprintf(stderr, "can't allocate sectorsize buffers\n");
			return 0;
		}
	}
	for (ok = 1 ; ok && (cc = read(fileno(fpin), buf, readsize)) > 0 ; ) {
		if (encoding) {
			outcc = circa_encode(circa, buf, (size_t)cc, out, sectorsize);
		} else {
			outcc = circa_decode(circa, buf, (size_t)cc, out, sectorsize);
		}
		if (write(fileno(fpout), out, outcc) != outcc) {
			(void) fprintf(stderr, "short write (%zd)\n", outcc);
			ok = 0;
		}
	}
	if (fpin != stdin) {
		fclose(fpin);
	}
	if (fpout != stdout) {
		fclose(fpout);
	}
	free(out);
	free(buf);
	return ok;
}
