/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/******************************************************************************/
/*                                                                            */
/* Author  : Ross Williams (ross@guest.adelaide.edu.au.).                     */
/* Date    : 3 June 1993.                                                     */
/* Version : 1.0.                                                             */
/* Status  : Public domain.                                                   */
/*                                                                            */
/* Description : This program writes a CRC lookup table (suitable for         */
/* inclusion in a C program) to a designated output file. The program can be  */
/* statically configured to produce any table covered by the Rocksoft^tm      */
/* Model CRC Algorithm. For more information on the Rocksoft^tm Model CRC     */
/* Algorithm, see the document titled "A Painless Guide to CRC Error          */
/* Detection Algorithms" by Ross Williams (ross@guest.adelaide.edu.au.). This */
/* document is likely to be in "ftp.adelaide.edu.au/pub/rocksoft".            */
/*                                                                            */
/* Note: Rocksoft is a trademark of Rocksoft Pty Ltd, Adelaide, Australia.    */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "libcrc.h"

#define DEF_OUT   "crctab.h"

/******************************************************************************/

/* If mess is non-empty, write it out and abort. Otherwise, check the error   */
/* status of outfile and abort if an error has occurred.                      */
static void
chk_err(FILE *fp, const char *mess)
{
	if (mess[0] != 0x0) {
		printf("%s\n",mess);
		exit(EXIT_FAILURE);
	}
	if (ferror(fp)) {
		perror("chk_err");
		exit(EXIT_FAILURE);
	}
}

/******************************************************************************/

static void
chkparam(FILE *fp, int width, unsigned poly, int reflect)
{
	if ((width != 2) && (width != 4)) {
		chk_err(fp, "chkparam: Width parameter is illegal.");
	}
	if ((width == 2) && (poly & 0xFFFF0000L)) {
		chk_err(fp, "chkparam: Poly parameter is too wide.");
	}
	if ((reflect != 0) && (reflect != 1)) {
		chk_err(fp, "chkparam: Reverse parameter is not boolean.");
	}
}

/******************************************************************************/

static void
gentable(FILE *fp, int width, unsigned poly, int reflect)
{
	CRC_CTX	 crcmodel;
	int	 perline = (width == 2) ? 8 : 4;
	int	 i;

	(void) fprintf(fp, "#ifndef CRCTAB_H_\n#define CRCTAB_H_ 1\n");
	(void) fprintf(fp, "/* automatically generated by gencrctab */\n\n");
	(void) fprintf(fp, "#include <inttypes.h>\n\n");
	(void) fprintf(fp, "/*\n");
	(void) fprintf(fp, " * CRC LOOKUP TABLE\n");
	(void) fprintf(fp, " * ================\n");
	(void) fprintf(fp, " * The following CRC lookup table was generated automagically\n");
	(void) fprintf(fp, " * by the Rocksoft^tm Model CRC Algorithm Table Generation\n");
	(void) fprintf(fp, " * Program V1.0 using the following model parameters:\n");
	(void) fprintf(fp, " *\n");
	(void) fprintf(fp, " *    Width   : %1lu bytes\n", (unsigned long) width);
	(void) fprintf(fp, " *    Poly    : ");
	(void) fprintf(fp, (width == 2) ? "0x%04x" : "0x%08x", poly);
	(void) fprintf(fp, "\n *    Reverse : %s.\n", (reflect) ? "TRUE" : "FALSE");
	(void) fprintf(fp, " *\n");
	(void) fprintf(fp, " * For more information on the Rocksoft^tm Model CRC Algorithm,\n");
	(void) fprintf(fp, " * see the document titled \"A Painless Guide to CRC Error\n");
	(void) fprintf(fp, " * Detection Algorithms\" by Ross Williams\n");
	(void) fprintf(fp, " * (ross@guest.adelaide.edu.au.). This document is likely to be\n");
	(void) fprintf(fp, " * in the FTP archive \"ftp.adelaide.edu.au/pub/rocksoft\".\n");
	(void) fprintf(fp, " */\n\n\n");
	(void) fprintf(fp, "\n");
	switch (width) {
	case 2:
	case 4:
		(void) fprintf(fp, "static uint32_t crctable[256] = {\n");
		break;
	default:
		chk_err(fp, "gentable: width is invalid.");
	}
	chk_err(fp, "");
	CRC_Init(&crcmodel);
	CRC_Args(&crcmodel, width, poly, reflect);
	for (i = 0; i < 256; i++) {
		if (i % perline == 0) {
			(void) fprintf(fp, "\t/* %d */\t", i);
		}
		(void) fprintf(fp, (width == 2) ? "0x%04x" : "0x%08x",
			CRC_tabentry(&crcmodel, i));
		if (i != 255) {
			(void) fprintf(fp, ", ");
		}
		if (((i+1) % perline) == 0) {
			(void) fprintf(fp, "\n");
		}
		chk_err(fp, "");
	}
	(void) fprintf(fp, "};\n\n");
	(void) fprintf(fp, "static inline uint32_t\n");
	(void) fprintf(fp, "crc_get(uint8_t ch)\n");
	(void) fprintf(fp, "{\n");
	(void) fprintf(fp, "\treturn crctable[ch];\n");
	(void) fprintf(fp, "}\n#endif /* !CRCTAB_H_ */\n");
	chk_err(fp, "");
}

int
main(int argc, char **argv)
{
	const char	*out;
	unsigned	 poly;
	FILE		*outfile;
	int		 reflect;
	int		 width;
	int		 i;

	out = DEF_OUT;
	width = CRC_DEF_WIDTH;
	poly = CRC_DEF_POLY;
	reflect = CRC_DEF_REVER;
	while ((i = getopt(argc, argv, "o:p:r:w:")) != -1) {
		switch(i) {
		case 'o':
			out = optarg;
			break;
		case 'p':
			poly = strtol(optarg, NULL, 16);
			break;
		case 'r':
			reflect = (strcmp(optarg, "true") == 0);
			break;
		case 'w':
			width = atoi(optarg);
			break;
		default:
			break;
		}
	}
	if ((outfile = fopen(out, "w")) == NULL) {
		(void) fprintf(stderr, "can't open output file '%s'\n", out);
		exit(EXIT_FAILURE);
	}
	chkparam(outfile, width, poly, reflect);
	chk_err(outfile, "");
	gentable(outfile, width, poly, reflect);
	(void) fclose(outfile);
	exit(EXIT_SUCCESS);
}
