/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/******************************************************************************/
/*                             Start of crcmodel.c                            */
/******************************************************************************/
/*                                                                            */
/* Author : Ross Williams (ross@guest.adelaide.edu.au.).                      */
/* Date   : 3 June 1993.                                                      */
/* Status : Public domain.                                                    */
/*                                                                            */
/* Description : This is the implementation (.c) file for the reference       */
/* implementation of the Rocksoft^tm Model CRC Algorithm. For more            */
/* information on the Rocksoft^tm Model CRC Algorithm, see the document       */
/* titled "A Painless Guide to CRC Error Detection Algorithms" by Ross        */
/* Williams (ross@guest.adelaide.edu.au.). This document is likely to be in   */
/* "ftp.adelaide.edu.au/pub/rocksoft".                                        */
/*                                                                            */
/* Note: Rocksoft is a trademark of Rocksoft Pty Ltd, Adelaide, Australia.    */
/*                                                                            */
/******************************************************************************/
#include <sys/types.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "libcrc.h"

/******************************************************************************/

/* The following definitions make the code more readable. */

#define BITMASK(X)	(1 << (X))
#define MASK32		0xFFFFFFFFL

/******************************************************************************/

/* Returns the value v with the bottom b [0, 32] bits reflected. */
/* Example: reflect(0x3e23L, 3) == 0x3e26                        */
static unsigned
reflect(unsigned v, int b)
{
	unsigned	t = v;
	int		i;

	for (i = 0; i < b; i++) {
		if (t & 1L) {
			v |=  BITMASK((b-1)-i);
		} else {
			v &= ~BITMASK((b-1)-i);
		}
		t >>= 1;
	}
	return v;
}

/******************************************************************************/

/* Returns a longword whose value is (2^model->width)-1.     */
/* The trick is to do this portably (e.g. without doing <<32). */
static unsigned
widmask(CRC_CTX *model)
{
	return (((1 << (model->width - 1)) - 1) << 1) | 1;
}

static void
crcbyte(CRC_CTX *model, int ch)
{
	int   i;
	unsigned uch  = (unsigned) ch;
	unsigned topbit = BITMASK(model->width - 1);

	if (model->refin) {
		uch = reflect(uch, 8);
	}
	model->ctxt ^= (uch << (model->width - 8));
	for (i = 0; i < 8; i++) {
		if (model->ctxt & topbit) {
			model->ctxt = (model->ctxt << 1) ^ model->poly;
		} else {
			model->ctxt <<= 1;
		}
		model->ctxt &= widmask(model);
	}
}

/******************************************************************************/

void
CRC_Init(CRC_CTX *model)
{
	(void) memset(model, 0x0, sizeof(*model));
	model->ctxt = model->init;
}

void
CRC_Args(CRC_CTX *model, int width, unsigned poly, int reverse)
{
	if (model->width == 0) {
		model->width = width * 8;
	}
	if (model->poly == 0) {
		model->poly = poly;
	}
	model->refin = reverse;
}

void
CRC_Update(CRC_CTX *model, uint8_t *blk_adr, size_t blk_len)
{
	if (model->width == 0 || model->poly == 0) {
		CRC_Args(model, CRC_DEF_WIDTH, CRC_DEF_POLY, CRC_DEF_REVER);
	}
	while (blk_len-- > 0) {
		crcbyte(model, *blk_adr++);
	}
}

void
CRC_Final(unsigned *crc, CRC_CTX *model)
{
	*crc = (model->refot) ?
			(unsigned)(model->xorot ^ reflect(model->ctxt, model->width)) :
			(unsigned)(model->xorot ^ model->ctxt);
}

char *
CRC_Data(uint8_t *data, size_t len, char *buf)
{
	CRC_CTX	model;

	(void) memset(&model, 0x0, sizeof(model));
	CRC_Init(&model);
	CRC_Update(&model, data, len);
	CRC_Final((unsigned *)(void *)buf, &model);
	(void) memset(&model, 0x0, sizeof(model));
	return buf;
}

char *
CRC_File(const char *filename, char *buf)
{
	CRC_CTX	 model;
	uint8_t	 rbuf[BUFSIZ];
	ssize_t	 cc;
	FILE	*fp;

	(void) memset(&model, 0x0, sizeof(model));
	if ((fp = fopen(filename, "r")) == NULL) {
		(void) fprintf(stderr, "CRC_File: can't open '%s'\n", filename);
		return NULL;
	}
	CRC_Init(&model);
	while ((cc = read(fileno(fp), rbuf, sizeof(rbuf))) > 0) {
		CRC_Update(&model, rbuf, (size_t)cc);
	}
	(void) fclose(fp);
	CRC_Final((unsigned *)(void *)buf, &model);
	(void) memset(&model, 0x0, sizeof(model));
	return buf;
}

char *
CRC_FileChunk(const char *filename, char *buf, off_t off, off_t len)
{
	CRC_CTX	 model;
	uint8_t	 rbuf[BUFSIZ];
	ssize_t	 cc;
	off_t	 done;
	FILE	*fp;

	(void) memset(&model, 0x0, sizeof(model));
	if ((fp = fopen(filename, "r")) == NULL) {
		(void) fprintf(stderr, "CRC_FileChunk: can't open '%s'\n", filename);
		return NULL;
	}
	CRC_Init(&model);
	if (lseek(fileno(fp), off, SEEK_SET) < 0) {
		(void) fprintf(stderr, "CRC_FileChunk: can't open '%s'\n", filename);
		(void) fclose(fp);
		return NULL;
	}
	for (done = 0 ; done < len ; done += cc) {
		if ((cc = read(fileno(fp), rbuf, sizeof(rbuf))) < 0) {
			break;
		}
		CRC_Update(&model, rbuf, (size_t)cc);
	}
	(void) fclose(fp);
	CRC_Final((unsigned *)(void *)buf, &model);
	(void) memset(&model, 0x0, sizeof(model));
	return buf;
}

unsigned
CRC_tabentry(CRC_CTX *model, int ind)
{
	unsigned	topbit;
	unsigned	inbyte;
	unsigned	r;
	int		i;

	topbit = BITMASK(model->width - 1);
	inbyte = (unsigned) ind;
	if (model->refin) {
		inbyte = reflect(inbyte, 8);
	}
	r = inbyte << (model->width - 8);
	for (i=0; i<8; i++) {
		if (r & topbit) {
			r = (r << 1) ^ model->poly;
		} else {
			r <<= 1;
		}
	}
	if (model->refin) {
		r = reflect(r, model->width);
	}
	return r & widmask(model);
}
