/* $NetBSD$ */

/*-
 * Copyright (c) 2005,2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/cdefs.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/mount.h>
#ifdef HAVE_STRUCT_STATVFS
#include <sys/statvfs.h>
#endif

#include <ctype.h>
#include <dirent.h>
#include <err.h>
#include <fnmatch.h>
#include <grp.h>
#include <pwd.h>
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <utmp.h>

#include "defs.h"
#include "find2.h"

enum {
	NONE,
	EQ,
	NE,
	GT,
	GE,
	LT,
	LE,
	ANDAND,
	OROR,
	AND,
	OR,
	XOR,
	IN,
	NOTIN
};

enum {
	Pathname,
	Fnmatch,
	Regexp,
	Name,
	Size,
	Mtime,
	Atime,
	Ctime,
	Nlink,
	Ino,
	Mode,
	Blocks,
	Type,
	Dev,
	Uid,
	Gid,
	Fstype,
	Depth,
	Mage,
	Aage,
	Cage
};

/* a str_t consists of a string, its length, and any associated op */
typedef struct str_t {
	const char     *name;		/* string */
	unsigned	len;		/* its length */
	int64_t		op;		/* associated operator or value */
} str_t;

/* array of operators */
static str_t	ops[] = {
	{ "==",		2,	EQ },
	{ "!=",		2,	NE },
	{ ">=",		2,	GE },
	{ ">",		1,	GT },
	{ "<=",		2,	LE },
	{ "<",		1,	LT },
	{ "&&",		2,	ANDAND },
	{ "||",		2,	OROR },
	{ "&",		1,	AND },
	{ "|",		1,	OR },
	{ "^",		1,	XOR },
	{ "in",		2,	IN },
	{ "notin",	5,	NOTIN },
	{ "!in",	3,	NOTIN },
	{ NULL,		0,	0 }
};

/* array of logical boolean operators */
static str_t	logicals[] = {
	{ "&&",		2,	ANDAND },
	{ "||",		2,	OROR },
	{ NULL,		0,	0 }
};

/* names of fields */
static str_t	names[] = {
	{ "pathname",	8,	Pathname },
	{ "fnmatch",	7,	Fnmatch },
	{ "regexp",	6,	Regexp },
	{ "d_name",	6,	Name },
	{ "name",	4,	Name },
	{ "st_size",	7,	Size },
	{ "size",	4,	Size },
	{ "st_mtime",	8,	Mtime },
	{ "mtime",	5,	Mtime },
	{ "st_atime",	8,	Atime },
	{ "atime",	5,	Atime },
	{ "st_ctime",	8,	Ctime },
	{ "ctime",	5,	Ctime },
	{ "st_nlink",	8,	Nlink },
	{ "nlink",	5,	Nlink },
	{ "st_ino",	6,	Ino },
	{ "ino",	3,	Ino },
	{ "st_mode",	7,	Mode },
	{ "mode",	4,	Mode },
	{ "st_blocks",	9,	Blocks },
	{ "blocks",	6,	Blocks },
	{ "st_dev",	6,	Dev },
	{ "dev",	3,	Dev },
	{ "type",	4,	Type },
	{ "st_uid",	6,	Uid },
	{ "uid",	3,	Uid },
	{ "st_gid",	6,	Gid },
	{ "gid",	3,	Gid },
	{ "fstype",	6,	Fstype },
	{ "depth",	5,	Depth },
	{ "mage",	4,	Mage },
	{ "aage",	4,	Aage },
	{ "cage",	4,	Cage },
	{ NULL,		0,	0 }
};

/* definitions of types */
static str_t	types[] = {
	{ "file",	4,	S_IFREG },
	{ "S_IFREG",	7,	S_IFREG },
	{ "f",		1,	S_IFREG },
	{ "directory",	9,	S_IFDIR },
	{ "dir",	3,	S_IFDIR },
	{ "S_IFDIR",	7,	S_IFDIR },
	{ "d",		1,	S_IFDIR },
	{ "char",	4,	S_IFCHR },
	{ "S_IFCHR",	7,	S_IFCHR },
	{ "c",		1,	S_IFCHR },
	{ "block",	5,	S_IFBLK },
	{ "S_IFBLK",	7,	S_IFBLK },
	{ "b",		1,	S_IFBLK },
	{ "symlink",	7,	S_IFLNK },
	{ "S_IFLNK",	7,	S_IFLNK },
	{ "l",		1,	S_IFLNK },
	{ "whiteout",	8,	S_IFWHT },
	{ "S_IFWHT",	7,	S_IFWHT },
	{ "W",		1,	S_IFWHT },
	{ "fifo",	4,	S_IFIFO },
	{ "S_IFIFO",	7,	S_IFIFO },
	{ "p",		1,	S_IFIFO },
	{ "socket",	6,	S_IFSOCK },
	{ "S_IFSOCK",	8,	S_IFSOCK },
	{ "s",		1,	S_IFSOCK },
	{ NULL,		0,	0 }
};

static str_t	mults[] = {
	{ "years",	5,	(365 * 24 * 60 * 60) },
	{ "months",	6,	(31 * 24 * 60 * 60) },
	{ "days",	4,	(24 * 60 * 60) },
	{ "hours",	5,	(60 * 60) },
	{ "minutes",	7,	60 },
	{ "K",		1,	1024 },
	{ "k",		1,	1024 },
	{ "M",		1,	(1024 * 1024) },
	{ "m",		1,	(1024 * 1024) },
	{ "G",		1,	(1024 * 1024 * 1024) },
	{ "g",		1,	(1024 * 1024 * 1024) },
	{ "T",		1,	1099511627776LL },
	{ "t",		1,	1099511627776LL },
	{ NULL,		0,	0 }
};

/* save `n' chars of `s' in malloc'd memory */
static char *
strnsave(const char *s, ssize_t n)
{
	char	*cp;

	if (n < 0) {
		n = strlen(s);
	}
	NEWARRAY(char, cp, n + 1, "strnsave", exit(EXIT_FAILURE));
	(void) memcpy(cp, s, (unsigned)n);
	cp[n] = 0;
	return cp;
}

/* recognise a string from an array of names */
static int
recog(const char *s, int *ip, str_t *array, str_t **np)
{
	*ip += (int)strspn(&s[*ip], " \t\n\r");
	for (*np = array ; (*np)->name ; *np += 1) {
		if (strncmp(&s[*ip], (*np)->name, (*np)->len) == 0) {
			*ip += (*np)->len;
			return 1;
		}
	}
	return 0;
}

/* parse a phrase */
static int
parse_phrase(f2phrase_t *phrase, const char *s, int *ip)
{
	const char     *cp;
	str_t	       *found;
	int		n;

	if (!recog(s, ip, names, &found)) {
		warnx("can't recognise lhs of `%s'", &s[*ip]);
		return 0;
	}
	phrase->name = (int)found->op;
	if (!recog(s, ip, ops, &found)) {
		warnx("can't recognise operator of `%s'", &s[*ip]);
		return 0;
	}
	phrase->op = (int)found->op;
	*ip += (int)strspn(&s[*ip], " \t\n\r");
	switch(s[*ip]) {
	case '"':
		for (cp = &s[*ip + 1]; *cp && *cp != '"' ; cp++) {
			if (*cp == '\\') {
				cp++;
			}
		}
		if (*cp == '"') {
			phrase->value = strnsave(&s[*ip + 1], (int)(cp - &s[*ip]) - 1);
			*ip = (int)(cp - &s[*ip]) + 1;
		} else {
			phrase->value = strdup(&s[*ip + 1]);
			*ip = (int)(cp - &s[*ip]);
		}
		break;
	case '{':
		if ((cp = strchr(&s[*ip], '}')) == NULL) {
			warnx("missing '}' in `%s'", &s[*ip]);
			return 0;
		}
		phrase->value = strnsave(&s[*ip] + 1, (int)(cp - &s[*ip]) - 1);
		*ip = (int)(cp - &s[*ip]) + 1;
		break;
	default:
		*ip += (int)strspn(&s[*ip], " \t\n\r");
		if (isdigit((unsigned)s[*ip])) {
			for (n = *ip ; isdigit((unsigned)s[*ip]) ; *ip += 1) {
			}
			(void) recog(s, ip, mults, &found);
			phrase->value = strnsave(&s[n], *ip - n);
		} else {
			n = (int)strcspn(&s[*ip], ") \t\n\r");
			phrase->value = strnsave(&s[*ip], n);
			*ip += n;
		}
	}
	return 1;
}

/* parse an expression */
static int
parse_expr(f2clause_t *clause, const char *s, int *ip)
{
	str_t *found;
	int	parenc;

	parenc = 0;
	*ip += (int)strspn(&s[*ip], " \t\r\n");
	if (s[*ip] == '(') {
		parenc += 1;
		*ip += 1;
		*ip += (int)strspn(&s[*ip], " \t\r\n");
	}
	NEW(f2phrase_t, clause->lhs, "parse_expr: lhs", exit(EXIT_FAILURE));
	if (!parse_phrase(clause->lhs, s, ip)) {
		return 0;
	}
	*ip += (int)strspn(&s[*ip], " \t\r\n");
	if (s[*ip] != ')') {
		if (s[*ip] == 0) {
			return 1;
		}
		if (!recog(s, ip, logicals, &found)) {
			warnx("parse_expr: can't recognise logical operator in `%s'", &s[*ip]);
			return 0;
		}
		clause->op = (int)found->op;
		NEW(f2phrase_t, clause->rhs, "parse_expr: rhs", exit(EXIT_FAILURE));
		if (!parse_phrase(clause->rhs, s, ip)) {
			return 0;
		}
	}
	if (parenc > 0) {
		*ip += (int)strspn(&s[*ip], " \t\r\n");
		if (s[*ip] != ')') {
			warnx("parse_expr: expected closing parenthesis at `%s'", &s[*ip]);
			return 0;
		}
		*ip += 1;
		*ip += (int)strspn(&s[*ip], " \t\r\n");
	}
	return 1;
}

/* match a string */
static int
match_str(const char *lhs, int op, char *rhs, int ign)
{
	int	cmp;

	cmp = (ign) ? strcasecmp(lhs, rhs) : strcmp(lhs, rhs);
	switch(op) {
	case EQ:
		return cmp == 0;
	case NE:
		return cmp != 0;
	case GT:
		return cmp > 0;
	case GE:
		return cmp >= 0;
	case LT:
		return cmp < 0;
	case LE:
		return cmp <= 0;
	default:
		warnx("invalid op `%d'", op);
		return -1;
	}
}

/* match a number */
static int
match_num(int64_t lhs, int op, int64_t rhs)
{
	switch(op) {
	case EQ:
		return lhs == rhs;
	case NE:
		return lhs != rhs;
	case GT:
		return lhs > rhs;
	case GE:
		return lhs >= rhs;
	case LT:
		return lhs < rhs;
	case LE:
		return lhs <= rhs;
	case AND:
		return (int)(lhs & rhs);
	case OR:
		return (int)(lhs | rhs);
	case XOR:
		return (int)(lhs ^ rhs);
	default:
		warnx("invalid op `%d'", op);
		return -1;
	}
}

/* cache the number associated with the entry */
static int64_t
getfield(find_t *find, f2phrase_t *phrase)
{
	struct stat	st;
	str_t	       *mp;
	char	       *endptr;
	int		n;

	if (!phrase->init) {
		phrase->init = 1;
		if (strchr(phrase->value, '/') == NULL) {
			phrase->num = (long)strtoll(phrase->value, &endptr, 0);
			endptr += (int)strspn(endptr, " \t\r\n");
			for (mp = mults ; mp->name && strncmp(mp->name, endptr, mp->len) != 0 ; mp++) {
			}
			if (mp->name != NULL) {
				phrase->num *= mp->op;
			}
		} else {
			n = (find->usestat) ? stat(phrase->value, &st) : lstat(phrase->value, &st);
			if (n < 0) {
				warnx("`%s' is a filename, but doesn't exist", phrase->value);
				return 0;
			}
			switch(phrase->name) {
			case Size:
				return phrase->num = st.st_size;
			case Mtime:
				return phrase->num = st.st_mtime;
			case Atime:
				return phrase->num = st.st_atime;
			case Ctime:
				return phrase->num = st.st_ctime;
			case Mage:
				return phrase->num = (find->started - st.st_mtime);
			case Aage:
				return phrase->num = (find->started - st.st_atime);
			case Cage:
				return phrase->num = (find->started - st.st_ctime);
			case Nlink:
				return phrase->num = st.st_nlink;
			case Ino:
				return phrase->num = st.st_ino;
			case Mode:
				return phrase->num = (st.st_mode & ~S_IFMT);
			case Blocks:
				return phrase->num = st.st_blocks;
			case Dev:
				return phrase->num = st.st_dev;
			}
		}
	}
	return phrase->num;
}

/* cache the type of the directory entry */
static int64_t
gettype(find_t *find, f2phrase_t *phrase)
{
	struct stat	st;
	str_t	       *found;
	int		i;
	int		n;

	if (!phrase->init) {
		phrase->init = 1;
		if (strchr(phrase->value, '/') == NULL) {
			i = 0;
			if (recog(phrase->value, &i, types, &found)) {
				return phrase->num = found->op;
			}
			warnx("unknown type `%s'", phrase->value);
			return phrase->num = 0;
		} else {
			n = (find->usestat) ? stat(phrase->value, &st) : lstat(phrase->value, &st);
			if (n < 0) {
				warnx("`%s' is a filename, but doesn't exist", phrase->value);
				return 0;
			}
			return phrase->num = (st.st_mode & S_IFMT);
		}
	}
	return phrase->num;
}

/* cache the uid of a directory entry */
static int64_t
getusername(find_t *find, f2phrase_t *phrase)
{
	struct passwd  *pwp;
	struct stat	st;
	int		n;

	if (!phrase->init) {
		phrase->init = 1;
		if (strchr(phrase->value, '/') == NULL) {
			if ((pwp = getpwnam(phrase->value)) != NULL) {
				return phrase->num = pwp->pw_uid;
			} else if ((pwp = getpwuid((unsigned)atoi(phrase->value))) != NULL) {
				return phrase->num = pwp->pw_uid;
			}
			warnx("unknown uid `%s'", phrase->value);
			return phrase->num = 0;
		} else {
			n = (find->usestat) ? stat(phrase->value, &st) : lstat(phrase->value, &st);
			if (n < 0) {
				warnx("`%s' is a filename, but doesn't exist", phrase->value);
				return 0;
			}
			return phrase->num = st.st_uid;
		}
	}
	return phrase->num;
}

/* cache the gid of a directory entry */
static int64_t
getgroupname(find_t *find, f2phrase_t *phrase)
{
	struct group   *grp;
	struct stat	st;
	int		n;

	if (!phrase->init) {
		phrase->init = 1;
		if (strchr(phrase->value, '/') == NULL) {
			if ((grp = getgrnam(phrase->value)) != NULL) {
				return phrase->num = grp->gr_gid;
			} else if ((grp = getgrgid((unsigned)atoi(phrase->value))) != NULL) {
				return phrase->num = grp->gr_gid;
			}
			warnx("unknown gid `%s'", phrase->value);
			return phrase->num = 0;
		} else {
			n = (find->usestat) ? stat(phrase->value, &st) : lstat(phrase->value, &st);
			if (n < 0) {
				warnx("`%s' is a filename, but doesn't exist", phrase->value);
				return 0;
			}
			return phrase->num = st.st_gid;
		}
	}
	return phrase->num;
}

/* this struct defines a regexp cache entry */
typedef struct recache_t {
	char	       *s;		/* string representation */
	int		flags;		/* regcomp flags */
	regex_t		r;		/* the regexp */
	u_int64_t	lru;		/* last time accessed */
} recache_t;

enum {
	MaxREs = 10
};

static recache_t	rev[MaxREs];
static u_int64_t	lru;

/* find the regexp `re' in the cache */
static regex_t *
getre(char *re, int flags)
{
	int	oldest;
	int	i;

	for (i = 0 ; i < MaxREs ; i++) {
		if (rev[i].s && strcmp(rev[i].s, re) == 0 && rev[i].flags == flags) {
			break;
		}
	}
	if (i == MaxREs) {
		for (oldest = 0, i = 1 ; i < MaxREs ; i++) {
			if (rev[i].lru < rev[oldest].lru) {
				oldest = i;
			}
		}
		if (rev[i = oldest].lru != 0) {
			regfree(&rev[i].r);
		}
		if (regcomp(&rev[i].r, re, REG_EXTENDED | flags) != 0) {
			warnx("can't compile `%s'", re);
			return NULL;
		}
	}
	rev[i].flags = flags;
	rev[i].lru = ++lru;
	return &rev[i].r;
}

/* returns 1 if `re' matches `name' */
static int
regexmatch(find_t *find, char *re, const char *name)
{
	regmatch_t	matchv[10];
	regex_t	       *rp;

	if ((rp = getre(re, (find->ignorecase) ? REG_ICASE : 0)) == NULL) {
		return 0;
	}
	return regexec(rp, name, 10, matchv, 0) == 0;
}

/* match a file system type */
static int
match_fstype(const char *fullname, f2phrase_t *phrase)
{
#ifdef HAVE_STRUCT_STATVFS
	struct statvfs	fs;

	if (statvfs(fullname, &fs) < 0) {
		warn("can't statvfs `%s'", fullname);
		return 0;
	}
	if (strcasecmp(phrase->value, "local") == 0) {
		return (int)(fs.f_flag & MNT_LOCAL);
	}
	if (strcasecmp(phrase->value, "rdonly") == 0 ||
	    strcasecmp(phrase->value, "ro") == 0) {
		return (int)(fs.f_flag & MNT_RDONLY);
	}
	return match_str(fs.f_fstypename, phrase->op, phrase->value, 1);
#else
	struct statfs	fs;

	if (statfs(fullname, &fs) < 0) {
		warn("can't statfs `%s'", fullname);
		return 0;
	}
	if (strcasecmp(phrase->value, "local") == 0) {
		return fs.f_flags & MNT_LOCAL;
	}
	if (strcasecmp(phrase->value, "rdonly") == 0 ||
	    strcasecmp(phrase->value, "ro") == 0) {
		return fs.f_flags & MNT_RDONLY;
	}
	return match_str(fs.f_fstypename, phrase->op, phrase->value, 1);
#endif
}

/* match a phrase */
static int
match_phrase(find_t *find, f2phrase_t *phrase, const char *fullname, const char *name, struct stat *sp, int depth)
{
	switch(phrase->name) {
	case Pathname:
		return match_num((int64_t)(fnmatch(fullname, name, FNM_PATHNAME | ((find->ignorecase) ? FNM_CASEFOLD : 0)) == 0), phrase->op, (int64_t)1);
	case Fnmatch:
		return match_num((int64_t)(fnmatch(phrase->value, name, FNM_PATHNAME | ((find->ignorecase) ? FNM_CASEFOLD : 0)) == 0), phrase->op, (int64_t)1);
	case Regexp:
		return match_num((int64_t)regexmatch(find, phrase->value, name), phrase->op, (int64_t)1);
	case Name:
		return match_str(name, phrase->op, phrase->value, find->ignorecase);
	case Size:
		return match_num(sp->st_size, phrase->op, getfield(find, phrase));
	case Mtime:
		return match_num((int64_t)sp->st_mtime, phrase->op, getfield(find, phrase));
	case Atime:
		return match_num((int64_t)sp->st_atime, phrase->op, getfield(find, phrase));
	case Ctime:
		return match_num((int64_t)sp->st_ctime, phrase->op, getfield(find, phrase));
	case Mage:
		return match_num((int64_t)(find->started - sp->st_mtime), phrase->op, getfield(find, phrase));
	case Aage:
		return match_num((int64_t)(find->started - sp->st_atime), phrase->op, getfield(find, phrase));
	case Cage:
		return match_num((int64_t)(find->started - sp->st_ctime), phrase->op, getfield(find, phrase));
	case Nlink:
		return match_num((int64_t)sp->st_nlink, phrase->op, getfield(find, phrase));
	case Ino:
		return match_num((int64_t)sp->st_ino, phrase->op, getfield(find, phrase));
	case Mode:
		return match_num((int64_t)(sp->st_mode & ~S_IFMT), phrase->op, getfield(find, phrase));
	case Blocks:
		return match_num(sp->st_blocks, phrase->op, getfield(find, phrase));
	case Dev:
		return match_num((int64_t)sp->st_dev, phrase->op, getfield(find, phrase));
	case Type:
		return match_num((int64_t)(sp->st_mode & S_IFMT), phrase->op, gettype(find, phrase));
	case Uid:
		return match_num((int64_t)sp->st_uid, phrase->op, getusername(find, phrase));
	case Gid:
		return match_num((int64_t)sp->st_gid, phrase->op, getgroupname(find, phrase));
	case Fstype:
		return match_fstype(fullname, phrase);
	case Depth:
		return match_num((int64_t)depth, phrase->op, getfield(find, phrase));
	}
	warnx("match_phrase: `%d' not recognised", phrase->name);
	return 0;
}

/* match an expression */
static int
match_clause(find_t *find, f2clause_t *clause, const char *fullname, const char *name, struct stat *sp, int depth)
{
	int	lhs;
	int	rhs;

	lhs = match_phrase(find, clause->lhs, fullname, name, sp, depth);
	switch(clause->op) {
	case NONE:
		return lhs;
	}
	rhs = match_phrase(find, clause->rhs, fullname, name, sp, depth);
	switch(clause->op) {
	case ANDAND:
		return lhs && rhs;
	case OROR:
		return lhs || rhs;
	}
	warnx("match_clause: `%d' not recognised", clause->op);
	return -1;
}

/* match a sentence */
static int
match_sentence(find_t *find, const char *fullname, const char *name, struct stat *sp, int depth)
{
	int	lhs;
	int	rhs;

	lhs = match_clause(find, find->sent.lhs, fullname, name, sp, depth);
	switch(find->sent.op) {
	case NONE:
		return lhs;
	}
	rhs = match_clause(find, find->sent.rhs, fullname, name, sp, depth);
	switch(find->sent.op) {
	case ANDAND:
		return lhs && rhs;
	case OROR:
		return lhs || rhs;
	}
	warnx("match_sentence: `%d' not recognised", find->sent.op);
	return -1;
}

/* print information about a symbolic link */
static void
printlink(const char *name, FILE *filep)
{
	ssize_t	cc;
	char	path[MAXPATHLEN + 1];

	if ((cc = readlink(name, path, MAXPATHLEN)) < 0) {
		warn("%s", name);
		return;
	}
	(void) fprintf(filep, " -> %.*s", (int)cc, path);
}

#define SIXMONTHS	(365 * 60 * 60 * 24 / 2)

/* print the time */
static void
ptime(find_t *find, time_t ftime, FILE *filep)
{
	char   *longstring;

	longstring = ctime(&ftime);
	(void) fprintf(filep, "%.7s", &longstring[4]);
	if (ftime + SIXMONTHS > find->started) {
		(void) fprintf(filep, "%.5s ", &longstring[11]);
	} else {
		(void) fprintf(filep, " %.4s ", &longstring[20]);
	}
}

/* long listing of directory entry */
static void
longlist(find_t *find, const char *name, const char *accpath, struct stat *sp, FILE *filep)
{
	char	modep[15];

	(void) fprintf(filep, "%7lu %6lld ", (u_long) sp->st_ino, (long long) sp->st_blocks);
	(void) strmode(sp->st_mode, modep);
	(void) fprintf(filep, "%s %3lu %-*s %-*s ",
		modep,
		(unsigned long) sp->st_nlink,
		UT_NAMESIZE, user_from_uid(sp->st_uid, 0),
		UT_NAMESIZE, group_from_gid(sp->st_gid, 0));
	if (S_ISCHR(sp->st_mode) || S_ISBLK(sp->st_mode)) {
		(void) fprintf(filep, "%3d,%5d ", major(sp->st_rdev), minor(sp->st_rdev));
	} else {
		(void) fprintf(filep, "%9lld ", (long long) sp->st_size);
	}
	ptime(find, sp->st_mtime, filep);
	(void) fprintf(filep, "%s", name);
	if (S_ISLNK(sp->st_mode)) {
		printlink(accpath, filep);
	}
	(void) fputc('\n', filep);
}

/* do a directory */
static int
dodir(find_t *find, const char *d, int depth, FILE *filep)
{
	struct dirent  *dp;
	struct stat	lst;
	struct stat	st;
	const char     *nul = (find->nulterm) ? "\0" : "";
	char		buf[MAXPATHLEN];
	DIR	       *dirp;
	int		enter;
	int		ret;
	int		n;

	if (find->verbose) {
		(void) fprintf(stderr, "Directory %s\n", d);
	}
	if ((dirp = opendir(d)) == NULL) {
		warn("can't open directory `%s'", d);
		return 1;
	}
	for (ret = 1 ; (dp = readdir(dirp)) != NULL ; ) {
		if (strcmp(dp->d_name, ".") == 0 || strcmp(dp->d_name, "..") == 0) {
			continue;
		}
		(void) snprintf(buf, sizeof(buf), "%s%s%s",
				d, (strcmp(d, "/") == 0) ? "" : "/", dp->d_name);
		n = (find->usestat) ? stat(buf, &st) : lstat(buf, &st);
		if (n < 0) {
			warnx("can't access `%s'", buf);
			continue;
		}
		if (S_ISDIR(st.st_mode)) {
			enter = 1;
			if (find->nofollow && lstat(buf, &lst) == 0 && S_ISLNK(lst.st_mode)) {
				enter = 0;
			}
			if (find->samedev && find->dev != st.st_dev) {
				enter = 0;
			}
			if (enter) {
				(void) dodir(find, buf, depth + 1, filep);
			}
		}
		switch (match_sentence(find, buf, dp->d_name, &st, depth + 1)) {
		case -1:
			ret = 0;
			break;
		case 0:
			break;
		default:
			if (find->longlist) {
				longlist(find, buf, dp->d_name, &st, filep);
			} else {
				(void) fprintf(filep, "%s%s%s%s\n", d,
					(strcmp(d, "/") == 0) ? "" : "/", dp->d_name, nul);
			}
			break;
		}
	}
	(void) closedir(dirp);
	return ret;
}

/* do a directory entry */
static int
dodirent(find_t *find, const char *f, int depth, FILE *filep)
{
	struct stat	st;
	const char     *nul = (find->nulterm) ? "\0" : "";
	int		n;

	if (find->verbose) {
		(void) fprintf(stderr, "Directory entry %s\n", f);
	}
	n = (find->usestat) ? stat(f, &st) : lstat(f, &st);
	if (n < 0) {
		warnx("can't access `%s'", f);
	}
	switch (match_sentence(find, f, f, &st, depth + 1)) {
	case -1:
		return 0;
	case 0:
		return 1;
	default:
		if (find->longlist) {
			longlist(find, f, f, &st, filep);
		} else {
			(void) fprintf(filep, "%s%s\n", f, nul);
		}
	}
	return 1;
}

/******************** exported functions *******************************/

/* parse sentence */
int
find2_parse(find_t *find, const char *s)
{
	str_t *found;
	int	i;

	find->sent.op = NONE;
	i = 0;
	NEW(f2clause_t, find->sent.lhs, "find2_parse: sent lhs", exit(EXIT_FAILURE));
	if (!parse_expr(find->sent.lhs, s, &i)) {
		return 0;
	}
	while (s[i] != 0) {
		if (!recog(s, &i, logicals, &found)) {
			warnx("find2_parse: can't recognise logical operator in `%s'", &s[i]);
			return 0;
		}
		find->sent.op = (int)found->op;
		NEW(f2clause_t, find->sent.rhs, "find2_parse: sent rhs", exit(EXIT_FAILURE));
		if (!parse_expr(find->sent.rhs, s, &i)) {
			return 0;
		}
	}
	return 1;
}

/* do the find command */
int
find2(find_t *find, const char *f, FILE *filep)
{
	struct stat	lst;
	struct stat	st;
	int		depth;
	int		ent;

	depth = 0;
	(void) time(&find->started);
	if (stat(f, &st) < 0) {
		warnx("find2: can't access `%s'", f);
		return 0;
	}
	if (find->samedev) {
		find->dev = st.st_dev;
	}
	ent = (S_ISDIR(st.st_mode)) ? (find->nofollow && lstat(f, &lst) == 0 && S_ISLNK(lst.st_mode)) : 1;
	return (ent) ? dodirent(find, f, depth, filep) : dodir(find, f, depth, filep);
}
