#! /bin/sh

# $NetBSD: genraidconf.sh,v 1.4 2011/05/18 19:17:21 agc Exp $

# Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# small script to set up a RAID configuration
# usage:
#	genraidconf [-d raid3] [-n] [-o raid1] [-o raid5] [-o root]
#		[-o spare] [-o zero] [-r 1|5] [-v] components...
#
# e.g.
#	genraidconf -r 1 -o root -d raid0 sd0 sd1
# makes a RAID1 configuration on /dev/raid0 using sd0 and sd1 disks as
# components which can be use as a root device
#
#	genraidconf -o raid5 -o spare -o zero wd0 wd1 wd2 wd3
# makes a RAID5 configuration using wd0, wd1 and wd2 disks as components,
# and wd3 as hot spare, zeroing the component disks before labelling them

# utility function
die() {
	echo "$1" >&2
	exit 1
}


doit=""
raiddev=raid0
netbsdpartid=169
minraid5=3 # per mrg, no hotswap
raidlevel=1 # default is RAID1 protection
spare=0
zerocomponents=false
enableroot=false
while [ $# -gt 0 ]; do
	case "$1" in
	-d)	raiddev=$2; shift ;;
	-n)	doit="echo" ;;
	-o)	case "$2" in
		raid1) raidlevel=1 ;;
		raid5) raidlevel=5 ;;
		root) enableroot=true ;;
		spare) spare=1 ;;
		zero) zerocomponents=true ;;
		*) die "unknown option \"$2\"" ;;
		esac
		shift ;;
	-r)	raidlevel=$2; shift ;;
	-v)	set -x ;;
	-*)	die "Usage: genraidconf [-d raidN] [-n] [-o raid1|raid5|root|spare|zero] [-r 1|5] [-v] components..." ;;
	*)	break ;;
	esac
	shift
done

# expect them in wd0 wd1 wd2 form
disks=$@

# work out whether we use a dos disk lebel
case $(sysctl kern.rawpartition) in
*2)	wholedisk=c
	firstpart=d
	doslabel=false
	;;
*3)	wholedisk=d
	firstpart=e
	doslabel=true
	;;
*)	die "can't find raw partition number"
	;;
esac

# count number of devices
devc=$#
lastdev=0
case ${spare} in
1)
	devc=$(expr ${devc} - 1)
	lastdev=1
	;;
esac

# count devices, get dimensions, make them into a list
sectorc=0
devs=""
while [ $# -gt ${lastdev} ]; do
	case "$1" in
	vnd*)	part1=a ;;
	*)	part1=${firstpart} ;;
	esac
	devs="${devs}|/dev/$1${part1}"
	nsectorc=$(disklabel $1 2>/dev/null | awk '/^total sectors:/ {print $3;exit}')
	case "${nsectorc}" in
	"")	die "device $1 not found on computer" ;;
	esac
	case "${sectorc}" in
	0)	sectorc=${nsectorc} ;;
	esac
	if [ ${sectorc} -ne ${nsectorc} ]; then
		die "device ${d} has ${nsectorc}, not ${sectorc}"
	fi
	shift
done
devs=$(echo "${devs}" | tr '|' '\n')

# sanity checks
case "${raidlevel}" in
1)	if [ ${devc} -ne 2 ]; then
		die "RAIDframe RAID1 requires 2 disks"
	fi
	case ${spare} in
	1)
		die "RAID 1 protection has no spare"
		;;
	esac
	;;
5)	if [ ${devc} -lt ${minraid5} ]; then
		die "RAIDframe RAID5 requires ${minraid5} data disks"
	fi
	;;
*)	die "Raid 1 or 5 needed (1 is default)"
	;;
esac

# zero out component disks first
if ${zerocomponents}; then
	for d in ${disks}; do
		${doit} dd if=/dev/zero bs=32k of=/dev/r${d}a
	done
fi

# put an fdisk label on there
if ${doslabel}; then
	for d in ${disks}; do
		secs=$(expr ${sectorc} - 2048)
		${doit} fdisk -u -a -i -f -3 -s ${netbsdpartid}/2048/${secs} -b 100/1/${sectorc} -A 2048/2048 ${d}a
	done
fi

# put disklabel onto component disks
if ${doslabel}; then
	nbsectorc=$(expr ${sectorc} - 2048)
	label=$(mktemp /tmp/label.XXXXXX)
	cylc=$(expr ${sectorc} / 2)
	nbcylc=$(expr ${cylc} - 2)
	cat > ${label} << EOF
5 partitions:
#        size  offset  fstype [fsize bsize cpg/sgs]
c: ${nbsectorc}    2048  unused      0     0        # (Cyl.    2*- ${nbcylc})
d: ${sectorc}       0  unused      0     0        # (Cyl.    0 - ${cylc})
e: ${nbsectorc}    2048    RAID                     # (Cyl.    2*- ${nbcylc})
EOF
else
	label=$(mktemp /tmp/label.XXXXXX)
	cylc=$(expr ${sectorc} / 2)
	cat > ${label} << EOF
4 partitions:
#        size  offset  fstype [fsize bsize cpg/sgs]
c: ${sectorc}    2048  unused      0     0        # (Cyl.    2*- ${cylc})
d: ${sectorc}    2048    RAID                     # (Cyl.    2*- ${cylc})
EOF
fi

for d in ${disks}; do
	${doit} disklabel -R ${d} ${label}
done
rm -f ${label}

# if we already have a config file, preserve it in RCS
if [ -e /etc/${raiddev}.conf ]; then
	echo "Preserving previous ${raiddev}.conf in /etc/RCS"
	${doit} mkdir -p /etc/RCS
	(cd /etc && ${doit} ci -l -m"creating new ${raiddev}.conf" -t-"new ${raiddev}.conf file" ${raiddev}.conf)
fi

# set up spare definitions if we need one
case ${spare} in
0)
	spare1="# START spare"
	spare2=""
	;;
1)
	case "$1" in
	vnd*)	part1=a ;;
	*)	part1=${firstpart} ;;
	esac
	spare1="START spare"
	spare2="/dev/${1}${part1}"
	;;
esac
case "${doit}" in
"")
	cat > /etc/${raiddev}.conf << EOF
# automatically generated by genraidconf at $(date)

START array
#numrow numcol numspare
1 ${devc} 0

# Identify physical disks
START disks
${devs}

# Layout is simple - 64 sectors per stripe
START layout
#Sect/StripeUnit StripeUnit/ParityUnit StripeUnit/ReconUnit RaidLevel
128 1 1 ${raidlevel}

${spare1}
${spare2}

# Command queueing
START queue
fifo 100
EOF
	${doit} raidctl -C /etc/${raiddev}.conf ${raiddev}
	;;
esac

# initialise components within raid device
${doit} raidctl -I $(date '+%Y%m%d%H%M%S') ${raiddev}

# initialise parity
${doit} raidctl -i ${raiddev}

# RAIDframe autoconfig
${doit} raidctl -A yes ${raiddev}

# enable as root device
if ${enableroot}; then
	${doit} raidctl -A root ${raiddev}
fi

# unconfigure the raid device
${doit} raidctl -u ${raiddev}

exit 0
