/*-
 * Copyright (c) 2010 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/param.h>

#include <ctype.h>
#include <regex.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define SETLIST_VERSION	"20110522"

enum {
	MAX_DEFS	= 64,
	MAX_VERSION	= 20,
	MAX_SCRIPTS	= 32,
	MAX_LINKS	= 512
};

/* definitions we read or pass in */
typedef struct vars_t {
	char		*mans[MAX_DEFS];	/* manual pages */
	int		 manc;			/* # of man pages */
	char		*incs[MAX_DEFS];	/* headers */
	int		 incc;			/* # of headers */
	char		*defs[MAX_DEFS];	/* definitions */
	int		 defc;			/* # of definitions */
	char		*mlinks[MAX_LINKS];	/* man page hard links */
	int		 mlinkc;		/* # of links */
	char		*links[MAX_LINKS];	/* hard links */
	int		 linkc;			/* # of links */
	char		*symlinks[MAX_LINKS];	/* symbolic links */
	int		 symlinkc;		/* # of symbolic links */
	char		*scripts[MAX_SCRIPTS];	/* scripts */
	int		 scriptc;		/* # of scripts */
	char		 progname[BUFSIZ];	/* program name */
	char		 libname[BUFSIZ];	/* library name */
	char		 bindir[MAXPATHLEN];	/* program directory */
	char		 scriptsdir[MAXPATHLEN];/* scripts directory */
	char		 libdir[MAXPATHLEN];	/* library directory */
	char		 incsdir[MAXPATHLEN];	/* include directory */
	char		 mandir[MAXPATHLEN];	/* manual page directory */
	int		 shlib;			/* non-zero if lib is shared */
	char		 major[MAX_VERSION];	/* shlib major */
	char		 minor[MAX_VERSION];	/* shlib minor */
	char		 teeny[MAX_VERSION];	/* possible minimus version */
	char		*category;		/* category for syspkgs */
} vars_t;

/* for machine arches which use compat libs, define them here */
typedef struct compat_t {
	const char	*type;		/* arch or mach dependent */
	const char	*arch64;	/* 64-bit arch */
	const char	*arch32;	/* equivalent 32-bit arch */
} compat_t;

static compat_t	compats[] = {
	{	"ad",	"mips64eb",	"64"	},
	{	"ad",	"mips64eb",	"o32"	},
	{	"ad",	"mips64el",	"64"	},
	{	"ad",	"mips64el",	"o32"	},
	{	"md",	"amd64",	"i386"	},
	{	"md",	"sparc64",	"sparc"	},
	{	NULL,	NULL,		NULL	}
};

/* format defs */
static char *
pdefs(vars_t *vars, char *buf, size_t size, const char *linedefs)
{
	size_t	cc;
	int	i;
	int	ret;

	cc = snprintf(buf, size, "%s", (linedefs) ? linedefs : "");
	for (i = 0 ; i < vars->defc && cc < size ; cc += ret, i++) {
		ret = snprintf(&buf[cc], size - cc,
			"%s%s",
			((i == 0 && linedefs) || i < vars->defc - 1) ? "," : "",
			vars->defs[i]);
	}
	return buf;
}

/* look for regexps matching wanted fields */
static int
parse(vars_t *vars, const char *in)
{
	static regex_t	 scriptsdir;
	static regex_t	 subscripts;
	static regex_t	 sublinks;
	static regex_t	 symlinks;
	static regex_t	 incsdir;
	static regex_t	 scripts;
	static regex_t	 bindir;
	static regex_t	 libdir;
	static regex_t	 mandir;
	static regex_t	 sub;
	static regex_t	 mlinks;
	static regex_t	 links;
	static regex_t	 shlib;
	static regex_t	 incs;
	static regex_t	 prog;
	static regex_t	 lib;
	static regex_t	 man;
	static int	 compiled;
	regmatch_t	 matches[10];
	const char	*cp;

	if (!compiled) {
		(void) regcomp(&lib, "^[ \t]*LIB[ \t]*=[ \t]*([^ \t]+)",
				REG_EXTENDED);
		(void) regcomp(&libdir, "^[ \t]*LIBDIR[ \t]*=[ \t]*([^ \t]+)",
				REG_EXTENDED);
		(void) regcomp(&scriptsdir,
				"^[ \t]*SCRIPTSDIR[ \t]*=[ \t]*([^ \t]+)",
				REG_EXTENDED);
		(void) regcomp(&subscripts, "^[ \t]*([^ \t.]+)(\\.sh)?",
				REG_EXTENDED);
		(void) regcomp(&scripts, "^[ \t]*SCRIPTS[ \t]*=",
				REG_EXTENDED);
		(void) regcomp(&man, "^[ \t]*MAN[ \t]*=", REG_EXTENDED);
		(void) regcomp(&sub, "^[ \t]*([^ \t]+)", REG_EXTENDED);
		(void) regcomp(&prog, "^[ \t]*PROG[ \t]*=[ \t]*([^ \t]+)",
				REG_EXTENDED);
		(void) regcomp(&links, "^[ \t]*LINKS[ \t]*=",
				REG_EXTENDED);
		(void) regcomp(&bindir, "^[ \t]*BINDIR[ \t]*=[ \t]*([^ \t]+)",
				REG_EXTENDED);
		(void) regcomp(&mandir, "^[ \t]*MANDIR[ \t]*=[ \t]*([^ \t]+)",
				REG_EXTENDED);
		(void) regcomp(&incs, "^[ \t]*INCS[ \t]*=", REG_EXTENDED);
		(void) regcomp(&incsdir, "^[ \t]*INCSDIR[ \t]*=[ \t]*([^ \t]+)",
				REG_EXTENDED);
		(void) regcomp(&mlinks, "^[ \t]*MLINKS[ \t]*=",
				REG_EXTENDED);
		(void) regcomp(&symlinks, "^[ \t]*SYMLINKS[ \t]*=",
				REG_EXTENDED);
		(void) regcomp(&sublinks, "^[ \t]*([^ \t]+)[ \t]+([^ \t]+)",
				REG_EXTENDED);
		(void) regcomp(&shlib, "^[ \t]*SHLIB_VERSION[ \t]*=[ \t]*"
				"([0-9]*)[:.]([0-9]*)[:.]([0-9]*)",
				REG_EXTENDED);
	}
	if (regexec(&lib, in, 10, matches, 0) == 0) {
		(void) snprintf(vars->libname, sizeof(vars->libname), 
			"%.*s",
			(int)(matches[1].rm_eo - matches[1].rm_so),
			&in[(int)matches[1].rm_so]);
		return 1;
	}
	if (regexec(&libdir, in, 10, matches, 0) == 0) {
		(void) snprintf(vars->libdir, sizeof(vars->libdir), 
			"%.*s",
			(int)(matches[1].rm_eo - matches[1].rm_so),
			&in[(int)matches[1].rm_so]);
		return 1;
	}
	if (regexec(&man, in, 10, matches, 0) == 0) {
		for (cp = &in[(int)matches[0].rm_eo] ; 
		     regexec(&sub, cp, 10, matches, 0) == 0 ;
		     cp += matches[0].rm_eo) {
			(void) asprintf(&vars->mans[vars->manc++], 
				"%.*s",
				(int)(matches[1].rm_eo - matches[1].rm_so),
				&cp[(int)matches[1].rm_so]);
		}
		return 1;
	}
	if (regexec(&scripts, in, 10, matches, 0) == 0) {
		for (cp = &in[(int)matches[0].rm_eo] ; 
		     regexec(&subscripts, cp, 10, matches, 0) == 0 ;
		     cp += matches[0].rm_eo) {
			(void) asprintf(&vars->scripts[vars->scriptc++], 
				"%.*s",
				(int)(matches[1].rm_eo - matches[1].rm_so),
				&cp[(int)matches[1].rm_so]);
		}
		return 1;
	}
	if (regexec(&prog, in, 10, matches, 0) == 0) {
		(void) snprintf(vars->progname, sizeof(vars->progname), 
			"%.*s",
			(int)(matches[1].rm_eo - matches[1].rm_so),
			&in[(int)matches[1].rm_so]);
		return 1;
	}
	if (regexec(&bindir, in, 10, matches, 0) == 0) {
		(void) snprintf(vars->bindir, sizeof(vars->bindir), 
			"%.*s",
			(int)(matches[1].rm_eo - matches[1].rm_so),
			&in[(int)matches[1].rm_so]);
		return 1;
	}
	if (regexec(&mandir, in, 10, matches, 0) == 0) {
		(void) snprintf(vars->mandir, sizeof(vars->mandir), 
			"%.*s",
			(int)(matches[1].rm_eo - matches[1].rm_so),
			&in[(int)matches[1].rm_so]);
		return 1;
	}
	if (regexec(&scriptsdir, in, 10, matches, 0) == 0) {
		(void) snprintf(vars->scriptsdir, sizeof(vars->scriptsdir), 
			"%.*s",
			(int)(matches[1].rm_eo - matches[1].rm_so),
			&in[(int)matches[1].rm_so]);
		return 1;
	}
	if (regexec(&incs, in, 10, matches, 0) == 0) {
		for (cp = &in[(int)matches[0].rm_eo] ; 
		     regexec(&sub, cp, 10, matches, 0) == 0 ;
		     cp += matches[0].rm_eo) {
			(void) asprintf(&vars->incs[vars->incc++], 
				"%.*s",
				(int)(matches[1].rm_eo - matches[1].rm_so),
				&cp[(int)matches[1].rm_so]);
		}
		return 1;
	}
	if (regexec(&incsdir, in, 10, matches, 0) == 0) {
		(void) snprintf(vars->incsdir, sizeof(vars->incsdir), 
			"%.*s",
			(int)(matches[1].rm_eo - matches[1].rm_so),
			&in[(int)matches[1].rm_so]);
		return 1;
	}
	if (regexec(&mlinks, in, 10, matches, 0) == 0) {
		for (cp = &in[(int)matches[0].rm_eo] ; 
		     regexec(&sublinks, cp, 10, matches, 0) == 0 ;
		     cp += matches[0].rm_eo) {
			(void) asprintf(&vars->mlinks[vars->mlinkc++], 
				"%.*s",
				(int)(matches[2].rm_eo - matches[2].rm_so),
				&cp[(int)matches[2].rm_so]);
		}
		return 1;
	}
	if (regexec(&links, in, 10, matches, 0) == 0) {
		for (cp = &in[(int)matches[0].rm_eo] ; 
		     regexec(&sublinks, cp, 10, matches, 0) == 0 ;
		     cp += matches[0].rm_eo) {
			(void) asprintf(&vars->links[vars->linkc++], 
				"%.*s",
				(int)(matches[2].rm_eo - matches[2].rm_so),
				&cp[(int)matches[2].rm_so]);
		}
		return 1;
	}
	if (regexec(&symlinks, in, 10, matches, 0) == 0) {
		for (cp = &in[(int)matches[0].rm_eo] ; 
		     regexec(&sublinks, cp, 10, matches, 0) == 0 ;
		     cp += matches[0].rm_eo) {
			(void) asprintf(&vars->symlinks[vars->symlinkc++], 
				"%.*s",
				(int)(matches[2].rm_eo - matches[2].rm_so),
				&cp[(int)matches[2].rm_so]);
		}
		return 1;
	}
	if (regexec(&shlib, in, 10, matches, 0) == 0) {
		if (matches[1].rm_so >= 0) {
			(void) snprintf(vars->major, sizeof(vars->major),
				"%.*s",
				(int)(matches[1].rm_eo - matches[1].rm_so),
				&in[(int)matches[1].rm_so]);
			(void) snprintf(vars->minor, sizeof(vars->minor),
				"%.*s",
				(int)(matches[2].rm_eo - matches[2].rm_so),
				&in[(int)matches[2].rm_so]);
			(void) snprintf(vars->teeny, sizeof(vars->teeny),
				"%.*s",
				(int)(matches[3].rm_eo - matches[3].rm_so),
				&in[(int)matches[3].rm_so]);
			vars->shlib = 1;
		}
		return 1;
	}
	return 0;
}

/* lower-case a string */
static char *
lowerit(const char *in, char *out, size_t size)
{
	const char	*ip;
	char		*op;

	for (op = out, ip = in ;
	     *in && (int)(op - out) < (int)(size - 1) ;
	     ip++, op++) {
		*op = tolower((unsigned char)*ip);
	}
	*op = 0x0;
	return out;
}

/* format then print the setname */
static void
psetname(char *setname, size_t size, const char *fmt, ...)
{
	va_list	args;
	char	newset[BUFSIZ];

	va_start(args, fmt);
	(void) vsnprintf(newset, sizeof(newset), fmt, args);
	va_end(args);
	if (strcmp(newset, setname) != 0) {
		printf("%s", newset);
		(void) snprintf(setname, size, "%s", newset);
	}
}

int
main(int argc, char **argv)
{
	compat_t	*compat;
	vars_t		 vars;
	char		 setname[BUFSIZ];
	char		 buf[BUFSIZ];
	int		 i;

	(void) memset(&vars, 0x0, sizeof(vars));
	while ((i = getopt(argc, argv, "D:Vc:v:")) != -1) {
		switch(i) {
		case 'D':
			if (strncmp(optarg, "MK", 2) == 0) {
				optarg += 2;
			}
			vars.defs[vars.defc++] = strdup(lowerit(optarg, buf,
				sizeof(buf)));
			break;
		case 'V':
			(void) fprintf(stderr, "distlist version: %s\n",
				SETLIST_VERSION);
			exit(EXIT_SUCCESS);
		case 'c':
			vars.category = optarg;
			break;
		case 'v':
			if (!parse(&vars, optarg)) {
				(void) fprintf(stderr, "bad arg: -v %s\n",
					optarg);
			}
			break;
		default:
			(void) fprintf(stderr, "Unknown option '%c'\n", i);
			break;
		}
	}
	(void) snprintf(setname, sizeof(setname), "-- none --");
	/* base sets */
	if (vars.libname[0]) {
		/* libraries */
		if (vars.shlib) {
			/* base compat entries */
			for (compat = compats ; compat->arch64 ; compat++) {
				psetname(setname, sizeof(setname),
					"\n[src/distrib/sets/lists/"
						"base/%s.%s]\n",
					compat->type, compat->arch64);
				printf(".%s/%s/lib%s.so.%s\t\t"
						"base-compat-shlib\t%s\n",
					vars.libdir, compat->arch32,
					vars.libname, vars.major,
					pdefs(&vars, buf, sizeof(buf),
						"compat,pic"));
				printf(".%s/%s/lib%s.so.%s.%s\t\t"
						"base-compat-shlib\t%s\n",
					vars.libdir, compat->arch32,
					vars.libname, vars.major, vars.minor,
					pdefs(&vars, buf, sizeof(buf),
						"compat,pic"));
			}
			psetname(setname, sizeof(setname),
				"\n[src/distrib/sets/lists/base/shl.mi]\n");
			printf(".%s/lib%s.so\t\tbase-%s-shlib\t%s\n",
				vars.libdir, vars.libname, vars.category,
				pdefs(&vars, buf, sizeof(buf), NULL));
			printf(".%s/lib%s.so.%s\t\tbase-%s-shlib\t%s\n",
				vars.libdir, vars.libname, vars.major,
				vars.category,
				pdefs(&vars, buf, sizeof(buf), NULL));
			printf(".%s/lib%s.so.%s.%s\t\tbase-%s-shlib\t%s\n",
				vars.libdir, vars.libname, vars.major,
				vars.minor, vars.category,
				pdefs(&vars, buf, sizeof(buf), NULL));
		}
	}
	/* if a program or a script, then print the set name */
	if (vars.progname[0] || vars.scriptc > 0) {
		psetname(setname, sizeof(setname),
			"\n[src/distrib/sets/lists/base/mi]\n");
	}
	/* programs */
	if (vars.progname[0]) {
		printf(".%s/%s\t\tbase-%s-bin\t%s\n",
			vars.bindir, vars.progname, vars.category,
			pdefs(&vars, buf, sizeof(buf), NULL));
	}
	/* binary links */
	for (i = 0 ; i < vars.linkc ; i++) {
		printf(".%s\t\tbase-%s-bin\t%s\n",
			vars.links[i], vars.category,
			pdefs(&vars, buf, sizeof(buf), NULL));
	}
	/* scripts */
	for (i = 0 ; i < vars.scriptc ; i++) {
		printf(".%s/%s\t\tbase-%s-bin\t%s\n",
			vars.scriptsdir,
			vars.scripts[i],
			vars.category,
			pdefs(&vars, buf, sizeof(buf), NULL));
	}
	/* symbolic links */
	if (vars.symlinkc > 0) {
		for (i = 0 ; i < vars.symlinkc ; i++) {
			printf(".%s\t\tbase-%s-bin\t%s\n",
				vars.symlinks[i], vars.category,
				pdefs(&vars, buf, sizeof(buf), NULL));
		}
	}
	/* comp sets */
	if (vars.incc > 0 || vars.manc > 0 || vars.libname[0]) {
		psetname(setname, sizeof(setname),
			"\n[src/distrib/sets/lists/comp/mi]\n");
	}
	/* header files */
	for (i = 0 ; i < vars.incc ; i++) {
		printf(".%s/%s\t\tcomp-c-include\t%s\n",
			vars.incsdir, vars.incs[i],
			pdefs(&vars, buf, sizeof(buf), NULL));
	}
	/* debug programs */
	if (vars.progname[0]) {
		printf("./usr/libdata/debug%s/%s.debug\t\tcomp-%s-debug\t%s\n",
			vars.bindir, vars.progname, vars.category,
			pdefs(&vars, buf, sizeof(buf), "debug"));
	}
	/* man pages */
	for (i = 0 ; i < vars.manc ; i++) {
		printf(".%s/cat%c/%.*s.0\t\tcomp-%s-catman\t\t%s\n",
			vars.mandir,
			vars.mans[i][strlen(vars.mans[i]) - 1],
			(int)strlen(vars.mans[i]) - 2,
			vars.mans[i], vars.category,
			pdefs(&vars, buf, sizeof(buf), ".cat"));
		printf(".%s/html%c/%.*s.html\t\tcomp-%s-htmlman\t\t%s\n",
			vars.mandir,
			vars.mans[i][strlen(vars.mans[i]) - 1],
			(int)strlen(vars.mans[i]) - 2,
			vars.mans[i], vars.category,
			pdefs(&vars, buf, sizeof(buf), ".html"));
		printf(".%s/man%c/%s\t\tcomp-%s-man\t\t%s\n",
			vars.mandir,
			vars.mans[i][strlen(vars.mans[i]) - 1],
			vars.mans[i], vars.category,
			pdefs(&vars, buf, sizeof(buf), ".man"));
	}
	/* man page links */
	for (i = 0 ; i < vars.mlinkc ; i++) {
		printf(".%s/cat%c/%.*s.0\t\tcomp-%s-catman\t\t%s\n",
			vars.mandir,
			vars.mlinks[i][strlen(vars.mlinks[i]) - 1],
			(int)strlen(vars.mlinks[i]) - 2,
			vars.mlinks[i], vars.category,
			pdefs(&vars, buf, sizeof(buf), ".cat"));
		printf(".%s/html%c/%.*s.html\t\tcomp-%s-htmlman\t\t%s\n",
			vars.mandir,
			vars.mlinks[i][strlen(vars.mlinks[i]) - 1],
			(int)strlen(vars.mlinks[i]) - 2,
			vars.mlinks[i], vars.category,
			pdefs(&vars, buf, sizeof(buf), ".html"));
		printf(".%s/man%c/%s\t\tcomp-%s-man\t\t%s\n",
			vars.mandir,
			vars.mlinks[i][strlen(vars.mlinks[i]) - 1],
			vars.mlinks[i], vars.category,
			pdefs(&vars, buf, sizeof(buf), ".man"));
	}
	/* libs */
	if (vars.libname[0]) {
		printf(".%s/lib%s.a\t\tcomp-%s-lib\t%s\n",
			vars.libdir, vars.libname, vars.category,
			pdefs(&vars, buf, sizeof(buf), NULL));
		printf(".%s/lib%s_g.a\t\tcomp-%s-lib\t%s\n",
			vars.libdir, vars.libname, vars.category,
			pdefs(&vars, buf, sizeof(buf), "debuglib"));
		printf(".%s/lib%s_p.a\t\tcomp-%s-proflib\t%s\n",
			vars.libdir, vars.libname, vars.category,
			pdefs(&vars, buf, sizeof(buf), "profile"));
		printf("./usr/libdata/lint/llib-l%s.ln\t\t"
				"comp-%s-lintlib\t%s\n",
			vars.libname, vars.category,
			pdefs(&vars, buf, sizeof(buf), "lint"));
		if (vars.shlib) {
			psetname(setname, sizeof(setname),
				"\n[src/distrib/sets/lists/comp/shl.mi]\n");
			printf(".%s/lib%s_pic.a\t\tcomp-%s-piclib\t%s\n",
				vars.libdir, vars.libname, vars.category,
				pdefs(&vars, buf, sizeof(buf), NULL));
			printf("./usr/libdata/debug%s/lib%s.so.%s.%s."
					"debug\t\tcomp-%s-debug\t%s\n",
				vars.libdir, vars.libname, vars.major,
				vars.minor, vars.category,
				pdefs(&vars, buf, sizeof(buf), "debug"));
		}
		/* compat entries */
		for (compat = compats ; compat->arch64 ; compat++) {
			psetname(setname, sizeof(setname),
				"\n[src/distrib/sets/lists/comp/%s.%s]\n",
				compat->type, compat->arch64);
			printf(".%s/%s/lib%s.a\t\tcomp-c-lib\t%s\n",
				vars.libdir, compat->arch32, vars.libname,
				pdefs(&vars, buf, sizeof(buf), "compat"));
			printf(".%s/%s/lib%s.so\t\tcomp-sys-shlib\t%s\n",
				vars.libdir, compat->arch32, vars.libname,
				pdefs(&vars, buf, sizeof(buf), "compat"));
			printf(".%s/%s/lib%s_g.a\t\tcomp-c-proflib\t%s\n",
				vars.libdir, compat->arch32, vars.libname,
				pdefs(&vars, buf, sizeof(buf),
					"compat,debuglib"));
			printf(".%s/%s/lib%s_p.a\t\tcomp-c-proflib\t%s\n",
				vars.libdir, compat->arch32, vars.libname,
				pdefs(&vars, buf, sizeof(buf),
					"compat,profile"));
			printf(".%s/%s/lib%s_pic.a\t\tcomp-c-piclib\t%s\n",
				vars.libdir, compat->arch32, vars.libname,
				pdefs(&vars, buf, sizeof(buf),
					"compat,pic"));
			if (vars.shlib) {
				printf("./usr/libdata/debug%s/%s/lib%s.so.%s."
					"%s.debug\t\tcomp-compat-shlib\t%s\n",
					vars.libdir, compat->arch32,
					vars.libname, vars.major, vars.minor,
					pdefs(&vars, buf, sizeof(buf),
						"compat,pic,debug"));
			}
		}
	}
	exit(EXIT_SUCCESS);
}
