/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "hash64.h"

/* hashtest file from Simon Burge:

	ftp://ftp.netbsd.org/pub/NetBSD/misc/simonb/hashtest.tar.gz

See

	http://mail-index.netbsd.org/tech-perform/2001/11/29/0002.html

*/

/* choose some sane defaults for the hashing algorithms */
#define DEFAULT_HASH32_ALG	"perl"
#define DEFAULT_HASH64_ALGS	"perl,mouse"

#ifndef USE_ARG
#define USE_ARG(x)       /*LINTED*/(void)&(x)
#endif

/*
 * Each of the hashing algorithms is split into 3 functions: 
 * initialisation, update, and finalisation.  This forms a template,
 * and is modelled on the way the larger message digest functions work. 
 * Some hashing algorithms from the original article do not map well
 * onto this framework, since they either use the length of the string
 * in the finalisation part, or use more than 32 bits of context (for a
 * 32-bit hashing algorithm).  These hashing functions have not been
 * implemented here.
 */

/************/

/* hash function from perl */

static void
perlhash_init(uint32_t *hash)
{
	*hash = 0;
}

static uint32_t
perlhash(uint32_t *hash, const unsigned char *str, size_t len)
{
	while (len-- > 0) {
		*hash = *hash * 33 + *str++;
	}
	return *hash;
}

static uint32_t
perlhash_final(uint32_t *hash)
{
	*hash = *hash + (*hash >> 5);
	return *hash;
}

/*
 * we special-case this hashing function to provide a _data function,
 * since we use perlhash internally
 */
static uint32_t
perlhash_data(uint32_t *hash, const unsigned char *str, size_t len)
{
	perlhash_init(hash);
	perlhash(hash, str, len);
	return perlhash_final(hash);
}

/************/

/* hash function from der Mouse */

static void
mousehash_init(uint32_t *hash)
{
	*hash = 0xffffffff;
}

static uint32_t
mousehash(uint32_t *hash, const unsigned char *str, size_t len)
{
	while (len-- > 0) {
		*hash ^= *str++;
		*hash = (*hash >> 1) ^ ((*hash & 1) ? 0xedb88320 : 0);
	}
	return *hash;
}

static uint32_t
mousehash_final(uint32_t *hash)
{
	return *hash;
}

/************/

/* traditional "dumb" hash function */

static void
dumbhash_init(uint32_t *hash)
{
	*hash = 0;
}

static uint32_t
dumbhash(uint32_t *hash, const unsigned char *str, size_t len)
{
	while (len-- > 0) {
		*hash += *str++;
	}
	return *hash;
}

static uint32_t
dumbhash_final(uint32_t *hash)
{
	return *hash;
}

/************/

/* a bit more sophisticated - "dumb" hash function via multiplication */

static void
dumbmulhash_init(uint32_t *hash)
{
	*hash = 0;
}

static uint32_t
dumbmulhash(uint32_t *hash, const unsigned char *str, size_t len)
{
	size_t i;

	for (i = 1; i <= len; i++) {
		*hash += (uint32_t)(*str++ * i);
	}
	return *hash;
}

static uint32_t
dumbmulhash_final(uint32_t *hash)
{
	return *hash;
}

/************/

/* CRC-derived hashing function */

static void
crchash_init(uint32_t *hash)
{
	*hash = 0;
}

static uint32_t
crchash(uint32_t *hash, const unsigned char *str, size_t len)
{
	unsigned	highorder;

	while (len-- > 0) {
		highorder = *hash & 0xf8000000;
		*hash = *hash << 5;
		*hash = *hash ^ (highorder >> 27);
		*hash = *hash ^ *str++;
	}
	return *hash;
}

static uint32_t
crchash_final(uint32_t *hash)
{
	return *hash;
}

/************/

/* the perl xor-based hashing function */

static void
perlxorhash_init(uint32_t *hash)
{
	*hash = 0;
}

static uint32_t
perlxorhash(uint32_t *hash, const unsigned char *str, size_t len)
{
	while (len-- > 0) {
		*hash = *hash * 33 ^ *str++;
	}
	return *hash;
}

static uint32_t
perlxorhash_final(uint32_t *hash)
{
	*hash = *hash + (*hash >> 5);
	return *hash;
}

/************/

/* the python hashing function */

static void
pythonhash_init(uint32_t *hash)
{
	*hash = 0;
}

static uint32_t
pythonhash(uint32_t *hash, const unsigned char *str, size_t len)
{
	while (len--) {
		*hash += *str++;
		*hash += (*hash << 10);
		*hash ^= (*hash >> 6);
	}
	return *hash;
}

static uint32_t
pythonhash_final(uint32_t *hash)
{
	*hash += (*hash << 3);
	*hash ^= (*hash >> 11);
	*hash = (*hash + (*hash << 15));
	return *hash;
}

/************/

/* hashing function by Dan Bernstein */

static void
bernsteinhash_init(uint32_t *hash)
{
	*hash = 5381;
}

static uint32_t
bernsteinhash(uint32_t *hash, const unsigned char *str, size_t len)
{
	while (len-- > 0) {
		*hash = *hash * 33 + *str++;
	}
	return *hash;
}

static uint32_t
bernsteinhash_final(uint32_t *hash)
{
	*hash = *hash + (*hash >> 5);
	return *hash;
}

/************/

/* hashing function by Peter Honeyman */

static uint32_t honeymanCrcTable[128] = {
	0, 1207959552, 603979776, 1811939328, 301989888, 1509949440, 905969664,
	2113929216, 150994944, 1090519040, 754974720, 1694498816, 452984832,
	1392508928, 1056964608, 1996488704, 75497472, 1283457024, 545259520,
	1753219072, 377487360, 1585446912, 847249408, 2055208960, 226492416,
	1166016512, 696254464, 1635778560, 528482304, 1468006400, 998244352,
	1937768448, 37748736, 1245708288, 641728512, 1849688064, 272629760,
	1480589312, 876609536, 2084569088, 188743680, 1128267776, 792723456,
	1732247552, 423624704, 1363148800, 1027604480, 1967128576, 113246208,
	1321205760, 583008256, 1790967808, 348127232, 1556086784, 817889280,
	2025848832, 264241152, 1203765248, 734003200, 1673527296, 499122176,
	1438646272, 968884224, 1908408320, 18874368, 1226833920, 622854144,
	1830813696, 320864256, 1528823808, 924844032, 2132803584, 136314880,
	1075838976, 740294656, 1679818752, 438304768, 1377828864, 1042284544,
	1981808640, 94371840, 1302331392, 564133888, 1772093440, 396361728,
	1604321280, 866123776, 2074083328, 211812352, 1151336448, 681574400,
	1621098496, 513802240, 1453326336, 983564288, 1923088384, 56623104,
	1264582656, 660602880, 1868562432, 291504128, 1499463680, 895483904,
	2103443456, 174063616, 1113587712, 778043392, 1717567488, 408944640,
	1348468736, 1012924416, 1952448512, 132120576, 1340080128, 601882624,
	1809842176, 367001600, 1574961152, 836763648, 2044723200, 249561088,
	1189085184, 719323136, 1658847232, 484442112, 1423966208, 954204160,
	1893728256,
};

static void
honeymanhash_init(uint32_t *hash)
{
	*hash = 0;
}

static uint32_t
honeymanhash(uint32_t *hash, const unsigned char *str, size_t len)
{
	while (len-- > 0) {
		*hash = (*hash >> 7) ^ honeymanCrcTable[(*hash ^ (*str++)) & 0x7f];
	}
	return *hash;
}

static uint32_t
honeymanhash_final(uint32_t *hash)
{
	return *hash;
}

/************/

/* hashing function by Peter Weinberger */

static void
pjwhash_init(uint32_t *hash)
{
	*hash = 0;
}

/* Defines the so called `hashpjw' function by P.J. Weinberger
   [see Aho/Sethi/Ullman, COMPILERS: Principles, Techniques and Tools,
   1986, 1987 Bell Telephone Laboratories, Inc.]  */

/* We assume to have `unsigned long int' value with at least 32 bits.  */
#define	HASHWORDBITS	32

static uint32_t
pjwhash(uint32_t *hash, const unsigned char *str, size_t len)
{
	unsigned g;

	/* Compute the hash value for the given string.  */
	while (len-- > 0) {
		*hash <<= 4;
		*hash += *str++;
		g = *hash & (0xf << (HASHWORDBITS - 4));
		if (g != 0) {
			*hash ^= g >> (HASHWORDBITS - 8);
			*hash ^= g;
		}
	}
	return *hash;
}

static uint32_t
pjwhash_final(uint32_t *hash)
{
	return *hash;
}

/************/

/* hashing function by Chris Torek */

static void
torekhash_init(uint32_t *hash)
{
	*hash = 0;
}

static uint32_t
torekhash(uint32_t *hash, const unsigned char *str, size_t len)
{
	size_t loop;

#define HASH4a   *hash = (*hash << 5) - *hash + *str++;
#define HASH4b   *hash = (*hash << 5) + *hash + *str++;
#define HASH4 HASH4b

	if (len > 0) {
		loop = (len + 8 - 1) >> 3;

		switch (len & (8 - 1)) {
		case 0:
			do {
				HASH4;
				/* FALLTHROUGH */
		case 7:
				HASH4;
				/* FALLTHROUGH */
		case 6:
				HASH4;
				/* FALLTHROUGH */
		case 5:
				HASH4;
				/* FALLTHROUGH */
		case 4:
				HASH4;
				/* FALLTHROUGH */
		case 3:
				HASH4;
				/* FALLTHROUGH */
		case 2:
				HASH4;
				/* FALLTHROUGH */
		case 1:
				HASH4;
			} while (--loop);
		}
	}
	return *hash;
}

static uint32_t
torekhash_final(uint32_t *hash)
{
	return *hash;
}

/************/

/* hashing function from Berkeley yacc */

static void
byacchash_init(uint32_t *hash)
{
	*hash = 0;
}

static uint32_t
byacchash(uint32_t *hash, const unsigned char *str, size_t len)
{
	while (len-- > 0) {
		*hash = 31 * *hash + *str++;
	}
	return *hash;
}

static uint32_t
byacchash_final(uint32_t *hash)
{
	return *hash;
}

/************/

/* hashing function from tcl */

static void
tclhash_init(uint32_t *hash)
{
	*hash = 0;
}

static uint32_t
tclhash(uint32_t *hash, const unsigned char *str, size_t len)
{
	while (len-- > 0) {
		*hash = (*hash << 3) + *hash + *str++;
	}
	return *hash;
}

static uint32_t
tclhash_final(uint32_t *hash)
{
	return *hash;
}

/************/

/* Oz's original sdbm hash, copied from Margo Seltzers db package. */

static void
gawkhash_init(uint32_t *hash)
{
	*hash = 0;
}

static uint32_t
gawkhash(uint32_t *hash, const unsigned char *s, size_t len)
{
	/*
	 * This is INCREDIBLY ugly, but fast.  We break the string up into
	 * 8 byte units.  On the first time through the loop we get the
	 * "leftover bytes" (strlen % 8).  On every other iteration, we
	 * perform 8 HASHC's so we handle all 8 bytes.  Essentially, this
	 * saves us 7 cmp & branch instructions.  If this routine is
	 * heavily used enough, it's worth the ugly coding.
	 *
	 * OZ's original sdbm hash, copied from Margo Seltzers db package.
	 */

	/*
	 * Even more speed:
	 * #define HASHC   h = *s++ + 65599 * h
	 * Because 65599 = pow(2, 6) + pow(2, 16) - 1 we multiply by shifts
	 */
#define HASHC   htmp = (*hash << 6); *hash = *s++ + htmp + (htmp << 10) - *hash

	unsigned int htmp;

	/* "Duff's Device" for those who can handle it */
	if (len > 0) {
		size_t loop = (len + 8 - 1) >> 3;

		switch (len & (8 - 1)) {
		case 0:
			do {    /* All fall throughs */
				HASHC; /*FALLTHROUGH*/
		case 7:         HASHC; /*FALLTHROUGH*/
		case 6:         HASHC; /*FALLTHROUGH*/
		case 5:         HASHC; /*FALLTHROUGH*/
		case 4:         HASHC; /*FALLTHROUGH*/
		case 3:         HASHC; /*FALLTHROUGH*/
		case 2:         HASHC; /*FALLTHROUGH*/
		case 1:         HASHC; /*FALLTHROUGH*/
			} while (--loop);
		}
	}
	return *hash;
}

static uint32_t
gawkhash_final(uint32_t *hash)
{
	return *hash;
}

/************/

/* unknown origin "nem" hashing algorithm */

static void
nemhash_init(uint32_t *hash)
{
	*hash = 0;
}

static uint32_t
nemhash(uint32_t *hash, const unsigned char *str, size_t len)
{
	while (len--) {
		*hash += (*hash << 8) + *str++;
	}
	return *hash;
}

static uint32_t
nemhash_final(uint32_t *hash)
{
	*hash = (*hash + (*hash << 8));
	return *hash;
}

/************/

/* simple size indication of the hashed data - useful for a number of things */

static void
size_init(uint32_t *hash)
{
	*hash = 0;
}

static uint32_t
size_update(uint32_t *hash, const unsigned char *str, size_t len)
{
	USE_ARG(str);
	*hash += (uint32_t) len;
	return *hash;
}

static uint32_t
size_final(uint32_t *hash)
{
	return *hash;
}

/************/

/*
 * use perlhash internally to store the hash value for each of the
 * hashing function.  we can then use a sparse jump table in the form
 * of a switch statement to get the required hash function.  This
 * initialises the context for the required hash algorithm, which is
 * not necessarily always 0
 */

static uint32_t
hashinit(HASH64_CTX *ctx, int indx, const char *alg, size_t len)
{
	uint32_t	hash;

	switch(ctx->algs[indx] = perlhash_data(&hash,
				(const uint8_t *)alg, len)) {
	case /* perlhash */ 0xd3babdd6:
	case /* perl */ 0x0041206c:
		perlhash_init(&ctx->hashes[indx]);
		break;
	case /* mousehash */ 0x4de0f41a:
	case /* mouse */ 0x0832e858:
		mousehash_init(&ctx->hashes[indx]);
		break;
	case /* dumbhash */ 0x816b47b6:
	case /* dumb */ 0x003a9cb8:
		dumbhash_init(&ctx->hashes[indx]);
		break;
	case /* dumbmulhash */ 0xe163b02d:
	case /* dumbmul */ 0x2bec12f7:
		dumbmulhash_init(&ctx->hashes[indx]);
		break;
	case /* crchash */ 0xdc9b2197:
	case /* crc */ 0x0001c1d9:
		crchash_init(&ctx->hashes[indx]);
		break;
	case /* perlxorhash */ 0xa150cff6:
	case /* perlxor */ 0xa66ac9e8:
		perlxorhash_init(&ctx->hashes[indx]);
		break;
	case /* pythonhash */ 0x336ac3ec:
	case /* python */ 0x0e814686:
		pythonhash_init(&ctx->hashes[indx]);
		break;
	case /* bernsteinhash */ 0x9419892e:
	case /* bernstein */ 0x1ef544a8:
		bernsteinhash_init(&ctx->hashes[indx]);
		break;
	case /* honeymanhash */ 0xc9bfa906:
	case /* honeyman */ 0x66581aeb:
		honeymanhash_init(&ctx->hashes[indx]);
		break;
	case /* pjwhash */ 0xb2cc8a47:
	case /* pjw */ 0x0001f9e5:
		pjwhash_init(&ctx->hashes[indx]);
		break;
	case /* torekhash */ 0x775f94ee:
	case /* torek */ 0x08b57a3c:
		torekhash_init(&ctx->hashes[indx]);
		break;
	case /* byacchash */ 0x5e989f90:
	case /* byacc */ 0x076af02a:
		byacchash_init(&ctx->hashes[indx]);
		break;
	case /* tclhash */ 0xd6b0c1a2:
	case /* tcl */ 0x00020a78:
		tclhash_init(&ctx->hashes[indx]);
		break;
	case /* gawkhash */ 0x5963f3ab:
	case /* gawk */ 0x003bf8a5:
		gawkhash_init(&ctx->hashes[indx]);
		break;
	case /* nemhash */ 0x0746b6c9:
	case /* nem */ 0x0001f06b:
		nemhash_init(&ctx->hashes[indx]);
		break;
	case /* size */ 0x0042e54d:
		size_init(&ctx->hashes[indx]);
		break;
	default:
		(void) fprintf(stderr, "%.*s not found\n", (int)len, alg);
		return 0;
	}
	return ctx->algs[indx];
}

/* perform the update function per hash */
static uint32_t
hashupdate(HASH64_CTX *ctx, int indx, const uint8_t *str, size_t len)
{
	switch(ctx->algs[indx]) {
	case /* perlhash */ 0xd3babdd6:
	case /* perl */ 0x0041206c:
		perlhash(&ctx->hashes[indx], str, len);
		break;
	case /* mousehash */ 0x4de0f41a:
	case /* mouse */ 0x0832e858:
		mousehash(&ctx->hashes[indx], str, len);
		break;
	case /* dumbhash */ 0x816b47b6:
	case /* dumb */ 0x003a9cb8:
		dumbhash(&ctx->hashes[indx], str, len);
		break;
	case /* dumbmulhash */ 0xe163b02d:
	case /* dumbmul */ 0x2bec12f7:
		dumbmulhash(&ctx->hashes[indx], str, len);
		break;
	case /* crchash */ 0xdc9b2197:
	case /* crc */ 0x0001c1d9:
		crchash(&ctx->hashes[indx], str, len);
		break;
	case /* perlxorhash */ 0xa150cff6:
	case /* perlxor */ 0xa66ac9e8:
		perlxorhash(&ctx->hashes[indx], str, len);
		break;
	case /* pythonhash */ 0x336ac3ec:
	case /* python */ 0x0e814686:
		pythonhash(&ctx->hashes[indx], str, len);
		break;
	case /* bernsteinhash */ 0x9419892e:
	case /* bernstein */ 0x1ef544a8:
		bernsteinhash(&ctx->hashes[indx], str, len);
		break;
	case /* honeymanhash */ 0xc9bfa906:
	case /* honeyman */ 0x66581aeb:
		honeymanhash(&ctx->hashes[indx], str, len);
		break;
	case /* pjwhash */ 0xb2cc8a47:
	case /* pjw */ 0x0001f9e5:
		pjwhash(&ctx->hashes[indx], str, len);
		break;
	case /* torekhash */ 0x775f94ee:
	case /* torek */ 0x08b57a3c:
		torekhash(&ctx->hashes[indx], str, len);
		break;
	case /* byacchash */ 0x5e989f90:
	case /* byacc */ 0x076af02a:
		byacchash(&ctx->hashes[indx], str, len);
		break;
	case /* tclhash */ 0xd6b0c1a2:
	case /* tcl */ 0x00020a78:
		tclhash(&ctx->hashes[indx], str, len);
		break;
	case /* gawkhash */ 0x5963f3ab:
	case /* gawk */ 0x003bf8a5:
		gawkhash(&ctx->hashes[indx], str, len);
		break;
	case /* nemhash */ 0x0746b6c9:
	case /* nem */ 0x0001f06b:
		nemhash(&ctx->hashes[indx], str, len);
		break;
	case /* size */ 0x0042e54d:
		size_update(&ctx->hashes[indx], str, len);
		break;
	default:
		(void) fprintf(stderr, "%08x not found\n", ctx->algs[indx]);
		return 0;
	}
	return ctx->hashes[indx];
}

/* perform the finalising function per hash */
static uint32_t
hashfinal(HASH64_CTX *ctx, int indx)
{
	switch(ctx->algs[indx]) {
	case /* perlhash */ 0xd3babdd6:
	case /* perl */ 0x0041206c:
		perlhash_final(&ctx->hashes[indx]);
		break;
	case /* mousehash */ 0x4de0f41a:
	case /* mouse */ 0x0832e858:
		mousehash_final(&ctx->hashes[indx]);
		break;
	case /* dumbhash */ 0x816b47b6:
	case /* dumb */ 0x003a9cb8:
		dumbhash_final(&ctx->hashes[indx]);
		break;
	case /* dumbmulhash */ 0xe163b02d:
	case /* dumbmul */ 0x2bec12f7:
		dumbmulhash_final(&ctx->hashes[indx]);
		break;
	case /* crchash */ 0xdc9b2197:
	case /* crc */ 0x0001c1d9:
		crchash_final(&ctx->hashes[indx]);
		break;
	case /* perlxorhash */ 0xa150cff6:
	case /* perlxor */ 0xa66ac9e8:
		perlxorhash_final(&ctx->hashes[indx]);
		break;
	case /* pythonhash */ 0x336ac3ec:
	case /* python */ 0x0e814686:
		pythonhash_final(&ctx->hashes[indx]);
		break;
	case /* bernsteinhash */ 0x9419892e:
	case /* bernstein */ 0x1ef544a8:
		bernsteinhash_final(&ctx->hashes[indx]);
		break;
	case /* honeymanhash */ 0xc9bfa906:
	case /* honeyman */ 0x66581aeb:
		honeymanhash_final(&ctx->hashes[indx]);
		break;
	case /* pjwhash */ 0xb2cc8a47:
	case /* pjw */ 0x0001f9e5:
		pjwhash_final(&ctx->hashes[indx]);
		break;
	case /* torekhash */ 0x775f94ee:
	case /* torek */ 0x08b57a3c:
		torekhash_final(&ctx->hashes[indx]);
		break;
	case /* byacchash */ 0x5e989f90:
	case /* byacc */ 0x076af02a:
		byacchash_final(&ctx->hashes[indx]);
		break;
	case /* tclhash */ 0xd6b0c1a2:
	case /* tcl */ 0x00020a78:
		tclhash_final(&ctx->hashes[indx]);
		break;
	case /* gawkhash */ 0x5963f3ab:
	case /* gawk */ 0x003bf8a5:
		gawkhash_final(&ctx->hashes[indx]);
		break;
	case /* nemhash */ 0x0746b6c9:
	case /* nem */ 0x0001f06b:
		nemhash_final(&ctx->hashes[indx]);
		break;
	case /* size */ 0x0042e54d:
		size_final(&ctx->hashes[indx]);
		break;
	default:
		(void) fprintf(stderr, "%08x not found\n", ctx->algs[indx]);
		return 0;
	}
	return ctx->hashes[indx];
}

/*****************************************************************************/

/* initialise the 64-bit contexts according to the algorithms suggested */
int
HASH64_Init(HASH64_CTX *ctx, const char *algs)
{
	const char	*comma;

	(void) memset(ctx, 0x0, sizeof(*ctx));
	if (algs == NULL) {
		algs = DEFAULT_HASH64_ALGS;
	}
	if ((comma = strchr(algs, ',')) == NULL) {
		(void) fprintf(stderr, "need 2 hash algs in '%s'\n", algs);
		return 0;
	}
	if (!hashinit(ctx, 0, algs, (size_t)(comma - algs)) ||
	    !hashinit(ctx, 1, comma + 1, strlen(comma + 1))) {
		(void) fprintf(stderr, "hash algs not found '%s'\n", algs);
		return 0;
	}
	return 1;
}

/* update the 64-bit contexts according to the algorithms stored */
int
HASH64_Update(HASH64_CTX *ctx, const uint8_t *str, size_t len)
{
	if (!hashupdate(ctx, 0, str, len) || !hashupdate(ctx, 1, str, len)) {
		(void) fprintf(stderr, "HASH64_Update failed\n");
		return 0;
	}
	return 1;
}

/* finalise the 64-bit contexts according to the algorithms stored */
int
HASH64_Final(uint64_t *hash, HASH64_CTX *ctx)
{
	if (!hashfinal(ctx, 0) || !hashfinal(ctx, 1)) {
		(void) fprintf(stderr, "HASH64_Final failed\n");
		return 0;
	}
	*hash = ((uint64_t)ctx->hashes[0] << 32) | ctx->hashes[1];
	return 1;
}

/* finalise the 64-bit contexts according to the algorithms stored */
uint64_t
HASH64_End(HASH64_CTX *ctx, uint64_t *hash)
{
	if (!hashfinal(ctx, 0) || !hashfinal(ctx, 1)) {
		(void) fprintf(stderr, "HASH64_End failed\n");
		return 0;
	}
	*hash = ((uint64_t)ctx->hashes[0] << 32) | ctx->hashes[1];
	return *hash;
}

/* higher level function - calculate the 64-bit hash over a file */
uint64_t
HASH64_File(char *filename, uint64_t *hash, const char *algs)
{   
	HASH64_CTX	ctx;
	uint8_t		buffer[BUFSIZ];
	ssize_t		num;
	int		fd;
	int		oerrno;

	HASH64_Init(&ctx, algs);
	if ((fd = open(filename, O_RDONLY)) < 0) {
		return 0;
	}
	while ((num = read(fd, buffer, sizeof(buffer))) > 0) {
		HASH64_Update(&ctx, buffer, (size_t)num);
	}
	oerrno = errno;
	close(fd);
	errno = oerrno;
	return (num < 0) ? 0 : HASH64_End(&ctx, hash);
}

/* higher level function - calculate the 64-bit hash over bounded data */
uint64_t
HASH64_Data(const uint8_t *data, size_t len, uint64_t *hash, const char *algs)
{   
	HASH64_CTX	ctx;

	HASH64_Init(&ctx, algs);
	HASH64_Update(&ctx, data, len);
	return HASH64_End(&ctx, hash);
}

/* print a list of algorithms on the file pointer */
int
HASH64_ListAlgs(char *buf, size_t size)
{ 
	return snprintf(buf, size,
			"perlhash/perl\n"
			"mousehash/mouse\n"
			"dumbhash/dumb\n"
			"dumbmulhash/dumbmul\n"
			"crchash/crc\n"
			"perlxorhash/perlxor\n"
			"pythonhash/python\n"
			"bernsteinhash/bernstein\n"
			"honeymanhash/honeyman\n"
			"pjwhash/pjw\n"
			"torekhash/torek\n"
			"byacchash/byacc\n"
			"tclhash/tcl\n"
			"gawkhash/gawk\n"
			"nemhash/nem\n"
			"size\n"
		);
}

/* initialise the 32-bit contexts according to the algorithm suggested */
int
HASH32_Init(HASH64_CTX *ctx, const char *algs)
{
	(void) memset(ctx, 0x0, sizeof(*ctx));
	if (algs == NULL) {
		algs = DEFAULT_HASH32_ALG;
	}
	if (!hashinit(ctx, 0, algs, strlen(algs))) {
		(void) fprintf(stderr, "hash algs not found '%s'\n", algs);
		return 0;
	}
	return 1;
}

/* update the 32-bit contexts according to the algorithm stored */
int
HASH32_Update(HASH64_CTX *ctx, const uint8_t *str, size_t len)
{
	return hashupdate(ctx, 0, str, len);
}

/* finalise the 32-bit contexts according to the algorithm stored */
int
HASH32_Final(uint32_t *hash, HASH64_CTX *ctx)
{
	if (!hashfinal(ctx, 0)) {
		(void) fprintf(stderr, "HASH32_Final failed\n");
		return 0;
	}
	*hash = ctx->hashes[0];
	return 1;
}

/* finalise the 32-bit contexts according to the algorithm stored */
uint32_t
HASH32_End(HASH64_CTX *ctx, uint32_t *hash)
{
	if (!hashfinal(ctx, 0)) {
		(void) fprintf(stderr, "HASH32_End failed\n");
		return 0;
	}
	*hash = ctx->hashes[0];
	return *hash;
}

/* higher level function - calculate the 32-bit hash over a file */
uint32_t
HASH32_File(char *filename, uint32_t *hash, const char *algs)
{   
	HASH64_CTX	ctx;
	uint8_t		buffer[BUFSIZ];
	ssize_t		num;
	int		fd;
	int		oerrno;

	HASH32_Init(&ctx, algs);
	if ((fd = open(filename, O_RDONLY)) < 0) {
		return 0;
	}
	while ((num = read(fd, buffer, sizeof(buffer))) > 0) {
		HASH32_Update(&ctx, buffer, (size_t)num);
	}
	oerrno = errno;
	close(fd);
	errno = oerrno;
	return (num < 0) ? 0 : HASH32_End(&ctx, hash);
}

/* higher level function - calculate the 32-bit hash over bounded data */
uint32_t
HASH32_Data(const uint8_t *data, size_t len, uint32_t *hash, const char *algs)
{   
	HASH64_CTX	ctx;

	HASH32_Init(&ctx, algs);
	HASH32_Update(&ctx, data, len);
	return HASH32_End(&ctx, hash);
}
