/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "hmac.h"

/* calculate the hmac on a file */
static int
dofile(const char *f, const char *alg, uint8_t *key)
{
	struct stat	 st;
	uint8_t		 mac[HMAC_MAX_DIGEST_SIZE];
	uint8_t		*msg;
	char		 strmac[HMAC_MAX_STRING_SIZE];
	FILE		*fp;
	int		 len;
	int		 ret;

	if ((fp = fopen(f, "r")) == NULL) {
		(void) fprintf(stderr, "hmac: can't open '%s'\n", f);
		return 0;
	}
	(void) fstat(fileno(fp), &st);
	msg = mmap(NULL, (size_t)st.st_size, PROT_READ, MAP_FILE, fileno(fp), 0);
	if (msg == MAP_FAILED) {
		(void) fprintf(stderr, "hmac: mmap failed '%s'\n", f);
		return 0;
	}
	if ((len = hmac(alg, key, strlen((char *)key), msg, st.st_size, mac, sizeof(mac))) != 0) {
		hmac_bin2str(mac, len, strmac, sizeof(strmac));
		if (strncmp(alg, "hmac", 4) != 0) {
			printf("hmac-");
		}
		printf("%s (%s) = %s\n", alg, f, strmac);
		ret = 1;
	} else {
		ret = 0;
	}
	munmap(msg, st.st_size);
	return ret;
}

/* read a key from a file */
static ssize_t
readkey(const char *f, uint8_t *key, size_t keysize)
{
	ssize_t	 cc;
	FILE	*fp;

	if ((fp = fopen(f, "r")) == NULL) {
		(void) fprintf(stderr, "hmac: can't open keyfile '%s'\n", f);
		return -1;
	}
	if ((cc = read(fileno(fp), key, keysize)) < 0) {
		(void) fprintf(stderr, "hmac: can't read keyfile '%s'\n", f);
		cc = -1;
	}
	(void) fclose(fp);
	return cc;
}

#define MB(x)	(1024 * 1024 * (x))

/* read from stdin (up to a max size) */
static int
hmacstdin(const char *alg, uint8_t *key, int sz)
{
	uint8_t	*buf;
	uint8_t	 mac[HMAC_MAX_DIGEST_SIZE];
	char	 strmac[HMAC_MAX_STRING_SIZE];
	int	 len;
	int	 cc;
	int	 i;

	if ((buf = calloc(1, (unsigned)sz)) == NULL) {
		(void) fprintf(stderr, "can't allocate %d bytes\n", sz);
		return 0;
	}
	for (cc = 0 ; (i = read(STDIN_FILENO, &buf[cc], (unsigned)(sz - cc))) > 0 && cc < sz ; cc += i) {
	}
	if ((len = hmac(alg, key, strlen((char *)key), buf, cc, mac, sizeof(mac))) != 0) {
		len = hmac_bin2str(mac, len, strmac, sizeof(strmac));
		printf("%.*s\n", len, strmac);
		free(buf);
		return 1;
	}
	free(buf);
	return 0;
}

int
main(int argc, char **argv)
{
	const char	*alg;
	uint8_t		 key[128];
	char		*keyfile;
	char		*s;
	int		 sz;
	int		 ex;
	int		 i;

	alg = "sha512";
	keyfile = NULL;
	(void) memset(key, 0x0, sizeof(key));
	sz = MB(4);
	while ((i = getopt(argc, argv, "a:f:k:s:")) != -1) {
		switch(i) {
		case 'a':
			alg = optarg;
			break;
		case 'f':
			keyfile = optarg;
			break;
		case 'k':
			(void) snprintf((char *)key, sizeof(key), "%s", optarg);
			break;
		case 's':
			sz = atoi(optarg);
			break;
		default:
			break;
		}
	}
	if (keyfile) {
		if (readkey(keyfile, key, sizeof(key)) < 0) {
			exit(EXIT_FAILURE);
		}
	} else if (key[0] == 0x0) {
		s = getpass("hmac key: ");
		(void) snprintf((char *)key, sizeof(key), "%s", s);
	}
	if (optind == argc) {
		if (!hmacstdin(alg, key, sz)) {
			ex = EXIT_FAILURE;
		}
	} else {
		for (ex = EXIT_SUCCESS, i = optind ; i < argc ; i++) {
			if (!dofile(argv[i], alg, key)) {
				ex = EXIT_FAILURE;
			}
		}
	}
	exit(EXIT_SUCCESS);
}
