/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <string.h>
#include <stdio.h>

#include "hmacwrap.h"
#include "hmac.h"

/*************************************************************************/

void
hmac_rmd160_init(hmac_rmd160_ctx_t *ctx, const uint8_t *keyarg, size_t key_size)
{
	const uint8_t		*key;
	uint8_t		 key_dig[(RMD160_DIGEST_SIZE * 2) + 1];
	uint8_t		 key_temp[RMD160_DIGEST_SIZE];
	size_t		 fillc;
	size_t		 keylen;
	size_t		 i;

	if (key_size > RMD160_BLOCK_SIZE){
		key = key_temp;
		keylen = RMD160_DIGEST_SIZE;
		RMD160Data(keyarg, key_size, (char *)key_dig);
		hmac_str2bin(key_dig, sizeof(key_dig), key_temp, sizeof(key_temp));
	} else {
		/* key_size <= RMD160_BLOCK_SIZE */
		key = keyarg;
		keylen = key_size;
	}
	fillc = RMD160_BLOCK_SIZE - keylen;
	memset(ctx->block_ipad + keylen, 0x36, fillc);
	memset(ctx->block_opad + keylen, 0x5c, fillc);
	for (i = 0; i < keylen; i++) {
		ctx->block_ipad[i] = key[i] ^ 0x36;
		ctx->block_opad[i] = key[i] ^ 0x5c;
	}

	RMD160Init(&ctx->ctx_inside);
	RMD160Update(&ctx->ctx_inside, ctx->block_ipad, RMD160_BLOCK_SIZE);

	RMD160Init(&ctx->ctx_outside);
	RMD160Update(&ctx->ctx_outside, ctx->block_opad,
		RMD160_BLOCK_SIZE);

	/* for hmac_reinit */
	memcpy(&ctx->ctx_inside_reinit, &ctx->ctx_inside,
		sizeof(ctx->ctx_inside_reinit));
	memcpy(&ctx->ctx_outside_reinit, &ctx->ctx_outside,
		sizeof(ctx->ctx_outside_reinit));
}

void
hmac_rmd160_reinit(hmac_rmd160_ctx_t *ctx)
{
	memcpy(&ctx->ctx_inside, &ctx->ctx_inside_reinit,
		sizeof(ctx->ctx_inside));
	memcpy(&ctx->ctx_outside, &ctx->ctx_outside_reinit,
		sizeof(ctx->ctx_outside));
}

void
hmac_rmd160_update(hmac_rmd160_ctx_t *ctx, const uint8_t *message,
                        size_t message_len)
{
	RMD160Update(&ctx->ctx_inside, message, (unsigned)message_len);
}

void
hmac_rmd160_final(hmac_rmd160_ctx_t *ctx, uint8_t *mac, size_t mac_size)
{
	uint8_t digest_inside[RMD160_DIGEST_SIZE];
	uint8_t mac_temp[RMD160_DIGEST_SIZE];

	RMD160Final(digest_inside, &ctx->ctx_inside);
	RMD160Update(&ctx->ctx_outside, digest_inside, RMD160_DIGEST_SIZE);
	RMD160Final(mac_temp, &ctx->ctx_outside);
	memcpy(mac, mac_temp, mac_size);
}

void
hmac_rmd160(const uint8_t *key, size_t key_size,
          const uint8_t *message, size_t message_len,
          uint8_t *mac, size_t mac_size)
{
	hmac_rmd160_ctx_t ctx;

	hmac_rmd160_init(&ctx, key, key_size);
	hmac_rmd160_update(&ctx, message, message_len);
	hmac_rmd160_final(&ctx, mac, mac_size);
}

/*************************************************************************/

void
hmac_md5_init(hmac_md5_ctx_t *ctx, const uint8_t *keyarg, size_t key_size)
{
	const uint8_t		*key;
	uint8_t		 key_dig[(MD5_DIGEST_SIZE * 2) + 1];
	uint8_t		 key_temp[MD5_DIGEST_SIZE];
	size_t		 fillc;
	size_t		 keylen;
	size_t		 i;

	if (key_size > MD5_BLOCK_SIZE){
		key = key_temp;
		keylen = MD5_DIGEST_SIZE;
		MD5Data(keyarg, (unsigned)key_size, (char *)key_dig);
		hmac_str2bin(key_dig, sizeof(key_dig), key_temp, sizeof(key_temp));
	} else {
		/* key_size <= MD5_BLOCK_SIZE */
		key = keyarg;
		keylen = key_size;
	}
	fillc = MD5_BLOCK_SIZE - keylen;
	memset(ctx->block_ipad + keylen, 0x36, fillc);
	memset(ctx->block_opad + keylen, 0x5c, fillc);
	for (i = 0; i < keylen; i++) {
		ctx->block_ipad[i] = key[i] ^ 0x36;
		ctx->block_opad[i] = key[i] ^ 0x5c;
	}

	MD5Init(&ctx->ctx_inside);
	MD5Update(&ctx->ctx_inside, ctx->block_ipad, MD5_BLOCK_SIZE);

	MD5Init(&ctx->ctx_outside);
	MD5Update(&ctx->ctx_outside, ctx->block_opad,
		MD5_BLOCK_SIZE);

	/* for hmac_reinit */
	memcpy(&ctx->ctx_inside_reinit, &ctx->ctx_inside,
		sizeof(ctx->ctx_inside_reinit));
	memcpy(&ctx->ctx_outside_reinit, &ctx->ctx_outside,
		sizeof(ctx->ctx_outside_reinit));
}

void
hmac_md5_reinit(hmac_md5_ctx_t *ctx)
{
	memcpy(&ctx->ctx_inside, &ctx->ctx_inside_reinit,
		sizeof(ctx->ctx_inside));
	memcpy(&ctx->ctx_outside, &ctx->ctx_outside_reinit,
		sizeof(ctx->ctx_outside));
}

void
hmac_md5_update(hmac_md5_ctx_t *ctx, const uint8_t *message,
                        size_t message_len)
{
	MD5Update(&ctx->ctx_inside, message, (unsigned)message_len);
}

void
hmac_md5_final(hmac_md5_ctx_t *ctx, uint8_t *mac, size_t mac_size)
{
	uint8_t digest_inside[MD5_DIGEST_SIZE];
	uint8_t mac_temp[MD5_DIGEST_SIZE];

	MD5Final(digest_inside, &ctx->ctx_inside);
	MD5Update(&ctx->ctx_outside, digest_inside, MD5_DIGEST_SIZE);
	MD5Final(mac_temp, &ctx->ctx_outside);
	memcpy(mac, mac_temp, mac_size);
}

void
hmac_md5(const uint8_t *key, size_t key_size,
          const uint8_t *message, size_t message_len,
          uint8_t *mac, size_t mac_size)
{
	hmac_md5_ctx_t ctx;

	hmac_md5_init(&ctx, key, key_size);
	hmac_md5_update(&ctx, message, message_len);
	hmac_md5_final(&ctx, mac, mac_size);
}

/*************************************************************************/

void
hmac_sha1_init(hmac_sha1_ctx_t *ctx, const uint8_t *keyarg, size_t key_size)
{
	const uint8_t		*key;
	uint8_t		 key_dig[(SHA1_DIGEST_SIZE * 2) + 1];
	uint8_t		 key_temp[SHA1_DIGEST_SIZE];
	size_t		 fillc;
	size_t		 keylen;
	size_t		 i;

	if (key_size > SHA1_BLOCK_SIZE){
		key = key_temp;
		keylen = SHA1_DIGEST_SIZE;
		SHA1Data(keyarg, key_size, (char *)key_dig);
		hmac_str2bin(key_dig, sizeof(key_dig), key_temp, sizeof(key_temp));
	} else {
		/* key_size <= SHA1_BLOCK_SIZE */
		key = keyarg;
		keylen = key_size;
	}
	fillc = SHA1_BLOCK_SIZE - keylen;
	memset(ctx->block_ipad + keylen, 0x36, fillc);
	memset(ctx->block_opad + keylen, 0x5c, fillc);
	for (i = 0; i < keylen; i++) {
		ctx->block_ipad[i] = key[i] ^ 0x36;
		ctx->block_opad[i] = key[i] ^ 0x5c;
	}

	SHA1Init(&ctx->ctx_inside);
	SHA1Update(&ctx->ctx_inside, ctx->block_ipad, SHA1_BLOCK_SIZE);

	SHA1Init(&ctx->ctx_outside);
	SHA1Update(&ctx->ctx_outside, ctx->block_opad,
		SHA1_BLOCK_SIZE);

	/* for hmac_reinit */
	memcpy(&ctx->ctx_inside_reinit, &ctx->ctx_inside,
		sizeof(ctx->ctx_inside_reinit));
	memcpy(&ctx->ctx_outside_reinit, &ctx->ctx_outside,
		sizeof(ctx->ctx_outside_reinit));
}

void
hmac_sha1_reinit(hmac_sha1_ctx_t *ctx)
{
	memcpy(&ctx->ctx_inside, &ctx->ctx_inside_reinit,
		sizeof(ctx->ctx_inside));
	memcpy(&ctx->ctx_outside, &ctx->ctx_outside_reinit,
		sizeof(ctx->ctx_outside));
}

void
hmac_sha1_update(hmac_sha1_ctx_t *ctx, const uint8_t *message,
                        size_t message_len)
{
	SHA1Update(&ctx->ctx_inside, message, (unsigned)message_len);
}

void
hmac_sha1_final(hmac_sha1_ctx_t *ctx, uint8_t *mac,
                       size_t mac_size)
{
	uint8_t digest_inside[SHA1_DIGEST_SIZE];
	uint8_t mac_temp[SHA1_DIGEST_SIZE];

	SHA1Final(digest_inside, &ctx->ctx_inside);
	SHA1Update(&ctx->ctx_outside, digest_inside, SHA1_DIGEST_SIZE);
	SHA1Final(mac_temp, &ctx->ctx_outside);
	memcpy(mac, mac_temp, mac_size);
}

void
hmac_sha1(const uint8_t *key, size_t key_size,
          const uint8_t *message, size_t message_len,
          uint8_t *mac, size_t mac_size)
{
	hmac_sha1_ctx_t ctx;

	hmac_sha1_init(&ctx, key, key_size);
	hmac_sha1_update(&ctx, message, message_len);
	hmac_sha1_final(&ctx, mac, mac_size);
}

/****************************************************************************/

void
hmac_sha224_init(hmac_sha224_ctx_t *ctx, const uint8_t *keyarg, size_t key_size)
{
	const uint8_t		*key;
	uint8_t		 key_dig[(SHA224_DIGEST_SIZE * 2) + 1];
	uint8_t		 key_temp[SHA224_DIGEST_SIZE];
	size_t		 fillc;
	size_t		 keylen;
	size_t		 i;

	if (key_size > SHA224_BLOCK_SIZE){
		key = key_temp;
		keylen = SHA224_DIGEST_SIZE;
		SHA224_Data(keyarg, key_size, (char *)key_dig);
		hmac_str2bin(key_dig, sizeof(key_dig), key_temp, sizeof(key_temp));
	} else {
		/* key_size <= SHA224_BLOCK_SIZE */
		key = keyarg;
		keylen = (int)key_size;
	}
	fillc = SHA224_BLOCK_SIZE - keylen;
	memset(ctx->block_ipad + keylen, 0x36, fillc);
	memset(ctx->block_opad + keylen, 0x5c, fillc);
	for (i = 0; i < keylen; i++) {
		ctx->block_ipad[i] = key[i] ^ 0x36;
		ctx->block_opad[i] = key[i] ^ 0x5c;
	}

	SHA224_Init(&ctx->ctx_inside);
	SHA224_Update(&ctx->ctx_inside, ctx->block_ipad, SHA224_BLOCK_SIZE);

	SHA224_Init(&ctx->ctx_outside);
	SHA224_Update(&ctx->ctx_outside, ctx->block_opad,
		SHA224_BLOCK_SIZE);

	/* for hmac_reinit */
	memcpy(&ctx->ctx_inside_reinit, &ctx->ctx_inside,
		sizeof(ctx->ctx_inside_reinit));
	memcpy(&ctx->ctx_outside_reinit, &ctx->ctx_outside,
		sizeof(ctx->ctx_outside_reinit));
}

void
hmac_sha224_reinit(hmac_sha224_ctx_t *ctx)
{
	memcpy(&ctx->ctx_inside, &ctx->ctx_inside_reinit,
		sizeof(ctx->ctx_inside));
	memcpy(&ctx->ctx_outside, &ctx->ctx_outside_reinit,
		sizeof(ctx->ctx_outside));
}

void
hmac_sha224_update(hmac_sha224_ctx_t *ctx, const uint8_t *message,
                        size_t message_len)
{
	SHA224_Update(&ctx->ctx_inside, message, message_len);
}

void
hmac_sha224_final(hmac_sha224_ctx_t *ctx, uint8_t *mac,
                       size_t mac_size)
{
	uint8_t digest_inside[SHA224_DIGEST_SIZE];
	uint8_t mac_temp[SHA224_DIGEST_SIZE];

	SHA224_Final(digest_inside, &ctx->ctx_inside);
	SHA224_Update(&ctx->ctx_outside, digest_inside, SHA224_DIGEST_SIZE);
	SHA224_Final(mac_temp, &ctx->ctx_outside);
	memcpy(mac, mac_temp, mac_size);
}

void
hmac_sha224(const uint8_t *key, size_t key_size,
          const uint8_t *message, size_t message_len,
          uint8_t *mac, size_t mac_size)
{
	hmac_sha224_ctx_t ctx;

	hmac_sha224_init(&ctx, key, key_size);
	hmac_sha224_update(&ctx, message, message_len);
	hmac_sha224_final(&ctx, mac, mac_size);
}

/****************************************************************************/

void
hmac_sha256_init(hmac_sha256_ctx_t *ctx, const uint8_t *keyarg, size_t key_size)
{
	const uint8_t		*key;
	uint8_t		 key_dig[(SHA256_DIGEST_SIZE * 2) + 1];
	uint8_t		 key_temp[SHA256_DIGEST_SIZE];
	size_t		 fillc;
	size_t		 keylen;
	size_t		 i;

	if (key_size > SHA256_BLOCK_SIZE){
		key = key_temp;
		keylen = SHA256_DIGEST_SIZE;
		SHA256_Data(keyarg, key_size, (char *)key_dig);
		hmac_str2bin(key_dig, sizeof(key_dig), key_temp, sizeof(key_temp));
	} else {
		/* key_size <= SHA256_BLOCK_SIZE */
		key = keyarg;
		keylen = key_size;
	}
	fillc = SHA256_BLOCK_SIZE - keylen;

	memset(ctx->block_ipad + keylen, 0x36, fillc);
	memset(ctx->block_opad + keylen, 0x5c, fillc);

	for (i = 0; i < keylen; i++) {
		ctx->block_ipad[i] = key[i] ^ 0x36;
		ctx->block_opad[i] = key[i] ^ 0x5c;
	}

	SHA256_Init(&ctx->ctx_inside);
	SHA256_Update(&ctx->ctx_inside, ctx->block_ipad, SHA256_BLOCK_SIZE);

	SHA256_Init(&ctx->ctx_outside);
	SHA256_Update(&ctx->ctx_outside, ctx->block_opad,
		SHA256_BLOCK_SIZE);

	/* for hmac_reinit */
	memcpy(&ctx->ctx_inside_reinit, &ctx->ctx_inside,
		sizeof(ctx->ctx_inside_reinit));
	memcpy(&ctx->ctx_outside_reinit, &ctx->ctx_outside,
		sizeof(ctx->ctx_outside_reinit));
}

void
hmac_sha256_reinit(hmac_sha256_ctx_t *ctx)
{
	memcpy(&ctx->ctx_inside, &ctx->ctx_inside_reinit,
		sizeof(ctx->ctx_inside));
	memcpy(&ctx->ctx_outside, &ctx->ctx_outside_reinit,
		sizeof(ctx->ctx_outside));
}

void
hmac_sha256_update(hmac_sha256_ctx_t *ctx, const uint8_t *message,
                        size_t message_len)
{
	SHA256_Update(&ctx->ctx_inside, message, message_len);
}

void
hmac_sha256_final(hmac_sha256_ctx_t *ctx, uint8_t *mac, size_t mac_size)
{
	uint8_t digest_inside[SHA256_DIGEST_SIZE];
	uint8_t mac_temp[SHA256_DIGEST_SIZE];

	SHA256_Final(digest_inside, &ctx->ctx_inside);
	SHA256_Update(&ctx->ctx_outside, digest_inside, SHA256_DIGEST_SIZE);
	SHA256_Final(mac_temp, &ctx->ctx_outside);
	memcpy(mac, mac_temp, mac_size);
}

void
hmac_sha256(const uint8_t *key, size_t key_size,
          const uint8_t *message, size_t message_len,
          uint8_t *mac, size_t mac_size)
{
	hmac_sha256_ctx_t ctx;

	hmac_sha256_init(&ctx, key, key_size);
	hmac_sha256_update(&ctx, message, message_len);
	hmac_sha256_final(&ctx, mac, mac_size);
}

/****************************************************************************/

void
hmac_sha384_init(hmac_sha384_ctx_t *ctx, const uint8_t *keyarg, size_t key_size)
{
	const uint8_t		*key;
	uint8_t		 key_dig[(SHA384_DIGEST_SIZE * 2) + 1];
	uint8_t		 key_temp[SHA384_DIGEST_SIZE];
	size_t		 fillc;
	size_t		 keylen;
	size_t		 i;

	if (key_size > SHA384_BLOCK_SIZE){
		key = key_temp;
		keylen = SHA384_DIGEST_SIZE;
		SHA384_Data(keyarg, key_size, (char *)key_dig);
		hmac_str2bin(key_dig, sizeof(key_dig), key_temp, sizeof(key_temp));
	} else {
		/* key_size <= SHA384_BLOCK_SIZE */
		key = keyarg;
		keylen = key_size;
	}
	fillc = SHA384_BLOCK_SIZE - keylen;
	memset(ctx->block_ipad + keylen, 0x36, fillc);
	memset(ctx->block_opad + keylen, 0x5c, fillc);
	for (i = 0; i < keylen; i++) {
		ctx->block_ipad[i] = key[i] ^ 0x36;
		ctx->block_opad[i] = key[i] ^ 0x5c;
	}

	SHA384_Init(&ctx->ctx_inside);
	SHA384_Update(&ctx->ctx_inside, ctx->block_ipad, SHA384_BLOCK_SIZE);

	SHA384_Init(&ctx->ctx_outside);
	SHA384_Update(&ctx->ctx_outside, ctx->block_opad,
		SHA384_BLOCK_SIZE);

	/* for hmac_reinit */
	memcpy(&ctx->ctx_inside_reinit, &ctx->ctx_inside,
		sizeof(ctx->ctx_inside_reinit));
	memcpy(&ctx->ctx_outside_reinit, &ctx->ctx_outside,
		sizeof(ctx->ctx_outside_reinit));
}

void
hmac_sha384_reinit(hmac_sha384_ctx_t *ctx)
{
	memcpy(&ctx->ctx_inside, &ctx->ctx_inside_reinit,
		sizeof(ctx->ctx_inside));
	memcpy(&ctx->ctx_outside, &ctx->ctx_outside_reinit,
		sizeof(ctx->ctx_outside));
}

void
hmac_sha384_update(hmac_sha384_ctx_t *ctx, const uint8_t *message,
                        size_t message_len)
{
	SHA384_Update(&ctx->ctx_inside, message, message_len);
}

void
hmac_sha384_final(hmac_sha384_ctx_t *ctx, uint8_t *mac, size_t mac_size)
{
	uint8_t digest_inside[SHA384_DIGEST_SIZE];
	uint8_t mac_temp[SHA384_DIGEST_SIZE];

	SHA384_Final(digest_inside, &ctx->ctx_inside);
	SHA384_Update(&ctx->ctx_outside, digest_inside, SHA384_DIGEST_SIZE);
	SHA384_Final(mac_temp, &ctx->ctx_outside);
	memcpy(mac, mac_temp, mac_size);
}

void
hmac_sha384(const uint8_t *key, size_t key_size,
          const uint8_t *message, size_t message_len,
          uint8_t *mac, size_t mac_size)
{
	hmac_sha384_ctx_t ctx;

	hmac_sha384_init(&ctx, key, key_size);
	hmac_sha384_update(&ctx, message, message_len);
	hmac_sha384_final(&ctx, mac, mac_size);
}

/****************************************************************************/

void
hmac_sha512_init(hmac_sha512_ctx_t *ctx, const uint8_t *keyarg, size_t key_size)
{
	const uint8_t		*key;
	uint8_t		 key_dig[(SHA512_DIGEST_SIZE * 2) + 1];
	uint8_t		 key_temp[SHA512_DIGEST_SIZE];
	size_t		 fillc;
	size_t		 keylen;
	size_t		 i;

	if (key_size > SHA512_BLOCK_SIZE){
		key = key_temp;
		keylen = SHA512_DIGEST_SIZE;
		SHA512_Data(keyarg, key_size, (char *)key_dig);
		hmac_str2bin(key_dig, sizeof(key_dig), key_temp, sizeof(key_temp));
	} else {
		/* key_size <= SHA512_BLOCK_SIZE */
		key = keyarg;
		keylen = key_size;
	}
	fillc = SHA512_BLOCK_SIZE - keylen;
	memset(ctx->block_ipad + keylen, 0x36, fillc);
	memset(ctx->block_opad + keylen, 0x5c, fillc);
	for (i = 0; i < keylen; i++) {
		ctx->block_ipad[i] = key[i] ^ 0x36;
		ctx->block_opad[i] = key[i] ^ 0x5c;
	}

	SHA512_Init(&ctx->ctx_inside);
	SHA512_Update(&ctx->ctx_inside, ctx->block_ipad, SHA512_BLOCK_SIZE);

	SHA512_Init(&ctx->ctx_outside);
	SHA512_Update(&ctx->ctx_outside, ctx->block_opad,
		SHA512_BLOCK_SIZE);

	/* for hmac_reinit */
	memcpy(&ctx->ctx_inside_reinit, &ctx->ctx_inside,
		sizeof(ctx->ctx_inside_reinit));
	memcpy(&ctx->ctx_outside_reinit, &ctx->ctx_outside,
		sizeof(ctx->ctx_outside_reinit));
}

void
hmac_sha512_reinit(hmac_sha512_ctx_t *ctx)
{
	memcpy(&ctx->ctx_inside, &ctx->ctx_inside_reinit,
		sizeof(ctx->ctx_inside));
	memcpy(&ctx->ctx_outside, &ctx->ctx_outside_reinit,
		sizeof(ctx->ctx_outside));
}

void
hmac_sha512_update(hmac_sha512_ctx_t *ctx, const uint8_t *message,
                        size_t message_len)
{
	SHA512_Update(&ctx->ctx_inside, message, message_len);
}

void
hmac_sha512_final(hmac_sha512_ctx_t *ctx, uint8_t *mac, size_t mac_size)
{
	uint8_t digest_inside[SHA512_DIGEST_SIZE];
	uint8_t mac_temp[SHA512_DIGEST_SIZE];

	SHA512_Final(digest_inside, &ctx->ctx_inside);
	SHA512_Update(&ctx->ctx_outside, digest_inside, SHA512_DIGEST_SIZE);
	SHA512_Final(mac_temp, &ctx->ctx_outside);
	memcpy(mac, mac_temp, mac_size);
}

void
hmac_sha512(const uint8_t *key, size_t key_size,
          const uint8_t *message, size_t message_len,
          uint8_t *mac, size_t mac_size)
{
	hmac_sha512_ctx_t ctx;

	hmac_sha512_init(&ctx, key, key_size);
	hmac_sha512_update(&ctx, message, message_len);
	hmac_sha512_final(&ctx, mac, mac_size);
}
