/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "hmacwrap.h"

#include "hmac.h"

typedef struct hmac_t {
	const char	*alg;		/* digest algorithm name */
	const size_t	 macsize;	/* size of raw MAC digest */
	void		(*func)(const uint8_t *, size_t, const uint8_t *, size_t, uint8_t *, size_t);
} hmac_t;

static hmac_t 	hmacs[] = {
	{	"hmac-sha512",	SHA512_DIGEST_SIZE,	hmac_sha512	},
	{	"hmac-sha256",	SHA256_DIGEST_SIZE,	hmac_sha256	},
	{	"hmac-sha224",	SHA224_DIGEST_SIZE,	hmac_sha224	},
	{	"hmac-sha384",	SHA384_DIGEST_SIZE,	hmac_sha384	},
	{	"hmac-sha1",	SHA1_DIGEST_SIZE,	hmac_sha1	},
	{	"hmac-rmd160",	RMD160_DIGEST_SIZE,	hmac_rmd160	},
	{	"hmac-md5",	MD5_DIGEST_SIZE,	hmac_md5	},
	{	"sha512",	SHA512_DIGEST_SIZE,	hmac_sha512	},
	{	"sha256",	SHA256_DIGEST_SIZE,	hmac_sha256	},
	{	"sha224",	SHA224_DIGEST_SIZE,	hmac_sha224	},
	{	"sha384",	SHA384_DIGEST_SIZE,	hmac_sha384	},
	{	"sha1",		SHA1_DIGEST_SIZE,	hmac_sha1	},
	{	"rmd160",	RMD160_DIGEST_SIZE,	hmac_rmd160	},
	{	"md5",		MD5_DIGEST_SIZE,	hmac_md5	},
	{	NULL,		0,	NULL }
};

#ifndef USE_ARG
#define USE_ARG(x)	/*LINTED*/(void)&(x)
#endif

/* convert printable string to a binary digest */
size_t
hmac_str2bin(uint8_t *in, size_t insize, uint8_t *out, size_t osize)
{
	static const char	*hexes = "0123456789abcdef";
	const char		*cp1;
	const char		*cp2;
	size_t			 i;
	size_t			 o;

	for (i = 0, o = 0; i < insize && o < osize ; i += 2, o++) {
		if ((cp1 = strchr(hexes, in[i])) == NULL ||
		    (cp2 = strchr(hexes, in[i + 1])) == NULL) {
			return 0;
		}
		out[o] = (uint8_t)(((cp1 - hexes) << 4) | (cp2 - hexes));
	}
	return o;
}

/* convert to a printable string */
size_t
hmac_bin2str(uint8_t *digest, size_t digsize, char *str, size_t strsize)
{
	size_t	i;
	size_t	o;

	for (i = 0, o = 0 ; i < digsize && o < strsize ; i++, o += 2) {
		str[o] = "0123456789abcdef"[(uint8_t)((digest[i]) / 16) & 0xfU];
		str[o + 1] = "0123456789abcdef"[(digest[i]) & 0xf];
	}
	str[o] = 0x0;
	return o;
}

/* wrapper function around hmac: just need hash alg, key, msg and a place for the mac */
int
hmac(const char *alg, uint8_t *key, size_t keylen, const uint8_t *msg, size_t msglen, uint8_t *mac, size_t macsize)
{
	hmac_t	*hp;

	USE_ARG(macsize);
	if (alg == NULL) {
		(void) fprintf(stderr, "hmac: unknown algorithm: '%s'\n", alg);
		return 0;
	}
	for (hp = hmacs ; hp->alg ; hp++) {
		if (strcasecmp(hp->alg, alg) == 0) {
			(*hp->func)(key, keylen, msg, msglen, mac, hp->macsize);
			return (int)hp->macsize;
		}
	}
	(void) fprintf(stderr, "hmac: unknown algorithm: '%s'\n", alg);
	return 0;
}
