/*-
 * Copyright (c) 2012 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/param.h>

#include <err.h>
#include <fcntl.h>
#include <netdb.h>
#include <regex.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "http.h"

#ifndef USER_AGENT
#define USER_AGENT		"libhttp/20120208"
#endif

/* default port for HTTP as a string */
#define DEFAULT_HTTP_PORT	"80"

/* offset in buffer where the numeric response code is */
#define RESPONSE_OFFSET		9

/* HTTP 1.1 response codes */
#define	HTTP_SUCCESS		200
#define	HTTP_PARTIAL_CONTENT	206
#define	HTTP_PERM_REDIRECT	301
#define	HTTP_TEMP_REDIRECT	307
#define HTTP_UNAUTHORISED	403
#define HTTP_NOT_FOUND		404

enum {
	HEADER_LEN		= 4096
};

/* this describes an http request */
typedef struct url_t {
	char	*host;		/* host serving url */
	char	*port;		/* string for port number */
	char	*path;		/* base uri to retrieve */
	int	 server;	/* socket descriptor */
	char	*agent;		/* user-agent string we announce */
} url_t;

/* HTTP header */
typedef struct header_t {
	char	 buf[HEADER_LEN];	/* buffer to hold header */
	char	*body;		/* pointer to start of body in header */
	int	 httpresp;	/* HTTP response code */
	size_t	 cc;		/* # of chars we've read so far */
	size_t	 contlength;	/* total size we will read */
	int	 chunked;	/* Transfer-Encoding: chunked */
} header_t;

#ifndef USE_ARG
#define USE_ARG(x)         /*LINTED*/(void)&(x)
#endif

/* open a socket connection to the given server */
static int
connect_to_server(const char *host, const char *port)
{
	struct addrinfo	 hints;
	struct addrinfo	*res;
	struct addrinfo	*res0;
	const char	*cause = NULL;
	int		 server;
	int		 error;

	(void) memset(&hints, 0x0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_NUMERICSERV;
	if ((error = getaddrinfo(host, (port) ? port : DEFAULT_HTTP_PORT, &hints, &res0)) != 0) {
		warnx("%s", gai_strerror(error));
		return -1;
	}
	for (server = -1, res = res0; res && server < 0; res = res->ai_next) {
		server = socket(res->ai_family, res->ai_socktype,
		    res->ai_protocol);
		if (server < 0) {
			cause = "socket";
			continue;
		}
		if (connect(server, res->ai_addr, res->ai_addrlen) < 0) {
			cause = "connect";
			close(server);
			server = -1;
			continue;
		}
	}
	freeaddrinfo(res0);
	if (server < 0) {
		warn("%s", cause);
	}
	return server;
}

/* read the header returned by the server */
static int
readheader(url_t *url, header_t *h)
{
	static regex_t	 r;
	static int	 compiled;
	regmatch_t	 match[10];
	ssize_t		 rc;

	if (!compiled) {
		compiled = 1;
		(void) regcomp(&r, "(Content-Length: ([0-9]+)|Transfer-Encoding: chunked)\r\n", REG_EXTENDED);
	}
	for (h->cc = 0 ; (rc = read(url->server, &h->buf[h->cc], sizeof(h->buf) - h->cc)) > 0 ; h->cc += rc) {
		if ((h->body = strstr(h->buf, "\r\n\r\n")) != NULL) {
			h->body += 4; /* jump over end of header (\r\n\r\n) */
			h->cc += rc;
			break;
		}
	}
	switch(h->httpresp = (int)strtol(&h->buf[RESPONSE_OFFSET], NULL, 10)) {
	case HTTP_SUCCESS:
	case HTTP_PARTIAL_CONTENT:
		break;
	case HTTP_PERM_REDIRECT:
	case HTTP_TEMP_REDIRECT:
		warnx("redirection");
		return 0;
	case HTTP_UNAUTHORISED:
	case HTTP_NOT_FOUND:
		warnx("%s: No such file or directory", url->path);
		return 0;
	default:
		warnx("weird response code '%s'", &h->buf[RESPONSE_OFFSET]);
		return 0;
	}
	if (regexec(&r, h->buf, 10, match, 0) != 0) {
		warn("no content-length in header");
		return 0;
	}
	if (match[2].rm_so >= 0) {
		h->contlength = (size_t)strtoull(&h->buf[match[2].rm_so], NULL, 10);
	} else {
		h->chunked = 1;
	}
	return 1;
}

/* receive a buffer from the server */
static ssize_t
recv_buffer(url_t *url, header_t *head, char *output)
{
	unsigned	 i;
	ssize_t		 rc;

	head->cc -= (size_t)(head->body - head->buf);
	(void) memcpy(output, head->body, head->cc);
	for (i = 0 ; head->cc < head->contlength ; i++) {
		if ((rc = read(url->server, &output[head->cc], head->contlength - head->cc)) <= 0) {
			break;
		}
		head->cc += rc;
	}
	return (ssize_t)head->cc;
}

/* parse a url into constituent part */
static int
parse_url(url_t *url, const char *s)
{
	static regex_t	r;
	regmatch_t	match[10];
	static int	compiled;

	if (!compiled) {
		compiled = 1;
		regcomp(&r, "(http://)?([^:/]+)(:([0-9]+))?(.*)", REG_EXTENDED);
	}
	if (regexec(&r, s, 10, match, 0) != 0) {
		return 0;
	}
	asprintf(&url->host, "%.*s", (int)(match[2].rm_eo - match[2].rm_so), &s[match[2].rm_so]);
	if (match[4].rm_so >= 0) {
		asprintf(&url->port, "%.*s", (int)(match[4].rm_eo - match[4].rm_so), &s[match[4].rm_so]);
	}
	asprintf(&url->path, "%s", &s[match[5].rm_so]);
	return 1;
}

/* send request to the server */
static int
send_request(url_t *url, const char *cmd, int64_t from, int64_t to, const char *agent)
{
	ssize_t		 cc;
	char		 request[4096];
	char		 range[128];

	if ((url->server = connect_to_server(url->host, url->port)) < 0) {
		return 0;
	}
	range[0] = 0x0;
	if (from >= 0 && to >= 0) {
		(void) snprintf(range, sizeof(range), "Range: bytes=%" PRId64 "-%" PRId64 "\r\n",
			from, to);
	}
	cc = snprintf(request, sizeof(request),
		"%s %s HTTP/1.1\r\nHost: %s:%s\r\nUser-Agent: %s\r\n%s\r\n",
		cmd, url->path, url->host, url->port,
		(agent) ? agent : USER_AGENT, range);
	return (write(url->server, request, (size_t)cc) == cc);
}

/* close connection to server and free resources */
static int
close_connection(url_t *url)
{
	if (url->host) {
		free(url->host);
	}
	if (url->port) {
		free(url->port);
	}
	if (url->path) {
		free(url->path);
	}
	close(url->server);
	return 1;
}

/* put/post method */
static ssize_t
post(const char *method, char *buf, size_t size, const char *uri, int64_t from, int64_t to)
{
	ssize_t		wc;
	ssize_t		cc;
	size_t		n;
	url_t		url;

	(void) memset(&url, 0x0, sizeof(url));
	cc = -1;
	if (uri == NULL || buf == NULL) {
		return -1;
	}
	if (parse_url(&url, uri) && send_request(&url, method, from, to, USER_AGENT)) {
		for (cc = 0 ; cc < (ssize_t)size ; cc += wc) {
			n = MAX(BUFSIZ, size - cc);
			if ((wc = write(url.server, &buf[cc], n)) != (ssize_t)n) {
				goto done;
			}
		}
	}
done:
	close_connection(&url);
	return cc;
}

/**************************************************************************/

/* HTTP get into a buffer */
ssize_t
http_get(char *output, size_t size, const char *uri, int64_t from, int64_t to)
{
	header_t	h;
	ssize_t		ret;
	url_t		url;

	ret = -1;
	(void) memset(&url, 0x0, sizeof(url));
	if (uri == NULL || output == NULL) {
		return -1;
	}
	if (parse_url(&url, uri) && send_request(&url, "GET", from, to, USER_AGENT)) {
		(void) memset(&h, 0x0, sizeof(h));
		if (!readheader(&url, &h)) {
			return -1;
		}
		if (h.contlength >= size) {
			h.contlength = size;
		}
		ret = recv_buffer(&url, &h, output);
	}
	close_connection(&url);
	return ret;
}

ssize_t
http_head(char *output, size_t size, const char *uri)
{
	header_t	h;
	ssize_t		ret;
	url_t		url;

	ret = -1;
	(void) memset(&url, 0x0, sizeof(url));
	if (uri == NULL || output == NULL) {
		return -1;
	}
	if (parse_url(&url, uri) && send_request(&url, "HEAD", -1, -1, USER_AGENT)) {
		(void) memset(&h, 0x0, sizeof(h));
		if (!readheader(&url, &h)) {
			return -1;
		}
		if ((ret = (ssize_t)(h.body - h.buf)) > (ssize_t)size - 1) {
			ret = size - 1;
		}
		(void) memcpy(output, h.buf, (size_t)ret);
		h.buf[ret] = 0x0;
	}
	close_connection(&url);
	return ret;
}

/* HTTP PUT from a buffer */
ssize_t
http_put(char *buf, size_t size, const char *uri, int64_t from, int64_t to)
{
	return post("PUT", buf, size, uri, from, to);
}

/* HTTP POST from a buffer */
ssize_t
http_post(char *buf, size_t size, const char *uri, int64_t from, int64_t to)
{
	return post("POST", buf, size, uri, from, to);
}
