/*-
 * Copyright (c) 2012 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/socket.h>

#include <arpa/inet.h>

#define FUSE_USE_VERSION	26

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <fuse.h>
#include <netdb.h>
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "http.h"
#include "virtdir.h"

#if defined(__NetBSD__) && defined(USE_LIBKMOD)
#include "libkmod.h"
#endif

static virtdir_t	 httpdev;		/* virtual dir for httpdev devices */

/* this struct keeps information on the target */
typedef struct targetinfo_t {
	const char	*uri;		/* uri of the target */
	char		*host;		/* resolvable host name */
	char		*address;	/* textual network address */
	char		*name;		/* name of target prog */
	char		*service;	/* name of target port */
	uint64_t	 size;		/* size of target */
	uint64_t	 lbac;		/* number of LBAs */
	uint32_t	 blocksize;	/* size of device blocks */
	const char	*vendor;	/* device vendor information */
	const char	*product;	/* device product information */
	char		*version;	/* device version information */
	char		*pid;		/* pid of httpdev process */
} targetinfo_t;

/* small target info, embedded into the 'storage' file in the target dir */
typedef struct smalltgt_t {
	struct stat      st;		/* normal stat info */
	const char	*uri;		/* uri of target */
} smalltgt_t;

#ifndef __UNCONST
#define __UNCONST(x)	(x)
#endif

#ifndef KB
#define KB(x)	((x) * 1024)
#endif

#ifndef DEFAULT_BLOCKSIZE
#define DEFAULT_BLOCKSIZE	KB(4)
#endif

/* read the capacity (maximum LBA and blocksize) from the target */
static int 
read_capacity(const char *uri, uint64_t *size, uint64_t *maxlba, uint32_t *blocklen)
{
	char	 buf[1024];
	char	*s;

	/* discovery - find size (via HEAD) of remote file */
	http_head(buf, sizeof(buf), uri);
	if ((s = strstr(buf, "Content-Length:")) == NULL) {
		return 0;
	}
	if (*blocklen == 0) {
		*blocklen = DEFAULT_BLOCKSIZE;
	}
	*size = (int64_t)strtoll(s + 16, NULL, 10);
	*maxlba = *size / *blocklen;
	return 1;
}

/* send inquiry command to the target, to get it to identify itself */
static int 
inquiry(const char *uri, targetinfo_t *info)
{
	char	 buf[1024];
	char	*eol;
	char	*server;

	/* discovery - find size (via HEAD) of remote file */
	http_head(buf, sizeof(buf), uri);
	if ((server = strstr(buf, "Server: ")) == NULL) {
		return 0;
	}
	server += 8;
	eol = strchr(server, '/');
	*eol = '-';
	eol = strchr(eol, '\r');
	asprintf(&info->version, "%.*s", (int)(eol - server), server);
	return 1;
}

/* read info from the target - method depends on size of data being read */
static ssize_t 
targetop(const char *uri, uint64_t offset, uint32_t length, char *buf, const int writing)
{
	if (writing) {
		return http_post(buf, length, uri, offset, offset + length);
	} else {
		return http_get(buf, length, uri, offset, offset + length);
	}
}


/****************************************************************************/

/* perform the stat operation */
/* if this is the root, then just synthesise the data */
/* otherwise, retrieve the data, and be sure to fill in the size */
static int 
httpdev_getattr(const char *path, struct stat *st)
{
	virt_dirent_t	*ep;
	smalltgt_t	*p;

	if (strcmp(path, "/") == 0) {
		(void) memset(st, 0x0, sizeof(*st));
		st->st_mode = S_IFDIR | 0755;
		st->st_nlink = 2;
		return 0;
	}
	if ((ep = virtdir_find(&httpdev, path, strlen(path))) == NULL) {
		return -ENOENT;
	}
	switch(ep->type) {
	case 'b':
		(void) memcpy(st, &httpdev.file, sizeof(*st));
		st->st_mode = (S_IFBLK | 0644);
		break;
	case 'c':
		(void) memcpy(st, &httpdev.file, sizeof(*st));
		st->st_mode = (S_IFCHR | 0644);
		break;
	case 'd':
		(void) memcpy(st, &httpdev.dir, sizeof(*st));
		break;
	case 'f':
		(void) memcpy(st, &httpdev.file, sizeof(*st));
		p = (smalltgt_t *) ep->tgt;
		st->st_size = p->st.st_size;
		break;
	case 'l':
		(void) memcpy(st, &httpdev.lnk, sizeof(*st));
		st->st_size = ep->tgtlen;
		break;
	default:
		warn("unknown directory type `%c'", ep->type);
		return -ENOENT;
	}
	st->st_ino = ep->ino;
	return 0;
}

/* readdir operation */
static int 
httpdev_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
	      off_t offset, struct fuse_file_info * fi)
{
	virt_dirent_t	*dp;
	VIRTDIR		*dirp;

	if ((dirp = openvirtdir(&httpdev, path)) == NULL) {
		return 0;
	}
	filler(buf, ".", NULL, 0);
	filler(buf, "..", NULL, 0);
	while ((dp = readvirtdir(dirp)) != NULL) {
		filler(buf, dp->d_name, NULL, 0);
	}
	closevirtdir(dirp);
	return 0;
}

/* open the file in the file system */
static int 
httpdev_open(const char *path, struct fuse_file_info *fi)
{
	virt_dirent_t	*ep;
	const char	*slash;

	if ((ep = virtdir_find(&httpdev, path, strlen(path))) == NULL) {
		return -ENOENT;
	}
	/* check path is the correct one */
	if ((slash = strrchr(path, '/')) == NULL) {
		slash = path;
	} else {
		slash += 1;
	}
	if (strcmp(slash, "storage") != 0) {
		return -ENOENT;
	}
	return 0;
}

/* read the storage from the iSCSI target */
static int 
httpdev_read(const char *path, char *buf, size_t size, off_t offset,
	   struct fuse_file_info * fi)
{
	virt_dirent_t	*ep;
	smalltgt_t	*smalltgt;
	const int	 reading = 0;
	ssize_t		 rc;

	if ((ep = virtdir_find(&httpdev, path, strlen(path))) == NULL) {
		return -ENOENT;
	}
	smalltgt = (smalltgt_t *)ep->tgt;
	if ((rc = targetop(smalltgt->uri, offset, size, buf, reading)) < 0) {
		return -EPERM;
	}
	return rc;
}

/* write the file's contents to the file system */
static int 
httpdev_write(const char *path, const char *buf, size_t size, off_t offset,
	   struct fuse_file_info * fi)
{
	virt_dirent_t	*ep;
        smalltgt_t	*smalltgt;
	const int	 writing = 1;
	ssize_t		 wc;

	if ((ep = virtdir_find(&httpdev, path, strlen(path))) == NULL) {
		return -ENOENT;
	}
	smalltgt = (smalltgt_t *)ep->tgt;
	if ((wc = targetop(smalltgt->uri, offset, size, __UNCONST(buf), writing)) < 0) {
		return -EPERM;
	}
	return wc;
}

/* fill in the statvfs struct */
static int
httpdev_statfs(const char *path, struct statvfs *st)
{
	(void) memset(st, 0x0, sizeof(*st));
	return 0;
}

/* read the symbolic link */
static int
httpdev_readlink(const char *path, char *buf, size_t size)
{
	virt_dirent_t	*ep;

	if ((ep = virtdir_find(&httpdev, path, strlen(path))) == NULL) {
		return -ENOENT;
	}
	if (ep->tgt == NULL) {
		return -ENOENT;
	}
	(void) strlcpy(buf, ep->tgt, size);
	return 0;
}

/* operations struct */
static struct fuse_operations httpdevops = {
	.getattr = httpdev_getattr,
	.readlink = httpdev_readlink,
	.readdir = httpdev_readdir,
	.open = httpdev_open,
	.read = httpdev_read,
	.write = httpdev_write,
	.statfs = httpdev_statfs
};

/* try to get the presentation mode for the remote address */
static int
getaddress(const char *host, const char *service, char *addr, size_t size)
{
	struct sockaddr_in6	*ipv6;
	struct sockaddr_in	*ipv4;
	struct addrinfo		 hints;
	struct addrinfo		*res;
	void			*vp;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	if (getaddrinfo(host, service, &hints, &res) != 0) {
		return 0;
	}
	switch(res->ai_family) {
	case AF_INET:
		ipv4 = (struct sockaddr_in *)res->ai_addr;
		vp = &ipv4->sin_addr;
		break;
	case AF_INET6:
		ipv6 = (struct sockaddr_in6 *)res->ai_addr;
		vp = &ipv6->sin6_addr;
		break;
	default:
		return 0;
	}
	inet_ntop(res->ai_family, vp, addr, size);
	freeaddrinfo(res);
	return 1;
}

/* fill in the targetinfo struct */
static int
gettargetinfo(const char *uri, targetinfo_t *target, const char *targetname)
{
	regmatch_t	 match[10];
	regex_t		 r;
	char	         targethost[256];
	char	         service[256];
	char	         targetpath[1024];
	char		 targetip[256];

	target->uri = strdup(uri);
	(void)regcomp(&r, "(http://)?([^:/]+)(:([0-9]+))?(.*)", REG_EXTENDED);
	if (regexec(&r, uri, 10, match, 0) != 0) {
		warnx("bad uri '%s'", uri);
		return 0;
	}
	(void) snprintf(targethost, sizeof(targethost), "%.*s",
		(int)(match[2].rm_eo - match[2].rm_so), &uri[match[2].rm_so]);
	if (match[4].rm_so >= 0) {
		(void) snprintf(service, sizeof(service), "%.*s",
			(int)(match[4].rm_eo - match[4].rm_so), &uri[match[4].rm_so]);
	} else if (match[1].rm_so >= 0) {
		(void) snprintf(service, sizeof(service), "%.*s",
			(int)(match[1].rm_eo - match[1].rm_so - 3), &uri[match[1].rm_so]);
	} else {
		(void) snprintf(service, sizeof(service), "http");
	}
	(void) snprintf(targetpath, sizeof(targetpath), "%.*s",
		(int)(match[5].rm_eo - match[5].rm_so), &uri[match[5].rm_so]);
	if (targetname == NULL) {
		targetname = targethost;
	}
	target->name = strdup(targetname);
	target->host = strdup(targethost);
	target->service = strdup(service);
	target->vendor = "NetBSD";
	target->product = "NetBSD httpdev";
	getaddress(targethost, service, targetip, sizeof(targetip));
	target->address = strdup(targetip);
	asprintf(&target->pid, "%d", (int)getpid());
	return 1;
}

/* set up a single virtdir entry */
static void
setup_entry(const char *name, const char *entname, char type, const void *ent, ssize_t entsize)
{
	ssize_t	cc;
	char	f[1024];

	cc = snprintf(f, sizeof(f), "/%s/%s", name, entname);
	if (entsize < 0) {
		entsize = strlen(ent);
	}
	virtdir_add(&httpdev, f, cc, type, ent, entsize);
}

/* set up the virtual directory tree */
static void
setup_virtdir(targetinfo_t *target, smalltgt_t *smalltgt)
{
	ssize_t	cc;
	char	buf[1024];

	cc = snprintf(buf, sizeof(buf), "/%s", target->name);
	virtdir_add(&httpdev, buf, cc, 'd', buf, cc);
	setup_entry(target->name, "address", 'l', target->address, -1);
	cc = snprintf(buf, sizeof(buf), "%u", target->blocksize);
	setup_entry(target->name, "blocksize", 'l', buf, cc);
	setup_entry(target->name, "hostname", 'l', target->host, -1);
	setup_entry(target->name, "pid", 'l', target->pid, -1);
	setup_entry(target->name, "product", 'l', target->product, -1);
	setup_entry(target->name, "service", 'l', target->service, -1);
	setup_entry(target->name, "storage", 'f', smalltgt, sizeof(*smalltgt));
	setup_entry(target->name, "targetname", 'l', target->name, -1);
	setup_entry(target->name, "vendor", 'l', target->vendor, -1);
	setup_entry(target->name, "version", 'l', target->version, -1);
}

int 
main(int argc, char **argv)
{
	targetinfo_t	 target;
	const char	*targetname;
	smalltgt_t	 smalltgt;
	int		 retry;
	int		 i;

	(void) memset(&target, 0x0, sizeof(target));
	(void) memset(&smalltgt, 0x0, sizeof(smalltgt));
	(void) stat("/etc/hosts", &smalltgt.st);
	targetname = NULL;
	while ((i = getopt(argc, argv, "b:n:")) != -1) {
		switch(i) {
		case 'b':
			target.blocksize = (uint32_t)strtol(optarg, NULL, 0);
			break;
		case 'n':
			targetname = optarg;
			break;
		default:
			break;
		}
	}
	smalltgt.st.st_ino = 0x15c51;
#if defined(__NetBSD__) && defined(USE_LIBKMOD)
	/* check that the puffs module is loaded on NetBSD */
	if (kmodstat("puffs", NULL) == 0 && !kmodload("puffs")) {
		(void) fprintf(stderr, "initiator: can't load puffs module\n");
	}
#endif
	/* get size from the target */
	for (retry = 5 ; retry > 0 ; retry--) {
		if (read_capacity(argv[optind], &target.size, &target.lbac, &target.blocksize)) {
			break;
		}
		warn("read_capacity failed - retrying %d\n", retry);
		sleep(1);
	}
	if (retry == 0) {
		errx(EXIT_FAILURE, "read_capacity failed - can't continue");
	}
	smalltgt.st.st_size = (off_t)target.size;

	/* fill in target information */
	gettargetinfo(smalltgt.uri = argv[optind], &target, targetname);

	/* get iSCSI target information */
	inquiry(argv[optind], &target);

	/* create the tree using virtdir routines */
	setup_virtdir(&target, &smalltgt);

	return fuse_main(argc - optind - 1, argv + optind + 1, &httpdevops, NULL);
}
