/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <sha2.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "lamport.h"

/*
see:
	http://en.wikipedia.org/wiki/Lamport_signature

256 pairs of hashes of 32 bytes each
*/

#ifndef KiB
#define KiB(x)	((x) * 1024)
#endif

#ifndef USE_ARG
#define USE_ARG(x)    /*LINTED*/(void)&(x)
#endif

#define PRIsize "z"

/* compare two binary arrays */
static int
cmp(uint8_t *a, uint8_t *b, size_t size)
{
	size_t	i;
	int	ok;

	for (ok = 1, i = 0 ; i < size ; i++) {
		if (a[i] != b[i]) {
			printf("[%zu] %02x != %02x\n", i, a[i], b[i]);
			ok = 0;
		}
	}
	return ok;
}

/************************************************************************/

/* initialise the random number generator from the microsecs time */
int
lamport_init(void)
{
	struct timeval	tv;
	unsigned long	l;

	if (gettimeofday(&tv, NULL) < 0) {
		return 0;
	}
	l = (tv.tv_sec + tv.tv_usec);
	arc4random_addrandom((unsigned char *)(void *)&l, (int)sizeof(l));
	return 1;
}

/* generate a secret key */
void
lamport_mk_sec(lamport_key_t *sec)
{
	uint32_t	r;
	int		h;
	int		d;
	int		p;

	for (h = 0 ; h < LAMPORT_MAX_PAIRS ; h++) {
		for (p = 0 ; p < 2 ; p++) {
			for (d = 0 ; d < LAMPORT_DIGEST_SIZE ; d += sizeof(r)) {
				r = arc4random();
				(void) memcpy(&sec->data[LAMPORT_OFFSET(h, p, d)], &r, sizeof(r));
			}
		}
	}
}

/* generate a semi-public key from the secret key */
void
lamport_mk_semipub(lamport_key_t *sec, /* uint8_t *key, size_t keysize, */ lamport_key_t *pub)
{
	SHA256_CTX	sha256;
	int		h;
	int		p;

	for (h = 0 ; h < LAMPORT_MAX_PAIRS ; h++) {
		for (p = 0 ; p < 2 ; p++) {
			SHA256_Init(&sha256);
			SHA256_Update(&sha256, &sec->data[LAMPORT_OFFSET(h, p, 0)], LAMPORT_DIGEST_SIZE);
			SHA256_Final(&pub->data[LAMPORT_OFFSET(h, p, 0)], &sha256);
		}
	}
}

/* generate a public key from the secret key */
void
lamport_mk_pub(lamport_key_t *sec, lamport_key_t *pub)
{
	SHA256_CTX	sha256;
	int		h;
	int		p;

	for (h = 0 ; h < LAMPORT_MAX_PAIRS ; h++) {
		for (p = 0 ; p < 2 ; p++) {
			SHA256_Init(&sha256);
			SHA256_Update(&sha256, &sec->data[LAMPORT_OFFSET(h, p, 0)], LAMPORT_DIGEST_SIZE);
			SHA256_Final(&pub->data[LAMPORT_OFFSET(h, p, 0)], &sha256);
		}
	}
}

/* generate a signature on data */
int
lamport_sign_data(lamport_key_t *sec, const void *vmsg, size_t size, uint8_t *sig, size_t sigsize)
{
	const uint8_t	*msg = (const uint8_t *)vmsg;
	SHA256_CTX	 sha256;
	unsigned	 bit;
	unsigned	 h;
	uint8_t		 digest[LAMPORT_DIGEST_SIZE];

	USE_ARG(sigsize);
	SHA256_Init(&sha256);
	SHA256_Update(&sha256, msg, size);
	SHA256_Final(digest, &sha256);
	for (h = 0 ; h < LAMPORT_MAX_PAIRS ; h++) {
		for (bit = 0 ; bit < 8 ; bit++) {
			(void) memcpy(&sig[h * LAMPORT_DIGEST_SIZE],
					&sec->data[LAMPORT_OFFSET(h, ((unsigned)(digest[h / 8]) >> bit) & 1, 0)],
					LAMPORT_DIGEST_SIZE);
		}
	}
	return 1;
}

/* verify a signature on data */
int
lamport_verify_data(lamport_key_t *pub, const void *vmsg, size_t size, uint8_t *sig, size_t sigsize)
{
	const uint8_t	*msg = (const uint8_t *)vmsg;
	SHA256_CTX	 sha256;
	unsigned	 bit;
	unsigned	 h;
	uint8_t		 hashsig[LAMPORT_MAX_PAIRS * LAMPORT_DIGEST_SIZE];
	uint8_t		 calcsig[LAMPORT_MAX_PAIRS * LAMPORT_DIGEST_SIZE];
	uint8_t		 digest[LAMPORT_DIGEST_SIZE];

	/* get a digest from the message */
	SHA256_Init(&sha256);
	SHA256_Update(&sha256, msg, size);
	SHA256_Final(digest, &sha256);
	for (h = 0 ; h < LAMPORT_MAX_PAIRS ; h++) {
		for (bit = 0 ; bit < 8 ; bit++) {
			(void) memcpy(&calcsig[h * LAMPORT_DIGEST_SIZE],
					&pub->data[LAMPORT_OFFSET(h, ((unsigned)(digest[h / 8]) >> bit) & 1, 0)],
					LAMPORT_DIGEST_SIZE);
		}
	}
	for (h = 0 ; h < LAMPORT_MAX_PAIRS ; h++) {
		SHA256_Init(&sha256);
		SHA256_Update(&sha256, &sig[h * LAMPORT_DIGEST_SIZE], LAMPORT_DIGEST_SIZE);
		SHA256_Final(&hashsig[h * LAMPORT_DIGEST_SIZE], &sha256);
	}
	if (sigsize != LAMPORT_MAX_PAIRS * LAMPORT_DIGEST_SIZE) {
		(void) fprintf(stderr, "size mismatch when validating\n");
		return 0;
	}
	cmp(hashsig, calcsig, LAMPORT_MAX_PAIRS * LAMPORT_DIGEST_SIZE);
	return (memcmp(hashsig, calcsig, LAMPORT_MAX_PAIRS * LAMPORT_DIGEST_SIZE) == 0);
}

/* sign a file */
int
lamport_sign_file(lamport_key_t *sec, const char *f, uint8_t *sig, size_t sigsize)
{
	struct stat	 st;
	size_t		 size;
	FILE		*fp;
	char		*mapped;
	int		 ret;

	if ((fp = fopen(f, "r")) == NULL) {
		(void) fprintf(stderr, "can't read file '%s'\n", f);
		return 0;
	}
	(void) fstat(fileno(fp), &st);
	size = (size_t)st.st_size;
	mapped = mmap(NULL, size, PROT_READ, MAP_FILE, fileno(fp), 0);
	ret = lamport_sign_data(sec, mapped, size, sig, sigsize);
	(void) munmap(mapped, size);
	(void) fclose(fp);
	return ret;
}

/* sign a file */
int
lamport_verify_file(lamport_key_t *pub, const char *f, uint8_t *sig, size_t sigsize)
{
	struct stat	 st;
	size_t	 	 size;
	FILE		*fp;
	char		*mapped;
	int		 ret;

	if ((fp = fopen(f, "r")) == NULL) {
		(void) fprintf(stderr, "can't read file '%s'\n", f);
		return 0;
	}
	(void) fstat(fileno(fp), &st);
	size = (size_t)st.st_size;
	mapped = mmap(NULL, size, PROT_READ, MAP_FILE, fileno(fp), 0);
	ret = lamport_verify_data(pub, mapped, size, sig, sigsize);
	(void) munmap(mapped, size);
	(void) fclose(fp);
	return ret;
}

/************************************************************************/
