/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>

#include <sha2.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "lamport.h"

#ifndef KiB
#define KiB(x)	((x) * 1024)
#endif

#ifndef LSIG_PATH_MAX
#define LSIG_PATH_MAX	1024
#endif

/* save a key in a file */
static int
savedata(uint8_t *data, size_t size, const char *fmt, ...)
{
	ssize_t	 wc;
	va_list	 args;
	size_t	 cc;
	FILE	*fp;
	char	 path[LSIG_PATH_MAX];

	va_start(args, fmt);
	(void) vsnprintf(path, sizeof(path), fmt, args);
	va_end(args);
	if ((fp = fopen(path, "w")) == NULL) {
		(void) fprintf(stderr, "can't open '%s'\n", path);
		return 0;
	}
	for (cc = 0 ; cc < size ; cc += wc) {
		if ((wc = write(fileno(fp), &data[cc], size - cc)) < 0) {
			break;
		}
	}
	(void) fprintf(fp, "\n");
	(void) fclose(fp);
	return cc == size;
}

/* save a key in a file */
static int
loaddata(uint8_t *data, size_t size, const char *fmt, ...)
{
	va_list	 args;
	ssize_t	 rc;
	size_t	 cc;
	FILE	*fp;
	char	 path[LSIG_PATH_MAX];

	va_start(args, fmt);
	(void) vsnprintf(path, sizeof(path), fmt, args);
	va_end(args);
	if ((fp = fopen(path, "r")) == NULL) {
		(void) fprintf(stderr, "can't open '%s'\n", path);
		return 0;
	}
	for (cc = 0 ; cc < size ; cc += rc) {
		if ((rc = read(fileno(fp), &data[cc], size - cc)) < 0) {
			break;
		}
	}
	(void) fclose(fp);
	return cc = size;
}

int
main(int argc, char **argv)
{
	lamport_key_t	 sec;
	lamport_key_t	 pub;
	uint8_t		 sig[KiB(8)];
	char		 action;
	int		 ok;
	int		 i;

	action = 0x0;
	while ((i = getopt(argc, argv, "csv")) != -1) {
		switch(i) {
		case 'c':
		case 's':
		case 'v':
			action = (char)i;
			break;
		default:
			break;
		}
	}
	lamport_init();
	ok = 1;
	switch(action) {
	case 'c':
		/* create a key */
		(void) memset(&sec, 0x0, sizeof(sec));
		(void) memset(&pub, 0x0, sizeof(pub));
		lamport_mk_sec(&sec);
		savedata(sec.data, sizeof(sec.data), "%s", "secring.lamport");
		lamport_mk_pub(&sec, &pub);
		savedata(pub.data, sizeof(sec.data), "%s", "pubring.lamport");
		break;
	case 's':
		/* sign a message */
		(void) memset(&sec, 0x0, sizeof(sec));
		loaddata(sec.data, sizeof(sec.data), "%s", "secring.lamport");
		for (i = optind ; i < argc ; i++) {
			if (!lamport_sign_file(&sec, argv[i], sig, sizeof(sig))) {
				ok = 0;
			}
			savedata(sig, sizeof(sig), "%s.lsig", argv[i]);
		}
		break;
	case 'v':
		/* verify a message */
		(void) memset(&pub, 0x0, sizeof(pub));
		loaddata(pub.data, sizeof(pub.data), "%s", "pubring.lamport");
		for (i = optind ; i < argc ; i++) {
			loaddata(sig, sizeof(sig), "%s.lsig", argv[i]);
			if (!lamport_verify_file(&pub, argv[i], sig, sizeof(sig))) {
				ok = 0;
			}
		}
		break;
	default:
		break;
	}
	exit((ok) ? EXIT_SUCCESS : EXIT_FAILURE);
}
