/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>

#include <regex.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "getopt2.h"

#include "mat.h"

#ifndef TMPDIR
#define TMPDIR	"/tmp"
#endif

/* read input from stdin into a temp file */
static int
dostdin(const char *tmpdir, char *temp, size_t size)
{
	ssize_t	cc;
	char	buf[BUFSIZ * 10];
	int	fd;

	(void) snprintf(temp, size, "%s/mat.XXXXXX", tmpdir);
	if ((fd = mkstemp(temp)) < 0) {
		(void) fprintf(stderr, "can't make temporary file '%s'\n", temp);
		return -1;
	}
	while ((cc = read(STDIN_FILENO, buf, sizeof(buf))) > 0) {
		if (write(fd, buf, (size_t)cc) != cc) {
			(void) fprintf(stderr, "short write to '%s'\n", temp);
			(void) unlink(temp);
			close(fd);
			return -1;
		}
	}
	return fd;
}

/* return non-zero if we're reading the archive */
static int
reading(const char *action)
{
	return (strcmp(action, "verify") == 0 ||
		strcmp(action, "audit") == 0 ||
		strcmp(action, "list") == 0 ||
		strcmp(action, "extract") == 0);
}

/* take a list of files from a metalog */
static int
read_metalog(mat_t *mat, char *metalog)
{
	regmatch_t	 matches[10];
	regex_t		 r;
	FILE		*fp;
	char		 buf[MAT_PATH_MAX * 2];
	char		 f[MAT_PATH_MAX];
	int		 ok;

	mat->recursing = 0;
	if ((fp = fopen(metalog, "r")) == NULL) {
		(void) fprintf(stderr, "can't open metalog '%s'\n", metalog);
		return 0;
	}
	(void) regcomp(&r, "^([^ \t]+)[ \t]+type=[a-z]*[ \t]+uname=([^ \t]+)[ \t]+gname=([^ \t\n]+)", REG_EXTENDED);
	for (ok = 1 ; fgets(buf, (int)sizeof(buf), fp) != NULL ; ) {
		if (regexec(&r, buf, 10, matches, 0) == 0) {
			(void) snprintf(mat->user, sizeof(mat->user), "%.*s",
				(int)(matches[2].rm_eo - matches[2].rm_so),
				&buf[matches[2].rm_so]);
			(void) snprintf(mat->group, sizeof(mat->group), "%.*s",
				(int)(matches[3].rm_eo - matches[3].rm_so),
				&buf[matches[3].rm_so]);
			(void) snprintf(f, sizeof(f), "%.*s",
				(int)(matches[1].rm_eo - matches[1].rm_so),
				&buf[matches[1].rm_so]);
			if (!mat_add(mat, f)) {
				ok = 0;
			}
		}
	}
	(void) fclose(fp);
	regfree(&r);
	return ok;
}

/* compile the regular expression we have */
static int
compile_subst(mat_t *mat, char *arg)
{
	char	*mid;
	char	*s;

	for (s = arg + 1 ; *s && *s != *arg ; s++) {
		if (*s == '\\') {
			s += 1;
		}
	}
	(void) snprintf(mat->from, sizeof(mat->from), "%.*s",
		(int)(s - arg) - 1, arg + 1);
	for (mid = s++ ; *s && *s != *arg ; s++) {
		if (*s == '\\') {
			s += 1;
		}
	}
	(void) snprintf(mat->to, sizeof(mat->to), "%.*s",
		(int)(s - mid) - 1, mid + 1);
	if ((mat->re = calloc(sizeof(regex_t), 1)) == NULL) {
		(void) fprintf(stderr, "bad calloc '%s'\n", mat->from);
		return 0;
	}
	if (regcomp(mat->re, mat->from, REG_EXTENDED) != 0) {
		(void) fprintf(stderr, "bad regexp '%s'\n", mat->from);
		return 0;
	}
	return 1;
}

/***************************************************************************/

/* tar personality for mat */
int
mat_tar(int argc, char **argv)
{
	const char	*action;
	getopt2_t	 options;
	mat_t	 	 mat;
	char		 newarg[MAT_PATH_MAX + 256];
	char		 temp[MAT_PATH_MAX];
	char		 cwd[MAT_PATH_MAX];
	char		*metalog;
	char		*dir;
	char		*f;
	int		 deltemp;
	int		 fd;
	int		 ok;
	int		 i;

	(void) memset(&mat, 0x0, sizeof(mat));
	action = NULL;
	f = NULL;
	metalog = NULL;
	dir = NULL;
	deltemp = 0;
	if (*argv[1] != '-') {
		(void) snprintf(newarg, sizeof(newarg), "-%s", argv[1]);
		argv[1] = newarg;
	}
	getopt2_init(&options);
	while ((i = getopt2(&options, argc, argv, "C:T:Vacf:ps:tvx")) != -1) {
		switch(i) {
		case 'C':
			dir = options.optarg;
			break;
		case 'T':
			metalog = options.optarg;
			break;
		case 'V':
			action = "verify";
			break;
		case 'a':
			action = "audit";
			break;
		case 'c':
			action = "create";
			break;
		case 'f':
			f = options.optarg;
			break;
		case 'p':
			mat.preserve = 1;
			break;
		case 's':
			if (!compile_subst(&mat, options.optarg)) {
				return 0;
			}
			break;
		case 't':
			action = "list";
			break;
		case 'v':
			mat.verbose += 1;
			break;
		case 'x':
			action = "extract";
			break;
		default:
			break;
		}
	}
	if (action == NULL) {
		(void) fprintf(stderr, "No action specified\n");
		return 0;
	}
	if (f == NULL) {
		(void) fprintf(stderr, "No file specified\n");
		return 0;
	}
	if (reading(action) && strcmp(f, "-") == 0) {
		/* read archive from stdin */
		if ((fd = dostdin(TMPDIR, f = temp, sizeof(temp))) < 0) {
			return 0;
		}
		deltemp = 1;
		(void) close(fd);
	}
	if (!mat_init(&mat, f, action)) {
		(void) fprintf(stderr, "can't initialise mat\n");
		return 0;
	}
	getcwd(cwd, sizeof(cwd));
	if (dir != NULL && chdir(dir) < 0) {
		(void) fprintf(stderr, "can't chdir to '%s'\n", dir);
		return 0;
	}
	if (metalog) {
		ok = read_metalog(&mat, metalog);
	} else if (argc == options.optind) {
		if (strcmp(action, "list") == 0) {
			ok = mat_list(&mat, NULL, stdout);
		} else if (strcmp(action, "verify") == 0) {
			ok = mat_verify(&mat);
		} else if (strcmp(action, "extract") == 0) {
			ok = mat_extract(&mat, NULL);
		} else if (strcmp(action, "audit") == 0) {
			ok = mat_audit(&mat, NULL, stdout);
		} else {
			(void) fprintf(stderr, "unknown action specified (%s)\n", action);
			return 0;
		}
	} else {
		for (ok = 1, i = options.optind ; i < argc ; i++) {
			if (strcmp(action, "create") == 0) {
				if (!mat_add(&mat, argv[i])) {
					ok = 0;
				}
			} else if (strcmp(action, "extract") == 0) {
				if (!mat_extract(&mat, argv[i])) {
					ok = 0;
				}
			} else if (strcmp(action, "list") == 0) {
				if (!mat_list(&mat, argv[i], stdout)) {
					ok = 0;
				}
			} else if (strcmp(action, "audit") == 0) {
				if (!mat_audit(&mat, argv[i], stdout)) {
					ok = 0;
				}
			}
		}
	}
	(void) mat_end(&mat);
	if (deltemp) {
		(void) unlink(temp);
	}
	if (dir != NULL && chdir(cwd) < 0) {
		(void) fprintf(stderr, "can't chdir back to '%s'\n", cwd);
	}
	return ok;
}

#ifndef MAT_ARGC_MAX
#define MAT_ARGC_MAX	2048
#endif

/* varargs version of tar */
int
mat_vtar(char *arg, ...)
{
	va_list	 args;
	char	*argv[MAT_ARGC_MAX];
	char	*s;
	int	 argc;

	argv[0] = arg;
	va_start(args, arg);
	for (argc = 1 ; (s = va_arg(args, char *)) != NULL && argc < MAT_ARGC_MAX ; argc++) {
		argv[argc] = s;
	}
	va_end(args);
	return mat_tar(argc, argv);
}

/* pax personality for mat */
int
mat_pax(int argc, char **argv)
{
	const char	*action;
	getopt2_t	 options;
	mat_t		 mat;
	char		 temp[MAT_PATH_MAX];
	char		 cwd[MAT_PATH_MAX];
	char		*metalog;
	char		*archive;
	char		*dir;
	int		 deltemp;
	int		 fd;
	int		 ok;
	int		 i;

	(void) memset(&mat, 0x0, sizeof(mat));
	dir = NULL;
	metalog = NULL;
	archive = NULL;
	ok = 1;
	action = "list";
	deltemp = 0;
	getopt2_init(&options);
	while ((i = getopt2(&options, argc, argv, "C:Vaf:lrs:wv")) != -1) {
		switch(i) {
		case 'C':
			dir = options.optarg;
			break;
		case 'V':
			action = "verify";
			break;
		case 'a':
			action = "audit";
			break;
		case 'f':
			archive = options.optarg;
			break;
		case 'l':
			action = "list";
			break;
		case 'r':
			action = "extract";
			break;
		case 's':
			if (!compile_subst(&mat, options.optarg)) {
				(void) fprintf(stderr, "bad regexp '%s'\n", options.optarg);
				return 0;
			}
			break;
		case 'w':
			action = "create";
			break;
		case 'v':
			mat.verbose += 1;
			break;
		default:
			(void) fprintf(stderr, "Unknown flag \"%c\"\n", i);
		}
	}
	if (archive == NULL) {
		(void) fprintf(stderr, "No file specified\n");
		return 0;
	}
	if (reading(action) && strcmp(archive, "-") == 0) {
		/* read archive from stdin */
		if ((fd = dostdin(TMPDIR, archive = temp, sizeof(temp))) < 0) {
			return 0;
		}
		deltemp = 1;
		(void) close(fd);
	}
	if (!mat_init(&mat, archive, action)) {
		(void) fprintf(stderr, "can't initialise mat\n");
		return 0;
	}
	getcwd(cwd, sizeof(cwd));
	if (dir != NULL && chdir(dir) < 0) {
		(void) fprintf(stderr, "can't chdir to '%s'\n", dir);
		return 0;
	}
	if (metalog != NULL) {
		ok = read_metalog(&mat, metalog);
	} else if (argc == options.optind) {
		if (strcmp(action, "list") == 0) {
			ok = mat_list(&mat, NULL, stdout);
		} else if (strcmp(action, "verify") == 0) {
			ok = mat_verify(&mat);
		} else if (strcmp(action, "extract") == 0) {
			ok = mat_extract(&mat, NULL);
		} else {
			(void) fprintf(stderr, "unknown action specified (%s)\n", action);
			return 0;
		}
	} else {
		for (ok = 1, i = options.optind ; i < argc ; i++) {
			if (strcmp(action, "create") == 0) {
				if (!mat_add(&mat, argv[i])) {
					ok = 0;
				}
			} else if (strcmp(action, "extract") == 0) {
				if (!mat_extract(&mat, argv[i])) {
					ok = 0;
				}
			} else if (strcmp(action, "list") == 0) {
				if (!mat_list(&mat, argv[i], stdout)) {
					ok = 0;
				}
			}
		}
	}
	(void) mat_end(&mat);
	if (deltemp) {
		(void) unlink(temp);
	}
	if (dir != NULL && chdir(cwd) < 0) {
		(void) fprintf(stderr, "can't chdir back to '%s'\n", cwd);
	}
	return ok;
}

/* varargs version of pax */
int
mat_vpax(char *arg, ...)
{
	va_list	 args;
	char	*argv[MAT_ARGC_MAX];
	char	*s;
	int	 argc;

	argv[0] = arg;
	va_start(args, arg);
	for (argc = 1 ; (s = va_arg(args, char *)) != NULL && argc < MAT_ARGC_MAX ; argc++) {
		argv[argc] = s;
	}
	va_end(args);
	return mat_pax(argc, argv);
}
