/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <dirent.h>
#include <fnmatch.h>
#include <grp.h>
#include <inttypes.h>
#include <pwd.h>
#include <regex.h>
#include <sha2.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "mat.h"

#define BSWAP32(x)	((((x) & 0xff) << 24) |				\
			 (((x) & 0xff00) << 8) |			\
			 (((x) & 0xff0000) >> 8) |			\
			 (((x) & 0xff000000) >> 24))

#define BSWAP64(x)	((((x) & 0xffULL) << 56) |			\
			 (((x) & 0xff00ULL) << 40) |			\
			 (((x) & 0xff0000ULL) << 24) |			\
			 (((x) & 0xff000000ULL) << 8) |			\
			 (((x) & 0xff00000000ULL) >> 8) |		\
			 (((x) & 0xff0000000000ULL) >> 24) |		\
			 (((x) & 0xff000000000000ULL) >> 40) |		\
			 (((x) & 0xff00000000000000ULL) >> 56))

/* XXX - incorprate into autoconf - agc */
#ifdef __APPLE__
#define LISTXATTR_ARGS	4
#define GETXATTR_ARGS	6
#define SETXATTR_ARGS	6
#else
#define LISTXATTR_ARGS	3
#define GETXATTR_ARGS	4
#define SETXATTR_ARGS	5
#endif

/* XXX - incorprate into autoconf - agc */
#define HAVE_FUTIMES	1
#define HAVE_LUTIMES	1

/* structure used in extracting components from archives */
typedef struct extract_t {
	uint64_t	 off;		/* offset in archive */
	int		 preserve;	/* preserve file times and owners */
	int		 extract;	/* really extract? */
	uint8_t		*s;		/* archive itself */
	char		 newpath[MAT_PATH_MAX]; /* if it needs to change */
} extract_t;

/* byte-swap 32 bits, if necessary */
static uint32_t
local_htobe32(uint32_t in)
{
	int	indian = 1;

	return (*(char *)(void *)&indian) ? BSWAP32(in) : in;
}

/* byte-swap 32 bits, if necessary */
static uint32_t
local_be32toh(uint32_t in)
{
	return local_htobe32(in);
}

/* byte-swap 64 bits, if necessary */
static uint64_t
local_htobe64(uint64_t in)
{
	int	indian = 1;

	return (*(char *)(void *)&indian) ? (uint64_t)BSWAP64(in) : in;
}

/* byte-swap 64 bits, if necessary */
static uint64_t
local_be64toh(uint64_t in)
{
	return local_htobe64(in);
}

/* format a user or group name in text */
static char *
find_name(const char *type, uint32_t id, char *buf, size_t size)
{
	struct passwd	*pwp;
	struct group	*grp;

	if (strcmp(type, "user") == 0) {
		if ((pwp = getpwuid(id)) == NULL) {
			(void) snprintf(buf, size, "%u", id);
		} else {
			(void) snprintf(buf, size, "%s", pwp->pw_name);
		}
	} else {
		if ((grp = getgrgid(id)) == NULL) {
			(void) snprintf(buf, size, "%u", id);
		} else {
			(void) snprintf(buf, size, "%s", grp->gr_name);
		}
	}
	return buf;
}

/* format a user or group name in text */
static int
find_id(const char *type, const char *name, int32_t *id)
{
	struct passwd	*pwp;
	struct group	*grp;

	if (strcmp(type, "user") == 0) {
		if ((pwp = getpwnam(name)) == NULL) {
			return 0;
		}
		*id = pwp->pw_uid;
	} else {
		if ((grp = getgrnam(name)) == NULL) {
			return 0;
		}
		*id = grp->gr_gid;
	}
	return 1;
}

/* fill the meta information in from knowledge about the directory entry */
static void
fillmeta(matent_t *ent, const char *f, struct stat *st, const char *user, const char *group)
{
	(void) memset(ent, 0x0, sizeof(ent->meta));
	ent->meta.size = st->st_size;
	ent->meta.mode = st->st_mode;
	ent->meta.nlink = st->st_nlink;
	ent->meta.atime = st->st_atime;
	ent->meta.mtime = st->st_mtime;
	if (user && *user) {
		(void) snprintf(ent->meta.user, sizeof(ent->meta.user), "%s", user);
	} else {
		(void) find_name("user", st->st_uid, ent->meta.user, sizeof(ent->meta.user));
	}
	if (group && *group) {
		(void) snprintf(ent->meta.group, sizeof(ent->meta.group), "%s", group);
	} else {
		(void) find_name("group", st->st_gid, ent->meta.group, sizeof(ent->meta.group));
	}
	ent->meta.namelen = (uint32_t)(strlen(f) + 1);
	ent->meta.ino = st->st_ino;
	ent->meta.dev = st->st_dev;
	ent->meta.flags = st->st_flags;
}

/* write the entry's meta information to the archive */
static int
writemeta(int fd, matent_t *ent)
{
	matent_t	lent;

	lent.meta.size = local_htobe64(ent->meta.size);
	lent.meta.mode = local_htobe32(ent->meta.mode);
	lent.meta.nlink = local_htobe32(ent->meta.nlink);
	lent.meta.atime = local_htobe64((uint64_t)ent->meta.atime);
	lent.meta.mtime = local_htobe64((uint64_t)ent->meta.mtime);
	(void) memcpy(lent.meta.user, ent->meta.user, sizeof(lent.meta.user));
	(void) memcpy(lent.meta.group, ent->meta.group, sizeof(lent.meta.group));
	lent.meta.namelen = local_htobe32(ent->meta.namelen);
	lent.meta.ino = local_htobe64(ent->meta.ino);
	lent.meta.dev = local_htobe64(ent->meta.dev);
	lent.meta.flags = local_htobe32(ent->meta.flags);
	lent.meta.linklen = local_htobe32(ent->meta.linklen);
	lent.meta.type = ent->meta.type;
	return write(fd, &lent, sizeof(lent.meta)) == sizeof(lent.meta);
}

/* read the entry's meta information from the archive */
static int
readmeta(uint8_t *s, matent_t *ent)
{
	matent_t	lent;

	(void) memcpy(&lent, s, sizeof(lent.meta));
	ent->meta.size = local_be64toh(lent.meta.size);
	ent->meta.mode = local_be32toh(lent.meta.mode);
	ent->meta.nlink = local_be32toh(lent.meta.nlink);
	ent->meta.atime = local_be64toh((uint64_t)lent.meta.atime);
	ent->meta.mtime = local_be64toh((uint64_t)lent.meta.mtime);
	(void) memcpy(ent->meta.user, lent.meta.user, sizeof(ent->meta.user));
	(void) memcpy(ent->meta.group, lent.meta.group, sizeof(ent->meta.group));
	ent->meta.namelen = local_be32toh(lent.meta.namelen);
	ent->meta.ino = local_be64toh(lent.meta.ino);
	ent->meta.dev = local_be64toh(lent.meta.dev);
	ent->meta.flags = local_be32toh(lent.meta.flags);
	ent->meta.linklen = local_be32toh(lent.meta.linklen);
	ent->meta.type = lent.meta.type;
	return 1;
}

/* calculate a combo digest */
static void
digest_update(uint8_t *dig, uint8_t *s, size_t size)
{
	SHA256_CTX	sha256;

	SHA256_Init(&sha256);
	SHA256_Update(&sha256, s, size);
	SHA256_Final(dig, &sha256);
}

/* calculate a combo digest */
static int
digest_file(uint8_t *dig, const char *f)
{
	struct stat	 st;
	SHA256_CTX	 sha256;
	ssize_t		 cc;
	size_t		 size;
	uint8_t		*mapped;
	FILE		*fp;
	char		 buf[BUFSIZ * 10];

	if ((fp = fopen(f, "r")) == NULL) {
		return 0;
	}
	SHA256_Init(&sha256);
	(void) fstat(fileno(fp), &st);
	size = (size_t)st.st_size;
	mapped = mmap(NULL, size, PROT_READ, MAP_FILE, fileno(fp), 0);
	if (mapped == MAP_FAILED) {
		SHA256_Update(&sha256, mapped, size);
	} else {
		while ((cc = read(fileno(fp), buf, sizeof(buf))) > 0) {
			SHA256_Update(&sha256, (const uint8_t *)buf, (size_t)cc);
		}
	}
	SHA256_Final(dig, &sha256);
	if (mapped != MAP_FAILED) {
		munmap(mapped, size);
	}
	(void) fclose(fp);
	return 1;
}

/* make intermediate dirs */
static int
intermediates(const char *path)
{
	struct stat	 st;
	char		 d[MAT_PATH_MAX];
	char		*slash;

	(void) snprintf(d, sizeof(d), "%s", path);
	for (slash = d + 2 ; (slash = strchr(slash, '/')) != NULL ; slash += 1) {
		*slash = 0x0;
		if (stat(d, &st) < 0 && mkdir(d, 0777) < 0) {
			(void) fprintf(stderr, "can't create intermediate dir '%s'\n", d);
			return 0;
		}
		*slash = '/';
	}
	return 1;
}

/* preserve modes and ownership */
static int
fpreserve(int fd, const char *lpath, const char *path, matent_t *ent)
{ 
	uint32_t	mode;
	int32_t		uid;
	int32_t		gid;
	int		ret;

	mode = (geteuid() == 0) ? (ent->meta.mode & 07777) : (ent->meta.mode & 0777);
	ret = (fd >= 0) ? fchmod(fd, mode) :
		(lpath) ? lchmod(lpath, mode) : chmod(path, mode);
	if (ret < 0) {
		(void) fprintf(stderr, "fchmod failed for '%s'\n", lpath);
		return 0;
	}
	if (!find_id("user", ent->meta.user, &uid) ||
	    !find_id("group", ent->meta.group, &gid)) {
		(void) fprintf(stderr, "no such user/group '%s/%s'\n",
			ent->meta.user, ent->meta.group);
		return 0;
	}
	ret = (fd >= 0) ? fchown(fd, (uid_t)uid, (gid_t)gid) :
		(lpath) ? lchown(lpath, (uid_t)uid, (gid_t)gid) :
		chown(path, (uid_t)uid, (gid_t)gid);
	if (ret < 0) {
		(void) fprintf(stderr, "fchown failed for '%s'\n", path);
		return 0;
	}
	return 1;
}

/* find a hardlink, get index in link table */
static int
findlink(mat_t *mat, uint64_t ino)
{
	int	i;

	for (i = 0 ; i < (int)mat->linkc ; i++) {
		if (mat->links[i].ino == ino) {
			return i;
		}
	}
	return -1;
}

/* add a link's inode and primary name */
static int
isprimary(mat_t *mat, uint64_t ino, const char *name, int *primary)
{
	matlink_t	*lnk;

	if (mat->linkc == MAT_LINKTAB_SIZE) {
		(void) fprintf(stderr, "performance warning: hard link table full\n");
		return 1;
	}
	if ((*primary = findlink(mat, ino)) >= 0) {
		return 0;
	}
	lnk = &mat->links[mat->linkc];
	lnk->ino = ino;
	lnk->namelen = (uint32_t)(strlen(name) + 1);
	lnk->name = strdup(name);
	*primary = mat->linkc++;
	return 1;
}

/* substitute for the name */
static int
substname(mat_t *mat, extract_t *extract, const char *name)
{
	regmatch_t	matches[10];

	if (mat->re && regexec(mat->re, name, 10, matches, 0) == 0) {
		(void) snprintf(extract->newpath, sizeof(extract->newpath), "%.*s%.*s%.*s",
			(int)matches[0].rm_so, name,
			(int)strlen(mat->to), mat->to,
			(int)strlen(&name[matches[0].rm_eo]), &name[matches[0].rm_eo]);
		return 1;
	}
	return 0;
}

/* extract a hardlink from the archive */
static int
extract_link(mat_t *mat, extract_t *archive, matent_t *ent, const char *name)
{
	if (archive->extract) {
		if (substname(mat, archive, name)) {
			name = archive->newpath;
		}
		(void) unlink(name);
		if (!intermediates(name)) {
			(void) fprintf(stderr, "can't create intermediate dirs '%s'\n", name);
			return 0;
		}
		if (link((char *)&archive->s[archive->off], name) < 0) {
			(void) fprintf(stderr, "can't link '%s'\n", name);
			return 0;
		}
	}
	archive->off += ent->meta.linklen;
	archive->off += sizeof(ent->digest);
	if (archive->extract && archive->preserve && !fpreserve(-1, NULL, name, ent)) {
		(void) fprintf(stderr, "preserve info failed for '%s'\n", name);
		return 0;
	}
	return 1;
}

/* extract a file from the archive */
static int
extract_file(mat_t *mat, extract_t *archive, matent_t *ent, const char *name)
{
	struct timeval	 tv[2];
	FILE		*fp;

	if (archive->extract) {
		if (substname(mat, archive, name)) {
			name = archive->newpath;
		}
		(void) unlink(name);
		if (!intermediates(name)) {
			(void) fprintf(stderr, "can't create intermediate dirs '%s'\n", name);
			return 0;
		}
		if ((fp = fopen(name, "w")) == NULL) {
			(void) fprintf(stderr, "can't create '%s'\n", name);
			return 0;
		}
		write(fileno(fp), &archive->s[archive->off], ent->meta.size);
		tv[0].tv_sec = ent->meta.atime;
		tv[0].tv_usec = 0;
		tv[1].tv_sec = ent->meta.mtime;
		tv[1].tv_usec = 0;
#ifdef HAVE_FUTIMES
		if (futimes(fileno(fp), tv) < 0) {
			(void) fprintf(stderr, "futimes failed for '%s'\n", name);
		}
#else
		if (utimes(name, tv) < 0) {
			(void) fprintf(stderr, "utimes failed for '%s'\n", name);
		}
#endif
		if (archive->preserve && !fpreserve(fileno(fp), NULL, NULL, ent)) {
			(void) fprintf(stderr, "preserve info failed for '%s'\n", name);
			return 0;
		}
		(void) fclose(fp);
	}
	/* next are the contents */
	archive->off += ent->meta.size;
	/* and last is the digest */
	archive->off += sizeof(ent->digest);
	return 1;
}

/* extract a symlink from the archive */
static int
extract_symlink(mat_t *mat, extract_t *archive, matent_t *ent, const char *name)
{
	struct timeval	 tv[2];

	if (archive->extract) {
		if (substname(mat, archive, name)) {
			name = archive->newpath;
		}
		(void) unlink(name);
		if (!intermediates(name)) {
			(void) fprintf(stderr, "can't create intermediate dirs '%s'\n", name);
			return 0;
		}
		if (symlink((char *)&archive->s[archive->off], name) < 0) {
			(void) fprintf(stderr, "can't symlink '%s'\n", name);
			return 0;
		}
#ifdef HAVE_LUTIMES
		tv[0].tv_sec = ent->meta.atime;
		tv[0].tv_usec = 0;
		tv[1].tv_sec = ent->meta.mtime;
		tv[1].tv_usec = 0;
		if (lutimes(name, tv) < 0) {
			(void) fprintf(stderr, "futimes failed for '%s'\n", name);
			return 0;
		}
#endif
		if (archive->preserve && !fpreserve(-1, name, NULL, ent)) {
			(void) fprintf(stderr, "preserve info failed for '%s'\n", name);
			return 0;
		}
	}
	archive->off += ent->meta.linklen;
	archive->off += sizeof(ent->digest);
	return 1;
}

/* extract a directory from the archive */
static int
extract_dir(mat_t *mat, extract_t *archive, matent_t *ent, const char *name)
{
	if (archive->extract) {
		if (substname(mat, archive, name)) {
			name = archive->newpath;
		}
		if (!intermediates(name)) {
			(void) fprintf(stderr, "can't create intermediate dirs '%s'\n", name);
			return 0;
		}
		if (mkdir(name, ent->meta.mode & 07777) < 0) {
			(void) fprintf(stderr, "can't mkdir '%s'\n", name);
			return 0;
		}
		if (archive->preserve && !fpreserve(-1, NULL, name, ent)) {
			(void) fprintf(stderr, "preserve info failed for '%s'\n", name);
			return 0;
		}
	}
	return 1;
}

/* extract a special device from the archive */
static int
extract_special(mat_t *mat, extract_t *archive, matent_t *ent, const char *name)
{
	if (archive->extract) {
		if (substname(mat, archive, name)) {
			name = archive->newpath;
		}
		if (!intermediates(name)) {
			(void) fprintf(stderr, "can't create intermediate dirs '%s'\n", name);
			return 0;
		}
		if (mknod(name, ent->meta.mode & 07777, ent->meta.dev) < 0) {
			(void) fprintf(stderr, "can't mknod '%s'\n", name);
			return 0;
		}
		if (archive->preserve && !fpreserve(-1, NULL, name, ent)) {
			(void) fprintf(stderr, "preserve info failed for '%s'\n", name);
			return 0;
		}
	}
	return 1;
}

/* translate old file name to new one */
static int
write_meta_info(mat_t *mat, matent_t *ent, const char *f)
{
	regmatch_t	matches[10];
	char		newpath[MAT_PATH_MAX];

	if (mat->re && regexec(mat->re, f, 10, matches, 0) == 0) {
		ent->meta.namelen = snprintf(newpath, sizeof(newpath), "%.*s%.*s%.*s",
			(int)matches[0].rm_so, f,
			(int)strlen(mat->to), mat->to,
			(int)strlen(&f[matches[0].rm_eo]), &f[matches[0].rm_eo]) + 1;
		f = newpath;
	}
	writemeta(fileno(mat->fp), ent);
	write(fileno(mat->fp), f, ent->meta.namelen);
	return 1;
}

/*******************************************************************************/

/* add an entry to the archive */
int
mat_add(mat_t *mat, const char *f)
{
	struct dirent	 d;
	struct dirent	*dp;
	struct stat	 st;
	matent_t	 ent;
	uint8_t		*mapped;
	ssize_t		 cc;
	char		 newpath[MAT_PATH_MAX];
	char		 dotpath[MAT_PATH_MAX];
	FILE		*fp;
	DIR		*dirp;
	int		 original;
	int		 primary;

	(void) memset(&ent, 0x0, sizeof(ent));
	if (*f == '/') {
		(void) snprintf(dotpath, sizeof(dotpath), ".%s", f);
		f = dotpath;
	}
	if (lstat(f, &st) < 0) {
		(void) fprintf(stderr, "can't stat '%s'\n", f);
		return 0;
	}
	primary = -1;
	/* categorise entries first */
	switch(st.st_mode & S_IFMT) {
	case S_IFREG:
		if (st.st_ino == mat->archive) {
			(void) fprintf(stderr, "ignoring attempt to add ourselves '%s'\n", f);
			return 1;
		}
		fillmeta(&ent, f, &st, mat->user, mat->group);
		ent.meta.type = 'f';
		if (ent.meta.nlink > 1) {
			original = isprimary(mat, ent.meta.ino, f, &primary);
			if (!original) {
				ent.meta.type = 'h';
			}
		}
		break;
	case S_IFLNK:
		fillmeta(&ent, f, &st, mat->user, mat->group);
		ent.meta.type = 'l';
		break;
	case S_IFBLK:
	case S_IFCHR:
		fillmeta(&ent, f, &st, mat->user, mat->group);
		ent.meta.type = ((st.st_mode & S_IFMT) == S_IFBLK) ? 'b' : 'c';
		break;
	case S_IFDIR:
		fillmeta(&ent, f, &st, mat->user, mat->group);
		ent.meta.type = 'd';
		break;
	default:
		(void) fprintf(stderr, "unknown type '%s'\n", f);
		return 0;
	}
	/* now switch on type of entry */
	switch(ent.meta.type) {
	case 'f':
		if ((fp = fopen(f, "r")) == NULL) {
			(void) fprintf(stderr, "can't open '%s'\n", f);
			return 0;
		}
		mapped = mmap(NULL, (size_t)st.st_size, PROT_READ, MAP_FILE, fileno(fp), 0);
		if (mapped == MAP_FAILED) {
			(void) fclose(fp);
			(void) fprintf(stderr, "can't mmap '%s'\n", f);
			return 0;
		}
		write_meta_info(mat, &ent, f);
		write(fileno(mat->fp), mapped, (size_t)st.st_size);
		digest_update(ent.digest, mapped, (size_t)st.st_size);
		if (ent.meta.nlink > 1 && original) {
			/* multi-linked, and we're the primary */
			(void) memcpy(&mat->links[primary].digest, ent.digest, sizeof(ent.digest));
		}
		write(fileno(mat->fp), ent.digest, sizeof(ent.digest));
		munmap(mapped, (size_t)st.st_size);
		(void) fclose(fp);
		break;
	case 'h':
		/* store hardlink to primary */
		ent.meta.linklen = mat->links[primary].namelen;
		write_meta_info(mat, &ent, f);
		write(fileno(mat->fp), mat->links[primary].name, (size_t)ent.meta.linklen);
		/* get digest from primary */
		(void) memcpy(ent.digest, mat->links[primary].digest, sizeof(ent.digest));
		write(fileno(mat->fp), ent.digest, sizeof(ent.digest));
		break;
	case 'l':
		if ((cc = readlink(f, newpath, sizeof(newpath))) < 0) {
			(void) fprintf(stderr, "can't readlink '%s'\n", f);
			return 0;
		}
		newpath[cc] = 0x0;
		ent.meta.linklen = (uint32_t)(cc + 1);
		write_meta_info(mat, &ent, f);
		write(fileno(mat->fp), newpath, (size_t)ent.meta.linklen);
		digest_update(ent.digest, (uint8_t *)newpath, (size_t)ent.meta.linklen);
		write(fileno(mat->fp), ent.digest, sizeof(ent.digest));
		break;
	case 'b':
	case 'c':
		write_meta_info(mat, &ent, f);
		break;
	case 'd':
		write_meta_info(mat, &ent, f);
		if (mat->recursing) {
			if ((dirp = opendir(f)) == NULL) {
				(void) fprintf(stderr, "can't opendir '%s'\n", f);
				return 0;
			}
			(void) memset(&d, 0x0, sizeof(d));
			while (readdir_r(dirp, &d, &dp) == 0 && dp != NULL) {
				if (strcmp(dp->d_name, ".") == 0 || strcmp(dp->d_name, "..") == 0) {
					continue;
				}
				(void) snprintf(newpath, sizeof(newpath), "%s/%s", f, dp->d_name);
				mat_add(mat, newpath);
			}
			(void) closedir(dirp);
		}
		break;
	default:
		(void) fprintf(stderr, "unknown type '%s'\n", f);
		break;
	}
	return 1;
}

#define MONTHS(x)	((x) * 30 * 24 * 60 * 60)

/* list the archive */
int
mat_list(mat_t *mat, const char *pat, FILE *fp)
{
	uint64_t	 off;
	unsigned	 i;
	matent_t	 ent;
	uint8_t		*contents;
	uint8_t		*name;
	uint8_t		*s;
	time_t		 t;
	char		 buf[MAT_PATH_MAX * 2];
	char		 mbuf[12];
	char		*tim;
	char		 date[16];
	int		 list;
	int		 cc;

	if (!mat_verify(mat)) {
		return 0;
	}
	for (off = sizeof(matheader_t), s = mat->mapped; off < mat->size ; ) {
		(void) memset(&ent, 0x0, sizeof(ent));
		contents = NULL;
		readmeta(&s[off], &ent);
		/* skip over meta info */
		off += sizeof(ent.meta);
		/* next is name */
		name = &s[off];
		off += ent.meta.namelen;
		list = (pat == NULL) ? 1 : (fnmatch(pat, (char *)name, FNM_PATHNAME) == 0);
		switch(ent.meta.type) {
		case 'f':
			/* next are the contents */
			digest_update(ent.digest, &s[off], ent.meta.size);
			off += ent.meta.size;
			/* and last is the digest */
			off += sizeof(ent.digest);
			break;
		case 'h':
			/* next are the contents - the target of the symlink */
			contents = &s[off];
			off += ent.meta.linklen;
			(void) memcpy(ent.digest, &s[off], sizeof(ent.digest));
			off += sizeof(ent.digest);
			break;
		case 'l':
			/* next are the contents - the target of the symlink */
			contents = &s[off];
			digest_update(ent.digest, &s[off], ent.meta.linklen);
			off += ent.meta.linklen;
			off += sizeof(ent.digest);
			break;
		case 'b':
		case 'c':
		case 'd':
			break;
		default:
			(void) fprintf(stderr, "unrecognised type '%c' '%s'\n",
				ent.meta.type, name);
			break;
		}
		cc = 0;
		if (mat->verbose > 0) {
			mbuf[0] = (ent.meta.type == 'f' || ent.meta.type == 'h') ? '-' : ent.meta.type;
			mbuf[1] = (ent.meta.mode & 0400) ? 'r' : '-';
			mbuf[2] = (ent.meta.mode & 0200) ? 'w' : '-';
			mbuf[3] = (ent.meta.mode & 04000) ? 's' : (ent.meta.mode & 0100) ? 'x' : '-';
			mbuf[4] = (ent.meta.mode & 0040) ? 'r' : '-';
			mbuf[5] = (ent.meta.mode & 0020) ? 'w' : '-';
			mbuf[6] = (ent.meta.mode & 02000) ? 's' : (ent.meta.mode & 0010) ? 'x' : '-';
			mbuf[7] = (ent.meta.mode & 0004) ? 'r' : '-';
			mbuf[8] = (ent.meta.mode & 0002) ? 'w' : '-';
			mbuf[9] = (ent.meta.mode & 01000) ? 't' : (ent.meta.mode & 0001) ? 'x' : '-';
			mbuf[10] = 0x0;
			t = ent.meta.mtime;
			tim = ctime(&t);
			t = time(NULL);
			if (t - ent.meta.mtime >= MONTHS(6)) {
				(void) snprintf(date, sizeof(date), "%.7s %.4s", &tim[4], &tim[20]);
			} else {
				(void) snprintf(date, sizeof(date), "%.12s", &tim[4]);
			}
			cc += snprintf(&buf[cc], sizeof(buf) - cc, "%s %2d %s  %s %6" PRIu64 " %s ",
				mbuf, ent.meta.nlink, ent.meta.user, ent.meta.group, ent.meta.size, date);
			if (mat->verbose > 1) {
				for (i = 0 ; i < sizeof(ent.digest) ; i++) {
					cc += snprintf(&buf[cc], sizeof(buf) - cc, "%02x", ent.digest[i]);
				}
				buf[cc++] = ' ';
			}
		}
		cc += snprintf(&buf[cc], sizeof(buf) - cc, "%s", name);
		if (mat->verbose > 0 && ent.meta.type == 'l') {
			cc += snprintf(&buf[cc], sizeof(buf) - cc, " -> %s", contents);
		}
		if (mat->verbose > 0 && ent.meta.type == 'h') {
			cc += snprintf(&buf[cc], sizeof(buf) - cc, " == %s", contents);
		}
		cc += snprintf(&buf[cc], sizeof(buf) - cc, "\n");
		if (list && fp) {
			(void) fprintf(fp, "%s", buf);
		}
	}
	return 1;
}

/* verify the archive */
int
mat_verify(mat_t *mat)
{
	matheader_t	 header;
	uint64_t	 off;
	matent_t	 ent;
	uint8_t		*name;
	uint8_t		*s;

	if (mat == NULL) {
		return 0;
	}
	(void) memcpy(&header, mat->mapped, sizeof(header));
	if (local_be32toh(header.magic) != MAT_MAGIC) {
		(void) fprintf(stderr, "not a mat archive\n");
		return 0;
	}
	for (off = sizeof(header), s = mat->mapped; off < mat->size ; ) {
		(void) memset(&ent, 0x0, sizeof(ent));
		readmeta(&s[off], &ent);
		/* skip over meta info */
		off += sizeof(ent.meta);
		/* next is name */
		name = &s[off];
		off += ent.meta.namelen;
		if (*name == '/') {
			(void) fprintf(stderr, "leading '/' - dangerous '%.*s'\n",
				(int)ent.meta.namelen, name);
			return 0;
		}
		if (strstr((const char *)name, "/../") != NULL) {
			(void) fprintf(stderr, "name contains '/../' '%.*s'\n",
				(int)ent.meta.namelen, name);
			return 0;
		}
		if (ent.meta.namelen >= mat->size) {
			(void) fprintf(stderr, "bad name length (%u) for '%.*s'\n",
				ent.meta.namelen, (int)ent.meta.namelen, name);
			return 0;
		}
		switch(ent.meta.type) {
		case 'f':
			/* next are the contents */
			digest_update(ent.digest, &s[off], ent.meta.size);
			off += ent.meta.size;
			/* and last is the digest */
			if (memcmp(ent.digest, &s[off], sizeof(ent.digest)) != 0) {
				(void) fprintf(stderr, "bad digest for '%.*s'\n",
					(int)ent.meta.namelen, name);
				return 0;
			}
			off += sizeof(ent.digest);
			if (ent.meta.size >= mat->size) {
				(void) fprintf(stderr, "bad size (%" PRIu64 ") for '%.*s'\n",
					ent.meta.size, (int)ent.meta.namelen, name);
				return 0;
			}
			break;
		case 'h':
			off += ent.meta.linklen;
			off += sizeof(ent.digest);
			break;
		case 'l':
			/* next are the contents - the target of the symlink */
			digest_update(ent.digest, &s[off], ent.meta.linklen);
			off += ent.meta.linklen;
			if (memcmp(ent.digest, &s[off], sizeof(ent.digest)) != 0) {
				fprintf(stderr, "bad link digest for '%.*s'\n",
					(int)ent.meta.namelen, name);
				return 0;
			}
			off += sizeof(ent.digest);
			break;
		case 'b':
		case 'c':
		case 'd':
			break;
		default:
			(void) fprintf(stderr, "unrecognised type '%c' '%.*s'\n",
				ent.meta.type, (int)ent.meta.namelen, name);
			break;
		}
	}
	return 1;
}

/* extract a single entry from the archive */
int
mat_extract(mat_t *mat, const char *pat)
{
	extract_t	 archive;
	matent_t	 ent;
	char		*name;

	if (!mat_verify(mat)) {
		return 0;
	}
	(void) memset(&archive, 0x0, sizeof(archive));
	archive.preserve = mat->preserve;
	for (archive.off = sizeof(matheader_t), archive.s = mat->mapped; archive.off < mat->size ; ) {
		archive.extract = 0;
		(void) memset(&ent, 0x0, sizeof(ent));
		readmeta(&archive.s[archive.off], &ent);
		/* skip over meta info */
		archive.off += sizeof(ent.meta);
		/* next is name */
		name = (char *)&archive.s[archive.off];
		archive.off += ent.meta.namelen;
		archive.extract = (pat == NULL) ? 1 : (fnmatch(pat, name, FNM_PATHNAME) == 0);
		switch(ent.meta.type) {
		case 'h':
			if (!extract_link(mat, &archive, &ent, name)) {
				return 0;
			}
			break;
		case 'f':
			if (!extract_file(mat, &archive, &ent, name)) {
				return 0;
			}
			break;
		case 'l':
			if (!extract_symlink(mat, &archive, &ent, name)) {
				return 0;
			}
			break;
		case 'd':
			if (!extract_dir(mat, &archive, &ent, name)) {
				return 0;
			}
			break;
		case 'b':
		case 'c':
			if (!extract_special(mat, &archive, &ent, name)) {
				return 0;
			}
			break;
		default:
			(void) fprintf(stderr, "unrecognised type '%s'\n", name);
			break;
		}
		if (archive.extract && mat->verbose > 0) {
			(void) fprintf(stdout, "%s\n", name);
		}
	}
	return 1;
}

/* initialise the archive */
int
mat_init(mat_t *mat, const char *f, const char *mode)
{
	struct stat	st;
	matheader_t	header;

	mat->recursing = 1; /* default behavior */
	if (strcmp(mat->action = mode, "create") == 0) {
		if (f == NULL || strcmp(f, "-") == 0) {
			mat->fp = stdout;
		} else if ((mat->fp = fopen(f, "w")) == NULL) {
			(void) fprintf(stderr, "can't create archive '%s'\n", f);
			return 0;
		}
		fstat(fileno(mat->fp), &st);
		mat->archive = st.st_ino;
		(void) memset(&header, 0x0, sizeof(header));
		header.magic = local_htobe32(MAT_MAGIC);
		write(fileno(mat->fp), &header, sizeof(header));
		return 1;
	}
	if ((mat->fp = fopen(f, "r")) == NULL) {
		(void) fprintf(stderr, "can't read archive '%s'\n", f);
		return 0;
	}
	fstat(fileno(mat->fp), &st);
	mat->archive = st.st_ino;
	mat->size = (size_t)st.st_size;
	mat->mapped = mmap(NULL, (size_t)st.st_size, PROT_READ, MAP_FILE, fileno(mat->fp), 0);
	if (mat->mapped == MAP_FAILED) {
		(void) fclose(mat->fp);
		(void) fprintf(stderr, "can't mmap '%s'\n", f);
		return 0;
	}
	return 1;
}

/* finish up, closing everything */
int
mat_end(mat_t *mat)
{
	if (mat->mapped) {
		munmap(mat->mapped, mat->size);
	}
	(void) fclose(mat->fp);
	return 1;
}

/* list the contents of the archive on the file system */
int
mat_audit(mat_t *mat, const char *pat, FILE *fp)
{
	struct stat	 st;
	uint64_t	 off;
	unsigned	 i;
	matent_t	 fsent;
	matent_t	 ent;
	uint8_t		*contents;
	uint8_t		*name;
	uint8_t		*s;
	time_t		 t;
	char		 buf[MAT_PATH_MAX * 2];
	char		 mbuf[12];
	char		*tim;
	char		 date[16];
	int		 list;
	int		 cc;
	int		 ok;

	if (!mat_verify(mat)) {
		return 0;
	}
	for (ok = 1, off = sizeof(matheader_t), s = mat->mapped; off < mat->size ; ) {
		(void) memset(&ent, 0x0, sizeof(ent));
		(void) memset(&fsent, 0x0, sizeof(fsent));
		readmeta(&s[off], &ent);
		/* skip over meta info */
		off += sizeof(ent.meta);
		/* next is name */
		name = &s[off];
		off += ent.meta.namelen;
		list = (pat == NULL) ? 1 : (fnmatch(pat, (char *)name, FNM_PATHNAME) == 0);
		if (list) {
			if (lstat((const char *)name, &st) < 0) {
				(void) fprintf(stderr, "can't lstat '%s'\n", name);
				list = 0;
			} else {
				fillmeta(&fsent, (const char *)name, &st, NULL, NULL);
			}
		}
		contents = NULL;
		switch(ent.meta.type) {
		case 'f':
			digest_file(fsent.digest, (const char *)name);
			off += ent.meta.size;
			off += sizeof(ent.digest);
			break;
		case 'h':
			contents = &s[off];
			off += ent.meta.linklen;
			digest_file(fsent.digest, (const char *)name);
			off += sizeof(ent.digest);
			break;
		case 'l':
			contents = &s[off];
			if ((cc = (int)readlink((const char *)name, buf, sizeof(buf))) < 0) {
				(void) fprintf(stderr, "bad readlink for '%s'\n", name);
				ok = 0;
			}
			buf[cc] = 0x0;
			digest_update(fsent.digest, (uint8_t *)buf, ent.meta.linklen);
			off += ent.meta.linklen;
			(void) memcpy(ent.digest, &s[off], sizeof(ent.digest));
			off += sizeof(ent.digest);
			break;
		case 'b':
		case 'c':
		case 'd':
			break;
		default:
			(void) fprintf(stderr, "unrecognised type '%c' '%.*s'\n",
				ent.meta.type, (int)ent.meta.namelen, name);
			break;
		}
		cc = 0;
		if (mat->verbose > 0) {
			mbuf[0] = ((st.st_mode & S_IFMT) == S_IFREG) ? '-' :
					((st.st_mode & S_IFMT) == S_IFLNK) ? 'l':
					((st.st_mode & S_IFMT) == S_IFDIR) ? 'd':
					((st.st_mode & S_IFMT) == S_IFBLK) ? 'b':
					((st.st_mode & S_IFMT) == S_IFCHR) ? 'c': '?';
			mbuf[1] = (st.st_mode & 0400) ? 'r' : '-';
			mbuf[2] = (st.st_mode & 0200) ? 'w' : '-';
			mbuf[3] = (st.st_mode & 04000) ? 's' : (st.st_mode & 0100) ? 'x' : '-';
			mbuf[4] = (st.st_mode & 0040) ? 'r' : '-';
			mbuf[5] = (st.st_mode & 0020) ? 'w' : '-';
			mbuf[6] = (st.st_mode & 02000) ? 's' : (st.st_mode & 0010) ? 'x' : '-';
			mbuf[7] = (st.st_mode & 0004) ? 'r' : '-';
			mbuf[8] = (st.st_mode & 0002) ? 'w' : '-';
			mbuf[9] = (st.st_mode & 01000) ? 't' : (st.st_mode & 0001) ? 'x' : '-';
			mbuf[10] = 0x0;
			t = st.st_mtime;
			tim = ctime(&t);
			t = time(NULL);
			if (t - st.st_mtime >= MONTHS(6)) {
				(void) snprintf(date, sizeof(date), "%.7s %.4s", &tim[4], &tim[20]);
			} else {
				(void) snprintf(date, sizeof(date), "%.12s", &tim[4]);
			}
			find_name("user", st.st_uid, fsent.meta.user, sizeof(fsent.meta.user));
			find_name("group", st.st_gid, fsent.meta.group, sizeof(fsent.meta.group));
			cc += snprintf(&buf[cc], sizeof(buf) - cc, "%s %2d %s  %s %6" PRIu64 " %s ",
				mbuf, st.st_nlink, fsent.meta.user, fsent.meta.group, st.st_size, date);
			if (mat->verbose > 1) {
				for (i = 0 ; i < sizeof(fsent.digest) ; i++) {
					cc += snprintf(&buf[cc], sizeof(buf) - cc, "%02x", fsent.digest[i]);
				}
				buf[cc++] = ' ';
			}
		}
		cc += snprintf(&buf[cc], sizeof(buf) - cc, "%s", name);
		if (mat->verbose > 0 && ent.meta.type == 'l') {
			cc += snprintf(&buf[cc], sizeof(buf) - cc, " -> %s", contents);
		}
		if (mat->verbose > 0 && ent.meta.type == 'h') {
			cc += snprintf(&buf[cc], sizeof(buf) - cc, " == %s", contents);
		}
		cc += snprintf(&buf[cc], sizeof(buf) - cc, "\n");
		if (list && fp) {
			(void) fprintf(fp, "%s", buf);
		}
	}
	return ok;
}
