/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef MAT_H_
#define MAT_H_	20110612

#include <sys/types.h>

#include <inttypes.h>
#include <stdio.h>

#ifndef __BEGIN_DECLS
#  if defined(__cplusplus)
#  define __BEGIN_DECLS           extern "C" {
#  define __END_DECLS             }
#  else
#  define __BEGIN_DECLS
#  define __END_DECLS
#  endif
#endif

__BEGIN_DECLS

enum {
	MAT_MAGIC		= 0x19980626,
	MAT_NAME_LEN		= 32,
	MAT_DIGEST_LEN		= 32,
	MAT_PATH_MAX		= 1024,
	MAT_LINKTAB_SIZE	= 2048
};

/* mat archive entry metadata */
typedef struct matmeta_t {
	char		 user[MAT_NAME_LEN];	/* user name */
	char		 group[MAT_NAME_LEN];	/* primary group name */
	uint64_t	 size;			/* size of data */
	int64_t		 atime;			/* last access time */
	int64_t		 mtime;			/* modification time */
	uint64_t	 ino;			/* inode number */
	uint64_t	 dev;			/* device (for specials) */
	uint32_t	 mode;			/* file mode */
	uint32_t	 nlink;			/* number of links */
	uint32_t	 namelen;		/* size of name */
	uint32_t	 linklen;		/* size of link target */
	uint32_t	 flags;			/* fs flags */
#if 0
	uint32_t	 xasize;		/* extattr size */
#endif
	char		 type;			/* single char type */
} matmeta_t;

/* an entry in a mat archive */
typedef struct matent_t {
	matmeta_t	 meta;			/* meta data */
	char		*name;			/* name of entry */
	uint8_t		*data;			/* associated data */
#if 0
	uint8_t		*extattrs;		/* extended attributes */
#endif
	uint8_t		 digest[MAT_DIGEST_LEN]; /* sha256 */
} matent_t;

/* describe a hard link <=> name mapping */
typedef struct matlink_t {
	uint64_t	 ino;			/* inode number */
	char		*name;			/* name of primary */
	uint32_t	 namelen;		/* length of name */
	uint8_t		 digest[MAT_DIGEST_LEN]; /* digest of contents */
} matlink_t;

/* the header for the archive file */
typedef struct matheader_t {
	uint32_t	 magic;			/* magic number */
} matheader_t;

/* the minimalist archive tool structure */
typedef struct mat_t {
	const char	*action;	/* what action we're doing */
	FILE		*fp;		/* archive pointer */
	uint8_t		*mapped;	/* memory mapped archive */
	size_t		 size;		/* size of mmap */
	int		 verbose;	/* be gobby */
	uint64_t	 archive;	/* inode of the archive */
	int		 recursing;	/* when we get a directory, recurse */
	char		 user[MAT_NAME_LEN]; /* when creating, user info */
	char		 group[MAT_NAME_LEN]; /* when creating, group info */
	int		 preserve;	/* restore user and group info */
	uint32_t	 linkc;		/* # of hard links */
	matlink_t	 links[MAT_LINKTAB_SIZE];	/* hard link entries */
	char		 from[MAT_PATH_MAX];	/* path to change from */
	char		 to[MAT_PATH_MAX];	/* path to change to */
	void		*re;		/* pointer to compiled regexp */
} mat_t;

int mat_init(mat_t */*mat*/, const char */*f*/, const char */*mode*/);
int mat_end(mat_t */*mat*/);

int mat_add(mat_t */*mat*/, const char */*f*/);
int mat_list(mat_t */*mat*/, const char */*pat*/, FILE */*fp*/);
int mat_verify(mat_t */*mat*/);
int mat_extract(mat_t */*mat*/, const char */*pat*/);
int mat_audit(mat_t */*mat*/, const char */*pat*/, FILE */*fp*/);

int mat_tar(int /*argc*/, char **/*argv*/);
int mat_vtar(char */*arg*/, ...);

int mat_pax(int /*argc*/, char **/*argv*/);
int mat_vpax(char */*arg*/, ...);

__END_DECLS

#endif
