/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "rs.h"

#ifndef PKG_VERSION
#define PKG_VERSION	20110309
#endif

#ifndef PKG_AUTHOR
#define PKG_AUTHOR	"Alistair Crooks <agc@netbsd.org>"
#endif

/* process a file - encoding or decoding */
static int
reedsolomon(rs_t *rs, uint8_t data, uint8_t parity, int encoding, char *in, char *fout)
{
	uint8_t	 buf[256];
	uint8_t	 out[256];
	size_t	 readsize;
	FILE	*fpout;
	FILE	*fpin;
	int	 outcc;
	int	 cc;
	int	 ok;

	if (in == NULL) {
		fpin = stdin;
	} else if ((fpin = fopen(in, "r")) == NULL) {
		(void) fprintf(stderr, "can't read '%s'\n", in);
		return 0;
	}
	if (fout == NULL) {
		fpout = stdout;
	} else if ((fpout = fopen(fout, "w")) == NULL) {
		(void) fprintf(stderr, "can't write '%s'\n", fout);
		if (fpin != stdin) {
			(void) fclose(fpin);
		}
		return 0;
	}
	if (encoding) {
		/* copy the encoding paramters to output */
		readsize = data;
		outcc = rs_put_header(rs, out, sizeof(out));
		if (write(fileno(fpout), out, outcc) != outcc) {
			(void) fprintf(stderr, "short write\n");
			return 0;
		}
	} else {
		/* read the encoding paramters from input */
		if (read(fileno(fpin), &rs->header, sizeof(rs->header)) != sizeof(rs->header)) {
			(void) fprintf(stderr, "short read\n");
			return 0;
		}
		rs_get_header(rs, &rs->header, sizeof(rs->header));
		readsize = rs->header.datac + rs->header.parityc;
	}
	for (ok = 1 ; ok && (cc = read(fileno(fpin), buf, readsize)) > 0 ; ) {
		if (encoding) {
			outcc = rs_encode(rs, buf, cc, out, sizeof(out));
		} else {
			outcc = rs_decode(rs, buf, cc, out, sizeof(out));
		}
		if (write(fileno(fpout), out, outcc) != outcc) {
			(void) fprintf(stderr, "short write\n");
			ok = 0;
		}
	}
	if (fpin != stdin) {
		fclose(fpin);
	}
	if (fpout != stdout) {
		fclose(fpout);
	}
	return ok;
}

int
main(int argc, char **argv)
{
	unsigned	 poly;
	uint8_t		 databytes;
	uint8_t		 parity;
	uint8_t		 data2;
	uint8_t		 par2;
	rs_t		 rs;
	char		*out;
	char		*s;
	int		 encoding;
	int		 i;

	databytes = RS_DEFAULT_DATA;
	parity = RS_DEFAULT_PARITY;
	poly = RS_DEFAULT_POLYNOMIAL;
	encoding = 1;
	out = NULL;
	while ((i = getopt(argc, argv, "P:b:do:p:v")) != -1) {
		switch(i) {
		case 'P':
			poly = (unsigned)strtol(optarg, NULL, 16);
			break;
		case 'd':
			encoding = 0;
			break;
		case 'o':
			out = optarg;
			break;
		case 'b':
		case 'p':
			for (s = optarg, data2 = 0 ; *s && *s != ',' ; s++) {
				data2 = (data2 * 10) + (*s - '0');
			}
			if (*s == ',') {
				for (s++, par2 = 0 ; *s ; s++) {
					par2 = (par2 * 10) + (*s - '0');
				}
				parity = data2 - par2;
				databytes = par2;
			}
			break;
		case 'v':
			printf("Version: %u\n", PKG_VERSION);
			printf("All praise, bug reports and chocolate to %s, please\n", PKG_AUTHOR);
			exit(EXIT_SUCCESS);
		default:
			break;
		}
	}
	(void) memset(&rs, 0x0, sizeof(rs));
	/* Initialization of the ECC library */
	if (!rs_init(&rs, poly, databytes, parity)) {
		exit(EXIT_FAILURE);
	}
	if (optind == argc) {
		reedsolomon(&rs, databytes, parity, encoding, NULL, out);
	} else {
		for (i = optind ; i < argc ; i++) {
			reedsolomon(&rs, databytes, parity, encoding, argv[i], out);
		}
	}
	exit(EXIT_SUCCESS);
}
