/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * Reed Solomon Encoder/Decoder
 * 
 * (c) Henry Minsky (hqm@ua.com), Universal Access 1991-1995
 */
#include <sys/types.h>
#include <sys/param.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rs.h"

/*****************************
 *
 *
 * Multiplication and Arithmetic on Galois Field GF(256)
 *
 * From Mee, Daniel, "Magnetic Recording, Volume III", Ch. 5 by Patel.
 *
 * (c) 1991 Henry Minsky
 *
 *
 ******************************/

enum {
	MAX_PARITY	= 256,	/* not strictly necessary, just useful */
	MAX_DEGREE	= 256,	/* not strictly necessary, just useful */
	MAX_GF8		= 256
};

static uint8_t	gfexp[MAX_GF8 * 2];
static uint8_t	gflog[MAX_GF8];
static uint8_t	gfinit;
 
#ifndef USE_ARG
#define USE_ARG(x)		/*LINTED*/(void)&(x)
#endif

#define RS_MAGIC		"rs1"

/* set up the gf8 tables */
static int
galois_inits(const unsigned poly)
{
	uint8_t	p[9];
	int	i;
	int	z;
	int	j;

	if (!gfinit) {
		(void) memset(p, 0x0, sizeof(p));
		gfexp[MAX_GF8 - 1] = gfexp[0] = p[1] = 1;
		gflog[0] = 0;		/* shouldn't log[0] be an error? */
		for (i = 1; i < MAX_GF8; i++) {
			for (p[0] = p[8], j = 8 ; j > 0 ; --j) {
				p[j] = (j > 1 && (poly & (1 << (j - 1)))) ? 
					p[j - 1] ^ p[0] : p[j - 1];
			}
			gfexp[i + MAX_GF8 - 1] = gfexp[i] =
				p[1] + (p[2] * 2) + (p[3] * 4) + (p[4] * 8) +
				(p[5] * 16) + (p[6] * 32) + (p[7] * 64) +
				(p[8] * 128);
		}
		for (i = 1; i < MAX_GF8; i++) {
			for (z = 0; z < MAX_GF8; z++) {
				if (gfexp[z] == i) {
					gflog[i] = z;
					break;
				}
			}
		}
		gfinit = 1;
		return 1;
	}
	return 0;
}

/* multiplication using logarithms */
static __inline uint8_t 
gfmult(uint8_t a, uint8_t b)
{
	return (a == 0 || b == 0) ? 0 : gfexp[gflog[a] + gflog[b]];
}

/* inversion using logarithms */
static __inline uint8_t 
gfinv(uint8_t elt)
{
	return (gfexp[MAX_GF8 - 1 - gflog[elt]]);
}

/***********************************************************************
 * Berlekamp-Peterson and Berlekamp-Massey Algorithms for error-location
 *
 * From Cain, Clark, "Error-Correction Coding For Digital Communications", pp. 205.
 *
 * This finds the coefficients of the error locator polynomial.
 *
 * The roots are then found by looking for the values of a^n
 * where evaluating the polynomial yields zero.
 *
 * Error correction is done using the error-evaluator equation  on pp 207.
 *
 * hqm@ai.mit.edu   Henry Minsky
 */

/* add a gf8 polynomial */
static void 
poly_add(uint8_t *dst, uint8_t *src, size_t size)
{
	size_t	i;

	for (i = 0; i < size; i++) {
		dst[i] ^= src[i];
	}
}

/* multiply by a constant a polynomial */
static void 
poly_scale(uint8_t *poly, uint8_t k, size_t size)
{
	size_t	i;

	for (i = 0; i < size; i++) {
		poly[i] = gfmult(k, poly[i]);
	}
}

/* zero out a polynomial */
static void 
poly_zero(uint8_t *poly, size_t size)
{
	(void) memset(poly, 0x0, size * sizeof(*poly));
}


/* multiply by z, i.e., shift right by 1 */
static void 
poly_mul_z(uint8_t *src, size_t size)
{
	size_t	i;

	for (i = size - 1; i > 0; i--) {
		src[i] = src[i - 1];
	}
	src[0] = 0;
}

/* multiply a GF8 polynomial */
static void
poly_mult(uint8_t *dst, uint8_t *p1, uint8_t *p2, size_t size)
{
	uint8_t	tmp1[MAX_DEGREE * 2];
	ssize_t	j;
	size_t	i;

	(void) memset(dst, 0x0, (size * 2) * sizeof(dst[0]));
	for (i = 0; i < size; i++) {
		(void) memset(&tmp1[size], 0x0, size * sizeof(tmp1[0]));
		/* scale tmp1 by p1[i] */
		for (j = 0; j < (ssize_t)size; j++) {
			tmp1[j] = gfmult(p2[j], p1[i]);
		}
		/* and mult (shift) tmp1 right by i */
		for (j = (size * 2) - 1; j >= (ssize_t)i; j--) {
			tmp1[j] = tmp1[j - i];
		}
		(void) memset(tmp1, 0x0, i * sizeof(tmp1[0]));
		/* add into partial product */
		for (j = 0; j < (ssize_t)(size * 2); j++) {
			dst[j] ^= tmp1[j];
		}
	}
}

/*
 * given Psi (called Lambda in Modified_Berlekamp_Massey) and syndrome,
 * compute the combined erasure/error evaluator polynomial as Psi*S mod z^4
 */
static void
compute_modified_omega(rs_t *rs)
{
	uint8_t	product[MAX_DEGREE * 2];
	int	i;

	poly_mult(product, rs->lambda, rs->syndrome, rs->header.parityc * 2);
	poly_zero(rs->omega, rs->header.parityc * 2);
	for (i = 0; i < rs->header.parityc; i++) {
		rs->omega[i] = product[i];
	}
}

/* gamma = product (1-z*a^Ij) for erasure locs Ij */
static void
init_gamma(rs_t *rs, uint8_t *gamma, size_t size)
{
	uint8_t	tmp[MAX_GF8];
	int	e;

	poly_zero(gamma, size);
	poly_zero(tmp, size);
	gamma[0] = 1;
	for (e = 0; e < rs->nerasures; e++) {
		(void) memcpy(tmp, gamma, size);
		poly_scale(tmp, gfexp[rs->erasurelocs[e]], size);
		poly_mul_z(tmp, size);
		poly_add(gamma, tmp, size);
	}
}

static uint8_t
compute_discrepancy(uint8_t *lambda, uint8_t *S, int L, int n)
{
	uint8_t	sum;
	int	i;

	for (sum = 0, i = 0; i <= L; i++) {
		sum ^= gfmult(lambda[i], S[n - i]);
	}
	return sum;
}

/*
 * From  Cain, Clark, "Error-Correction Coding For Digital Communications",
 * pp. 216.
 */
static void
Modified_Berlekamp_Massey(rs_t *rs)
{
	uint8_t	gamma[MAX_DEGREE];
	uint8_t	psi[MAX_DEGREE];
	uint8_t	psi2[MAX_DEGREE];
	uint8_t	D[MAX_DEGREE];
	uint8_t	d;
	int	L2;
	int	L;
	int	k;
	int	i;
	int	n;

	/* initialize Gamma, the erasure locator polynomial */
	init_gamma(rs, gamma, rs->header.parityc * 2);

	/* initialize to z */
	(void) memcpy(D, gamma, rs->header.parityc * 2);
	poly_mul_z(D, rs->header.parityc * 2);

	(void) memcpy(psi, gamma, rs->header.parityc * 2);
	k = -1;

	for (L = n = 0; n < rs->header.parityc; n++) {
		if ((d = compute_discrepancy(psi, rs->syndrome, L, n)) != 0) {
			/* psi2 = psi - d*D */
			for (i = 0; i < rs->header.parityc * 2; i++) {
				psi2[i] = psi[i] ^ gfmult(d, D[i]);
			}
			if (L < (n - k)) {
				L2 = n - k;
				k = n - L;
				/* D = poly_scale(psi, gfinv(d), rs->parityc * 2); */
				for (i = 0; i < rs->header.parityc * 2; i++) {
					D[i] = gfmult(psi[i], gfinv(d));
				}
				L = L2;
			}
			/* psi = psi2 */
			(void) memcpy(psi, psi2, rs->header.parityc * 2);
		}
		poly_mul_z(D, rs->header.parityc * 2);
	}
	(void) memcpy(rs->lambda, psi, rs->header.parityc * 2);
	compute_modified_omega(rs);
}

/* polynomial multiplication */

/*
 * Finds all the roots of an error-locator polynomial with coefficients
 * Lambda[j] by evaluating Lambda at successive values of alpha.
 * 
 * This can be tested with the decoder's equations case.
 */
static int
find_roots(rs_t *rs)
{
	uint8_t	sum;
	int	r;
	int	k;

	for (rs->nerrors = 0, r = 1; r < MAX_GF8; r++) {
		/* evaluate lambda at r */
		for (sum = 0, k = 0; k < rs->header.parityc + 1; k++) {
			sum ^= gfmult(gfexp[(k * r) % (MAX_GF8 - 1)], rs->lambda[k]);
		}
		if (sum == 0) {
			rs->errorlocs[rs->nerrors++] = (MAX_GF8 - 1 - r);
			if (rs->debug) {
				fprintf(stderr, "Root found at r = %d, (255-r) = %d\n", r, (MAX_GF8 - 1 - r));
			}
		}
	}
	return rs->nerrors;
}

/*
 * Create a generator polynomial for an n byte RS code. The coefficients are
 * returned in the genPoly arg. Make sure that the genPoly array which is
 * passed in is at least n+1 bytes long.
 */
static void
compute_genpoly(int nbytes, uint8_t *genpoly, size_t size)
{
	uint8_t	tp1[MAX_GF8];
	uint8_t	tp[MAX_GF8];
	int	i;

	/* multiply (x + a^n) for n = 1 to nbytes */
	poly_zero(tp1, size);
	tp1[0] = 1;
	for (i = 1; i <= nbytes; i++) {
		poly_zero(tp, size);
		tp[0] = gfexp[i];/* set up x+a^n */
		tp[1] = 1;
		poly_mult(genpoly, tp, tp1, size);
		(void) memcpy(tp1, genpoly, size);
	}
}

/*
 * Combined Erasure And Error Magnitude Computation
 * 
 * Pass in the codeword, its size in bytes, as well as an array of any known
 * erasure locations, along the number of these erasures.
 * 
 * Evaluate Omega(actually Psi)/Lambda' at the roots alpha^(-i) for error locs
 * i.
 * 
 * Returns 1 if everything ok, or 0 if an out-of-bounds error is found
 * 
 */
static int
correct_errors(rs_t *rs, uint8_t *in, size_t size)
{
	int	err;
	int	r;
	int	i;
	int	j;

	Modified_Berlekamp_Massey(rs);
	find_roots(rs);
	if ((rs->nerrors <= rs->header.parityc) && rs->nerrors > 0) {
		/* first check for illegal error locs */
		for (r = 0; r < rs->nerrors; r++) {
			if (rs->errorlocs[r] >= size) {
				if (rs->debug) {
					fprintf(stderr, "Error loc i=%u outside of codeword length %zu\n",
						rs->errorlocs[r], size);
				}
				return 0;
			}
		}
		for (r = 0; r < rs->nerrors; r++) {
			int             num, denom;

			i = rs->errorlocs[r];
			/* evaluate Omega at alpha^(-i) */

			num = 0;
			for (j = 0; j < rs->header.parityc * 2; j++) {
				num ^= gfmult(rs->omega[j], gfexp[((MAX_GF8 - 1 - i) * j) % (MAX_GF8 - 1)]);
			}
			/*
			 * evaluate Lambda' (derivative) at alpha^(-i) ; all
			 * odd powers disappear
			 */
			denom = 0;
			for (j = 1; j < rs->header.parityc * 2; j += 2) {
				denom ^= gfmult(rs->lambda[j], gfexp[((MAX_GF8 - 1 - i) * (j - 1)) % (MAX_GF8 - 1)]);
			}
			err = gfmult(num, gfinv(denom));
			if (rs->debug) {
				fprintf(stderr, "Error magnitude %#x at loc %zu\n", err, size - i);
			}
			in[size - i - 1] ^= err;
		}
		return 1;
	}
	if (rs->debug && rs->nerrors) {
		fprintf(stderr, "Uncorrectable codeword\n");
	}
	return 0;
}


/* Check if the syndrome is zero - return 1 if syndrome is nonzero */
static int
nonzero_syndrome(rs_t *rs)
{
	int	i;

	for (i = 0; i < rs->header.parityc; i++) {
		if (rs->syndrome[i] != 0) {
			return 1;
		}
	}
	return 0;
}

/*
 * Simulate a LFSR with generator polynomial for n byte RS code. Pass in a
 * pointer to the data array, and amount of data.
 * 
 * The parity bytes are deposited into parity[], and the whole message and
 * parity are copied to dest to make a codeword.
 * 
 */
static ssize_t
encode_chunk(rs_t *rs, const uint8_t *msg, size_t size, uint8_t *dst, int include)
{
	uint8_t	lfsr[MAX_PARITY + 1];
	uint8_t	dbyte;
	size_t	off;
	size_t	i;
	int	j;

	(void) memset(lfsr, 0x0, sizeof(lfsr));
	for (i = 0; i < size; i++) {
		dbyte = msg[i] ^ lfsr[rs->header.parityc - 1];
		for (j = rs->header.parityc - 1; j > 0; j--) {
			lfsr[j] = lfsr[j - 1] ^ gfmult(rs->genPoly[j], dbyte);
		}
		lfsr[0] = gfmult(rs->genPoly[0], dbyte);
	}
	if (include) {
		(void) memcpy(dst, msg, size);
		off = size;
	} else {
		off = 0;
	}
	for (i = 0; i < rs->header.parityc; i++) {
		dst[off + i] = lfsr[rs->header.parityc - 1 - i];
	}
	return off + rs->header.parityc;
}

/* decode one chunk of the input */
static ssize_t
decode_chunk(rs_t *rs, const uint8_t *in, size_t size, uint8_t *dst)
{
	uint8_t	sum;
	size_t	i;
	int	j;

	(void) memcpy(dst, in, size);
	for (j = 0; j < rs->header.parityc; j++) {
		for (sum = 0, i = 0; i < size; i++) {
			sum = dst[i] ^ gfmult(gfexp[j + 1], sum);
		}
		rs->syndrome[j] = sum;
	}
	if (nonzero_syndrome(rs) && !correct_errors(rs, dst, size)) {
		return -1;
	}
	return size - rs->header.parityc;
}

#define BSWAP32(x)	((((x) & 0xff) << 24) |				\
			 (((x) & 0xff00) << 8) |			\
			 (((x) & 0xff0000) >> 8) |			\
			 (((x) & 0xff000000) >> 24))

/* byte-swap 32 bits, if necessary, to get little-endian */
static uint32_t
le32(uint32_t in)
{
	int	indian = 1;

	return (*(char *)(void *)&indian) ? in : BSWAP32(in);
}

/****************************************************************************/
/* exported functions from here on in */
/****************************************************************************/

/* Initialize lookup tables, polynomials, etc. */
int
rs_init(rs_t *rs, const unsigned poly, const uint8_t databytes, const uint8_t paritybytes)
{
	if (paritybytes > databytes) {
		(void) fprintf(stderr, "paritybytes %u larger than data bytes %u\n",
			paritybytes, databytes);
		return 0;
	}
	if (paritybytes + databytes > 256) {
		(void) fprintf(stderr, "paritybytes %u and data bytes %u > 256\n",
			paritybytes, databytes);
		return 0;
	}
	(void) memset(rs, 0x0, sizeof(*rs));
	rs->header.poly = poly;
	rs->header.datac = databytes;
	rs->header.parityc = paritybytes;
	/* Initialize the galois field arithmetic tables */
	galois_inits(poly);
	/* Compute the encoder generator polynomial */
	compute_genpoly(rs->header.parityc, rs->genPoly, rs->header.parityc * 2);
	return 1;
}

/* Reed Solomon Encoder */
ssize_t
rs_encode(rs_t *rs, const void *inp, size_t size, uint8_t *out, size_t outsize)
{
	const uint8_t	*in = (const uint8_t *)inp;
	const int	 includemsg = 1;
	size_t	 	 outcc;
	size_t		 incc;
	size_t		 chunk;

	if (!rs->header.poly) {
		rs_init(rs, RS_DEFAULT_POLYNOMIAL, RS_DEFAULT_DATA, RS_DEFAULT_PARITY);
	}
	for (outcc = 0, incc = 0 ; incc < size && outcc < outsize ; incc += chunk) {
		chunk = MIN(size - incc, rs->header.datac);
		outcc += encode_chunk(rs, &in[incc], chunk, &out[outcc], includemsg);
	}
	return (ssize_t)outcc;
}

/* Reed Solomon Decoder */
ssize_t
rs_decode(rs_t *rs, const void *inp, size_t size, uint8_t *out, size_t outsize)
{
	const uint8_t	*in = (const uint8_t *)inp;
	ssize_t		 outcc;
	ssize_t		 cc;
	size_t		 incc;
	size_t		 chunk;

	for (outcc = 0, incc = 0 ; incc < size && outcc < (ssize_t)outsize ; incc += chunk) {
		chunk = MIN(size - incc, rs->header.datac + rs->header.parityc);
		if ((cc = decode_chunk(rs, &in[incc], chunk, &out[outcc])) < 0) {
			break;
		}
		outcc += cc;
	}
	return outcc;
}

/* get a header from an area of memory */
ssize_t
rs_get_header(rs_t *rs, const void *in, size_t insize)
{
	rs_header_t	head;

	USE_ARG(insize);
	(void) memcpy(&head, in, sizeof(head));
	head.poly = le32(head.poly);
	(void) memcpy(&rs->header, &head, sizeof(rs->header));
	return sizeof(rs->header);
}

/* put a header into an area of memory */
ssize_t
rs_put_header(rs_t *rs, void *out, size_t outsize)
{
	rs_header_t	head;

	USE_ARG(outsize);
	(void) memcpy(&head, &rs->header, sizeof(head));
	(void) snprintf(head.magic, sizeof(head.magic), "%s", RS_MAGIC);
	head.poly = le32(head.poly);
	(void) memcpy(out, &head, sizeof(head));
	return sizeof(head);
}

/* Reed Solomon Encoder */
ssize_t
rs_parity(rs_t *rs, const void *inp, size_t size, uint8_t *out, size_t outsize)
{
	const uint8_t	*in = (const uint8_t *)inp;
	const int	 nomsg = 0;
	size_t	 	 outcc;
	size_t		 incc;
	size_t		 chunk;

	if (!rs->header.poly) {
		rs_init(rs, RS_DEFAULT_POLYNOMIAL, RS_DEFAULT_DATA, RS_DEFAULT_PARITY);
	}
	for (outcc = 0, incc = 0 ; incc < size && outcc < outsize ; incc += chunk) {
		chunk = MIN(size - incc, rs->header.datac);
		outcc += encode_chunk(rs, &in[incc], chunk, &out[outcc], nomsg);
	}
	return (ssize_t)outcc;
}
