/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/* Reed Solomon Coding for glyphs
 * 
 * (c) Henry Minsky (hqm@ua.com), Universal Access Inc.  (1991-1996)
 *     
 *
 */
#ifndef RS_H_
#define RS_H_	20110301

#include <sys/types.h>

#include <inttypes.h>

#ifndef __BEGIN_DECLS
#  if defined(__cplusplus)
#  define __BEGIN_DECLS           extern "C" {
#  define __END_DECLS             }
#  else
#  define __BEGIN_DECLS
#  define __END_DECLS
#  endif
#endif

__BEGIN_DECLS

typedef struct rs_header_t {
	char		magic[4];		/* magic number */
	uint32_t	poly;			/* polynomial used */
	uint8_t		datac;			/* # of data bytes per packet */
	uint8_t		parityc;		/* # of parity bytes per packet */
	uint8_t		interleave;		/* interleave signature */
	uint8_t		convolution;		/* convolution factor */
} rs_header_t;

typedef struct rs_t {
	uint8_t		syndrome[256]; 		/* Decoder syndrome bytes */
	uint8_t		genPoly[256 * 2];	/* generator polynomial */
	int		debug;			/* verbosity switch */
	int		nerasures;
	int		erasurelocs[256];
	uint8_t		lambda[256];		/* error locator polynomial */
	uint8_t		omega[256];		/* error evaluator polynomial */
	/* error locations found using Chien's search */
	int		nerrors;
	uint8_t		errorlocs[256];
	rs_header_t	header;			/* header for transportation */
} rs_t;

enum {
	/*
	 * This is one of 14 irreducible polynomials of degree 8 and
	 * cycle length 255.  (Ch 5, pp.  275, Magnetic Recording) The
	 * high order 1 bit is implicit
	 */
	/* x^8 + x^4 + x^3 + x^2 + 1 */
	RS_DEFAULT_POLYNOMIAL	= 0x11d,
	RS_DEFAULT_DATA		= 224,
	RS_DEFAULT_PARITY	= 32
};

/* initialisations */
int rs_init(rs_t *, const unsigned, const uint8_t, const uint8_t);

/* Reed Solomon encode/decode routines */
ssize_t rs_decode(rs_t *, const void *, size_t, uint8_t *, size_t);
ssize_t rs_encode(rs_t *, const void *, size_t, uint8_t *, size_t);
ssize_t rs_parity(rs_t *, const void *, size_t, uint8_t *, size_t);

/* Reed Solomon header routines */
ssize_t rs_get_header(rs_t *, const void *, size_t);
ssize_t rs_put_header(rs_t *, void *, size_t);

__END_DECLS

#endif
