/*-
 * Copyright (c) 2010,2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * Copyright (c) 2010 Juan Camilo Corena
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/param.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "threshold.h"
#include "internal.h"

enum {
	ALLOCATED	= 1,
	MMAPPED		= 2
};

/* information dispersal algorithm matrix */
typedef struct ida_t {
	uint8_t		*matrix;	/* encoding/decoding matrix */
} ida_t;

/* static invariants over GF(256) */
static uint8_t	gfmuls[THRESH_MAX_SHARES][THRESH_MAX_SHARES];
static uint8_t	gfinvs[THRESH_MAX_SHARES];
static unsigned	gfinit;

#define SUBSCRIPT(r,c,rowlen)	(((r) * (rowlen)) + (c))

#define IDA_MAGIC	"ida"

#define BSWAP64(x)      ((((x) & 0xffULL) << 56) |			\
                         (((x) & 0xff00ULL) << 40) |			\
                         (((x) & 0xff0000ULL) << 24) |			\
                         (((x) & 0xff000000ULL) << 8) |			\
                         (((x) & 0xff00000000ULL) >> 8) |		\
                         (((x) & 0xff0000000000ULL) >> 24) |		\
                         (((x) & 0xff000000000000ULL) >> 40) |		\
                         (((x) & 0xff00000000000000ULL) >> 56))

/* add 2 gf8 numbers inline */
static __inline uint8_t
gfadd(uint8_t a, uint8_t b)
{
	return a ^ b;
}

/* multiply 2 gf8 numbers inline */
static __inline uint8_t
gfmul(uint8_t a, uint8_t b)
{
	unsigned	p;

	for (p = 0 ; b != 0 ; b >>= 1) {
		p ^= (a * (b & 1));
		a = ((((unsigned)a >> 7) == 1) ? ((a << 1) ^ 0x1b) : (a << 1)) & 0xff;
	}
	return p & 0xff;
}

/* find the inverse of the vandermonde matrix */
static void
vandermondeinverse(ida_t *ida, unsigned threshold, unsigned *coeff)
{
	unsigned	i;
	unsigned	j;
	uint8_t		ak1[THRESH_MAX_SHARES + 1];
	uint8_t		ak2[THRESH_MAX_SHARES + 1];
	int		p;
	int		q;
	int		r;

	(void) memset(ak1, 0x0, sizeof(ak1));
	(void) memset(ak2, 0x0, sizeof(ak2));
	ak1[0] = coeff[0];
	ak1[1] = 1;
	for (i = 1 ; i < threshold ; i++) {
		for (j = 0 ; j <= i + 1 ; j++) {
			ak2[j] = gfadd((j > 0) ? ak1[j - 1] : 0,
				gfmuls[coeff[i]][(j <= i) ? ak1[j] : 0]);
		}
		(void) memcpy(ak1, ak2, (i + 2));
	}
	for (i = 0 ; i < threshold ; i++) {
		p = coeff[i];
		q = ida->matrix[SUBSCRIPT(threshold - 1, i, threshold)] = 1;
		r = ak1[1];
		for (j = 1 ; j < threshold ; j++, p = gfmuls[p][coeff[i]]) {
			q = ida->matrix[SUBSCRIPT(threshold - 1 - j, i, threshold)] =
				gfadd(gfmuls[coeff[i]][q], ak1[threshold - j]);
			if ((j & 1) == 0) {
				r = gfadd(r, gfmuls[p][ak1[j + 1]]);
			}
		}
		for (j = 0 ; j < threshold ; j++) {
			ida->matrix[SUBSCRIPT(j, i, threshold)] =
				gfmuls[ida->matrix[SUBSCRIPT(j, i, threshold)]][gfinvs[r]];
		}
	}
}

/* initialise the gf8 arrays */
static void
gf_initialisations(void)
{
	int	i;
	int	j;

	if (!gfinit) {
		gfinit = 1;
		for (i = 0 ; i < THRESH_MAX_SHARES ; i++) {
			for (j = 0 ; j < THRESH_MAX_SHARES ; j++) {
				gfmuls[i][j] = gfmul(i, j);
			}
		}
		for (i = 1; i < THRESH_MAX_SHARES ; i++) {
			if (gfinvs[i] == 0) {
				for (j = 1 ; j < THRESH_MAX_SHARES ; j++) {
					if (gfmuls[i][j] == 1) {
						gfinvs[i] = j;
						gfinvs[j] = i;
					}
				}
			}
		}
	}
}

/*************************************************************************/

/* split original data into shares */
int
ida_split(threshold_t *thresh)
{
	thresh_head_t	 head;
	thresh_str_t	*mem;
	unsigned	 i;
	unsigned	 j;
	uint8_t		 ch;
	uint8_t		 arr[THRESH_MAX_SHARES];
	size_t		 outsize;
	size_t		 insize;
	ida_t		*ida;

	ida = (ida_t *)thresh->handle;
	mem = &thresh->shares[THRESH_MAX_SHARES];
	thresh_fmt_header(&head, IDA_MAGIC, thresh->threshold, thresh->sharesc, mem->io.size);
	outsize = (mem->io.size / thresh->threshold) + sizeof(head) + 10;
	for (i = 0 ; i < thresh->sharesc ; i++) {
		if (thresh->shares[i].io.base == NULL &&
		    (thresh->shares[i].io.base = calloc(outsize, 1)) == NULL) {
			(void) fprintf(stderr, "Bad alloc\n");
			return 0;
		}
		head.coeff = (thresh->threshold == 1) ? 0 : ida->matrix[SUBSCRIPT(i, 1, thresh->threshold)];
		(void) memcpy(thresh->shares[i].io.base, &head, sizeof(head));
		thresh->shares[i].c = sizeof(head);
	}
	for (insize = 0 ; insize < mem->io.size ; ) {
		(void) memset(arr, 0x0, thresh->sharesc);
		for (j = 0 ; j < thresh->threshold ; j++) {
			ch = mem->io.base[insize++];
			for (i = 0 ; i < thresh->sharesc ; i++) {
				arr[i] = gfadd(arr[i], gfmuls[ida->matrix[SUBSCRIPT(i, j, thresh->threshold)]][ch]);
			}
		}
		for (i = 0 ; i < thresh->sharesc ; i++) {
			thresh->shares[i].io.base[thresh->shares[i].c++] = arr[i];
		}
	}
	for (i = 0 ; i < thresh->sharesc ; i++) {
		thresh->shares[i].io.size = thresh->shares[i].c;
	}
	return 1;
}

/* combine shares into output */
int
ida_combine(threshold_t *thresh)
{
	thresh_head_t	 heads[THRESH_MAX_SHARES];
	thresh_str_t	*out;
	unsigned	 coeff[THRESH_MAX_SHARES];
	unsigned	 i;
	unsigned	 j;
	uint8_t		 arr[THRESH_MAX_SHARES];
	uint8_t		 ch;
	ida_t		*ida;
	int		 cc;

	ida = (ida_t *)thresh->handle;
	(void) memset(coeff, 0x0, sizeof(coeff));
	for (i = 0 ; i < thresh->availc ; i++) {
		thresh_get_header(&heads[i], thresh->shares[i].io.base);
		if (!thresh_sane_header(&heads[i], &heads[0])) {
			(void) fprintf(stderr, "insane ida header %d\n", i);
			return 0;
		}
		thresh->shares[i].c = sizeof(heads[i]);
		coeff[i] = heads[i].coeff;
	}
	if (thresh->availc < heads[0].thresh) {
		(void) fprintf(stderr, "avail shares %u, threshold %u\n",
			thresh->availc, heads[0].thresh);
		return 0;
	}
	ida_init(thresh, (unsigned)heads[0].thresh, (unsigned)heads[0].sharesc);
	out = &thresh->shares[THRESH_MAX_SHARES];
	if (out->io.base == NULL &&
	    (out->io.base = calloc(1, heads[0].size)) == NULL) {
		(void) fprintf(stderr, "can't allocate space for output\n");
		return 0;
	}
	thresh->mapped[THRESH_MAX_SHARES] = ALLOCATED;
	vandermondeinverse(ida, thresh->threshold, coeff);
	(void) memset(arr, 0x0, sizeof(arr));
	for (out->c = 0 ; out->c < heads[0].size ; ) {
		(void) memset(arr, 0x0, thresh->threshold);
		for (j = 0 ; j < thresh->threshold ; j++) {
			ch = thresh->shares[j].io.base[thresh->shares[j].c++];
			for (i = 0 ; i < thresh->threshold ; i++) {
				arr[i] = gfadd(arr[i], gfmuls[ida->matrix[SUBSCRIPT(i, j, thresh->threshold)]][ch]);
			}
		}
		cc = MIN(thresh->threshold, (heads[0].size - out->c));
		(void) memcpy(&out->io.base[out->c], arr, cc);
		out->c += cc;
	}
	out->io.size = out->c;
	for (i = 0 ; i < thresh->availc ; i++) {
		thresh->shares[i].io.size = thresh->shares[i].c;
	}
	return 1;
}

/* initialise */
int
ida_init(threshold_t *threshold, unsigned thresh, unsigned sharesc)
{
	unsigned	i;
	unsigned	j;
	ida_t		*ida;

	if ((ida = calloc(1, sizeof(ida_t))) == NULL) {
		(void) fprintf(stderr, "can't allocate ida\n");
		return 0;
	}
	threshold->handle = ida;
	threshold->sharesc = sharesc;
	threshold->threshold = thresh;
	if ((ida->matrix = calloc(sharesc, sharesc)) == NULL) {
		(void) fprintf(stderr, "ida_init: bad calloc\n");
		return 0;
	}
	gf_initialisations();
	for (i = 0 ; i < sharesc ; i++) {
		for (j = 0 ; j < thresh ; j++) {
			ida->matrix[SUBSCRIPT(i, j, thresh)] = (j == 0) ? 1 :
				gfmuls[ida->matrix[SUBSCRIPT(i, j - 1, thresh)]][i];
		}
	}
	return 1;
}

/* finalisation routine */
int
ida_end(threshold_t *threshold)
{
	ida_t	*ida;

	if ((ida = threshold->handle) != NULL) {
		if (ida->matrix) {
			free(ida->matrix);
		}
		(void) memset(ida, 0x0, sizeof(*ida));
	}
	return 1;
}
