/*-
 * Copyright (c) 2010,2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <codecs.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "threshold.h"
#include "internal.h"

/* our own small raid structure */
typedef struct raid_t {
	unsigned	raidlevel;	/* numeric raid number */
	unsigned	disks;		/* # of disks */
	unsigned	blocksize;	/* size of a block of data */
} raid_t;

/*
group of blocks as input
a1, a2, a3, ar

b1, b2, br, b3
c1, cr, c2, c3
dr, d1, d2, d3

# disks == 4, stripe == (block#/disks), parity pos == (stripe + 1) * (disks - 1)
i0  i1  i2  P1	=> blocks(0, 1, 2) -> P1 	(P1 == o3)
i3  i4  P2  i5	=> blocks(3, 4, 5) -> P2	(P2 == o6)
i6  P3  i7  i8	=> blocks(6, 7, 8) -> P3	(P3 == o9)
P4  i9  i10 i11	=> blocks(9, 10, 11) -> P4	(P4 == o12)

*/

enum {
	DEFAULT_BLOCKSIZE	= 512
};

/* calculate the XOR parity for the block */
/* we use 32-bit operations for efficiency */
static void
calcparity(const void *in8, void *parity, unsigned blocksize)
{
	const uint32_t	*in32 = (const uint32_t *)in8;
	uint32_t	*parity32 = (uint32_t *)parity;
	unsigned	 i;

	for (i = 0 ; i < blocksize / sizeof(uint32_t) ; i++) {
		parity32[i] ^= in32[i];
	}
}

/* this gives the stripe number */
#define STRIPE(blk, dsk)	(blk/dsk)

/* return 1 if this is a parity block */
static int
isparity(const unsigned raid, unsigned blk, unsigned dsk)
{
	switch(raid) {
	case 5:
		return blk == (STRIPE(blk, dsk) + 1) * ((dsk) - 1);
	case 4:
		return blk % (dsk - 1) == 0;
	default:
		(void) fprintf(stderr, "raid%u not supported\n", raid);
		return 0;
	}
}

/* shares == # of disks */
static int
raid_assemble(thresh_str_t *inputv, raid_t *raid, int repair, thresh_str_t *str)
{
	unsigned	i;
	unsigned	o;
	unsigned	d;
	unsigned	b;

	if (repair >= 0) {
		/* recover via xor the lost disk */
		for (i = 0, b = 0 ; i < inputv[repair].io.size / raid->blocksize ; i += raid->disks, b += raid->blocksize) {
			for (d = 0 ; d < raid->disks ; d++) {
				if ((int)d != repair) {
					calcparity(&inputv[d].io.base[b], 
						&inputv[repair].io.base[b], raid->blocksize);
				}
			}
		}
	}
	/* recover into output */
	for (i = 0, o = 0 ; i < inputv[repair].io.size / raid->blocksize ; i += raid->disks - 1) {
		for (d = 0 ; d < raid->disks ; o++, d++) {
			if (!isparity(raid->raidlevel, o, raid->disks)) {
				(void) memcpy(&str->io.base[str->c],
					&inputv[d].io.base[inputv[d].c],
					raid->blocksize);
				str->c += raid->blocksize;
			}
			inputv[d].c += raid->blocksize;
		}
	}
	return 1;
}

/* check blocksize and disk numbers are sane */
static int
check_values(unsigned blocksize, unsigned disks)
{
	if (disks < 2 || disks >= 256 - 1) {
		(void) fprintf(stderr, "Invalid number of disks (%d).\n", disks);
		return 0;
	}
	if (blocksize % 2 != 0 || blocksize > 32768) {
		(void) fprintf(stderr, "Invalid block size %d\n", blocksize);
		return 0;
	}
	return 1;
}

/*****************************************************************************/
/* external functions from here */
/*****************************************************************************/

int
raid_init(threshold_t *thresh, unsigned raidlevel, unsigned blocksize, unsigned disks)
{
	raid_t	*raid;

	if ((raid = calloc(1, sizeof(*raid))) == NULL) {
		fprintf(stderr, "raid_init: unable to allocate space\n");
		return 0;
	}
	thresh->handle = raid;
	raid->raidlevel = raidlevel;
	raid->disks = disks;
	raid->blocksize = blocksize;
	return 1;
}

/* split memory into different disks, rotate parity */
int
raid_split_memory(threshold_t *thresh, const void *input, unsigned insize, unsigned threshold, unsigned sharesc)
{
	const uint8_t	*in;
	unsigned	 blocksize;
	unsigned	 blockc;
	unsigned	 disks;
	unsigned	 i;
	unsigned	 d;
	unsigned	 o;
	uint8_t		*parity;
	raid_t		*raid;

	in = (const uint8_t *)input;
	raid = thresh->handle;
	blocksize = threshold;	/* overloaded here */
	blockc = (insize / blocksize) + 1;
	if ((parity = calloc(1, blocksize)) == NULL) {
		(void) fprintf(stderr, "Bad alloc\n");
		return 0;
	}
	disks = sharesc;
	thresh->sharesc = (unsigned)disks;
	for (i = 0 ; i < disks ; i++) {
		thresh->shares[i].io.size = blockc * blocksize;
		thresh->shares[i].c = 0;
		if ((thresh->shares[i].io.base = calloc(1, thresh->shares[i].io.size)) == NULL) {
			free(parity);
			(void) fprintf(stderr, "Bad alloc\n");
			return 0;
		}
	}
	/* the main loop */
	for (i = 0, o = 0 ; i < insize / blocksize ; i += disks - 1) {
		/* calculate parity from the disks in the stripe */
		(void) memset(parity, 0x0, blocksize);
		for (d = 0 ; d < disks - 1 ; d++) {
			calcparity(&in[(i + d) * blocksize], parity, blocksize);
		}
		/* write to output either copy of input, or parity */
		for (d = 0 ; d < disks ; o++, d++) {
			(void) memcpy(&thresh->shares[d].io.base[thresh->shares[d].c],
				isparity(raid->raidlevel, o, disks) ? parity :
					&in[(i + d) * blocksize],
				blocksize);
			thresh->shares[d].c += blocksize;
		}
	}
	free(parity);
	return 1;
}

int
raid_split(threshold_t *thresh)
{
	thresh_str_t	*mem;

	mem = &thresh->shares[THRESH_MAX_SHARES];
	return raid_split_memory(thresh, mem->io.base, mem->io.size, thresh->threshold, thresh->sharesc);
}

/* split the file into shares */
int 
raid_split_file(threshold_t *thresh, const char *f, unsigned blocksize, unsigned disks)
{
	struct stat	 st;
	size_t		 bytes;
	char		*mem;
	FILE		*fp;
	int		 ret;

	if (!check_values(blocksize, disks)) {
		return 0;
	}
	if ((fp = fopen(f, "r")) == NULL) {
		(void) fprintf(stderr, "Unable to open file %s\n", f);
		return 0;
	}
	(void) fstat(fileno(fp), &st);
	bytes = (size_t)st.st_size;
	if ((mem = mmap(NULL, bytes, PROT_READ, MAP_PRIVATE, fileno(fp), 0)) == MAP_FAILED) {
		(void) fprintf(stderr, "raid_split: can't mmap file '%s'\n", f);
		(void) fclose(fp);
		return 0;
	}
	/* Do the work */
	ret = raid_split_memory(thresh, mem, bytes, blocksize, disks);
	(void) fclose(fp);
	(void) munmap(mem, bytes);
	return ret;
}

/* reconstruct the shares */
int 
raid_join_file(threshold_t *thresh, char **files, unsigned disks, thresh_str_t *str)
{
	thresh_str_t	*inputv;
	struct stat	 st;
	unsigned	 i;
	raid_t		*raid;
	FILE		*f_in[THRESH_MAX_SHARES];
	int		 repair;
	int		 good;
	int		 ret;

	raid = thresh->handle;
	if (raid->blocksize == 0) {
		raid->blocksize = DEFAULT_BLOCKSIZE;
	}
	repair = -1;
	good = -1;
	/* filename list */
	if ((inputv = calloc((unsigned)disks, sizeof(*inputv))) == NULL) {
		fprintf(stderr, "bad alloc 0\n");
		return 0;
	}
	/* Open input files */
	for (i = 0; i < disks; i++) {
		if (files[i] == NULL) {
			repair = i;
			continue;
		} else if (good < 0) {
			good = i;
		}
		if ((f_in[i] = fopen(files[i], "r")) == NULL) {
			fprintf(stderr, "Unable to open %d file %s\n", i, files[i]);
			return 0;
		}
		(void) fstat(fileno(f_in[i]), &st);
		inputv[i].io.size = (unsigned)st.st_size;
		inputv[i].c = 0;
		inputv[i].io.base = mmap(NULL, inputv[i].io.size, PROT_READ, MAP_PRIVATE, fileno(f_in[i]), 0);
	}
	if (good < 0) {
		fprintf(stderr, "No input files\n");
		return 0;
	}
	inputv[repair].io.size = inputv[good].io.size;
	if ((inputv[repair].io.base = calloc(1, inputv[repair].io.size)) == NULL) {
		fprintf(stderr, "can't allocate repair space\n");
		return 0;
	}
	(void) memset(str, 0x0, sizeof(*str));
	if ((str->io.base = calloc(1, str->io.size = (inputv[0].io.size * (disks - 1)))) == NULL) {
		(void) fprintf(stderr, "raid_join_files: bad alloc %zu\n", str->io.size);
		ret = 0;
	} else {
		/* Do the work */
		ret = raid_assemble(inputv, raid, repair, str);
	}
	/* Clean up and exit */
	for (i = 0; i < disks; i++) {
		if (files[i] == NULL) {
			free(inputv[i].io.base);
		} else {
			fclose(f_in[i]);
			munmap(inputv[i].io.base, inputv[i].io.size);
		}
	}
	free(inputv);
	return ret;
}

/* reconstruct the shares from memory */
int 
raid_join_memory(threshold_t *thresh, thresh_str_t *inputv, unsigned bytes, unsigned disks, thresh_str_t *str)
{
	unsigned	 i;
	raid_t		*raid;
	int		 repair;

	raid = thresh->handle;
	if (raid->blocksize == 0) {
		raid->blocksize = DEFAULT_BLOCKSIZE;
	}
	(void) memset(str, 0x0, sizeof(*str));
	if ((str->io.base = calloc(1, str->io.size = bytes)) == NULL) {
		(void) fprintf(stderr, "Bad alloc %u bytes\n", bytes);
		return 0;
	}
	for (repair = -1, i = 0 ; i < disks ; i++) {
		if (inputv[i].io.base == NULL) {
			inputv[i].io.size = bytes;
			inputv[i].io.base = calloc(1, bytes);
			repair = i;
		}
		inputv[i].c = 0;
	}
	/* Do the work */
	return raid_assemble(inputv, raid, repair, str);
}

/* finish up */
int
raid_end(threshold_t *thresh)
{
	if (thresh->handle) {
		free(thresh->handle);
	}
	return 1;
}

#define RAID4_MAGIC		"r4"
#define RAID5_MAGIC		"r5"

/* read a header from the share, and return any info desired */
int
raid_read_header(const char *mem, uint8_t *sharesc, uint8_t *threshold, uint8_t *order)
{
	if (memcmp(mem, RAID5_MAGIC, strlen(RAID5_MAGIC)) == 0) {
		if (threshold) {
			*threshold = mem[strlen(RAID5_MAGIC)];
		}
		if (sharesc) {
			*sharesc = mem[strlen(RAID5_MAGIC) + 1];
		}
		if (order) {
			*order = mem[strlen(RAID5_MAGIC) + 2];
		}
		return 5;
	}
	if (memcmp(mem, RAID4_MAGIC, strlen(RAID4_MAGIC)) == 0) {
		if (threshold) {
			*threshold = mem[strlen(RAID4_MAGIC)];
		}
		if (sharesc) {
			*sharesc = mem[strlen(RAID4_MAGIC) + 1];
		}
		if (order) {
			*order = mem[strlen(RAID4_MAGIC) + 2];
		}
		return 4;
	}
	return 0;
}

#ifdef RAID5_MAIN
int
main(int argc, char **argv)
{
	thresh_str_t	 repair;
	threshold_t	 thresh;
	codecs_t	 codecs;
	unsigned	 blocksize;
	char		*mem;
	char		 buf[12000];
	int		 size;
	int		 cc;
	int		 i;

	(void) memset(&thresh, 0x0, sizeof(thresh));
	(void) memset(&codecs, 0x0, sizeof(codecs));
	blocksize = DEFAULT_BLOCKSIZE;
	size = 677891;
	mem = calloc(1, size);
	for (i = 0 ; i < size ; i += cc) {
		cc = snprintf(&mem[i], size - i, "The quick brown fox jumps over the %d liquor jugs\n", i);
	}
	raid_split_memory(&thresh, mem, i, blocksize, 4);
	for (i = 0 ; i < 13 ; i++) {
		cc = codecs_transform(&codecs,
			(const char *)&thresh.shares[i % 4].v[(i / 4) * blocksize],
			blocksize, "hexdump", buf, sizeof(buf));
		printf("block %d:\n%.*s\n", i, cc, buf);
	}
	(void) memset(&thresh.shares[2], 0x0, sizeof(thresh.shares[2]));
	(void) memset(&repair, 0x0, sizeof(repair));
	raid_join_memory(&thresh, thresh.shares, thresh.shares[0].io.size, 4, &repair);
	cc = codecs_transform(&codecs,
		(const char *)repair.v,
		12000, "hexdump", buf, sizeof(buf));
	printf("repair block:\n%.*s\n", cc, buf);
	exit(EXIT_SUCCESS);
}
#endif
