/*-
 * Copyright (c) 2011 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "internal.h"
#include "threshold.h"

enum {
	ALLOCATED	= 1,
	MMAPPED		= 2
};

/* text mapping */
typedef struct map_t {
	const char	*s;	/* string */
	const int	 val;	/* related value */
} map_t;

static const map_t	algnames[] = {
	{	"ida",		THRESH_IDA	},
	{	"rabin",	THRESH_IDA	},
	{	"vdm",		THRESH_IDA	},
	{	"fec",		THRESH_IDA	},
	{	"raidnm",	THRESH_IDA	},
	{	"raid4",	THRESH_RAID4	},
	{	"r4",		THRESH_RAID4	},
	{	"raid5",	THRESH_RAID5	},
	{	"r5",		THRESH_RAID5	},
	{	NULL,		0 }
};

/* find a text mapping in an array */
static const map_t *
findmap(const map_t *array, const char *s)
{
	const map_t	*map;

	for (map = array ; map->s ; map++) {
		if (strcasecmp(map->s, s) == 0) {
			return map;
		}
	}
	return NULL;
}

#define BSWAP64(x)      ((((x) & 0xffULL) << 56) |			\
                         (((x) & 0xff00ULL) << 40) |			\
                         (((x) & 0xff0000ULL) << 24) |			\
                         (((x) & 0xff000000ULL) << 8) |			\
                         (((x) & 0xff00000000ULL) >> 8) |		\
                         (((x) & 0xff0000000000ULL) >> 24) |		\
                         (((x) & 0xff000000000000ULL) >> 40) |		\
                         (((x) & 0xff00000000000000ULL) >> 56))

/* do 64bit integers in little-endian order */
static void
swapheader(thresh_head_t *head)
{
        int     indian = 1;

        head->size = (*(char *)(void *)&indian) ? head->size : (uint64_t)BSWAP64(head->size);
}

/**************************************************************************/

/* check we have a sane header */
int
thresh_sane_header(thresh_head_t *head, thresh_head_t *refhead)
{
	if (head->coeff > head->sharesc ||
	    head->thresh > head->sharesc) {
		return 0;
	}
	if (head->thresh != refhead->thresh ||
	    head->size != refhead->size ||
	    head->sharesc != refhead->sharesc) {
		return 0;
	}
	return 1;
}

/* initialise */
int
thresh_init(threshold_t *thresh, const char *type, unsigned threshold, unsigned sharesc)
{
	const map_t	*map;

	if (type == NULL) {
		(void) fprintf(stderr, "bad type\n");
		return 0;
	}
	if ((map = findmap(algnames, type)) == NULL) {
		(void) fprintf(stderr, "unknown type '%s'\n", type);
		return 0;
	}
	thresh->sharesc = sharesc;
	thresh->threshold = threshold;
	switch(thresh->type = map->val) {
	case THRESH_IDA:
		return ida_init(thresh, threshold, sharesc);
	case THRESH_RAID4:
		return raid_init(thresh, 4, threshold, sharesc);
	case THRESH_RAID5:
		return raid_init(thresh, 5, threshold, sharesc);
	default:
		(void) fprintf(stderr, "bad type\n");
		return 0;
	}
}

/* add a share */
int
thresh_add_share(threshold_t *thresh, unsigned n, const void *data, ssize_t size)
{
	thresh_str_t	*share;
	struct stat	 st;
	FILE		*fp;

	if (n != THRESH_MAX_SHARES) {
		thresh->availc += 1;
	}
	share = &thresh->shares[n];
	if (size < 0) {
		if ((fp = fopen(data, "r")) == NULL) {
			(void) fprintf(stderr, "can't open file '%s'\n", (const char *)data);
			return 0;
		}
		(void) fstat(fileno(fp), &st);
		share->io.size = (size_t)st.st_size;
		share->io.base = mmap(NULL, share->io.size, PROT_READ, MAP_PRIVATE, fileno(fp), 0);
		if (share->io.base == MAP_FAILED) {
			(void) fprintf(stderr, "ida_add_share: can't mmap file '%s'\n", (const char *)data);
			(void) fclose(fp);
			return 0;
		}
		thresh->mapped[n] = MMAPPED;
	} else {
		if ((share->io.base = calloc(1, (size_t)size)) == NULL) {
			(void) fprintf(stderr, "ida_add_share: can't calloc %zu\n",
				(size_t)size);
			return 0;
		}
		(void) memcpy(share->io.base, data, (size_t)size);
		share->io.size = (size_t)size;
		thresh->mapped[n] = ALLOCATED;
	}
	return 1;
}

/* split via different algorithms */
int
thresh_split(threshold_t *thresh)
{
	switch(thresh->type) {
	case THRESH_IDA:
		return ida_split(thresh);
	case THRESH_RAID5:
		return ida_split(thresh);
	default:
		return 0;
	}
}


/* read a header from the share, and return any info desired */
int
thresh_get_header(thresh_head_t *head, const uint8_t *mem)
{
	if (mem == NULL) {
		return 0;
	}
	(void) memcpy(head, mem, sizeof(*head));
	swapheader(head);
	return 1;
}

/* format the output header for split files */
int
thresh_fmt_header(thresh_head_t *head, const char *type, unsigned threshold, unsigned sharesc, uint64_t size)
{
	(void) memset(head, 0x0, sizeof(*head));
	(void) snprintf(head->magic, sizeof(head->magic), "%s", type);
	head->sharesc = sharesc;
	head->thresh = threshold;
	head->size = size;
	swapheader(head);
	return 1;
}

/* combine/join/decode into one */
int
thresh_combine(threshold_t *thresh)
{
	thresh_head_t	 head;
	thresh_str_t	*out;

	(void) memset(&head, 0x0, sizeof(head));
	if (!thresh_get_header(&head, thresh->shares[0].io.base)) {
		return -1;
	}
	if (thresh->threshold == 0) {
		thresh->threshold = head.thresh;
	}
	if (thresh->sharesc == 0) {
		thresh->sharesc = head.sharesc;
	}
	if (thresh->availc < thresh->threshold) {
		return -1;
	}
	thresh_init(thresh, head.magic, head.thresh, head.sharesc);
	out = &thresh->shares[THRESH_JOIN_DEST];
	switch(thresh->type) {
	case THRESH_IDA:
		return ida_combine(thresh);
	case THRESH_RAID5:
		if (!raid_join_memory(thresh, thresh->shares,
			thresh->shares[0].io.size, thresh->threshold, out)) {
			return -1;
		}
		break;
	default:
		break;
	}
	return -1;
}

/* write a share to a file */
int
thresh_write_share(threshold_t *thresh, unsigned n, const char *f)
{
	thresh_str_t	*out;
	ssize_t		 wc;
	size_t		 cc;
	FILE		*fp;
	int		 ok;

	if (n >= thresh->sharesc && n != THRESH_SPLIT_SOURCE) {
		(void) fprintf(stderr, "share %u out of range (max %u)\n", n, thresh->sharesc);
		return 0;
	}
	out = &thresh->shares[n];
	if (f == NULL) {
		fp = stdout;
	} else {
		if ((fp = fopen(f, "w")) == NULL) {
			(void) fprintf(stderr, "can't write to '%s'\n", f);
			return 0;
		}
	}
	for (ok = 1, cc = 0 ; ok && cc < out->io.size ; cc += wc) {
		if ((wc = write(fileno(fp), &out->io.base[cc], out->io.size - cc)) <= 0) {
			(void) fprintf(stderr, "Short write\n");
			ok = 0;
		}
	}
	if (f != NULL) {
		(void) fclose(fp);
	}
	return ok;
}

/* get a share */
int
thresh_get_share(threshold_t *thresh, unsigned n, void **outp, size_t *size)
{
	thresh_str_t	*share;

	if (n >= thresh->sharesc && n != THRESH_SPLIT_SOURCE) {
		(void) fprintf(stderr, "share %u out of range (max %u)\n", n, thresh->sharesc);
		return 0;
	}
	share = &thresh->shares[(n == THRESH_SPLIT_SOURCE) ? THRESH_MAX_SHARES : n];
	*outp = (char *)share->io.base;
	*size = share->c;
	return 1;
}

/* finish up */
int
thresh_end(threshold_t *thresh)
{
	unsigned	i;

	switch(thresh->type) {
		break;
	case THRESH_IDA:
		ida_end(thresh);
		break;
	case THRESH_RAID5:
		raid_end(thresh);
		break;
	default:
		break;
	}
	for (i = 0 ; i < thresh->sharesc ; i++) {
		switch (thresh->mapped[i]) {
		case MMAPPED:
			(void) munmap(thresh->shares[i].io.base, thresh->shares[i].io.size);
			break;
		case ALLOCATED:
			free(thresh->shares[i].io.base);
			break;
		default:
			break;
		}
	}
	return 1;
}
