/*-
 * Copyright (c) 2012 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>

#include <err.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "vmdk.h"

enum {
	VMDK_DEFAULT_GRAINSIZE		= 128,

	VMDK_VALID_NL_TEST		= 1,
	VMDK_REDUNDANT_GRAIN_TABLE	= (1 << 1),
	VMDK_ZERO_GRAIN_GTE		= (1 << 2),
	VMDK_COMPRESSED_GRAINS		= (1 << 16),
	VMDK_MARKERS			= (1 << 17)
};

#ifndef __packed
#  if defined(__lint__)
#    define	__packed	__packed
#  elif defined(__GNUC__)
#    define	__packed	__attribute__((__packed__))
#  elif defined(__PCC__)
#    define	__packed	_Pragma("packed 1")
#  else
#    define	__packed	error: no __packed for this compiler
#  endif
#endif

/*
 * VMDK information taken from
 * http://www.vmware.com/support/developer/vddk/vmdk_50_technote.pdf?src=vmdk
 */
struct vmdk_extent_header_t {
	uint32_t	magicNumber;
	uint32_t	version;
	uint32_t	flags;
	uint64_t	capacity;	/* capacity of extent in sectors, must be multiple of grainSize */
	uint64_t	grainSize;	/* size of grain in sectors, power of 2 */
	uint64_t	descriptorOffset;
	uint64_t	descriptorSize;
	uint32_t	numGTEsPerGT;
	uint64_t	rgdOffset;
	uint64_t	gdOffset;
	uint64_t	overHead;
	uint8_t		uncleanShutdown;
	int8_t		singleEndLineChar;
	int8_t		nonEndLineChar;
	int8_t		doubleEndLineChar1;
	int8_t		doubleEndLineChar2;
	uint16_t	compressAlgorithm;
	uint8_t		pad[433];
} __packed;

typedef struct vmdk_extent_header_t	vmdk_extent_header_t;

#define	VMDK_SPARSE_MAGICNUMBER	0x564d444b /* 'V' 'M' 'D' 'K' */

/* write the vmdk info to a file descriptor */
static int
write_vmdk_info(FILE *vmdk, const char *vmname, int slices, int flat, uint64_t bytes)
{
	uint64_t	 slicesize;
	uint64_t	 cc;
	uint32_t	 r[4];
	int		 i;

	for (i = 0 ; i < 4 ; i++) {
		r[i] = (uint32_t)random();
	}
	fprintf(vmdk, "# Disk DescriptorFile\n");
	fprintf(vmdk, "version=1\n");
	fprintf(vmdk, "CID=%08x\n", r[3]);
	fprintf(vmdk, "parentCID=ffffffff\n");
	if (flat) {
		fprintf(vmdk, "createType=\"monolithicSparse\"\n\n");
	} else {
		fprintf(vmdk, "encoding=\"UTF-8\"\n");
		fprintf(vmdk, "isNativeSnapshot=\"no\"\n");
		fprintf(vmdk, "createType=\"twoGbMaxExtentSparse\"\n\n");
	}
	fprintf(vmdk, "# Extent description\n");
	if (flat) {
		fprintf(vmdk, "RW %" PRIu64 " SPARSE \"%s.vmdk\"\n",
				bytes / VMDK_SECTOR_SIZE, vmname);
	} else {
		for (i = 0, cc = 0 ; i < slices ; i++, cc += slicesize) {
			slicesize = MIN(VMDK_SLICE_SIZE, bytes - cc);
			fprintf(vmdk, "RW %" PRIu64 " SPARSE \"%s-s%03d.vmdk\"\n",
				slicesize / VMDK_SECTOR_SIZE, vmname, i + 1);
		}
	}
	fprintf(vmdk, "\n# The Disk Data Base\n");
	fprintf(vmdk, "#DDB\n\n");
	fprintf(vmdk, "ddb.adapterType = \"ide\" \n");
	fprintf(vmdk, "ddb.geometry.sectors = \"63\" \n");
	fprintf(vmdk, "ddb.geometry.heads = \"16\" \n");
	fprintf(vmdk, "ddb.geometry.cylinders = \"%" PRIu64 "\"\n",
		(uint64_t)((MIN(VMDK_SLICE_SIZE, bytes)) / (63 * 16 * VMDK_SECTOR_SIZE)));
	if (!flat) {
		fprintf(vmdk, "ddb.uuid = \"");
		for (i = 0 ; i < 4 ; i++) {
			fprintf(vmdk, "%02x %02x %02x %02x%s",
				(r[i] >> 24) & 0xff,
				(r[i] >> 16) & 0xff,
				(r[i] >> 8) & 0xff,
				r[i] & 0xff,
				(i == 1) ? "-" : (i < 3) ? " " : "");
		}
		fprintf(vmdk, "\"\n");
		fprintf(vmdk, "ddb.longContentID = \"%08x%08x%08x%08x\"\n", r[0], r[1], r[2], r[3]);
	}
	fprintf(vmdk, "ddb.virtualHWVersion = \"%d\"\n", (flat) ? 4 : 8);
	return 1;
}

/* write the extent header to the extent */
static int
write_extent_header(FILE *fp, const char *img, uint64_t bytes, const int flat)
{
	vmdk_extent_header_t	 header;

	(void) memset(&header, 0x0, sizeof(header));
	header.magicNumber = VMDK_SPARSE_MAGICNUMBER;
	header.version = 1;
	header.flags = (VMDK_VALID_NL_TEST | VMDK_REDUNDANT_GRAIN_TABLE);
	header.capacity = (bytes / VMDK_SECTOR_SIZE);
	header.grainSize = VMDK_DEFAULT_GRAINSIZE;
	header.descriptorOffset = (flat) ? 1 : 0;
	header.descriptorSize = (flat) ? 20 : 0;
	header.numGTEsPerGT = 512;
	header.rgdOffset = (flat) ? 21 : 1;
	header.gdOffset = (flat) ? 38 : 66;
	header.overHead = (flat) ? 128 : 256;
	header.uncleanShutdown = 0;
	header.singleEndLineChar = '\n';
	header.nonEndLineChar = ' ';
	header.doubleEndLineChar1 = '\r';
	header.doubleEndLineChar2 = '\n';
	header.compressAlgorithm = 0;
	if (fwrite(&header, sizeof(header), 1, fp) != 1) {
		warn("can't write to image '%s'", img);
		return 0;
	}
	return 1;
}

/* extend a sparse file by writing its last byte */
static int
write_last_byte(FILE *fp, off_t last, const char *img)
{
	char	ch;

	if (fseeko(fp, (off_t)last, SEEK_SET) != 0) {
		warn("can't fseeko on image '%s'", img);
		return 0;
	}
	ch = 0x0;
	if (fwrite(&ch, 1, 1, fp) != 1) {
		warn("can't write on image '%s'", img);
		return 0;
	}
	return 1;
}

/* copy a file from img to fp */
static int
copyfile(FILE *fp, FILE *img, uint64_t bytes)
{
	ssize_t	rc;
	size_t	cc;
	char	buf[BUFSIZ * 10];

	fsync(fileno(fp));
	for (cc = 0 ; cc < bytes ; cc += rc) {
		if ((rc = read(fileno(img), buf, sizeof(buf))) <= 0) {
			break;
		}
		if (write(fileno(fp), buf, (size_t)rc) != rc) {
			warn("short write");
			return 0;
		}
	}
	return 1;
}

/*******************************************************************************/

/* make a new VMDK descriptor file, and point it to the image we want to use */
int
vmdk_make_vmdk(const char *vmname, int slices, uint64_t bytes)
{
	char	 vmdkname[MAXPATHLEN];
	FILE	*vmdk;
	int	 ret;

	(void) snprintf(vmdkname, sizeof(vmdkname), "%s.vmdk", vmname);
	if ((vmdk = fopen(vmdkname, "w")) == NULL) {
		warn("can't open vmdk file '%s'", vmdkname);
		return 0;
	}
	printf("[%s %" PRIu64 " bytes]\n", vmdkname, bytes);
	ret = write_vmdk_info(vmdk, vmname, slices, 0, bytes);
	(void) fclose(vmdk);
	return ret;
}

/* construct a VMDK descriptor file from an existing file */
int
vmdk_file_to_vmdk(const char *vmname, const char *imgname)
{
	struct stat	 st;
	const int	 flat = 1;
	uint64_t	 bytes;
	char		 vmdkname[MAXPATHLEN];
	FILE		*vmdk;
	FILE		*img;
	int		 ret;

	(void) snprintf(vmdkname, sizeof(vmdkname), "%s.vmdk", vmname);
	if ((vmdk = fopen(vmdkname, "w")) == NULL) {
		warn("can't open vmdk file '%s'", vmdkname);
		return 0;
	}
	if ((img = fopen(imgname, "r")) == NULL) {
		warn("can't open image file '%s'", imgname);
		fclose(vmdk);
		return 0;
	}
	(void) fstat(fileno(img), &st);
	ret = 0;
	if (write_extent_header(vmdk, imgname,
		bytes = vmdk_nearest_size((uint64_t)st.st_size), flat) &&
	    fseeko(vmdk, (off_t)VMDK_SECTOR_SIZE, SEEK_SET) == 0 &&
	    write_vmdk_info(vmdk, vmname, 1, flat, bytes) &&
	    write_last_byte(vmdk, (off_t)(VMDK_SECTOR_SIZE * 2) - 1, vmdkname) &&
	    copyfile(vmdk, img, (uint64_t)st.st_size)) {
		ret = 1;
	}
	(void) fclose(vmdk);
	(void) fclose(img);
	return ret;
}

/* make a new sparse extent of the required size in bytes */
int
vmdk_make_extent(const char *img, uint64_t bytes)
{
	const int	 sparse = 0;
	FILE		*fp;
	int		 ret;

	if ((fp = fopen(img, "w")) == NULL) {
		warn("can't write to image '%s'", img);
		return 0;
	}
	ret = 0;
	if (write_extent_header(fp, img, bytes, sparse) &&
	    write_last_byte(fp, (off_t)(VMDK_SECTOR_SIZE + bytes - 1), img)) {
		ret = 1;
	}
	(void) fclose(fp);
	return ret;
}

/* list an extent's information */
int
vmdk_list_extent(const char *img)
{
	vmdk_extent_header_t	 header;
	FILE			*fp;
	int			 ret;

	(void) memset(&header, 0x0, sizeof(header));
	if ((fp = fopen(img, "r")) == NULL) {
		warn("can't read from image '%s'", img);
		return 0;
	}
	if (fread(&header, sizeof(header), 1, fp) != 1) {
		warn("can't read from image '%s'", img);
		ret = 0;
	} else {
		printf("magicNumber: %#08x\n", header.magicNumber);
		printf("version: %d\n", header.version);
		printf("flags: %#08x\n", header.flags);
		printf("capacity: %" PRIu64 "\n", header.capacity);
		printf("grainSize: %" PRIu64 "\n", header.grainSize);
		printf("descriptorOffset: %" PRIu64 "\n", header.descriptorOffset);
		printf("descriptorSize: %" PRIu64 "\n", header.descriptorSize);
		printf("numGTEsPerGT: %d\n", header.numGTEsPerGT);
		printf("rgdOffset: %" PRIu64 "\n", header.rgdOffset);
		printf("gdOffset: %" PRIu64 "\n", header.gdOffset);
		printf("overHead: %" PRIu64 "\n", header.overHead);
		printf("uncleanShutdown: %d\n", header.uncleanShutdown);
		printf("singleEndLineChar: %d\n", header.singleEndLineChar);
		printf("nonEndLineChar: %d\n", header.nonEndLineChar);
		printf("dounleEndLineChar1: %d\n", header.doubleEndLineChar1);
		printf("dounleEndLineChar2: %d\n", header.doubleEndLineChar2);
		printf("compressAlgorithm: %#04x\n", header.compressAlgorithm);
		ret = 1;
	}
	fclose(fp);
	return ret;
}

/* return the nearest socially acceptable capacity to 'n' bytes */
uint64_t
vmdk_nearest_size(uint64_t bytes)
{
	return (bytes / (VMDK_DEFAULT_GRAINSIZE * VMDK_SECTOR_SIZE)) *
		(VMDK_DEFAULT_GRAINSIZE * VMDK_SECTOR_SIZE);
}
