/****************************************************************
Copyright (C) Lucent Technologies 1997
All Rights Reserved

Permission to use, copy, modify, and distribute this software and
its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that the copyright notice and this
permission notice and warranty disclaimer appear in supporting
documentation, and that the name Lucent Technologies or any of
its entities not be used in advertising or publicity pertaining
to distribution of the software without specific, written prior
permission.

LUCENT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
IN NO EVENT SHALL LUCENT OR ANY OF ITS ENTITIES BE LIABLE FOR ANY
SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
THIS SOFTWARE.
****************************************************************/

/* lasciate ogne speranza, voi ch'intrate. */

#define	DEBUG

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "awkre.h"
#include "eawk.h"
#include "awkgram.h"
#include "proto.h"

#define NCHARS	(256+3)		/* 256 handles 8-bit chars; 128 does 7-bit */
				/* watch out in match(), etc. */

#define	HAT	(NCHARS+2)	/* matches ^ in regular expr */
				/* NCHARS is 2**n */
#define MAXLIN 22

#define TYPE(v)		(v)->nobj	/* badly overloaded here */
#define INFO(v)		(v)->ntype	/* badly overloaded here */
#define LEFT(v)		(v)->narg[0]
#define RIGHT(v)	(v)->narg[1]
#define PARENT(v)	(v)->nnext

#define LEAF	case CCL: case NCCL: case CHAR: case DOT: case FINAL: case ALL:
#define ELEAF	case EMPTYRE:		/* empty string in regexp */
#define UNARY	case STAR: case PLUS: case QUEST:

/* encoding in tree awknode_t:
	leaf (CCL, NCCL, CHAR, DOT, FINAL, ALL, EMPTYRE):
		left is index, right contains value or pointer to value
	unary (STAR, PLUS, QUEST): left is child, right is null
	binary (CAT, OR): left and right are children
	parent contains pointer to parent
*/


static	int	makeinit(eawk_t *, fa_t *, int);
static	int	penter(eawk_t *, awknode_t *);
static	int	freetr(eawk_t *, awknode_t *);
static	int	hexstr(const uint8_t **);
static	int	cfoll(eawk_t *eawk, fa_t *, awknode_t *);
static	int	first(eawk_t *, awknode_t *);
static	void	follow(eawk_t *, awknode_t *);
static	int	member(int, const char *);
static	awknode_t	*regexp(eawk_t *);
static	awknode_t	*primary(eawk_t *);
static	awknode_t	*concat(eawk_t *, awknode_t *);
static	awknode_t	*alt(eawk_t *, awknode_t *);
static int relex(eawk_t *);
static awknode_t *unary(eawk_t *eawk, awknode_t *);

static int
resize_state(eawk_t *eawk, fa_t *fap, int state)
{
	void *p;
	int	new_count;
	int	i;

	if (++state < fap->state_count)
		return 1;

	new_count = state + 10; /* needs to be tuned */

	p = realloc(fap->gototab, new_count * sizeof(fap->gototab[0]));
	if (p == NULL)
		goto out;
	fap->gototab = p;

	p = realloc(fap->out, new_count * sizeof(fap->out[0]));
	if (p == NULL)
		goto out;
	fap->out = p;

	p = realloc(fap->posns, new_count * sizeof(fap->posns[0]));
	if (p == NULL)
		goto out;
	fap->posns = p;

	for (i = fap->state_count; i < new_count; ++i) {
		fap->gototab[i] = calloc(1, NCHARS * sizeof (**fap->gototab));
		if (fap->gototab[i] == NULL)
			goto out;
		fap->out[i]  = 0;
		fap->posns[i] = NULL;
	}
	fap->state_count = new_count;
	return 1;
out:
	eawk_warning(eawk, "out of memory in resize_state");
	return 0;
}

/* parses regular expression pointed to by p */
/* uses relex() to scan regular expression */
static awknode_t *
reparse(eawk_t *eawk, const char *p)
{
	awknode_t *np;

	DPRINTF(eawk, ("reparse <%s>\n", p) );
	/* prestr points to string to be parsed */
	eawk->lastre = eawk->prestr = (uint8_t *)__UNCONST(p);
	eawk->rtok = relex(eawk);
	/* GNU compatibility: an empty regexp matches anything */
	if (eawk->rtok == '\0') {
		/* FATAL(eawk, "empty regular expression"); previous */
		return(eawk_op2(eawk, EMPTYRE, NIL, NIL));
	}
	np = regexp(eawk);
	if (eawk->rtok != '\0') {
		eawk_warning(eawk, "syntax error in regular expression %s at %s", eawk->lastre, eawk->prestr);
		return NULL;
	}
	return np;
}

/* does the real work of making a dfa */
/* anchor = 1 for anchored matches, else 0 */
static fa_t *
mkdfa(eawk_t *eawk, const char *s, int anchor)
{
	awknode_t	*p;
	awknode_t	*p1;
	fa_t		*re;

	p = reparse(eawk, s);
	p1 = eawk_op2(eawk, CAT, eawk_op2(eawk, STAR, eawk_op2(eawk, ALL, NIL, NIL), NIL), p);
		/* put ALL STAR in front of reg.  exp. */
	p1 = eawk_op2(eawk, CAT, p1, eawk_op2(eawk, FINAL, NIL, NIL));
		/* put FINAL after reg.  exp. */

	eawk->poscnt = 0;
	penter(eawk, p1);	/* enter parent pointers and leaf indices */
	if ((re = calloc(1, sizeof(fa_t) + eawk->poscnt*sizeof(awkrrow_t))) == NULL) {
		eawk_warning(eawk, "out of space for fa_t");
		return NULL;
	}
	re->accept = eawk->poscnt-1;	/* penter has computed number of positions in re */
	cfoll(eawk, re, p1);	/* set up follow sets */
	if (!freetr(eawk, p1)) {
		return NULL;
	}
	if (!resize_state(eawk, re, 1)) {
		return NULL;
	}
	if ((re->posns[0] = calloc(1, *(re->re[0].lfollow)*sizeof(int))) == NULL) {
		eawk_warning(eawk, "out of space in mkdfa");
		return NULL;
	}
	if ((re->posns[1] = calloc(1, sizeof(int))) == NULL) {
		eawk_warning(eawk, "out of space in mkdfa");
		return NULL;
	}
	*re->posns[1] = 0;
	re->initstat = makeinit(eawk, re, anchor);
	re->anchor = anchor;
	re->restr = (uint8_t *) eawk_tostring(eawk, s);
	return re;
}

static int
cgoto(eawk_t *eawk, fa_t *re, int s, int c)
{
	int	i;
	int	j;
	int	k;
	int	*p;
	int	*q;

	if (!(c == HAT || c < NCHARS)) {
		eawk_warning(eawk, "cgoto: not HAT or < NCHARS");
		return 0;
	}
	while (re->accept >= eawk->maxsetvec) {	/* guessing here! */
		eawk->maxsetvec *= 4;
		eawk->setvec = realloc(eawk->setvec, eawk->maxsetvec * sizeof(i));
		eawk->tmpset = realloc(eawk->tmpset, eawk->maxsetvec * sizeof(i));
		if (eawk->setvec == NULL || eawk->tmpset == NULL) {
			eawk_warning(eawk, "out of space in cgoto()");
			return 0;
		}
	}
	for (i = 0; i <= re->accept; i++)
		eawk->setvec[i] = 0;
	eawk->setcnt = 0;
	if (!resize_state(eawk, re, s)) {
		return 0;
	}
	/* compute positions of gototab[s,c] into setvec */
	p = re->posns[s];
	for (i = 1; i <= *p; i++) {
		if ((k = (int)re->re[p[i]].ltype) != FINAL) {
			if ((k == CHAR && c == eawk_ptoi(re->re[p[i]].lval.np))
			 || (k == DOT && c != 0 && c != HAT)
			 || (k == ALL && c != 0)
			 || (k == EMPTYRE && c != 0)
			 || (k == CCL && member(c, (char *) re->re[p[i]].lval.up))
			 || (k == NCCL && !member(c, (char *) re->re[p[i]].lval.up) && c != 0 && c != HAT)) {
				q = re->re[p[i]].lfollow;
				for (j = 1; j <= *q; j++) {
					if (q[j] >= eawk->maxsetvec) {
						eawk->maxsetvec *= 4;
						eawk->setvec = realloc(eawk->setvec, eawk->maxsetvec * sizeof(i));
						eawk->tmpset = realloc(eawk->tmpset, eawk->maxsetvec * sizeof(i));
						if (eawk->setvec == NULL || eawk->tmpset == NULL) {
							eawk_warning(eawk, "cgoto overflow");
							return 0;
						}
					}
					if (eawk->setvec[q[j]] == 0) {
						eawk->setcnt++;
						eawk->setvec[q[j]] = 1;
					}
				}
			}
		}
	}
	/* determine if setvec is a previous state */
	eawk->tmpset[0] = eawk->setcnt;
	j = 1;
	for (i = re->accept; i >= 0; i--)
		if (eawk->setvec[i]) {
			eawk->tmpset[j++] = i;
		}

	if (!resize_state(eawk, re, re->curstat > s ? re->curstat : s)) {
		return 0;
	}
	/* tmpset == previous state? */
	for (i = 1; i <= re->curstat; i++) {
		p = re->posns[i];
		if ((k = eawk->tmpset[0]) != p[0])
			goto different;
		for (j = 1; j <= k; j++)
			if (eawk->tmpset[j] != p[j])
				goto different;
		/* setvec is state i */
		if (c != HAT)
			re->gototab[s][c] = i;
		return i;
different:;
	}

	/* add tmpset to current set of states */
	++(re->curstat);
	if (!resize_state(eawk, re, re->curstat)) {
		return 0;
	}
	for (i = 0; i < NCHARS; i++)
		re->gototab[re->curstat][i] = 0;
	XFREE(re->posns[re->curstat]);
	if ((p = calloc(1, (eawk->setcnt+1)*sizeof(i))) == NULL) {
		eawk_warning(eawk, "out of space in cgoto");
		return 0;
	}

	re->posns[re->curstat] = p;
	re->gototab[s][c] = re->curstat;
	for (i = 0; i <= eawk->setcnt; i++)
		p[i] = eawk->tmpset[i];
	if (eawk->setvec[re->accept])
		re->out[re->curstat] = 1;
	else
		re->out[re->curstat] = 0;
	return re->curstat;
}

static int
makeinit(eawk_t *eawk, fa_t *re, int anchor)
{
	int	i;
	int	k;

	if (!resize_state(eawk, re, 2)) {
		return 0;
	}
	re->curstat = 2;
	re->out[2] = 0;
	k = *(re->re[0].lfollow);
	XFREE(re->posns[2]);			
	if ((re->posns[2] = calloc(1, (k+1)*sizeof(i))) == NULL) {
		eawk_warning(eawk, "out of space in makeinit");
		return 0;
	}
	for (i=0; i <= k; i++) {
		(re->posns[2])[i] = (re->re[0].lfollow)[i];
	}
	if ((re->posns[2])[1] == re->accept)
		re->out[2] = 1;
	for (i=0; i < NCHARS; i++)
		re->gototab[2][i] = 0;
	re->curstat = cgoto(eawk, re, 2, HAT);
	if (anchor) {
		*re->posns[2] = k-1;	/* leave out position 0 */
		for (i=0; i < k; i++) {
			(re->posns[0])[i] = (re->posns[2])[i];
		}

		re->out[0] = re->out[2];
		if (re->curstat != 2) {
			resize_state(eawk, re, re->curstat);
			--(*re->posns[re->curstat]);
		}
	}
	return re->curstat;
}

/* set up parent pointers and leaf indices */
static int
penter(eawk_t *eawk, awknode_t *p)
{
	switch (TYPE(p)) {
	ELEAF
	LEAF
		INFO(p) = eawk->poscnt;
		eawk->poscnt++;
		break;
	UNARY
		if (!penter(eawk, LEFT(p))) {
			return 0;
		}
		PARENT(LEFT(p)) = p;
		break;
	case CAT:
	case OR:
		if (!penter(eawk, LEFT(p))) {
			return 0;
		}
		if (!penter(eawk, RIGHT(p))) {
			return 0;
		}
		PARENT(LEFT(p)) = p;
		PARENT(RIGHT(p)) = p;
		break;
	default:	/* can't happen */
		eawk_warning(eawk, "can't happen: unknown type %d in penter", TYPE(p));
		return 0;
	}
	return 1;
}

/* free parse tree */
static int
freetr(eawk_t *eawk, awknode_t *p)
{
	switch (TYPE(p)) {
	ELEAF
	LEAF
		XFREE(p);
		break;
	UNARY
		if (!freetr(eawk, LEFT(p))) {
			return 0;
		}
		XFREE(p);
		break;
	case CAT:
	case OR:
		if (!freetr(eawk, LEFT(p))) {
			return 0;
		}
		if (!freetr(eawk, RIGHT(p))) {
			return 0;
		}
		XFREE(p);
		break;
	default:	/* can't happen */
		eawk_warning(eawk, "can't happen: unknown type %d in freetr", TYPE(p));
		return 0;
	}
	return 1;
}

/* in the parsing of regular expressions, metacharacters like . have */
/* to be seen literally;  \056 is not a metacharacter. */

/* find and eval hex string at pp, return new p */
/* only pick up one 8-bit byte (2 chars) */
static int
hexstr(const uint8_t **pp)
{
	const uint8_t *p;
	int n = 0;
	int i;

	for (i = 0, p = *pp; i < 2 && isxdigit(*p); i++, p++) {
		if (isdigit(*p)) {
			n = 16 * n + *p - '0';
		} else if (*p >= 'a' && *p <= 'f') {
			n = 16 * n + *p - 'a' + 10;
		} else if (*p >= 'A' && *p <= 'F') {
			n = 16 * n + *p - 'A' + 10;
		}
	}
	*pp = p;
	return n;
}

#define ISOCTDIGIT(c) ((c) >= '0' && (c) <= '7')	/* multiple use of arg */

/* pick up next thing after a \\ */
/* and increment *pp */
static int
quoted(const uint8_t **pp)
{
	const uint8_t *p = *pp;
	int c;

	if ((c = *p++) == 't')
		c = '\t';
	else if (c == 'n')
		c = '\n';
	else if (c == 'f')
		c = '\f';
	else if (c == 'r')
		c = '\r';
	else if (c == 'b')
		c = '\b';
	else if (c == '\\')
		c = '\\';
	else if (c == 'x') {	/* hexadecimal goo follows */
		c = hexstr(&p);	/* this adds a null if number is invalid */
	} else if (ISOCTDIGIT(c)) {	/* \d \dd \ddd */
		int n = c - '0';
		if (ISOCTDIGIT(*p)) {
			n = 8 * n + *p++ - '0';
			if (ISOCTDIGIT(*p))
				n = 8 * n + *p++ - '0';
		}
		c = n;
	} /* else */
		/* c = c; */
	*pp = p;
	return c;
}

/* add a character class */
static char *
cclenter(eawk_t *eawk, const char *argp)
{
	int	i;
	int	c;
	int	c2;
	const uint8_t *p = (const uint8_t *) argp;
	const uint8_t *op;
	uint8_t	*bp;

	op = p;
	if (eawk->cclenterbuf == NULL && (eawk->cclenterbuf = malloc(eawk->cclenterbufsz)) == NULL) {
		eawk_warning(eawk, "out of space for character class [%.10s...] 1", p);
		return NULL;
	}
	bp = eawk->cclenterbuf;
	for (i = 0; (c = *p++) != 0; ) {
		if (c == '\\') {
			c = quoted(&p);
		} else if (c == '-' && i > 0 && bp[-1] != 0) {
			if (*p != 0) {
				c = bp[-1];
				c2 = *p++;
				if (c2 == '\\')
					c2 = quoted(&p);
				if (c > c2) {	/* empty; ignore */
					bp--;
					i--;
					continue;
				}
				while (c < c2) {
					if (!eawk_adjbuf(eawk, &eawk->cclenterbuf, &eawk->cclenterbufsz, (int)(bp - eawk->cclenterbuf+2), 100, &bp, "cclenter1")) {
						eawk_warning(eawk, "out of space for character class [%.10s...] 2", p);
						return NULL;
					}
					*bp++ = ++c;
					i++;
				}
				continue;
			}
		}
		if (!eawk_adjbuf(eawk, &eawk->cclenterbuf, &eawk->cclenterbufsz, (int)(bp - eawk->cclenterbuf+2), 100, &bp, "cclenter2")) {
			eawk_warning(eawk, "out of space for character class [%.10s...] 3", p);
			return NULL;
		}
		*bp++ = c;
		i++;
	}
	*bp = 0;
	DPRINTF(eawk, ("cclenter: in = |%s|, out = |%s|\n", op, eawk->cclenterbuf) );
	XFREE(op);
	return (char *) eawk_tostring(eawk, (char *) eawk->cclenterbuf);
}

/* enter follow set of each leaf of vertex v into lfollow[leaf] */
static int
cfoll(eawk_t *eawk, fa_t *re, awknode_t *v)
{
	int i;
	int *p;

	switch (TYPE(v)) {
	ELEAF
	LEAF
		re->re[INFO(v)].ltype = TYPE(v);
		re->re[INFO(v)].lval.np = RIGHT(v);
		while (re->accept >= eawk->maxsetvec) {	/* guessing here! */
			eawk->maxsetvec *= 4;
			eawk->setvec = realloc(eawk->setvec, eawk->maxsetvec * sizeof(i));
			eawk->tmpset = realloc(eawk->tmpset, eawk->maxsetvec * sizeof(i));
			if (eawk->setvec == NULL || eawk->tmpset == NULL) {
				eawk_warning(eawk, "out of space in cfoll()");
				return 0;
			}
		}
		for (i = 0; i <= re->accept; i++)
			eawk->setvec[i] = 0;
		eawk->setcnt = 0;
		follow(eawk, v);	/* computes setvec and setcnt */
		if ((p = calloc(eawk->setcnt + 1, sizeof(i))) == NULL) {
			eawk_warning(eawk, "out of space building follow set");
			return 0;
		}
		re->re[INFO(v)].lfollow = p;
		*p = eawk->setcnt;
		for (i = re->accept; i >= 0; i--)
			if (eawk->setvec[i] == 1)
				*++p = i;
		break;
	UNARY
		cfoll(eawk, re,LEFT(v));
		break;
	case CAT:
	case OR:
		cfoll(eawk, re,LEFT(v));
		cfoll(eawk, re,RIGHT(v));
		break;
	default:	/* can't happen */
		eawk_warning(eawk, "can't happen: unknown type %d in cfoll", TYPE(v));
		return 0;
	}
	return 1;
}

/* collects initially active leaves of p into setvec */
/* returns 0 if p matches empty string */
static int
first(eawk_t *eawk, awknode_t *p)
{
	int	lp;
	int	b;

	switch (TYPE(p)) {
	ELEAF
	LEAF
		lp = INFO(p);	/* look for high-water mark of subscripts */
		while (eawk->setcnt >= eawk->maxsetvec || lp >= eawk->maxsetvec) {	/* guessing here! */
			eawk->maxsetvec *= 4;
			eawk->setvec = realloc(eawk->setvec, eawk->maxsetvec * sizeof(b));
			eawk->tmpset = realloc(eawk->tmpset, eawk->maxsetvec * sizeof(b));
			if (eawk->setvec == NULL || eawk->tmpset == NULL) {
				eawk_warning(eawk, "out of space in first()");
				return 0;
			}
		}
		if (TYPE(p) == EMPTYRE) {
			eawk->setvec[lp] = 0;
			return(0);
		}
		if (eawk->setvec[lp] != 1) {
			eawk->setvec[lp] = 1;
			eawk->setcnt++;
		}
		if (TYPE(p) == CCL && (*(const char *)(const void *)RIGHT(p)) == '\0')
			return(0);		/* empty CCL */
		else return(1);
	case PLUS:
		if (first(eawk, LEFT(p)) == 0) return(0);
		return(1);
	case STAR:
	case QUEST:
		first(eawk, LEFT(p));
		return(0);
	case CAT:
		if (first(eawk, LEFT(p)) == 0 && first(eawk, RIGHT(p)) == 0) return(0);
		return(1);
	case OR:
		b = first(eawk, RIGHT(p));
		if (first(eawk, LEFT(p)) == 0 || b == 0) return(0);
		return(1);
	}
	/* can't happen */
	eawk_warning(eawk, "can't happen: unknown type %d in first", TYPE(p));
	return(-1);
}

/* collects leaves that can follow v into setvec */
static void
follow(eawk_t *eawk, awknode_t *v)
{
	awknode_t *p;

	if (TYPE(v) == FINAL)
		return;
	p = PARENT(v);
	switch (TYPE(p)) {
	case STAR:
	case PLUS:
		first(eawk, v);
		follow(eawk, p);
		return;

	case OR:
	case QUEST:
		follow(eawk, p);
		return;

	case CAT:
		if (v == LEFT(p)) {	/* v is left child of p */
			if (first(eawk, RIGHT(p)) == 0) {
				follow(eawk, p);
				return;
			}
		} else		/* v is right child */
			follow(eawk, p);
		return;
	}
}

static int
member(int c, const char *sarg)	/* is c in s? */
{
	const uint8_t *s = (const uint8_t *) sarg;

	while (*s)
		if (c == *s++)
			return(1);
	return(0);
}

/* shortest match ? */
int
eawk_match(eawk_t *eawk, void *fa, const char *p0)
{
	const uint8_t	*p = (const uint8_t *) p0;
	fa_t		*re = (fa_t *)(void *)fa;
	int		 ns;
	int		 s;

	s = re->initstat;
	if (s >= re->state_count) {
		eawk_warning(eawk, "eawk_match: s >= re->state_count");
		return 0;
	}

	if (re->out[s])
		return(1);
	do {
		/* assert(*p < NCHARS); */
		if ((ns = re->gototab[s][*p]) != 0)
			s = ns;
		else
			s = cgoto(eawk, re, s, *p);

		if (s >= re->state_count) {
			eawk_warning(eawk, "eawk_match: s >= re->state_count");
			return 0;
		}

		if (re->out[s])
			return(1);
	} while (*p++ != 0);
	return(0);
}

/* longest match, for sub */
int
eawk_pmatch(eawk_t *eawk, void *fa, const char *p0)
{
	int s, ns;
	fa_t	*re = (fa_t *)(void *)fa;
	uint8_t *p = (uint8_t *)__UNCONST(p0);
	uint8_t *q;

	s = re->initstat;
	if (s >= re->state_count) {
		eawk_warning(eawk, "eawk_pmatch: s >= re->state_count");
		return 0;
	}
	eawk->patbeg = p;
	eawk->patlen = -1;
	do {
		q = p;
		do {
			if (re->out[s])		/* final state */
				eawk->patlen = (int)(q-p);
			/* assert(*q < NCHARS); */
			if ((ns = re->gototab[s][*q]) != 0)
				s = ns;
			else
				s = cgoto(eawk, re, s, *q);

			if (s >= re->state_count) {
				eawk_warning(eawk, "eawk_pmatch: s >= re->state_count");
				return 0;
			}

			if (s == 1) {	/* no transition */
				if (eawk->patlen >= 0) {
					eawk->patbeg = p;
					return(1);
				}
				else
					goto nextin;	/* no match */
			}
		} while (*q++ != 0);
		if (re->out[s])
			eawk->patlen = (int)(q-p-1);	/* don't count $ */
		if (eawk->patlen >= 0) {
			eawk->patbeg = p;
			return(1);
		}
nextin:
		s = 2;
	} while (*p++ != 0);
	return (0);
}

/* non-empty match, for sub */
int
eawk_nematch(eawk_t *eawk, void *fa, const char *p0)
{
	int s, ns;
	fa_t	*re = (fa_t *)(void *)fa;
	uint8_t *p = (uint8_t *)__UNCONST(p0);
	uint8_t *q;

	s = re->initstat;
	if (s >= re->state_count) {
		eawk_warning(eawk, "eawk_nematch: s >= re->state_count");
		return 0;
	}

	eawk->patlen = -1;
	while (*p) {
		q = p;
		do {
			if (re->out[s])		/* final state */
				eawk->patlen = (int)(q-p);
			/* assert(*q < NCHARS); */
			if ((ns = re->gototab[s][*q]) != 0)
				s = ns;
			else
				s = cgoto(eawk, re, s, *q);

			if (s >= re->state_count) {
				eawk_warning(eawk, "eawk_nematch: s >= re->state_count");
				return 0;
			}

			if (s == 1) {	/* no transition */
				if (eawk->patlen > 0) {
					eawk->patbeg = p;
					return(1);
				} else
					goto nnextin;	/* no nonempty match */
			}
		} while (*q++ != 0);
		if (re->out[s])
			eawk->patlen = (int)(q-p-1);	/* don't count $ */
		if (eawk->patlen > 0 ) {
			eawk->patbeg = p;
			return(1);
		}
nnextin:
		s = 2;
		p++;
	}
	return (0);
}

/* top-level parse of reg expr */
static awknode_t *
regexp(eawk_t *eawk)
{
	return (alt(eawk, concat(eawk, primary(eawk))));
}

static awknode_t *
primary(eawk_t *eawk)
{
	awknode_t *np;

	switch (eawk->rtok) {
	case CHAR:
		np = eawk_op2(eawk, CHAR, NIL, eawk_itonp(eawk->rlxval));
		eawk->rtok = relex(eawk);
		return (unary(eawk, np));
	case ALL:
		eawk->rtok = relex(eawk);
		return (unary(eawk, eawk_op2(eawk, ALL, NIL, NIL)));
	case EMPTYRE:
		eawk->rtok = relex(eawk);
		return (unary(eawk, eawk_op2(eawk, ALL, NIL, NIL)));
	case DOT:
		eawk->rtok = relex(eawk);
		return (unary(eawk, eawk_op2(eawk, DOT, NIL, NIL)));
	case CCL:
		np = eawk_op2(eawk, CCL, NIL, (awknode_t*)(void *)cclenter(eawk, (char *)(void *)eawk->rlxstr));
		eawk->rtok = relex(eawk);
		return (unary(eawk, np));
	case NCCL:
		np = eawk_op2(eawk, NCCL, NIL, (awknode_t *)(void *)cclenter(eawk, (char *)(void *)eawk->rlxstr));
		eawk->rtok = relex(eawk);
		return (unary(eawk, np));
	case '^':
		eawk->rtok = relex(eawk);
		return (unary(eawk, eawk_op2(eawk, CHAR, NIL, eawk_itonp(HAT))));
	case '$':
		eawk->rtok = relex(eawk);
		return (unary(eawk, eawk_op2(eawk, CHAR, NIL, NIL)));
	case '(':
		eawk->rtok = relex(eawk);
		if (eawk->rtok == ')') {	/* special pleading for () */
			eawk->rtok = relex(eawk);
			return unary(eawk, eawk_op2(eawk, CCL, NIL, (awknode_t *)(void *)eawk_tostring(eawk, "")));
		}
		np = regexp(eawk);
		if (eawk->rtok == ')') {
			eawk->rtok = relex(eawk);
			return (unary(eawk, np));
		}
		eawk_warning(eawk, "syntax error in regular expression %s at %s", eawk->lastre, eawk->prestr);
		return 0;
	default:
		eawk_warning(eawk, "illegal primary in regular expression %s at %s", eawk->lastre, eawk->prestr);
		return 0;
	}
}

static awknode_t *
concat(eawk_t *eawk, awknode_t *np)
{
	switch (eawk->rtok) {
	case CHAR:
	case DOT:
	case ALL:
	case EMPTYRE:
	case CCL:
	case NCCL:
	case '$':
	case '(':
		return (concat(eawk, eawk_op2(eawk, CAT, np, primary(eawk))));
	}
	return (np);
}

static awknode_t *
alt(eawk_t *eawk, awknode_t *np)
{
	if (eawk->rtok == OR) {
		eawk->rtok = relex(eawk);
		return (alt(eawk, eawk_op2(eawk, OR, np, concat(eawk, primary(eawk)))));
	}
	return (np);
}

static awknode_t *
unary(eawk_t *eawk, awknode_t *np)
{
	switch (eawk->rtok) {
	case STAR:
		eawk->rtok = relex(eawk);
		return (unary(eawk, eawk_op2(eawk, STAR, np, NIL)));
	case PLUS:
		eawk->rtok = relex(eawk);
		return (unary(eawk, eawk_op2(eawk, PLUS, np, NIL)));
	case QUEST:
		eawk->rtok = relex(eawk);
		return (unary(eawk, eawk_op2(eawk, QUEST, np, NIL)));
	default:
		return (np);
	}
}

/*
 * Character class definitions conformant to the POSIX locale as
 * defined in IEEE P1003.1 draft 7 of June 2001, assuming the source
 * and operating character sets are both ASCII (ISO646) or supersets
 * thereof.
 *
 * Note that to avoid overflowing the temporary buffer used in
 * relex(), the expanded character class (prior to range expansion)
 * must be less than twice the size of their full name.
 */

/* Because isblank doesn't show up in any of the header files on any
 * system i use, it's defined here.  if some other locale has a richer
 * definition of "blank", define HAS_ISBLANK and provide your own
 * version.
 * the parentheses here are an attempt to find a path through the maze
 * of macro definition and/or function and/or version provided.  thanks
 * to nelson beebe for the suggestion; let's see if it works everywhere.
 */

#ifndef HAS_ISBLANK

int (isblank)(int c)
{
	return c==' ' || c=='\t';
}

#endif

static const struct charclass {
	const char *cc_name;
	int cc_namelen;
	int (*cc_func)(int);
} charclasses[] = {
	{ "alnum",	5,	isalnum },
	{ "alpha",	5,	isalpha },
	{ "blank",	5,	isblank },
	{ "cntrl",	5,	iscntrl },
	{ "digit",	5,	isdigit },
	{ "graph",	5,	isgraph },
	{ "lower",	5,	islower },
	{ "print",	5,	isprint },
	{ "punct",	5,	ispunct },
	{ "space",	5,	isspace },
	{ "upper",	5,	isupper },
	{ "xdigit",	6,	isxdigit },
	{ NULL,		0,	NULL },
};

/* lexical analyzer for reparse */
static int
relex(eawk_t *eawk)
{
	int c, n;
	int cflag;
	uint8_t *bp;
	const struct charclass *cc;
	int i;

	switch (c = *eawk->prestr++) {
	case '|': return OR;
	case '*': return STAR;
	case '+': return PLUS;
	case '?': return QUEST;
	case '.': return DOT;
	case '\0': eawk->prestr--; return '\0';
	case '^':
	case '$':
	case '(':
	case ')':
		return c;
	case '\\':
		eawk->rlxval = quoted((const uint8_t **)&eawk->prestr);
		return CHAR;
	default:
		eawk->rlxval = c;
		return CHAR;
	case '[': 
		if (eawk->relexbuf == NULL && (eawk->relexbuf = malloc(eawk->relexbufsz)) == NULL) {
			eawk_warning(eawk, "out of space in reg expr %.10s..", eawk->lastre);
			return 0;
		}
		bp = eawk->relexbuf;
		if (*eawk->prestr == '^') {
			cflag = 1;
			eawk->prestr++;
		}
		else
			cflag = 0;
		n = (int)(2 * strlen((const char *) eawk->prestr)+1);
		if (!eawk_adjbuf(eawk, &eawk->relexbuf, &eawk->relexbufsz, n, n, &bp, "relex1")) {
			eawk_warning(eawk, "out of space for reg expr %.10s...", eawk->lastre);
			return 0;
		}
		for (;;) {
			if ((c = *eawk->prestr++) == '\\') {
				*bp++ = '\\';
				if ((c = *eawk->prestr++) == '\0') {
					eawk_warning(eawk, "nonterminated character class %.20s...", eawk->lastre);
					return 0;
				}
				*bp++ = c;
			/* } else if (c == '\n') { */
			/* 	FATAL(eawk, "newline in character class %.20s...", eawk->lastre); */
			} else if (c == '[' && *eawk->prestr == ':') {
				/* POSIX char class names, Dag-Erling Smorgrav, des@ofug.org */
				for (cc = charclasses; cc->cc_name; cc++)
					if (strncmp((const char *) eawk->prestr + 1, (const char *) cc->cc_name, cc->cc_namelen) == 0)
						break;
				if (cc->cc_name != NULL && eawk->prestr[1 + cc->cc_namelen] == ':' &&
				    eawk->prestr[2 + cc->cc_namelen] == ']') {
					eawk->prestr += cc->cc_namelen + 3;
					for (i = 1; i < NCHARS; i++) {
						if (!eawk_adjbuf(eawk, &eawk->relexbuf, &eawk->relexbufsz, (int)(bp - eawk->relexbuf+1), 100, &bp, "relex2")) {
						    eawk_warning(eawk, "out of space for reg expr %.10s...", eawk->lastre);
						    return 0;
						}
						if (cc->cc_func(i)) {
							*bp++ = i;
							n++;
						}
					}
				} else
					*bp++ = c;
			} else if (c == '\0') {
				eawk_warning(eawk, "nonterminated character class %.20s", eawk->lastre);
				return 0;
			} else if (bp == eawk->relexbuf) {	/* 1st char is special */
				*bp++ = c;
			} else if (c == ']') {
				*bp++ = 0;
				eawk->rlxstr = (uint8_t *) eawk_tostring(eawk, (char *) eawk->relexbuf);
				if (cflag == 0)
					return CCL;
				else
					return NCCL;
			} else
				*bp++ = c;
		}
	}
}


static void
freefa(fa_t *re)	/* free a finite automaton */
{
	int i;

	if (re == NULL)
		return;
	for (i = 0; i < re->state_count; i++) {
		XFREE(re->gototab[i]);
		XFREE(re->posns[i]);
	}
	for (i = 0; i <= re->accept; i++) {
		XFREE(re->re[i].lfollow);
		if (re->re[i].ltype == CCL || re->re[i].ltype == NCCL)
			XFREE((re->re[i].lval.np));
	}
	XFREE(re->restr);
	XFREE(re->out);
	XFREE(re->posns);
	XFREE(re->gototab);
	XFREE(re);
}

/* returns dfa for reg expr s */
void *
eawk_regcomp(eawk_t *eawk, const char *s, int anchor)
{
	int i, use, nuse;
	fa_t	*re;
	fa_t *pfa;

	if (eawk->setvec == NULL) {	/* first time through any RE */
		eawk->maxsetvec = MAXLIN;
		eawk->setvec = malloc(eawk->maxsetvec * sizeof(i));
		eawk->tmpset = malloc(eawk->maxsetvec * sizeof(i));
		if (eawk->setvec == NULL || eawk->tmpset == NULL) {
			eawk_warning(eawk, "out of space initializing eawk_regcomp");
			return NULL;
		}
	}

	if (strcmp(eawk_getsval(eawk, eawk_lookup("stage", eawk->control)), "compiling") == 0) {
		/* a constant for sure */
		return mkdfa(eawk, s, anchor);
	}
	for (i = 0; i < eawk->nfatab; i++) {
		/* is it there already? */
		re = eawk->fatab[i];
		if (re->anchor == anchor &&
		    strcmp((const char *) re->restr, s) == 0) {
			re->use = eawk->make_dfa_now++;
			return re;
		}
	}
	pfa = mkdfa(eawk, s, anchor);
	if (eawk->nfatab < AWKLIB_NFA) {	/* room for another */
		re = eawk->fatab[eawk->nfatab] = pfa;
		re->use = eawk->make_dfa_now++;
		eawk->nfatab++;
		return pfa;
	}
	re = eawk->fatab[0];
	use = re->use;	/* replace least-recently used */
	nuse = 0;
	for (i = 1; i < eawk->nfatab; i++) {
		re = eawk->fatab[i];
		if (re->use < use) {
			use = re->use;
			nuse = i;
		}
	}
	freefa(eawk->fatab[nuse]);
	eawk->fatab[nuse] = pfa;
	pfa->use = eawk->make_dfa_now++;
	return pfa;
}
