/****************************************************************
Copyright (C) Lucent Technologies 1997
All Rights Reserved

Permission to use, copy, modify, and distribute this software and
its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that the copyright notice and this
permission notice and warranty disclaimer appear in supporting
documentation, and that the name Lucent Technologies or any of
its entities not be used in advertising or publicity pertaining
to distribution of the software without specific, written prior
permission.

LUCENT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
IN NO EVENT SHALL LUCENT OR ANY OF ITS ENTITIES BE LIABLE FOR ANY
SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
THIS SOFTWARE.
****************************************************************/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "eawk.h"
#include "awkgram.h"
#include "proto.h"

typedef struct Keyword {
	const char *word;
	int	sub;
	int	type;
} Keyword;

static int peek(eawk_t *);
static int gettok(eawk_t *, char **, int *);
static int binsearch(const char *, const Keyword *, int);

const Keyword keywords[] ={	/* keep sorted: binary searched */
	{ "BEGIN",	XBEGIN,		XBEGIN },
	{ "END",	XEND,		XEND },
	{ "NF",		VARNF,		VARNF },
	{ "and",	BITAND,		BITAND },
	{ "atan2",	FATAN,		BLTIN },
	{ "break",	BREAK,		BREAK },
	{ "close",	CLOSE,		CLOSE },
	{ "compl",	BITCOMPL,	BITCOMPL },
	{ "continue",	CONTINUE,	CONTINUE },
	{ "cos",	FCOS,		BLTIN },
	{ "delete",	DELETE,		DELETE },
	{ "dlalloc",	DLALLOC,	DLALLOC },
	{ "dlcall",	DLCALL,		DLCALL },
	{ "dlclose",	DLCLOSE,	DLCLOSE },
	{ "dlfield",	DLFIELD,	DLFIELD },
	{ "dlfree",	DLFREE,		DLFREE },
	{ "dlopen",	DLOPEN,		DLOPEN },
	{ "dlproto",	DLPROTO,	DLPROTO },
	{ "do",		DO,		DO },
	{ "else",	ELSE,		ELSE },
	{ "exit",	EXIT,		EXIT },
	{ "exp",	FEXP,		BLTIN },
	{ "fflush",	FFLUSH,		BLTIN },
	{ "for",	FOR,		FOR },
	{ "func",	FUNC,		FUNC },
	{ "function",	FUNC,		FUNC },
	{ "gensub",	GENSUB,		GENSUB },
	{ "getline",	GETLINE,	GETLINE },
	{ "gsub",	GSUB,		GSUB },
	{ "if",		IF,		IF },
	{ "in",		IN,		IN },
	{ "index",	INDEX,		INDEX },
	{ "int",	FINT,		BLTIN },
	{ "length",	FLENGTH,	BLTIN },
	{ "log",	FLOG,		BLTIN },
	{ "lshift",	BITLSHIFT,	BITLSHIFT },
	{ "match",	MATCHFCN,	MATCHFCN },
	{ "next",	NEXT,		NEXT },
	{ "nextfile",	NEXTFILE,	NEXTFILE },
	{ "or",		BITOR,		BITOR },
	{ "print",	PRINT,		PRINT },
	{ "printf",	PRINTF,		PRINTF },
	{ "rand",	FRAND,		BLTIN },
	{ "return",	RETURN,		RETURN },
	{ "rshift",	BITRSHIFT,	BITRSHIFT },
	{ "sin",	FSIN,		BLTIN },
	{ "split",	SPLIT,		SPLIT },
	{ "sprintf",	SPRINTF,	SPRINTF },
	{ "sqrt",	FSQRT,		BLTIN },
	{ "srand",	FSRAND,		BLTIN },
	{ "strftime",	FSTRFTIME,	BLTIN },
	{ "sub",	SUB,		SUB },
	{ "substr",	SUBSTR,		SUBSTR },
	{ "system",	FSYSTEM,	BLTIN },
	{ "systime",	FSYSTIME,	BLTIN },
	{ "tolower",	FTOLOWER,	BLTIN },
	{ "toupper",	FTOUPPER,	BLTIN },
	{ "while",	WHILE,		WHILE },
	{ "xor",	BITXOR,		BITXOR }
};

#define	RET(eawk, x)							\
	if(eawk->dbg) {							\
		char	token[100];					\
		printf("lex %s\n", eawk_tokname(x, token, sizeof(token))); \
	}								\
	return(x);							\

static int
peek(eawk_t *eawk)
{
	int c = input(eawk);

	if (!unput(eawk, c)) {
		return EOF;
	}
	return c;
}

/* get next input token */
static int
gettok(eawk_t *eawk, char **pbuf, int *psz)
{
	int c, retc;
	uint8_t *buf = (uint8_t *) *pbuf;
	int sz = *psz;
	uint8_t *bp = buf;

	c = input(eawk);
	if (c == 0)
		return 0;
	buf[0] = c;
	buf[1] = 0;
	if (!isalnum(c) && c != '.' && c != '_')
		return c;

	*bp++ = c;
	if (isalpha((uint8_t)c) || c == '_') {	/* it's a varname */
		for ( ; (c = input(eawk)) != 0; ) {
			if (bp-buf >= sz) {
				if (!eawk_adjbuf(eawk, &buf, &sz, (int)(bp-buf+2), 100, &bp, "gettok")) {
					eawk_warning(eawk,  "out of space for name %.10s...", buf );
					return 0;
				}
			}
			if (isalnum((uint8_t)c) || c == '_')
				*bp++ = c;
			else {
				*bp = 0;
				if (!unput(eawk, c)) {
					return 0;
				}
				break;
			}
		}
		*bp = 0;
		retc = 'a';	/* alphanumeric */
	} else {	/* maybe it's a number, but could be . */
		char *rem;
		/* read input until can't be a number */
		for ( ; (c = input(eawk)) != 0; ) {
			if (bp-buf >= sz) {
				if (!eawk_adjbuf(eawk, &buf, &sz, (int)(bp-buf+2), 100, &bp, "gettok")) {
					eawk_warning(eawk,  "out of space for number %.10s...", buf );
					return 0;
				}
			}
			if (isdigit((uint8_t)c) || c == 'e' || c == 'E' ||
			    c == '.' || c == '+' || c == '-')
				*bp++ = c;
			else {
				if (!unput(eawk, c)) {
					return 0;
				}
				break;
			}
		}
		*bp = 0;
		strtod((char *)buf, &rem);	/* parse the number */
		if (rem == (char *)buf) {	/* it wasn't a valid number at all */
			buf[1] = 0;	/* return one character as token */
			retc = buf[0];	/* character is its own type */
			if (!unputstr(eawk, rem+1)) {
				/* put rest back for later */
				return 0;
			}
		} else {	/* some prefix was a number */
			if (!unputstr(eawk, rem)) {
				/* put rest back for later */
				return 0;
			}
			rem[0] = 0;	/* truncate buf after number part */
			retc = '0';	/* type is number */
		}
	}
	*pbuf = (char *)buf;
	*psz = sz;
	return retc;
}

static int	word(char *, YYSTYPE *, eawk_t *);
static int	string(YYSTYPE *, eawk_t *);
static int	regexpr(YYSTYPE *, eawk_t *);

static int
string(YYSTYPE *lvalp, eawk_t *eawk)
{
	int c, n;
	uint8_t *s, *bp;

	if (eawk->stringbuf == NULL && (eawk->stringbuf = malloc(eawk->stringbufsz)) == NULL) {
		eawk_warning(eawk, "out of space for strings");
		return 0;
	}
	for (bp = eawk->stringbuf; (c = input(eawk)) != '"'; ) {
		if (!eawk_adjbuf(eawk, &eawk->stringbuf, &eawk->stringbufsz, (int)(bp - eawk->stringbuf+2), 500, &bp, "string")) {
			eawk_warning(eawk, "out of space for string %.10s...", eawk->stringbuf);
			return 0;
		}
		switch (c) {
		case '\n':
		case '\r':
		case 0:
			eawk_syntax(eawk,  "non-terminated string %.10s...", eawk->stringbuf );
			eawk->lineno++;
			if (c == 0) {
				/* hopeless */
				eawk_warning(eawk,  "giving up" );
				return 0;
			}
			break;
		case '\\':
			c = input(eawk);
			switch (c) {
			case '\n': break;
			case '"': *bp++ = '"'; break;
			case 'n': *bp++ = '\n'; break;	
			case 't': *bp++ = '\t'; break;
			case 'f': *bp++ = '\f'; break;
			case 'r': *bp++ = '\r'; break;
			case 'b': *bp++ = '\b'; break;
			case 'v': *bp++ = '\v'; break;
			case 'a': *bp++ = '\007'; break;
			case '\\': *bp++ = '\\'; break;

			case '0': case '1': case '2': /* octal: \d \dd \ddd */
			case '3': case '4': case '5': case '6': case '7':
				n = c - '0';
				if ((c = peek(eawk)) >= '0' && c < '8') {
					n = 8 * n + input(eawk) - '0';
					if ((c = peek(eawk)) >= '0' && c < '8')
						n = 8 * n + input(eawk) - '0';
				}
				*bp++ = n;
				break;

			case 'x':	/* hex  \x0-9a-fA-F + */
			    {	char xbuf[100], *px;
				for (px = xbuf; (c = input(eawk)) != 0 && px-xbuf < 100-2; ) {
					if (isdigit((uint8_t)c)
					 || (c >= 'a' && c <= 'f')
					 || (c >= 'A' && c <= 'F'))
						*px++ = c;
					else
						break;
				}
				*px = 0;
				if (!unput(eawk, c)) {
					return 0;
				}
	  			sscanf(xbuf, "%x", &n);
				*bp++ = n;
				break;
			    }

			default:
				eawk_warning(eawk, "warning: escape sequence `\\%c' "
				    "treated as plain `%c'", c, c);
				*bp++ = c;
				break;
			}
			break;
		default:
			*bp++ = c;
			break;
		}
	}
	*bp = 0; 
	s = (uint8_t *)eawk_tostring(eawk, (char *)eawk->stringbuf);
	*bp++ = ' '; *bp++ = 0;
	(*lvalp).cp = eawk_var_set(eawk, (char *)eawk->stringbuf, (char *)s, 0.0, CON|STR|DONTFREE, eawk->symtab);
	RET(eawk, STRING);
}


static int
binsearch(const char *w, const Keyword *kp, int n)
{
	int cond, low, mid, high;

	low = 0;
	high = n - 1;
	while (low <= high) {
		mid = (low + high) / 2;
		if ((cond = strcmp(w, kp[mid].word)) < 0)
			high = mid - 1;
		else if (cond > 0)
			low = mid + 1;
		else
			return mid;
	}
	return -1;
}

static int
word(char *w, YYSTYPE *lvalp, eawk_t *eawk) 
{
	const Keyword *kp;
	int c, n;

	kp = NULL;
	n = binsearch(w, keywords, (int)(sizeof(keywords)/sizeof(keywords[0])));
	if (n != -1) {	/* found in table */
		/* this ought to be inside the if; in theory could fault (daniel barrett) */
		kp = keywords + n;
		(*lvalp).i = kp->sub;
		switch (kp->type) {	/* special handling */
		case BLTIN:
			if (kp->sub == FSYSTEM && eawk->safe)
				eawk_syntax(eawk,  "system is unsafe" );
			RET(eawk, kp->type);
			/*FALLTHROUGH*/
		case FUNC:
			if (eawk->infunc)
				eawk_syntax(eawk,  "illegal nested function" );
			RET(eawk, kp->type);
			/*FALLTHROUGH*/
		case RETURN:
			if (!eawk->infunc)
				eawk_syntax(eawk,  "return not in function" );
			RET(eawk, kp->type);
			/*FALLTHROUGH*/
		case VARNF:
			(*lvalp).cp = eawk_var_set(eawk, "NF", "", 0.0, NUM, eawk->symtab);
			RET(eawk, VARNF);
			/*FALLTHROUGH*/
		default:
			RET(eawk, kp->type);
		}
	}
	c = peek(eawk);	/* look for '(' */
	if (c != '(' && eawk->infunc && (n=eawk_isarg(eawk, w)) >= 0) {
		(*lvalp).i = n;
		RET(eawk, ARG);
	} else {
		(*lvalp).cp = eawk_var_set(eawk, w, "", 0.0, STR|NUM|DONTFREE, eawk->symtab);
		if (c == '(') {
			RET(eawk, CALL);
		} else {
			RET(eawk, VAR);
		}
	}
}

/* next call to eawk_lex will return a regular expression */
void
startreg(eawk_t *eawk)
{
	eawk->reg = 1;
}

static int
regexpr(YYSTYPE *lvalp, eawk_t *eawk)
{
	int c;
	uint8_t *bp;

	if (eawk->regexprbuf == 0 && (eawk->regexprbuf = malloc(eawk->regexprbufsz)) == NULL) {
		eawk_warning(eawk, "out of space for rex expr");
		return 0;
	}
	bp = eawk->regexprbuf;
	for ( ; (c = input(eawk)) != '/' && c != 0; ) {
		if (!eawk_adjbuf(eawk, &eawk->regexprbuf, &eawk->regexprbufsz, (int)(bp - eawk->regexprbuf+3), 500, &bp, "regexpr")) {
			eawk_warning(eawk, "out of space for reg expr %.10s...", eawk->regexprbuf);
			return 0;
		}
		if (c == '\n') {
			eawk_syntax(eawk,  "newline in regular expression %.10s...", eawk->regexprbuf ); 
			if (!unput(eawk, '\n')) {
				return 0;
			}
			break;
		} else if (c == '\\') {
			*bp++ = '\\'; 
			*bp++ = input(eawk);
		} else {
			*bp++ = c;
		}
	}
	*bp = 0;
	if (c == 0)
		eawk_syntax(eawk, "non-terminated regular expression %.10s...", eawk->regexprbuf);
	(*lvalp).s = eawk_tostring(eawk, (char *)eawk->regexprbuf);
	if (!unput(eawk, '/')) {
		return 0;
	}
	RET(eawk, REGEXPR);
}

/* low-level lexical stuff, sort of inherited from lex */

static int
pgetc(eawk_t *eawk)		/* get 1 character from awk program */
{
	int c;

	for (;;) {
		if (eawk->yyin == NULL) {
			if (eawk->curscript >= eawk->scriptc)
				return EOF;
			if (strcmp(eawk->scripts[eawk->curscript], "-") == 0)
				eawk->yyin = stdin;
			else if ((eawk->yyin = fopen(eawk->scripts[eawk->curscript], "r")) == NULL) {
				eawk_warning(eawk, "can't open file %s", eawk->scripts[eawk->curscript]);
				return EOF;
			}
			eawk->lineno = 1;
		}
		if ((c = getc(eawk->yyin)) != EOF)
			return c;
		if (eawk->yyin != stdin)
			fclose(eawk->yyin);
		eawk->yyin = NULL;
		eawk->curscript++;
	}
}

int
input(eawk_t *eawk)	/* get next lexical input character */
{
	int c;

	if (eawk->yysptr > eawk->yysbuf)
		c = (uint8_t)*--eawk->yysptr;
	else if (eawk->lexprog != NULL) {	/* awk '...' */
		if ((c = (uint8_t)*eawk->lexprog) != 0)
			eawk->lexprog++;
	} else				/* awk -f ... */
		c = pgetc(eawk);
	if (c == '\n')
		eawk->lineno++;
	else if (c == EOF)
		c = 0;
	if (eawk->ep >= eawk->ebuf + sizeof(eawk->ebuf))
		eawk->ep = eawk->ebuf;
	return *eawk->ep++ = c;
}

int
unput(eawk_t *eawk, int c)	/* put lexical character back on input */
{
	if (c == '\n')
		eawk->lineno--;
	if (eawk->yysptr >= eawk->yysbuf + sizeof(eawk->yysbuf)) {
		eawk_warning(eawk, "pushed back too much: %.20s...", eawk->yysbuf);
		return 0;
	}
	*eawk->yysptr++ = c;
	if (--eawk->ep < eawk->ebuf)
		eawk->ep = eawk->ebuf + sizeof(eawk->ebuf) - 1;
	return 1;
}

int
unputstr(eawk_t *eawk, const char *s)	/* put a string back on input */
{
	int i;

	for (i = (int)strlen(s)-1; i >= 0; i--) {
		if (!unput(eawk, s[i])) {
			return 0;
		}
	}
	return 1;
}

int
eawk_lex(void *lval, void *vp)
{
	YYSTYPE	*lvalp = (YYSTYPE *)lval;
	eawk_t	*eawk = (eawk_t *)vp;
	int c;

	if (eawk->lexbuf == NULL && (eawk->lexbuf = malloc(eawk->lexbufsize)) == NULL) {
		eawk_warning(eawk,  "out of space in eawk_lex" );
		return 0;
	}
	if (eawk->sc) {
		eawk->sc = 0;
		RET(eawk, '}');
	}
	if (eawk->reg) {
		eawk->reg = 0;
		return regexpr(lvalp, eawk);
	}
	for (;;) {
		c = gettok(eawk, &eawk->lexbuf, &eawk->lexbufsize);
		if (c == 0)
			return 0;
		if (isalpha((uint8_t)c) || c == '_')
			return word(eawk->lexbuf, lvalp, eawk);
		if (isdigit((uint8_t)c)) {
			(*lvalp).cp = eawk_var_set(eawk, eawk->lexbuf, eawk_tostring(eawk, eawk->lexbuf), atof(eawk->lexbuf), CON|NUM, eawk->symtab);
			/* should this also have STR set? */
			RET(eawk, NUMBER);
		}
	
		(*lvalp).i = c;
		switch (c) {
		case '\n':	/* {EOL} */
			RET(eawk, NL);
		case '\r':	/* assume \n is coming */
		case ' ':	/* {WS}+ */
		case '\t':
			break;
		case '#':	/* #.* strip comments */
			while ((c = input(eawk)) != '\n' && c != 0)
				;
			if (!unput(eawk, c)) {
				return 0;
			}
			break;
		case ';':
			RET(eawk, ';');
		case '\\':
			if (peek(eawk) == '\n') {
				input(eawk);
			} else if (peek(eawk) == '\r') {
				input(eawk); input(eawk);	/* \n */
				eawk->lineno++;
			} else {
				RET(eawk, c);
			}
			break;
		case '&':
			if (peek(eawk) == '&') {
				input(eawk); RET(eawk, AND);
			} else 
				RET(eawk, '&');
		case '|':
			if (peek(eawk) == '|') {
				input(eawk); RET(eawk, BOR);
			} else
				RET(eawk, '|');
		case '!':
			if (peek(eawk) == '=') {
				input(eawk); (*lvalp).i = NE; RET(eawk, NE);
			} else if (peek(eawk) == '~') {
				input(eawk); (*lvalp).i = NOTMATCH; RET(eawk, MATCHOP);
			} else
				RET(eawk, NOT);
		case '~':
			(*lvalp).i = MATCH;
			RET(eawk, MATCHOP);
		case '<':
			if (peek(eawk) == '=') {
				input(eawk); (*lvalp).i = LE; RET(eawk, LE);
			} else {
				(*lvalp).i = LT; RET(eawk, LT);
			}
		case '=':
			if (peek(eawk) == '=') {
				input(eawk); (*lvalp).i = EQ; RET(eawk, EQ);
			} else {
				(*lvalp).i = ASSIGN; RET(eawk, ASGNOP);
			}
		case '>':
			if (peek(eawk) == '=') {
				input(eawk); (*lvalp).i = GE; RET(eawk, GE);
			} else if (peek(eawk) == '>') {
				input(eawk); (*lvalp).i = APPEND; RET(eawk, APPEND);
			} else {
				(*lvalp).i = GT; RET(eawk, GT);
			}
		case '+':
			if (peek(eawk) == '+') {
				input(eawk); (*lvalp).i = INCR; RET(eawk, INCR);
			} else if (peek(eawk) == '=') {
				input(eawk); (*lvalp).i = ADDEQ; RET(eawk, ASGNOP);
			} else
				RET(eawk, '+');
		case '-':
			if (peek(eawk) == '-') {
				input(eawk); (*lvalp).i = DECR; RET(eawk, DECR);
			} else if (peek(eawk) == '=') {
				input(eawk); (*lvalp).i = SUBEQ; RET(eawk, ASGNOP);
			} else
				RET(eawk, '-');
		case '*':
			if (peek(eawk) == '=') {	/* *= */
				input(eawk); (*lvalp).i = MULTEQ; RET(eawk, ASGNOP);
			} else if (peek(eawk) == '*') {	/* ** or **= */
				input(eawk);	/* eat 2nd * */
				if (peek(eawk) == '=') {
					input(eawk); (*lvalp).i = POWEQ; RET(eawk, ASGNOP);
				} else {
					RET(eawk, POWER);
				}
			} else
				RET(eawk, '*');
		case '/':
			RET(eawk, '/');
		case '%':
			if (peek(eawk) == '=') {
				input(eawk); (*lvalp).i = MODEQ; RET(eawk, ASGNOP);
			} else
				RET(eawk, '%');
		case '^':
			if (peek(eawk) == '=') {
				input(eawk); (*lvalp).i = POWEQ; RET(eawk, ASGNOP);
			} else
				RET(eawk, POWER);

		case '$':
			/* BUG: awkward, if not wrong */
			c = gettok(eawk, &eawk->lexbuf, &eawk->lexbufsize);
			if (isalpha((uint8_t)c)) {
				if (strcmp(eawk->lexbuf, "NF") == 0) {	/* very special */
					if (!unputstr(eawk, "(NF)")) {
						return 0;
					}
					RET(eawk, INDIRECT);
				}
				c = peek(eawk);
				if (c == '(' || c == '[' || (eawk->infunc && eawk_isarg(eawk, eawk->lexbuf) >= 0)) {
					if (!unputstr(eawk, eawk->lexbuf)) {
						return 0;
					}
					RET(eawk, INDIRECT);
				}
				(*lvalp).cp = eawk_var_set(eawk, eawk->lexbuf, "", 0.0, STR|NUM, eawk->symtab);
				RET(eawk, IVAR);
			} else if (c == 0) {	/*  */
				eawk_syntax(eawk,  "unexpected end of input after $" );
				RET(eawk, ';');
			} else {
				if (!unputstr(eawk, eawk->lexbuf)) {
					return 0;
				}
				RET(eawk, INDIRECT);
			}
	
		case '}':
			if (--eawk->bracecnt < 0)
				eawk_syntax(eawk,  "extra }" );
			eawk->sc = 1;
			RET(eawk, ';');
		case ']':
			if (--eawk->brackcnt < 0)
				eawk_syntax(eawk,  "extra ]" );
			RET(eawk, ']');
		case ')':
			if (--eawk->parencnt < 0)
				eawk_syntax(eawk,  "extra )" );
			RET(eawk, ')');
		case '{':
			eawk->bracecnt++;
			RET(eawk, '{');
		case '[':
			eawk->brackcnt++;
			RET(eawk, '[');
		case '(':
			eawk->parencnt++;
			RET(eawk, '(');
	
		case '"':
			return string(lvalp, eawk);	/* BUG: should be like tran.c ? */
	
		default:
			RET(eawk, c);
		}
	}
}
