/****************************************************************
Copyright (C) Lucent Technologies 1997
All Rights Reserved

Permission to use, copy, modify, and distribute this software and
its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that the copyright notice and this
permission notice and warranty disclaimer appear in supporting
documentation, and that the name Lucent Technologies or any of
its entities not be used in advertising or publicity pertaining
to distribution of the software without specific, written prior
permission.

LUCENT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
IN NO EVENT SHALL LUCENT OR ANY OF ITS ENTITIES BE LIABLE FOR ANY
SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
THIS SOFTWARE.
****************************************************************/

#define DEBUG
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <stdarg.h>

#include "awkre.h"
#include "eawk.h"
#include "awkgram.h"
#include "proto.h"

static int	makefields(eawk_t *, int, int);

/* get ARGV[n] */
static char *
getargv(eawk_t *eawk, int n)
{
	awkcell_t *x;
	char *s, temp[50];

	snprintf(temp, sizeof(temp), "%d", n);
	x = eawk_var_set(eawk, temp, "", 0.0, STR, eawk->ARGVtab);
	s = eawk_getsval(eawk, x);
	DPRINTF(eawk, ("getargv(%d) returns |%s|\n", n, s) );
	return s;
}

static void
initgetrec(eawk_t *eawk)
{
	int i;
	char *p;

	for (i = 1; i < *eawk->ARGC; i++) {
		if (!eawk_isclvar(p = getargv(eawk, i))) {	/* find 1st real filename */
			eawk_setsval(eawk, eawk_lookup("FILENAME", eawk->symtab), getargv(eawk, i));
			return;
		}
		eawk_setclvar(eawk, p);	/* a commandline assignment before filename */
		eawk->argno++;
	}
	eawk->infile = stdin;		/* no filenames, so use stdin */
}

/* clean out fields n1 .. n2 inclusive */
/* nvals remain intact */
static void
cleanfld(eawk_t *eawk, int n1, int n2)
{
	awkcell_t *p;
	int i;

	for (i = n1; i <= n2; i++) {
		p = eawk->fldtab[i];
		if (FREEABLE(p))
			XFREE(p->str);
		p->str = __UNCONST("");
		p->type = FLD | STR | DONTFREE;
	}
}

/* make new fields up to at least $n */
static int
growfldtab(eawk_t *eawk, unsigned n)
{
	unsigned nf = 2 * eawk->nfields;
	size_t s;

	if (n > nf)
		nf = n;
	s = (nf+1) * (sizeof(awkcell_t *));  /* freebsd: how much do we need? */
	if (s / sizeof(awkcell_t *) - 1 == nf) /* didn't overflow */
		eawk->fldtab = realloc(eawk->fldtab, s);
	else					/* overflow sizeof int */
		XFREE(eawk->fldtab);	/* make it null */
	if (eawk->fldtab == NULL) {
		eawk_warning(eawk, "out of space creating %d fields", nf);
		return 0;
	}
	if (!makefields(eawk, eawk->nfields+1, (int)nf)) {
		return 0;
	}
	eawk->nfields = nf;
	return 1;
}

/* build fields from reg expr in FS */
static int
refldbld(eawk_t *eawk, const char *rec, const char *fs)
{
	/* this relies on having fields[] the same length as $0 */
	/* the fields are all stored in this one array with \0's */
	char *fr;
	int i, tempstat, n;
	fa_t *pfa;

	n = (int)strlen(rec);
	if (n > eawk->fieldssize) {
		XFREE(eawk->fields);
		if ((eawk->fields = calloc(1, n+1)) == NULL) {
			eawk_warning(eawk, "out of space for fields in refldbld %d", n);
			return 0;
		}
		eawk->fieldssize = n;
	}
	fr = eawk->fields;
	*fr = '\0';
	if (*rec == '\0')
		return 0;
	pfa = eawk_regcomp(eawk, fs, 1);
	DPRINTF(eawk, ("into refldbld, rec = <%s>, pat = <%s>\n", rec, fs) );
	tempstat = pfa->initstat;
	for (i = 1; ; i++) {
		if (i > eawk->nfields)
			growfldtab(eawk, (unsigned)i);
		if (FREEABLE(eawk->fldtab[i]))
			XFREE(eawk->fldtab[i]->str);
		eawk->fldtab[i]->type = FLD | STR | DONTFREE;
		eawk->fldtab[i]->str = fr;
		DPRINTF(eawk, ("refldbld: i=%d\n", i) );
		if (eawk_nematch(eawk, pfa, rec)) {
			pfa->initstat = 2;	/* horrible coupling to b.c */
			DPRINTF(eawk, ("match %s (%d chars)\n", eawk->patbeg, eawk->patlen) );
			strncpy(fr, rec, (unsigned)(((char*)eawk->patbeg)-rec));
			fr += ((char*)eawk->patbeg) - rec + 1;
			*(fr-1) = '\0';
			rec = (char *)eawk->patbeg + eawk->patlen;
		} else {
			DPRINTF(eawk, ("no match %s\n", rec) );
			strcpy(fr, rec);
			pfa->initstat = tempstat;
			break;
		}
	}
	return i;		
}

static void
bclass(eawk_t *eawk, int c)
{
	switch (c) {
	case '{': eawk->bracecnt++; break;
	case '}': eawk->bracecnt--; break;
	case '[': eawk->brackcnt++; break;
	case ']': eawk->brackcnt--; break;
	case '(': eawk->parencnt++; break;
	case ')': eawk->parencnt--; break;
	}
}

/* try to print context around error */
static void
eprint(eawk_t *eawk)
{
	char	*stage;
	char *p, *q;

	stage = eawk_getsval(eawk, eawk_lookup("stage", eawk->control));
	if (strcmp(stage, "error display") == 0 || strcmp(stage, "running") == 0 || eawk->eprintc++ > 0) {
		return;
	}
	p = eawk->ep - 1;
	if (p > eawk->ebuf && *p == '\n')
		p--;
	for ( ; p > eawk->ebuf && *p != '\n' && *p != '\0'; p--)
		;
	while (*p == '\n')
		p++;
	(void) fprintf(stderr, " context is\n\t");
	for (q=eawk->ep-1; q>=p && *q!=' ' && *q!='\t' && *q!='\n'; q--)
		;
	for ( ; p < q; p++)
		if (*p)
			putc(*p, stderr);
	(void) fprintf(stderr, " >>> ");
	for ( ; p < eawk->ep; p++)
		if (*p)
			putc(*p, stderr);
	(void) fprintf(stderr, " <<< ");
#if 0
	/*
	 * The following code was used to print the rest of the line of
	 * error context. It naively counts brackets, parens and braces in
	 * order to minimize the parsing effect of dropping the rest of the
	 * line but it does not work in all the cases. It is too much work
	 * to save the current program input point and restore it in all the
	 * cases just for the benefit of error printing so for now this
	 * code is disabled. In particular this code is confused if the
	 * [ { ( ) } ] is inside a quoted string or a pattern.
	 */
	if (*eawk->ep) {
		int c;
		while ((c = input()) != '\n' && c != '\0' && c != EOF) {
			putc(c, stderr);
			bclass(eawk, c);
		}
	}
#endif
	putc('\n', stderr);
	eawk->ep = eawk->ebuf;
}

/* current source file name */
static char *
cursource(eawk_t *eawk)
{
	return (eawk->scriptc > 0) ? eawk->scripts[eawk->curscript] : NULL;
}

static void
bcheck2(int n, int c1, int c2)
{
	USE_ARG(c1);
	if (n == 1)
		(void) fprintf(stderr, "\tmissing %c\n", c2);
	else if (n > 1)
		(void) fprintf(stderr, "\t%d missing %c's\n", n, c2);
	else if (n == -1)
		(void) fprintf(stderr, "\textra %c\n", c2);
	else if (n < -1)
		(void) fprintf(stderr, "\t%d extra %c's\n", -n, c2);
}

static void
error(eawk_t *eawk)
{
	char	*stage;

	stage = eawk_getsval(eawk, eawk_lookup("stage", eawk->control));
	(void) fprintf(stderr, "\n");
	if (strcmp(stage, "error display") == 0 && eawk->NR && *eawk->NR > 0) {
		(void) fprintf(stderr, " input record number %d", (int) (*eawk->FNR));
		if (strcmp(*eawk->FILENAME, "-") != 0)
			(void) fprintf(stderr, ", file %s", *eawk->FILENAME);
		(void) fprintf(stderr, "\n");
	}
	if (strcmp(stage, "error display") == 0 && eawk->curnode) 
		(void) fprintf(stderr, " source line number %d", eawk->curnode->lineno);
	else if (strcmp(stage, "error display") != 0 && eawk->lineno) 
		(void) fprintf(stderr, " source line number %d", eawk->lineno);
	if (strcmp(stage, "compiling") == 0 && cursource(eawk) != NULL)
		(void) fprintf(stderr, " source file %s", cursource(eawk));
	(void) fprintf(stderr, "\n");
	eprint(eawk);
}


int
eawk_recinit(eawk_t *eawk, unsigned int n)
{
	if ((eawk->record = calloc(1, n)) == NULL ||
	    (eawk->fields = calloc(1, n)) == NULL ||
	    (eawk->fldtab = calloc((eawk->nfields+1), sizeof(awkcell_t *))) == NULL ||
	    (eawk->fldtab[0] = calloc(1, sizeof(awkcell_t))) == NULL) {
		eawk_warning(eawk, "out of space for $0 and fields");
		return 0;
	}
	*eawk->fldtab[0] = eawk->dollar0;
	eawk->fldtab[0]->str = (char *)eawk->record;
	eawk->fldtab[0]->name = eawk_tostring(eawk, "0");
	return makefields(eawk, 1, eawk->nfields);
}

/* create $n1..$n2 inclusive */
static int
makefields(eawk_t *eawk, int n1, int n2)
{
	char temp[50];
	int i;

	for (i = n1; i <= n2; i++) {
		eawk->fldtab[i] = malloc(sizeof (awkcell_t));
		if (eawk->fldtab[i] == NULL) {
			eawk_warning(eawk, "out of space in makefields %d", i);
			return 0;
		}
		*eawk->fldtab[i] = eawk->dollar1;
		snprintf(temp, sizeof(temp), "%d", i);
		eawk->fldtab[i]->name = eawk_tostring(eawk, temp);
	}
	return 1;
}

/* get next input record */
int
eawk_getrec(eawk_t *eawk, uint8_t **pbuf, int *pbufsize, int isrecord)
{			/* note: cares whether buf == record */
	int c;
	uint8_t *buf = *pbuf;
	uint8_t saveb0;
	int bufsize = *pbufsize, savebufsize = bufsize;

	if (!eawk->init_getrec) {
		eawk->init_getrec = 1;
		initgetrec(eawk);
	}
	DPRINTF(eawk, ("RS=<%s>, FS=<%s>, ARGC=%g, FILENAME=%s\n",
		*eawk->RS, *eawk->FS, *eawk->ARGC, *eawk->FILENAME) );
	if (isrecord) {
		eawk->donefld = 0;
		eawk->donerec = 1;
	}
	saveb0 = buf[0];
	buf[0] = 0;
	while (eawk->argno < *eawk->ARGC || eawk->infile == stdin) {
		DPRINTF(eawk, ("argno=%d, file=|%s|\n", eawk->argno, eawk->file) );
		if (eawk->infile == NULL) {	/* have to open a new file */
			eawk->file = getargv(eawk, eawk->argno);
			if (*eawk->file == '\0') {	/* it's been zapped */
				eawk->argno++;
				continue;
			}
			if (eawk_isclvar(eawk->file)) {	/* a var=value arg */
				eawk_setclvar(eawk, eawk->file);
				eawk->argno++;
				continue;
			}
			*eawk->FILENAME = eawk->file;
			DPRINTF(eawk, ("opening file %s\n", eawk->file) );
			if (*eawk->file == '-' && *(eawk->file+1) == '\0')
				eawk->infile = stdin;
			else if ((eawk->infile = fopen(eawk->file, "r")) == NULL) {
				eawk_warning(eawk, "can't open file %s", eawk->file);
				return 0;
			}
			eawk_setfval(eawk, eawk->fnrloc, 0.0);
		}
		c = eawk_readrec(eawk, &buf, &bufsize, eawk->infile);
		if (c != 0 || buf[0] != '\0') {	/* normal record */
			if (isrecord) {
				if (FREEABLE(eawk->fldtab[0]))
					XFREE(eawk->fldtab[0]->str);
				eawk->fldtab[0]->str = (char *)buf;	/* buf == record */
				eawk->fldtab[0]->type = REC | STR | DONTFREE;
				if (eawk_is_number(eawk->fldtab[0]->str)) {
					eawk->fldtab[0]->num = atof(eawk->fldtab[0]->str);
					eawk->fldtab[0]->type |= NUM;
				}
			}
			eawk_setfval(eawk, eawk->nrloc, eawk->nrloc->num+1);
			eawk_setfval(eawk, eawk->fnrloc, eawk->fnrloc->num+1);
			*pbuf = buf;
			*pbufsize = bufsize;
			return 1;
		}
		/* EOF arrived on this file; set up next */
		if (eawk->infile != stdin)
			fclose(eawk->infile);
		eawk->infile = NULL;
		eawk->argno++;
	}
	buf[0] = saveb0;
	*pbuf = buf;
	*pbufsize = savebufsize;
	return 0;	/* true end of file */
}

void
eawk_nextfile(eawk_t *eawk)
{
	if (eawk->infile != NULL && eawk->infile != stdin)
		fclose(eawk->infile);
	eawk->infile = NULL;
	eawk->argno++;
}

/* read one record into buf */
int
eawk_readrec(eawk_t *eawk, uint8_t **pbuf, int *pbufsize, FILE *inf)
{
	int sep, c;
	uint8_t *rr, *buf = *pbuf;
	int bufsize = *pbufsize;
	size_t len;

	if ((len = strlen(*eawk->FS)) <= eawk->len_inputFS) {
		strcpy(eawk->inputFS, *eawk->FS);	/* for subsequent field splitting */
	} else {
		if (eawk->inputFS != eawk->static_inputFS)
			free(eawk->inputFS);
		eawk->inputFS = malloc(len + 1);
		if (eawk->inputFS == NULL) {
			eawk_warning(eawk, "field separator %.10s... is too long", *eawk->FS);
			return 0;
		}
		eawk->len_inputFS = len;
		memcpy(eawk->inputFS, *eawk->FS, len + 1);
	}
	if ((sep = **eawk->RS) == 0) {
		sep = '\n';
		while ((c=getc(inf)) == '\n' && c != EOF)	/* skip leading \n's */
			;
		if (c != EOF)
			ungetc(c, inf);
	}
	for (rr = buf; ; ) {
		for (; (c=getc(inf)) != sep && c != EOF; ) {
			if (rr-buf+1 > bufsize) {
				if (!eawk_adjbuf(eawk, &buf, &bufsize, (int)(1+rr-buf), eawk->recsize, &rr, "readrec 1")) {
					eawk_warning(eawk, "input record `%.30s...' too long", buf);
					return 0;
				}
			}
			*rr++ = c;
		}
		if (**eawk->RS == sep || c == EOF)
			break;
		if ((c = getc(inf)) == '\n' || c == EOF) /* 2 in a row */
			break;
		if (!eawk_adjbuf(eawk, &buf, &bufsize, (int)(2+rr-buf), eawk->recsize, &rr, "readrec 2")) {
			eawk_warning(eawk, "input record `%.30s...' too long", buf);
			return 0;
		}
		*rr++ = '\n';
		*rr++ = c;
	}
	if (!eawk_adjbuf(eawk, &buf, &bufsize, (int)(1+rr-buf), eawk->recsize, &rr, "readrec 3")) {
		eawk_warning(eawk, "input record `%.30s...' too long", buf);
		return 0;
	}
	*rr = 0;
	DPRINTF(eawk, ("readrec saw <%s>, returns %d\n", buf, c == EOF && rr == buf ? 0 : 1) );
	*pbuf = buf;
	*pbufsize = bufsize;
	return c == EOF && rr == buf ? 0 : 1;
}

/* set var=value from s */
void
eawk_setclvar(eawk_t *eawk, char *s)
{
	char *p;
	awkcell_t *q;

	for (p=s; *p != '='; p++)
		;
	*p++ = 0;
	p = eawk_qstring(eawk, p, '\0');
	q = eawk_var_set(eawk, s, p, 0.0, STR, eawk->symtab);
	eawk_setsval(eawk, q, p);
	if (eawk_is_number(q->str)) {
		q->num = atof(q->str);
		q->type |= NUM;
	}
	DPRINTF(eawk, ("command line set %s to |%s|\n", s, p) );
}


/* create fields from current record */
int
eawk_fldbld(eawk_t *eawk)
{
	/* this relies on having fields[] the same length as $0 */
	/* the fields are all stored in this one array with \0's */
	char *r, *fr, sep;
	awkcell_t *p;
	unsigned	i;
	int j, n;

	if (eawk->donefld)
		return 1;
	if (!ISSTR(eawk->fldtab[0]))
		eawk_getsval(eawk, eawk->fldtab[0]);
	r = eawk->fldtab[0]->str;
	n = (int)strlen(r);
	if (n > eawk->fieldssize) {
		XFREE(eawk->fields);
		if ((eawk->fields = malloc(n+1)) == NULL) {
			eawk_warning(eawk, "out of space for fields in eawk_fldbld %d", n);
			return 0;
		}
		eawk->fieldssize = n;
	}
	fr = eawk->fields;
	i = 0;	/* number of fields accumulated here */
	if (eawk->inputFS[0] && eawk->inputFS[1]) {	/* it's a regular expression */
		i = refldbld(eawk, r, eawk->inputFS);
	} else if ((sep = *eawk->inputFS) == ' ') {	/* default whitespace */
		for (i = 0; ; ) {
			while (*r == ' ' || *r == '\t' || *r == '\n')
				r++;
			if (*r == 0)
				break;
			i++;
			if (i > (unsigned)eawk->nfields)
				growfldtab(eawk, i);
			if (FREEABLE(eawk->fldtab[i]))
				XFREE(eawk->fldtab[i]->str);
			eawk->fldtab[i]->str = fr;
			eawk->fldtab[i]->type = FLD | STR | DONTFREE;
			do {
				*fr++ = *r++;
			} while (*r != ' ' && *r != '\t' && *r != '\n' && *r != '\0');
			*fr++ = 0;
		}
		*fr = 0;
	} else if ((sep = *eawk->inputFS) == 0) {		/* new: FS="" => 1 char/field */
		for (i = 0; *r != 0; r++) {
			char buf[2];
			i++;
			if (i > (unsigned)eawk->nfields)
				growfldtab(eawk, i);
			if (FREEABLE(eawk->fldtab[i]))
				XFREE(eawk->fldtab[i]->str);
			buf[0] = *r;
			buf[1] = 0;
			eawk->fldtab[i]->str = eawk_tostring(eawk, buf);
			eawk->fldtab[i]->type = FLD | STR;
		}
		*fr = 0;
	} else if (*r != 0) {	/* if 0, it's a null field */
		/* subtlecase : if length(FS) == 1 && length(RS > 0)
		 * \n is NOT a field separator (cf awk book 61,84).
		 * this variable is tested in the inner while loop.
		 */
		int rtest = '\n';  /* normal case */
		if (strlen(*eawk->RS) > 0)
			rtest = '\0';
		for (;;) {
			i++;
			if (i > (unsigned)eawk->nfields)
				growfldtab(eawk, i);
			if (FREEABLE(eawk->fldtab[i]))
				XFREE(eawk->fldtab[i]->str);
			eawk->fldtab[i]->str = fr;
			eawk->fldtab[i]->type = FLD | STR | DONTFREE;
			while (*r != sep && *r != rtest && *r != '\0')	/* \n is always a separator */
				*fr++ = *r++;
			*fr++ = 0;
			if (*r++ == 0)
				break;
		}
		*fr = 0;
	}
	if (i > (unsigned)eawk->nfields) {
		eawk_warning(eawk, "record `%.30s...' has too many fields; can't happen", r);
		return 0;
	}
	cleanfld(eawk, (int)(i+1), eawk->lastfld);	/* clean out junk from previous record */
	eawk->lastfld = i;
	eawk->donefld = 1;
	for (j = 1; j <= eawk->lastfld; j++) {
		p = eawk->fldtab[j];
		if(eawk_is_number(p->str)) {
			p->num = atof(p->str);
			p->type |= NUM;
		}
	}
	eawk_setfval(eawk, eawk->nfloc, (awknum_t) eawk->lastfld);
	if (eawk->dbg) {
		for (j = 0; j <= eawk->lastfld; j++) {
			p = eawk->fldtab[j];
			printf("field %d (%s): |%s|\n", j, p->name, p->str);
		}
	}
	return 1;
}

/* add field n after end of existing lastfld */
void
eawk_newfld(eawk_t *eawk, int n)
{
	if (n > eawk->nfields)
		growfldtab(eawk, (unsigned)n);
	cleanfld(eawk, eawk->lastfld+1, n);
	eawk->lastfld = n;
	eawk_setfval(eawk, eawk->nfloc, (awknum_t) n);
}

/* get nth field */
awkcell_t *
eawk_fieldadr(eawk_t *eawk, int n)
{
	if (n < 0) {
		eawk_warning(eawk, "trying to access out of range field %d", n);
		return NULL;
	}
	if (n > eawk->nfields)	/* fields after NF are empty */
		growfldtab(eawk, (unsigned)(n));	/* but does not increase NF */
	return eawk->fldtab[n];
}

/* create $0 from $1..$NF if necessary */
int
eawk_recbld(eawk_t *eawk)
{
	int i;
	uint8_t *r;
	char *p;

	if (eawk->donerec == 1)
		return 1;
	r = eawk->record;
	for (i = 1; i <= *eawk->NF; i++) {
		p = eawk_getsval(eawk, eawk->fldtab[i]);
		if (!eawk_adjbuf(eawk, &eawk->record, &eawk->recsize, (int)(1+strlen(p)+(int)(r - eawk->record)), eawk->recsize, &r, "recbld 1")) {
			eawk_warning(eawk, "created $0 `%.30s...' too long", eawk->record);
			return 0;
		}
		while ((*r = *p++) != 0)
			r++;
		if (i < *eawk->NF) {
			if (!eawk_adjbuf(eawk, &eawk->record, &eawk->recsize, (int)(2+strlen(*eawk->OFS)+(int)(r - eawk->record)), eawk->recsize, &r, "recbld 2")) {
				eawk_warning(eawk, "created $0 `%.30s...' too long", eawk->record);
				return 0;
			}
			for (p = *eawk->OFS; (*r = *p++) != 0; )
				r++;
		}
	}
	if (!eawk_adjbuf(eawk, &eawk->record, &eawk->recsize, 2+(int)(r - eawk->record), eawk->recsize, &r, "recbld 3")) {
		eawk_warning(eawk, "built giant record `%.30s...'", eawk->record);
		return 0;
	}
	*r = '\0';
	DPRINTF(eawk, ("in recbld inputFS=%s, fldtab[0]=%p\n", eawk->inputFS, eawk->fldtab[0]) );

	if (FREEABLE(eawk->fldtab[0]))
		XFREE(eawk->fldtab[0]->str);
	eawk->fldtab[0]->type = REC | STR | DONTFREE;
	eawk->fldtab[0]->str = (char *)eawk->record;

	DPRINTF(eawk, ("in recbld inputFS=%s, fldtab[0]=%p\n", eawk->inputFS, eawk->fldtab[0]) );
	DPRINTF(eawk, ("recbld = |%s|\n", eawk->record) );
	eawk->donerec = 1;
	return 1;
}

void
eawk_error(const char *s)
{
	(void) fprintf(stderr, "%s", s);
}

void
eawk_syntax(eawk_t *eawk, const char *fmt, ...)
{
	va_list varg;
	char	*stage;

	if (eawk->syntaxc++ > 2)
		return;
	(void) fprintf(stderr, "%s: ", eawk_getsval(eawk, eawk_lookup("cmdname", eawk->control)));
	va_start(varg, fmt);
	vfprintf(stderr, fmt, varg);
	va_end(varg);
	(void) fprintf(stderr, " at source line %d", eawk->lineno);
	if (eawk->curfname != NULL)
		(void) fprintf(stderr, " in function %s", eawk->curfname);
	stage = eawk_getsval(eawk, eawk_lookup("stage", eawk->control));
	if (strcmp(stage, "compiling") == 0 && cursource(eawk) != NULL)
		(void) fprintf(stderr, " source file %s", cursource(eawk));
	(void) fprintf(stderr, "\n");
	eawk->errorflag = 2;
	eprint(eawk);
}

void
eawk_err_display(eawk_t *eawk)
{
	int c;

	if (eawk->bracechecked++)
		return;
	while ((c = input(eawk)) != EOF && c != '\0')
		bclass(eawk, c);
	bcheck2(eawk->bracecnt, '{', '}');
	bcheck2(eawk->brackcnt, '[', ']');
	bcheck2(eawk->parencnt, '(', ')');
}

void
eawk_warning(eawk_t *eawk, const char *fmt, ...)
{
	va_list varg;

	fflush(stdout);
	(void) fprintf(stderr, "%s: ",
		eawk_getsval(eawk, eawk_lookup("cmdname", eawk->control)));
	va_start(varg, fmt);
	vfprintf(stderr, fmt, varg);
	va_end(varg);
	error(eawk);
}

awknum_t
errcheck(eawk_t *eawk, awknum_t x, const char *s)
{

	if (errno == EDOM) {
		errno = 0;
		eawk_warning(eawk, "%s argument out of domain", s);
		x = 1;
	} else if (errno == ERANGE) {
		errno = 0;
		eawk_warning(eawk, "%s result out of range", s);
		x = 1;
	}
	return x;
}

/* is s of form var=something ? */
int
eawk_isclvar(const char *s)
{
	const char *os = s;

	if (!(isalpha((uint8_t) *s) || *s == '_')) {
		return 0;
	}
	for ( ; *s; s++) {
		if (!(isalnum((uint8_t) *s) || *s == '_')) {
			break;
		}
	}
	return *s == '=' && s > os && *(s+1) != '=';
}

/* strtod is supposed to be a proper test of what's a valid number */
/* appears to be broken in gcc on linux: thinks 0x123 is a valid FP number */
/* wrong: violates 4.10.1.4 of ansi C standard */

#include <math.h>
int eawk_is_number(const char *s)
{
	awknum_t r;
	char *ep;

	errno = 0;
	r = strtod(s, &ep);
	if (ep == s || r == HUGE_VAL || errno == ERANGE)
		return 0;
	while (*ep == ' ' || *ep == '\t' || *ep == '\n')
		ep++;
	if (*ep == '\0')
		return 1;
	else
		return 0;
}
