#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<fcntl.h>
#include	<string.h>
#include	<errno.h>
#include	<pwd.h>
#include	<grp.h>
#include	<sys/stat.h>
#include	<sys/time.h>
#include	<sys/param.h>
#include	"libpq-fe.h"

/*
**	Check MS66 database consistency.  Look for any errors
**	and optionally generate a directory tree with empty files.
*/
#define	DBNAME		"ms66"

PGconn		*conn = NULL;
char		*newroot = NULL;
char		checkpath[PATH_MAX+1];
int		verbose = 0;
int		status;

/*
**	Functions and structures to keep track of usernames and
**	groupnames.
*/
struct hashent	{
	char	*name;
	long	value;
	struct	hashent	*link;
};

#define	HASH	0x1f

struct	hashent	*username[HASH+1];
struct	hashent	*groupname[HASH+1];

int
hashfunc(char *name)
{
	char	*cp;
	int	x = 0;

	for (cp=name; *cp; cp++)
		x ^= *cp;

	x &= HASH;
	return x;
}

struct hashent *
hashfind(struct hashent **array, char *name)
{
	struct	hashent	*hp;
	int		x = hashfunc(name);

	for (hp=array[x]; hp; hp=hp->link) {
		if (strcmp(hp->name, name) == 0)
			break;
	}
	return hp;
}

void
hashadd(struct hashent **spot, char *name, long value)
{
	struct hashent	*hp;

	hp = (struct hashent *)malloc(sizeof(struct hashent));
	hp->name = strdup(name);
	hp->value = value;
	hp->link = *spot;
	*spot = hp;
}

/*
**	Output an error message.
*/
void
prterr(char *call, char *name)
{
	char	buf[256];

	sprintf(buf, "%s: %s", call, name);
	perror(buf);
}

/*
**	Send an sql statement with retrys.
*/
PGresult *
send_sql(char *statement)
{
	int		i;
	PGresult	*res = NULL;

	for (i=0; i<2; i++) {
		res = PQexec(conn, statement);
		status = PQresultStatus(res);

		switch (status) {

		case PGRES_BAD_RESPONSE:
		case PGRES_NONFATAL_ERROR:
		case PGRES_FATAL_ERROR:
			fprintf(stderr, "'%s'\n\t(status %d) %s",
				statement, status,
				PQresultErrorMessage(res));
			if (PQstatus(conn) != CONNECTION_BAD)
				return res;

		default:
			return res;
		}

		PQclear(res);
		PQreset(conn);
	}
	return res;
}

/*
**	Get the segments for a particular uid.  A select is used rather
**	then a cursor because the number of records should be small
**	enough to hold in memory.
**	Return 0 if all is well, otherwise return 1.
*/
int
check_fileseg(char *uid, char *name, off_t flength)
{
	PGresult	*res;
	int		i, tuples;
	char		*segname;
	int		segment, nextseg;
	int		copynum, nextcopy;
	int		stripe, nextstripe;
	int		currwidth, width;
	off_t		foffset;
	off_t		newoff;
	off_t		length;
	char		buf[1024];

	sprintf(buf,
		"SELECT name, f.copynum, f.segment, stripe, width, "
			"foffset, length, cart_id "
		"FROM fileseg f, stripe s "
		"WHERE f.uid = '%s' AND f.uid = s.uid AND "
			"f.segment = s.segment AND f.copynum = s.copynum "
		"ORDER BY f.copynum, f.segment, stripe", uid);

	res = send_sql(buf);
	if (status != PGRES_TUPLES_OK) {
		PQclear(res);
		return 1;
	}

	if ((tuples = PQntuples(res)) == 0) {
		PQclear(res);
		return 1;
	}

	nextseg = 0;
	nextcopy = 0;
	nextstripe = 0;
	for (i=0; i<tuples; i++) {
		segname = PQgetvalue(res, i, 0);

		if (verbose && strcmp(name, segname) != 0) {
			fprintf(stderr, "%s: name mismatch\n\t%s\n\t%s\n",
				uid, name, segname);
		}

		copynum = atoi(PQgetvalue(res, i, 1));
		segment = atoi(PQgetvalue(res, i, 2));
		stripe = atoi(PQgetvalue(res, i, 3));
		width = atoi(PQgetvalue(res, i, 4));

		if (nextstripe != stripe)
			break;
		if (nextseg != segment)
			break;
		if (nextcopy != copynum)
			break;

		if (segment == 0) {
			currwidth = width;
			newoff = 0;
		}
		else if (currwidth != width)
			break;

		foffset = strtoq(PQgetvalue(res, i, 5), NULL, 10);
		length = strtoq(PQgetvalue(res, i, 6), NULL, 10);
		if (length % width) {
			fprintf(stderr,
			"%s: seg %d len %lld not divisable by width %d\n",
				uid, segment, length, width);
			break;
		}

		if (foffset != newoff) {
			fprintf(stderr,
				"%s: seg %d file hole %lld -> %lld\n",
				uid, segment, newoff, foffset);
			break;
		}

		if (++nextstripe == width) {
			nextstripe = 0;
			newoff = foffset + length;

			if (newoff == flength) {
				nextseg = 0;
				nextcopy++;
			}
			else
				nextseg++;
		}
	}
	PQclear(res);

	return (i == tuples) ? 0 : 1;
}

/*
**	Look through the cartpart table to make sure there
**	is the right number of sides and partitions.
**	Return 0 if all is well, otherwise return 1.
*/
int
check_cartpart(char *cartid, int sides, int parts)
{
	PGresult	*res;
	int		i, tuples;
	int		partition, nextpart;
	int		side, nextside;
	char		*partname;
	char		buf[1024];

	sprintf(buf,
		"SELECT side, partition, part_name "
			"FROM cartpart "
			"WHERE cart_id = '%s' "
		"ORDER BY side, partition", cartid);

	res = send_sql(buf);
	if (status != PGRES_TUPLES_OK) {
		PQclear(res);
		return 1;
	}

	if ((tuples = PQntuples(res)) == 0) {
		PQclear(res);
		return 1;
	}

	nextside = 0;
	nextpart = 0;
	for (i=0; i<tuples; i++) {

		side = atoi(PQgetvalue(res, i, 0));
		partition = atoi(PQgetvalue(res, i, 1));
		partname = PQgetvalue(res, i, 2);

		if (nextpart != partition)
			break;
		if (nextside != side)
			break;

		if (verbose) {
			fprintf(stderr,
				"%s: side %d partition %d %s\n",
				cartid, side, partition, partname);
		}

		if (++nextpart == parts) {
			nextpart = 0;
			nextside++;
		}
	}
	PQclear(res);

	if (nextside == sides && nextpart == 0)		/* all there */
		return 0;
	else
		return 1;
}

/*
**	Create a file path if the user specified a root dir.
**	Return 0 if all is well, otherwise return 1.
*/
int
create_path(char *name)
{
	char		*ch;
	int		i, num, len, end;
	struct stat	sbuf;

	if (newroot == NULL)
		return 1;

	ch = strrchr(name, '/');	/* last slash */
	end = ch - name;

	/*
	**	Compare the passed in name with the last seen
	**	name up to and including the last slash.
	**	If they are different, we have to create some
	**	intermidiate directories.
	*/
	if (memcmp(newroot, name, end+1) != 0) {
		memcpy(newroot, name, end);
		newroot[end] = '\0';

		for (num=0;; num++) {
			if (stat(checkpath, &sbuf) == 0)
				break;

			if (errno != ENOENT) {
				prterr("stat", checkpath);
				return 1;
			}

			ch = strrchr(newroot, '/');
			*ch = '\0';
		}

		for (i=0; i<num; i++) {
			len = strlen(newroot);
			newroot[len] = '/';
			if (mkdir(checkpath, 0700) == -1) {
				prterr("mkdir", checkpath);
				return 1;
			}
		}
	}
	strcpy(&newroot[end], &name[end]);
	return 0;
}

int
main(int argc, char *argv[])
{
	PGresult	*res = NULL;
	struct	stat	sbuf;
	struct	passwd	*pwent;
	struct	group	*grent;
	int		c, i, j;
	int		fd;
	int		numroot, tuples;
	char		**root;
	int		*rootlen;
	char		*pghost = NULL;
	char		*pgport = NULL;
	char		*pgopts = NULL;
	char		*pgtty  = NULL;
	char		*pgname = DBNAME;
	char		*login  = NULL;
	char		*pwd    = NULL;
	char		*prevname;
	char		*prevuid = NULL;
	int		totfiles = 0;
	int		totcarts = 0;
	int		goodfiles = 0;
	int		badfile;
	int		goodcarts = 0;
	int		errflg = 0;
	extern char	*optarg;
	extern int	optind, opterr;


	while ((c = getopt(argc, argv, "vh:p:o:t:n:l:r:")) != EOF) {
		switch (c) {

		case 'v':
			verbose = 1;
			break;

		case 'h':
			pghost = optarg;
			break;

		case 'p':
			pgport = optarg;
			break;

		case 'o':
			pgopts = optarg;
			break;

		case 't':
			pgtty = optarg;
			break;

		case 'n':
			pgname = optarg;
			break;

		case 'l':
			login = optarg;
			break;

		case 'r':
			if ((stat(optarg, &sbuf)) == -1) {
				prterr("stat", optarg);
				errflg = 1;
			}
			else if (!S_ISDIR(sbuf.st_mode)) {
				fprintf(stderr, "%s: not a directory\n",
					optarg);
				errflg = 1;
			}
			else {
				int	 len = strlen(optarg);

				memcpy(checkpath, optarg, len);
				newroot = &checkpath[len];
			}
			break;

		case '?':
		default:
			errflg = 1;
			break;
		}
	}

	if (errflg) {
		fprintf(stderr, "usage: %s [-v][-h host][-p port]\n", argv[0]);
		fprintf(stderr, "\t[-o opts][-t tty][-n dbname]\n");
		fprintf(stderr, "\t[-l login][-r root]\n");
		return 2;
	}

	while ((pwent = getpwent()) != NULL) {
		int	x = hashfunc(pwent->pw_name);

		hashadd(&username[x], pwent->pw_name, (long)pwent->pw_uid);
	}
	endpwent();

	while ((grent = getgrent()) != NULL) {
		int	x = hashfunc(grent->gr_name);

		hashadd(&groupname[x], grent->gr_name, (long)grent->gr_gid);
	}
	endgrent();

	conn = PQsetdbLogin(pghost, pgport, pgopts, pgtty, pgname, login, pwd);
	if (PQstatus(conn) == CONNECTION_BAD) {
		fprintf(stderr, "could not connect to database: %s\n",
			PQerrorMessage(conn));
		goto done;
	}

	/*
	**	Read table fsms and save each export_root.
	**	This should be a small number of records so we
	**	just do a simple select and allocate memory to hold
	**	the result.
	*/
	res = send_sql("SELECT export_root FROM fsms");
	if (status != PGRES_TUPLES_OK)
		goto done;

	numroot = PQntuples(res);
	root = (char **)calloc(numroot, sizeof(char *));
	rootlen = (int *)calloc(numroot, sizeof(int));
	for (i=0; i<numroot; i++) {
		root[i] = strdup(PQgetvalue(res, i, 0));
		rootlen[i] = strlen(root[i]);
	}
	PQclear(res);

	/*
	**	Now, use a cursor to read each record out of the
	**	file table.
	*/
	res = send_sql("BEGIN WORK");
	if (status != PGRES_COMMAND_OK)
		goto done;
	PQclear(res);

	res = send_sql(
		"DECLARE getfile CURSOR FOR "
		"SELECT uid, name, type, length, "
			"permissions, ownerid, groupid, "
			"date_part('epoch', accessed), "
			"date_part('epoch', modified), symlink "
		"FROM file "
		"ORDER BY uid");
	if (status != PGRES_COMMAND_OK)
		goto done;
	PQclear(res);

	if (verbose)
		fprintf(stderr, "checking file table\n");
	for (;;) {
		res = send_sql("FETCH 50 FROM getfile");
		if (status != PGRES_TUPLES_OK)
			break;

		if ((tuples = PQntuples(res)) == 0)
			break;

		totfiles += tuples;

		for (i=0; i<tuples; i++) {
			char		*uid, *name, *type;
			off_t		length;
			mode_t		perms;
			char		*owner;
			char		*group;
			char		*access;
			char		*modify;
			char		*linkname;
			uid_t		userid;
			gid_t		groupid;
			struct	timeval	times[2];
			struct	hashent	*hp;

			if (PQgetisnull(res, i, 0)) {	/* no uid */
				fprintf(stderr, "record with null uid\n");
				continue;
			}

			uid = PQgetvalue(res, i, 0);
			name = PQgetvalue(res, i, 1);
			type = PQgetvalue(res, i, 2);
			length = strtoq(PQgetvalue(res, i, 3), NULL, 10);
			perms = (mode_t)strtol(PQgetvalue(res, i, 4), NULL, 8);
			owner = PQgetvalue(res, i, 5);
			group = PQgetvalue(res, i, 6);
			access = PQgetvalue(res, i, 7);
			modify = PQgetvalue(res, i, 8);
			linkname = PQgetvalue(res, i, 9);

			if ((hp=hashfind(username, owner)) == NULL) {
				int	x = hashfunc(owner);

				fprintf(stderr,
					"userid\t%s\tnot found\n", owner);
				hashadd(&username[x], owner, -1);
				userid = (uid_t)-1;
			}
			else
				userid = (uid_t)hp->value;

			if ((hp=hashfind(groupname, group)) == NULL) {
				int	x = hashfunc(group);

				fprintf(stderr,
					"groupid\t%s\tnot found\n", group);
				hashadd(&groupname[x], group, -1);
				groupid = (gid_t)-1;
			}
			else
				groupid = (gid_t)hp->value;

			times[0].tv_sec = atol(access);
			times[0].tv_usec = 0;
			times[1].tv_sec = atol(modify);
			times[1].tv_usec = 0;

			for (j=0; j<numroot; j++) {
				if (strncmp(name, root[j], rootlen[j]) == 0)
					break;
			}

			if (j == numroot) {
				fprintf(stderr, "%s\t%s\n", uid, name);
				fprintf(stderr, "\tdoes not start with root\n");
			}

			/*
			**	Check for hard link (same uid).
			*/
			if (prevuid) {
				if (strcmp(uid, prevuid) == 0) {
					if (create_path(name))
						continue;
					if (link(prevname, checkpath) == -1)
						prterr("link", checkpath);
					continue;
				}
				free(prevname);
				free(prevuid);
				prevuid = NULL;
			}

			badfile = 0;
			switch (*type) {

			case 'F':
				if (check_fileseg(uid, name, length)) {
					badfile++;
					fprintf(stderr,
					"%s: %s fileseg sequence incomplete\n",
						uid, name);
					break;
				}

				if (create_path(name))
					break;
				fd = open(checkpath,
					O_WRONLY|O_CREAT|O_EXCL, perms);
				if (fd == -1) {
					prterr("open", checkpath);
					break;
				}
				if (ftruncate(fd, length) == -1)
					prterr("ftruncate", checkpath);
				if (futimes(fd, times) == -1)
					prterr("futimes", checkpath);
				if (fchown(fd, userid, groupid) == -1)
					prterr("fchown", checkpath);

				(void)close(fd);
				printf("%s %s\n", uid, checkpath);
				break;

			case 'C':
				(void)create_path(name);
				break;

			case 'D':
				if (create_path(name))
					break;
				if (mkdir(checkpath, perms) == -1) {
					if (errno != EEXIST) {
						prterr("mkdir", name);
						break;
					}
					if (chmod(checkpath, perms) == -1)
						prterr("chmod", checkpath);
				}
				if (utimes(checkpath, times) == -1)
					prterr("futimes", checkpath);
				if (chown(checkpath, userid, groupid) == -1)
					prterr("fchown", checkpath);
				break;

			case 'S':
				if (create_path(name))
					break;
				if (symlink(linkname, checkpath) == -1) {
					prterr("symlink", checkpath);
					break;
				}
				if (lchmod(checkpath, perms) == -1)
					prterr("lchmod", checkpath);
				if (lutimes(checkpath, times) == -1)
					prterr("lutimes", checkpath);
				if (lchown(checkpath, userid, groupid) == -1)
					prterr("lchown", checkpath);
				break;

			default:
				badfile++;
				fprintf(stderr,
					"%s: %s unknown type %s\n",
					uid, name, type);
				break;
			}

			if (badfile)
				continue;

			goodfiles++;
			if (verbose)
				fprintf(stderr, "%s %s\n", uid, name);
			prevname = strdup(checkpath);
			prevuid = strdup(uid);
		}

		PQclear(res);
	}
	PQclear(res);
	(void)send_sql("CLOSE getfile");

	/*
	**	Now check carts.
	*/
	res = send_sql("DECLARE getcart CURSOR for "
		"SELECT cart_id, partitions, sides, media, blocklen, blocksize "
		"FROM cart");

	if (status != PGRES_COMMAND_OK)
		goto done;
	PQclear(res);

	if (verbose)
		fprintf(stderr, "checking cart table\n");
	for (;;) {
		res = send_sql("FETCH 50 FROM getcart");
		if (status != PGRES_TUPLES_OK)
			break;

		if ((tuples = PQntuples(res)) == 0)
			break;

		totcarts += tuples;

		for (i=0; i<tuples; i++) {
			char	*cartid, *media;
			int	parts, sides;
			int	blocklen, blocksize;
			char	*szcheck = "";

			if (PQgetisnull(res, i, 0)) {	/* no cart_id */
				fprintf(stderr, "record with null cart_id\n");
				continue;
			}

			cartid = PQgetvalue(res, i, 0);
			parts = atoi(PQgetvalue(res, i, 1));
			sides = atoi(PQgetvalue(res, i, 2));
			media = PQgetvalue(res, i, 3);
			blocklen = atoi(PQgetvalue(res, i, 4));
			blocksize = atoi(PQgetvalue(res, i, 5));

			if (blocklen && blocksize && (blocklen != blocksize))
				szcheck = " BAD";

			if (check_cartpart(cartid, parts, sides)) {
				fprintf(stderr,
					"%s: partition sequence incomplete\n",
					cartid);
				continue;
			}

			goodcarts++;
			if (verbose) {
				fprintf(stderr,
					"cartid\t%s\t%s blksize %d/%d%s\n",
					cartid, media, blocklen, blocksize,
					szcheck);
			}
		}
		PQclear(res);
	}
	PQclear(res);

	(void)send_sql("CLOSE getcart");
	(void)send_sql("COMMIT WORK");

 done:
	if (res)
		PQclear(res);
	PQfinish(conn);

	if (verbose) {
		fprintf(stderr, "files: %d total %d good\n",
			totfiles, goodfiles);
		fprintf(stderr, "carts: %d total %d good\n",
			totcarts, goodcarts);
	}
	return 0;
}
