#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<fcntl.h>
#include	<string.h>
#include	<errno.h>
#include	<pwd.h>
#include	<grp.h>
#include	<sys/stat.h>
#include	<sys/time.h>
#include	<sys/param.h>
#include	<sys/ioctl.h>
#include	<sys/mtio.h>

#include	"libpq-fe.h"

/*
**	Use the information in an MS66 postgres database to
**	mount and position tapes and read the file data.
*/
#define	DBNAME		"ms66"

PGconn		*conn = NULL;
int		verbose = 0;
int		status;

char		filename[MAXPATHLEN+1];
int		filefd = -1;

int		tapefd = -1;
char		tapebuf[64*1024];
int		tmblkskip;
int		tmbyteskip;

char		*mountedcart = NULL;
int		mountedpart, mountedside;

/*
**	Output an error message.
*/
void
prterr(char *call, char *name)
{
	char	buf[256];

	sprintf(buf, "%s: %s", call, name);
	perror(buf);
}

/*
**	Send an sql statement with retrys.
*/
PGresult *
send_sql(char *statement)
{
	int		i;
	PGresult	*res = NULL;

	for (i=0; i<2; i++) {
		res = PQexec(conn, statement);
		status = PQresultStatus(res);

		switch (status) {

		case PGRES_BAD_RESPONSE:
		case PGRES_NONFATAL_ERROR:
		case PGRES_FATAL_ERROR:
			fprintf(stderr, "'%s'\n\t(status %d) %s",
				statement, status,
				PQresultErrorMessage(res));
			if (PQstatus(conn) != CONNECTION_BAD)
				return res;

		default:
			return res;
		}

		PQclear(res);
		PQreset(conn);
	}
	return res;
}

/*
**	Read a block from tape.  There may be a tape mark which
**	needs to be skipped with control blocks that follow.
*/
size_t
readblock(void)
{
	static	char	id[] = "readblock";
	int	i;
	size_t	len;

	if ((len = read(tapefd, tapebuf, sizeof(tapebuf))) == -1)
		prterr("read", mountedcart);
	else if (len == 0) {
		int	control = 0;

		for (i=0; i<tmblkskip; i++) {
			len = read(tapefd, tapebuf, sizeof(tapebuf));
			if (len <= 0) {
				prterr("read", mountedcart);
				fprintf(stderr,
					"%s: read %d in control skip\n",
					id, len);
				return -1;
			}
			control += len;
		}
		if (control != tmbyteskip) {
			fprintf(stderr,
				"%s: control skip %d bytes, should be %d\n",
				id, control, tmbyteskip);
			return -1;
		}
		if ((len = read(tapefd, tapebuf, sizeof(tapebuf))) <= 0) {
			prterr("read", mountedcart);
			fprintf(stderr,
				"%s: read %d after control skip\n", id, len);
			return -1;
		}
	}
	return len;
}

/*
**	Clean up tape state.
*/
void
failtape(char *op)
{
	if (op)
		prterr(op, mountedcart);

	if (tapefd != -1) {
		(void)close(tapefd);
		tapefd = -1;
	}
	if (mountedcart) {
		free(mountedcart);
		mountedcart = NULL;
	}
	return;
}
	
/*
**	Get the segments for a particular uid.  A select is used rather
**	then a cursor because the number of records should be small
**	enough to hold in memory.
*/
void
get_fileseg(char *uid, off_t flength)
{
	PGresult	*res;
	int		i, tuples;
	int		segment, wantseg;
	int		copynum, wantcopy;
	int		stripe, wantstripe;
	int		currwidth, width;
	off_t		foffset, fs_done, file_done;
	off_t		newoff, fs_end;
	off_t		length;
	int		fsheader, fstrailer;
	char		*cartid;
	int		partition, side;
	int		tapemark, block, record, byte;
	int		startpoint, endpoint, amount;
	int		logical_block_size;
	char		buf[1024];
	struct	mtop	mtop;

	sprintf(buf,
		"SELECT f.copynum, f.segment, stripe, width, "
			"foffset, length, header, trailer, "
			"cart_id, partition, side, tapemark, block, "
			"record, byte "
		"FROM fileseg f, stripe s "
		"WHERE f.uid = '%s' AND f.uid = s.uid AND "
			"f.segment = s.segment AND f.copynum = s.copynum "
		"ORDER BY f.copynum, f.segment, stripe", uid);

	res = send_sql(buf);
	if (status != PGRES_TUPLES_OK) {
		PQclear(res);
		return;
	}

	if ((tuples = PQntuples(res)) == 0) {
		fprintf(stderr, "%s: no file segments found\n", uid);
		PQclear(res);
		return;
	}
	if (verbose)
		printf("retrieved %d records for %s\n", tuples, uid);

	wantcopy = 0;
	wantseg = 0;
	wantstripe = 0;
	file_done = 0;

	/*
	**	Loop over all records from above SELECT statement.
	*/
	for (i=0; i<tuples; i++) {
		PGresult	*cartres;
		int		physicalblk;
		int		logicalblk;
		int		logicalhdr;
		int		logicaltrl;

		copynum = atoi(PQgetvalue(res, i, 0));
		segment = atoi(PQgetvalue(res, i, 1));

		if (wantcopy != copynum) {
			if (verbose) {
				printf("%s: skipping copy %d seg %d\n",
					uid, copynum, segment);
			}
			continue;
		}

		if (wantseg != segment) {
			wantcopy++;
			wantseg = wantstripe = 0;
			fprintf(stderr,
				"%s: missing segment %d, goto copy %d\n",
				uid, segment, wantcopy);
			continue;
		}

		stripe = atoi(PQgetvalue(res, i, 2));
		if (wantstripe != stripe) {
			wantcopy++;
			wantseg = wantstripe = 0;
			fprintf(stderr,
				"%s: missing stripe %d, goto copy %d\n",
				uid, stripe, wantcopy);
			continue;
		}

		width = atoi(PQgetvalue(res, i, 3));
		if (segment == 0) {		/* first segment */
			currwidth = width;
			newoff = 0;
		}
		else if (currwidth != width) {
			wantcopy++;
			wantseg = wantstripe = 0;
			fprintf(stderr,
				"%s: seg %d:%d width %d should be %d, "
				"goto copy %d\n",
				uid, segment, stripe, width,
				currwidth, wantcopy);
			continue;
		}

		foffset = strtoq(PQgetvalue(res, i, 4), NULL, 10);
		length = strtoq(PQgetvalue(res, i, 5), NULL, 10) / width;
		foffset += stripe * length;
		fs_end = foffset + length;

		if (fs_end <= file_done) {
			if (verbose) {
				printf("skip segment %d copy %d stripe %d "
					"offset %lld - %lld file@ %lld\n",
					segment, copynum, stripe,
					foffset, fs_end, file_done);
			}
			goto skip;
		}

		/*
		**	Check for file hole.
		*/
		if (foffset != newoff) {
			wantcopy++;
			wantseg = wantstripe = 0;
			fprintf(stderr,
				"%s: seg %d:%d offset %lld should be %lld, "
				"goto copy %d\n",
				uid, segment, stripe,
				foffset, newoff, wantcopy);
			continue;
		}

		fsheader = strtoq(PQgetvalue(res, i, 6), NULL, 10);
		fstrailer = strtoq(PQgetvalue(res, i, 7), NULL, 10);
		cartid = PQgetvalue(res, i, 8);
		partition = atoi(PQgetvalue(res, i, 9));
		side = atoi(PQgetvalue(res, i, 10));
		tapemark = strtol(PQgetvalue(res, i, 11), NULL, 10);
		block = strtol(PQgetvalue(res, i, 12), NULL, 10);
		record = strtol(PQgetvalue(res, i, 13), NULL, 10);
		byte = strtol(PQgetvalue(res, i, 14), NULL, 10);

		sprintf(buf,
			"SELECT blocklen, data_len, hdr_len, trl_len, "
				"blocksaftertm, bytesaftertm "
			"FROM cart WHERE cart_id = '%s'", cartid);
		cartres = send_sql(buf);
		if (status != PGRES_TUPLES_OK) {
			wantcopy++;
			wantseg = wantstripe = 0;
			fprintf(stderr,
				"Could not get record for %s, "
				"goto copy %d\n",
				cartid, wantcopy);
			PQclear(cartres);
			continue;
		}

		if (PQntuples(cartres) != 1) {
			wantcopy++;
			wantseg = wantstripe = 0;
			fprintf(stderr,
				"%s has %d records, should be 1, "
				"goto copy %d\n",
				cartid, PQntuples(cartres), wantcopy);
			PQclear(cartres);
			continue;
		}
		physicalblk = atoi(PQgetvalue(cartres, 0, 0));
		logicalblk = atoi(PQgetvalue(cartres, 0, 1));
		logicalhdr = atoi(PQgetvalue(cartres, 0, 2));
		logicaltrl = atoi(PQgetvalue(cartres, 0, 3));
		tmblkskip = atoi(PQgetvalue(cartres, 0, 4));
		tmbyteskip = atoi(PQgetvalue(cartres, 0, 5));
		PQclear(cartres);

		/*
		**	Calculate where the file segment starts
		**	in the first block of the tape.
		*/
		logical_block_size = logicalhdr + logicalblk + logicaltrl;
		if (physicalblk == 0)
			physicalblk = sizeof(tapebuf);

		startpoint = record*logical_block_size + byte;
		if (startpoint >= physicalblk) {
			wantcopy++;
			wantseg = wantstripe = 0;
			fprintf(stderr,
				"%s: fs%d blkoff %d >= size %d, "
				"goto copy %d\n",
				uid, segment, startpoint,
				physicalblk, wantcopy);
			continue;
		}

		if (verbose) {
			printf("segment %d copy %d stripe %d width %d\n",
				segment, copynum, stripe, width);
		}
		if (mountedcart == NULL ||
				(strcmp(cartid, mountedcart) != 0 ||
				mountedpart != partition ||
				mountedside != side)) {

			char	tapedev[128];

			for (;;) {
				int	len;

				printf("Mount %s side %d, partition %d "
					"and input the device\n"
					"(or return if not available).\n",
					cartid, side, partition);
				(void)fgets(tapedev, sizeof(tapedev), stdin);
				if (tapedev[0] == '\n')
					break;

				len = strlen(tapedev);
				tapedev[len-1] = '\0';

				failtape(NULL);
				if ((tapefd = open(tapedev, O_RDONLY)) != -1)
					break;

				prterr("open", tapedev);
			}
			if (tapedev[0] == '\n') {		/* skip cart */
				wantcopy++;
				wantseg = wantstripe = 0;
				continue;
			}

			mountedcart = strdup(cartid);
			mountedpart = partition;
			mountedside = side;
		}

		if (verbose)
			printf("%s: rewind tape\n", mountedcart);
		mtop.mt_op = MTREW;
		mtop.mt_count = 1;
		if (ioctl(tapefd, MTIOCTOP, &mtop) == -1) {
			failtape("ioctl(MTIOCTOP, MTREW)");
			wantcopy++;
			wantseg = wantstripe = 0;
			continue;
		}

		if (verbose) {
			printf("%s: forward space file %d\n",
				mountedcart, tapemark);
		}
		mtop.mt_op = MTFSF;
		mtop.mt_count = tapemark;
		if (ioctl(tapefd, MTIOCTOP, &mtop) == -1) {
			failtape("ioctl(MTIOCTOP, MTFSF)");
			wantcopy++;
			wantseg = wantstripe = 0;
			continue;
		}

		if (verbose) {
			printf("%s: forward space block %d\n",
				mountedcart, block);
		}
		mtop.mt_op = MTFSR;
		mtop.mt_count = block;
		if (ioctl(tapefd, MTIOCTOP, &mtop) == -1) {
			failtape("ioctl(MTIOCTOP, MTFSR)");
			wantcopy++;
			wantseg = wantstripe = 0;
			continue;
		}
		/*
		**	Now read the tape.  If this is the first element
		**	of a stripe set, read past any file segment header.
		*/
		if (stripe == 0) {
			for (;;) {
				if ((endpoint = readblock()) == -1)
					break;
				amount = endpoint - startpoint;
				if (amount > fsheader) {
					startpoint += fsheader;
					break;
				}
				startpoint = 0;
				fsheader -= amount;
			}
		}
		if (lseek(filefd, foffset, SEEK_SET) != foffset) {
			prterr("lseek", filename);
			break;
		}

		if (verbose)
			printf("%s: reading data\n", mountedcart);
		/*
		**	Now we can read file data.
		*/
		fs_done = 0;
		if (logical_block_size == 0)	/* no logical block */
			logicalblk = endpoint - startpoint;

		for (;;) {
			for (amount = endpoint - startpoint;
					amount > 0;
					amount = endpoint - startpoint) {
				int	outlen, len;

				/* check for end of file segment */
				outlen = length-fs_done;
				outlen = MIN(logicalblk, outlen);
				if (outlen == 0)
					break;

				startpoint += logicalhdr;

				/* check for last (possibly short) record */
				len = endpoint-logicaltrl-startpoint;
				outlen = MIN(outlen, len);

				len = write(filefd, &tapebuf[startpoint],
						outlen);
				if (len != outlen) {
					prterr("write", filename);
					goto done;
				}

				fs_done += outlen;
				if (foffset + fs_done > file_done)
					file_done = foffset + fs_done;
				startpoint += outlen;
				startpoint += logicaltrl;
			}
			if (fs_done == length)
				break;

			endpoint = readblock();
			if (endpoint == -1)
				break;
			if (logical_block_size == 0)	/* no logical block */
				logicalblk = endpoint;
			startpoint = 0;
		}

 skip:
		newoff = fs_end;
		if (++wantstripe == width) {
			if (newoff == flength)
				break;

			wantstripe = 0;
			wantseg++;
		}

	}
 done:
	PQclear(res);

	if (file_done != flength) {
		fprintf(stderr, "%s: file size %lld should be %lld\n",
			uid, file_done, flength);
	}
	else if (verbose)
		printf("%s: good file length %lld\n", uid, flength);
	return;
}

int
main(int argc, char *argv[])
{
	PGresult	*res = NULL;
	char		buf[1024];
	int		c, i;
	int		tuples;
	char		*pghost = NULL;
	char		*pgport = NULL;
	char		*pgopts = NULL;
	char		*pgtty  = NULL;
	char		*pgname = DBNAME;
	char		*login  = NULL;
	char		*pwd    = NULL;
	struct	passwd	*pent;
	struct	group	*gent;
	int		errflg = 0;
	extern char	*optarg;
	extern int	optind, opterr;


	while ((c = getopt(argc, argv, "vh:p:o:t:n:l:")) != EOF) {
		switch (c) {

		case 'v':
			verbose = 1;
			break;

		case 'h':
			pghost = optarg;
			break;

		case 'p':
			pgport = optarg;
			break;

		case 'o':
			pgopts = optarg;
			break;

		case 't':
			pgtty = optarg;
			break;

		case 'n':
			pgname = optarg;
			break;

		case 'l':
			login = optarg;
			break;

		case '?':
		default:
			errflg = 1;
			break;
		}
	}

	if (errflg) {
		fprintf(stderr, "usage: %s [-v][-h host][-p port]\n", argv[0]);
		fprintf(stderr, "\t[-o opts][-t tty][-n dbname]\n");
		fprintf(stderr, "\t[-l login] uid ...\n");
		return 2;
	}

	conn = PQsetdbLogin(pghost, pgport, pgopts, pgtty, pgname, login, pwd);
	if (PQstatus(conn) == CONNECTION_BAD) {
		fprintf(stderr, "could not connect to database: %s\n",
			PQerrorMessage(conn));
		return 1;
	}

	for (i=optind; i<argc; i++, PQclear(res)) {
		char		*name, *type;
		off_t		length;
		mode_t		perms;
		char		*owner;
		char		*group;
		char		*access;
		char		*modify;
		char		*linkname;
		uid_t		userid = -1;
		gid_t		groupid = -1;
		struct	timeval	times[2];

		sprintf(buf,
			"SELECT name, type, length, permissions, "
				"ownerid, groupid, "
				"date_part('epoch', accessed), "
				"date_part('epoch', modified), symlink "
			"FROM file "
			"WHERE uid='%s'", argv[i]);

		res = send_sql(buf);
		if (status != PGRES_TUPLES_OK)
			break;

		if ((tuples = PQntuples(res)) == 0) {
			fprintf(stderr, "%s: uid not found\n", argv[i]);
			continue;
		}
		name = PQgetvalue(res, 0, 0);
		type = PQgetvalue(res, 0, 1);
		length = strtoq(PQgetvalue(res, 0, 2), NULL, 10);
		perms = (mode_t)strtol(PQgetvalue(res, 0, 3), NULL, 8);
		owner = PQgetvalue(res, 0, 4);
		group = PQgetvalue(res, 0, 5);
		access = PQgetvalue(res, 0, 6);
		modify = PQgetvalue(res, 0, 7);
		linkname = PQgetvalue(res, 0, 8);

		if (type == NULL)
			continue;

		if (*type != 'F') {
			fprintf(stderr, "%s: not a plain file\n", argv[i]);
			continue;
		}

		if ((pent = getpwnam(owner)) == NULL) {
			fprintf(stderr, "%s: user '%s' not found\n",
				argv[i], owner);
		}
		else
			userid = pent->pw_uid;
		if ((gent = getgrnam(group)) == NULL) {
			fprintf(stderr, "%s: group '%s' not found\n",
				argv[i], group);
		}
		else
			groupid = gent->gr_gid;

		times[0].tv_sec = atol(access);
		times[0].tv_usec = 0;
		times[1].tv_sec = atol(modify);
		times[1].tv_usec = 0;

		if ((name = strrchr(name, '/')) != NULL)
			name++;
		printf("filename [%s]? ", name ? name : "");
		(void)fgets(filename, sizeof(filename), stdin);
		if (filename[0] == '\n')
			strcpy(filename, name);
		else {
			int	len = strlen(filename);

			filename[len-1] = '\0';
		}

		filefd = open(filename, O_WRONLY|O_CREAT|O_TRUNC, 0644);
		if (filefd == -1) {
			prterr("open", filename);
			continue;
		}

		get_fileseg(argv[i], length);

		close(filefd);
	}
	PQfinish(conn);
	close(tapefd);

	return 0;
}
