
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/proc.h>
#include <sys/uio.h>
#include <sys/vnode.h>
#include <sys/select.h>
#include <sys/file.h>
#include <sys/kernel.h>
#include <miscfs/dmfs/dmfs.h>
#include <miscfs/genfs/layer_extern.h>
#include <miscfs/dmfs/dmfs_extern.h>
#include <miscfs/dmfs/kdcom.h>
#include <miscfs/dmfs/rash.h>

struct timespec dmfs_nulltime = { VNOVAL, VNOVAL};

int
dm_read(v)
	void *v;
{
	struct vop_read_args /* {
		struct vnode *a_vp;
		struct uio *a_uio;
		int a_ioflag;
		struct ucred *a_cred;
	} */ *ap = v;
	struct vnode *vp = ap->a_vp;
	struct dm_node *dp = VTODM(vp);
	int error;

	if (!ISSET(dp->dm_flags, DM_HASMETA)) {
		return (dm_bypass(ap));
	}
	if (ISSET(ap->a_ioflag, IO_ALTSEMANTICS)) {
		/* make sure super user? */
		return (dm_bypass(ap));
	}

	if (ISSET(dp->dm_flags, DM_NONR)) {
		if ((error = dmfs_ioblock(vp, ap->a_uio, ap->a_ioflag, 0,
				ap->a_cred->cr_uid)) > 0)
			return (error);
	}

	if ((error = dm_bypass(ap))) {
		return (error);
	}
	SET(dp->dm_flags, DM_UPDATE_ATIME);
	VOP_UPDATE(vp, NULL, NULL, 1);
	return (0);
}

int
dm_write(v)
	void *v;
{
	struct vop_write_args /* {
		struct vnode *a_vp;
		struct uio *a_uio;
		int a_ioflag;
		struct ucred *a_cred;
	} */ *ap = v;
	struct vnode *vp = ap->a_vp;
	struct dm_node *dp = VTODM(vp);
	int error = 0, error1;
	uid_t theuser;

	if (!ISSET(dp->dm_flags, DM_HASMETA))
		return (dm_bypass(ap));
	if (ISSET(ap->a_ioflag, IO_ALTSEMANTICS)) {
		/* make sure super user? */
		return (dm_bypass(ap));
	}

	if (ISSET(dp->dm_flags, DM_NONR)) {
		if (ap->a_cred)
			theuser = ap->a_cred->cr_uid;
		else
			theuser = 0;
		if ((error = dmfs_ioblock(vp, ap->a_uio, ap->a_ioflag, 1,
				theuser)) > 0)
			return (error);
	}

	error = dm_bypass(ap);
	if (error)
		return (error1);
	if (ISSET(dp->dm_flags, DM_ARCH)) {
		if (ap->a_cred)
			theuser = ap->a_cred->cr_uid;
		else
			theuser = 0;
		dmfs_rashdelete(vp, theuser, NULL, 0);
		CLR(dp->dm_flags, DM_ARCH);
	}
	if (ISSET(dp->dm_flags, DM_AIP))
		CLR(dp->dm_flags, DM_AIP);
	SET(dp->dm_flags, DM_MODIFIED | DM_UPDATE_MTIME);
	VOP_UPDATE(vp, NULL, NULL, 1);
	return (0);
}

/*
 * dmfs_ioblock handles the core blocking for both reads & writes through
 * the dmfs layer.
 *
 * Procedure is to check & see if io touches non-restored data. If so,
 * send a message to the userland daemon indicating we want data. We pass
 * a flag indicating if we see a restore in progress, where the block boundry
 * needs to be for this access to proceed, and how much data needs to be
 * present.
 *
 * We then release the lock on the vnode, indicate that we want data, and
 * sleep. When we awake, we look to see if we are still non-resident, the
 * boundry's not beyond what we touch, and the restore hasn't failed. If so,
 * go back to sleep & repeat. If restore failed, return an error. Otherwise
 * return 0.
 *
 * We pass messages out even if DM_RIP so that a userland daemon can deal
 * cleverly with multiple requests if it so chooses. It could cancel the
 * restoration of data we're about to overwrite if it so chooses.
 *
 * We require the block boundry to be beyond the end of a write so that we
 * ensure that there are no pending restores which might overwrite our
 * changes to the file.
 *
 * We special case writes which would leave no unrestored bytes & no other
 * access are waiting. We return an error code which will tell dmfs_write
 * to update the byte boundry if the write succeeds.
 */
int
dmfs_ioblock(vp, uio, flags, iswrite, user)
	struct vnode *vp;
	struct uio *uio;
	int flags, iswrite;
	uid_t user;
{
	struct dm_node *dp = VTODM(vp);
	off_t	start, last;
	int	error, error1, remessage = 0;

	start = uio->uio_offset;
	last = uio->uio_offset + uio->uio_resid -1;
	if (last < start)
		return (0);	/* Let the VOP_ deal with this (0-byte io?) */

	if ((last < dp->dm_bbarrier) && (iswrite == 0)) {
		/*
		 * early reads are fine.
		 */
		return (0);
	}

	/* Now send message */
	if (ISSET(dp->dm_flags, DM_RIP) == 0) {
		if ((error = dmfs_callrash(vp, user)))
			return (error);
	}

	/*
	 * Was the I/O requested with IO_KNONBLOCK? If so, say we can't do it
	 * right now. The nfs server will make all nfs i/o requests
	 * using this flag. Note, we've already asked the daemon to
	 * restore the file.
	 */
	if (ISSET(flags, IO_KNONBLOCK))
		return (EAGAIN);

	/*
	 * Now we have to sleep until data becomes available.
	 */

	for (;;) {
		if (remessage) {
			remessage = 0;
			printf("Timeout in dmfs_ioblock - remessaging!\n");
			if ((error = dmfs_callrash(vp, user)))
				return (error);
		}
		SET(dp->dm_flags, DM_IOWANT);

		VOP_UNLOCK(vp, 0);
		error = tsleep(vp, PRIBIO | PCATCH, "dmfsio", 5 * 60 * hz);
		error1 = vn_lock(vp, LK_EXCLUSIVE | LK_RETRY);
		if (error1)
			return (error1);
		if (error == EINTR) {
			return (EINTR);
		} else if (error == ERESTART) {
			return (ERESTART);
		} else if (error == EWOULDBLOCK) {
			if (dp->dm_restorepid == 0)
				remessage = 1;
		}
		/* Can we proceed? */
		if (vp->v_op != dm_vnodeop_p)
			return (EIO);
		if (!ISSET(dp->dm_flags, DM_HASMETA))
			/* no longer a dmfs file, sure! */
			return (0);
		if (ISSET(dp->dm_flags, DM_RESTOREFAIL))
			return (EIO);
		if ((ISSET(dp->dm_flags, DM_NONR) == 0) || 
				((last < dp->dm_bbarrier) && (iswrite == 0)))
			return (0);
	}
}

/*
 * Wait for an in-progress restore to finish (for DM_RIP to clear).
 *
 * Is passed a locked vnode, which we unlock while sleeping. vnode
 * locked on exit.
 */

int
dmfs_wait(vp)
	struct vnode *vp;
{
	struct dm_node	*dp = VTODM(vp);
	int error, error1, remessage = 0;

	if (!ISSET(dp->dm_flags, DM_HASMETA)) {
		return (0);
	}

	for (;ISSET(dp->dm_flags, DM_RIP);) {
		if (remessage) {
			remessage = 0;
			printf("Timeout in dmfs_wait - remessaging!\n");
			if ((error = dmfs_callrash(vp, 0)))
				return (error);
		}
		SET(dp->dm_flags, DM_IOWANT);

		VOP_UNLOCK(vp, 0);
		error = tsleep(vp, PRIBIO | PCATCH, "dmfswait", 5 * 60 * hz);
		error1 = vn_lock(vp, LK_EXCLUSIVE | LK_RETRY);
		if (error1)
			return (error1);
		if (error == EINTR) {
			return (EINTR);
		} else if (error == ERESTART) {
			return (ERESTART);
		} else if (error == EWOULDBLOCK) {
			if (dp->dm_restorepid == 0)
				remessage = 1;
		}
		/* Can we proceed? */
		if (vp->v_op != dm_vnodeop_p)
			return (EIO);
		if (!ISSET(dp->dm_flags, DM_HASMETA))
			/* no longer a dmfs file, sure! */
			return (0);
		if (ISSET(dp->dm_flags, DM_RESTOREFAIL))
			return (EIO);
	}

	return (0);
}

/*
 * We do two things with VOP_BMAP. First, we make sure the file is restored
 * (as we don't know if we're reading or writing), and also do diagnostics
 * on the returned device node.
 */
int
dm_bmap(v)
	void *v;
{
	struct vop_bmap_args /* {
		struct vnode *a_vp;
		daddr_t  a_bn;
		struct vnode **a_vpp;
		daddr_t *a_bnp;
		int *a_runp;
	} */ *ap = v;
	struct vnode *vp;
	struct dm_node *dp;
	int error;

	vp = ap->a_vp;
	dp = VTODM(vp);

	/*
	 * Make sure a non-resident file is fully resident before
	 * proceeding.
	 */
	if ((ISSET(dp->dm_flags, DM_HASMETA)) &&
			(ISSET(dp->dm_flags, DM_NONR))) {
		if ((error = dmfs_callwait(vp, 0)))
			return (error);
	}

	error = VOP_BMAP(dp->dm_lowervp, ap->a_bn, ap->a_vpp, ap->a_bnp,
			ap->a_runp);

	return (error);
}

/*
 * Some other routine, which has vp locked, has determined that we might
 * want to wake up our sleepers.
 */

void
dmfs_wakeup(vp)
	struct vnode *vp;
{
	struct dm_node	*dp = VTODM(vp);

	if ((ISSET(dp->dm_flags, DM_HASMETA)) &&
				(ISSET(dp->dm_flags, DM_IOWANT))) {
		CLR(dp->dm_flags, DM_IOWANT);
		wakeup(vp);
	}
}

int
dmfs_callrash(vp, user)
	struct vnode *vp;
	uid_t user;
{
	struct	dm_node		*dp = VTODM(vp);
	int	error;

	if ((error = dmfs_sendmessage(vp, RASH_RESTORE, 0, user))) {
		SET(dp->dm_flags, DM_RESTOREFAIL);
	} else {
		CLR(dp->dm_flags, DM_RESTOREFAIL);
		SET(dp->dm_flags, DM_RIP);
	}
	return (error);
}

int
dmfs_callwait(vp, user)
	struct vnode *vp;
	uid_t user;
{
	int error;

	if (!ISSET(VTODM(vp)->dm_flags, DM_RIP))
		if ((error = dmfs_callrash(vp, user)))
			return (error);

	return (dmfs_wait(vp));
}

int
dmfs_rashdelete(vp, user, meta, flag)
	struct vnode *vp;
	uid_t user;
	struct dmfs_mdata *meta;
	int flag;
{
	struct dm_node	*dp = VTODM(vp);
	if (flag != 0) {
		/* This metadata was supposed to be for this node but isn't. */
		dp->dm_bfid	= meta->dmfsm_bfid;
		dp->dm_dbuid	= meta->dmfsm_dbuid;
	}

	return (dmfs_sendmessage(vp, RASH_OBSOLETE, 0, user));
}

static int dmfs_message_seq = 0;

int
dmfs_sendmessage(vp, message_type, todaemon, user)
	struct vnode *vp;
	u_int32_t	message_type;
	dev_t		todaemon;
	uid_t		user;
{
	struct	dm_node		*dp = VTODM(vp);
	union	kdcommsg	kmsg;
	struct	rashmsg		*rmsg;
	int			error;

	rmsg = &kmsg.rash_msg;
	rmsg->rm_type		= message_type;
	rmsg->rm_sequence	= dmfs_message_seq++;
	rmsg->rm_bfid		= dp->dm_bfid;
	rmsg->rm_dbuid		= dp->dm_dbuid;
	rmsg->rm_uid		= user;
	rmsg->rm_fh.fh_fsid = vp->v_mount->mnt_stat.f_fsid;
	if ((error = VFS_VPTOFH(vp, &rmsg->rm_fh.fh_fid)))
		return (error);

	error = kdcput(todaemon, &kmsg);

	return (error);
}
