/*	$NetBSD: dmfs_subr.c,v 1.13 1998/12/29 06:50:10 waddell Exp $	*/

/*
 * Copyright (c) 1992, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software donated to Berkeley by
 * Jan-Simon Pendry.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	from: Id: lofs_subr.c,v 1.11 1992/05/30 10:05:43 jsp Exp
 *	@(#)dmfs_subr.c	8.7 (Berkeley) 12/29/98
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/proc.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/vnode.h>
#include <sys/fcntl.h>
#include <sys/mount.h>
#include <sys/namei.h>
#include <sys/malloc.h>
#include <machine/bswap.h>
#include <miscfs/specfs/specdev.h>
#include <miscfs/dmfs/dmfs.h>
#include <miscfs/dmfs/dmfs_mdata.h>
#include <miscfs/dmfs/dmfs_extern.h>
#include <miscfs/genfs/layer_extern.h>

/*
 * Make a new dm_node node.
 * Vp is the alias vnode, lowervp is the lower vnode.
 * Maintain a reference to lowervp.
 */
int
dm_node_alloc(mp, lowervp, vpp)
	struct mount *mp;
	struct vnode *lowervp;
	struct vnode **vpp;
{
	struct layer_node_hashhead *hd;
	struct layer_mount *lmp = MOUNTTOLAYERMOUNT(mp);
	struct dm_node *xp;
	struct vnode *vp, *nvp;
	int error;
	extern int (**dead_vnodeop_p) __P((void *));

	if ((error = getnewvnode(VT_DMFS, mp, dm_vnodeop_p, &vp)) != 0)
		return (error);
	vp->v_type = lowervp->v_type;
	vp->v_flag |= VLAYER;

	MALLOC(xp, struct dm_node *, sizeof(struct dm_node), M_TEMP,
	    M_WAITOK);
	if (vp->v_type == VBLK || vp->v_type == VCHR) {
		MALLOC(vp->v_specinfo, struct specinfo *,
		    sizeof(struct specinfo), M_VNODE, M_WAITOK);
		vp->v_hashchain = NULL;
		vp->v_rdev = lowervp->v_rdev;
	}

	vp->v_data = xp;
	xp->dm_vnode = vp;
	xp->dm_lowervp = lowervp;
	xp->dm_flags = 0;
	xp->dm_restorepid = 0;
	xp->dm_archpid = 0;

	/* Get opaque data for node, if it exists. */
	error = dmfs_read_mdata(vp, 0);

	/*
	 * Before we insert our new node onto the hash chains,
	 * check to see if someone else has beaten us to it.
	 * (We could have slept in MALLOC.)
	 *
	 * Though this will only be a problem over NFS as we had
	 * the lower node locked, and only NFS doesn't do locking..
	 */
	if ((error) || ((nvp = layer_node_find(mp, lowervp)) != NULL)) {
		*vpp = nvp;

		/* free the substructures we've allocated. */
		FREE(xp, M_TEMP);
		if (vp->v_type == VBLK || vp->v_type == VCHR)
			FREE(vp->v_specinfo, M_VNODE);

		vp->v_type = VBAD;		/* node is discarded */
		vp->v_op = dead_vnodeop_p;	/* so ops will still work */
		vrele(vp);			/* get rid of it. */
		return (0);
	}

	simple_lock(&lmp->layerfs_hashlock);
	/*
	 * Lock fix
	 */

	vp->v_vnlock = lowervp->v_vnlock;
	LAYERFS_UPPERLOCK(vp, LK_EXCLUSIVE, error);

	*vpp = vp;
	VREF(lowervp);	/* will be vrele'd in dm_node_create */
	hd = LAYER_NHASH(lmp, lowervp);
	LIST_INSERT_HEAD(hd, (struct layer_node *)xp, layer_hash);
	simple_unlock(&lmp->layerfs_hashlock);
	return (0);
}

void
dmfs_nullmeta(dp)
	struct dm_node *dp;
{
	memset(&dp->dm_bfid, 0, sizeof(dp->dm_bfid));
	dp->dm_flags		= 0;
	dp->dm_archsize		= 0;
	dp->dm_bbarrier		= 0;
	dp->dm_atime		= 0;
	dp->dm_atimensec	= 0;
	dp->dm_mtime		= 0;
	dp->dm_mtimensec	= 0;
	dp->dm_rtime		= 0;
	dp->dm_artime		= 0;
	dp->dm_gen		= 0;
}

int
dmfs_mount_mfile(vp, mp)
	struct	vnode		*vp;
	struct	mount		*mp;
{
	union {
		struct vattr	uva;
		struct uio	uuio;
		struct statfs	ustatfs;
	} uarea;
	struct	vattr		*va;
	struct	iovec		iov;
	struct	uio		*uio;
	struct	statfs		*fsstat;
	struct	dmfs_mount_data	*dmpmd;
	struct	vnode		*nvp;
	struct	dm_mount	*dmp = MOUNTTODMMOUNT(mp);
	struct	mount		*mp1;
	int			error;
	long			filecount;
	u_quad_t		expsize;

	dmpmd = (struct dmfs_mount_data *)malloc(sizeof(struct dmfs_mount_data),
			M_UFSMNT, M_WAITOK);	/* XXX */
	if (dmpmd == NULL)
		return (ENOMEM);

	uio = &uarea.uuio;
	iov.iov_base	= dmpmd;
	iov.iov_len	= sizeof(*dmpmd);
	uio->uio_iov	= &iov;
	uio->uio_iovcnt	= 1;
	uio->uio_offset	= 0;
	uio->uio_resid	= sizeof(*dmpmd);
	uio->uio_segflg	= UIO_SYSSPACE;
	uio->uio_rw	= UIO_READ;
	uio->uio_procp	= curproc;

	error = VOP_READ(vp, uio, IO_NODELOCKED | IO_UNIT, dmp->dmm_ucred);
	if (error)
		goto out;

	/* now check things out. */
	if (dmpmd->dmmnt_magic != DMFS_MOUNT_MAGIC) {
		if (bswap32(dmpmd->dmmnt_magic) != DMFS_MOUNT_MAGIC) {
printf("Bad magic\n");
			error = EINVAL;
			goto out;
		} else {
printf("Swapped meta\n");
			SET(dmp->dmm_flags, DMM_SWAPMD);
			dmpmd->dmmnt_mntsize = bswap16(dmpmd->dmmnt_mntsize);
			dmpmd->dmmnt_mntvers = bswap16(dmpmd->dmmnt_mntvers);
			dmpmd->dmmnt_nodesize = bswap16(dmpmd->dmmnt_nodesize);
			dmpmd->dmmnt_nodevers = bswap16(dmpmd->dmmnt_nodevers);
		}
	}

	/*
	 * Now the size & versions for the mount and the node data are in
	 * host byte order. Do they make sense?
	 */
	switch (dmpmd->dmmnt_mntvers) {
	case 0:
printf("Version 0 mount");
		if ((dmpmd->dmmnt_mntsize != sizeof(*dmpmd)) ||
			(dmpmd->dmmnt_nodevers != 0) ||
			(dmpmd->dmmnt_nodesize != sizeof(struct dmfs_mdata))) {
printf(" either with bad lengths, or vers != 0 nodes\n");
		    error = EINVAL;
		    goto out;
		}
printf(" with good node info\n");
		if (ISSET(dmp->dmm_flags, DMM_SWAPMD)) {
			dmpmd->dmmnt_flags = bswap32(dmpmd->dmmnt_flags);
			dmpmd->dmmnt_files = bswap32(dmpmd->dmmnt_files);
			dmpmd->dmmnt_fsid.val[0] =
					bswap32(dmpmd->dmmnt_fsid.val[0]);
			dmpmd->dmmnt_fsid.val[1] =
					bswap32(dmpmd->dmmnt_fsid.val[1]);
		}
		break;	

	default:
printf("Bad mount version\n");
		error = EINVAL;
		goto out;
	}

	mp->mnt_stat.f_fsid.val[1] = dmpmd->dmmnt_fsid.val[1];
	mp->mnt_stat.f_fsid.val[0] = 0;
printf("Checking fsid %x:%x.\n", dmpmd->dmmnt_fsid.val[0],
			dmpmd->dmmnt_fsid.val[1]);
	if ((mp1 = vfs_getvfs(&dmpmd->dmmnt_fsid))) {
		error = EINVAL;
printf("Dup fsid from fs on %s.\n", mp1->mnt_op->vfs_name);
		goto out;
	}
	mp->mnt_stat.f_fsid = dmpmd->dmmnt_fsid;

printf("using fsid %x:%x\n", mp->mnt_stat.f_fsid.val[0],
		mp->mnt_stat.f_fsid.val[1]);

	fsstat = &uarea.ustatfs;
	if ((error = VFS_STATFS(dmp->dmm_vfs, fsstat, curproc)))
		goto out;
	filecount = fsstat->f_files;
#if 0
	if (dmpmd->dmmnt_files != filecount) {
		error = EINVAL;
		goto out;
	}
#endif

	expsize = (u_quad_t)filecount * dmpmd->dmmnt_nodesize +
			dmpmd->dmmnt_mntsize;

printf("Expect %ld files\n", filecount);
	va = &uarea.uva;
	if ((error = VOP_GETATTR(vp, va, dmp->dmm_ucred, curproc)))
		goto out;

	if (ISSET(dmpmd->dmmnt_flags, DMFS_MOUNT_ONDEV)) {
printf("Should be device\n");
		if (va->va_type != VCHR) {
			error = ENODEV;
			goto out;
		}
	} else {
printf("Should be file, expect %lld get %lld\n", (long long)expsize, (long long)va->va_size);
		/*
		 * The file must be at least as big as we expect.
		 * bigger than we expect is only ok iff so flagged.
		 */
		if ((va->va_size < expsize) || ((va->va_size > expsize) &&
			    !ISSET(dmpmd->dmmnt_flags, DMFS_MOUNT_BIGGEROK))) {
			error = EINVAL;
			goto out;
		}
	}
printf("File looks fine.. now update all mdata in nodes.\n");

	dmp->dmm_mnode = vp;
	dmp->dmm_mdata = dmpmd;

	simple_lock(&mntvnode_slock);
loop:
	for (vp = mp->mnt_vnodelist.lh_first; vp != NULL; vp = nvp) {
		/*
		 * If the vnode we're looking at is no longer associated
		 * with our mount point, start over.
		 */
		if (vp->v_mount != mp)
			goto loop;
		simple_lock(&vp->v_interlock);
		nvp = vp->v_mntvnodes.le_next;
		simple_unlock(&mntvnode_slock);
		error = vget(vp, LK_EXCLUSIVE | LK_NOWAIT | LK_INTERLOCK);
		if (error) {
			simple_lock(&mntvnode_slock);
			if (error == ENOENT)
				goto loop;
			continue;
		}
		dmfs_read_mdata(vp, 1);
		vput(vp);
		simple_lock(&mntvnode_slock);
	}
	simple_unlock(&mntvnode_slock);

	return (0);

out:
	free(dmpmd, M_UFSMNT);
	return (error);
}

/*
 * Read metadata from mfile in for given dmfs node
 */
int
dmfs_read_mdata(vp, islocked)
	struct	vnode		*vp;
	int			islocked;
{
	struct	dmfs_mdata	the_data;
	struct	vattr		va;
	struct	iovec		iov;
	struct	uio		uio;
	struct	dm_mount	*dmm = MOUNTTODMMOUNT(vp->v_mount);
	struct	dm_node		*dp = VTODM(vp);
	struct	dmfs_mount_data	*dmpmd;
	int			error;
	u_int32_t		gen;
	long			filenum;

	if ((dmpmd = dmm->dmm_mdata) == NULL) {
		dmfs_nullmeta(dp);
		return (0);
	}

	if ((error = VOP_GETATTR(DMVPTOLOWERVP(vp), &va, NULL, curproc)))
		return (error);
	gen = va.va_gen & 0xffffffff;
	filenum = va.va_fileid;

	iov.iov_base	= &the_data;
	iov.iov_len	= sizeof(struct dmfs_mdata);
	uio.uio_iov	= &iov;
	uio.uio_iovcnt	= 1;
	uio.uio_offset	= (off_t)dmpmd->dmmnt_mntsize +
			((off_t)dmpmd->dmmnt_nodesize) * filenum;
	uio.uio_resid	= sizeof(struct dmfs_mdata);
	uio.uio_segflg	= UIO_SYSSPACE;
	uio.uio_rw	= UIO_READ;
	uio.uio_procp	= NULL;

	if (!islocked) 
	    if ((error = vn_lock(dmm->dmm_mnode, LK_EXCLUSIVE | LK_RETRY)))
		return (error);
	error = VOP_READ(dmm->dmm_mnode, &uio, IO_NODELOCKED | IO_UNIT,
			dmm->dmm_ucred);
	if (!islocked)
		VOP_UNLOCK(dmm->dmm_mnode, 0);
	if (error)
		return (error);

	if (MOUNTTODMMOUNT(vp->v_mount)->dmm_flags & DMM_SWAPMD)
		dmfs_byteswap_meta(&the_data);

	/* If no valid metadata, return ok. But set the generation #
	 * to save a stat in dmfs_write_mdata() */
	if (!ISSET(the_data.dmfsm_flags, DM_HASMETA)) {
		dmfs_nullmeta(VTODM(vp));
		dp->dm_gen = gen;
		dp->dm_filenum = filenum;
		return (0);
	}

	/*
	 * If the generation # doesn't match, someone deleted the underlying
	 * file. Signal a delete message and clear the metadata, leaving
	 * the correct generation # in place.
	 */
	if (gen != the_data.dmfsm_gen) {
		dmfs_rashdelete(vp, the_data.dmfsm_dbuid, &the_data, 1);
		dmfs_nullmeta(VTODM(vp));
		SET(dp->dm_flags, DM_MODIFIED);
		dp->dm_gen = gen;
		dp->dm_filenum = filenum;
		return (0);
	}
	dp->dm_gen	 = gen;
	dp->dm_filenum	 = filenum;

	dmfs_impmeta(vp, &the_data);
	return (0);
}

int
dmfs_write_mdata(vp, islocked)
	struct	vnode		*vp;
	int			islocked;
{
	struct	dm_mount	*dmm = MOUNTTODMMOUNT(vp->v_mount);
	struct	dm_node		*dp = VTODM(vp);
	struct	dmfs_mdata	the_data;
	struct	iovec		iov;
	struct	uio		uio;
	struct	dmfs_mount_data	*dmpmd;
	int			ioflags, error;

	if (((dmpmd = dmm->dmm_mdata) == NULL) ||
			ISSET(dmm->dmm_flags, DMM_METARO)) {
		return (EINVAL);
	}
	if ISSET(dmm->dmm_flags, DMM_METALOCKED)
		islocked = 1;
	if ISSET(dmm->dmm_flags, DMM_NODES_LOCKED)
		if ((error = dmfs_node_wait(vp)))
			return (error);

	dmfs_expmeta(vp, &the_data);
	the_data.dmfsm_gen	= dp->dm_gen;
	the_data.dmfsm_spare[0]	= 0;
	the_data.dmfsm_spare[1]	= 0;
	the_data.dmfsm_spare[2]	= 0;

	iov.iov_base	= &the_data;
	iov.iov_len	= sizeof(struct dmfs_mdata);
	uio.uio_iov	= &iov;
	uio.uio_iovcnt	= 1;
	uio.uio_offset	= (off_t)dmpmd->dmmnt_mntsize +
			((off_t)dmpmd->dmmnt_nodesize) * dp->dm_filenum;
	uio.uio_resid	= sizeof(struct dmfs_mdata);
	uio.uio_segflg	= UIO_SYSSPACE;
	uio.uio_rw	= UIO_WRITE;
	uio.uio_procp	= NULL;

	ioflags = IO_NODELOCKED | IO_UNIT;
	if ISSET(dmm->dmm_flags, DMM_SYNCMDATA)
		ioflags |= IO_SYNC;
	if (!islocked) 
	    if ((error = vn_lock(dmm->dmm_mnode, LK_EXCLUSIVE | LK_RETRY)))
		return (error);
	error = VOP_WRITE(dmm->dmm_mnode, &uio, ioflags, dmm->dmm_ucred);
	if (!islocked)
		VOP_UNLOCK(dmm->dmm_mnode, 0);
	return (error);
}

void
dmfs_byteswap_meta(meta)
	struct dmfs_mdata *meta;
{
	meta->dmfsm_flags	= bswap32(meta->dmfsm_flags);
	meta->dmfsm_gen		= bswap32(meta->dmfsm_gen);
	meta->dmfsm_arch_size	= bswap64(meta->dmfsm_arch_size);
	meta->dmfsm_bbarrier	= bswap64(meta->dmfsm_bbarrier);
	meta->dmfsm_dbuid	= bswap32(meta->dmfsm_dbuid);
	/* dmfsm_bfid is endian-independent */
	meta->dmfsm_atime	= bswap32(meta->dmfsm_atime);
	meta->dmfsm_atimensec	= bswap32(meta->dmfsm_atimensec);
	meta->dmfsm_mtime	= bswap32(meta->dmfsm_mtime);
	meta->dmfsm_mtimensec	= bswap32(meta->dmfsm_mtimensec);
	meta->dmfsm_rtime	= bswap32(meta->dmfsm_rtime);
	meta->dmfsm_artime	= bswap32(meta->dmfsm_artime);
}

/*
 * Following routines copy data between vnode private storage & a
 * metadata storage struct. Certain dmfs_mdata fields aren't copied, like
 * the generation number and internal state flags.
 *
 * dmfs_impmeta imports from a dmfs_mdata to a vnode, and dmfs_expmeta
 * exports from the vnode to a dmfs_mdata. dmfs_blankmeta clears out
 * fields not touched by dmfs_expmeta.
 */

void
dmfs_impmeta(vp, meta)
	struct	vnode		*vp;
	struct	dmfs_mdata	*meta;
{
	struct	dm_node *dp = VTODM(vp);

	CLR(dp->dm_flags, DM_FLAGMASK);
	SET(dp->dm_flags, meta->dmfsm_flags & DM_FLAGMASK);
	dp->dm_dbuid		= meta->dmfsm_dbuid;
	dp->dm_archsize		= meta->dmfsm_arch_size;
	dp->dm_bbarrier		= meta->dmfsm_bbarrier;
	dp->dm_atime		= meta->dmfsm_atime;
	dp->dm_atimensec	= meta->dmfsm_atimensec;
	dp->dm_mtime		= meta->dmfsm_mtime;
	dp->dm_mtimensec	= meta->dmfsm_mtimensec;
	dp->dm_rtime		= meta->dmfsm_rtime;
	dp->dm_artime		= meta->dmfsm_artime;
	dp->dm_bfid		= meta->dmfsm_bfid;
}

void
dmfs_expmeta(vp, meta)
	struct vnode *vp;
	struct dmfs_mdata *meta;
{
	struct dm_node	*dp = VTODM(vp);

	meta->dmfsm_flags	= dp->dm_flags & DM_FLAGMASK;
	meta->dmfsm_dbuid	= dp->dm_dbuid;
	meta->dmfsm_arch_size	= dp->dm_archsize;
	meta->dmfsm_bbarrier	= dp->dm_bbarrier;
	meta->dmfsm_atime	= dp->dm_atime;
	meta->dmfsm_atimensec	= dp->dm_atimensec;
	meta->dmfsm_mtime	= dp->dm_mtime;
	meta->dmfsm_mtimensec	= dp->dm_mtimensec;
	meta->dmfsm_rtime	= dp->dm_rtime;
	meta->dmfsm_artime	= dp->dm_artime;
	meta->dmfsm_bfid        = dp->dm_bfid;
}

void
dmfs_blankmeta(meta)
	struct dmfs_mdata *meta;
{
	meta->dmfsm_gen = 0;
}
