/*	$NetBSD: dmfs_vfsops.c,v 1.20 1998/12/29 20:51:09 waddell Exp $	*/

/*
 * Copyright (c) 1992, 1993, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software donated to Berkeley by
 * Jan-Simon Pendry.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	from: Id: lofs_vfsops.c,v 1.9 1992/05/30 10:26:24 jsp Exp
 *	from: @(#)lofs_vfsops.c	1.2 (Berkeley) 6/18/92
 *	@(#)dmfs_vfsops.c	12/29/98
 */

/*
 * DMFS Layer
 * (See dmfs_vnops.c for a description of what this does.)
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/time.h>
#include <sys/proc.h>
#include <sys/types.h>
#include <sys/vnode.h>
#include <sys/fcntl.h>
#include <sys/mount.h>
#include <sys/namei.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <miscfs/dmfs/dmfs.h>
#include <miscfs/dmfs/dmfs_extern.h>
#include <miscfs/dmfs/dmfs_mdata.h>
#include <miscfs/genfs/layer_extern.h>

int	dmfs_mount __P((struct mount *, const char *, void *,
			  struct nameidata *, struct proc *));
int	dmfs_unmount __P((struct mount *, int, struct proc *));
int	dmfs_sync __P((struct mount *, int, struct ucred *, struct proc *));

#define NDMNODECACHE 64

/*
 * Mount dm layer
 */
int
dmfs_mount(mp, path, data, ndp, p)
	struct mount *mp;
	const char *path;
	void *data;
	struct nameidata *ndp;
	struct proc *p;
{
	int error;
	struct dm_args args;
	struct vnode *lowerrootvp, *vp, *mdatavp;
	struct dm_mount *dmp;
	struct dm_node *dp;
	struct ucred *credp;
	size_t size;

#ifdef DMFS_DIAGNOSTIC
	printf("dmfs_mount(mp = %p)\n", mp);
#endif

	/*
	 * Get argument
	 */
	error = copyin(data, (caddr_t)&args, sizeof(struct dm_args));
	if (error)
		return (error);

	/*
	 * Update: mainly handle export updates.
	 */
	if (mp->mnt_flag & MNT_UPDATE) {
		dmp = MOUNTTODMMOUNT(mp);
		/*
		 * Might add a lot of things later, like add/del writability
		 * or backup quiescence.
		 */
		if (args.la.target == 0)
		      return (vfs_export(mp, &dmp->dmm_export, &args.la.export));
		else
			return (EOPNOTSUPP);
	}

	/*
	 * Find lower node
	 */
	lowerrootvp = mp->mnt_vnodecovered;
	if ((error = vget(lowerrootvp, LK_EXCLUSIVE | LK_RETRY)))
		return (error);

	/*
	 * First cut at fixing up upper mount point
	 */
	dmp = (struct dm_mount *) malloc(sizeof(struct dm_mount),
				M_UFSMNT, M_WAITOK);	/* XXX */
	if (dmp == NULL)
		return (ENOMEM);
	memset((caddr_t)dmp, 0, sizeof(struct dm_mount));

	mp->mnt_data = (qaddr_t) dmp;
	dmp->dmm_vfs = lowerrootvp->v_mount;
	if (dmp->dmm_vfs->mnt_flag & MNT_LOCAL)
		mp->mnt_flag |= MNT_LOCAL;

	/*
	 * Make sure that the mount point is sufficiently initialized
	 * that the node create call will work.
	 */
	dmp->dmm_size = sizeof(struct dm_node);
	dmp->dmm_tag = VT_DMFS;
	dmp->dmm_bypass = layer_bypass;	/* default bypass is fine */
	dmp->dmm_alloc = dm_node_alloc;	/* default alloc isn't */
	dmp->dmm_vnodeop_p = dm_vnodeop_p;
	simple_lock_init(&dmp->dmm_hashlock);
	dmp->dmm_node_hashtbl = hashinit(NDMNODECACHE, M_CACHE, M_WAITOK,
			&dmp->dmm_node_hash);
	/*
	 * Mark metadata R/O until we're sure it's writable.
	 */
	dmp->dmm_flags = DMM_METARO;

	/*
	 * Save reference.  Each mount also holds
	 * a reference on the root vnode.
	 */
	error = layer_node_create(mp, lowerrootvp, &vp);
	/*
	 * Make sure the node alias worked. If so, lowerrootvp is
	 * a dangling reference (shouldn't vrele it). If there was an
	 * error, lowerrootvp is still there and vp is nothing.
	 */
	if (error) {
		vput(lowerrootvp);
		goto out1;
	}
	dp = VTODM(vp);

	/*
	 * Keep a held reference to the root vnode.
	 * It is vrele'd in dmfs_unmount.
	 */
	vp->v_flag |= VROOT;
	dmp->dmm_rootvp = vp;
	/*
	 * Unlock the root node
	 */
	VOP_UNLOCK(vp, 0);

	NDINIT(ndp, LOOKUP, FOLLOW | LOCKLEAF,
		UIO_USERSPACE, args.la.target, p);
	if ((error = namei(ndp)) != 0)
		goto out2;

	/* Now see about mountpoint meta data */
	credp = crdup(p->p_ucred);
	mdatavp = ndp->ni_vp;
	if ((error = VOP_OPEN(mdatavp, FREAD | FWRITE, credp, p)))
		goto out3;

	dmp->dmm_ucred = credp;
	if ((error = dmfs_mount_mfile(mdatavp, mp)))
		goto out3;

	CLR(dmp->dmm_flags, DMM_METARO);
	VOP_UNLOCK(mdatavp, 0);

	(void) copyinstr(path, mp->mnt_stat.f_mntonname, MNAMELEN - 1, &size);
	memset(mp->mnt_stat.f_mntonname + size, 0, MNAMELEN - size);
	(void) copyinstr(args.la.target, mp->mnt_stat.f_mntfromname, MNAMELEN - 1, 
	    &size);
	memset(mp->mnt_stat.f_mntfromname + size, 0, MNAMELEN - size);
#ifdef DMFS_DIAGNOSTIC
	printf("dmfs_mount: lower %s, alias at %s\n",
	    mp->mnt_stat.f_mntfromname, mp->mnt_stat.f_mntonname);
#endif
	return (0);

out3:				/* got (locked) mdata vnode */
	vput(mdatavp);
	crfree(credp);
out2:				/* got root vnode for our mp */
	vrele(vp);
	vgone(vp);
out1:				/* malloc succeeded. */
	free(dmp, M_UFSMNT);	/* XXX */
	return (error);
}

/*
 * Free reference to dm layer
 */
int
dmfs_unmount(mp, mntflags, p)
	struct mount *mp;
	int mntflags;
	struct proc *p;
{
	struct	dm_mount	*dmp = MOUNTTODMMOUNT(mp);
	struct	vnode		*the_rootvp = dmp->dmm_rootvp, *mdata;
	int error;
	int flags = 0;

#ifdef DMFS_DIAGNOSTIC
	printf("dmfs_unmount(mp = %p)\n", mp);
#endif

	if (mntflags & MNT_FORCE)
		flags |= FORCECLOSE;

	/*
	 * Clear out buffer cache.  I don't think we
	 * ever get anything cached at this level at the
	 * moment, but who knows...
	 */
#if 0
	mntflushbuf(mp, 0); 
	if (mntinvalbuf(mp, 1))
		return (EBUSY);
#endif
	if (the_rootvp->v_usecount > 1) {
		return (EBUSY);
	}
	if ((error = vflush(mp, the_rootvp, flags)) != 0) {
		return (error);
	}

#ifdef DMFS_DIAGNOSTIC
	vprint("alias root of lower", the_rootvp);
#endif	 
	mdata = dmp->dmm_mnode;

	if (vn_lock(mdata, LK_EXCLUSIVE | LK_RETRY)) {
		/* Oh well, we're not syncing mdata! */
		dmp->dmm_mnode = NULL;	/* dmfs_sync() now fails */
		vrele(mdata);	/* bye bye mdata node */
	} else {
		dmfs_dosync(mp);
		VOP_CLOSE(mdata, FREAD | FWRITE, dmp->dmm_ucred, p);
		dmp->dmm_mnode = NULL;	/* dmfs_sync() now fails */
		vput(mdata);
	}
	crfree(dmp->dmm_ucred);

	/*
	 * Release reference on underlying root vnode
	 */
	vrele(the_rootvp);
	/*
	 * And blow it away for future re-use
	 */
	vgone(the_rootvp);
	/*
	 * Finally, throw away the dm_mount structure
	 */
	free(dmp->dmm_mdata, M_UFSMNT);	/* XXX */
	free(dmp, M_UFSMNT);	/* XXX */
	mp->mnt_data = 0;
	return 0;
}

int
dmfs_sync(mp, waitfor, cred, p)
	struct mount *mp;
	int waitfor;
	struct ucred *cred;
	struct proc *p;
{
	struct	dm_mount	*dmp = MOUNTTODMMOUNT(mp);
	struct	vnode		*mdata;
	int error;

	if ((mdata = dmp->dmm_mnode) == NULL)
		return (0);
	if ((error = vget(mdata, LK_EXCLUSIVE | LK_RETRY))) {
		/* Oh well, we're not syncing mdata! */
	} else {
		dmfs_dosync(mp);
		vput(mdata);
	}
	return (0);
}

/*
 * Assumes that the mfile node is locked before entry
 */
void
dmfs_dosync(mp)
	struct	mount *mp;
{
	struct	vnode	*vp, *nvp;
	int		error;

	if ISSET(MOUNTTODMMOUNT(mp)->dmm_flags, DMM_METARO)
		return;

	simple_lock(&mntvnode_slock);
	SET(MOUNTTODMMOUNT(mp)->dmm_flags, DMM_METALOCKED);
loop:
	for (vp = mp->mnt_vnodelist.lh_first; vp != NULL; vp = nvp) {
		/*
		 * If the vnode we're looking at is no longer associated
		 * with our mount point, start over.
		 */
		if (vp->v_mount != mp)
			goto loop;
		simple_lock(&vp->v_interlock);
		nvp = vp->v_mntvnodes.le_next;
		simple_unlock(&mntvnode_slock);
		error = vget(vp, LK_EXCLUSIVE | LK_NOWAIT | LK_INTERLOCK);
		if (error) {
			simple_lock(&mntvnode_slock);
			if (error == ENOENT)
				goto loop;
			continue;
		}
		if (ISSET(VTODM(vp)->dm_flags, DM_MODIFIED)) {
			dmfs_write_mdata(vp, 1);
		}
		vput(vp);
		simple_lock(&mntvnode_slock);
	}
	CLR(MOUNTTODMMOUNT(mp)->dmm_flags, DMM_METALOCKED);
	simple_unlock(&mntvnode_slock);

	return;
}


extern struct vnodeopv_desc dm_vnodeop_opv_desc;

struct vnodeopv_desc *dmfs_vnodeopv_descs[] = {
	&dm_vnodeop_opv_desc,
	NULL,
};

struct vfsops dmfs_vfsops = {
	MOUNT_DMFS,
	dmfs_mount,
	layerfs_start,
	dmfs_unmount,
	layerfs_root,
	layerfs_quotactl,
	layerfs_statfs,
	dmfs_sync,
	layerfs_vget,
	layerfs_fhtovp,
	layerfs_vptofh,
	layerfs_init,
	layerfs_sysctl,
	NULL,				/* vfs_mountroot */
	layerfs_checkexp,
	dmfs_vnodeopv_descs,
};
