
/*
 * kdcom - kernel <-> daemon communication driver
 *
 * This pseudo device driver is designed to provide a communications
 * vehicle for two way message passing between the kernel and a daemon
 * process.  
 *
 * This driver is setup as a character driver, though the message passed is
 * limited in size by the kdcom data structure.  A user process interfaces 
 * with this driver as it does with any device, via open(), read(), write(),
 * ioctl(), and close() calls.
 *
 * The kernel has two entry points that roughly correspond to the interrupt
 * level routines of real devices. Kdcput() is used by the kernel to send a
 * message to a daemon; kdcget() is used by athe kernel to receive a message
 * from a daemon.
 */
/*
 * Originally written by Alan Poston.
 *
 * Ported to NetBSD by Bill Studenmund and Matt Jacob.
 */

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/systm.h>
#include <sys/buf.h>
#include <sys/conf.h>
#include <sys/device.h>
#include <sys/disk.h>
#include <sys/poll.h>
#include <sys/proc.h>
#include <sys/select.h>
#include <sys/conf.h>
#include <sys/fcntl.h>
#include <sys/ioccom.h>
#include <sys/vnode.h>

#include "kdc.h"
#include "rash.h"
#include "kdcom.h"

#define	NKDCBUF		(64*NKDC)

struct kdcom kdcom[NKDCBUF];
struct kdchead kdcfree;
struct kdchead kdcin[NKDC];
struct kdchead kdcout[NKDC];

static int kdcio __P((struct kdcom *, struct uio *));
static void put_frame __P((struct kdcom *, struct kdchead *));
static int get_frame __P((struct kdchead *, int, struct kdcom **));

cdev_decl(kdc_);

void	kdcattach	__P((int));
int	kdcdetach	__P((void));

int	nkdcom = 0;

void
kdcattach(n)
	int n;
{
	int i;
	struct kdchead *khp;

	khp = &kdcfree;
	SIMPLEQ_INIT(&khp->kdqh);
	for(i = 0; i < NKDCBUF; i++) {
		SIMPLEQ_INSERT_TAIL(&khp->kdqh, &kdcom[i], kc_list);
	}
	/*qsema_init(&kdcfree.kch_lock, 1, QSTC_BUFHD, &bufhd_qclassinfo); */

	nkdcom = NKDC;

	for (i = 0; i < nkdcom; i++) {
		khp = &kdcin[i];
		khp->kch_open = 0;
		khp->kch_wait = 0;
		SIMPLEQ_INIT(&khp->kdqh);
		khp->kch_selinfo.si_pid = 0;
		khp->kch_selinfo.si_flags = 0;
	/* qsema_init(&khp->kch_lock, 1, QSTC_BUFHD, &bufhd_qclassinfo); */
		khp = &kdcout[i];
		khp->kch_open = 0xdeadbeef; /* only kept on kdcin[] */
		khp->kch_wait = 0;
		SIMPLEQ_INIT(&khp->kdqh);
		khp->kch_selinfo.si_pid = 0;
		khp->kch_selinfo.si_flags = 0;
	/* qsema_init(&khp->kch_lock, 1, QSTC_BUFHD, &bufhd_qclassinfo); */
	}
}

/*
 * Detach the kdcom's. Only called from lkm when we're detaching.
 * Return EBUSY if there are open kdcom's.
 *
 * As we have static storage at the moment, we don't need to free any memory.
 */
int
kdcdetach(void)
{
	int i;

	for (i = 0; i < nkdcom; i++)
		if (kdcin[i].kch_open)
			return (EBUSY);

	nkdcom = 0;
	return (0);
}

/*
 * Open kernel - daemon pseudo device
 */

int
kdc_open(dev, flags, mode, p)
	dev_t dev;
	int flags;
	int mode;
	struct proc *p;
{
	struct kdchead *khp;
	int error = 0;

	if (minor(dev) >= nkdcom) {
		return(ENXIO);
	}

	/*
	 * set open flag for device, note: open flag kept only in kdcin header,
	 * not in kdcout header
	 */
	khp = &kdcin[minor(dev)];
/*	qsema_acq(&khp->kch_lock, PZERO, "kdcop"); */
	if (khp->kch_open) {
		error = EBUSY;	/* this pseudo device is busy  */
	} else {
		khp->kch_open = flags & 0xf;
	}
/*	qsema_rel(&khp->kch_lock,"kdcop"); */
	return (error);
}

int
kdc_close(dev, flags, mode, p)
	dev_t dev;
	int flags;
	int mode;
	struct proc *p;
{
	struct kdchead *khp;
	int	s;

	khp = &kdcin[minor(dev)];
	/* qsema_acq(&khp->kch_lock, PZERO, "kdccl"); */
	s = splhigh();
	khp->kch_open = 0;
	splx(s);
	if (khp->kch_wait) {
		khp->kch_wait = 0;
		wakeup((caddr_t)khp);
	}
	/* qsema_rel(&khp->kch_lock, "kdccl"); */
	khp = &kdcout[minor(dev)];
	/* qsema_acq(&khp->kch_lock, PZERO, "kdccl"); */
	if (khp->kch_wait) {
		khp->kch_wait = 0;
		wakeup((caddr_t)khp);
	}
	/* qsema_rel(&khp->kch_lock, "kdccl"); */
	return(0);
}

int
kdc_read(dev, uio, flags)
	dev_t dev;
	struct uio *uio;
	int flags;
{
	struct kdchead *khp;
	struct kdcom   *kp;
	int	error;

	khp = &kdcin[minor(dev)];
	if ((error = get_frame(khp, (flags & IO_NDELAY), &kp)) == 0) {
		error = kdcio(kp, uio);
		put_frame(kp, &kdcfree);
	}
	return (error);
}

int
kdc_write(dev, uio, flags)
	dev_t dev;
	struct uio *uio;
	int flags;
{
	struct kdcom *kp;
	struct kdchead *khp;
	int error;

	khp = &kdcout[minor(dev)];
	if ((error = get_frame(&kdcfree, (flags & IO_NDELAY), &kp)) == 0) {
		error = kdcio(kp, uio);
		put_frame(kp,khp);
	}
	return (error);
}

/* ARGSUSED */
int
kdc_ioctl(dev, cmd, data, flag, p)
	dev_t dev;
	u_long cmd;
	caddr_t data;
	int flag;
	struct proc *p;
{
	struct kdcom   *kp;
	dev_t other_dev;
	int error = 0;

	other_dev = minor(*(dev_t *)data);
	if (other_dev >= nkdcom) {
		 return (ENXIO);
	}

	switch (cmd) {
	case KDCIOCTL_LOOP_OUT2IN:
		if ((error = get_frame(&kdcout[minor(dev)], 0, &kp)) != 0)
			break;
		error = kdcput(other_dev, &kp->kc_message);
		put_frame(kp, &kdcfree);
		break;
	case KDCIOCTL_LOOP_IN2OUT:
		if ((error = get_frame(&kdcfree, 0, &kp)) != 0)
			break;
		kdcget(other_dev, &kp->kc_message, 0);
		put_frame(kp, &kdcin[minor(dev)]);
		break;
	default:
		error = EINVAL;
		break;
	}
	return (error);
}

int
kdc_poll(dev, events, p)
	dev_t dev;
	int events;
	struct proc *p;
{
	int revents;
	struct kdchead *khp;

	revents = 0;

	khp = &kdcin[minor(dev)];
	if (events & (POLLIN | POLLRDNORM)) {
		if (SIMPLEQ_FIRST(&khp->kdqh) != NULL)
			revents |= events & (POLLIN | POLLRDNORM);
	}

	if (revents == 0) {
		if (events & (POLLIN | POLLRDNORM))
			selrecord(p, &khp->kch_selinfo);
	}

	return (revents);
}


/* 
 * kdcio - i/o routine for kdcread and kdcwrite
 *         this routines moves the data between the kernel and the user
 */
static
int
kdcio(kp, uio)
	struct kdcom *kp;     /* pointer to message struct */
	struct uio *uio;
{
	int	cnt;

	cnt = KDCMESSAGE_SIZE;
	if (cnt > uio->uio_resid)
		cnt = uio->uio_resid;
	return (uiomove(&kp->kc_message, cnt, uio));
}

/*
 * kdcput - put (send) kernel message to daemon; via dev's in queue
 *	returns: 0 = ok
 *		 n = Error:
 *
 *	ENODEV	No such device
 *	ENXIO	Device not open for reading
 */
int
kdcput(dev, buf)
	int dev;           /* minor dev number   */
	union kdcommsg *buf;         /* pointer to message */
{
	struct kdchead *khp;
	struct kdcom   *pkc;
	int error;
	int dv = minor(dev);

	if (dv >= nkdcom) {
		return (ENODEV);
	}
	khp = &kdcin[dv];
	if (khp->kch_open & FREAD) {
		/* somebody reading and frame free */
		if ((error = get_frame(&kdcfree, 0, &pkc)) != 0)
			return (error);
		memcpy(&pkc->kc_message, buf, KDCMESSAGE_SIZE);
		put_frame(pkc, khp);
		return (0);
	} else {
		return (ENXIO);
	}
}

/*
 * kdcget - get a message from daemon; dev's out queue
 *
 * returns: 0	= message transferred to buffer
 *      EAGAIN	= no message available and "nowait"  set
 *	ENXIO	= no such kdcom or kdcom not open for write.
 */
int
kdcget(dev, buf, nowait)
	int dev;	/* minor dev number from which to get message      */
	union kdcommsg *buf;	/* kernel buffer into which to move message        */
	int nowait;	/* nowait = 1->do NOT sleep if message unavailable */
{
	struct kdchead *khp;
	struct kdcom   *pkc;
	int error = 0;

	if (dev >= nkdcom) {
		return (ENXIO);
	}
	khp = &kdcout[dev];
	if ((kdcin[dev].kch_open & FWRITE)) {
		/* somebody has it open for write */
		if ((error = get_frame(khp, nowait, &pkc)) != 0) 
			return (error);
		/* message available              */
		memcpy(buf, &pkc->kc_message, KDCMESSAGE_SIZE);
		put_frame(pkc, &kdcfree);
	} else {
		error = ENXIO;
	}
	return (error);
}

static int
get_frame(quep, nowait, theframe)
	struct kdchead *quep;
	int nowait;
	struct kdcom **theframe;
{
	struct kdcom *fp;
	int	error = 0;

	while (1) {
		fp = SIMPLEQ_FIRST(&quep->kdqh);
		if (fp) {
			/* frame is available from queue */
			SIMPLEQ_REMOVE_HEAD(&quep->kdqh, fp, kc_list);
			SIMPLEQ_NEXT(fp, kc_list) = NULL;
			break;
		} else if (nowait) {
			fp = NULL;
			error = EAGAIN;
			break;
		} else {
			/* nowait not set - sleep        */
			quep->kch_wait++;
			error = tsleep((caddr_t)quep, PCATCH | (PZERO+1),
				"kdcom", 0);
			if ((quep->kch_open == 0)  || (error != 0)) {
				/*
				 * Oops! device got closed, or there was
				 * an error
				 */
				fp = NULL;
				break;
			}
		}
	}
	*theframe = fp;
	return(error);
}

static void
put_frame(fp, quep)
	struct kdcom *fp;
	struct kdchead *quep;
{
	SIMPLEQ_INSERT_TAIL(&quep->kdqh, fp, kc_list);
	selwakeup(&quep->kch_selinfo);
	if (quep->kch_wait) {
		quep->kch_wait = 0;
		wakeup((caddr_t)quep);
	}
}

