#include <_dba_impl.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#ifndef PATH_MAX
#define PATH_MAX 1024
#endif

#include <dba.h>
/*******************************************************
   utility functions 

*******************************************************/

/* 
   read the config file and set variables
*/

int _read_dba_conf(DbA_impl* impl)
{
  FILE* fp;
  char* linebuffer, *p;
  struct stat foo;
  char* cfile;

  if(getenv("DEBUG") != NULL)
    cfile = "./metriclib.cf";
  else
    cfile = DBA_CONFFILE;
  
  if((fp = fopen(cfile, "r"))== NULL){
    if((stat(cfile, &foo)) < 0){
      perror(cfile);
    }
    else{
      fprintf(stderr,"libdba: unknown error while opening %s\n", DBA_CONFFILE);
    }

    return(DBA_ERROR);
  }

  if((linebuffer = (char* )(malloc(1025))) == NULL){
    perror("libdba:malloc");
    return(DBA_ERROR);
  }
  
  impl->backend_host = NULL;
  
  while(fgets(linebuffer, 1024, fp) != NULL){
    if((p = strtok(linebuffer, " \t\n")) == NULL)
      continue;
    else if(strcmp(p, "BACKEND_HOST") == 0){
      if((p = strtok(NULL,  " \t\n")) == NULL){
	fprintf(stderr, "libdba: error in config file %s",  DBA_CONFFILE);
	fprintf(stderr, ": no value given after BACKEND_HOST\n");
	free(linebuffer);
	return(DBA_ERROR);
      }
      
      if((impl->backend_host = (char *)(malloc(strlen(p) + 1))) == NULL){
	perror("libdba: malloc");
	free(linebuffer);
	return(DBA_ERROR);
      }

      strcpy(impl->backend_host, p);
    }
    else if(strcmp(p, "USE_HOSTNAME") == 0){
      if((p = strtok(NULL,  " \t\n")) == NULL){
	fprintf(stderr, "libdba: error in config file %s",  DBA_CONFFILE);
	fprintf(stderr, ": no value given after USE_HOSTNAME\n");
	free(linebuffer);
	return(DBA_ERROR);
      }
      free(impl->pg_db);
      if((impl->pg_db = (char *)(malloc(strlen(p) + 1))) == NULL){
	perror("libdba: malloc");
	return(DBA_ERROR);
      }
      strcpy(impl->pg_db, p);
    }
    else if(strcmp(p, "BACKEND_PORT") == 0){
      if((p = strtok(NULL,  " \t\n")) == NULL){
        fprintf(stderr, "libdba: error in config file %s",  DBA_CONFFILE);
        fprintf(stderr, ": no value given after BACKEND_PORT\n");
        free(linebuffer);
        return(DBA_ERROR);
      }
      strcpy(impl->backend_port, p);
    }
    else
      continue;
  }

  free(linebuffer);

  if(impl->backend_host == NULL){
    fprintf(stderr, "libdba: error in config file %s",  DBA_CONFFILE);
    fprintf(stderr, ": BACKEND_HOST not defined\n");
    return(DBA_ERROR);
  }
  return(0);
}


/* 
   destroy memory allocated to an impl, including the impl itself 
*/

void cleanup_impl(DbA_impl *impl)
{
  if(impl == NULL)
    return;

  if( impl->tablename != NULL )
    free(impl->tablename);

  if( impl->_dba_ref != NULL )
    free(impl->_dba_ref);

  if( impl->pg_db != NULL )
    free(impl->pg_db);

  if( impl->backend_host != NULL )
    free(impl->backend_host);

  if( impl->backend_port != NULL )
    free(impl->backend_port);

  if( impl->result != NULL )
    PQclear(impl->result);

  if( impl->db != NULL)
    PQfinish(impl->db );

  if(impl->q_buffer != NULL)
    free(impl->q_buffer);

  free( impl );
}


void cleanup_query(DbA_impl *impl)
{
  if(impl->result != NULL){
    PQclear(impl->result);
    impl->result = NULL;
  }
  if(impl->db != NULL){
    PQfinish(impl->db );
    impl->db = NULL;
  }
}

/* 
   handle the sql commands, select or otherwise. check exit status. 
   also, clean up memory from previous results. 
*/

int sql_exec(DbA_impl *impl)
{

  if(impl->result != NULL)
    PQclear(impl->result);

  if(impl->db == NULL){
    if(impl->open(impl)){
      impl->rows = 0;
      impl->current_row=0;
      return(impl->db_errno);
    }
  }

  impl->result = PQexec(impl->db, impl->q_buffer);

  impl->pg_status = PQresultStatus(impl->result);

  switch(impl->pg_status){
  case PGRES_COMMAND_OK:
    impl->db_errno = DBA_SUCCESS;
    impl->rows = 0;
    break;
  case PGRES_TUPLES_OK:
    impl->db_errno = DBA_SUCCESS;
    impl->rows = PQntuples(impl->result);
    impl->current_row = 0;
    break;
  case PGRES_COPY_OUT:
    impl->db_errno = DBA_SUCCESS;
    break;
  case PGRES_COPY_IN:
    impl->db_errno = DBA_SUCCESS;
    break;
  case PGRES_BAD_RESPONSE:
    impl->db_errno = DBA_ERROR;
    cleanup_query(impl);
    impl->current_row = 0;
    impl->rows = 0;
    impl->result = NULL;
    return(-1);
  case PGRES_NONFATAL_ERROR:
    impl->db_errno = DBA_ERROR;
    cleanup_query(impl);
    impl->result = NULL;
    impl->current_row = 0;
    impl->rows = 0;
    return(-1);
  default:
    impl->db_errno = DBA_ERROR;
    cleanup_query(impl);
    impl->result = NULL;
    impl->current_row = 0;
    impl->rows = 0;
    return(-1);
  }

  return(0);
}
/*
  delete table from database and remove
  table_name from descriptions table.
*/

int remove_table(DbA_impl *impl)
{
  sprintf(impl->q_buffer, "Drop table %s", impl->tablename);

  impl->db_errno = DBA_ERROR;
  if(sql_exec(impl) != 0){
    fprintf(stderr, "libdba: Error DROPing %s table\n", impl->tablename);
    return(impl->db_errno);
  }
  sprintf(impl->q_buffer, "Delete from descriptions where table_name = '%s'",
          impl->tablename);
  if(sql_exec(impl) != 0){
    fprintf(stderr, "libdba: Error DELETEing %s from descriptions table\n",
            impl->tablename);
    return(impl->db_errno);
  }
  return(0);
}


/*
   check if table exists
   return -1 on error, 0 if does not exist, +1 if exists 
*/

int check_table_exists(DbA_impl *impl, char* table_name)
{
  sprintf(impl->q_buffer, "select relname from pg_class where relname = '%s'", 
	  table_name);
  
  if(sql_exec(impl) != 0){
    fprintf(stderr, "libdba: Error checking for table existence\n");
    return(-1);
  }
  else if(impl->rows == 0)
    return(0);
  else
    return(1);
}




int create_table(DbA_impl *impl)
{
  char buf[20];
  sprintf(impl->q_buffer, "BEGIN");

  if(sql_exec(impl) != 0){
    fprintf(stderr, "libdba: Error BEGINing the table creation seqeuence\n");
    return(1);
  }

  if(impl->value_type=='i')
    strcpy(buf, "int4");
  else if(impl->value_type == 'f')
    strcpy(buf, "float8");
  else
    sprintf(buf, "text");

  sprintf(impl->q_buffer, "CREATE TABLE %s (dtime datetime, utime int4, value %s, ctime datetime, start_dtime datetime, start_utime int4, subset varchar(32), tag varchar(32))", impl->tablename, buf);

  if(sql_exec(impl) != 0){
    sprintf(impl->q_buffer, "ABORT");
    fprintf(stderr, "libdba: Error creating table\n");
    return(1);
  }

  sprintf(impl->q_buffer, "INSERT into descriptions (table_name, value_type) VALUES ('%s', '%c')", impl->tablename, (char)(impl->value_type));
  
  if(sql_exec(impl) != 0){
    sprintf(impl->q_buffer, "ABORT");
    fprintf(stderr, "libdba: Error creating table description\n");
    return(1);
  }
  else{
    sprintf(impl->q_buffer, "END");
    if(sql_exec(impl)  != 0){
      fprintf(stderr, "libdba: Error ENDing table create. Contact db system admin\n");   
      return(1);
    }
    else
      return(0);
  }
}


int utime_to_dtime(time_t utime, char* return_dtime, int dtime_size)
{
  struct tm* foo;
  char buffer[128];

  if((foo = gmtime(&utime)) == NULL)
    return(DBA_ERROR);
  
  sprintf(buffer, "%04d-%02d-%02d %02d:%02d:%02d UTC", 
	  foo->tm_year + 1900, foo->tm_mon + 1, foo->tm_mday, 
	  foo->tm_hour,  foo->tm_min, foo->tm_sec);

  if(strlen(buffer) > ( dtime_size - 1 ))
    return(DBA_EBUFFER);
  else
    strcpy(return_dtime, buffer);

  return(0);

}
  

/* changes a date string to utime 
   
   returns 

   0 on underflow
   -3 on overflow 
   -2 if bad args
   -1 if the mktime call fails
   
*/

time_t dtime_to_utime(char* dtime)
{
  char *p, timebuf[9], word[4][32];
  char buffer[124];
  int  wordcnt = 0; 
  int year, month, day;
  int hour = 0, min = 0, sec = 0;
  struct tm foo;

  timebuf[0] = '\0';
  strncpy(buffer, dtime, 123);
  buffer[123]='\0';
  
  for(p = strtok(buffer, " \t"); p!= NULL; p = strtok(NULL, " \t")){
    strncpy(word[wordcnt], p, 31);
    word[wordcnt++][31] = '\0';
  }

  if(wordcnt > 4){
    fprintf(stderr, "libdba: Erorr, invalid date string\n");
    return(-2);
  }
  
  if(strchr(word[0], '-') == NULL)
    return(-2);

  
  if((p = strtok(word[0], "-")) == NULL){
    fprintf(stderr, "libdba: Erorr, invalid date string\n");
    return(-2);
  }
  year = atoi(p);

  if((p = strtok(NULL, "-")) == NULL){
    fprintf(stderr, "libdba: Erorr, invalid date string\n");
    return(-2);
  }
  month=atoi(p);

  if((p = strtok(NULL, " ")) == NULL){
    fprintf(stderr, "libdba: Erorr, invalid date string\n");
    return(-2);
  }
  day=atoi(p);

  /* underflow */
  if(year < 1970)
    return(0);
      
  /* overflow */
  if(year > 2034)
    return(-3);

  if(wordcnt > 1){
    

    if(!isdigit(word[1][0])){ /* other than hour:min:sec */
      if(strncmp(word[1], "BC", 2) == 0)
	return(0);
    }

    else if(strchr(word[1], ':') != NULL){ /* hour:min:sec */
      if((p = strtok(word[1], ":")) == NULL){
	fprintf(stderr, "libdba: Erorr, invalid date string\n");
	return(-2);
      }
      hour = atoi(p);

      if(hour > 23 || hour < 0){
	fprintf(stderr, "libdba: hour in string out of range"); 
	return(-2);
      }
      if((p = strtok(NULL, ":")) == NULL){
	fprintf(stderr, "libdba: Erorr, invalid date string\n");
	return(-2);
      }
      min = atoi(p);

      if(min > 59 || min < 0){
	fprintf(stderr, "libdba: min in string out of range"); 
	return(-2);
      }
      if((p = strtok(NULL, ":")) == NULL){
	fprintf(stderr, "libdba: Erorr, invalid date string\n");
	return(-2);
      }
      sec = atoi(p);
      if(sec > 60 || sec < 0){ /* 60 is for a leap second */
	fprintf(stderr, "libdba: sec in string out of range"); 
	return(-2);
      }
    }
  }

  foo.tm_year = year - 1900;
  foo.tm_mon = month;
  foo.tm_mday = day;
  foo.tm_hour = hour;
  foo.tm_min = min;
  foo.tm_sec = sec; 
  foo.tm_isdst = -1 ;

  return(mktime(&foo));
}


void return_value(char* val, DbA_impl* impl, void *value, int value_size)
{
 
  if(impl->value_type == 'i'){
    if(value_size < 4){
      fprintf(stderr, "libdba: value buffer too small for result\n");
      impl->db_errno = DBA_EBUFFER;
      return;
    }
    *((long *)value) = strtol(val, (char **)NULL, 10);
  }
  if(impl->value_type == 'f'){
    if(value_size < 8){
      impl->db_errno = DBA_EBUFFER;
      fprintf(stderr, "libdba: value buffer too small for result\n");
      return;
    }
    *((double*)value) = strtod(val, (char **)NULL);
  }
  if(impl->value_type == 'c'){
    if((strlen(val) + 1) > value_size){
      fprintf(stderr, "libdba: value buffer too small for result\n");
      impl->db_errno = DBA_EBUFFER;
      return;
    }
    strcpy((char*)value, val);
  }

  impl->db_errno = DBA_SUCCESS;
}







