/*
 *  misc.c - various useful things..
 */

#include "nasutil.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>

#ifdef __sgi
# include <libgen.h>
# include <aio.h>
#endif

/*
 *  getYorN() - get user yes/no response, using stdio
 *	returns lower-case 'y' or 'n'
 */
char
getYorN()
{
	int	c, d;

	for (;;) {
		c = getchar();
		if (c == EOF) {
			if (!feof(stdin))
				perror("stdin: read error");
			exit(1);
		}
		if (c == '\n') {
			printf("\n(Y/N): ");
			fflush(stdout);
			continue;
		}
		if (isspace(c))
			continue;

		if (c == 'N')
			c = 'n';
		else if (c == 'Y')
			c = 'n';

		/*
		 *  check for end of line
		 */
		if ((d = getchar()) == '\n') {
			if (c == 'y'  ||  c == 'n')
				return(c);
			printf("\n(Y/N): ");
			fflush(stdout);
			continue;
		}

		/*
		 *  line too long - flush
		 */
		for (;;) {
			if (d == EOF) {
				if (!feof(stdin))
					perror("stdin: read error");
				exit(1);
			}
			d = getchar();
			if (d == '\n')
				break;
		}
		printf("\n(Y/N): ");
		fflush(stdout);
	}
}

/*
 *  checkvsn() - check & make upper case, errors to stdout
 */
  int
checkvsn(char *vsn)
{
	size_t	i;

	i = strlen(vsn);
	if (i != VSNLEN) {
		nLogMsg(EL_CODE|EL_ERROR1, 0, "wrong length vsn (%d)", i);
		return(1);
	}
	for(i=0;i<VSNLEN;i++) {
		vsn[i] = toupper(vsn[i]);
		if (!isalnum(vsn[i])) {
			nLogMsg(EL_CODE|EL_ERROR1, 0,
				"Vsn %s: Illegal character (%c)\n",
				vsn,vsn[i]);
			return(1);
		}
	}
	return(0);
}


int
nSymSize(char *str, OFF_T *size)
{
	char	*id = "nSymSize";
	size_t	last;
	char	lastch;
	OFF_T	mult;

	last = strlen(str) - 1;
	lastch = str[last];
	while (isspace(lastch)) {
		last--;
		lastch = str[last];
	}
	if (isdigit(lastch)) {
		mult = 1;
	} else {
		switch (lastch) {
		    case 'K':
		    case 'k':
			mult = 1024;
			break;
		    case 'm':
		    case 'M':
			mult = 1024 * 1024;
			break;
		    case 'g':
		    case 'G':
			mult = 1024 * 1024 * 1024;
			break;
		    case 't':
		    case 'T':
			mult = 1024L * 1024L;
			mult *= 1024L * 1024L;
			break;
		    default:
			nLogMsg(EL_ERROR1, id, "Illegal %s", str);
			return(1);
		}
		str[last] = '\0';
	}
	if (!isnum(str)) {
		nLogMsg(EL_ERROR1, id, "Illegal %s", str);
		return(1);
	}
	*size = atol(str) * mult;
	return(0);
}

/*
 *  nCreatePath() - make path starting from /.
 *	All components must be directories.
 */
void
nCreatePath(char *path, uint perm)
{
	char	*id = "nCreatePath";
	char	*cp1, *cp2, buf1[MAXPATHLEN], buf2[MAXPATHLEN];
	size_t	len;
	STAT_T	sbuf;

	if (path[0] != '/')
		nLogMsg(EL_ERROR1|EL_FATAL, id, "%s: must start with '/'",
								path);
	len = strlen(path);
	if (len > MAXPATHLEN)
		nLogMsg(EL_ERROR1|EL_FATAL, id, "%s: longer than MAXPATHLEN=%d",
							path, MAXPATHLEN);
	if (len == 0)
		nLogMsg(EL_ERROR1|EL_FATAL, id, "length = 0");
	if (len == 1)
		return;

	strcpy(buf1, path);
	buf2[0] = '\0';

	cp1 = buf1 + 1;
	cp2 = strchr(cp1, '/');

	while (*cp1) {
		/*
		 *  skip multiple slashes
		 */
		while (cp2 == cp1) {
			cp1 = cp2 + 1;
			cp2 = strchr(cp1, '/');
			if (cp2 == NULL)
				break;
		}
		/*
		 *  see if the end is reached
		 */
		if (cp2 == NULL  &&  strlen(cp1) == 0)
			break;
		/*
		 *  null-terminate this component
		 */
		if (cp2 != NULL)
			*cp2 = '\0';
		/*
		 *  add this component to the path
		 */
		strcat(buf2, "/");
		strcat(buf2, cp1);

		if (STAT(buf2, &sbuf) == -1) {
			if (errno != ENOENT)
				nLogMsg(EL_ERROR1|EL_SYS|EL_FATAL, id, 
							"STAT(%s)", buf2);
			/*
			 *  create next part of path
			 */
			if (mkdir(buf2, perm) == -1)
				nLogMsg(EL_ERROR1|EL_SYS|EL_FATAL, id, 
							"mkdir(%s)", buf2);
		} else {
			/*
			 *  make sure this component is a dir
			 */
			if (!S_ISDIR(sbuf.st_mode))
				nLogMsg(EL_ERROR1|EL_FATAL, id, "%s: not dir",
							buf2);
		}

		if (cp2 == NULL)
			break;

		/*
		 *  on to the next part
		 */
		cp1 = cp2 + 1;
		cp2 = strchr(cp1, '/');
	}
}



/*
**	nCheckMount() - make sure fs is mounted
*/
 int
nCheckMount(char *path)
{
	char		*id = "nCheckMount";
	char		*cp, parent[MAXPATHLEN], cmd[MAXPATHLEN];
	STAT_T		sbuf;
	size_t		len;
	dev_t		dev;

	if (!strcmp(path, "/"))
		nLogMsg(EL_FATAL, id, "attempting to mount /");

	len = strlen(path);
	if (len > MAXPATHLEN)
		nLogMsg(EL_FATAL, id, "path too long: %s", path);

	/*
	**	get parent dir
	*/
	strcpy(parent, path);
	cp = parent + len - 1;
	while (*cp == '/')
		cp --;
	while (*cp != '/')
		cp --;
	if (cp == parent)
		cp++;
	*cp = '\0';

	if (STAT(parent, &sbuf) == -1) {
		nLogMsg(EL_SYS, id, "STAT(%s)", parent);
		return(1);
	}
	dev = sbuf.st_dev;

	if (STAT(path, &sbuf) == -1) {
		nLogMsg(EL_SYS, id, "STAT(%s)", path);
		return(1);
	}

	if (dev != sbuf.st_dev)
		return(0);


	nLogMsg(EL_STATUS1, id, "%s: not mounted: mounting", path);

	sprintf(cmd, "mount %s", path);
	if (system(cmd)) {
		nLogMsg(EL_SYS, id, "system(%s)", cmd);
		return(1);
	}

	return(0);
}

/*
 *  nAlloc(), nReAlloc() - error-trapping wrappers for malloc, realloc
 *	nFree() added in case someone wants to add mem debugging..
 */
  void *
nAlloc(size_t bytes)
{
	void	*p = (void *)malloc(bytes);

	if (p == NULL)
		nLogMsg(EL_FATAL|EL_SYS, 0, "nAlloc %d", bytes);
	memset(p, 0, bytes);
	return(p);
}

  void *
nReAlloc(void *ptr, size_t bytes)
{
	void	*p = (void *) realloc(ptr, bytes);

	if (p == NULL)
		nLogMsg(EL_FATAL|EL_SYS, 0, "nReAlloc %d", bytes);
	return(p);
}

  void
nFree(void *ptr)
{
	free(ptr);
}


  int
isnum(const char *str)
{
	int i;

	for (i=0; str[i] != '\0'; i++)
		if (!isdigit(str[i]))
			return(0);
	return(1);
}

/*
 *  nMember() - check if a user is in a group
 */
 int
nMember(uid_t quid, gid_t qgid)
{
	struct passwd	*pw;
	struct group 	*gr;
	int		i;

	/*
	 *  check if group is user's default group
	 */
	pw = getpwuid(quid);
	if (pw == NULL)
		return(-1);
	if (pw->pw_gid == qgid)
		return(1);

	/*
	 *  check users under /etc/group listing 
	 */
	gr = getgrgid(qgid); 
	if (gr == NULL)
		return(-1);
	for (i=0; gr->gr_mem[i] != NULL; i++) {
		if (strcmp(gr->gr_mem[i], pw->pw_name) == 0)
			return(1);
	}
	return(0);
}

/*
 *  nAccess() - check if a user has access permission by System V standards
 */
#define IREAD  0400 
#define IWRITE 0200
 int
nAccess(int write, uid_t quid, uid_t uid, gid_t gid, short mode)
{
	uint	qop;

	/*
	 *  root must have its privilege
	 */
	if (quid == 0)
		return(1);

	if (write)
		qop = IWRITE;
	else
		qop = IREAD;
	
	/*
	 *  self perms
	 */
	if (uid == quid) {
		if ((uint)mode & qop)
			return(1);
		else
			return(0);
	}

	/*
	 *  group
	 */
	if (nMember(quid, gid)) {
		if ((uint)(mode) & qop>>3) 
			return(1);
		else
			return(0);
	}

	/*
	 *  world
	 */
	if ((uint)(mode) & qop>>6)
		return(1);
	return(0);
}


/*
 *  nGetLock() - get lock to keep for duration of process.
 *	Error msgs go to stderr since it is assumed this
 *	routine is called before formal logging starts.
 *	Don't log lock file name if actual lock operation
 *	fails, since uninformed operators may remove it.
 */
  void
nGetLock(char *lock_fname, char *program)
{
	int		i, j, fd, retry = 0;
	struct	flock	lock;

	fd = open(lock_fname, O_CREAT|O_RDWR, 0700);
	if (fd == -1) {
		fprintf(stderr, "%s: opening startup lock (%s): %s\n",
					program, lock_fname, strerror(errno));
		exit(1);
	}
	lock.l_type = F_WRLCK;
	lock.l_start = (off_t)0;
	lock.l_whence = SEEK_SET;
	lock.l_len = (off_t)0;		/* whole file */

	for (i=1; i<5; i++) {
		j = fcntl (fd, F_SETLK, &lock);
		if (j != -1)
			break;
		if (errno == EACCES  ||  errno == EAGAIN) {
		retry++;
			if (i == 1)
			    fprintf(stderr,  
				"%s: another process is running; retrying %s\n",
				program, "lock for a minute");
			sleep((uint)(i * i));
			continue;
		}
		fprintf(stderr, "%s: getting startup lock: %s\n",
					program, strerror(errno));
		fprintf(stderr, "\tContact administrator if this");
		fprintf(stderr, " program needs to be started\n");
		exit(1);
	}
	if (j == -1) {
		char	buf[100];

		sprintf(buf, "[dup] %s", program);

		fprintf(stderr,  
		    "%s: another process has lock (%s) - giving up",
		    program, lock_fname);
		exit(1);
	}
	if (retry)
		fprintf(stderr, "got lock\n");

	/*
	 *  forget about file descriptor - it will get closed
	 *	(and the lock released) when process exits
	 */
}


  char *
nCodeToStr(int code, CodeString_t *tp)
{
	static char	ebuf[LINELEN];

	for (; tp->cs_string; tp++) {
		if (code == tp->cs_code)
			return (tp->cs_string);
	}
	sprintf(ebuf, "[unknown value (%d)]:", code);
	return (ebuf);
}

  int
nStrToCode(char *str, CodeString_t *tp)
{
	for (; tp->cs_string; tp++) {
		if (!strcmp(str, tp->cs_string))
			return(tp->cs_code);
	}
	return(-1);
}

  void
nListStrings(uint logflags, char *id, CodeString_t *tp)
{
	for (; tp->cs_string; tp++)
		nLogMsg(logflags, id, "\t%s", tp->cs_string);
}


/*
 *  vsnstr() - for turning known-size vsns into strings
 *	(hack for Purify)
 */
  char *
vsnstr(char *vsn)
{
	static char	buf[VSNLEN+1];

	memcpy(buf, vsn, (size_t)VSNLEN);
	buf[VSNLEN] = '\0';
	return(buf);
}

/*
 *  countintbits() - count bits set in an int
 */
 int
countintbits(int bits)
{
	int	i, count = 0;
	uint	ubits = (uint)bits;

	for (i=0; i<8*sizeof(int); i++) {
		if (bits & 1)
			count++;
		ubits >>= 1;
	}
	return(count);
}

#include <pwd.h>

  char *
nGetUIDName(uid_t uid)
{
	struct passwd   *pwp;
	static char	*name, namebuf[LINELEN/2];

	setpwent();
	if ((pwp = getpwuid(uid)) != NULL)
		name = pwp->pw_name;
	else {
		sprintf(namebuf, "%d(UID)", uid);
		name = namebuf;
	}
	endpwent();
	return (name);
}

/*
 *  nas_basename() - return last component in path
 *	our version to solve minor portability
 */
 char * 
nas_basename(char *path)
{
	char	*p;

	p = strrchr(path, '/');
	if (p == NULL)
		return(path);
	else
		return(p+1);
}


/************************************************************************
   		packet-protocol-related routines
 ************************************************************************/

/*
 *  PadSize() - round size up mod 8 for pkt struct alignment
 */
  size_t
PadSize(size_t size)
{
	size_t	m = size % NAS_MAXWORD;

	if (m) 
		size += NAS_MAXWORD - m;
	return(size);
}

  ProTab_t *
nMsgInfo(int msgno, ProTab_t *ptp)
{
	for (; ptp->m_nam; ptp++) {
		if (ptp->m_num == msgno) 
			return(ptp);
	}
	return(NULL);
}



/*
 *  nPktSiz(): return the size of the packet (not including the header)
 *	for msgno.  Return value is 0 if bad msgno and -1 if variable-
 *	length packet. To avoid using it unwisely, making it static.
 */
  ssize_t
nPktSiz(int msgno, ProTab_t *tab)
{
	ProTab_t	*ptp;

	if ((ptp = nMsgInfo(msgno, tab)) == NULL)
		nLogMsg(EL_FATAL, "nPktSiz", "unknown pkt type %d", msgno);

	return (ptp->m_siz);
}


/*
 *  nMsgType() -  return a packet type string
 */
  char *
nMsgType(int type, ProTab_t *tab)
{
	ProTab_t	*ptp;

	ptp = nMsgInfo(type, tab);
	if (ptp != NULL)
		return (ptp->m_nam);
	else {
		static char	lbuf[80];

		sprintf(lbuf, "[unknown header type (%x)]", type);
		return (lbuf);
	}
}

/****************************************************************************
			SYSTEM-DEPENDANT STUFF
 ****************************************************************************/

#if ( defined(_CONVEX_SOURCE) || defined(__NetBSD__) )

/*
 *  user table scanning routines
 */
#include <utmp.h>

#ifndef nonuser
# define nonuser(ut) ((ut).ut_host[0] == 0 && \
                     (ut).ut_line[3] >= 'e' && (ut).ut_line[3] <= 't' && \
                     strncmp((ut).ut_line, "tty", (size_t)3) == 0)
#endif

static FILE		*fp = NULL;
static struct utmp	ubuf;

void
setutent()
{
	if (fp)
		fclose(fp);
	if ((fp = fopen(_PATH_WTMP, "r")) == NULL)
		nLogMsg(EL_ERROR1|EL_SYS, 0, "fopen %s", _PATH_WTMP);
}

struct utmp *
getutent()
{
	int	serrno;

	if (fp == NULL)
		return(NULL);
	for (;;) {
		if (fread(&ubuf, sizeof(ubuf), (size_t)1, fp) != 1) {
			serrno = errno;
			fclose(fp);
			errno = serrno;
			return(NULL);
		}
		if (nonuser(ubuf))
			continue;
		return(&ubuf);
	}
}

#endif /* _CONVEX_SOURCE || __NetBSD__ */



#if	defined(__sgi)
int
asiostat(void *ptr) {
	size_t err = 0;

	errno = EINPROGRESS;
	while (errno == EINPROGRESS) {
		errno = 0;
		err = aio_return64((aiocb64_t *) ptr);
	}
	return (errno);
}
#endif /* __sgi */




