/*
 *  nas_ascii.c - conversion to/from ascii formats;
 *	used in database tables and tape labels.
 *	Errors are logged with nLogMsg(EL_ERROR1|EL_CODE, ...).
 *
 *	Naming convention for conversion routines: 
 *
 *		[Set] ascii_format number_format
 *
 *	E.g.
 *
 *		SetHexOFF()	- convert OFF_T to hex
 *		HexOFF()	- convert hex to OFF_T
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include "nasutil.h"

static char	*hex1 = NULL, *hex2 = NULL;
static char	badchar[] = "field (%.*s) has bad char (%0o) at %d";

  static void
InitHexTrans()
{
	int	i;

	/*
	 *  set up hex translation arrays
	 */
	hex1 = (char *) nAlloc((size_t)128);

	for (i='0'; i<='9'; i++)
		hex1[i] = i - '0';
	hex1['A'] = 10;
	hex1['B'] = 11;
	hex1['C'] = 12;
	hex1['D'] = 13;
	hex1['E'] = 14;
	hex1['F'] = 15;

	hex2 = (char *) nAlloc((size_t)16);
	for (i=0; i<=9; i++)
		hex2[i] = '0' + i;
	hex2[10] = 'A';
	hex2[11] = 'B';
	hex2[12] = 'C';
	hex2[13] = 'D';
	hex2[14] = 'E';
	hex2[15] = 'F';
}

#define	BINCHK	u_char

/*
 *  Hex2Bin() - convert hex string to binary
 */
 int
Hex2Bin(void *nump, size_t numsize, char *buf, size_t bufsize)
{
	char	*id = "Hex2Bin";
	int	i, j;
	void	*copy;
	BINCHK	num, val;

	assert(numsize % sizeof(num) == 0);
	if (!hex1)
		InitHexTrans();

	memset(nump, 0, numsize);
	copy = (char *)nump + (numsize - sizeof(num));
	/*
	 *  should be hex numerals
	 */
	for (i=(int)bufsize-1; i>=0;) {
		num = 0;
		for (j=0; j<sizeof(num)*2; j++) {
			int    ch = (int)buf[i];

			if (ch == ' ')
				break;

			if (!isxdigit(ch)) {
				nLogMsg(EL_ERROR1, id, badchar,
					bufsize, buf, ch, i);
				return(1);
			}

			val = (BINCHK)(hex1[ch] << (j*4));
			num |= val;

			if (i-- == 0)
				break;
		}
		memcpy(copy, &num, sizeof(num));
		if (copy == nump) {
			if (i >= 0 && buf[i] != ' ') {
				nLogMsg(EL_ERROR1, id,
					"overflow %d into %d",
					bufsize, numsize);
				return(1);
			}
			break;
		}
		copy = (void *)((char *)copy - sizeof(num));
	}

	return(0);
}

/*
 *  HexUshort() - get unsigned short from hex string
 */
 int
HexUshort(unsigned short *nump, char *buf, size_t size)
{
	char	*id = "HexUshort";
	int     i;

	if (!hex1)
		InitHexTrans();

	*nump = 0;

	/*
	 *  skip leading space
	 */
	for (i=0; i<size && buf[i] == ' '; i++);

	/*
	 *  if all spaces, number is 0
	 */
	if (i == size)
		return(0);

	/*
	 *  should be hex numerals
	 */
	for (; i<size; i++) {
		if (!isxdigit(buf[i])) {
			nLogMsg(EL_ERROR1|EL_CODE, id, badchar,
				size, buf, buf[i], i);
			return(1);
		}
		*nump *= 16;
		*nump += hex1[(int)buf[i]];
	}
	return(0);
}

/*
 *  Bin2Hex() - convert binary to hex string
 */
  int
Bin2Hex(char *buf, size_t bufsize, void *nump, size_t numsize)
{
	char	*id = "Bin2Hex";
	int	i, j, ch, mask;
	BINCHK	num;

	assert(numsize % sizeof(num) == 0);
	if (!hex1)
		InitHexTrans();

	ch = (int)bufsize;

	/*
	**	Start with least significant bits of nump
	**	and process in BINCHK sized chunks.
	*/
	for (i=(int)(numsize-sizeof(num)); i>=0; i -= sizeof(num)) {
		memcpy(&num, (char *)nump+i, sizeof(num));

		for (j=0; j<sizeof(num)*2; j++) {
			/*
		 	*  get least significant 4 bits
		 	*/
			mask = num & 0xf;

			ch--;

			if (ch > -1)
				buf[ch] = hex2[mask];
			else if (mask) {
				nLogMsg(EL_ERROR1|EL_CODE, id, 
					"bin len %d overflow reserved %d", 
					numsize, bufsize);
				return(1);
			}
			num >>= 4;
		}
	}

	return(0);
}

/*
 *  SetHexUshort() - convert unsigned short to hex string
 */
static int	shorthexsize = sizeof(short) * 2;
  int
SetHexUshort(char *buf, size_t size, unsigned short num)
{
	char	*id = "SetHexUshort";
	int	i, ch;
	uint	mask;
	uint	copy;

	if (!hex1)
		InitHexTrans();

	ch = (int)size;
	copy = num;
	for (i=0; i<shorthexsize; i++) {
		/*
		 *  get least significant 4 bits
		 */
		mask = copy & 0xf;

		ch--;

		if (ch > -1)
			buf[ch] = hex2[mask];
		else if (mask) {
			nLogMsg(EL_ERROR1|EL_CODE, id, 
						"%d: hex len > reserved %d", 
						num, size);
			return(1);
		}
		copy >>= 4;
	}

	return(0);
}

/*
 *  HexInt() - get int from hex string
 */
 int
HexInt(uint *nump, char *buf, size_t size)
{
	char	*id = "HexInt";
	int     i;

	if (!hex1)
		InitHexTrans();

	*nump = 0;

	/*
	 *  skip leading space
	 */
	for (i=0; i<size && buf[i] == ' '; i++);

	/*
	 *  if all spaces, number is 0
	 */
	if (i == size)
		return(0);

	/*
	 *  should be hex numerals
	 */
	for (; i<size; i++) {
		if (!isxdigit(buf[i])) {
			nLogMsg(EL_ERROR1|EL_CODE, id, badchar,
				size, buf, buf[i], i);
			return(1);
		}
		*nump *= 16;
		*nump += hex1[(int)buf[i]];
	}
	return(0);
}

static int	inthexsize = sizeof(int) * 2;

  int
SetHexInt(char *buf, size_t size, uint num)
{
	char	*id = "SetHexInt";
	int	i, ch;
	uint	mask;
	ulong	copy;

	if (!hex1)
		InitHexTrans();

	ch = (int)size;
	copy = num;
	for (i=0; i<inthexsize; i++) {
		/*
		 *  get least significant 4 bits
		 */
		mask = (uint)copy & 0xf;

		ch--;

		if (ch > -1) {
			buf[ch] = hex2[mask];
		} else if (mask) {
			nLogMsg(EL_ERROR1|EL_CODE, id, 
						"%d: hex len > reserved %d", 
						num, size);
			return(1);
		}
		copy >>= 4;
	}

	return(0);
}


/*
 *  HexOFF() - get OFF_T from hex string
 */
 int
HexOFF(OFF_T *nump, char *buf, size_t size)
{
	char	*id = "HexOFF";
	int     i;

	if (!hex1)
		InitHexTrans();

	*nump = 0;

	/*
	 *  skip leading space
	 */
	for (i=0; i<size && buf[i] == ' '; i++);

	/*
	 *  if all spaces, number is 0
	 */
	if (i == size)
		return(0);

	/*
	 *  should be hex numerals
	 */
	for (; i<size; i++) {
		if (!isxdigit(buf[i])) {
			nLogMsg(EL_ERROR1|EL_CODE, id, badchar,
				size, buf, buf[i], i);
			return(1);
		}
		*nump *= 16;
		*nump += hex1[(int)buf[i]];
	}
	return(0);
}

static int	offhexsize = sizeof(OFF_T) * 2;

  int
SetHexOFF(char *buf, size_t size, OFF_T num)
{
	char	*id = "SetHexOFF";
	int	i, ch, mask;
	OFF_T	copy;

	if (!hex1)
		InitHexTrans();

	ch = (int)size;
	copy = num;
	for (i=0; i<offhexsize; i++) {
		/*
		**	get least significant bits
		*/
		mask = (int)copy & 0xf;

		ch--;

		if (ch >= 0) {
			buf[ch] = hex2[mask];
		} else if (mask) {
			nLogMsg(EL_ERROR1|EL_CODE, id, 
						"0x%llx: hex len > reserved %d", 
						(long long)num, size);
			return(1);
		}
		copy >>= 4;
	}

	return(0);
}

/*
 *  IncHex() - increment an ascii hex field by 1
 */
 int
IncHex(char *buf, size_t size)
{
	char	*id = "IncHex";

	if (size == 0) {
		nLogMsg(EL_ERROR1|EL_CODE, id, "overflow");
		return(1);
	}
	size--;
	if (buf[size] == '9') {
		buf[size] = 'A';
		return(0);
	}
	if (buf[size] == 'F') {
		buf[size] = '0';
		return(IncHex(buf, size));
	}
	if (buf[size] == ' ') {
		buf[size] = '1';
		return(0);
	}
	buf[size]++;
	return(0);
}


/*
 *  DecInt() - translate decimal char -> int
 */
 int
DecInt(int *nump, char *str, size_t size)
{
	char	*id = "DecInt";
	int     i;

	/*
	 *  skip leading space
	 */
	for (i=0; i<size && str[i] == ' '; i++);

	/*
	 *  if all spaces, number is 0
	 */

	if (i == size) {
		*nump = 0;
		return(0);
	}

	/*
	 *  leading '-' ok
	 */
	if (str[i] == '-') {
		if (i > size-1) {
			nLogMsg(EL_ERROR1|EL_CODE, id, 
						"leading '-' w/ no # (%.*s)",
						size, str);
			return(1);
		}
		i++;
	}

	/*
	 *  rest should be numerals
	 */
	for (; i<size; i++) {
		if (!isdigit(str[i])) {
			nLogMsg(EL_ERROR1|EL_CODE, id, badchar,
				size, str, str[i], i);
			return(1);
		}
	}
	*nump = atoi(str);
	return(0);
}

/*
 *  SetDecInt() - translate int into decimal char field
 *	return: 0 if ok, otherwise 1
 */
 int
SetDecInt(char *str, size_t size, int num)
{
	char	*id = "SetDecInt";
	char	arr[100];
	size_t	len;

	sprintf(arr, "%d", num);
	len = strlen(arr);

	if (len > size) {
		nLogMsg(EL_CODE|EL_ERROR1, id, "%d: int len %u > reserved %u", 
						num, len, size);
		return(1);
	}
	sprintf(arr, "%*d", (int)size, num);
	memcpy(str, arr, size);
	return(0);
}

/*
 *  DecUint() - translate decimal char -> unsigned int
 */
 int
DecUint(unsigned int *nump, char *str, size_t size)
{
	char	*id = "DecInt";
	int     i;

	/*
	 *  skip leading space
	 */
	for (i=0; i<size && str[i] == ' '; i++);

	/*
	 *  if all spaces, number is 0
	 */

	if (i == size) {
		*nump = 0;
		return(0);
	}

	/*
	 *  leading '-' ok
	 */
	if (str[i] == '-') {
		if (i > size-1) {
			nLogMsg(EL_ERROR1|EL_CODE, id, 
						"leading '-' w/ no # (%.*s)",
						size, str);
			return(1);
		}
		i++;
	}

	/*
	 *  rest should be numerals
	 */
	for (; i<size; i++) {
		if (!isdigit(str[i])) {
			nLogMsg(EL_ERROR1|EL_CODE, id, badchar,
				size, str, str[i], i);
			return(1);
		}
	}
	*nump = atoi(str);
	return(0);
}

/*
 *  SetDecUint() - translate unsigned int into decimal char field
 *	return: 0 if ok, otherwise 1
 */
 int
SetDecUint(char *str, size_t size, unsigned int num)
{
	char	*id = "SetDecInt";
	char	arr[100];
	size_t	len;

	sprintf(arr, "%u", num);
	len = strlen(arr);

	if (len > size) {
		nLogMsg(EL_CODE|EL_ERROR1, id, "%u: int len %u > reserved %u", 
						num, len, size);
		return(1);
	}
	sprintf(arr, "%*u", (int)size, num);
	memcpy(str, arr, size);
	return(0);
}

/*
 *  IncDecInt() - increment by 1; decrement by 1 if already negative
 */
int
IncDecInt(char *str, size_t size)
{
	char	*id = "IncDec";
	int	n;

	if (DecInt(&n, str, size)) {
		nLogMsg(EL_ERROR1, id, "can't convert to int");
		return(1);
	}

	if (n < 0)
		n--;
	else
		n++;
	if (SetDecInt(str, size, n)) {
		nLogMsg(EL_ERROR1, id, "can't set field");
		return(1);
	}
	return(0);
}

/*
 *  DecUlong() - check & translate char -> unsigned long
 */
 int
DecUlong(unsigned long *nump, char *str, size_t size)
{
	char	*id = "DecUlong";
	int     i;

	/*
	 *  skip leading space
	 */
	for (i=0; i<size && str[i] == ' '; i++);

	/*
	 *  if all spaces, number is 0
	 */

	if (i == size) {
		*nump = 0;
		return(0);
	}

	/*
	 *  rest should be numerals
	 */
	for (; i<size; i++) {
		if (!isdigit(str[i])) {
			nLogMsg(EL_ERROR1|EL_CODE, id, badchar,
				size, str, str[i], i);
			return(1);
		}
	}
	*nump = atoi(str);
	return(0);
}

/*
 *  SetDecUlong() - translate unsigned long into char field
 *
 */
 int
SetDecUlong(char *str, size_t size, unsigned long num)
{
	char	*id = "SetDecUlong";
	char	arr[100];
	size_t	len;

	sprintf(arr, "%lu", num);
	len = strlen(arr);

	if (len > size) {
		nLogMsg(EL_CODE|EL_ERROR1, id, 
					"%lu: ulong len %u > reserved %u",
					num, len, size);
		return(1);
	}
	sprintf(arr, "%*lu", (int)size, num);
	memcpy(str, arr, size);
	return(0);
}

/*
 *  SetTime() - 64-bit version of unix time in anticipation
 *	of growth of this standard; hex for size.
 */
 void
SetTime(char *buf, time_t time)
{
	char	arr[DATELEN+1];

	if (time == -1) {
		memset(buf, ' ', (size_t)DATELEN);
	} else {
		sprintf(arr, "%*llX", DATELEN, (long long)time);
		memcpy(buf, arr, (size_t)DATELEN);
	}
}


/*
 *  GetTime() - 64-bit version of unix time in anticipation
 *      of growth of this standard
 */
 time_t
GetTime(char *buf)
{
	long long	t;

	if (buf[DATELEN-1] == ' ')
		return(-1);

	sscanf(buf, "%llX", &t);
	return((time_t)t);
}
