/*
 *  NAStore
 *
 *  nas_logmsg.c: Event Log Manager
 *
 */

#include <time.h>			/* for localtime()		*/

#ifdef AMDAHL_UTS
#include <sys/timeb.h>
#endif

#include "nasutil.h"


#include <ctype.h>
#include <syslog.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <stdarg.h>
#include <stdio.h>

#define IDLEN	40		/* length of per-program tag string */
#define LINELEN	80
#define LINC	2		/* chunks of logfile_t to expand by */

/*
 *  global vars that callers may set/read
 */
char	*nErrStr = NULL;	/* last error string when EL_CODE */

/*
 *  internally-used struct for tracking file logging by flags
 */
typedef struct {
	int	log_flags;		/* 'log level' flags that trigger */
	char	log_fname[MAXPATHLEN];
	int	log_fd;
} logfile_t;

static logfile_t	*Logs = NULL;
static int		n_logfiles = 0, log_alloc = 0;

static int		stderrflags = 0;

static int		consoles = 0;	/* EL_CONSOLES bitmask */
static int		console0_fac = -1;
static int		console1_fac = -1;
static int		console0flags = 0, console1flags = 0;

static char		Caller[IDLEN];	/* program ID for log messages	*/
static int		callerset = 0;
static int		LogLevel = (EL_ALWAYS|EL_ERROR1|EL_ERROR2|
					EL_STATUS1|EL_STATUS2);

static void 		vConWrite(const char *msg, int pri, char c);

 static int
defaultfatal()
{
	(void)fprintf(stderr, "%s fatal: exiting\n", (callerset ? Caller : ""));
	exit(1);
	return(1);	/* for compiler warning */
}

static int	(*OnFatal)() = defaultfatal;	/* called on fatal errors */

CodeString_t	SysLogArr[] = {
	{ LOG_LOCAL0,	"LOG_LOCAL0" }, 
	{ LOG_LOCAL1,	"LOG_LOCAL1" },
	{ LOG_LOCAL2,	"LOG_LOCAL2" },
	{ LOG_LOCAL3,	"LOG_LOCAL3" },
	{ LOG_LOCAL4,	"LOG_LOCAL4" },
	{ LOG_LOCAL5,	"LOG_LOCAL5" },
	{ LOG_LOCAL6,	"LOG_LOCAL6" },
	{ LOG_LOCAL7,	"LOG_LOCAL7" },
	{ 0,	NULL }
};

/*
 *  nSysLogCode() - look up LOG_LOCALx from string
 */
  int
nSysLogCode(char *str)
{
	return(nStrToCode(str, SysLogArr));
}


/*
 *  nLogInit() - set Caller, routine to call for EL_FATAL
 */
 void
nLogInit( const char *caller, int (*onfatal)() )
{
	if (caller == NULL) {
		callerset = 0;
	} else {
		(void)strncpy(Caller, caller, (size_t)(IDLEN-1));
		callerset = 1;
	}

	if (onfatal == NULL)
		OnFatal = defaultfatal;
	else
		OnFatal = onfatal;
}

/*
 *  nLogInitStderr() - set 'log level' flags that trigger stderr logging
 */
 void
nLogInitStderr(int flags)
{
	if (flags & ~EL_LEVEL) {
		(void)fprintf(stderr, 
			"%s (nLogInitStderr): non-EL_LEVEL flag(s) 0x%x\n",
			(callerset ? Caller : ""), flags);
		exit(1);
	}
	stderrflags = flags;
}

/*
 *  nLogInitFile() - open/reopen file, setting 'log level' flags 
 *	for that file
 */
 void
nLogInitFile(const char *fname, 
	     int flags, 
	     int renamelog,
	     const char *rname)
{
	char	*id = "nLogInitFile";
	int	i;
	STAT_T	sbuf;

	if (fname == NULL) {
		(void)fprintf(stderr, "%s (%s): fname NULL\n",
					(callerset ? Caller : ""), id);
		exit(1);
	}

	if (strlen(fname) >= MAXPATHLEN) {
		(void)fprintf(stderr, "%s (%s): fname too long (%s)\n",
					(callerset ? Caller : ""), id, fname);
		exit(1);
	}

	if (flags & ~EL_LEVEL) {
		(void)fprintf(stderr, "%s (%s): non-EL_LEVEL flag(s) 0x%x\n",
			(callerset ? Caller : ""), id, flags);
		exit(1);
	}

	/*
	 *  see if this re-inits an already opened logfile
	 */
	for (i=0; i<n_logfiles; i++) {
		if (!strcmp(fname, Logs[i].log_fname))
			break;
	}
	if (i == n_logfiles) {
		/*
		 *  adding a new logfile - see if a closed slot
		 *	can be reused
		 */
		for (i=0; i<n_logfiles; i++) {
			if (Logs[i].log_fd == -1) {
				/*
				 *  reuse slot
				 */
				(void)strcpy(Logs[i].log_fname, fname);
				break;
			}
		}
	}
	if (i == n_logfiles) {
		/*
		 *  neither current nor reused slot available
		 */
		if (n_logfiles == log_alloc) {
			/*
			 *  need more space
			 */
			if (!log_alloc) 
				Logs = (logfile_t *) nAlloc(LINC *
							sizeof(logfile_t));
			else
				Logs = (logfile_t *) nReAlloc(Logs, 
							(LINC + log_alloc)  * 
							sizeof(logfile_t));
			log_alloc += LINC;
		}
		/*
		 *  give new slot
		 */
		n_logfiles++;
		(void)strcpy(Logs[i].log_fname, fname);
	} else {
		/*
		 *  resetting existing log
		 */
		if (Logs[i].log_fd != -1)
			(void)close(Logs[i].log_fd);
	}

	/*
	 *  install new flags & note that log isn't open
	 */
	Logs[i].log_flags = flags;
	Logs[i].log_fd = -1;
	if (flags == 0)
		return;

	/*
	 *  some flags are set, so log must be opened
	 *	move old one aside if necc
	 */
	if (renamelog  &&  STAT(fname, &sbuf) != -1) {
		time_t		timeval;
		char		dir[MAXPATHLEN], newlogname[MAXPATHLEN],
				*logpath;
		struct tm	*tp;

		/*
		 *  log exists, so rename
		 */
		(void)time(&timeval);
		tp = bsd_localtime(&timeval);

		/*
		 *  see if the logfile has a path component;
		 *	if so, it must be prepended to newlogname
		 */
		if ((logpath = strrchr(fname, '/')) != 0) {
			logpath++;	/* point one past the '/' */
			(void)strncpy(dir, fname, (size_t)(logpath - fname));
			dir[(int)(logpath - fname)] = '\0';
		} else
			dir[0] = '\0';
		(void)sprintf(newlogname, "%s%slog%02d-%02d.%02d%02d", 
			    dir, (rname ? rname : ""),
			    tp->tm_mon+1, tp->tm_mday, tp->tm_hour, tp->tm_min);
		if (link(fname, newlogname) == -1) {
			if (errno != EEXIST)
				(void)fprintf(stderr, 
					"link(%s, %s) failed: %s\n",
				 	fname, newlogname, strerror(errno));
		} else
			(void)unlink(fname);

	}

	/*
	 *  open log file
	 */
	if ((Logs[i].log_fd =
		  	   open(fname, O_WRONLY|O_APPEND|O_CREAT, 0644)) < 0) {
		(void)fprintf(stderr, 
				"%s: cannot open/create log file (%s): %s\n",
			  	(callerset ? Caller : "?"), fname, 
				strerror(errno));
		if (OnFatal)
			(*OnFatal)();
	}
}

/*
 *  nLogCloseFile() - close a given file
 */
 void
nLogCloseFile( const char *fname )
{
	int	i;

	for (i=0; i<n_logfiles; i++) {
		if (!strcmp(fname, Logs[i].log_fname))
			break;
	}
	if (i == n_logfiles)
		return;
	if (Logs[i].log_fd == -1)
		return;

	(void)close(Logs[i].log_fd);
	Logs[i].log_fd = -1;
}

/*
 *  nLogInitConsole() - set up console logging.
 * 	syslog_fac is e.g. LOG_LOCAL0 which points to 
 *	/etc/syslog.conf local0. Syslog stuff is described 
 *	in syslog(3) and syslogd(1M) man pages
 */
 void
nLogInitConsole(int syslog_fac, int console, int flags)
{
	char	*id = "nLogInitConsole";

	if (flags & ~EL_LEVEL) {
		(void)fprintf(stderr, "%s (%s): non-EL_LEVEL flag(s) 0x%x\n", 
				(callerset ? Caller : ""), id, flags);
		exit(1);
	}
	if (console & ~EL_CONSOLES) {
		(void)fprintf(stderr, "%s (%s): non-EL_CONSOLE flag(s) 0x%x\n",
				(callerset ? Caller : ""), id, flags);
		exit(1);
	}
	if (countintbits(console & EL_CONSOLES) != 1) {
		(void)fprintf(stderr, 
			"%s (%s): exactly 1 EL_CONSOLE flag req'd 0x%x\n",
			(callerset ? Caller : ""), id, flags);
		exit(1);
	}
	if (console & EL_CONSOLE0) {
		if (console0_fac != syslog_fac) {
			console0_fac = syslog_fac;
			openlog("", LOG_NDELAY | LOG_CONS, console0_fac);
		}
		console0flags = flags;
		consoles |= EL_CONSOLE0;
	} else {
		if (console1_fac != syslog_fac) {
			console1_fac = syslog_fac;
			openlog("", LOG_NDELAY | LOG_CONS, console1_fac);
		}
		console1flags = flags;
		consoles |= EL_CONSOLE1;
	}
}

/*
 *  nLogSetLevel() - set all bits < level req'd
 */
 void
nLogSetLevel(int level)
{
	uint	lll;

	if (level & ~EL_LEVEL) {
		(void)fprintf(stderr, 
			"%s (nLogSetLogLevel): non-EL_LEVEL flag(s) 0x%x\n",
			(callerset ? Caller : ""), level);
		exit(1);
	}
	if (level == 0) {
		LogLevel = EL_ALWAYS;
		return;
	}
	LogLevel = level;
	lll = level;
	while (lll >>= 1)
		LogLevel |= lll;
}

 int 
nLogLevel()
{
	return(LogLevel);
}

 void
nClearError()
{
	nErrStr = NULL;
}


/*
 *  nLogMsg() - write a message to log file(s) &/or stderr &/or syslog
 */
int	fatalflag = 0;

void
nLogMsg(unsigned int flag, char *id, char *fmt, ...) 
{
	va_list		args;
	char		buf[LINELEN * 20],
			tbuf[LINELEN * 20],
			*errstrptr = "";		/* for nErrStr */
	static int	broken = 0;	/* if true don't attempt to write */
	int		i, level, action, serrno, wrotecon0 = 0,
			fatalerror = 0;
	size_t		bcnt;
	ssize_t		wcnt;

	if (flag & EL_FATAL  &&  fatalflag)
		exit(1);

	/*
	 *  save errno to restore after logging
	 */
	serrno = errno;

	/*
	 *  if logging is broken, still handle EL_FATAL
	 */
	if (broken)
		goto DONE;

	/*
	 *  make mask of level bits that satisfy LogLevel
	 */
 	level = flag & LogLevel;

	/*
	 *  add action-based loglevels
	 */
	action = flag & EL_ACTION;
	if (action & EL_SYS  &&  !(level & (EL_ERROR1|EL_ERROR2)))
		level |= EL_ERROR1;
	if (action & EL_FATAL)
		level |= EL_ALWAYS;

	/*
	 *  see if msg setup really required
	 */
	if ( !(action & (EL_CODE|EL_TOSTDERR))  &&  
	     !(action & EL_CONSOLES & consoles) ) {
		/*
		 *  no 'action' to perform - see if level cutoff met
		 */
		if (level == 0)
			goto DONE;

		/*
		 *  see if stderr or some file would get msg
		 */
		if (!(level & stderrflags)) {
			for (i=0; i<n_logfiles; i++)
				if (Logs[i].log_flags & level)
					break;
			if (i == n_logfiles)
				goto DONE;
		}
	}

	/*
	 *  set up msg
	 */
	buf[0] = '\0';
	if (!(action & EL_NOTIME)) {
		time_t		secs;
		struct tm	*tp;

		if (time(&secs) < 0)
			secs = 0;
		tp = bsd_localtime(&secs);
#ifdef DEBUG
		/* give day hr:min:sec in debug mode */
		(void)sprintf(buf, "%d %d:%02d:%02d", tp->tm_mday,
		  tp->tm_hour, tp->tm_min, tp->tm_sec);
#else	/* DEBUG */
		(void)sprintf(buf, "%02d/%02d/%04d %02d:%02d:%02d", 
			(tp->tm_mon + 1), tp->tm_mday, (tp->tm_year + 1900),
			tp->tm_hour, tp->tm_min, tp->tm_sec);
#endif	/* DEBUG */
		if (callerset) 
			(void)strcat(buf, ": ");
	}
	
	if (callerset) 
		(void)strcat(buf, Caller);
	if (id != NULL) {
		(void)strcat(buf, ": ");
		(void)strcat(buf, id);
	}

	if (action & EL_FATAL) {
		if (level & (EL_ERROR1|EL_ERROR2))
			(void)strcat(buf, ": *Error (FATAL)");
		else
			(void)strcat(buf, ": (FATAL)");
	} else {
		if (level & (EL_ERROR1|EL_ERROR2))
			(void)strcat(buf, ": *Error");
	}

	(void)strcat(buf, ": ");
	if (action & (EL_CODE|EL_CODEONLY))	/* start nErrStr from here */
		errstrptr = buf + strlen(buf);

	va_start(args, fmt);
	(void)vsprintf(tbuf, fmt, args);
	va_end(args);

	(void)strcat(buf, tbuf);
		
	if (action & EL_SYS_) {
		int	erno = (flag & ERRNOMASK) >> ERRNOSHFT;

		(void)strcat(buf, ": ");
		(void)strcat(buf, strerror(erno));
	}

	(void)strcat(buf, "\n");			/* terminate message */
	buf[sizeof (buf) - 1] = '\0';			/* safety null */

	/*
	 *  msg built: write to stderr if req
	 */
	if (action & EL_TOSTDERR  ||  level & stderrflags) {
		register char	*msgptr = NULL;

		/*
		 *  don't print time info to stderr; use strchr to skip it
		 */
		if (!(action & EL_NOTIME) &&
				(msgptr = strchr(buf, ':')) != NULL) {
			if ((msgptr = strchr(msgptr+1, ':')) != NULL) {
				if ((msgptr = strchr(msgptr+1, ':')) != NULL)
					msgptr += 2;
			}
		}
		if (msgptr)
			(void)fputs(msgptr, stderr);
		else
			(void)fputs(buf, stderr);
		(void)fflush(stderr);
	}

	/*
	 *  write to any open logfiles that satisfy level flags
	 */
	if (!(action & EL_CODEONLY)) {
		bcnt = strlen(buf);
		for (i=0; i<n_logfiles; i++) {
			if (Logs[i].log_flags & level && Logs[i].log_fd != -1) {
				wcnt = write(Logs[i].log_fd, buf, bcnt);
				if (wcnt != bcnt) {
		    			broken = 1;
		    			(void)sprintf(tbuf, 
						"nLogMsg (%s): %s %s: %s\n",
		    				Caller, 
						"can't write to event log",
						Logs[i].log_fname, 
						strerror(errno));
		    			(void)fputs(tbuf, stderr);
		    			(void)fprintf(stderr, 
						"msg was >>>> %s", buf);
		    			fatalerror = 1;
				}
			}
		}
	}

	/*
	 *  perform writes to console(s) if requested
	 */
	if ( (action & EL_CONSOLE0 & consoles)  &&  (level & console0flags) ) {
		vConWrite(buf, (console0_fac | LOG_NOTICE), ' ');
		wrotecon0 = 1;
	}
	if ( (action & EL_CONSOLE1 & consoles)  &&  (level & console1flags)) {
		/*
		 *  The syslogd has a bug in that if the same message is
		 *  sent to a different facility, the message is written 
		 *  to the first facility twice. Therefore, if the message was
		 *  sent to console0, make the message different so syslogd
		 *  can write it to the second facility
		 */
		if (wrotecon0)
			vConWrite(buf, console1_fac | LOG_NOTICE, '^');
		else
			vConWrite(buf, console1_fac | LOG_NOTICE, ' ');
	}

	/*
	 *  copy error info into global nErrStr if EL_CODE
	 */
	if (action & (EL_CODE|EL_CODEONLY)) {
		static char	esbuf[LINELEN*2];	/* for ErrStr */
		char		*nlp;
		(void)strncpy(esbuf, errstrptr, sizeof (esbuf));

		esbuf[sizeof (esbuf) - 1] = '\0';
		while ((nlp = strrchr(esbuf, '\n')) != 0)
			*nlp = ' ';

		nErrStr = esbuf;
	}

DONE:

	if (flag & EL_FATAL  ||  fatalerror) {
		fatalflag = 1;
		if (OnFatal)
			(*OnFatal)(id, errstrptr);
		else
			exit(1);
	}
	errno = serrno;
}



#include	<utmp.h>
#include 	<signal.h>

/*
 *  vConWrite() - broadcast a message to the syslog
 */
 static void
vConWrite(const char *msg, int pri, char c)
{
	register char   *msgptr = NULL;

	/*
	 *  don't print time info to syslog: use strchr to skip it
	 */
	if ((msgptr  = strchr(msg, ':')) != 0) {
		if ((msgptr = strchr(msgptr+1, ':')) != 0) {
			if ((msgptr = strchr(msgptr+1, ':')) != 0)
				msgptr += 2;
		}
	}
	if (msgptr)
		syslog(pri,"%c%s", c, msgptr);

	else
		syslog(pri,"%c%s", c, msg);
}


