
/*
 *  nas_net.c: NAStore server net comm stuff, using sockets
 */

#include "nasutil.h"

#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <errno.h>
#include <string.h>

extern int	errno;
extern int	h_errno;

/*
 *  local version of hstrerror() since some platforms lack it
 *
 *	error codes from /usr/include/netdb.h
 */
CodeString_t nHostErrs[] = {
	{ HOST_NOT_FOUND,	"Authoritative Answer Host not found" },
	{ TRY_AGAIN,	"Non-Authoritive Host not found, or SERVERFAIL" },
	{ NO_RECOVERY,	"Non recoverable errors, FORMERR, REFUSED, NOTIMP" },
	{ NO_DATA,	"Valid name, no data record of requested type" },
	{ 0,		"no error" }
};

  char *
nHostErr()
{
#ifdef NETDB_INTERNAL	/* e.g. linux */
	if (h_errno == NETDB_INTERNAL)
		return(strerror(errno));
#endif
	return(nCodeToStr(h_errno, nHostErrs));
}

/*
 *  ipc family names
 */
CodeString_t nIPCNames[] = {
	{ AF_UNSPEC,	"unspecified" },
	{ AF_UNIX,	"AF_UNIX (local to host (pipes, portals))" },
	{ AF_INET,	"AF_INET (internetwork: UDP, TCP, etc.)" },
#ifdef AF_IMPLINK
	{ AF_IMPLINK,	"AF_IMPLINK (arpanet imp addresses)" },
#endif
#ifdef AF_PUP
	{ AF_PUP,		"AF_PUP (pup protocols: e.g. BSP)" },
#endif
#ifdef AF_CHAOS
	{ AF_CHAOS,	"AF_CHAOS (mit CHAOS protocols)" },
#endif
#ifdef AF_NS
	{ AF_NS,		"AF_NS (XEROX NS protocols)" },
#endif
#ifdef AF_NBS
	{ AF_NBS,		"AF_NBS (nbs protocols)" },
#endif
#ifdef AF_ECMA
	{ AF_ECMA,	"AF_ECMA (european computer manufacturers)" },
#endif
#ifdef AF_DATAKIT
	{ AF_DATAKIT,	"AF_DATAKIT (datakit protocols)" },
#endif
#ifdef AF_CCITT
	{ AF_CCITT,	"AF_CCITT (CCITT protocols, X.25 etc)" },
#endif
#ifdef AF_SNA
	{ AF_SNA,		"AF_SNA (IBM SNA)" },
#endif
#ifdef AF_DECnet
	{ AF_DECnet,	"AF_DECnet" },
#endif
#ifdef AF_DLI
	{ AF_DLI,		"AF_DLI (Direct data link interface)" },
#endif
#ifdef AF_LAT
	{ AF_LAT,		"AF_LAT" },
#endif
#ifdef AF_HYLINK
	{ AF_HYLINK,	"AF_HYLINK (NSC Hyperchannel)" },
#endif
#ifdef AF_APPLETALK
	{ AF_APPLETALK,	"AF_APPLETALK (Apple Talk)" },
#endif
#ifdef AF_NIT
	{ AF_NIT,		"AF_NIT (Network Interface Tap)" },
#endif
#ifdef AF_802
	{ AF_802,		"AF_802 (IEEE 802.2, also ISO 8802)" },
#endif
#ifdef AF_OSI
	{ AF_OSI,		"AF_OSI (umbrella for all families used)" },
#endif
#ifdef AF_X25
	{ AF_X25,		"AF_X25 (CCITT X.25 in particular)" },
#endif
#ifdef AF_OSINET
	{ AF_OSINET,	"AF_OSINET (AFI = 47, IDI = 4)" },
#endif
#ifdef AF_GOSIP
	{ AF_GOSIP,	"AF_GOSIP (U.S. Government OSI)" },
#endif
#ifdef AF_IPX
	{ AF_IPX,		"AF_IPX (Novell Internet Protocol)" },
#endif
#ifdef AF_ROUTE
	{ AF_ROUTE,	"AF_ROUTE (Internal Routing Protocol)" },
#endif
#ifdef AF_LINK
	{ AF_LINK,	"AF_LINK (Link-layer interface)" },
#endif
	{ 0,		"" }
};

  char *
nIPCFamily(int af_code)
{
	return(nCodeToStr(af_code, nIPCNames));
}


  struct hostent *
nGetHostInfo(char *hostname)
{
	char			*id = "nGetHostInfo";
	char 			myhost[MAXHOSTNAMELEN];
	struct hostent 		*htp;

	for (;;) {
		if (hostname == NULL) {
			if (gethostname(myhost, (size_t)MAXHOSTNAMELEN) == -1) {
				nLogMsg(EL_ERROR2|EL_SYS, id, "gethostname");
				sleep(20);
				continue;
			}
			hostname = myhost;
		}
		htp = gethostbyname(hostname);
		if (htp == NULL) {
			if (errno != TRY_AGAIN) {
#ifdef __i386__
				nLogMsg(EL_ERROR2|EL_SYS, id, 
						"gethostbyname (%s)", hostname);
#else
				nLogMsg(EL_ERROR2, id, 
						"gethostbyname (%s): %s",
						hostname, nHostErr());
#endif
				return(NULL);
			}
			sleep(20);
			continue;
		}
		break;
	}
	nLogMsg(EL_DEBUG1, id, "host info ok (%s: %s)", hostname, htp->h_name);
	return(htp);
}


/*
 *  nStartListening() - Open a 'well-known' socket to accept connections
 */
 int
nStartListening(u_long port)
{
	char			*id = "nStartListening";
	struct sockaddr_in 	ip_addr;
	int 			listen_sd, on;
	uint			i;
	socklen_t		socksize = sizeof( struct sockaddr_in );


	/*
	 *  create master socket
	 */
	listen_sd = socket(AF_INET, SOCK_STREAM, 0);
	if (listen_sd == -1)
		nLogMsg(EL_FATAL|EL_SYS|EL_TOSTDERR, id, "socket() failed");
	nLogMsg(EL_DEBUG1, id, "listen_sd %d", listen_sd);

	/*
	 *  set socket to recycle quickly .. SO_LINGER too?
	 */
	on = 1;
	if (setsockopt(listen_sd, SOL_SOCKET, SO_REUSEADDR, 
					&on, (socklen_t)sizeof(int)) == -1) {
		nLogMsg(EL_FATAL|EL_SYS|EL_TOSTDERR, id, "setsockopt() failed");
	}

	ip_addr.sin_addr.s_addr = htonl( INADDR_ANY );
	ip_addr.sin_family = AF_INET;
	ip_addr.sin_port = htons((in_port_t)port);

	for (i=1; i<16; i++) {
	    if ((bind(listen_sd, (struct sockaddr *)&ip_addr, socksize)) ==-1) {
		if (errno == EADDRINUSE) {
		    nLogMsg(EL_TOSTDERR|EL_STATUS2, id,
			"Main port address already in use - retrying");
		    sleep(5*i);
		    continue;
		}
		nLogMsg(EL_FATAL|EL_SYS|EL_TOSTDERR, id, "bind() failed");
	    }
	    break;
	}
	if (i == 16)
		nLogMsg(EL_FATAL|EL_SYS|EL_TOSTDERR, id, 
					"giving up - wait & retry later");
	if (i != 1)
		nLogMsg(EL_TOSTDERR|EL_STATUS2, id, "Main port address OK");

	if (listen(listen_sd, 5) == -1)
		nLogMsg(EL_FATAL|EL_SYS|EL_TOSTDERR, id, "listen() failed");
	return(listen_sd);
}

/*
 *  nStopConn() - stop accepting new connections
 */
  void
nStopConn(int sd)
{
	char	*id = "nStopConn";

	if (close(sd) == -1)
		nLogMsg(EL_SYS|EL_ERROR1, id, "closing sock");
}

  int
nCloseConn(int sd)
{
	char	*id = "nCloseConn";

	if (shutdown(sd, 2) == -1) 
		nLogMsg(EL_SYS|EL_CONSOLE0, id, "socket shutdown");
	if (close(sd) == -1) {
		nLogMsg(EL_SYS|EL_CONSOLE0, id, "socket close");
		return(1);
	}
	return(0);
}


