/*
 *  nas_pwd_grp.c - load passwd & group files for faster searching
 */

#include "nasutil.h"
#include <pwd.h>
#include <grp.h>
#include <string.h>


/*
 * HASHMASK must be a power of 2 less 1. This is to speed the hashing
 * function by reducing it to a masking operation.
 */

#define HASHMASK	0xfff
#define NAMESIZE	14 

typedef struct pwe {
	uid_t		pw_uid;
	char		pw_name[NAMESIZE];
	struct pwe 	*pw_link;
} passwd_ent;

typedef struct gpe {
	gid_t		gr_gid;
	char		gr_name[NAMESIZE];
	struct gpe	*gr_link;
} group_ent;


static passwd_ent	*pwarray[HASHMASK + 1];
static group_ent	*grarray[HASHMASK + 1];


/*
 *  nas_readpwd() - read password file info into hash table in memory
 */
static int	pwd_loaded = 0;
void
nas_readpwd()
{
	struct passwd	*pwent, *getpwent();

	while ((pwent = getpwent()) != (struct passwd *)NULL) {
		int hashbucket = pwent->pw_uid & HASHMASK;
		passwd_ent *new =
			(passwd_ent *)nAlloc(sizeof(passwd_ent));
	
		new->pw_uid = pwent->pw_uid;
		strcpy(new->pw_name, pwent->pw_name);
		new->pw_link = pwarray[hashbucket];
		pwarray[hashbucket] = new;
	}
	endpwent();
	pwd_loaded = 1;
}

char   *
nas_uidname(uid_t uid)
{
	passwd_ent		*next;
	int			hashbucket = uid & HASHMASK;
	static passwd_ent	*last = NULL; /* remember last one */

	if (!pwd_loaded)
		nLogMsg(EL_FATAL|EL_TOSTDERR|EL_ERROR1, "nas_uidname",
					"init 1st with nas_readpwd");

	if (last  &&  last->pw_uid == uid)
		return (last->pw_name);

	next = pwarray[hashbucket];
	while (next != (passwd_ent *)NULL && next->pw_uid != uid) {
		next = next->pw_link;
	}
	if (next == (passwd_ent *)NULL)
		return(NULL);

	last = next;
	return (next->pw_name);
}

static int	grp_loaded = 0;
void
nas_readgrp()
{
	struct group	*grent, *getgrent();

	while ((grent = getgrent()) != (struct group *)NULL) {
		int hashbucket = grent->gr_gid & HASHMASK;
		group_ent *new = (group_ent *) nAlloc(sizeof(group_ent));
		
		new->gr_gid = grent->gr_gid;
		strcpy(new->gr_name, grent->gr_name);
		new->gr_link = grarray[hashbucket];
		grarray[hashbucket] = new;
	}
	endgrent();
	grp_loaded = 1;
}

char   *
nas_gidname(gid_t gid)
{
	int			hashbucket = gid & HASHMASK;
	group_ent		*next;
	static group_ent	*last = NULL; /* remember last one */

	if (!grp_loaded)
		nLogMsg(EL_FATAL|EL_TOSTDERR|EL_ERROR1, "nas_gidname",
					"init 1st with nas_readgrp");

	if (last  &&  last->gr_gid == gid)
		return (last->gr_name);

	for (next=grarray[hashbucket]; next; next = next->gr_link) {
		if (next->gr_gid != gid)
			break;
	}
	if (next == (group_ent *)NULL)
		return(NULL);

	last = next;
	return (next->gr_name);
}

