/*
 **  This program provides the RASH archive "File Selector" function.
 **  Written by Tom Proett, NASA Ames Research Center.
 **  This program is public domain software.
 **
 **  The usage is as follows:
 **  archive
 **	[-A resi]
 **	[-a]
 **     [-d xx]
 **	[-c copies]
 **	[-s megabytes]
 **	inputfile
 **
 **  -a After a file is archived, release the blocks from the file
 **	following the default truncate point.
 **     Default: leave the data blocks around.
 **
 **  -A Release the blocks after the resi offset.
 **
 **  -c	Specifies the maximum number of copies to be run at one time.
 **	Default: 1 (the parent process will process the archives).
 **
 **  -s Specifies the size in megabytes to allocate for a chunk of
 **	memory to hold filenames to process.
 **	Default: 20.
 **
 **  -d Turn debugging level to the value xx
 **
 **  inputfile is a file containing full path names of all files to be
 **  archived.  This list is generally created by arcbuild.
 */

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/param.h>
#include <miscfs/dmfs/rash.h>

#include "nas_logmsg.h"
#include "vv_cl.h"
#include "dblib.h"
#include "nasutil.h"
#include "rashcom.h"
#include "dmfsd.h"


int		maxkids = 1;		/* maximum number of procs */
int		aflg;
int     	procs;

char    	mypid[20];
char		*com;

extern OFF_T	leave_resi;

typedef	struct	{
	char	name[PATH_MAX+1];
	uid_t	uid;
	gid_t	gid;
	off_t	size;
} arc_ent_t;

arc_ent_t	*arc_array;
int		arc_max;
time_t  	list_ctime;

/*
 *  die_archive
 *  clean up and exit
 */
static int
die_archive(void)
{
	int  errflg;

	while (procs > 0)		/* wait for any kids to die */
		waitproc(&errflg);

	exit(1);
	return 0;	/* not reached */
}

/*
**	Shutdown for parent archive process.
*/
static void
bomb_archive(int signum)
{
	char		*id = "bomb_archive    ";

	nLogMsg(EL_ERROR1, "bomb_archive", "caught signal %d", signum);

	(void)sig_ign(SIGHUP);
	(void)sig_ign(SIGINT);
	(void)sig_ign(SIGTERM);

	/*
	 * SIGHUP and SIGINT are usually propagated to the process group from
	 * the shell.  If we do it here too, the kids could get double
	 * signals and screw up.
	 */
	if (signum == SIGTERM) {	/* probably a kill */
		if (kill(0, signum) == -1)	/* send sig to all kids */
			nLogMsg(EL_ERROR1 | EL_SYS, id, "Kill failed ");
	}
	die_archive();
}

int
archive_list(int num)
{
	static	char	id[] = "archive_list";
	int		rc, i;
	arc_ent_t	*aP;
	struct	stat	sbuf;

	nLogMsg(EL_STATUS1, id, "entries %d", num);

	rc = 0;
	for (i=0, aP=arc_array; i<num; i++, aP++) {
		if (lstat(aP->name, &sbuf) < 0) {
			nLogMsg(EL_ERROR1|EL_SYS, id, "lstat %s", aP->name);
			rc++;
			continue;
		}

		/* 
	 	**  ensure same owner
	 	*/
		if (aP->uid != sbuf.st_uid) {
			nLogMsg(EL_STATUS2, id,
				"%s uid match failed %d vs %d",
				aP->name, (int)aP->uid, (int)sbuf.st_uid);
			continue;
		}

		/* 
	 	**  ensure same group
	 	*/
		if (aP->gid != sbuf.st_gid) {
			nLogMsg(EL_STATUS2, id,
				"%s gid match failed %d vs %d",
				aP->name, (int)aP->gid, (int)sbuf.st_gid);
			continue;
		}

		/*  
	 	**  if size don't jive skip it  
	 	*/
		if (aP->size != sbuf.st_size) { 
			nLogMsg(EL_STATUS2, id,
				"%s size match failed %lld vs %lld",
				aP->name, aP->size, sbuf.st_size);
			continue;
		}
	
		/*
	 	**  if touched after list was generated then skip it  
	 	*/
		if (sbuf.st_ctime > list_ctime){
			nLogMsg(EL_STATUS2, id, 
				"%s touched after list was generated",
				aP->name);
			continue;
		}

		if (rashvv(aP->name, &sbuf) != 0) {
			rc++;
			continue;
		}
	}
	return rc;
}

int
archive_kid(void *ptr)
{
	static	char	id[] = "archive_kid";
	int		errs, *num;

	(void)signal(SIGHUP, intrpt);
	(void)signal(SIGINT, intrpt);
	(void)signal(SIGTERM, intrpt);

	/*
	 *  setup logging
	 */
	sprintf(mypid, "%s(%05d)", id, getpid());
	nLogInit(mypid, fatal);

	num = (int *)ptr;
	errs = archive_list(*num);
	if (errs > 0)
		nLogMsg(EL_STATUS1, id, "%d errors", errs);

	exit(0);
}

/*
**	Deal with a chunk of files by either forking or doing the
**	work itself depending on whether maxkids allows having children.
*/
int
process_arc(int num)
{
	static	char	id[] = "process_arc";
	int		errflg;
	int		pid;

	if (maxkids > 1) {
		while (procs >= maxkids) {	/* too many procs running */
			pid = waitproc(&errflg);
			nLogMsg(EL_STATUS1, id, "child %d status %d",
				pid, errflg);
		}
		pid = newproc(archive_kid, (void *)&num);
		nLogMsg(EL_STATUS1, id, "start child %d", pid);
		return 0;
	}
	else
		return(archive_list(num));

}

int
main(int argc, char *argv[])
{
	static	char	id[] = "main_archive";
	struct	stat 	statbuf;
	extern	char	*optarg;
	extern	int 	optind, opterr;
	FILE    	*fin;
	int		pid;
	int		errflg = 0;
	int		megs = 20;
	size_t		chunk;
	off_t		filesize;
	int		lineno;
	time_t		currtime;
	int		uid;
	int		gid;
	int		c, i;
	char		*chr;
	char		*infile, filename[PATH_MAX];
	char		parms[512];
	void		(*sigfunc)(int sig);

	time(&currtime);
	com 	= nas_basename(argv[0]);
	aflg 	= 0;

	while ((c = getopt(argc, argv, "aA:dc:s:")) != EOF) {
		switch (c) {

		case 'A':
			leave_resi = strtoq(optarg, &chr, 0);
			if (chr == optarg || *chr != '\0') {
				fprintf(stderr,
					"garbled resi value %s\n", optarg);
				errflg = 1;
			}
			else if (leave_resi < 0 || MAX_RESI < leave_resi) {
				fprintf(stderr,
					"resident size %s out of range\n",
					optarg);
				errflg = 1;
			}

		case 'a':		/* release blocks after archive */
			aflg++;
			break;

		case 'd':		/* Raise the debug level */
			nLogSetLevel(EL_LEVEL);
			break;

		case 'c':		/* copies */
			maxkids = atoi(optarg);
			if (maxkids < 1) {
				fprintf(stderr, "-c: %s must be > 0\n",
					optarg);
				errflg = 1;
			}
			break;

		case 's':		/* size */
			megs = atoi(optarg);
			if (megs < 1) {
				fprintf(stderr, "-s: %s must be > 0\n",
					optarg);
				errflg = 1;
			}
			break;

		case '?':
			errflg = 1;
			break;
		}
	}
	if (optind != (argc-1))		/* Any more options */
		errflg = 1;

	if (errflg) {
		fprintf(stderr, "usage: %s %s\n", com,
		    "[-a][-A resisize][-d][-c copies][-s memsize] inputfile");
		exit(1);
	}

	/*
	 * setup log file
	 */
	sprintf(mypid, "archv(%05d)", getpid());
	nLogInit(mypid, (maxkids > 1) ? die_archive : fatal);
	nLogInitFile(ERRORARC, EL_LEVEL, 0, NULL);
	nLogSetLevel(EL_STATUS2);

	for(i=0; i < argc; i++){
		strcat(parms, argv[i]);
		strcat(parms, " ");
	}
	nLogMsg(EL_DEBUG1, id,  "Entered: `%s' ", parms);

	/* try not to screw up if run from a terminal */
	if (!isatty(0)) {
		if (setpgid(0, 0) == -1) {
			nLogMsg(EL_ERROR1|EL_SYS, id, "setpgid");
			exit(1);
		}
	}

	sigfunc = (maxkids > 1) ? bomb_archive : intrpt;
	(void)signal(SIGHUP, sigfunc);
	(void)signal(SIGINT, sigfunc);
	(void)signal(SIGTERM, sigfunc);

	/*
	 * Serialize the access to the input file in case that there is  
	 * an outstanding arcbuild/narcbuild that is in the process of  
	 * updating it
	 */

	infile = argv[optind];
	if (strcmp("-", infile) != 0) {
		if ((fin = fopen(infile, "r")) == NULL) {
			nLogMsg(EL_ERROR1|EL_SYS, NULL,
					"unable to open file \"%s\"", infile);
			exit(1);
		}
	}
	else
		fin = stdin;

	flock(fileno(fin), LOCK_EX);
	if (fstat(fileno(fin), &statbuf) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "fstat %s", infile);
		exit(1);
	}
	list_ctime = statbuf.st_ctime;

	/*
	 *  load passwd/group files for quick access
	 */
	nas_readpwd();
	nas_readgrp();

	procs	= 0;
	chunk = megs*1024*1024;		/* convert to bytes */
	arc_max = chunk / sizeof(arc_ent_t);
	chunk = arc_max * sizeof(arc_ent_t);
	arc_array = (arc_ent_t *)malloc(chunk);
	if (arc_array == NULL) {
		nLogMsg(EL_ERROR1|EL_SYS, id,
			"array size (%d Mb) too large", megs);
		exit(1);
	}
	nLogMsg(EL_STATUS1, id, "allocated %d entries", arc_max);

	for (lineno = 1;; lineno++) {
		/*
	 	**  read the file from the list
	 	*/
		for (i=0; i<arc_max;) {

			c = file_line(fin, &uid, &gid, &filesize, filename);
			if (c) {
				if (c == -1) {
					nLogMsg(EL_ERROR1, id, 
						"garbled input line: %d",
						lineno);
				}
				break;
			}
			nLogMsg(EL_DEBUG4, id,
				"bad: %03d uid: %4d size: %lld %s name:%s",
				uid, gid, filesize, filename);

			strcpy(arc_array[i].name, filename);
			arc_array[i].uid = uid;
			arc_array[i].gid = gid;
			arc_array[i++].size = filesize;
		}
		if (process_arc(i))
			break;

		if (c)
			break;
	}

	nLogMsg(EL_DEBUG1, id, "End of the list encountered");
	/*
	 *  wait for any children
	 */
	while (procs > 0) {	/* wait for any children */
		pid = waitproc(&errflg);
		nLogMsg(EL_STATUS1, id, "child %d status %d",
			pid, errflg);
	}

	return 0;
}
