#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>

#include <sys/param.h>		/* these two to get bfid_t */
#ifdef __NetBSD__
#include <inttypes.h>
#endif
#include <miscfs/dmfs/rash.h>
#include "bfid.h"

static	char		pathname[PATH_MAX + 1];
static	char*		seq_pathname = NULL; 

static	u_quad_t	nextbfid = (u_quad_t)0, cacheceil = (u_quad_t)0;
static	uint32_t	site_id;
static	char		site_pattern[4];
static	u_int16_t	gen_id;
static	char		gen_pattern[2];

/*
**	A 6 byte (48 bit) value cut into three 2 byte (16 bit) pieces.
*/
typedef struct bsequence {
	u_int16_t	part[3];
} bsequence_t;

static	pid_t	pid = -1;


int readlinen(int fd, void* ptr, int nbytes);

/*
**	Take a 64 bit number and separate the lower 48 bits into
**	the three parts of a bsequence_t.  This is done in
**	network byte order.
*/
static	int int64_2_bsequence(bsequence_t* seq, int64_t val)
{
  seq->part[2] = htons(0x000000000000FFFF & val);
  seq->part[1] = htons((0x00000000FFFF0000 & val) >> 8);
  seq->part[0] = htons((0x0000FFFF00000000 & val) >> 16);
  return(0);
}

  
/*
**	Bump the value held in the bfid file and remember where
**	the new highwater point is.  This operation is done
**	to allocate bfids which can then be handed out by getbfid().
**	If the filename contained in seq_pathname is not already
**	open, open it and leave the file descriptor in seq_fd.
**	The value contained in the file will be the next bfid
**	returned out of the bfid cache.
**	Return 0 on success.
*/
int cache_bfid( int increment)
{
  static int	seq_fd = -1;
  u_quad_t	tmp_bfid;
  char		buffer[255];
  int		i;

  /* rudimentary arg checking */
  if(increment > MAX_BFID_REQUEST){
    fprintf(stderr,"request for %d cached bfids greater than max allowed %d\n",
	    increment, MAX_BFID_REQUEST);
    return(-2);
  } else if(increment < MIN_BFID_REQUEST){
    /* silently increase it to the miniumum */
    increment = MIN_BFID_REQUEST;
  }

  if(seq_fd == -1){
    if((seq_fd = open(seq_pathname, O_RDWR|O_CREAT|O_RSYNC|O_SYNC, 0644)) < 0){
      perror(seq_pathname);
      return(-1);
    }
  }

  /* read last value */

  lseek(seq_fd, 0L, SEEK_SET);
  flock(seq_fd, LOCK_EX);
  i = readlinen(seq_fd, buffer, 255);
  if(i > 0){
    /* check to see if the string is even remotely correct */
    if(strspn(buffer, "0123456789ABCDEFX") != strlen(buffer)){
      fprintf(stderr, "error parsing, \"%s\" in  %s", buffer, seq_pathname);
      flock(seq_fd, LOCK_UN);
      return(-1);
    }

    i = sscanf(buffer, "%qX", (unsigned long long *)&tmp_bfid);
    if (i != 1 || (tmp_bfid + increment) > 0x0000ffffffffffff) {
      fprintf(stderr, "bfid range error (%s)\n", buffer);
      flock(seq_fd, LOCK_UN);
      return(-1);
    }
  }
  else if (i == -1) {
    fprintf(stderr, "error reading bfid sequence from  %s", seq_pathname);
    flock(seq_fd, LOCK_UN);
    return(-1);
  }
  else
    tmp_bfid = 0;
  
  /* write new value, release lock and add in the desired increment */
  cacheceil = increment  + tmp_bfid;
  nextbfid = tmp_bfid;

  sprintf(buffer, "%016qX\n", (unsigned long long)cacheceil);
  lseek(seq_fd, 0L, SEEK_SET);
  write(seq_fd, buffer, strlen(buffer));

  /* remember who we are */
  pid = getpid();

  /* release lock */
  flock(seq_fd, LOCK_UN);
 
  return(0);
}

/*
**	Initialize the bfid server.  Read two files to get the
**	32 bit site id and the 16 bit generator id.  The files
**	are opened and read with no locking.  No file descriptors
**	are left open.
**	Return 0 on success.
*/
int bfid_init()
{
  int		i, len, fd;
  uint16_t	tmp16;
  uint32_t	tmp32;
  
  char buffer[255];

  len = strlen(SEQUENCE_FILE) + strlen(DEFAULT_BFID_HOME);

  seq_pathname = (char *)(malloc(len + 2));
  sprintf(seq_pathname, "%s/%s", DEFAULT_BFID_HOME, SEQUENCE_FILE);

  /* get site id */

  sprintf(pathname, "%s/%s", DEFAULT_BFID_HOME, SITE_ID_FILE );
  
  if((fd = open(pathname, O_RDONLY, 0)) < 0){
    perror(pathname);
    return(-1);
  }
  i = readlinen(fd, buffer, 255);
  close(fd);

  if(i < 0){
    fprintf(stderr, "error reading site id from %s", pathname);
    return(-1);
  }
  
  if(strspn(buffer, "0123456789ABCDEFX") != strlen(buffer)){
    fprintf(stderr, "error parsing %s, buffer = %s", pathname, buffer);
    return(-1);
  }
  sscanf(buffer, "%X", &site_id);

  tmp32 = htonl(site_id);
  memcpy(site_pattern, (char *)(&tmp32), 4);


  /* get generator id */

  sprintf(pathname, "%s/%s", DEFAULT_BFID_HOME, GENERATOR_ID_FILE);
  
  if((fd = open(pathname, O_RDONLY, 0)) < 0){
    perror(pathname);
    return(-1);
  }
  i = readlinen(fd, buffer, 255);
  close(fd);
  if(i < 0){
    fprintf(stderr, "error reading generator id from %s", pathname);
    return(-1);
  }
  
  if(strspn(buffer, "0123456789ABCDEFX") != strlen(buffer)){
    fprintf(stderr, "error parsing %s", pathname);
    return(-1);
  }
  sscanf(buffer, "%hX", &gen_id);

  tmp16 = htons(gen_id);
  memcpy(gen_pattern, (char *)(&tmp16), 2);
  
  return(0);
}


/*
**	Get a bfid.  If the generator is not initialized, call bfid_init().
**	If the bfid cache is exhausted, call cache_bfid() to add
**	more to the cache.
**	Return 0 on sussess.
*/
int getbfid(bfid_t *bP)
{
  bsequence_t seq;

  if(seq_pathname == NULL){
    if(bfid_init() != 0)
      return(-1);
  }
  if(getpid() != pid || nextbfid >= cacheceil){
    if(cache_bfid(MIN_BFID_REQUEST) != 0)
      return(-1);
  }
  int64_2_bsequence(&seq, nextbfid++);

  /*
  **	16 byte (128 bit) bfid
  **	byte	byte	len
  **	0	3	4	cleared
  **	4	7	4	site pattern
  **	8	9	2	generator pattern
  **	10	15	6	local bfid cache
  */
  memset(&(bP->bfid[0]), 0, 4);
  memcpy(&(bP->bfid[4]), site_pattern, 4);
  memcpy(&(bP->bfid[8]), gen_pattern, 2);
  memcpy(&(bP->bfid[10]), &seq, 6);

  return(0);
}
