/*
**  This program is the DMFS RASH restore daemon.
**
*/

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/param.h>
#include <miscfs/dmfs/dmfs.h>
#include <miscfs/dmfs/rash.h>
#include <miscfs/dmfs/dmfs_fcntl.h>

#include "nas_logmsg.h"
#include "vv_cl.h"
#include "dblib.h"
#include "nasutil.h"
#include "rashcom.h"
#include "dmfsd.h"

#define	KDEV		"/dev/kdcom"
#define NUMTYPE		2		/* number of message types */

struct link {
	struct link    *forw;
	struct link    *back;
};

typedef struct qent {
	struct link	p;
	fhandle_t	fh;
	bfid_t		bfid;
	int		pid;
	int		type;
	uid_t		dbuid;
	uid_t		uid;
} qent_t;

char		*com;
char		mypid[20];	/* nLogMsg eye catcher		*/
char		*arg0;

bfid_t		zerobfid;	/* zero fill */
char		*bfid;
int		type;
int		kfd;
sigset_t	sigkid;

struct	rashmsg	kc;
struct	link	active, waiting[NUMTYPE];
int		running[NUMTYPE], allowed[NUMTYPE];
ifunc		cleanup[NUMTYPE], handle[NUMTYPE];

int		resetflg = FALSE;	/* not used by dmfsd */
int		procs = 0;

int
dmfsd_fatal()
{
	/*
	 * If the disk file was opened then set the fail flag on
	 */
	if (rash_fd != -1) {
		fcntl(rash_fd, DMFS_SETRFAIL, NULL);
		close(rash_fd);
	}
	rashdb_close();
	vv_term();
	exit(1);
	return 0;	/* not reached */
}

/*
 *  getkid
 *  Enter on death of child signal.  SIGCHLD is automatically blocked
 *  in this function so we can manipulate the list.
 *
 *  Should change this so new processes are not forked from a signal
 *  handler function.  TODO
 */
void
getkid(int sig)
{
	char	*id = "getkid";
	int	pid, stat;
	qent_t	*ap;

	nLogMsg(EL_DEBUG3, id, "caught signal %d", sig);
	for (;;) {
		pid = getproc(&stat);
		if (pid == 0)
			break;

		ap = (qent_t *)active.forw;
		while (ap != (qent_t *)&active && ap->pid != pid)
			ap = (qent_t *)ap->p.forw;

		if (ap == (qent_t *)&active) {	/* not found! */
			nLogMsg(EL_ERROR2, id,
				"pid %d stat %d not found", pid, stat);
			return;
		}
		ap->p.back->forw = ap->p.forw;
		ap->p.forw->back = ap->p.back;
		type = ap->type;
		if (waiting[type].forw != &waiting[type])
			cleanup[type](NULL);
		else
			running[type]--;
		nFree(ap);
	}
	return;
}

/*
**	Return true if a type/bfid entry is found in a queue.
*/
int
dup_queue(struct link *queue)
{
	static	char	id[] = "dup_queue";
	qent_t		*oq;

	for (oq = (qent_t *)queue->forw; oq != (qent_t *)queue; 
				oq = (qent_t *)oq->p.forw) {
		if (oq->type == type &&
		    memcmp(&oq->bfid, &kc.rm_bfid, sizeof(bfid_t)) == 0) {
			nLogMsg(EL_STATUS1, id,
				"duplicate type %d bfid %s", type, bfid);
			return 1;
		}
	}
	return 0;
}

void
get_event()
{
	static	char	id[] = "get_event";
	int		nchar, errflg;
	qent_t		*ob;
	struct	link	*waitqueue;

	(void)sigprocmask(SIG_UNBLOCK, &sigkid, NULL);	/* enable SIGCHLD */
	nchar = read(kfd, &kc, sizeof(kc));
	errflg = errno;					/* save errno */
	(void)sigprocmask(SIG_BLOCK, &sigkid, NULL);	/* block SIGCHLD */
	errno = errflg;

	if (nchar != sizeof(kc)) {	/* check what we read from kernel */
		if (nchar == -1) {
			if (errflg != EINTR) {
				nLogMsg(EL_ERROR1|EL_SYS, id,
					"read from kfd %d", kfd);
			}
		}
		else if (nchar == 0)
			nLogMsg(EL_ERROR2, id, "kernel read EOF");
		else {
			nLogMsg(EL_ERROR2, id,
				"read struct rashmsg size %d", nchar);
		}
		return;
	}

	/*
	 *  elementary check of input
	 */
	if (memcmp(&kc.rm_bfid, &zerobfid, sizeof(bfid_t)) == 0) {
		nLogMsg(EL_ERROR1, id, "invalid bfid = 0");
		return;
	}

	/* set globals */
	type = kc.rm_type;
	bfid = bfidout(&kc.rm_bfid);

	switch (type) {

	case RASH_RESTORE:
	case RASH_OBSOLETE:
		break;

	default:
		nLogMsg(EL_ERROR1, id, "bad cmd=%d", type);
		return;
	}

	/*
	 *  Check queues to see if there is already 
	 *	a request in progress for this bfid.
	 */
	if (dup_queue(&active) || dup_queue(&waiting[type])) {
		return;
	}

	ob = (qent_t *) nAlloc(sizeof(qent_t));
	ob->bfid  = kc.rm_bfid;
	ob->type  = type;
	ob->dbuid = kc.rm_dbuid;
	ob->uid   = kc.rm_uid;
	ob->fh    = kc.rm_fh;

	if (running[type] < allowed[type]) {
		ob->pid = newproc(handle[type], ob);

		/* stick ob on the back of active */
		ob->p.forw = &active;
		ob->p.back = active.back;
		active.back->forw = &ob->p;
		active.back = &ob->p;

		running[type]++;
		nLogMsg(EL_DEBUG3, id,
			"start type %d running %d bfid %s",
			type, running[type], bfid);
	} else {
		waitqueue = &waiting[type];
		ob->p.forw = waitqueue;
		ob->p.back = waitqueue->back;
		waitqueue->back->forw = &ob->p;
		waitqueue->back = &ob->p;
	}

	return;
}

#define	FULLPRT(x)	(int)sizeof(x), (int)sizeof(x), x

char *
fhinfo(fhandle_t *fhp)
{
	static	char	info[200];
	char		*cp;
	int		i, len;

	len = fhp->fh_fid.fid_len;
	sprintf(info, "fsid %*.*x:%*.*x fid(%d)",
		FULLPRT(fhp->fh_fsid.val[0]),
		FULLPRT(fhp->fh_fsid.val[1]), len);
	cp = info + strlen(info);
	for (i=0; i<len; i++) {
		sprintf(cp, " %d", fhp->fh_fid.fid_data[i]);
		cp += strlen(cp);
	}
	*cp = '\0';

	return info;
}

/*
 *  Restore a file to disk from VV's. Input is global variable kc.
 */
int
restore()
{
	static	char	id[] = "restore";
	struct	stat	sbuf;
	rashstat_t	rstat;
	vvlist_t	*vvlist;
	USR_FILEINFO	filei;
	int		rc, nvv;
	char		mypid[32];

	/*
	 *  this is a forked process, so 
	 *	close kernel device left over from the parent
	 */
	close(kfd);

	/*
	 *  prepare logging
	 */
	sprintf(mypid, "resto(%05d)", getpid());
	nLogInit(mypid, dmfsd_fatal);
	nLogInitFile(ERRORRES, EL_LEVEL, 0, NULL);

	/*
	 *  open inode - Note: leave it open for restfile()
	 */
	if ((rash_fd = fhopen(&kc.rm_fh, O_WRONLY|O_ALT_IO)) < 0 ) {
		nLogMsg(EL_ERROR1|EL_SYS|EL_FATAL, id,
			"fhopen: %s", fhinfo(&kc.rm_fh));
	}

	/*
	 *  get/check rash-specific info
	 */
	if (fstat(rash_fd, &sbuf) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS|EL_FATAL, id,
			"FSTAT failed (kc.bfid %s kc.fh %s)",
			bfidout(&kc.rm_bfid), fhinfo(&kc.rm_fh));
	}

	if (fcntl(rash_fd, DMFS_GETRINFO, &rstat) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS|EL_FATAL, id,
			"rashcntl(R_GETST) bfid %s fh %s",
					bfidout(&kc.rm_bfid),
					fhinfo(&kc.rm_fh));
	}

	if (memcmp(&rstat.st_bfid, &kc.rm_bfid, sizeof(bfid_t))) {
		nLogMsg(EL_ERROR1|EL_FATAL, id,
			"bfid mismatch %s rstat=%llX kc=%llX",
			fhinfo(&kc.rm_fh),
			bfidout(&rstat.st_bfid),
			bfidout(&kc.rm_bfid));
	}

	if (!(rstat.st_rashflg & DM_NONR)) {
		nLogMsg(EL_ERROR1|EL_FATAL, com,
			"%s not in nonresident state rashflg=0x%x",
			fhinfo(&kc.rm_fh), rstat.st_rashflg);
	}

	vvlist = getdb(&filei, &sbuf, &rstat, &nvv);
	if (vvlist == NULL)
		nLogMsg(EL_ERROR1|EL_FATAL, id, "no VV's");

	/*
	**	Here, something could be done with the version
	**	number if needed.
	*/
	rc = restfile(&filei, &sbuf, &rstat, vvlist, nvv);

	nFree(vvlist);
	rashdb_close();
	vv_term();
	return rc;
}

/*
**	dmfsd_intrpt 
**	Come here if a sigterm signal is caught.  
**	Call dmfsd_fatal() to clean up and exit.
*/
void
dmfsd_intrpt(int signum)
{
	static	char	id[] = "dmfsd_intrpt";

	nLogMsg(EL_STATUS2, id,
		"processing signal %d", signum);

	(void)sig_ign(SIGTERM);
	dmfsd_fatal();
}

/*
**  sigusr1() - restart log if SIGUSR1 signal is caught
** 	(existing children will continue w/ old log)
*/
time_t	Start;

void 
sigusr1()
{
	static	char	id[] = "sigusr1";

	nLogMsg(EL_STATUS1, id, "-------- restarting log ----------");
	nLogInitFile(ERRORARC, EL_LEVEL, 1, "arc");
	nLogMsg(EL_STATUS1, id, "-------- restarted log ----------");
	nLogMsg(EL_STATUS1, id, "dmfsd started %s", ctime(&Start));

	nLogCloseFile(ERRORARC);

	nLogInitFile(ERRORRES, EL_LEVEL, 1, "rest");
	nLogMsg(EL_STATUS1, id, "-------- restarted log ----------");
	nLogMsg(EL_STATUS1, id, "dmfsd started %s", ctime(&Start));

	return;
}

/*
**	Mark a database record deleted.  The kernel sends a message
**	when rash can forget about a file.
*/
int
obsolete(void *notused)
{
	static	char	id[] = "obsolete";
	char		*pwname;
	char		mypid[32];
	extern	int	kfd;

	/*
	 *  this is a forked process, so 
	 *	close kernel device left over from the parent
	 */
	close(kfd);

	/*
	 *  prepare logging
	 */
	sprintf(mypid, "%s(%05d)", id, getpid());
	nLogInit(mypid, dmfsd_fatal);
	nLogInitFile(ERRORRES, EL_LEVEL, 0, NULL);
	nLogMsg(EL_STATUS1, id, "uid %d bfid %s",
		kc.rm_dbuid, bfidout(&kc.rm_bfid));

	/*
	 *  get the "database name" of this user
	 *	(user who owned the file when it 
	 *	was archived)
	 */
	pwname = nas_uidname(kc.rm_dbuid);
	if ( pwname == NULL ) {
		nLogMsg(EL_ERROR1|EL_FATAL, id,
			"dbuid %d not in the password file",
			kc.rm_dbuid);
		return NULL;
	}

	/* 
	 *  Get all of the records about this file
	 *	(bfid.VV) from the database
	 */
	if (rashdb_open(pwname, O_DB_RDWR))
		nLogMsg(EL_ERROR1|EL_FATAL, id, "db_open(%s)", pwname);

	if (rashdb_delete_bfid(&kc.rm_bfid) != RASHDB_OK) {
		nLogMsg(EL_ERROR1, id, "db_delete_bfid(%s)%s",
			pwname, bfidout(&kc.rm_bfid));
	}
	rashdb_close();

	return 0;
}

/*
**	Start a child to process a waiting request.
*/
int
startone()
{
	char		*id="startone";
	struct link	*waitqueue;
	struct qent	*ob;

	waitqueue = &waiting[type];
	ob = (struct qent *)waitqueue->forw;

	ob->p.forw->back = waitqueue;		/* dqueue ob from waitqueue */
	waitqueue->forw = ob->p.forw;

	kc.rm_bfid	= ob->bfid;
	kc.rm_dbuid	= ob->dbuid;
	kc.rm_uid	= ob->uid;
	kc.rm_fh	= ob->fh;

	nLogMsg(EL_DEBUG3, id, "bfid=%s", bfidout(&kc.rm_bfid));

	ob->pid = newproc(handle[type], NULL);

	ob->p.forw = &active;	/* stick ob on the back of active */
	ob->p.back = active.back;
	active.back->forw = &ob->p;
	active.back = &ob->p;

	return 0;
}

int
main(int argc, char *argv[])
{
	static	char	id[] = "main";
	static	char	devnull[] = "/dev/null";
	static	char	dev[] = "/dev/";
	char		*kdcom;
	int		nchar, c, i, fd;
	struct	link	*waitqueue;
	extern	char	*optarg;
	extern	int	optind, opterr;

	time(&Start);

	/*
	**	We don't want standard io to be connected
	**	but open them to devnull just in case they
	**	are used by some library.
	*/
	(void)close(0);
	(void)close(1);
	(void)close(2);

	fd = open(devnull, O_RDWR);
	assert(fd == 0);
	fd = open(devnull, O_RDWR);
	assert(fd == 1);
	fd = open(devnull, O_RDWR);
	assert(fd == 2);

	arg0 = argv[0];
	com  = nas_basename(arg0);

	handle[RASH_RESTORE] = restore; 
	allowed[RASH_RESTORE] =  MAXRESTORES;
	handle[RASH_OBSOLETE] = obsolete; 
	allowed[RASH_OBSOLETE] =  5;

	for (i=0; i<NUMTYPE; i++) {
		cleanup[i] = startone;
		running[i] = 0;

		waitqueue = &waiting[i];
		waitqueue->forw = waitqueue->back = waitqueue;
	}
	active.forw = active.back = &active;

	sprintf(mypid, "dmfsd(%05d)", getpid());
	nLogInit(mypid, dmfsd_fatal);
	nLogInitFile(ERRORRES, EL_LEVEL, 0, NULL);

	while ((c = getopt(argc, argv, "dr:")) != EOF) {
		switch (c) {
		case 'd':
			nLogSetLevel(EL_LEVEL);
			break;

		case 'r':		/* restores */
			if ((nchar = atoi(optarg)) == 0) {
				nLogMsg(EL_ERROR1|EL_FATAL,
					"bad number of restores %s", optarg);
			}
			allowed[RASH_RESTORE] = nchar;
			break;

		case '?':
		default:
			nLogMsg(EL_ERROR1|EL_FATAL, arg0, "see man page");
		}
	}
	/*
	**	If we are started from init, the device will
	**	be put on the command line.
	*/
	if (optind != argc-1)
		kdcom = KDEV;
	else {
		if (*argv[optind] == '/')
			kdcom = argv[optind];
		else {
			kdcom = nAlloc(strlen(argv[optind]) + strlen(dev) + 1);
			sprintf(kdcom, "%s%s", dev, argv[optind]);
		}
	}

	if (sig_ign(SIGHUP) || sig_ign(SIGINT))
		exit(1);

	(void)signal(SIGTERM, dmfsd_intrpt);
	(void)signal(SIGUSR1, sigusr1);
	(void)signal(SIGCHLD, getkid);

	(void)sigemptyset(&sigkid);
	(void)sigaddset(&sigkid, SIGCHLD);

	if ((kfd = open(kdcom, O_RDONLY)) == -1)
		nLogMsg(EL_ERROR1|EL_SYS|EL_FATAL, id, kdcom);

	if (fcntl(kfd, F_SETFD, 1) == -1) {	/* close on exec */
		nLogMsg(EL_ERROR1|EL_SYS|EL_FATAL, id,
			"fcntl F_SETFD %s", kdcom);
	}

	/*
	**  load passwd & group files for faster lookup
	*/
	nas_readpwd();
	nas_readgrp();

	nLogMsg(EL_STATUS2, id, "%d max restores allowed",
				allowed[RASH_RESTORE]);
	for (;;)
		get_event();

	return 0;	/* not reached */
}
