/*
**  This program provides the RASH "force archive" function.
**  Written by Tom Proett, NASA Ames Research Center.
**  This program is public domain software.
**
**  The usage is as follows:
**  forcearc [-v][-d][-r] [-f] file(s) ...
**
**  -v  Verbose output.
**
**  -d  Debug output.
**
**  -r	Normally, directories will be rejected from the list of
**	files.  If -r is given, any directories will be decended
**	and all plain files will be archived.
**      
**  -a  After a file is archived, release the used blocks from the file. 
**	Default: leave the data blocks around 
**
**  -A	Release the used blocks, giving the size to stay resident.
**
**  "file" specifies the file which is to be archived.  There may
**  be more than one.
*/

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/param.h>
#include <miscfs/dmfs/rash.h>

#include "nas_logmsg.h"
#include "vv_cl.h"
#include "dblib.h"
#include "nasutil.h"
#include "rashcom.h"
#include "dmfsd.h"

char		cwd[PATH_MAX+1];		/* current working dir	*/
uid_t		userid;
char		*arg0;
char		*com;
char		mypid[20];			/* nLogMsg eye catcher	*/
int		recurse=FALSE;			/* recursive flag 	*/
int		aflg;

extern OFF_T	leave_resi;

/*
**	Force a file to VV.
*/
int
forcefile(char *filename, struct stat *statP)
{
	static	char	*id = "forcefile";
	int	rc;

	if (userid != 0 && userid != statP->st_uid) {
		nLogMsg(EL_STATUS1, id, "%s: permision denied", filename);
		/* non root user can only do own files */
		return 0;
	}
	rc = rashvv(filename, statP);
	return rc;
}

int
main(int argc, char *argv[])
{
	char		*id = "main    ";
	char    	*filename;
	struct	stat	statbuf;
	char		*chr;
	int		c, errflg;
	int		verbose = FALSE;
	extern char	*optarg;
	extern int	optind, opterr;

	com = nas_basename(argv[0]);
	arg0 = argv[0];
	aflg = 0;

	errflg = FALSE;
	while ((c = getopt(argc, argv, "dvA:ar")) != EOF) {
		switch (c) {

		case 'd':
			nLogSetLevel(EL_LEVEL);
			break;

		case 'v':
			verbose = TRUE;
			break;

		case 'A':
			leave_resi = strtoq(optarg, &chr, 0);
			if (chr == optarg || *chr != '\0') {
				fprintf(stderr,
					"garbled resi value %s\n", optarg);
				errflg = 1;
			}
			else if (leave_resi < 0 || MAX_RESI < leave_resi) {
				fprintf(stderr,
					"resident size %s out of range\n",
					optarg);
				errflg = 1;
			}

		case 'a':
			aflg++;
			break;

		case 'r':
			recurse = TRUE;
			break;

		case '?':
			errflg = TRUE;
			break;
		}
	}

	if (optind == argc)		/* last option */
		errflg=TRUE;

	if (errflg == TRUE) {
		fprintf(stderr, "usage: %s %s %s\n", com,
		"[-a] [-d] [-v] [-r]", "file(s) ...");
		exit(1);
	}

	/*
         * setup log file
         */
        sprintf(mypid, "force(%05d)", getpid());
        nLogInit(mypid, fatal);
	if (verbose)
		nLogInitStderr(EL_STATUS1);
	nLogInitFile(ERRORARC, EL_LEVEL, 0, NULL);

	userid = getuid();		/* uid of callin proc */
	if ( (getcwd( cwd, sizeof(cwd) )) == NULL)
		nLogMsg(EL_ERROR1|EL_SYS|EL_FATAL, id, "getcwd failed");

	/*
         * if the process was started in the background then ignore
         * the keyboard interrupt signal by disconnecting from the terminal
         */
        if (!isatty(0)) {
                if ( setpgid(0,0) < 0)
			nLogMsg(EL_ERROR1|EL_SYS|EL_FATAL, id, "setpgid");
	}

	if (setuid(0) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS|EL_FATAL, id, 
			"unable to setuid to root");
	}

	(void)signal(SIGHUP, intrpt);
	(void)signal(SIGINT, intrpt);
	(void)signal(SIGTERM,intrpt);

	/*
	 *  load passwd & group files for faster lookup
	 */
	nas_readpwd();
	nas_readgrp();

	for (; optind < argc; optind++) {
		filename = fullname(argv[optind], cwd); /* file to archive */
		if (recurse == TRUE) {
			traverse(filename, forcefile);
		} else {
                        if (stat(filename, &statbuf) == -1) {
                                nLogMsg(EL_ERROR1|EL_SYS, id, filename);
                                continue;
                        }
			forcefile(filename, &statbuf);
		}
	}
	vv_term();

	fflush(stdout);
	fflush(stderr);

 	return 0;
}
