/*
**  This program provides the RASH "force restore" function.
**
**  The usage is as follows:
**  frestore  [-d xx][-rR] file(s) ...
**
**  -d  Debug value. 
**
**  -r	Normally, directories will be rejected from the list of
**	files.  If -r is given, any directories will be decended
**	and all plain files will be archived.
**    
**  -R  After the file is restored, clear the rash flag in the inode
**	that indicates that the file is archived on vv's. 
**
**  "file" specifies the file to be restored
*/

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/param.h>
#include <miscfs/dmfs/dmfs.h>
#include <miscfs/dmfs/rash.h>
#include <miscfs/dmfs/dmfs_fcntl.h>

#include "nas_logmsg.h"
#include "vv_cl.h"
#include "dblib.h"
#include "nasutil.h"
#include "rashcom.h"
#include "dmfsd.h"

char		cwd[PATH_MAX+1];		/* current working dir	*/
uid_t		userid;
char		*com;
int		recurse=FALSE;			/* recursive flag 	*/
int		resetflg = FALSE;

/*
**  Come here to die.
*/
static int
die()
{
	rashdb_close();

	exit(1);
}

/*
 *  intrpt_forcearc 
 *  Come here if a signal is caught.
 */
void
intrpt(int signum)
{
	static	char	id[] = "intrpt";

	nLogMsg(EL_STATUS2, id, "processing signal %d", signum);

	(void)sig_ign(SIGHUP);
	(void)sig_ign(SIGINT);
	(void)sig_ign(SIGTERM);

	die();
}

/*
**	Do a restore.  The rash_fd will be closed.
*/
int
frestore(char *fname, struct stat *sP)
{
	static	char	id[] = "frestore";
	rashstat_t	rstat;
	vvlist_t	*vvlist;
	USR_FILEINFO	filei;
	int		rc, nvv;

	rash_fd = open(fname, O_WRONLY|O_ALT_IO);
	if (rash_fd == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "rashopen(%s)", fname);
		return 1;
	}
	if (fcntl(rash_fd, DMFS_GETRINFO, &rstat)  == -1 ) {
		nLogMsg(EL_ERROR1|EL_SYS, id,
			"R_GETST failed for %s", fname);
		close(rash_fd);
		rash_fd = -1;
		return 1;
	}
	/*
 	*  If already on disk then skip it 
	*/
	if ( !(rstat.st_rashflg & DM_NONR) ) {
		nLogMsg(EL_STATUS1, id,
			"%s is already resident", fname);
		/*
	 	* If user requested to reset archive flag then 
	 	* comply
	 	*/
		if (resetflg)
			resetfd(id, fname);
		close(rash_fd);
		rash_fd = -1;
		return 0;
	}
	nLogMsg(EL_STATUS1, id, "%s: %s", bfidout(&rstat.st_bfid), fname);
	vvlist = getdb(&filei, sP, &rstat, &nvv);
	/*
	**	Here, something could be done with the version
	**	number if needed.
	*/
	rc = restfile(&filei, sP, &rstat, vvlist, nvv);

	nFree(vvlist);
	return rc;
}

int
main(int argc, char *argv[])
{
	char		*id = "main";
	char		mypid[20];
	char    	*filename;
	struct	stat	statbuf;
	int		c, errflg;
	extern char	*optarg;
	extern int	optind, opterr;

	com = nas_basename(argv[0]);

	/*
         * setup log file
         */
        sprintf(mypid, "frestore(%05d)", getpid());
        nLogInit(mypid, die);
	nLogInitStderr(EL_STATUS1);
	nLogInitFile(ERRORRES, EL_LEVEL, 0, NULL);

	userid = getuid();		/* uid of callin proc */
	if ( (getcwd( cwd, sizeof(cwd) )) == NULL)
		nLogMsg(EL_ERROR1|EL_SYS|EL_FATAL, id, "getcwd failed");

	/*
         * if the process was started in the background then ignore
         * the keyboard interrupt signal by disconnecting from the terminal
         */
        if (!isatty(0)) {
                if ( setpgid(0,0) < 0)
			nLogMsg(EL_ERROR1|EL_SYS|EL_FATAL, id, "setpgid");
	}

	errflg = FALSE;
	while ((c = getopt(argc, argv, "drR")) != EOF) {
		switch (c) {

		case 'd':
			nLogSetLevel(EL_LEVEL);
			break;

		case 'r':
			recurse = TRUE;
			break;

		case 'R':
			resetflg = TRUE;
			break;

		case '?':
		default:
			errflg = TRUE;
			break;
		}
	}

	if (optind == argc)		/* last option */
		errflg=TRUE;

	if (errflg == TRUE) {
		fprintf(stderr, "usage: %s %s %s\n", com,
		"[-d][-r][-R]", "file(s) ...");
		exit(1);
	}

	if (setuid(0) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS|EL_FATAL, id, 
			"unable to setuid to root");
	}

	(void)signal(SIGHUP, intrpt);
	(void)signal(SIGINT, intrpt);
	(void)signal(SIGTERM,intrpt);

	/*
	 *  load passwd & group files for faster lookup
	 */
	nas_readpwd();
	nas_readgrp();

	for (; optind < argc; optind++) {
		filename = fullname(argv[optind], cwd); /* file to restore */
		if (recurse == TRUE) {
			traverse(filename, frestore);
		} else {
                        if (stat(filename, &statbuf) == -1) {
                                nLogMsg(EL_ERROR1|EL_SYS, id, filename);
                                continue;
                        }
			frestore(filename, &statbuf);
		}
	}
	rashdb_close();
	vv_term();

	fflush(stdout);
	fflush(stderr);

 	return 0;
}
