/*
**	labels.c - rash labels that bracket file data in a VV
**
**		(Before MSS3.0, a series of ANSI labels was
**		written to tape.)
*/
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/param.h>
#include <miscfs/dmfs/rash.h>

#include "nas_logmsg.h"
#include "vv_cl.h"
#include "dblib.h"
#include "rashcom.h"
#include "labels.h"

static char		EndMark[] = "%END%";
int			EndMarkSize = ENDMRK_LEN;

static char		rashvvlbl[] = "RASH";		/* vv_hdr_lbl_t.hdr  */
static char		filelbl[] = "FILE";
static char		hdrlabel[] = "HDR";
static char		eoflabel[] = "EOF";
static char		eovlabel[] = "EOV";

static void
buildvvhdr(vv_hdr_lbl_t *lP, char *vvname, char *uname, uid_t dbuid)
{
	static char	id[] = "buildhdr";
	int		len;

	/*
	**	construct label
	*/
	memset(lP, ' ', sizeof(*lP));
	lP->end = '\n';

	memcpy(lP->hdr, rashvvlbl, sizeof(lP->hdr));
	memcpy(lP->vvname, vvname, strlen(vvname));

	if (C2ASC(SetDecInt, lP->version, RASH_VERSION))
		nLogMsg(EL_ERROR1|EL_FATAL, id, "setting RASH_VERSION");

	len = strlen(uname);
	if (len > sizeof(lP->dbuid_name)) {
		nLogMsg(EL_ERROR2, id, 
			"dbuid login name %s exceeds label format, truncating",
			uname);
		len = sizeof(lP->dbuid_name);
	}
	memcpy(lP->dbuid_name, uname, len);

	if (C2ASC(SetDecInt, lP->dbuid, dbuid)) 
		nLogMsg(EL_ERROR1|EL_FATAL, id, "setting dbuid");
}

int
lbl_vv_hdr(vv_handle_t *vhp, char *vvname, char *uname, uid_t dbuid)
{
	static char	id[] = "lbl_vv_hdr";
	vv_hdr_lbl_t	lbl;

	/*
	**	build basic label and date stamp it
	*/
	buildvvhdr(&lbl, vvname, uname, dbuid);
	SetTime(lbl.date, time((time_t *)0));

	/*
	**	write label out
	*/
	if (vv_write(vhp, &lbl, sizeof(lbl)) != sizeof(lbl)) {
		vvErrorMsg(EL_ERROR1, id, "vv_write");
		return 1;
	}
	return 0;
}

int
lbl_vv_hdrcheck(vv_handle_t *vhp, char *vvname, char *uname, uid_t dbuid)
{
	static char	id[] = "lbl_vvhdrcheck";
	vv_hdr_lbl_t	lbl1, lbl2;

	/*
	**	build basic label for comparison
	*/
	buildvvhdr(&lbl1, vvname, uname, dbuid);

	/*
	**	seek to beginning of vv and read label
	*/
	if (vv_lseek(vhp, (OFF_T)0, SEEK_SET) == (OFF_T)-1) {
		vvErrorMsg(EL_ERROR1, id, "vv_lseek");
		return 1;
	}
	if (vv_read(vhp, &lbl2, sizeof(lbl2)) != sizeof(lbl2)) {
		vvErrorMsg(EL_ERROR1, id, "vv_read");
		return 1;
	}

	/*
	**	'white-out' read-in date for easy comparison
	*/
	memset(lbl2.date, ' ', sizeof(lbl2.date));

	/*
	**	compare
	*/
	if (memcmp(&lbl1, &lbl2, sizeof(lbl1))) {
		lbl1.end = '\0';
		lbl2.end = '\0';
		nLogMsg(EL_ERROR1, id, "vv / db\n%s\n%s",
			(char *)&lbl2, (char *)&lbl1);
		return 1;
	}
	return 0;
}

/*
**	buildfilelbl() - generic info for 'HDR' 'EOV' or 'EOF'
*/
static void
buildfilelbl(file_hdr_lbl_t *lP, USR_FILEINFO *fP, char *groupname,
	char *vv0name, char *othervvname)
{
	static char	id[] = "buildfilelbl";
	int		len;

	/*
	**	basic stuff
	*/
	memset(lP, ' ', sizeof(*lP));
	lP->end = '\n';

	memcpy(lP->hdr, filelbl, sizeof(lP->hdr));

	if (C2ASC(SetDecInt, lP->version, RASH_VERSION))
		nLogMsg(EL_ERROR1|EL_FATAL, id, "setting RASH_VERSION");

	/*
	**	stuff for multi-vv files
	*/
	if (vv0name != NULL)
		memcpy(lP->vv0, vv0name, strlen(vv0name));

	if (C2ASC(SetDecInt, lP->vvno, fP->usr_vvno))
		nLogMsg(EL_ERROR1|EL_FATAL, id, "setting vol_order");

	if (othervvname != NULL) {
		len = strlen(othervvname);
		memcpy(lP->othervv, othervvname, len);
	}

	/*
	**	stuff for multi-file vv
	*/
	if (C2ASC(SetDecInt, lP->fno, fP->usr_fno))
		nLogMsg(EL_ERROR1|EL_FATAL, id, "setting fno");


	/*
	**	basic file info
	*/
	if (Bin2Hex(lP->bfid, BFIDLEN, &fP->usr_bfid, sizeof(bfid_t)))
		nLogMsg(EL_ERROR1|EL_FATAL, id, "setting bfid");

	if (C2ASC(SetDecInt, lP->uid, fP->usr_uid))
		nLogMsg(EL_ERROR1|EL_FATAL, id, "setting uid");

	if (groupname != NULL) {
		len = strlen(groupname);
		if (len > sizeof(lP->gname)) {
			nLogMsg(EL_ERROR2, id, 
				"grp name %s exceeds label format, truncating",
				groupname);
			len = sizeof(lP->gname);
		}
		memcpy(lP->gname, groupname, len);
	}
	if (C2ASC(SetDecInt, lP->gid, fP->usr_gid))
		nLogMsg(EL_ERROR1|EL_FATAL, id, "setting gid");

	if (C2ASC(SetHexInt, lP->mode, fP->usr_mode))
		nLogMsg(EL_ERROR1|EL_FATAL, id, "setting mode");

	SetTime(lP->mtime, fP->usr_mtime);
	SetTime(lP->ctime, fP->usr_ctime);
	SetTime(lP->arctm, fP->usr_arctm);

	if (C2ASC(SetHexOFF, lP->fsize, fP->usr_fsize))
		nLogMsg(EL_ERROR1|EL_FATAL, id, "setting size");

	if (C2ASC(SetHexOFF, lP->lseek, fP->usr_lseek))
		nLogMsg(EL_ERROR1|EL_FATAL, id, "setting lseek");

	if (C2ASC(SetHexOFF, lP->vvdata, fP->usr_vvdata))
		nLogMsg(EL_ERROR1|EL_FATAL, id, "setting vvdata");
	return;
}

file_hdr_lbl_t *
lbl_file_hdr(vv_handle_t *vhp, USR_FILEINFO *fP, char *groupname,
		char *vv0name, char *prevvname)
{
	static char	id[] = "lbl_file_hdr";
	file_hdr_lbl_t	*lP = nAlloc(sizeof(file_hdr_lbl_t));
	int		len = strlen(fP->usr_fname);

	/*
	**	build generic part of label
	*/
	buildfilelbl(lP, fP, groupname, vv0name, prevvname);

	/*
	**	make it a HDR and add file len
	*/
	memcpy(lP->label, hdrlabel, sizeof(lP->label)); 
	if (C2ASC(SetDecInt, lP->flen, len))
		nLogMsg(EL_ERROR1|EL_FATAL, id, "setting flen");

	/*
	**	append lbl, fname, and endmark
	*/
	if (vv_write(vhp, lP, sizeof(*lP)) != sizeof(*lP)) {
		vvErrorMsg(EL_ERROR1, id, "hdr");
		goto err;
	}
	if (vv_write(vhp, fP->usr_fname, len) != len) {
		vvErrorMsg(EL_ERROR1, id, "fname");
		goto err;
	}
	if (vv_write(vhp, EndMark, EndMarkSize) != EndMarkSize) {
		vvErrorMsg(EL_ERROR1, id, "hdr endmark");
		goto err;
	}

	return lP;
err:
	nFree(lP);
	return NULL;
}

/*
**	EOF means no further VV's and no filename.
*/
int
lbl_eof(vv_handle_t *vhp, file_hdr_lbl_t *lP)
{
	static	char	id[] = "lbl_eof";

	memcpy(lP->label, eoflabel, sizeof(lP->label)); 
	memset(lP->othervv, ' ', sizeof(lP->othervv));
	memset(lP->flen, ' ', sizeof(lP->flen));

	if (vv_write(vhp, lP, sizeof(*lP)) != sizeof(*lP)) {
		vvErrorMsg(EL_ERROR1, id, "eof");
		return 1;
	}
	if (vv_write(vhp, EndMark, EndMarkSize) != EndMarkSize) {
		vvErrorMsg(EL_ERROR1, id, "eof endmark");
		return 1;
	}
	return 0;
}

int
lbl_eof_check(vv_handle_t *vhp, USR_FILEINFO *fP)
{
	static char	id[] = "lbl_eof_check";
	char		buf[ENDMRK_LEN];
	file_hdr_lbl_t	lbl1, lbl2;
	OFF_T		off;

	/*
	**	seek back from end and check endmark
	*/
	off = vv_lseek(vhp, (OFF_T)-(ENDMRK_LEN), SEEK_END);
	if (off == (OFF_T)-1) {
		vvErrorMsg(EL_ERROR1, id, "lseek endmark");
		return 1;
	}
	if (vv_read(vhp, buf, ENDMRK_LEN) != ENDMRK_LEN) {
		vvErrorMsg(EL_ERROR1, id, "read endmark");
		return 1;
	}
	if (memcmp(buf, EndMark, ENDMRK_LEN)) {
		nLogMsg(EL_ERROR1, id, "expected %s got %.*s",
					EndMark, ENDMRK_LEN, buf);
		return 1;
	}

	/*
	**	build expected eof lbl
	*/
	buildfilelbl(&lbl1, fP, NULL, NULL, NULL);
	memcpy(lbl1.label, eoflabel, sizeof(lbl1.label));

	/*
	**	seek further back from end and read eof label
	*/
	off = (OFF_T)(sizeof(lbl2) + ENDMRK_LEN);
	off = vv_lseek(vhp, -off, SEEK_END);
	if (off == (OFF_T)-1) {
		vvErrorMsg(EL_ERROR1, id, "lseek eof");
		return 1;
	}
	if (vv_read(vhp, &lbl2, sizeof(lbl2)) != sizeof(lbl2)) {
		vvErrorMsg(EL_ERROR1, id, "read eof");
		return 1;
	}
	/*
	**	blank out the lbl parts that aren't in dbase
	*/
	memset(lbl2.vv0, ' ', sizeof(lbl2.vv0));
	memset(lbl2.gname, ' ', sizeof(lbl2.gname));
	memset(lbl2.othervv, ' ', sizeof(lbl2.othervv));

	/*
	**	compare
	*/
	if (memcmp(&lbl1, &lbl2, sizeof(lbl1))) {
		lbl1.end = '\0';
		lbl2.end = '\0';
		nLogMsg(EL_ERROR1, id, "vv / db\n%s\n%s",
			(char *)&lbl2, (char *)&lbl1);
		return 1;
	}
	return 0;
}

/*
**	EOV has the next vvname but no filename.
*/
int
lbl_eov(vv_handle_t *vhp, file_hdr_lbl_t *lP, char *nextvvname)
{
	static	char	id[] = "lbl_eov";

	memcpy(lP->label, eovlabel, sizeof(lP->label));
	memcpy(lP->othervv, nextvvname, strlen(nextvvname));
	memset(lP->flen, ' ', sizeof(lP->flen));

	if (vv_write(vhp, lP, sizeof(*lP)) != sizeof(*lP)) {
		vvErrorMsg(EL_ERROR1, id, "eov");
		return 1;
	}
	if (vv_write(vhp, EndMark, EndMarkSize) != EndMarkSize) {
		vvErrorMsg(EL_ERROR1, id, "endmark");
		return 1;
	}
	return 0;
}

/*
**	lbl_skipbof() - Check that the labels are correct and 
**		leave VV ready to read the file. Return non-0 if problem.
*/
int 
lbl_skipbof(vv_handle_t *vhp, vvlist_t *vlP, USR_FILEINFO *fP)
{
	static char	*id = "skipbof";
	file_hdr_lbl_t	lbl1, lbl2;
	char		buf[ENDMRK_LEN];
	int		flen;

	/*
	**	construct HDR label for comparison
	*/
	buildfilelbl(&lbl1, fP, NULL, NULL, NULL);
	memcpy(lbl1.label, hdrlabel, sizeof(lbl1.label)); 

	/*
	**	position to beginning of hdr label and read it
	*/
	if (vv_lseek(vhp, vlP->vvlseek, SEEK_SET) != vlP->vvlseek) {
		vvErrorMsg(EL_ERROR1, id, "lseek");
		return 1;
	}
	if (vv_read(vhp, &lbl2, sizeof(lbl2)) != sizeof(lbl2)) {
		vvErrorMsg(EL_ERROR1, id, "read hdr");
		return 1;
	}
	/*
	**	remember file len
	*/
	if (C2BIN(DecInt, &flen, lbl2.flen))
		nLogMsg(EL_ERROR1|EL_FATAL, id, "getting flen");
	/*
	**	blank out fields not in dbase for comparison
	*/
	memset(lbl2.vv0, ' ', sizeof(lbl2.vv0));
	memset(lbl2.gname, ' ', sizeof(lbl2.gname));
	memset(lbl2.othervv, ' ', sizeof(lbl2.othervv));
	memset(lbl2.flen, ' ', sizeof(lbl2.flen));

	/*
	**	compare
	*/
	if (memcmp(&lbl1, &lbl2, sizeof(lbl1))) {
		lbl1.end = '\0';
		lbl2.end = '\0';
		nLogMsg(EL_ERROR1, id, "vv / db\n%s\n%s",
				(char *)&lbl2, (char *)&lbl1);
		return 1;
	}

	/*
	**	skip over file name and check endmark
	*/
	if (vv_lseek(vhp, flen, SEEK_CUR) == -1) {
		vvErrorMsg(EL_ERROR1, id, "lseek end");
		return 1;
	}
	if (vv_read(vhp, buf, ENDMRK_LEN) != ENDMRK_LEN) {
		vvErrorMsg(EL_ERROR1, id, "endmark");
		return 1;
	}
	if (memcmp(buf, EndMark, ENDMRK_LEN)) {
		nLogMsg(EL_ERROR1, id, "expected %s got %.*s",
					EndMark, ENDMRK_LEN, buf);
		return 1;
	}

	return 0;
}
