/*********************************************************************
 *                              dblib.c                             
 *                                                                  
 * The enclosed routines perform various functions required to support
 * the RASH data base. 
 *********************************************************************/

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <assert.h>

#include "db3/db.h"

#include <sys/types.h>
#include <sys/param.h>
#include <miscfs/dmfs/rash.h>

#include "nasutil.h"
#include "dblib.h"
#include "nas_logmsg.h"

/*
**	Database info
*/
static	int		db_status;
static	char		db_username[20];	/* name of the user	*/
static	char		db_dir[PATH_MAX+1];	/* name of data db dir	*/

static	int		db_lock_fd = -1;	/* fd for lock file	*/

static	int		db_tbl_fd = -1;		/* data table fd	*/
static	int		db_cpy_fd = -1;		/* table copy fd	*/
static	int		db_fd;			/* descriptor in use 	*/

static	DB		*db_bfid_idx = NULL;
static	DB		*db_vv_idx = NULL;

/*
**	Table/record
*/
static	off_t		db_tbl_addr;		/* read position	*/
static	off_t		db_last_addr;		/* last read position	*/
static	USR_FILEINFO	db_binrec;
static	char		db_usr_fname[MAXPATHLEN+1];
static	char		db_fname_bfid[BFIDLEN];
static	DB_FILEINFO	db_asciirec;
static	DB_AUXREC	db_auxrec;


/*
**	These are the index numbers to give the various database
**	file names.
*/
#define		DBFILE_STATUS		0
#define		DBFILE_DATA		1
#define		DBFILE_BFID_INDEX	2
#define		DBFILE_VV_INDEX		3
#define		DBFILE_DATA_COPY	4
#define		DBFILE_BFID_COPY	5
#define		DBFILE_VV_COPY		6

#define		DBFILE_NUM		7
#define		COPY			3

static	char		*db_file_suffix[] = {
		"status",		/* 0	DBFILE_STATUS		*/
		"tbl",			/* 1	DBFILE_DATA		*/
		"bfid_idx",		/* 2	DBFILE_BFID_INDEX	*/
		"vv_idx",		/* 3	DBFILE_VV_INDEX		*/
		"tbl_cp",		/* 4	DBFILE_DATA_COPY	*/
		"bfid_idx_cp",		/* 5	DBFILE_BFID_COPY	*/
		"vv_idx_cp",		/* 6	DBFILE_VV_COPY		*/
};

static	char		db_file[DBFILE_NUM][PATH_MAX+1];

static	char	vv_keyfmt[] = "%s_%04d";
static	DBC	*vv_cursor;
static	char	vv_key[512];
static	DBT	vvkey, vvdata;

/*
 *	Unlock the database.  This only affects the status file.
 */
static	void
rashdb_unlock(void)
{
	static	char	id[] = "rashdb_unlock";

	assert(db_status != 0);
	assert(db_lock_fd != -1);

	if (ftruncate(db_lock_fd, (off_t)0) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "truncate %s",
			db_file[DBFILE_STATUS]);
	}
	if (close(db_lock_fd) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "close %s",
			db_file[DBFILE_STATUS]);
	}
	db_status = 0;
	db_lock_fd = -1;
	return;
}

/*
**	Close table and cleanup database state.
*/
static	void
rashdb_close_files(void)
{
	if (db_tbl_fd != -1) {
		(void)fsync(db_tbl_fd);
		(void)close(db_tbl_fd);
		db_tbl_fd = -1;
	}
	if (db_cpy_fd != -1) {
		(void)fsync(db_cpy_fd);
		(void)close(db_cpy_fd);
		db_cpy_fd = -1;
	}

	if (db_bfid_idx) {
		(void)db_bfid_idx->close(db_bfid_idx, 0);
		db_bfid_idx = NULL;
	}

	if (vv_cursor != NULL) {
		(void)vv_cursor->c_close(vv_cursor);
		vv_cursor = NULL;
	}
	if (db_vv_idx) {
		(void)db_vv_idx->close(db_vv_idx, 0);
		db_vv_idx = NULL;
	}
}

/*
**	Open an index.
*/
static	int
rashdb_openidx(DB **db, char *name, uint flags)
{
	int		ret;

	assert(*db == NULL);

	if ((ret = db_create(db, NULL, 0)) != 0)
		return ret;

	ret = (*db)->open(*db, name, NULL, DB_BTREE, flags, 0600);
	return ret;
}

/*
**	Close and unlock the database.
*/
void
rashdb_close(void)
{
	static	char	id[] = "db_close";
	int		i;

	if (db_status == 0)
		return;

	rashdb_close_files();

	/*
	**	If there is a copy to deal with, do some renames.
	*/
	if (db_status == O_DB_COPY) {
		for (i=DBFILE_DATA; i<DBFILE_DATA_COPY; i++) {
			if (unlink(db_file[i]) == -1) {
				nLogMsg(EL_ERROR1|EL_SYS, id,
					"unlink %s", db_file[i]);
				continue;
			}
			if (rename(db_file[i+COPY], db_file[i]) == -1) {
				nLogMsg(EL_ERROR1|EL_SYS, id,
					"rename %s to %s",
					db_file[i+COPY], 
					db_file[i]);
			}
		}
	}
	rashdb_unlock();
	return;
}

static int
table_bad(char *name, int *flag)
{
	static	char	id[] = "table_bad";
	struct	stat		sb;

	if (stat(name, &sb) != -1) {
		/*
		**	check size
		*/
		if (sb.st_size % sizeof(db_asciirec)) {
			nLogMsg(EL_ERROR1, id, 
				"table (%s) size %lld not integral record %d",
				name, sb.st_size, sizeof(db_asciirec));
			return 1;
		}
		*flag = 1;
	} else {
		if (errno != ENOENT) {
			nLogMsg(EL_ERROR1|EL_SYS, id, "stat(%s)",
				db_file[DBFILE_DATA]);
			return 1;
		}
		*flag = 0;
	}
	return 0;
}

/*
**	Read one record from the table with no seek.
*/
static int
read_rec(void *rec)
{
	static char	id[] = "read_rec";
	ssize_t		ret;

	ret = read(db_tbl_fd, rec, sizeof(db_asciirec));
	if (ret == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "read");
		return EIO;
	}
	if (ret != sizeof(db_asciirec)) {
		if (ret == 0)
			return RASHDB_EOF;

		nLogMsg(EL_ERROR1|EL_SYS, id, "reading record %d/%d",
						ret, sizeof(db_asciirec));
		return RASHDB_ERROR;
	}
	return RASHDB_OK;
}

/*
**	Make sure an ascii record has separators in the right places
**		(this is for db_asciirec, not db_auxrec).
*/
static int
check_rec()
{
	static char	*id = "check_rec";
	static char	*seploc[] = {
		&db_asciirec.b01,
		&db_asciirec.b02,
		&db_asciirec.b03,
		&db_asciirec.b04,
		&db_asciirec.b05,
		&db_asciirec.b06,
		&db_asciirec.b07,
		&db_asciirec.b08,
		&db_asciirec.b09,
		&db_asciirec.b10,
		&db_asciirec.b11,
		&db_asciirec.b12,
		&db_asciirec.b13,
		&db_asciirec.b14,
		&db_asciirec.b15,
		&db_asciirec.b16,
		&db_asciirec.b17,
		&db_asciirec.b18,
		&db_asciirec.b19,
		&db_asciirec.b20,
		&db_asciirec.b21,
		&db_asciirec.b22,
		NULL
	};
	int		i, bad = 0;

	for (i=0; seploc[i]; i++) {
		if (*seploc[i] != ' ') {
			bad++;
			nLogMsg(EL_ERROR1, id,
				"separator %d bad(%c)", i, *seploc[i]);
		}
	}
	if (db_asciirec.db_nrec[1] != '/'  &&
	    db_asciirec.db_nrec[1] != 't') {
		bad++;
		nLogMsg(EL_ERROR1, id, "nrec[1] != '/'");
	}
	if (db_asciirec.end1 != '|') {
		bad++;
		nLogMsg(EL_ERROR1, id, "end1 != '|'");
	}
	if (db_asciirec.end2 != '\n') {
		bad++;
		nLogMsg(EL_ERROR1, id, "end1 != newline");
	}
	return bad;
}

#define	ZERO_OBJ(obj)	memset(&obj, 0, sizeof(obj))

#define COPY_DBT(obj, dbt) \
	assert(sizeof(obj) == dbt.size); \
	memcpy(&obj, dbt.data, sizeof(obj))

#define	FILL_DBT(dbt, item) \
	ZERO_OBJ(dbt); \
	dbt.data = &(item); \
	dbt.size = sizeof(item)

#define	PTR_DBT(dbt, item) \
	ZERO_OBJ(dbt); \
	dbt.data = item; \
	dbt.size = sizeof(*item)

/*
**	Index a record by bfid and vv.
*/
static int
index_record(off_t addr)
{
	static	char	id[] = "index_rec";
	int		fno, err;
	DBT		key;
	DBT		data;

	FILL_DBT(data, addr);
	if (db_binrec.usr_vvno == 0  &&  
	    !(db_binrec.usr_rflg & (RFLG_DELETED|RFLG_NOIDX))) {
		/*
		**	Clean up any previous records with this bfid.
		*/
		if (rashdb_delete_bfid(&db_binrec.usr_bfid) != RASHDB_OK)
			return EINVAL;

		/*
		**	add to the bfid index 
		*/
		FILL_DBT(key, db_binrec.usr_bfid);
		if ((err = db_bfid_idx->put(db_bfid_idx, NULL, &key, &data,
				DB_NOOVERWRITE)) != 0) {
			nLogMsg(EL_ERROR1, id,
				"bfid add_key %s: %s",
				bfidout(&db_binrec.usr_bfid),
				db_strerror(err));
			return EINVAL;
		}
	}

	/*
	**	add to vv index
	**	skip tapes w/ tapedb-flagged problems
	*/
	if ( db_binrec.usr_vflg & (VFLG_BAD|VFLG_NONEXISTENT) )
		goto done;

	/*
	**	add vv_fileno to vv index
	**		file number in the index is 'reversed'
	**		in order to allow locate_key() searches 
	**		to find the last file on VV/tape 1st.
	*/
	fno = MAXTAPEFILES - db_binrec.usr_fno;
	sprintf(vv_key, vv_keyfmt, db_binrec.usr_vvname, fno);
	memset(&key, 0, sizeof(key));
	key.data = vv_key;
	key.size = strlen(vv_key);
	if (db_vv_idx->put(db_vv_idx, NULL,
			&key, &data, DB_NOOVERWRITE) != 0) {
		nLogMsg(EL_ERROR1, id,
			"vv add_key(%s)", vv_key);
		return EINVAL;
	}
  done:
	return RASHDB_OK;
}

static int
check_version(int version, char *id)
{
	if (version == RASH_VERSION)
		return 0;
	if (version == 211)
		return 0;

	nLogMsg(EL_ERROR1, id, "invalid version: %d", version);
	return 1;
}

/* 
**	rashdb_asc_to_bin() - convert record from ascii to binary format 
**		except, leave filename for caller to handle
*/
static int
rashdb_asc_to_bin(uint mask) 
{
	static char	*id = "rashdb_asc_to_bin";
	uint		uconv;
	int		iconv;

	/*
	**	always convert/check version
	*/
	if (C2BIN(DecInt, &db_binrec.usr_version, db_asciirec.db_version))
		return EINVAL;
	if (check_version(db_binrec.usr_version, id))
		return EINVAL;

	/*
	**	convert the requested fields
	*/
	if (mask & USR_BFID) {
		if (Hex2Bin(&db_binrec.usr_bfid, sizeof(bfid_t),
				db_asciirec.db_bfid, (size_t)BFIDLEN))
			return EINVAL;
	}

	if (mask & USR_FLG) {
		if (C2BIN(HexUshort, &db_binrec.usr_rflg, db_asciirec.db_rflg))
			return EINVAL;
	}

	if (mask & USR_FSIZE) {
		if (C2BIN(HexOFF, &db_binrec.usr_fsize, db_asciirec.db_fsize))
			return EINVAL;
	}

	if (mask & USR_UID) {
		if (C2BIN(DecInt, &iconv, db_asciirec.db_uid))
			return EINVAL;
		db_binrec.usr_uid = (uid_t)iconv;
	}

	if (mask & USR_GID) {
		if (C2BIN(DecInt, &iconv, db_asciirec.db_gid))
			return EINVAL;
		db_binrec.usr_gid = (gid_t)iconv;
	}

	if (mask & USR_MODE) {
		if (C2BIN(HexInt, &uconv, db_asciirec.db_mode))
			return EINVAL;
		db_binrec.usr_mode = (mode_t)uconv;
	}

	if (mask & USR_GROUP) {
		if (C2BIN(DecInt, &db_binrec.usr_group_no,
				db_asciirec.db_group_no))
			return EINVAL;
	}

	if (mask & USR_CTIME)
		db_binrec.usr_ctime = GetTime(db_asciirec.db_ctime);
	if (mask & USR_MTIME)
		db_binrec.usr_mtime = GetTime(db_asciirec.db_mtime);
	if (mask & USR_ARCTM)
		db_binrec.usr_arctm = GetTime(db_asciirec.db_arctm);
	if (mask & USR_RECTM)
		db_binrec.usr_rectm = GetTime(db_asciirec.db_rectm);
	if (mask & USR_DELTM)
		db_binrec.usr_deltm = GetTime(db_asciirec.db_deltm);

	/*
	**	leave filename for caller to deal with
	*/
	db_binrec.usr_fname = NULL;

	/*
	**	skip VV part if possible
	*/
	if ((mask & (USR_VV|USR_VVNO|USR_FNO|USR_BOF|USR_VFLG|
			USR_LSEEK|USR_TDATA)) ==0)
		return RASHDB_OK;

	/*
	**	do VV
	*/
	if (mask & USR_VV) {
		strncpy(db_binrec.usr_vvname, db_asciirec.db_vvname,
					(size_t)VVNAMELEN); 
	}

	if (mask & USR_VVNO) {
		if (C2BIN(DecInt, &db_binrec.usr_vvno, db_asciirec.db_vvno)) 
			return EINVAL;
	}

	if (mask & USR_FNO) {
		if (C2BIN(DecInt, &db_binrec.usr_fno, db_asciirec.db_fno))
			return EINVAL;
	}

	if (mask & USR_BOF) {
		if (C2BIN(HexInt, &db_binrec.usr_bof,
					db_asciirec.db_bof))
			return EINVAL;
	}

	if (mask & USR_VFLG) {
		if (C2BIN(HexUshort, &db_binrec.usr_vflg, 
					db_asciirec.db_vflg))
			return EINVAL;
	}

	if (mask & USR_LSEEK) {
		if (C2BIN(HexOFF, &db_binrec.usr_lseek,
					db_asciirec.db_lseek))
			return EINVAL;
	}

	if (mask & USR_TDATA) {
		if (C2BIN(HexOFF, &db_binrec.usr_vvdata, 
					db_asciirec.db_vvdata))
			return EINVAL;
	}
	return RASHDB_OK;
}

/*
**	rashdb_bld_indexes()
**	Recreate the indexes of a data base.  This is only called
**	right after the open of the table so no seek is done.
*/
static int
rashdb_bld_indexes()
{
	static	char		id[] = "rashdb_bld_indexes";
	int			i, ret;
	int			badrec;
	int			deadrec = 0;
	int			nbadrec = 0, nfiles = 0;
	int			nline2 = 0, nline3 = 0;
	struct	stat		sb;
	time_t			start, end;
	char			prev;

	time(&start);

	/*
	**	get/check tbl info (again)
	*/
	if (fstat(db_tbl_fd, &sb) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "fstat %s", db_file[DBFILE_DATA]);
		return EIO;
	}
	if (sb.st_size % sizeof(db_asciirec)) {
		nLogMsg(EL_ERROR1, id, "%s: bad size", db_file[DBFILE_DATA]);
		return EINVAL;
	}

	/*
	** 	make bfid & vv indexes
	*/
	ret = rashdb_openidx(&db_bfid_idx,
		db_file[DBFILE_BFID_INDEX], DB_CREATE|DB_TRUNCATE);
	if (ret != 0) {
	        nLogMsg(EL_ERROR1, id,
	                "make_index %s: %s",
			db_file[DBFILE_BFID_INDEX],
			db_strerror(ret));
	        return EIO;
	}
	ret = rashdb_openidx(&db_vv_idx,
		db_file[DBFILE_VV_INDEX], DB_CREATE|DB_TRUNCATE);
	if (ret != 0) {
	        nLogMsg(EL_ERROR1, id,
	                "make_index %s: %s",
			db_file[DBFILE_VV_INDEX],
			db_strerror(ret));
	        return EIO;
	}

	/* 
	**	read all the records from the data file and add them 
	**		to the index file.  
	*/
	for (i=1, db_tbl_addr = 0, prev = '1';
			db_tbl_addr < sb.st_size;
				i++, prev = db_asciirec.db_nrec[0],
				db_tbl_addr += sizeof(db_asciirec)) {

		/*
		**	Must seek here because rashdb_delete_bfid can
		**	change file position.
		*/
		if (lseek(db_tbl_fd, db_tbl_addr, SEEK_SET) != db_tbl_addr) {
			nLogMsg(EL_ERROR1|EL_SYS, id, "lseek");
			return EIO;
		}
		/*
		**	load record (using db_asciirec for aux recs
		**		since these are skipped anyway)
		*/
		if (read_rec(&db_asciirec) != RASHDB_OK) {
			nLogMsg(EL_ERROR1, id, "reading 'line' %d", i);
			return EIO;
		}

		/*
		**	act on 'nrec' part
		*/
		badrec = 0;
		switch (db_asciirec.db_nrec[0]) {

		    case '1':
			nfiles++;
			if (check_rec())
				badrec = 1;
			break;

		    case '2':
			if (prev == '1') {
				nline2++;
				continue;
			}
			badrec = 1;
			break;

		    case '3':
			if (prev == '2') {
				nline3++;
				continue;
			}
			badrec = 1;
			break;

		    default:
			badrec = 1;
			break;
		}
		if (badrec) {
			nLogMsg(EL_ERROR1, id, "record %d: bad nrec[0]: [%c]",
						i, db_asciirec.db_nrec[0]);
			nbadrec++;
			continue;
		}

		/*
		**	Got nrec[0] == '1' so this is a usr_fileinfo.
		*/

		/*
		**	Now convert all the fields (to make sure
		**		record is ok).
		*/
		if (rashdb_asc_to_bin(USR_ALL)) {
			nLogMsg(EL_ERROR1, id, "record %d", i);
			nbadrec++;
			continue;
		}
		if (db_binrec.usr_rflg & (RFLG_DELETED|RFLG_NOIDX))
			deadrec++;

		/*
		**	index the record - even if 'dead', since
		**		deleted bfid may still be on 'live' vv
		**		AND especially if it's last on the vv,
		**		and the vv is not finished, we need to 
		**		be able to find it to check when
		**		appending to vv.
		*/
		if (index_record(db_tbl_addr)) {
			nLogMsg(EL_ERROR1, id, "record %d", i);
	        	return EIO;
		}
		if (i % 10000 == 0)
			nLogMsg(EL_DEBUG1, id, "%d records processed", i);
	}

	time(&end);
	nLogMsg(EL_STATUS1, id, 
		"indexes rebuilt;  %d files  %d aux1  %d aux2  %d bad  time %d",
		nfiles, nline2, nline3, nbadrec, end - start);
	return RASHDB_OK;
}

/*
**	Open a user's rash data base using a default DBPATH.
*/
int
rashdb_open(char *username, uint access)
{
	static	char	id[] = "rashdb_open";
	ssize_t		oldlock_exist;
	size_t		len;
	int		i;
	int		ret, oflag, lktype;
	int		currlock;
	int		tbl_exists, tbl_copy_exists, build_indexes;
	union {
		int	val;
		char	buf[20];
	} old;

	assert( username != NULL);
	assert( access == O_DB_READ ||
		access == O_DB_RDWR ||
		access == O_DB_COPY);
#if	!defined(lint)
	assert( db_status == 0);
	assert( sizeof(db_asciirec) == 512);
	assert( sizeof(db_auxrec) == 512);
#endif

	nLogMsg(EL_DEBUG1, id, "%s type %d", username, access);

	/* set db_dir */
	strcpy(db_dir, DBPATH);
	len = strlen(db_dir);
	db_dir[len-1] = username[0];	/* first char of username */
	if (mkdir(db_dir, 0755) == -1) {
		if (errno != EEXIST) {
			nLogMsg(EL_ERROR1|EL_SYS, id, "stat %s", db_dir);
			return RASHDB_ERROR;
		}
	}
	else
		nLogMsg(EL_STATUS1, id, "CREATED DBDIR (%s)", db_dir);

	db_dir[len++] = '/';
	strcpy(&db_dir[len], username);
	if (mkdir(db_dir, 0755) == -1) {
		if (errno != EEXIST) {
			nLogMsg(EL_ERROR1|EL_SYS, id, "stat %s", db_dir);
			return RASHDB_ERROR;
		}
	}
	else
		nLogMsg(EL_STATUS1, id, "CREATED USRDIR (%s)", db_dir);

	/* set db_username */
	strcpy(db_username, username);

	/*
	**	generate file names
	*/
	for (i=0; i<DBFILE_NUM; i++) {
		sprintf(db_file[i], "%s/%s.%s", db_dir, db_username,
				db_file_suffix[i]);
	}

	/*
	**	open status file
	*/
	db_lock_fd = open(db_file[DBFILE_STATUS], O_RDWR|O_CREAT, 0600);
	if (db_lock_fd  == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "open %s",
			db_file[DBFILE_STATUS]);
		return RASHDB_ERROR;
	}

	/*
	**	Lock and read status file to see if any cleanup needs to
	**	be done.  Opening with O_DB_READ gives a shared lock.
	**	The only change that is allowed with this lock is to
	**	write O_DB_READ into the status file.  Moving, removing or
	**	creating files means coming back here to get a write
	**	lock.
	*/
	lktype = (access == O_DB_READ) ? LOCK_SH : LOCK_EX;
 dolock:
	if (flock(db_lock_fd, lktype) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "flock %s",
			db_file[DBFILE_STATUS]);
		goto bad;
	}
	if (lseek(db_lock_fd, (off_t)0, SEEK_SET) != 0) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "lseek %s",
			db_file[DBFILE_STATUS]);
		goto bad;
	}
	old.val = -1;
	if ((oldlock_exist = read(db_lock_fd, old.buf, sizeof(old))) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "read %s",
			db_file[DBFILE_STATUS]);
		goto bad;
	}
	if (oldlock_exist) {
		if ( oldlock_exist != sizeof(old.val)) {
			nLogMsg(EL_ERROR1, id, "read %d from %s, expected %d",
				oldlock_exist, db_file[DBFILE_STATUS],
				sizeof(old.val));
			goto bad;
		}
		if (old.val == O_DB_READ) {
			nLogMsg(EL_DEBUG1, id, "%s already open for read",
				username);
		}
		else if (old.val == O_DB_RDWR ||
			 old.val == O_DB_COPY) {
			nLogMsg(EL_STATUS1, id,
				"open %s for %d:%d not closed, state %d",
				username, access, lktype, old.val);
		}
		else {
			nLogMsg(EL_ERROR1, id, 
				 "%s: bad previous open state %d", old.val);
			goto bad;
		}
	}

	/*
	**	See if db table exists and db table copy exists
	**	(left over from a db rebuild), just for safety's sake.
	*/
	if (table_bad(db_file[DBFILE_DATA], &tbl_exists))
		goto bad;
	if (table_bad(db_file[DBFILE_DATA_COPY], &tbl_copy_exists))
		goto bad;
	build_indexes = 0;
	currlock = old.val;

	/*
	**	Check to see if the database was open for write.
	*/
	if (oldlock_exist && old.val != O_DB_READ) {
		/*
		**	Writable status existed from last rashdb_open()
		*/
		if (tbl_exists) {
			/*
			**	Get rid of any copy files.
			*/
			if (lktype == LOCK_SH) {	/* upgrade lock */
				lktype = LOCK_EX;
				goto dolock;
			}
			for (i=DBFILE_DATA_COPY; i<=DBFILE_VV_COPY; i++)
				(void)unlink(db_file[i]);
			build_indexes = 1;
		}
		else if (tbl_copy_exists) {
			/*
			**	No primary table but a copy exists.
			**	If we don't show that the db was left
			**	open for copy, get somebody to look.
			*/
			if (old.val != O_DB_COPY) {
				nLogMsg(EL_ERROR1, id, 
					"bad state for %s", db_dir);
				goto bad;
			}
			if (lktype == LOCK_SH) {	/* upgrade lock */
				lktype = LOCK_EX;
				goto dolock;
			}
			/*
			**	Rename table copy back to primary,
			**	remove all index copys and rebuild.
			*/
			nLogMsg(EL_STATUS1, id,
				"renaming copy in %s", db_dir);
			if (rename(db_file[DBFILE_DATA_COPY],
				db_file[DBFILE_DATA]) == -1) {
				nLogMsg(EL_ERROR1|EL_SYS, id,
					"rename(%s->%s)", 
					db_file[DBFILE_DATA_COPY],
					db_file[DBFILE_DATA]);
				goto bad;
			}
			for (i=DBFILE_BFID_COPY; i<=DBFILE_VV_COPY; i++)
				(void)unlink(db_file[i]);
			/*
			**	Since the rename worked, we've
			**	got a table for sure now.
			*/
			tbl_exists = 1;
			build_indexes = 1;
		}
		else {
			/*
			**	No table exists, better have a real
			**	person take a look.
			*/
			nLogMsg(EL_ERROR1, id, 
				"no primary or copy in %s", db_dir);
			goto bad;
		}
	}
	else if (lktype == LOCK_EX) {
		/*
		**	The database was not open or open for read
		**	and we are getting an exclusive lock to
		**	do something.
		**	We need to put O_DB_RDWR into the lock
		**	file so it will force another rebuild
		**	if we crash.
		*/
		currlock = (access == O_DB_READ) ? O_DB_RDWR : access;
		if (oldlock_exist &&
				lseek(db_lock_fd, (off_t)0, SEEK_SET) !=
					(off_t)0) {
			nLogMsg(EL_ERROR1|EL_SYS, id, "lseek %s",
				db_file[DBFILE_STATUS]);
			goto bad;
		}
		if (write(db_lock_fd, &currlock, sizeof(currlock)) !=
				sizeof(currlock)) {
			nLogMsg(EL_ERROR1|EL_SYS, id, "write %s",
				db_file[DBFILE_STATUS]);
			goto bad;
		}
		(void)fsync(db_lock_fd);
	}

	/*
	**	set up open(table) flags; and if the table is 
	**		being created, it should be a one-time
	**		thing, so log the fact for easy checking
	*/
	oflag = ((lktype == LOCK_EX) ? O_RDWR : O_RDONLY)|O_SYNC;
	if (!tbl_exists) {
		if (lktype == LOCK_SH) {		/* upgrade lock */
			lktype = LOCK_EX;
			goto dolock;
		}
		build_indexes = 1;
		oflag |= O_CREAT;
		nLogMsg(EL_STATUS1, id, "CREATING TABLE for %s", username);
	}

	/*
	**	open table
	*/
	if ((db_tbl_fd = open(db_file[DBFILE_DATA], oflag, 0600)) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "open %s", db_file[DBFILE_DATA]);
		goto bad;
	}
	db_fd = db_tbl_fd;	/* the default table */

	/*
	**	If this is an ordinary open (not a copy),
	**	open (or build) indexes.
	*/
	if (access != O_DB_COPY) {
		if (!build_indexes) {
			/*
			**	Try to open indexes to see if we need to
			**	rebuild them.
			*/
			uint	flags = (access == O_DB_READ) ?  DB_RDONLY : 0;

			build_indexes = 1;
			ret = rashdb_openidx(&db_bfid_idx,
				db_file[DBFILE_BFID_INDEX], flags);
			if (ret == 0) {
				ret = rashdb_openidx(&db_vv_idx,
					db_file[DBFILE_VV_INDEX], flags);
				if (ret == 0)
					build_indexes = 0;
			}
			if (build_indexes) {    
				if (db_bfid_idx) {
					db_bfid_idx->close(db_bfid_idx, 0);
					db_bfid_idx = NULL;
				}
				if (db_vv_idx) {
					db_vv_idx->close(db_vv_idx, 0);
					db_vv_idx = NULL;
				}
			}
		}
		if (build_indexes) {
			if (lktype == LOCK_SH) {	/* upgrade lock */
				lktype = LOCK_EX;
				goto dolock;
			}
			nLogMsg(EL_STATUS1, id,
				"BUILDING INDEXES for %s", username);
			if (rashdb_bld_indexes())
				goto bad;
		}
	} else {
		/*
		**	This is a copy so just make new indexes and
		**	create a new table.  The primary indexes need
		**	not be rebuilt even if they are junk because
		**	we don't use them for this open and they will
		**	be thrown away if this open aborts.
		**
		**	We must have a write lock
		**	here since the open is O_DB_COPY.
		*/
		db_cpy_fd = open(db_file[DBFILE_DATA_COPY],
				O_RDWR|O_CREAT|O_EXCL, 0600);
		if (db_cpy_fd == -1) {
			nLogMsg(EL_ERROR1|EL_SYS, id, "open %s",
				db_file[DBFILE_VV_COPY]);
			goto bad;
		}
		ret = rashdb_openidx(&db_bfid_idx,
			db_file[DBFILE_BFID_COPY], DB_CREATE|DB_TRUNCATE);
		if (ret != 0) {
	        	nLogMsg(EL_ERROR1, id,
	                	"make_index %s: %s",
				db_file[DBFILE_BFID_COPY],
				db_strerror(ret));
	        	goto bad;
		}
		ret = rashdb_openidx(&db_vv_idx,
			db_file[DBFILE_VV_COPY], DB_CREATE|DB_TRUNCATE);
		if (ret != 0) {
	        	nLogMsg(EL_ERROR1, id,
	                	"make_index %s: %s",
				db_file[DBFILE_VV_COPY],
				db_strerror(ret));
	        	goto bad;
		}
	}

	if (lktype == LOCK_EX && access == O_DB_READ) {
		/*
		**	We did an open for read and had to
		**	upgrade the lock.  Here we close everything
		**	and downgrade back to a shared lock.
		*/
		rashdb_close_files();
		if (ftruncate(db_lock_fd, (off_t)0) == -1) {
			nLogMsg(EL_ERROR1|EL_SYS, id, "truncate %s",
				db_file[DBFILE_STATUS]);
			goto bad;
		}
		lktype = LOCK_SH;
		goto dolock;
	}
	/*
	**	Now that things have been opened (cleaned up if necc), 
	**	put our status into the status file.
	*/
	if (currlock != access) {	/* write status file */
		if (lseek(db_lock_fd, (off_t)0, SEEK_SET) != (off_t)0) {
			nLogMsg(EL_ERROR1|EL_SYS, id, "lseek %s",
				db_file[DBFILE_STATUS]);
			goto bad;
		}
		if (write(db_lock_fd, &access, sizeof(access)) !=
				sizeof(access)) {
			nLogMsg(EL_ERROR1|EL_SYS, id, "write %s",
				db_file[DBFILE_STATUS]);
			goto bad;
		}
		(void)fsync(db_lock_fd);
	}

	db_status = access;
	db_last_addr = -1;
	return RASHDB_OK;

 bad:
	/*
	**	Close everything without calling rashdb_unlock which
	**	would truncate the status file.
	*/
	rashdb_close_files();
	close(db_lock_fd);
	return RASHDB_ERROR;
}

/*
**	Write a record at the current file location.
**	Do not update db_tbl_addr.
*/
static int
write_rec(void *rec)
{
	static char	id[] = "write_rec";
	ssize_t		ret;

	ret = write(db_fd, rec, sizeof(db_asciirec));
	if (ret == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "fd %d", db_fd);
		return EIO;
	}
	if (ret != sizeof(db_asciirec)) {
		nLogMsg(EL_ERROR1, id, "bad size %d/%d",
						ret, sizeof(db_asciirec));
		return RASHDB_ERROR;
	}
	return RASHDB_OK;
}

char	*
bfidout(bfid_t *bfidP)
{
	static	char	out[BFIDLEN+1];

	Bin2Hex(out, (size_t)BFIDLEN, bfidP, sizeof(bfid_t));
	out[BFIDLEN] = '\0';
	return out;
}

/*
**	Handle filename for the current asciirec, leaving db_tbl_addr 
**	pointing at the next record.  If requested, extract the filename 
**	from the asciirec and any followon auxrecs and set the binrec's 
**	filename pointer, otherwise just check the auxrecs if any.
*/
static int
rashdb_handle_fname(uint convert)
{
	static char	*id = "rashdb_handle_fname";
	char		*cp;
	int		vvno;

	/*
	**	We need the vv number to see if any filename (aux)
	**	records are to be read.
	*/
	if (C2BIN(DecInt, &vvno, db_asciirec.db_vvno)) 
		return EINVAL;
	if (vvno > 0 && convert) {
		if (memcmp(db_fname_bfid, db_asciirec.db_bfid,
				(size_t)BFIDLEN) == 0) {
			db_binrec.usr_fname = db_usr_fname;
			return RASHDB_OK;
		}
	}

	if (convert) {
		/*
		**	Convert filename for binary record too
		*/
		memcpy(db_usr_fname, db_asciirec.db_fname, 
				sizeof(db_asciirec.db_fname) + 1);
		cp = strchr(db_usr_fname, ' ');
		memcpy(db_fname_bfid, db_asciirec.db_bfid, (size_t)BFIDLEN);
	}
	if (vvno == 0 && db_asciirec.db_nrec[2] != '1') {
		if (read_rec(&db_auxrec) != RASHDB_OK) {
			nLogMsg(EL_ERROR1, id, "reading aux 1");
			return EIO;
		}
		db_tbl_addr += sizeof(db_asciirec);	/* bump db_tbl_addr */
		if (db_auxrec.end1 != '|'  ||  db_auxrec.end2 != '\n') {
			nLogMsg(EL_ERROR1, id, "auxrec 1 end-of-rec bad");
			return EINVAL;
		}
		if (db_auxrec.db_nrec[2] != db_asciirec.db_nrec[2]  ||
		    db_auxrec.db_nrec[1] != '/'  ||
		    db_auxrec.db_nrec[0] != '2') {
			nLogMsg(EL_ERROR1, id, "auxrec 1 nrec bad [%.*s]",
					sizeof(db_auxrec), db_auxrec.db_nrec);
			return EINVAL;
		}
		if (convert) {
			/* set end of field to ' ' for string op convenience */
			db_auxrec.end1 = ' ';
			memcpy(cp, db_auxrec.db_aux, 
					sizeof(db_auxrec.db_aux) + 1);
			cp = strchr(cp, ' ');
		}
		if (db_asciirec.db_nrec[2] == '3') {
			if (read_rec(&db_auxrec) != RASHDB_OK) {
				nLogMsg(EL_ERROR1, id, "reading aux 2");
				return EIO;
			}
			db_tbl_addr += sizeof(db_asciirec);
			if (db_auxrec.end1 != '|' || db_auxrec.end2 != '\n') {
				nLogMsg(EL_ERROR1, id, 
						"auxrec 2 end-of-rec bad");
				return EINVAL;
			}
			if (db_auxrec.db_nrec[2] != db_asciirec.db_nrec[2]  ||
			    db_auxrec.db_nrec[1] != '/'  ||
			    db_auxrec.db_nrec[0] != '3') {
				nLogMsg(EL_ERROR1, id, 
					"auxrec 2 nrec bad [%.*s]",
					sizeof(db_auxrec), db_auxrec.db_nrec);
				return EINVAL;
			}
			if (convert) {
				db_auxrec.end1 = ' ';
				memcpy(cp, db_auxrec.db_aux, 
						sizeof(db_auxrec.db_aux) + 1);
				cp = strchr(cp, ' ');
			}
		}
	}
	if (convert) {
		*cp = '\0';
		db_binrec.usr_fname = db_usr_fname;
	}
	return RASHDB_OK;
}

/*
**	Remove any index for this bfid.  If an entry exists, mark the
**	table record "deleted".
**	Does not change db_tbl_addr.
*/
int
rashdb_delete_bfid(bfid_t *bfidP)
{
	static char	id[] = "rashdb_delete_bfid";
	DBT		key;
	DBT		data;
	bfid_t		recbfid;
	int		rc, rr;
	unsigned short	usr_rflg;
	off_t		savepos;
	time_t		now, arctm, checktm;

	PTR_DBT(key, bfidP);
	ZERO_OBJ(data);
	if ((rc = db_bfid_idx->get(db_bfid_idx, NULL, &key, &data, 0)) != 0)
		return (rc == DB_NOTFOUND ? RASHDB_OK : RASHDB_ERROR);

	/*
	**	Save current position in file and set db_tbl_addr
	**	to this bfid's recptr so the read routines can work.
	*/
	savepos = db_tbl_addr;
	COPY_DBT(db_tbl_addr, data);
	rc = RASHDB_ERROR;

	/*
	**	Delete old index
	*/
	if ((rr = db_bfid_idx->del(db_bfid_idx, NULL, &key, 0)) != 0)
		nLogMsg(EL_ERROR1, id, "db delete: %s", db_strerror(rr));
	/*
	**	Load table record
	*/
	if (lseek(db_fd, db_tbl_addr, SEEK_SET) != db_tbl_addr) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "lseek 1");
		goto done;
	}
	if (read_rec(&db_asciirec) != RASHDB_OK)
		goto done;
	if (check_rec())
		goto done;
	/*
	**	Sanity check
	*/
	if (Hex2Bin(&recbfid, sizeof(recbfid), db_asciirec.db_bfid,
			(size_t)BFIDLEN))
		goto done;
	if (memcmp(&recbfid, bfidP, sizeof(bfid_t))) {
		char	ent[64];

		strcpy(ent, bfidout(bfidP));
		nLogMsg(EL_ERROR1, id, "bfid %s doesn't match %s", 
					ent, bfidout(&recbfid));
		goto done;
	}
	/*
	**	Remember the archive time for this sequence.
	*/
	arctm = GetTime(db_asciirec.db_arctm);

	/*
	**	Convert flag to binary, add 'deleted' flag,
	**		and convert back to hex in asciirec.
	*/
	if (C2BIN(HexUshort, &usr_rflg, db_asciirec.db_rflg))
		goto done;

	now = time(NULL);
	if (usr_rflg & RFLG_DELETED) {
		nLogMsg(EL_ERROR1, id, "bfid %s pos %lld already deleted",
			bfidout(bfidP), db_tbl_addr);
	}
	else {
		/*
		**	Set flag and put in the date of deletion.
		*/
		usr_rflg |= RFLG_DELETED;
		if (C2ASC(SetHexUshort, db_asciirec.db_rflg, usr_rflg))
			goto done;
		SetTime(db_asciirec.db_deltm, now);

		/*
		**	write back table record
		*/
		if (lseek(db_fd, db_tbl_addr, SEEK_SET) != db_tbl_addr) {
			nLogMsg(EL_ERROR1|EL_SYS, id, "lseek 2");
			goto done;
		}
		if (write_rec(&db_asciirec))
			goto done;
	}
	/*
	**	Read past any auxrecs and update secondary records.
	**	This uses db_file_addr which is why we set it above.
	*/
	db_tbl_addr += sizeof(db_asciirec);
	if (rashdb_handle_fname(0))
		goto done;
	for (;; db_tbl_addr += sizeof(db_asciirec)) {
		rr = read_rec(&db_asciirec);
		if (rr == RASHDB_EOF)
			break;
		if (rr != RASHDB_OK)
			goto done;
		if (check_rec())
			goto done;
		/*
		**	Sanity check
		*/
		if (Hex2Bin(&recbfid, sizeof(recbfid),
				db_asciirec.db_bfid, (size_t)BFIDLEN))
			goto done;
		if (memcmp(&recbfid, bfidP, sizeof(bfid_t)))
			break;
		checktm = GetTime(db_asciirec.db_arctm);
		if (checktm != arctm)
			break;

		/*
		**	Convert flag to binary, add 'deleted' flag,
		**		and convert back to hex in asciirec.
		*/
		if (C2BIN(HexUshort, &usr_rflg, db_asciirec.db_rflg))
			goto done;
		if (usr_rflg & RFLG_DELETED) {
			nLogMsg(EL_ERROR1, id,
				"bfid %s secondary already deleted",
				bfidout(bfidP));
			continue;
		}

		/*
		**	Set flag and put in the date of deletion.
		*/
		usr_rflg |= RFLG_DELETED;
		if (C2ASC(SetHexUshort, db_asciirec.db_rflg, usr_rflg))
			goto done;
		SetTime(db_asciirec.db_deltm, now);

		/*
		**	write back table record
		*/
		if (lseek(db_fd, db_tbl_addr, SEEK_SET) != db_tbl_addr) {
			nLogMsg(EL_ERROR1|EL_SYS, id, "lseek 2");
			goto done;
		}
		if (write_rec(&db_asciirec))
			goto done;
	}
	rc = RASHDB_OK;

 done:
	db_tbl_addr = savepos;
	return rc;
}

/*
**	Basic read function to get entire file record, modulo mask.  
**		Seek to db_tbl_addr, read, and convert the table 
**		entries to binary.
*/
static int
rashdb_read(uint mask)
{
	static char	id[] = "rashdb_read";
	int		ret;

	if (db_tbl_addr % sizeof(db_asciirec)) {
		nLogMsg(EL_ERROR1, id, "bad table size");
		return RASHDB_ERROR;
	}

	db_last_addr = db_tbl_addr;
	if (lseek(db_tbl_fd, db_tbl_addr, SEEK_SET) != db_tbl_addr) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "lseek");
		return RASHDB_ERROR;
	}

	if ((ret = read_rec(&db_asciirec)) != RASHDB_OK)
		return ret;
	if (check_rec())
		return RASHDB_ERROR;

	/*
	**	Convert asciirec to binary
	*/
	if (rashdb_asc_to_bin(mask))
		return RASHDB_ERROR;

	/*
	**	Get filename if req'd (including aux records if they exist)
	*/
	if (rashdb_handle_fname(mask & USR_FNAME))
		return RASHDB_ERROR;

	/*
	**	Set db_tbl_addr to point to next record
	*/
	db_tbl_addr += sizeof(db_asciirec);
	return RASHDB_OK;
}

/*
**	Position to beginning of table
*/
void
rashdb_firstrec(void)
{
	db_tbl_addr = 0;
}

/*
**	External read function to return error status
*/
USR_FILEINFO *
rashdb_nextrec(int *err, uint mask)
{
	db_fd = db_tbl_fd;

	*err = rashdb_read(mask);
	if (*err == RASHDB_OK)
		return &db_binrec;
	return NULL;
}

/*
**	db_add()
**	Convert binary struct to table ASCII format, append to table,
**	and index for bfid and vv.
*/
int
rashdb_add(USR_FILEINFO *uP)
{
	static	char	id[] = "db_add";
	size_t		lfname;
	int		bad = 0, i, j;
	off_t		addr;

	if (db_status == O_DB_READ) {
		nLogMsg(EL_ERROR1, id, "db not opened for write");
		return EINVAL;
	}
	else if (db_status == O_DB_COPY)
		db_fd = db_cpy_fd;
	else
		db_fd = db_tbl_fd;

	/*
	**  check input
	*/
	if (!uP->usr_fname) {
		nLogMsg(EL_ERROR1, id, "usr_fname passed is NULL");
		bad++;
	}
	if ( (lfname = strlen(uP->usr_fname)) == 0 ) {
		nLogMsg(EL_ERROR1, id, "Invalid file name specified");
		bad++;
	}
	if (uP->usr_vvno > 999) {
		nLogMsg(EL_ERROR1, id, "usr_vvno exceeds 999");
		bad++;
	}
	if ( uP->usr_fsize < (off_t) 0 ) {
		nLogMsg(EL_ERROR1, id, "usr_fsize is less than 0");
		bad++;
	}
	if (uP->usr_uid > 99999) {
		nLogMsg(EL_ERROR1, id, "usr_uid  exceeds 99999");
		bad++;
	}
	if (uP->usr_gid > 99999) {
		nLogMsg(EL_ERROR1, id, "usr_gid  exceeds 99999");
		bad++;
	}
	if (lfname > PATH_MAX) {
		nLogMsg(EL_ERROR1, id,
			"usr_fname is longer than PATH_MAX (%d) characters",
			PATH_MAX);
		bad++;
	}
	if (uP->usr_fno < 0) {
		nLogMsg(EL_ERROR1, id, "usr_fno < 0");
		bad++;
	}
	if (uP->usr_fno > MAXTAPEFILES) {
		nLogMsg(EL_ERROR1, id, "usr_fno exceeds MAXTAPEFILES");
		bad++;
	}
	if (uP->usr_vvname[0] == '\0') {
		nLogMsg(EL_ERROR1, id, "usr_vvname is null");
		bad++;
	}
	j = 0;
	for (i=0; i<VVNAMELEN; i++) {
		char	c = uP->usr_vvname[i];

		if (c == ' '  ||  c == '\0')
			break;
		if (c == '.') {
			j++;
			continue;
		}
		if (!isalnum(c)) {
			nLogMsg(EL_ERROR1, id, "bad char [%c] in vv name", c);
			bad++;
		}
	}
	if (j != 1) {
		nLogMsg(EL_ERROR1, id, "VV name needs 1, has %d '.'", j);
		bad++;
	}

	if (bad)
		return EINVAL;

	if (uP->usr_version == 0)
		uP->usr_version = RASH_VERSION;
	else if (check_version(uP->usr_version, id))
		return EINVAL;

	/*
	**	copy the user's record to local 
	**		static for simple/consistent indexing
	*/
	if (&db_binrec != uP)
		memcpy(&db_binrec, uP, sizeof(db_binrec));

	/*
	**  convert the fields to ascii in the global record
	*/
	memset(&db_asciirec, ' ', sizeof(db_asciirec));
	db_asciirec.end1 = '|';
	db_asciirec.end2 = '\n';

	/* version */
	if (C2ASC(SetDecInt, db_asciirec.db_version, uP->usr_version))
		return EINVAL;

	/* db time=now */
	SetTime(db_asciirec.db_rectm, time( (time_t *) 0));

	/* VV */
	memcpy(db_asciirec.db_vvname, uP->usr_vvname, (size_t)VVNAMELEN);
	if (C2ASC(SetHexUshort, db_asciirec.db_vflg, uP->usr_vflg))
		return EINVAL;
	if (C2ASC(SetHexInt, db_asciirec.db_bof, uP->usr_bof))
		return EINVAL;
	if (C2ASC(SetDecInt, db_asciirec.db_fno, uP->usr_fno))
		return EINVAL;
	if (C2ASC(SetHexOFF, db_asciirec.db_lseek, uP->usr_lseek))
		return EINVAL;
	if (C2ASC(SetHexOFF, db_asciirec.db_vvdata, uP->usr_vvdata))
		return EINVAL;

	/* file info */
	if (Bin2Hex(db_asciirec.db_bfid, (size_t)BFIDLEN,
			&uP->usr_bfid, sizeof(bfid_t)))
		return EINVAL;

	if (C2ASC(SetHexUshort, db_asciirec.db_rflg, uP->usr_rflg))
		return EINVAL;
	if (C2ASC(SetDecInt, db_asciirec.db_vvno, uP->usr_vvno))
		return EINVAL;
	if (C2ASC(SetHexOFF, db_asciirec.db_fsize, uP->usr_fsize))
		return EINVAL;
	if (C2ASC(SetDecInt, db_asciirec.db_uid, (int)uP->usr_uid))
		return EINVAL;
	if (C2ASC(SetDecInt, db_asciirec.db_gid, (int)uP->usr_gid))
		return EINVAL;
	if (C2ASC(SetHexInt, db_asciirec.db_mode, (uint)uP->usr_mode))
		return EINVAL;
	if (C2ASC(SetDecInt, db_asciirec.db_group_no, uP->usr_group_no))
		return EINVAL;

	SetTime(db_asciirec.db_ctime, uP->usr_ctime);
	SetTime(db_asciirec.db_mtime, uP->usr_mtime);
	SetTime(db_asciirec.db_arctm, uP->usr_arctm);

	/*
	**	consider whether a long fname means it's an extended record
	*/
	db_asciirec.db_nrec[0] = '1';
	db_asciirec.db_nrec[1] = '/';
	lfname = strlen(uP->usr_fname);
	if (lfname <= sizeof(db_asciirec.db_fname)) {
		/*
		**	no need to continue pathname 
		**	across multiple records
		*/
		db_asciirec.db_nrec[2] = '1';
		memcpy(db_asciirec.db_fname, uP->usr_fname, lfname);
	} else {
		/*
		**	for vv's after the 1st, the longer
		**	fname isn't included, so the 1st
		**	record is marked 't' (e.g. "1t3" 
		**	instead of "1/3") to indicate that
		**	the filename is truncated.
		*/
		if (uP->usr_vvno > 0)
			db_asciirec.db_nrec[1] = 't';

		if (lfname <= sizeof(db_asciirec.db_fname) + 
			      sizeof(db_auxrec.db_aux))
			db_asciirec.db_nrec[2] = '2';
		else if (lfname <= sizeof(db_asciirec.db_fname) + 
			      sizeof(db_auxrec.db_aux) * 2)
			db_asciirec.db_nrec[2] = '3';
		else {
			nLogMsg(EL_ERROR1, id, "pathname %s too long", 
								uP->usr_fname);
			return EINVAL;
		}
		memcpy(db_asciirec.db_fname, uP->usr_fname,
					sizeof(db_asciirec.db_fname));
	}

	/*
	**	seek to end-of-table where the record will go
	**		and check tbl size while we're at it
	*/
	if ((addr = lseek(db_fd, (off_t)0, SEEK_END)) == (off_t)-1) {
		nLogMsg(EL_ERROR1|EL_SYS, id,
			"lseek(%s)", db_file[DBFILE_DATA]);
		return EIO;
	}
	if (addr % sizeof(db_asciirec)) {
		nLogMsg(EL_ERROR1, id, "bad table size");
		return EIO;
	}

	/*
	**	write the record
	*/
	if (write_rec(&db_asciirec))
		return EIO;

	/*
	**	if file name longer than 1st rec && this is 1st VV,
	**		write rest of filename in auxiliary record(s)
	*/
	if (db_asciirec.db_nrec[2] != '1'  &&  uP->usr_vvno == 0) {
		/*
		**	write auxiliary record(s) for rest of file name
		*/
		memset(&db_auxrec, ' ', sizeof(db_auxrec));
		db_auxrec.end1 = '|';
		db_auxrec.end2 = '\n';
		db_auxrec.db_nrec[2] = db_asciirec.db_nrec[2];
		db_auxrec.db_nrec[1] = '/';
		db_auxrec.db_nrec[0] = '2';
		memcpy(db_auxrec.db_aux,
			uP->usr_fname + sizeof(db_asciirec.db_fname),
			MIN(sizeof(db_auxrec.db_aux),
				lfname - sizeof(db_asciirec.db_fname)));
		/*
		**	(no need to lseek since continuing to write)
		*/
		if (write_rec(&db_auxrec))
			return EIO;
		if (db_asciirec.db_nrec[2] == '3') {
			/*
			**	write last record
			*/
			db_auxrec.db_nrec[0] = '3';
			memset(db_auxrec.db_aux, ' ', sizeof(db_auxrec.db_aux));
			memcpy(db_auxrec.db_aux, 
			       uP->usr_fname + sizeof(db_asciirec.db_fname) + 
					      sizeof(db_auxrec.db_aux),
			       lfname - sizeof(db_asciirec.db_fname) -
					      sizeof(db_auxrec.db_aux));
			/*
			**	(no need to lseek since continuing to write)
			*/
			if (write_rec(&db_auxrec))
				return EIO;
		}
	}

	/*
	**	index the tbl entry
	*/
	if (index_record(addr))
		return EIO;

	return RASHDB_OK;
}


USR_FILEINFO *
rashdb_get_bfid(int *err, bfid_t *bfidP)
{
	static char	id[] = "rashdb_get_bfid ";
	int		rc;
	DBT		key;
	DBT		data;

	nLogMsg(EL_DEBUG3, id, "Entered to read bfid: %s", bfidP);
	assert(db_status != O_DB_COPY);

	/*
	**	Look up in index
	*/
	PTR_DBT(key, bfidP);
	ZERO_OBJ(data);
	rc = db_bfid_idx->get(db_bfid_idx, NULL, &key, &data, 0);
	if (rc == DB_NOTFOUND) {
		nLogMsg(EL_DEBUG3, id, "bfid %s not found", bfidout(bfidP));
		*err = RASHDB_OK;
		return NULL;
	}
	else if (rc != 0) {
		nLogMsg(EL_ERROR1, id, "get bfid %s: %s",
			bfidout(bfidP), db_strerror(rc));
		*err = RASHDB_ERROR;
		return NULL;
	}
	COPY_DBT(db_tbl_addr, data);

	/*
	**	read table record
	*/
	if ((*err = rashdb_read(USR_ALL)) != RASHDB_OK) {
		nLogMsg(EL_ERROR1, id, "reading for bfid %s", bfidout(bfidP));
		return NULL;
	}

	/*
	**	Simple check
	*/
	if (memcmp(&db_binrec.usr_bfid, bfidP, sizeof(bfid_t))) {
		nLogMsg(EL_ERROR1, id, "bfid %s not found", bfidout(bfidP));
		*err = RASHDB_ERROR;
		return NULL;
	}
	return &db_binrec;
}

/*
**	db_next_bfidrec() - get next file record for this bfid
**		i.e. read next tbl record & see if it's for 
**		the same bfid
*/
USR_FILEINFO *
rashdb_next_bfidrec(int *err, bfid_t *bfidP)
{
	static char	id[] = "db_next_bfidrec";

	assert(db_status != O_DB_COPY);
	assert(memcmp(bfidP, &db_binrec.usr_bfid, sizeof(bfid_t)) == 0);

	/*
	**	point db_tbl_addr to next table record & load
	*/
	db_tbl_addr += sizeof(db_asciirec);
	*err = read_rec(&db_asciirec);
	if (*err != RASHDB_OK) {
		if (*err == RASHDB_EOF) {
			nLogMsg(EL_DEBUG2, id, "bfid %s not found",
				bfidout(bfidP));
		}
		else {
			nLogMsg(EL_ERROR1, id, "reading for %s",
				bfidout(bfidP));
		}
		return NULL;
	}
	if (check_rec()) {
		*err = RASHDB_ERROR;
		return NULL;
	}

	/*
	**	convert bfid to binary and check it
	*/
	if (rashdb_asc_to_bin(USR_BFID)) {
		nLogMsg(EL_ERROR1, id, "converting bfid for %s",
			bfidout(bfidP));
		*err = RASHDB_ERROR;
		return NULL;
	}
	if (memcmp(&db_binrec.usr_bfid, bfidP, sizeof(bfid_t)))
		return NULL;	/* no longer the same bfid */

	/*
	**	convert rest of record to binary
	*/
	if (rashdb_asc_to_bin(USR_ALL)) {
		nLogMsg(EL_ERROR1, id, "converting for %s", bfidout(bfidP));
		*err = RASHDB_ERROR;
		return NULL;
	}

	/*
	**	reset fname, which shouldn't change
	*/
	db_binrec.usr_fname = db_usr_fname;

	return &db_binrec;
}

/*
**	Common checks for get_vv and next_vv.
*/
static
USR_FILEINFO *
vvrec(int *err, char *id, char *vvname)
{
	int		fno;
	char		buf[512];

	/*
	**	get record from table
	*/
	if ((*err = rashdb_read(USR_ALL)) != RASHDB_OK)
		return NULL;

	/*
	**	check that vv_fileno agrees with index key
	*/
	if (strncmp(db_binrec.usr_vvname, vvname, (size_t)VVNAMELEN) != 0) {
		(void)vv_cursor->c_close(vv_cursor);
		vv_cursor = NULL;
		return NULL;
	}
	fno = MAXTAPEFILES - db_binrec.usr_fno;
	sprintf(buf, vv_keyfmt, vvname, fno);
	if (strlen(buf) != vvkey.size ||
			strncmp(buf, vvkey.data, (size_t)vvkey.size) != 0) {
		nLogMsg(EL_ERROR1, id, "%s != %*s",
			buf, vvkey.size, vvkey.data);
		*err = RASHDB_ERROR;
		(void)vv_cursor->c_close(vv_cursor);
		vv_cursor = NULL;
		return NULL;
	}
	return &db_binrec;
}

/*
**	db_get_vv() - look up last file on vv
*/
USR_FILEINFO *
rashdb_get_vv(int *err, char *vvname)
{
	static char	id[] = "db_get_vv";
	int		ret;

	assert(db_status != O_DB_COPY);
	/*
	**	Look up db record addr for the last file on vv.
	**	The vv part of the result is checked when checking
	**	the whole db record by vvrec(), i.e. if the vv is
	**	not in the dbase & locate_key() returns another vv,
	**	vvrec() detects this after loading the record. Since
	**	the record is expected to be there, this isn't a
	**	performance loss.
	*/
	if (vv_cursor != NULL) {
		(void)vv_cursor->c_close(vv_cursor);
		vv_cursor = NULL;
	}
	ret = db_vv_idx->cursor(db_vv_idx, NULL, &vv_cursor, 0);
	if (ret != 0) {
		nLogMsg(EL_ERROR1, id, "cursor %s: %s",
			vvname, db_strerror(ret));
		*err = RASHDB_ERROR;
		return NULL;
	}

	sprintf(vv_key, vv_keyfmt, vvname, 0);
	memset(&vvkey, 0, sizeof(vvkey));
	vvkey.data = vv_key;
	vvkey.size = strlen(vv_key);
	ZERO_OBJ(vvdata);

	ret = vv_cursor->c_get(vv_cursor, &vvkey, &vvdata, DB_SET_RANGE);
	if (ret != 0) {
		if (ret == DB_NOTFOUND)
			*err = RASHDB_OK;
		else {
			nLogMsg(EL_ERROR1, id,
				"locate_key %s: %s",
				vv_key, db_strerror(ret));
			*err = RASHDB_ERROR;
		}
		(void)vv_cursor->c_close(vv_cursor);
		vv_cursor = NULL;
		return NULL;
	}
	COPY_DBT(db_tbl_addr, vvdata);

	return (vvrec(err, id, vvname));
}

USR_FILEINFO *
rashdb_next_vvrec(int *err, char *vvname)
{
	static char	id[] = "db_next_vvrec";
	int		ret;

	assert(db_status != O_DB_COPY);

	if (vv_cursor == NULL) {
		nLogMsg(EL_ERROR1, id,
				"getvv not called for %s", vvname);
		*err = RASHDB_ERROR;
		return NULL;
	}

	ZERO_OBJ(vvkey);
	ZERO_OBJ(vvdata);
	ret = vv_cursor->c_get(vv_cursor, &vvkey, &vvdata, DB_NEXT);
	if (ret != 0) {
		if (ret == DB_NOTFOUND)
			*err = RASHDB_OK;
		else {
			nLogMsg(EL_ERROR1, id,
				"next_key %s: %s",
				vv_key, db_strerror(ret));
			*err = RASHDB_ERROR;
		}
		(void)vv_cursor->c_close(vv_cursor);
		vv_cursor = NULL;
		return NULL;
	}
	COPY_DBT(db_tbl_addr, vvdata);

	return (vvrec(err, id, vvname));
}

int
rashdb_set_flgs(unsigned short rflg, unsigned short vflg)
{
	static	char	id[] = "db_set_flgs";

	if (db_status != O_DB_RDWR) {
		nLogMsg(EL_ERROR1, id, "db not opened for write");
		return EINVAL;
	}
	if (db_last_addr % sizeof(db_asciirec)) {
		nLogMsg(EL_ERROR1, id, "bad table size");
		return EINVAL;
	}

	/*
	**	Make sure we are re-writting something resonable.
	*/
	if (check_rec())
		return RASHDB_ERROR;

	if (C2ASC(SetHexUshort, db_asciirec.db_rflg, rflg))
		return EINVAL;
	if (C2ASC(SetHexUshort, db_asciirec.db_vflg, vflg))
		return EINVAL;
	/*
	**	write back table record
	*/
	if (lseek(db_fd, db_last_addr, SEEK_SET) != db_last_addr) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "lseek");
		return RASHDB_ERROR;
	}
	if (write_rec(&db_asciirec))
		return RASHDB_ERROR;

	return RASHDB_OK;
}
