/*
 *  rashvv.c - archive a list of files to virtual volumes (VV's)
 *
 *  Bill Ross & Tom Proett, NASA Ames, July 1998
 *
 *  These functions are unencumbered by non-NASA proprietary information.
 */

#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/param.h>
#include <miscfs/dmfs/rash.h>
#include <miscfs/dmfs/dmfs_fcntl.h>
#include <miscfs/dmfs/dmfs.h>

#include "vv_cl.h"
#include "dblib.h"
#include "nas_logmsg.h"
#include "nasutil.h"
#include "rashcom.h"
#include "labels.h"
#include "bfid.h"
#include "dmfsd.h"

extern	char	*com;		/* command name 		 */
extern	int	logdest;	/* where log messages go 	 */
extern	int	debuglevel;

int	oktodie = TRUE;		/* set to FALSE in critical section */
int	dienow = FALSE;		/* test of exit from critical section */

off_t	leave_resi = 8*1024;	/* leave this amount resident	*/
bfid_t	zerobfid;		/* zero fill */
char	*username = NULL;
int	rash_fd;		/* 'file descriptor' to read from */

#define RASH_CLASS	"X"
#define	NAMESIZE	14

static	char	*filebuf = NULL;
static	int	bufsize = 0;
static	int	bnum;		/* block number in this file 	 */

/*
 *  info used in cleanup
 */
static vv_handle_t	*vhp = NULL, *prev_vhp = NULL;

/*
 *  GetClass() - figure out storage class to use for this file
 */
static char *
GetClass(uid_t uid, gid_t gid, off_t fsize)
{
	return(RASH_CLASS);
}

/*
 *  fatal
 *  clean up and exit.
 */
int
fatal()
{
	/*
	 *  If the disk file was opened then set the fail flag on
	 */
	if (rash_fd != -1) {
		close(rash_fd);
		rash_fd = -1;
	}
	rashdb_close();

	/*
	 *  close any open VV's
	 */
	if (prev_vhp != NULL)
		vv_close(prev_vhp, 0);
	if (vhp != NULL)
		vv_close(vhp, 0);
	vv_term();

	exit(1);
	return 0;	/* not reached */
}

/*
 *  intrpt
 *  Come here if a signal is caught.
 */
void
intrpt(int signum)
{
	static	char	id[] = "intrpt";

	nLogMsg(EL_STATUS2, id, "processing signal %d", signum);

	(void)sig_ign(SIGHUP);
	(void)sig_ign(SIGINT);
	(void)sig_ign(SIGTERM);

	if (oktodie)
		fatal();
	else
		dienow = TRUE;
}

/*
 *  archive_failed
 *  Print out an error message to the screen
 */
static	int
archive_failed(const char *filename)
{
	static	char	id[] = "Archive failed";

	close(rash_fd);
	rash_fd = -1;
	nLogMsg(EL_ERROR1, id, "%s", filename);
	fprintf(stderr, "%s: %s\n\t%s %s\n", filename,
			"Please see log file", ERRORARC,
			"for additional information");
	return 1;
}

/*
**	Make sure a VV has an EndOfTape mark and correct label info
**	for the file that the database shows as the last file.
*/
static int
lastfilecheck(vv_handle_t *vhp, vvlist_t *vvlP)
{
	static char	id[] = "lastfilecheck";
	USR_FILEINFO	*prev_fP;
	int		err;

	nLogMsg(EL_STATUS1, id, "vv %s", vvlP->vvname);

	/*
	**	get record for last file on vv
	*/
	prev_fP = rashdb_get_vv(&err, vvlP->vvname);
	if (prev_fP == NULL) {
		if (err != RASHDB_OK) {
			nLogMsg(EL_STATUS1, id,
				"no last file for vv %s err %d", 
				vvlP->vvname, err);
		}
		return 1;
	}
	if (prev_fP->usr_fno >= MAXTAPEFILES) {
		nLogMsg(EL_STATUS1, id, "vv %s already has %d files", 
				vvlP->vvname, prev_fP->usr_fno);
		return 1;
	}

	/*
	**	check eof label hdr
	*/
	if (lbl_eof_check(vhp, prev_fP))
		return 1;

	vvlP->lastfno = prev_fP->usr_fno;
	return 0;
}

/*
 *  writefile() - copy the file onto VV(s), writing labels. 
 *	Return list of VV's.
 */
static vvlist_t *
writefile(USR_FILEINFO *fP, int *nvv, struct stat *sP, rashstat_t *rP)
{
	static	char	id[] = "writefile";
	char		usertag[64], *vvclass;
	static	char	name[32];
	off_t		remaining, written;
	char		groupname[NAMESIZE], *prevvname, *cp, *ch;
	int		i, hdrsize, maxhdrsize, listsize, retried = 0;
	int		bytes_per_second;
	time_t		start_time = 0, end_time = 0;
	vvlist_t	*vvlist;
	extern	int	EndMarkSize;
	file_hdr_lbl_t	*lbl = NULL, *prev_lbl = NULL;

	/*
	**	Commit to archive.
	*/
	if (fcntl(rash_fd, DMFS_SETAIP, NULL) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "SETAIP (%s)", fP->usr_fname);
		return NULL;
	}
	time(&start_time);

	if (rP->st_rashflg & DM_ARCH) {
		nLogMsg(EL_ERROR1, id, "%s is already flagged as archived",
							fP->usr_fname);
		return NULL;
	}

	if (name[0] == '\0') {		/* need hostname */
		if (gethostname(name, sizeof(name)) == -1) {
			nLogMsg(EL_ERROR1|EL_SYS, id, "gethostname");
			return(NULL);
		}
		name[sizeof(name)-1] = '\0';
		if ((cp = index(name, '.')) != NULL)
			*cp = '\0';
	}

	/* generate VV id */
	sprintf(usertag, "%d%s%s", sP->st_uid, RASH_VVID, name);

	if ((ch = nas_gidname(sP->st_gid)) == NULL)
		sprintf(groupname, "%d", sP->st_gid);
	else
		strcpy(groupname, cp);

	vvclass = GetClass(sP->st_uid, sP->st_gid, sP->st_size);
	hdrsize = sizeof(file_hdr_lbl_t) + strlen(fP->usr_fname) + EndMarkSize;
	maxhdrsize = hdrsize + sizeof(vv_hdr_lbl_t);

	*nvv = 0;
	listsize = 20;
	vvlist = (vvlist_t *) nAlloc(listsize * sizeof(vvlist_t));

	remaining = sP->st_size;
	written = 0;
	bnum = 1;

	for (i=0; ; i++) {
		off_t	available_space;

		/*
		**	expand vvlist if necc
		*/
		if (i == listsize) {
			listsize += 20;
			vvlist = (vvlist_t *) 
				nReAlloc(vvlist, listsize * sizeof(vvlist_t));
		}
 retryvv:
		if (vhp != NULL) {
			/*
			**	throw away prev VV if possible
			*/
			nLogMsg(EL_STATUS1, id, "retry");
			
			/* TODO - throw away?? */
			if (vv_close(vhp, 1))
				vvErrorMsg(EL_ERROR2, id, "close");
			vv_destroyhandle(&vhp);
		}

		/*
		**	get a VV, allowing space for HDR & EOF labels
		*/
		vhp = vv_openwrite(usertag, vvclass, maxhdrsize+remaining,
					sizeof(file_hdr_lbl_t) + EndMarkSize);
		if (vhp == NULL) {
			vvErrorMsg(EL_ERROR1, id, "openwrite");

			switch (VVError) {

			case VVE_DOWN:
			case VVE_COMM:
				/*
				**  need to reconnect to vvmd
				*/
				vv_init(vvHost, RETRY_FOREVER);
				goto retryvv;

			case VVE_PERM:
			case VVE_NOTFOUND:
			case VVE_TAPE:
			case VVE_FILE:
				/*
				**  skip this file but continue processing
				*/
				goto clean;

			default:
				/*
				**  things are bad enough to bail
				*/
				fatal();
				break;		/* not reached */
			}
		}

		/*
		 *  note the VV name & the position in it and in the file
		 */
		strcpy(vvlist[i].vvname, vv_handle2name(vhp));
		if ((vvlist[i].vvlseek = vv_tell(vhp)) == (off_t) -1)
			vvErrorMsg(EL_FATAL, id, "vv_tell");
		vvlist[i].vvdata = 0;
		vvlist[i].flseek = written;

		nLogMsg(EL_STATUS1, id, vvlist[i].vvname);
		if (vvlist[i].vvlseek == (off_t) 0) {

			/*
			**	new VV: write vv_hdr indicating the user
			*/
			vvlist[i].new = 1;
			vvlist[i].lastfno = -1;
			if (lbl_vv_hdr(vhp, vvlist[i].vvname, 
						username, sP->st_uid))
				goto retryvv;

			/*
			**	get new start-of-file-in-vv location
			*/
			if ((vvlist[i].vvlseek = vv_tell(vhp)) == (off_t) -1)
				vvErrorMsg(EL_FATAL, id, "vv_tell");

		} else {
			vv_handle_t	*vhp2;

			/*
			**	previously written-to VV: open for read
			**		check vv_hdr and previous file's 
			**		end lbl
			**	vvlist[i].lastfno gets filled in by
			**		lastfilecheck
			*/
			vvlist[i].new = 0;
			if ((vhp2 = vv_openread(vvlist[i].vvname)) == NULL) {
				vvErrorMsg(EL_ERROR2, id, vvlist[i].vvname);
				if (retried)
					nLogMsg(EL_FATAL, id, "try 2b failed");
				retried = 1;
				goto retryvv;
			}
			if (lbl_vv_hdrcheck(vhp2, vvlist[i].vvname,
						username, sP->st_uid)) {
				if (retried)
					nLogMsg(EL_FATAL, id, "try 2c failed");
				retried = 1;
				if (vv_close(vhp2, 0)) {
					vvErrorMsg(EL_ERROR2, id,
						"vhp2 close");
				}
				vv_destroyhandle(&vhp2);
				goto retryvv;
			}
			if (lastfilecheck(vhp2, &vvlist[i])) {
				if (retried)
					nLogMsg(EL_FATAL, id, "try 2d failed");
				retried = 1;
				if (vv_close(vhp2, 0)) {
					vvErrorMsg(EL_ERROR2, id,
						"vhp2 close");
				}
				vv_destroyhandle(&vhp2);
				goto retryvv;
			}
			if (vv_close(vhp2, 0))
				vvErrorMsg(EL_ERROR2, id, "vv_close");
			vv_destroyhandle(&vhp2);
		}
		retried = 0;
		*nvv += 1;

		/*
		**	gather info for file header label
		*/

		available_space = vv_available(vhp);
		if (available_space == (off_t)-1)
			vvErrorMsg(EL_FATAL, id, "vv_available");
		available_space -= hdrsize;
		if (available_space < maxhdrsize) {
			/*
			**	this vv is too full to use
			*/
			nLogMsg(EL_ERROR1, id, "%s: only %lld available",
						vvlist[i].vvname,
						vv_available(vhp));
			goto retryvv;
		}

		/* amount of file data that will go on vv */
		fP->usr_vvdata = MIN(available_space, remaining);
		fP->usr_fno = vvlist[i].lastfno + 1;
		fP->usr_vvno = i;

		if (i)
			prevvname = vvlist[i-1].vvname;
		else
			prevvname = NULL;

		/*
		**	write the file header label to the VV
		*/
		nLogMsg(EL_DEBUG1, id, "writing hdr");
		lbl = lbl_file_hdr(vhp, fP, groupname, 
					vvlist[0].vvname, prevvname);
		if (lbl == NULL)
			goto clean;

		/*
		 *  copy file data to VV via buffer
		 */
		while (available_space  &&  remaining) {
			off_t	readsize = MIN(available_space, bufsize);
			off_t	retsize, writesize;

			/*
			 *  read next part of file into buffer
			 */
			retsize = read(rash_fd, filebuf, readsize);
			if (retsize == -1) {	/* read error */
				nLogMsg(EL_ERROR1|EL_SYS, id, "read(%s)",
						fP->usr_fname);
				goto clean;
			}
			writesize = retsize;

			/*
			 *  write buffer to VV
			 */
			retsize = vv_write(vhp, filebuf, writesize);
			if (retsize != writesize) {
				vvErrorMsg(EL_ERROR1, id, "vv_write");
				nLogMsg(EL_ERROR1, id, "wrote %d tried %d",
							retsize, writesize);
				goto clean;
			}

			/*
			 *  update per-VV and total accounts
			 */
			available_space -= retsize;
			vvlist[i].vvdata += retsize;
			remaining -= retsize;
			written += retsize;
		}

		if (fP->usr_vvdata != vvlist[i].vvdata) {
			nLogMsg(EL_ERROR1, id, "wrote %lld, expected %lld",
					vvlist[i].vvdata, fP->usr_vvdata);
			goto clean;
		}

		/*
		 *  data all written - prepare VV to write
		 *	trailing label
		 */
		if (vv_startreserve(vhp)) {
			vvErrorMsg(EL_ERROR1, id, "vv_startreserve");
			goto clean;
		}

		if (remaining == (off_t) 0) {
			/*
			**	write EOF label to this VV
			*/
			if (lbl_eof(vhp, lbl))
				goto clean;
			nFree(lbl);

			/*
			**	done w/ VV; close w/ 'finish' if not 1st
			*/
			if (vv_close(vhp, i))
				vvErrorMsg(EL_FATAL, id, "close vhp");
			vv_destroyhandle(&vhp);
		}

		/*
		**	if not 1st, write EOV label on prev
		*/
		if (i) {
			/*
			**	2nd or later VV in a multi-VV file:
			**		write EOV labels to prev VV
			**		containing the chain fwd info 
			**		to the VV just written.
			*/
			if (lbl_eov(prev_vhp, prev_lbl, vvlist[i].vvname))
				goto clean;
			nFree(prev_lbl);

			/*
			 *  finally done w/ prev VV
			 */
			if (vv_close(prev_vhp, 1))
				vvErrorMsg(EL_FATAL, id, "close prev_vhp");
			vv_destroyhandle(&prev_vhp);
		}
		if (remaining == (off_t) 0) {
			/*
			 *  done w/ file
			 */
			break;
		}

		/*
		**	more VV(s) req'd; write this VV's EOV labels when 
		**		next VV is known
		*/
		assert(prev_vhp == NULL);
		prev_vhp = vhp;
		prev_lbl = lbl;
		vhp = NULL;
	}

	/*
	**	Done copying file to VV's.
	**		Make sure that the number of bytes found in the file
	**		is the same as STAT() found. If not, the file
	**		was probably modified while it was being archived.
	*/
	if (written != sP->st_size) {
		nLogMsg(EL_STATUS2, id, 
			"%s archive ignored: size changed %lld->%lld.",
			fP->usr_fname, sP->st_size, written);
		goto clean;	/* TODO - could restart file.. */
	}

	/*
	 *  log statistics
	 */
	end_time = time((time_t *) 0) - start_time;
	if ( end_time )
                bytes_per_second = written / end_time;
        else
                bytes_per_second = written;

	nLogMsg(EL_STATUS2, id,
		"%s: %lld bytes copied to %d VV%s at %d bytes/sec",
		fP->usr_fname, sP->st_size, *nvv, 
		(*nvv == 1 ? "" : "'s"), bytes_per_second);
	return vvlist;

 clean:

	nLogMsg(EL_STATUS1, id, "%s: cleaning up VV's", fP->usr_fname);

	/*
	**	close and finish any open VV's
	*/
	if (vhp != NULL) {
		vv_close(vhp, 1);
		vv_destroyhandle(&vhp);
	}
	if (prev_vhp != NULL) {
		vv_close(prev_vhp, 1);
		vv_destroyhandle(&prev_vhp);
	}

	/*
	**	throw away any label copies
	*/
	if (lbl != NULL)
		nFree(lbl);
	if (prev_lbl != NULL)
		nFree(prev_lbl);

	/*
	**	throw away all the VV's that were newly alloc'd
	*/
	for (i=0; i<*nvv; i++) {
		if (vvlist[i].new == 0)
			continue;
		nLogMsg(EL_STATUS1, id, "recycling %s", vvlist[i].vvname);
		if (vv_recycle(vvlist[i].vvname))
			vvErrorMsg(EL_ERROR2, id, vvlist[i].vvname);
	}
	nFree(vvlist);
	*nvv = 0;
	return NULL;
}

/*
**  rashvv
**  Move a file to one or more virtual volumes.
*/
int
rashvv(char *filename, struct stat *statP)
{
	static	char	*id = "rashvv";
	static	char	ignore[] = "archive ignored";
	rashstat_t	rstat;
	struct	stat	sbuf;
	int 	  	err, nvv;
	off_t    	filesize;
	USR_FILEINFO    *db_fP, filei;
	vvlist_t	*vvlist, *vvp;
	int		j;

	rash_fd = -1;
	if (!S_ISREG(statP->st_mode)) {
		nLogMsg(EL_STATUS1, id, "%s: not a plain file", filename);
		return 1;
	}

	rash_fd = open(filename, O_RDONLY);
	if (rash_fd == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "open %s", filename);
		goto done;
	}
	if (fcntl(rash_fd, DMFS_GETRINFO, &rstat) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "rashstat(%s)", filename);
		goto done;
	}

	filesize = statP->st_size;
	if (filesize == 0) {	
		if (rstat.st_rashflg & DM_ARCH )
			nLogMsg(EL_ERROR1, id, "%s: EMPTY *AND* ARCHIVED",
								filename);
		else
			nLogMsg(EL_STATUS1, id, "%s: **EMPTY**", filename);
		goto done;
	}
	if (rstat.st_rashflg & (DM_NONR|DM_RIP)) {
		nLogMsg(EL_STATUS1, id, "%s: %sresident %s", filename,
			(rstat.st_rashflg & DM_NONR) ? "non-" : "",
			(rstat.st_rashflg & DM_RIP) ? ", restoring" : "");
		goto done;
	}

	if (rstat.st_rashflg & DM_ARCH) {
		/*
		**  The file is already archived; if we are 
		**	not releasing space then skip the  
		**	database check (which verifies if 
		**	the file is in fact archived)
		*/
		if (!aflg)			
			goto done;

		username = nas_uidname(rstat.st_dbuid);
		if ( username == NULL ) {
			nLogMsg(EL_ERROR1, id,
				"%s: dbuid %d not in the password file",
				filename, rstat.st_dbuid);
			goto done;
		}
		/* 
		**  Check the database to make sure that the file 
		**	is in fact archived.
		*/
		if ( rashdb_open(username, O_DB_READ) ){
			nLogMsg(EL_ERROR1, id, "rashdb_open(user %s)",
				username);
			goto done;
		}
		db_fP = rashdb_get_bfid(&err, &rstat.st_bfid);
		rashdb_close();
		if ( db_fP == NULL) {
			nLogMsg(EL_ERROR1, id,
				"%s rashdb_get_bfid(%s) failed %d",
				username, bfidout(&rstat.st_bfid), err);
			nLogMsg(EL_ERROR1, id, 
				"%s 'archived' but not in database",
				filename);
			goto done;
		}

		/*
		**  got the dbase info - check against the inode
	 	*/
		if ( filesize != db_fP->usr_fsize ) {
			nLogMsg(EL_ERROR1, id, 
				"%s: size diff: ino %lld  db %lld",
				filename, filesize, db_fP->usr_fsize);
			if ( rstat.st_rashflg & DM_NONR) {
				nLogMsg(EL_ERROR1, id,
			    	"LOST %s: nonres & db/inode disagree",
			    	filename);
			}
		}
		else {
			/*
			**  file is truly believed to be archived
			**  already
		 	*/
			if ( !(rstat.st_rashflg & DM_NONR) && aflg) {
				/* 
				**  File is resident && need to
				**  release space
			 	*/
				if (fcntl(rash_fd, DMFS_SETNONR,
						&leave_resi) != 0) {
					nLogMsg(EL_ERROR1|EL_SYS, id,
						"R_SETNONR failed %s",
						filename);
					goto done;
				}
				nLogMsg(EL_STATUS2, id, 
					"%s successfully archived",
						filename);
				goto done;
			}

			nLogMsg(EL_STATUS1, id, 
					"%s is already archived",
					filename);
			goto done;
		}
		/* re-archive if db doesn't disagree? */

		nLogMsg(EL_DEBUG1, id,
			"found bfid %s file %s",
			bfidout(&rstat.st_bfid), filename);
	}

	if (filebuf == NULL)
		filebuf = (char *) nAlloc(bufsize = VVBUFSIZE);

	memset(&filei, '\0', sizeof(filei));
	/*
	**	Get new bfid if needed.
	*/
	if (memcmp(&rstat.st_bfid, &zerobfid, sizeof(bfid_t)) == 0) {
		/*
		** The file id is zero, so we need to get a fresh
		** one from the bfid server.
		*/
		if (getbfid(&filei.usr_bfid))
			return(archive_failed(filename));

		nLogMsg(EL_STATUS2, id, 
			"new bfid %s file %s", 
			bfidout(&filei.usr_bfid), filename);
		/*
		**      set the info in the inode
		*/
		rstat.st_dbuid = statP->st_uid;
		memcpy(&rstat.st_bfid, &filei.usr_bfid, sizeof(bfid_t));
		if (fcntl(rash_fd, DMFS_SETRINFO, &rstat) != 0) {
			nLogMsg(EL_ERROR1|EL_SYS, id,
				"%s: R_SETARC failed", filename);
			return(archive_failed(filename));
		}
	}
	else {
		memcpy(&filei.usr_bfid, &rstat.st_bfid, sizeof(bfid_t));
		nLogMsg(EL_STATUS2, id, "do bfid %s file %s",
			bfidout(&rstat.st_bfid), filename);
	}
	username = nas_uidname(rstat.st_dbuid);
	if ( username == NULL ) {
		nLogMsg(EL_ERROR1|EL_SYS, id,
			"%s: dbuid %d not in the password file",
			filename, rstat.st_dbuid);
		goto done;
	}
	/*      
	**	initialize the virtual volume manager connection
	*/
	vv_init(vvHost, RETRY_FOREVER);
	/*
	**	open user's dbase
	*/
	if (rashdb_open(username, O_DB_RDWR)) {
		nLogMsg(EL_STATUS2, id,
			"rashdb_open for write failed for user %s",
			username);
		return (archive_failed(filename));
	}
	nLogMsg(EL_STATUS2, id, "rashdb_open ok, user %s", username);

	/*
	**	Note file characteristics
	*/
	filei.usr_fname = filename;
	filei.usr_fsize = statP->st_size;
	filei.usr_uid = statP->st_uid;
	filei.usr_gid = statP->st_gid;
	filei.usr_mode = statP->st_mode;
	filei.usr_ctime = statP->st_ctime;
	filei.usr_mtime = statP->st_mtime;

	/*
	**	use a single archive time across multi-vv files
	*/
	time(&filei.usr_arctm);

	oktodie = FALSE;
	vvlist = writefile(&filei, &nvv, statP, &rstat);
	if (vvlist == NULL) {
		rashdb_close();
		goto done;
	}
	/*
	**	add a record to the data base for each vv
	*/
	filei.usr_vflg = (short) 0;
	for (j=0,vvp=vvlist; j<nvv; j++, vvp++) {
		/*
		**	fill in the file info
		*/
		filei.usr_vvno = j;
		strcpy(filei.usr_vvname, vvp->vvname);
		filei.usr_fno = vvp->lastfno + 1;
		filei.usr_bof = vvp->vvlseek;
		filei.usr_lseek = vvp->flseek;
		filei.usr_vvdata = vvp->vvdata;

		/*
		**	add a record for the vv
		*/
		if (rashdb_add(&filei)) {
				nLogMsg(EL_STATUS2, id, 
				"rashdb_add failed for user %s",
					username);
			rashdb_close();
			return(archive_failed(filei.usr_fname));
		}
		nLogMsg(EL_STATUS2, id, "  rashdb_add ok, %s:%d %s",
			username, j, vvp->vvname);
	}
	rashdb_close();

	/*
	**	discard the vv list
	*/
	nFree(vvlist);

	/*
	**	check that the file on disk is still the 
	**	same file that was moved to VV. If is not,
	**	then the file was probably changed while 
	**	it was being copied to VV, so the disk
	**	version should be left alone. Otherwise, 
	**	set the archive bit on.
	*/
	if (fstat(rash_fd, &sbuf) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "fstat failed for %s",
			filei.usr_fname);
		goto done;
	}
	if (sbuf.st_size != filei.usr_fsize) {
		nLogMsg(EL_STATUS2, id,
			"%s %s: file size %lld changed to %lld",
			filei.usr_fname, ignore,
			filei.usr_fsize, sbuf.st_size);
		goto done;
	}
	if (sbuf.st_mtime != filei.usr_mtime) {
		nLogMsg(EL_STATUS2, id,
			"%s %s: mtime %d changed to %d",
			filei.usr_fname, ignore,
			filei.usr_mtime, sbuf.st_mtime);
		goto done;
	}

	if (fcntl(rash_fd, DMFS_FINISHARCH, NULL) != 0) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "%s: FINISHARCH failed",
			filei.usr_fname);
		goto done;
	}

	/*
	**	Clear metrics tcompact flag since file was archived
	*/
	fcntl(rash_fd, DMFS_CLRCOMPACT, NULL);

	/*
	**	Release the allocated blocks from the file if requested
	*/
	if (aflg) {
		if (fcntl(rash_fd, DMFS_SETNONR, &leave_resi) != 0) {
			nLogMsg(EL_ERROR1|EL_SYS, id,
				"R_SETNONR failed for %s",
				filei.usr_fname);
		}
	}

 done:
	close(rash_fd);
	rash_fd = -1;

	if (dienow)
		fatal();
	oktodie = TRUE;
	return 0;
}
