#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/mount.h>

#include <miscfs/dmfs/rash.h>
#include <miscfs/dmfs/dmfs_fcntl.h>
#include <miscfs/dmfs/dmfs.h>

#include "dmfsd.h"

char            buffer[MAXPATHLEN+256];
char            path[MAXPATHLEN+1];

off_t		leave_resi = 8*1024;		/* leave this resident */
int		verbose = 0;
int		hitfiles = 0;

#define	FLAGMASK	(DM_ARCH|DM_NONR)

void
reclaim(char *file)
{
	rashstat_t	rstat;
	int		fd;

	if ((fd = open(file, O_RDONLY)) == -1) {
		sprintf(buffer, "open: %s", file);
		perror(buffer);
		return;
	}
	if (fcntl(fd, DMFS_GETRINFO, &rstat) == -1) {
		sprintf(buffer, "DMFS_GETRINFO: %s", file);
		perror(buffer);
		close(fd);
		return;
	}
	if ((rstat.st_rashflg & FLAGMASK) == DM_ARCH) {
		if (verbose)
			printf("reclaim %s\n", file);

		if (fcntl(fd, DMFS_SETNONR, &leave_resi) != 0) {
			sprintf(buffer, "DMFS_SETNONR: %s", file);
			perror(buffer);
		}
		else
			hitfiles++;
	}
	else if ((rstat.st_rashflg & FLAGMASK) == (DM_ARCH|DM_NONR) &&
			rstat.st_bbarrier > leave_resi) {
		if (verbose) {
			printf("barrier %lld -> %lld %s\n",
				(long long)rstat.st_bbarrier,
				(long long)leave_resi, file);
		}

		if (fcntl(fd, DMFS_MOVENONR, &leave_resi) != 0) {
			sprintf(buffer, "DMFS_MOVENONR: %s", file);
			perror(buffer);
		}
		else
			hitfiles++;
	}
	else if (verbose)
		printf("%s: %x\n", file, rstat.st_rashflg);
	close(fd);
}

int
main(int argc, char *argv[])
{
	struct statfs	fsb;
	FILE		*listfp;
	int		lineno = 0;
	int		c;
	uid_t		uid;
	gid_t		gid;
	off_t		size;
	long long	totalblks;
	int		desired_free = 50;
	int		current_free;
	int		checkspace = 1;
	char		*chr;
	int		errflg = 0;


	while ((c = getopt(argc, argv, "A:f:vi")) != EOF) {
		switch (c) {

		case 'A':
			leave_resi = strtoq(optarg, &chr, 0);
			if (chr == optarg || *chr != '\0') {
				fprintf(stderr,
					"garbled resi value %s\n", optarg);
				errflg = 1;
			}
			else if (leave_resi < 0 ||
					1024*1024*1024 < leave_resi) {
				fprintf(stderr,
					"resident size %s out of range\n",
					optarg);
				errflg = 1;
			}
			break;

		case 'f':
			desired_free = atoi(optarg);
			if (desired_free < 10 || desired_free > 100) {
	  			fprintf(stderr, 
	    				"freespace%% must be 10 to 100\n");
				errflg = 1;
			}
			break;

		case 'v':
			verbose = 1;
			break;

		case 'i':
			checkspace = 0;
			break;

		case '?':
		default:
			errflg = 1;
			break;
		}
	}

	if (optind != argc-1 && optind != argc)
		errflg = 1;
	if (errflg) {
		fprintf(stderr,
			"Uasge: %s [-A resi] [-f free%%] [-vi] [list-file]\n",
			argv[0]);
		exit(1);
	}

	if (optind == argc)
		listfp = stdin;
	else {
		if ((listfp = fopen(argv[optind], "r")) == NULL) {
			fprintf(stderr, "Unable to open %s\n", argv[optind]);
			exit(1);
		}
	}

	/*
	 * read in the list and free files until the target amount of storage
	 * is available
	 * 
	 */

	for (;;) {
		c = file_line(listfp, &uid, &gid, &size, path);
		if (c) {
			if (c == -1)
				fprintf(stderr, "garbled line %d\n", lineno);
			break;
		}

		lineno++;
		if (checkspace) {
			if (statfs(path, &fsb) == -1) {
				fprintf(stderr, "statfs of %s failed\n", path);
				continue;
			}

			totalblks = fsb.f_blocks - fsb.f_bfree + fsb.f_bavail;
			current_free = 100 * fsb.f_bavail / totalblks;

			if (verbose) {
				printf("at %d want %d avail %ld total %lld\n", 
					current_free, desired_free,
		       			fsb.f_bavail, totalblks);
			}

			if (current_free >= desired_free)
				break;
		}
		reclaim(path);
	}
	if (checkspace || verbose)
		printf("examined %d files, reclaimed %d\n", lineno, hitfiles);
	exit(0);
}
