/*
 *  restfile.c - NAStore RASH restore of file given VV list
 *
 */

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/param.h>
#include <miscfs/dmfs/rash.h>
#include <miscfs/dmfs/dmfs_fcntl.h>
#include <miscfs/dmfs/dmfs.h>

#include "vv_cl.h"
#include "dblib.h"
#include "nas_logmsg.h"
#include "nasutil.h"
#include "rashcom.h"
#include "labels.h"
#include "dmfsd.h"

int		rash_fd;
static char    	*databufP = NULL;

/*
**	Turn off the archive flag off from a file on disk.
*/
int
resetfd(char *id, char *name)
{
	if (fcntl(rash_fd, DMFS_CLRARC, NULL) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id,
			"%s: R_CLRARC failed", name);
		return 1;
	}

	/*
	 * now set the tcompact flag to indicate for the metrics that
	 * this file was restored as a result of reset
	 */
	if (fcntl(rash_fd, DMFS_SETCOMPACT, NULL) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, 
			"%s: R_SETCOMPACT failed", name);
		return 1;
	}

	return 0;
}


/*
**  getdb() - get all VV info for file
*/
vvlist_t *
getdb(USR_FILEINFO *uhold, struct stat *sP, rashstat_t *rP, int *nvv)
{
	static	char	id[] = "getdb";
	static	int	dbuid = -1;	/* remember what is open */
	char		*username;
	USR_FILEINFO	*ufP;
	vvlist_t	*vvlist;
	int		err, i, listsize;
	time_t		arctm;

	if (dbuid != rP->st_dbuid) {
		if (dbuid != -1) {
			dbuid = -1;
			rashdb_close();
		}

		username = nas_uidname(rP->st_dbuid);
		if ( username == NULL ) {
			nLogMsg(EL_ERROR1, id,
				"dbuid %d not in the password file",
				rP->st_dbuid);
			return NULL;
		}
		if ( rashdb_open(username, O_DB_READ) ){
			nLogMsg(EL_ERROR1, id, "rashdb_open(user %s)",
				username);
			return NULL;
		}
		dbuid = rP->st_dbuid;
	}
	/*
	**	get 1st record for bfid
	*/
	ufP = rashdb_get_bfid(&err, &rP->st_bfid);
	if (ufP == NULL) {
		nLogMsg(EL_ERROR1, id, "%s: no record, err %d",
			bfidout(&rP->st_bfid), err);
		return NULL;
	}
	/*
	**	Save the database info except for the filename.
	*/
	memcpy(uhold, ufP, sizeof(*uhold));
	uhold->usr_fname = NULL;

	/*
	**	initialize vvlist
	*/
	*nvv = 0;
	arctm = ufP->usr_arctm;
	listsize = 5;
	vvlist = (vvlist_t *) nAlloc(listsize * sizeof(vvlist_t));

	for (i=0; ; i++) {

		/*
		**	perform sanity checks
		*/
		if (ufP->usr_arctm != arctm) {
			nLogMsg(EL_DEBUG1, id,
				"skipping %s: archive time mismatch",
				bfidout(&rP->st_bfid));
			nFree(vvlist);
			return NULL;
		}
		if (ufP->usr_fsize != sP->st_size) {
			nLogMsg(EL_ERROR1, id,
				"%s vvno %d fsize(%lld), ino(%lld)",
				bfidout(&rP->st_bfid), ufP->usr_vvno,
				ufP->usr_fsize, sP->st_size);
			nFree(vvlist);
			return NULL;
		}

		/*
		**	expand VV list if necc
		*/
		if (i == listsize) {
			listsize *= 2;
			vvlist = (vvlist_t *) nReAlloc((void *)vvlist,
						listsize * sizeof(vvlist_t));
		}

		/*
		**	save db info to VV list
		*/
		strcpy(vvlist[i].vvname, ufP->usr_vvname);
		vvlist[i].vvlseek = ufP->usr_bof;
		vvlist[i].flseek  = ufP->usr_lseek;
		vvlist[i].vvdata  = ufP->usr_vvdata;
		vvlist[i].fno	  = ufP->usr_fno;

		*nvv += 1;
		/*
		**	get next bfid.VV record
		*/
		if ((ufP = rashdb_next_bfidrec(&err, &rP->st_bfid)) == NULL) {
			if (err == RASHDB_ERROR) {
				nLogMsg(EL_ERROR1, id,
					"rashdb_next_bfidrec %s failed %d",
					bfidout(&rP->st_bfid), err);
			}
			break;	/* no more VV records */
		}
	} /* end for loop */
	vvlist = (vvlist_t *)nReAlloc((void *)vvlist, (i+1)*sizeof(vvlist_t));

	return vvlist;
}

/*
**	Restore a file given an open file descriptor, an optional name,
**	all inode info and a vvlist.
*/
int
restfile(USR_FILEINFO *fP, struct stat *sP, rashstat_t *rP,
		vvlist_t *vvlist, int nvv)
{
	char		*id = "restfile";
	char		*bfidstr;
	int     	i, bps1, bps2;
	off_t		remaining_bytes, chars;
	off_t		vvdata;
	off_t		numchars;
	off_t		size;
	time_t		start_time, end_time;
	time_t		time0, time1, mount_time = 0, copy_time = 0;
	vv_handle_t	*vhp = NULL;

	time(&start_time);

	bfidstr = bfidout(&rP->st_bfid);
	if (vvlist == NULL || nvv == 0)
		return 1;

	/* 
	 *  allocate a buffer for file system write
	 */
	if (databufP == NULL)
		databufP = nAlloc(VVBUFSIZE);

	/*  
	 * Set the RRIP flag on in the event we fail to 
	 * restore since the kernel close code will clean up 
	 * for us if rashclose is never issued against the 
	 * file (free allocated blocks)
	 */
	if (fcntl(rash_fd, DMFS_SETRIP, NULL) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "SET_RRIP failed for %s",
			bfidstr);
		close(rash_fd);
		rash_fd = -1;
		return 1;
	}
	nLogMsg(EL_STATUS1, id, "commit to restore %s (%d VV's)",
		bfidstr, nvv);
	/*
	 *  Skip any bytes that are already resident (0 to barrier-1).
	 */
	vvdata = 0;
	for (i=0; i<nvv; i++) {
		if (vvlist[i].flseek != vvdata) {
			nLogMsg(EL_ERROR1, id, 
				"%s: offset is %lld, but VV %s #%d has %lld",
				bfidstr, vvdata,
				vvlist[i].vvname, i, vvlist[i].flseek);
			goto error;
		}
		if ((vvlist[i].flseek + vvlist[i].vvdata) >= rP->st_bbarrier)
			break;
		nLogMsg(EL_STATUS1, id, "skip %s (%lld:%lld)",
			vvlist[i].vvname,
			vvlist[i].flseek,
			vvlist[i].vvdata);
		vvdata += vvlist[i].vvdata;
	}
	/*
	**	Do first VV with possible skip of initial data.
	*/
	time(&time0);
retryvv1:
	vv_init(vvHost, RETRY_FOREVER);
	nLogMsg(EL_STATUS2, id, "%s %s", bfidstr, vvlist[i].vvname);
	vhp = vv_openread(vvlist[i].vvname);
	if (vhp == NULL) {
		vvErrorMsg(EL_ERROR1, id, "vv_openread");
		goto retryvv1;
	}

	time(&time1);
	mount_time += (time1 - time0);
	fP->usr_vvdata = vvlist[i].vvdata;
	fP->usr_fno = vvlist[i].fno;
	fP->usr_vvno = i;
	/*
	 *	check labels, position to read VV
	 */
	if (lbl_skipbof(vhp, &vvlist[i], fP))
		goto error;
	/*
	 *	skip any unneeded data on the VV.
	 */
	chars = rP->st_bbarrier;
	numchars = chars - vvdata;
	if (vv_lseek(vhp, numchars, SEEK_CUR) == -1) {
		vvErrorMsg(EL_ERROR1, id, "vv_lseek");
		goto error;
	}
	/*
	**	track bytes
	*/
	remaining_bytes = sP->st_size - chars;
	vvdata = vvlist[i].vvdata;
	/*
	**	seek in file
	*/
	if (lseek(rash_fd, chars, SEEK_SET) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "%s: file seek (%lld)",
			bfidstr, chars);
		goto error;
	}
	/*
	 *  read from the VV and write to the file  
	 */
	for (; numchars<vvdata; numchars+=size) {
		OFF_T	read_size = MIN(remaining_bytes, VVBUFSIZE);
		int	ret;

		read_size = MIN(read_size, (vvdata-numchars));
		if (read_size == 0)	/* end of file */
			break;

		size = vv_read(vhp, databufP, read_size);
		if (size != read_size) {
			vvErrorMsg(EL_ERROR1, id, "vv_read");
			goto error;
		}

		ret = write(rash_fd, databufP, size);
		if (ret != size) {
			nLogMsg(EL_ERROR1|EL_SYS, id, 
				"%s: fd %d vv %d write %lld size %lld",
				bfidstr, rash_fd, i, ret, size);
			goto error;
		}
		chars += size;
		remaining_bytes -= size;
		if (fcntl(rash_fd, DMFS_SETBBOUND, &chars) == -1) {
			nLogMsg(EL_ERROR1|EL_SYS, id, 
				"%s: rashcntl(R_SETRSIZE) failed (%d)",
				bfidstr, i);
			goto error;
		}
	}
	copy_time += time((time_t *)0) - time1;
	if (numchars != vvdata) {
		nLogMsg(EL_ERROR1, id, 
			"%s: read from first VV %s %lld, db said %dll",
			bfidstr, vvlist[i].vvname,
			numchars, vvdata);
		goto error;
	}
	/* check EOV/EOF labels */

	vv_close(vhp, 0);
	vv_destroyhandle(&vhp);

	/*
	 *  mount, check, & read any other VV's
	 */
	for (i++; i<nvv; i++) {
		/*
		 * Make sure that the seek location represented by this 
		 * VV does not leave a gap within the file. 
		 */
		if (vvlist[i].flseek != chars) {
			nLogMsg(EL_ERROR1, id, 
				"%s: offset is %lld, but VV %s #%d has %lld",
				bfidstr, chars,
				vvlist[i].vvname, i, vvlist[i].flseek);
			goto error;
		}

		time(&time0);
retryvv2:
		/*
		 *  get a handle on the VV
		 */
		nLogMsg(EL_STATUS2, id, "%s %s", bfidstr, vvlist[i].vvname);
		vhp = vv_openread(vvlist[i].vvname);
		if (vhp == NULL) {
			vvErrorMsg(EL_ERROR1, id, "vv_openread 2");
			vv_init(vvHost, RETRY_FOREVER);
			goto retryvv2;
		}

		time(&time1);
		mount_time += (time1 - time0);
		fP->usr_vvdata = vvlist[i].vvdata;
		fP->usr_fno = vvlist[i].fno;
		fP->usr_vvno = i;

		/*
		 *  check labels, position to read VV
		 */
		if (lbl_skipbof(vhp, &vvlist[i], fP))
			goto error;

		/*  
		 *  read from the VV and write to the file  
		 */
		vvdata = vvlist[i].vvdata;
		for (numchars=0; numchars<vvdata; numchars+=size) {
			OFF_T	read_size = MIN(remaining_bytes, VVBUFSIZE);
			int	ret;

			read_size = MIN(read_size, (vvdata-numchars));
			if (read_size == 0)	/* end of file */
				break;

			size = vv_read(vhp, databufP, read_size);
			if (size != read_size) {
				vvErrorMsg(EL_ERROR1, id, "vv_read 2");
				goto error;
			}

			ret = write(rash_fd, databufP, size);
			if (ret != size) {
				nLogMsg(EL_ERROR1|EL_SYS, id, 
					"%s: fd %d vv %d write %lld size %lld",
					bfidstr, rash_fd, i, ret, size);
				goto error;
			}
			chars += size;
			remaining_bytes -= size;
			if (fcntl(rash_fd, DMFS_SETBBOUND, &chars) == -1) {
				nLogMsg(EL_ERROR1|EL_SYS, id, 
					"%s: rashcntl(R_SETRSIZE) failed (%d)",
					bfidstr, i);
				goto error;
			}
		}
		copy_time += time((time_t *)0) - time1;
		if (numchars != vvdata) {
			nLogMsg(EL_ERROR1, id, 
				"%s: read from VV #%d %s %lld, db said %lld",
				bfidstr, i, vvlist[i].vvname,
				numchars, vvdata);
			goto error;
		}
		/* check EOV/EOF labels */

		vv_close(vhp, 0);
		vv_destroyhandle(&vhp);

	} /* while/VV */

	/*
	 *  check size
	 */
	if (chars != sP->st_size) {
		nLogMsg(EL_ERROR1, id, "%s restored %lld != file size %lld",
				bfidstr, chars, sP->st_size);
		goto error;
	}

	if (resetflg) {
		/*
		 *  caller requested to clear inode flag
		 *	which indicates that there is a
		 *	copy of the file on VV
		 */
		resetfd(id, bfidstr);
	}

	/*
	**  done with rash-specific inode stuff
	*/
	if (fcntl(rash_fd, DMFS_FINISHRESTORE, NULL) == -1)
		nLogMsg(EL_ERROR1|EL_SYS, id, "DMFS_FINISHRESTORE");
	if (close(rash_fd) == -1)
		nLogMsg(EL_ERROR1|EL_SYS, id, "final close");
	rash_fd = -1;

	/*
	**  completely done w/ file - do accounting
	*/
	time(&end_time);
        end_time -= start_time;

	nLogMsg(EL_STATUS1, id, "%s GOOD %lld bytes in %d secs",
		bfidstr, chars, end_time);

	chars -= rP->st_bbarrier;
	if (end_time)
		bps1 = chars / end_time;
	else
		bps1 = chars;

	if (copy_time)
		bps2 = chars / copy_time;
	else
		bps2 = chars;

	nLogMsg(EL_STATUS2, id,
		"%d VVs total %d secs %d cps, active %d secs %d cps",
		nvv, end_time, bps1, copy_time, bps2);
	return 0;

error:
	nLogMsg(EL_STATUS1, id, "%s: restore failed", bfidstr);
	if (fcntl(rash_fd, DMFS_SETRFAIL, NULL) == -1) {
		nLogMsg(EL_ERROR1|EL_SYS, id,
			"R_SETRRFAIL failed %s", bfidstr);
	}
	if (close(rash_fd) == -1)
		nLogMsg(EL_ERROR1|EL_SYS, id, "%s: closeing failed", bfidstr);
	rash_fd = -1;

	if (vhp != NULL) {
		vv_close(vhp, 0);
		vv_destroyhandle(&vhp);
	}
	return 1;
}
