/*
**  rls.c  -  A tool which reads records from a RASH database
**            according to the criteria specified on the command line.
*/
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <assert.h>
#include <re_comp.h>

#include <sys/types.h>
#include <sys/param.h>
#include <miscfs/dmfs/rash.h>

#include "nasutil.h"
#include "dblib.h"
#include "dmfsd.h"

#define FLENGTH  	1024
#define LINELEN	  	80

#define QHEAD                  ((vvfact_t *) TFQ.q_head)
#define QNEXT(tfp)             ((vvfact_t *) ((tfp)->vv_q.q_next))

#define ENQ(tfp)               ENQUEUE((tfp), &TFQ, vv_q, vvfact_t)
#define DQ(tfp)                DEQUEUE((tfp), &TFQ, vv_q, vvfact_t)

typedef struct {
	QLink_t		vv_q;			/* queue */
	char		vv_name[VVNAMELEN];  	/* vv name */
	off_t		vv_tdata;		/* total data bytes on vv */
	off_t		vv_deldata;		/* data bytes deleted */
	unsigned int	vv_nfiles;		/* total files on vv */
	unsigned int	vv_ndel;		/* deleted files on vv */
} vvfact_t;

static QHead_t  TFQ;                    /* work q */

ushort	rec_flag = 1;
ushort	lflag = 0;
ushort	Tflag = 0,
	Vflag = 0,
	Fflag = 0;

short	tmflag = 0;		/* must be a signed short */

char	padbuf[LINELEN];

void
usage()
{
	fprintf(stderr, "USAGE:  ");
	fprintf(stderr, "rls [-l] [record-options] login-name\n");
	fprintf(stderr, "   OR:  ");
	fprintf(stderr, "rls [list-option] login-name\n");
	fprintf(stderr, "\trecord-options:\n");
	fprintf(stderr, "\t  -f filename \t\t list records containing file\n");
	fprintf(stderr, "\t  -m modify-time \t list records by mod time\n");
	fprintf(stderr, "\t  -c access-time \t list records by access time\n");
	fprintf(stderr, "\t  -a archive-time \t list records by arctime\n");
	fprintf(stderr, "\t  -b bfid \t\t list records by bfid\n");
	fprintf(stderr, "\t  -s sequence \t\t list records by vv number\n");
	fprintf(stderr, "\t  -v vvname \t\t list records containing vvname\n");
	fprintf(stderr, "\tlist-options:\n");
	fprintf(stderr, "\t  -T \t\t\t list vvnames in use by user\n");
	fprintf(stderr, "\t  -F filename \t\t list vvnames containing file\n");
	fprintf(stderr, "\t  -V vvname \t\t list files on a vv\n");
	fprintf(stderr, "\t  -g group_no \t\t list records by group no \n");
	exit(2);
}

void
diefast()
{
	printf("rls aborted\n");
	exit(1);
}

/*
**  FUNCTION:	bomb()
**		Signal handler.
*/
static void
bomb(int signum)
{
	fprintf(stderr, "\ncaught signal %d\n", signum);
	rashdb_close();
	exit(1);
}

/*
**  FUNCTION:	pad()
**		Generate blank character pad string.
*/
void
pad(int n)
{
	register int 	i;

	n = (n > LINELEN) ? LINELEN : n;

	for (i = 0; i < n; i++)
		padbuf[i] = ' ';

	padbuf[n] = '\0';
}

/*
**  FUNCTION:	print_rec()
**		Print a RASH database record.
*/
void
print_rec(char *usrname, USR_FILEINFO *up)
{
	static	char	acc_time[25], mod_time[25], arc_time[25];

	if (lflag) {
		printf("***************************************************\n");
		printf("   RASH DB RECORD FOR:  %s\n", up->usr_fname);
		printf("***************************************************\n");
		pad(12);
		printf("%s%s%s\n", padbuf, "Bitfile id:  ",
			bfidout(&up->usr_bfid));
		pad(10);
		printf("%s%s%d\n", padbuf, "Rash release:  ",
			up->usr_version);
		pad(9);
		printf("%s%s%d\n", padbuf, "Rash Group No:  ",
			up->usr_group_no);
		pad(10);
		printf("%s%s%s%s%s\n", padbuf, "Record flags:  ",
			(up->usr_rflg & RFLG_INODB) ?
				"*INODB" : "",
			(up->usr_rflg & RFLG_NOIDX) ?
				"*IGNORE IN INDEX" : "",
			(up->usr_rflg & RFLG_DELETED) ?
				"*DELETED" : "");
		pad(13);
		printf("%s%s%d\n\n", padbuf, "VV number:  ", up->usr_vvno); 
		pad(15);
		printf("%s%s%s\n", padbuf, "VV name:  ", up->usr_vvname);
		pad(11);
		printf("%s%s%d\n", padbuf, "File Number:  ",
			up->usr_fno);
		pad(10);
		printf("%s%s%08X\n", padbuf, "Block Offset:  ", 
			up->usr_bof & 0x03fffff);
		pad(15);
		printf("%s%s%s%s%s%s\n", padbuf, "VV Flag:  ",
			(up->usr_vflg & VFLG_BAD) ?
				"*FILE DAMAGED" : "",
			(up->usr_vflg & VFLG_NONEXISTENT) ?
				"*FILE NOT ON VV" : "",
			(up->usr_vflg & VFLG_RECYCLE) ?
				"*RECYCLED" : "",
			(up->usr_vflg & VFLG_OK) ?
				"*TAPEDB" : "");

		pad(4);
		printf("%s%s%lld\n", padbuf, "Lseek for disk loc:  ",
				(long long)up->usr_lseek);
		pad(5);
		printf("%s%s%lld\n\n", padbuf,
				"File's data on vv:  ",
				(long long)up->usr_vvdata);
		pad(13);
		printf("%s%s%lld\n", padbuf, "File Size:  ",
			(long long) up->usr_fsize);
		pad(13);
		printf("%s%s%s\n", padbuf, "User Name:  ", usrname);
		pad(15);
		printf("%s%s%ld\n", padbuf, "User Id:  ", (long) up->usr_uid);
		pad(7);
		printf("%s%s%ld\n", padbuf, "User's Group Id:  ",
			(long) up->usr_gid);
		pad(2);
		printf("%s%s%s", padbuf, "Record Creation Time:  ",
				ctime(&up->usr_rectm));
		pad(0);
		printf("%s%s%s", padbuf, "File Modification Time:  ",
			ctime(&up->usr_mtime));
		pad(6);
		printf("%s%s%s", padbuf, "File Access Time:  ",
			ctime(&up->usr_ctime));
		pad(5);
		printf("%s%s%s", padbuf, "File Archive Time:  ",
			ctime(&up->usr_arctm));
		if (up->usr_rflg & RFLG_DELETED) {
			pad(2);
			printf("%s%s%s", padbuf, "Record Deletion Time:  ",
				ctime(&up->usr_deltm));
		}
	} else {
		printf("%s %02d %02u ",
			bfidout(&up->usr_bfid),
			up->usr_vvno,
			up->usr_rflg);

		printf("%s ", up->usr_vvname);
		printf("%04d ", up->usr_fno);	
		printf("%06u ", up->usr_bof);
		printf("%02d ", up->usr_vflg);
		printf("%10lld ", (long long)up->usr_lseek);	

		strcpy(acc_time,ctime(&up->usr_ctime));   
		strcpy(mod_time,ctime(&up->usr_mtime));  
		strcpy(arc_time,ctime(&up->usr_arctm)); 
		printf("%020lld %05ld %05ld %-24.24s %-24.24s %-24.24s %s\n\n",
			(long long) up->usr_fsize,
			(long) up->usr_uid,
			(long) up->usr_gid,
			acc_time, mod_time, arc_time,
			up->usr_fname);
	}
}

/*
**  FUNCTION:	comp_time()
**		Compare time values.
*/
int
comp_time(time_t time1, time_t time2)
{
	int	retval = 1;

	switch (tmflag) {
	  case 0:
		if (time1 == time2)
			retval = 0;
		break;
	  case -1: 
		if (time1 < time2)
			retval = 0;
		break;
	  case 1: 
		if (time1 > time2)
			retval = 0;
		break;
	  default:
		break;
	}

	return(retval);
}

/*
**  FUNCTION:	add_vvs()
**		Add a new vv to the list.
*/
void
add_vvs(USR_FILEINFO *up)
{
	vvfact_t		*newtp;
	
	newtp = (vvfact_t *)nAlloc(sizeof(vvfact_t));
	strcpy(newtp->vv_name, up->usr_vvname);
	newtp->vv_tdata = up->usr_fsize;
	newtp->vv_nfiles = 1;
	if (up->usr_rflg & RFLG_DELETED) {
		newtp->vv_deldata = up->usr_fsize;
		newtp->vv_ndel = 1;
	}
	else {
		newtp->vv_deldata = 0;
		newtp->vv_ndel = 0;
	}

	ENQ(newtp);
}

/*
**  FUNCTION:	build_list()
**  		Build a linked list of vvs.  Note Primary or Secondary,
**		and keep track of block offsets for reporting total
**		number of bytes on vv.
*/
void
build_list(USR_FILEINFO *up)
{
	vvfact_t		*tp;

	if (!QHEAD) {
		add_vvs(up);
		return;
	}

	/*
	**  -  traverse list to check for dup vvs
	**  -  add on new file size to vv data accumulator
	**  -  when new vvs are found, add them to the list
	*/
	for (tp = QHEAD; tp != NULL; tp = QNEXT(tp)) {

		if (strcmp(tp->vv_name, up->usr_vvname) == 0) {
			tp->vv_tdata += up->usr_fsize;
			tp->vv_nfiles++;
			if (up->usr_rflg & RFLG_DELETED) {
				tp->vv_deldata += up->usr_fsize;
				tp->vv_ndel++;
			}
		} else {
			if (QNEXT(tp) == NULL) {
				add_vvs(up);
				break;
			}
		}
	}
	return;
}

/*
**  FUNCTION:	prt_vvs()
**		Lists vvs which are in use by user
**		and approximate total bytes on vv.
*/
void
prt_vvs()
{
	vvfact_t	*tp;

	printf("VV name\t\tFiles\t\t\tData\n");
	printf("\t\ttotal\tdeleted\t\ttotal\t\tdeleted\t\t%%dead\n");
	for (tp=QHEAD; tp != NULL; tp=QNEXT(tp)) {
		printf("%s\t\t%u\t%u\t\t%-12lld\t%-12lld\t%d\n", tp->vv_name,
			tp->vv_nfiles, tp->vv_ndel,
			(long long)tp->vv_tdata, (long long)tp->vv_deldata,
			(int)(100.0*(float)tp->vv_deldata/(float)tp->vv_tdata));
	}
}

/*
**  FUNCTION:	free_list()
*/
void
free_list()
{
	vvfact_t		*tp;

	for (tp=QHEAD; tp != NULL; tp=QNEXT(tp)) {
		DQ(tp);
		nFree(tp);
	}
}

/*
**  FUNCTION:	prt_filevols()
**		Lists all vvs where a file is archived
**		and the sequence number of the file on vv.
**
**	NOTE:	This routine will print out ALL instances of
**		a file on a vv.
*/		
void
prt_filevols(USR_FILEINFO *up, char *wildcard)
{

	if (wildcard != NULL) {
		printf("vv: %s  Fno: %d \tFile: %s\n",
				up->usr_vvname, up->usr_fno,
				up->usr_fname);
	} else {
		printf("vv: %s  Fno: %d\n",
				up->usr_vvname, up->usr_fno);
	}
}

/*
**  FUNCTION:	read_recs()
**		Read database records and act on them
**		according to mask and options.
*/
void
read_recs(char *usrname, USR_FILEINFO *uop, unsigned int mask)
{
	USR_FILEINFO	*uP;
	unsigned int	match_flag;
	int		ret;
	char		re[FLENGTH];
	char		*wildcard;	/* wild card flag                */
	char		*errstr;
	extern	char	*re_comp(const char *s);

	if ( (ret = rashdb_open(usrname, O_DB_READ)) != RASHDB_OK ) {
		fprintf(stderr, "open failed for user %s\n", usrname);
		exit(ret);
	}

	if (mask & USR_FNAME) {
		wildcard = strpbrk(uop->usr_fname,"?*");
		if (wildcard != NULL) {
			int	n = 0, i = 0;

			if (uop->usr_fname[0] != '*')
				re[i++] = '^';

			while (uop->usr_fname[n]) {
				if (uop->usr_fname[n] == '*') {
					re[i++] = '.';
					re[i++] = '*';
					n++;
				} else if (uop->usr_fname[n] == '?') {
					re[i++] = '.';
					n++;
				} else if (uop->usr_fname[n] == '.') {
					re[i++] = '\\';
					re[i++] = '.';
					n++;
				} else
					re[i++] = uop->usr_fname[n++];
			}
			if (re[i-1] != '*')
				re[i++] = '$';
			re[i] = '\0';
			if ((errstr = re_comp(re)) != (char *)NULL) {
				fprintf(stderr,"%s\n", errstr);
				exit(1);
			}
		}
	}

	rashdb_firstrec();
	for (;;) {
		match_flag = 0;

		uP = rashdb_nextrec(&ret, USR_ALL);
		if (ret == RASHDB_EOF)
			break;
		if (ret != RASHDB_OK) {
			fprintf(stderr, "Dbase error encountered: Quitting\n");
			break;
		}

		if (uP->usr_version < 300) {	/* 3.0.0 */
			fprintf(stderr, "cannot do version %d\n",
				uP->usr_version);
			abort();
		}

		if (Tflag) {
			build_list(uP);
			continue;
		}

		if (mask & USR_MTIME) {
			if (comp_time(uP->usr_mtime, uop->usr_mtime) != 0)
				continue;
			match_flag |= USR_MTIME;	
		}
		if (mask & USR_ARCTM) {
			if (comp_time(uP->usr_arctm, uop->usr_arctm) != 0)
				continue;
			match_flag |= USR_ARCTM;
		}
		if (mask & USR_CTIME) {
			if (comp_time(uP->usr_ctime, uop->usr_ctime) != 0)
				continue;
			match_flag |= USR_CTIME;
		}
		if (mask & USR_BFID) {
			if (memcmp(&uP->usr_bfid,
					&uop->usr_bfid,
					sizeof(bfid_t)) == 0)
				match_flag |= USR_BFID;
			else
				continue;
		}
		if (mask & USR_FNAME) {
			if (wildcard != NULL) { 
				if (re_exec(uP->usr_fname) == 0)
					continue;
				match_flag |= USR_FNAME;
			} else {
				if (strcmp(uP->usr_fname, uop->usr_fname) != 0)
					continue;
				match_flag |= USR_FNAME;
			}
		}
		if (mask & USR_VVNO) {
			if (uP->usr_vvno != uop->usr_vvno)
				continue;
			match_flag |= USR_VVNO;
		}
		if (mask & USR_GROUP) {
			if (uP->usr_group_no != uop->usr_group_no)
				continue;
			match_flag |= USR_GROUP;
		}
		if (mask & USR_VV) {
			if (strcmp(uP->usr_vvname, uop->usr_vvname) != 0)
				continue;
			match_flag |= USR_VV;
		}
		if (match_flag == mask) {
			if (rec_flag)
				print_rec(usrname, uP);
			if (Vflag) {
				printf("File Number: %d   Filename: %s\n",
					uP->usr_fno, uP->usr_fname);
			} else if (Fflag)
				prt_filevols(uP, wildcard);
		}
	}
	rashdb_close();

	if (Tflag) {
		prt_vvs();
		free_list();
	}
}

/*
**  FUNCTION:	conv_time()
**		Convert time values from char string to time_t.
*/
time_t
conv_time(char *bp)
{

	if (*bp == '<') {
		tmflag = -1;
		bp++;
	} else if (*bp == '>') {
		tmflag = 1;
		bp++;
	} else
		tmflag = 0;

	return(TimeStr(bp));
}

int
main(int argc, char *argv[])
{
	unsigned int	mask=0;		/* flag mask for different options */
	USR_FILEINFO 	uopt;		/* rash data base record structure */
	int			c, errflg=0;
	extern char		*optarg;
	extern int		optind;

	(void)signal(SIGHUP, bomb);
        (void)signal(SIGINT, bomb);
        (void)signal(SIGTERM, bomb);

	uopt.usr_fname = (char *)nAlloc(FLENGTH);

	/* parse command line  */
	while ((c = getopt(argc, argv, "lTg:b:f:v:s:m:a:c:V:F:")) != EOF) {

		switch (c) {
		 
		  case 'l':
			lflag++;
			rec_flag = 1;
			break;
		  case 'b': 
			mask |= USR_BFID;
			if (Hex2Bin(&uopt.usr_bfid, sizeof(bfid_t),
					optarg, BFIDLEN))
				diefast();
			rec_flag = 1;
			break;
		  case 'f':
			mask |= USR_FNAME;
			strcpy(uopt.usr_fname, optarg);
			rec_flag = 1;
			break;
		  case 'v':
			mask |= USR_VV;
			strcpy(uopt.usr_vvname, optarg);
			rec_flag = 1;
			break;
		  case 's':
			mask |= USR_VVNO;
			uopt.usr_vvno = atoi(optarg);
			rec_flag = 1;
			break;
		  case 'g':
			mask |= USR_GROUP;
			uopt.usr_group_no = atoi(optarg);
			rec_flag = 1;
			break;
		  case 'm':
			mask |= USR_MTIME;
			if ((uopt.usr_mtime = conv_time(optarg)) < 0)
				diefast();
			rec_flag = 1;
			break;
		  case 'a':
			mask |= USR_ARCTM;
			if ((uopt.usr_arctm = conv_time(optarg)) < 0)
				diefast();
			rec_flag = 1;
			break;
		  case 'c':
			mask |= USR_CTIME;
			if ((uopt.usr_ctime = conv_time(optarg)) < 0)
				diefast();
			rec_flag = 1;
			break;
		  case 'T':
			Tflag++;
			rec_flag = 0;
			TFQ.q_head = NULL; 		/* init TFQ */
			TFQ.q_tail = NULL;
			break;
		  case 'V':
			Vflag++;
			rec_flag = 0;
			mask |= USR_VV;
			strcpy(uopt.usr_vvname, optarg);
			break;
		  case 'F':
			Fflag++;
			rec_flag = 0;
			mask |= USR_FNAME;
			strcpy(uopt.usr_fname, optarg);
			break;
		  case '?':
		  default:
			errflg++;
			break;
		}
	}

	if (errflg)
		usage();

	if (Tflag && (optind > 2))
		usage();

	if ((Vflag || Fflag) && (optind > 3))
		usage();

	if (optind == argc)
		usage();

	/* read records from btree data base */
	for (c=optind; c<argc; c++)
		read_recs(argv[c], &uopt, mask);
	return 0;
}
