/*
 *  vchmod command -- user interface to change permissions on volumes in the
 *	volman(8) database.
 *
 *
 *	Bill Ross, NASA Ames, March 19992
 */


#include <stdio.h>
#include <signal.h>
#include <pwd.h>
#include <ctype.h>

#include "cl_space.h"

extern int errno;

short	subflag = 0666; 	/* default: subtract no perms */
short	addflag = 0;		/* default: add no perms */

/*
 * Print out usage message and exit.
 */
 void
usage(char *arg0)
{
	fprintf(stderr,
	  "usage: %s {[u|g|o|a]{+|-|=}{r|w}} | mod volsers \n", arg0);
	fprintf(stderr, "\tmod: only read/write permissions (2, 4, or 6) allowed\n");
	exit(1);
}


/*
 * subadd: update 'subtract' and 'add' permissions flags
 */
 void
subadd(short who, short mod, int op, int clear)
{
	short flag;

	/*
	 * subflag is to have 0's for perms to be subtracted: 
	 *	will '&' with volsers' mode
	 * addflag is to have 1's for perms to be added:
	 *	will '|' with volser's mode
	 * these flags update each other to be consistent, 
	 *	so the order they are applied to the volser's mode
	 *	won't matter
	 */	
#ifdef DEBUG
	printf("who %o mod %o op %o clear %o\n", who, mod, op, clear);
#endif
	if (who == 0)				/* default: a */
		who = 0666;
	flag = who & mod;

	if (clear) {				/* '=' */
		subflag	&= ~who;
		addflag &= ~who;
		subflag	|= flag;
		addflag |= flag;
	}
	if (op < 0) {				/* '-' */
		subflag &= ~flag;
		addflag &= subflag;
		return;
	}
	/* '+' */
	addflag |= flag;
	subflag |= addflag;
}

main(int argc, char *argv[])
{
	int		cc, op = 0, clear = 0, rval = 0;
	short		flag = 0, who = 0, mod = 0;
	char		*arg0 = argv[0];
	char		*ptr;
	char		pbuf[PKTSIZ(vmpVsn_t)];
	vmpHdr_t	*shp;
	vmpVsn_t	*vp;
	
	if (argc < 3)
		usage(arg0);

	/*
	 *  examine 1st arg
	 */

	argv++;
	if (vm_octalperm(*argv, &flag) != -1) {

		addflag = flag;
		subflag = flag;
	} else {

		/*
		 *  got a non-octal arg: process the [ugoa]{+-=}{rw}, ...
		 *	stuff
		 */

		int i;
		int opd = 0, oprw = 0;
	    	for (i=0; ((cc = argv[0][i]) != (int)NULL); i++) {
			switch (cc) {
		  	case 'u':				
				who |= 0600;
				break;
		  	case 'g':			
				who |= 060;
				break;
		  	case 'o':	
				who |= 06;
				break;
		  	case 'a':	
				who |= 0666;
				break;
		  	case '+':
				if (oprw) {
					subadd (who, mod, op, clear);
					oprw = 0;
					op = 0;
					opd = 0;
					clear = 0;
					mod = 0;
				}
				opd++;
				op++; 
				break;
		  	case '-':
				if (oprw) {
					subadd (who, mod, op, clear);
					oprw = 0;
					op = 0;
					opd = 0;
					clear = 0;
					mod = 0;
				}
				opd++;
				op--;
				break;
		  	case '=':
				if (oprw) {
					subadd (who, mod, op, clear);
					oprw = 0;
					op = 0;
					opd = 0;
					clear = 0;
					mod = 0;
				}
				opd++;
				op = 0;
				clear++;
				break;
		  	case 'r':
				if (!opd)
					usage(arg0);
				oprw++;
				mod |= 0444;
				break;
		  	case 'w':
				if (!opd)
					usage(arg0);
				oprw++;
				mod |= 0222;
				break;
		  	case ',':
				if (!opd)
					usage(arg0);
				subadd (who, mod, op, clear);
				op = 0;
				opd = 0;
				oprw = 0;
				clear = 0;
				mod = 0;
				who = 0;
				break;
			
		  	case '?':
		  	default:
				usage(arg0);
			}
	    	}
	    	subadd (who, mod, op, clear);
		if (addflag == 0  &&  subflag == 0666) {
			fprintf(stderr, "(null change specified)\n");
			exit(1);
		}
	}
#ifdef DEBUG
	printf("%o %o\n", addflag, subflag);
#endif

	/*
	 *  connect to vold & init packet
	 */

	if (vm_init(arg0)) {
		vm_prerr(stderr);
		exit(VMError);
	}
	shp = vBuildHdr(M_VCHMOD, vGetSeq(), (vmpHdr_t *) pbuf);
	vp = MSGPTR(shp, vmpVsn_t);

	vp->v_flag = addflag;
	vp->v_flag2 = subflag;
	
	/*
	 *  last arg(s) = volume serial number(s) (volser): try to vchmod them
	 */

	for (++argv; *argv != NULL; argv++) {
		if (checkvsn(*argv)) {
			printf("\n%s\n", nErrStr);
			rval++;
			continue;
		}

		printf("sending %s\n", *argv);
		memcpy(&vp->v_vsn, *argv, VSNLEN);
		if (lSnd2RP(shp)) {
			vm_prerr(stderr);
			exit(VMError);
		}
		rval += GetResponse(*argv, shp->vh_seq);
		shp->vh_seq++;
	}
	vm_term();
	exit(rval);
}

int
GetResponse(char *vol, Seq_t seq)
{
	vmpHdr_t	*rhp;
	vmpVsn_t	*vsp;
	vmpError_t	*ep;

	for (;;) {
		if ((rhp = lRcvFromRP(IPC_DELAY)) == NULL) {
			vm_prerr(stderr);
			exit(VMError);
		}
		if (seq != rhp->vh_seq  &&  rhp->vh_type != M_SHUTDOWN) {
			fprintf(stderr, "bad packet sequence (skipping): %s", 
								vHdrStr(rhp));
			continue;
		}
		switch (rhp->vh_type) {
		    case M_VCHMOD:
			vsp = MSGPTR(rhp, vmpVsn_t);
			if (strncmp((char *)&vsp->v_vsn, vol, VSNLEN)) {
				printf("%.6s: wrong vsn returned\n",
							&vsp->v_vsn);
				exit(1);
			}
			/* printf("%.6s ok\n"); */
			nFree(rhp);
			return(0);
		    case M_VERROR: 
			ep = MSGPTR(rhp, vmpError_t);
			printf("error from vold: %s\n",ep->vn_msg);
			if (ep->vn_error != E_USER)
				exit(1);
			nFree(rhp);
			return(1);
		    case M_SHUTDOWN:
			printf("shutdown received\n");
			vm_term();
			exit(1);
		    case M_NOP:
			nFree(rhp);
			continue;
		    default:
			printf("Unexpected packet: %s\n", vHdrStr(rhp));
			exit(1);
		}
	}
}
