/*
 *	NAStore vlabel, vwlabel utility - read header label on a
 *		mounted tape and in the volman(8) database (vlabel),
 *		and write same (vwlabel).
 *
 *		2 entry points - reading & writing: see READENTRY 
 *		& WRITEENTRY #defines
 *
 *		all labels considered ANSI-standard ASCII:
 *			col 1-4 = "VOL1"
 *			col 5-10 = 6 char volume serial #
 *			col 80  = (optional, = version of standard)
 *			2 tape marks for end-of-data
 *
 *		update/check volman(8) database
 *			writing: Verify that user has write permission on tape 
 *				and that tape owner doesn't have duplicate 
 *				internal label. Update internal label in dbase,
 *				set status to SCRATCH if VIRGIN or RECYCLE.
 *			reading: look up & print dbase version of intlabel too.
 *
 *		Note that status and RC names from the volume manager are
 *		hard coded into arrays here.
 *
 *
 * 	Bill Ross, NASA Ames Research Center, Feb 1991
 */

#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>

#include "cl_space.h"


#define READENTRY	"vlabel"
#define WRITEENTRY	"vwlabel"

extern int	errno;

static  char     *cmdname;

static  int     writeit = 0;
static  int     dbupdate = 0;
static	int	dbasetoo;

/*
 *  usage() - print out usage message and exit.
 */
 void
usage()
{
	if (!strcmp(cmdname, READENTRY)) 
		fprintf(stderr, 
			"usage: %s [-d] {vsn|dev}\n", cmdname);
	else if (!strcmp(cmdname, WRITEENTRY)) {
		fprintf(stderr, 
			"usage: %s [-i internal] vsn\n", cmdname);
		fprintf(stderr, 
			"       %s  -i internal  /dev/device\n", cmdname);
	} else
		fprintf(stderr,
			"%s: unknown command name for %s/%s \n", 
			cmdname, READENTRY, WRITEENTRY);
	exit(1);
}

char    	internal[7];
int		fd;
int     	status[2];

done(rc)
{
	if (fd != -1)
		close(fd);
	if (dbasetoo)
		vm_term();
	exit(rc);
	return(rc);	/* for compiler warning */
}


main(int argc, char *argv[])
{
	char		*arg0 = argv[0];
	int     	uid;
	int     	rc = 1;		/* return code for tape op */
	int		iflag = 0;
	char		*path;
	int		cc;
	char    	tape[VSNLEN+1];
	char    	vsn[VSNLEN+1];
	extern  int     optind;
	extern  char    *optarg;
	char		pbuf[PKTSIZ(vmpLblQry_t)];
	vmpHdr_t	*hp;
	vmpLblQry_t	*lp;
	
	uid = getuid();

	cmdname = nas_basename(argv[0]);
	if (!strcmp(cmdname, WRITEENTRY))
		writeit++;
	else if (strcmp(cmdname, READENTRY))
		usage();

	internal[0] = '\0';
	while ((cc = getopt(argc, argv, "di:a")) != EOF) {
		switch (cc) {
		    case 'd':

			/*
			 * read tape label & update dbase to correspond
			 */
			if (writeit)
				usage();
			dbupdate++;
			break;

		    case 'i':

			/*
			 * use a different internal label
			 */
			if (!writeit)
				usage();
			if (checkvsn(optarg)) {
				printf("\n%s\n", nErrStr);
				exit(1);
			}
			strncpy(internal, optarg, VSNLEN);
			iflag++;
			break;

		    case '?':
		    default:
			usage();
		}
	}

	nLogInitStderr(EL_ERROR1);

	argc -= optind;
	argv += optind;
	
	if (argc != 1)
		usage();
		
	path = *argv;

	/* 
	 * dbasetoo is a flag to indicate whether we have a vsn so we can
	 * operate on the dbase; it is false if device rather than vsn
	 * is given
	 */ 

	dbasetoo = strncmp(path, "/dev/", 5); 

	if (!dbasetoo  &&  uid != 0) {
		fprintf(stderr, 
			"Permission denied: must be root to access dev\n");
		exit(1);
	}
	if (!dbasetoo && dbupdate) {
		fprintf(stderr, 
			"%s: must use VSN argument to update database\n",
				cmdname);
		exit(1);
	}

	if (!dbasetoo && writeit && !iflag) {
		fprintf(stderr, 
		 "%s: Must use '-i internal' option when writing device\n",
			cmdname);
		exit(1);
	}

	if (dbasetoo) {
		if (checkvsn(nas_basename(path))) {
			printf("\n%s\n", nErrStr);
			exit(1);
		}
		strcpy(tape, nas_basename(path));
	}
	strncpy(vsn, nas_basename(path), VSNLEN);

	if (writeit && internal[0] == '\0')
		strncpy(internal, nas_basename(path), VSNLEN);
		
	if (writeit  &&  !dbasetoo   &&  uid == 0) {

		/*
		 *  consider the consequences of odd request
		 */

		if (isatty(fileno(stdin))  &&  isatty(fileno(stderr))) {
			fprintf(stderr, "%s %s \n%s",
			    "Writing internal label could lead to volman(8)",
			    "database inconsistency,\nsince external label",
			    "unknown. Continue? (Y/N) ");
			if (getYorN() == 'n')
				exit(0);
		} else
			fprintf(stderr, "%s %s %s\n",
			    "warning: writing internal label", internal,
			    "without updating volman(8) database");
	}

	/*
	 *  open the tape device - this constitutes the only permission
	 *	checking. If the open fails, the dbase can still be checked 
	 *	(but not written). If the open succeeds, position tape
         */
	fd = opendev(path);
	if (fd == -1  &&  (writeit  ||  dbupdate))
		exit(1);

	if (fd != -1) {
#ifdef AMDAHL_UTS
 		/*
		 *  clear the label bit, saving it, & rewind tape
		 */

		if (ioctl(fd, TAPGTTY, status) == -1) {
			perror("getting device status");
			exit(1);
		}
		if (ioctl(fd, TAPMOTION, status[1] & ~LABELED) == -1) {
			fprintf(stderr,
				"%s: unable to change tape state, errno = %d\n",
                        	cmdname, errno);
			exit(1);
		}
#endif
		if (ntp_taprew(fd, vsn, 0) == -1)
			exit(1);
	}

	/*
	 * assume that the user's access to tape means dbase update ok
	 * 	if requested
	 */

	if (dbasetoo) {

		/* 
		 *  contact vold
		 */
		if (vm_init(arg0)) {
			vm_prerr(stderr);
			exit(VMError);
		}

		/*
		 *  set up send pkt
		 */
		hp = vBuildHdr(M_LBLQRY, vGetSeq(), (vmpHdr_t *) pbuf);
		lp = MSGPTR(hp, vmpLblQry_t);
		lp->vq_obsolete = 0;
		memcpy(&lp->vq_vsn, tape, VSNLEN);
		memset(&lp->vq_label, ' ', VSNLEN);
	}
	if (writeit) {
		char	vol1buf[VOL1LEN];

		/*
		 *  check if valid label on tape already
		 */
		if (read(fd, vol1buf, VOL1LEN) == VOL1LEN) {
			if (!strncmp(vol1buf, "VOL1", 4)) {
				fprintf(stderr, "old label: %.*s\n", 
						VSNLEN, vol1buf+4);
				if (isatty(fileno(stdin))
					    && isatty(fileno(stderr))) {
					fprintf(stderr, "overwrite? (y/n) ");
					if (getYorN() == 'n')
						done(0);
				}
			}
		} 
		if (ntp_taprew(fd, vsn, 0) == -1)
			exit(1);

		if (dbasetoo) {

			/*
			 *  send request for authorization & get reply
			 */

			memcpy(&lp->vq_label, internal, VSNLEN);

			/*
			 *  send pkt & get reply
			 */
		
			if (lSnd2RP(hp)) {
				done(1);
			}
			if (GetResponse(tape, hp->vh_seq))
				done(1);
		}

		/*
		 *  now write label
		 */
		
		rc = vm_writelabel(fd, internal, vsn, cmdname);

	} else if (fd != -1) {	
	
		/* 
		 *  read label & query dbase
		 */
		rc = vm_readlabel(fd, internal, cmdname);
		if (!rc  &&  dbupdate) {

			/*
			 *  add "update" msg to packet in form of non-blanks
			 */

			if (internal[0] != '\0')
				memcpy(&lp->vq_label, internal, VSNLEN);
			else 
				memset(&lp->vq_label, 0, VSNLEN);
		}
		

		if (dbasetoo) {
			int	trc;
			/*
			 *  send pkt & get reply
			 */
			if (lSnd2RP(hp))
				done(1);

			if (GetResponse(tape, hp->vh_seq))
				done(1);

			printf("label according to volman database: (%s)\n",
					internal[0] ? internal : "none");
			if (!rc  &&  dbupdate)
			  printf("Updated database to reflect label on tape\n");
		}
	}
	done(0);
}

/*
 *  opendev() - open tape device
 */

opendev(char *path)
{
	int     fd;
	char    fullpath[256];

	/*
	 *  get the full path from root
	 */

	if (path[0] == '/') {

		/*
		 *  it's either /dev/vol/xxxxxx or /dev/rmt/cuu
		 */

		strcpy(fullpath, path);
	} else {

		/*
		 *  assume it's a vsn
		 */

		sprintf(fullpath, "%s/%s", vNodeDir, path);
	}
	if ((fd = ntp_open(fullpath, writeit, 0, 1, EL_TOSTDERR)) == -1)  {
		if (errno == ENOENT)
			fprintf(stderr, 
			    "(Volume must be mounted to access label)\n");
	}
	return(fd);
}


  int
GetResponse(char *vol, Seq_t seq)
{
	char		*id = "GetResponse";
	vmpHdr_t	*rhp;
	vmpLblQry_t	*lp;
	vmpError_t	*ep;

	for (;;) {
		if ((rhp = lRcvFromRP(IPC_DELAY)) == NULL) {
			done(1);
		}
		if (seq != rhp->vh_seq) {
			fprintf(stderr, "wrong seq\n");
			return(1);
		}
		switch (rhp->vh_type) {
		    case M_LBLQRY:
			lp = MSGPTR(rhp, vmpLblQry_t);
			if (strncmp((char *)&lp->vq_vsn, vol, VSNLEN)) {
				fprintf(stderr, "%.*s: wrong vsn returned\n",
							VSNLEN, &lp->vq_vsn);
				return(1);
			}
			strncpy(internal, (char *)&lp->vq_label,VSNLEN);
			return(0);
		    case M_VERROR: 
			ep = MSGPTR(rhp, vmpError_t);
			fprintf(stderr, "error from vold: %s\n",ep->vn_msg);
			return(1);
		    case M_SHUTDOWN:
			fprintf(stderr, "shutdown received\n");
			vm_term();
			return(1);
		    case M_NOP:
			nFree(rhp);
			continue;
		    default:
			fprintf(stderr, "Unexpected packet: %s\n", 
							vHdrStr(rhp));
			return(1);
		}
	}
}
