/*
 *  Volume Manager
 *
 *  cl_lib.c: Volume Manager client's protocol library 
 */

#include <ctype.h>
#include <signal.h>
#include <setjmp.h>

#include <pwd.h>
#include <grp.h>

#include "cl_space.h"
#include "nas_tape.h"


extern CodeString_t	vStatTab[];
extern DevEnt_t		*DriveArray;
extern int		DriveTotal;

extern int		vDeadWriter, vDeadReader;

static char		cmdnam[LINELEN];

int			lUid, lPid;

void	lProcPkts();
void	lPurgeCVReqs();

  int
defaultfatal()
{
	fprintf(stderr, "fatal error\n");
	exit(1);
	return(1);	/* avoid dumb compiler msg */
}

/*
 *  vm_init(): initialize pvrl (client connection), simple verwsion
 *	used by majority of applications that don't want to
 *	worry about all the options to nLogInit()
 */
  int
vm_init(char *id)
{
	/*
	 *  prepare for logging - all msgs go (at least) to stderr
	 */
	nLogInit(id, defaultfatal); 
	nLogInitStderr(EL_LEVEL);

	/*
	 *  start client connection to vold
	 */
	return (_vm_init(id, 0));
}

/*
 *  vm_initp(): initialize client connection, simple version
 *	used by applications that required priority status.
 *      Priority status allows the use of reserved connections
 *	in the event all the RP's normal ones have been exhausted. 
 */
  int
vm_initp(char *id)
{
	/*
	 *  prepare for logging - all msgs go (at least) to stderr
	 */
	nLogInit(id, defaultfatal); 
	nLogInitStderr(EL_LEVEL);
	return (_vm_init(id, VO_PRIORITY));
}


#if 0
static int	(*savedSigPipe)() = vm_init;	/* vm_init => not set yet */
#endif
static int	InitCalled = 0;

/*
 *  _vm_init(): initialize pvrl, power-user version;
 *	requires that nLogInit() called before _vm_init()
 *
 *	1) initialize data structures
 *	2) initialize connections to RP/VDM
 */
  int
_vm_init(char *id, int flag)
{

	strncpy(cmdnam, id, sizeof (cmdnam));
	if (InitCalled++) {
		nLogMsg(EL_ERROR1, NULL, "vm_init() called more than once");
	}
#ifdef DEBUG
	setbuf(stderr, NULL);
#endif
	vInitDirNames();
	vClearError();
	lPid = getpid();
	lUid = geteuid();	/* We could be called from a SUID */
				/* program so we need geteuid.  TMP */

	if (signal(SIGPIPE, vSigPipe) == SIG_ERR)
		nLogMsg(EL_FATAL|EL_SYS, id, "signal(SIGPIPE)");

	ClientInitIPC(flag, EL_CODE);
	if (VMError != E_NOERROR)
		InitCalled--;
	return (VMError);
}


/*
 *  vm_term() - terminate pvrl connections
 */
  int
vm_term()
{
	ClientTermIPC(NEWERROR);
	InitCalled = 0;
	return (VMError);
}


/*
 *  vm_prerr() - print all info about the current error condition on str
 */
  void
vm_prerr(FILE *fp)
{
	extern char	*vInstErrBuf;		/* see vInstallError() */

	fprintf(fp, "%s: %s", cmdnam, vErrStr(VMError));
	if (nErrStr && *nErrStr && nErrStr == vInstErrBuf)
		fprintf(fp, ": [%s]", nErrStr);
	fputc('\n', fp);
}


QHead_t		CVolReqQ;	/* queue of all requests sent out by this AC */
static int	vo_sig = 0;	/* signal for VO_SIGNAL if valid */

/*
 *  vm_signal() - register a signal for VO_SIGNAL
 */
  int
vm_signal(int sig)
{
	if (sig <= 0 || sig >= NSIG) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR1, "vm_signal", 
					"%d: bad signal\n", sig);
		vo_sig = 0;
		return (VMError);
	}
	vo_sig = sig;
	return (E_NOERROR);
}


/*
 *  vm_mount() - makes all the rest worthwhile
 *
 *	vsnp		name of volume to be mounted
 *	flags		see vm_volop.h for flag definitions
 *	rid		reservation id (optional)
 *	tag		optional tag field for volstat(1)
 *	waitnow		wait for vold response
 *	delmsg		when status rcvd, delete from msg buffer 
 */
  int
vm_mount(vsn_t *vsnp, int flags, char *rid, char *tag, 
					int waitnow, int delmsg)
{
	char		*id = "vm_mount";
	char		pbuf[PKTSIZ(vmpUsrMount_t)];
	vmpHdr_t	*hp;
	vmpUsrMount_t	*mntp;
	size_t		ridlen;

	vClearError();			/* clear global errors */
	if (CheckVO_Flags(flags))
		return(VMError);
	if ((flags & VO_ASYNC) && waitnow) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR1, id, 
					"coding error: waiting on ASYNC req");
		return (VMError);
	}

	if (rid) {
		ridlen = strlen(rid);
		if (ridlen > RIDLEN) {
			VMError = E_USER;
			nLogMsg(EL_CODE|EL_ERROR2, id, 
				"reservation id length %d - max %d",
						ridlen, RIDLEN);
			return (VMError);
		}
	} else {
		ridlen = 0;
	}

	/*
	 *  get the pkt (memory is on stack, so no need to free) 
	 *	& fill it in
	 */
	hp = vBuildHdr(M_USRMOUNT, vGetSeq(), (vmpHdr_t *)pbuf);
	mntp = MSGPTR(hp, vmpUsrMount_t);
	mntp->vu_vsn = *vsnp;
	mntp->vu_flags = flags;
	memset(mntp->vu_rid, '\0', sizeof(mntp->vu_rid));
	if (ridlen)
		memcpy(mntp->vu_rid, rid, sizeof(mntp->vu_rid));
	copytag(mntp->vu_tag, tag);

	/*
	 *  see if user wants a notification via signal
	 */
	if (flags & VO_SIGNAL) {
		/* if VO_SIGNAL and valid vo_sig, install vo_sig in flags */
		if (vo_sig) {
			flags &= ~VO_SIGMASK;
			flags |= VO_SIGSET(vo_sig);
		} else {
			VMError = E_USER;
			nLogMsg(EL_CODE|EL_ERROR2, id, 
			  "VO_SIGNAL flag set, but no vm_signal() not called");
			return (VMError);
		}
	}

	/*
	 *  Built request, now send it 
	 */

	if (lSnd2RP(hp)) {
		nLogMsg(EL_DEBUG2, id, "lSnd2RP failed VMError = %d",VMError);
		return (VMError);
	}
	nLogMsg(EL_DEBUG2, id, "SENT %s %s/%s", vHdrStr(hp),
		(mntp->vu_flags & VO_ASYNC ? "async" : "sync"),
		(waitnow ? "wait" : "nowait"));

	/*
	 *  make sure the request was initially accepted - get ACK
	 */
	if (vm_getstat(vsnp, hp->vh_seq, 1, 1, NULL, 1) == NULL)
		return(VMError);

	if (waitnow) {
		/* wait for status to arrive */
		nLogMsg(EL_DEBUG4, id, "Calling vm_getstat");
		(void) vm_getstat(vsnp, hp->vh_seq, 1, 0, NULL, delmsg);
	}

	nLogMsg(EL_DEBUG4, id, "Returning to caller VMError=%d",VMError);
	return (VMError);
}

/*
 *  vm_attach() - glom onto async mount
 */
  int
vm_attach(vsn_t *vsnp, int flags, int delmsg)
{
	char		*id = "vm_attach";
	char		pbuf[PKTSIZ(vmpVsn_t)];
	vmpVsn_t	*vsp;
	vmpHdr_t	*hp;

	vClearError();			/* clear global errors */
	if (flags != 0  &&  flags != M_USRMOUNT  &&  flags != M_MOVE) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR1, id, 
					"coding error: bad flag 0x%x", flags);
		return (VMError);
	}

	hp = vBuildHdr(M_ATTACH, vGetSeq(), (vmpHdr_t *) pbuf);
	vsp = MSGPTR(hp, vmpVsn_t);
	vsp->v_vsn = *vsnp;
	vsp->v_flag = flags;

	/*
	 * Built request, now send it
	 */

	if (lSnd2RP(hp)){
		nLogMsg(EL_DEBUG2, id, "lSnd2RP failed VMError = %d",VMError);
		return (VMError);
	}
	nLogMsg(EL_DEBUG2, id, "SENT %s", vHdrStr(hp));

	/* 
	 *  wait for status to arrive
	 */
	nLogMsg(EL_DEBUG4, id, "Calling vm_getstat");
	(void) vm_getstat(vsnp, hp->vh_seq, 1, 0, NULL, delmsg);

	nLogMsg(EL_DEBUG4, id, "Returning to caller VMError=%d",VMError);
	return (VMError);
}

/*
 *  vm_mount_pll() - mount a group of vsns at once - fail if any fail
 */
  int
vm_mount_pll(unsigned int nvsns, char **vsns, int flags, char *tag)
{
	char		*id = "vm_mount_pll";
	size_t		psize, i;
	vmpHdr_t	*hp, *rhp;
	vmpUsrMountN_t	*mp;
	char		*pktvsn;

	vClearError();			/* clear global errors */

	/*
	**	check args
	*/
	if (nvsns < 1) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR1, id, "coding error: nvsns < 1");
		return(VMError);
	}
	if (flags & (VO_ASYNC|VO_SIGNAL|VO_EXCLUSIVE)) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR1, id, 
			"coding error: ASYNC, SIGNAL, EXCLUSIVE not permitted");
		return (VMError);
	}
	if (CheckVO_Flags(flags))
		return(VMError);

	/*
	**	get pkt buf & set pointers into it
	*/
	psize = PadSize(sizeof(vmpHdr_t)) 
		+ sizeof(vmpUsrMountN_t)
		+ nvsns * VSNLEN;
	hp = (vmpHdr_t *) nAlloc(psize);
	mp = MSGPTR(hp, vmpUsrMountN_t);
	pktvsn = (char *) mp + sizeof(vmpUsrMountN_t);

	/*
	**	fill in pkt
	*/
        hp->vh_magic = VMP_MAGIC;
        hp->vh_type = M_USRMOUNT_PLL;
        hp->vh_seq = vGetSeq();
	hp->vh_size = (unsigned int)sizeof(vmpUsrMountN_t) + nvsns * VSNLEN;
	hp->vh_count = nvsns;
	
	mp->vx_flags = flags;
	copytag(mp->vx_tag, tag);

	for (i=0; i<nvsns; i++, pktvsn+=VSNLEN) {
		memcpy(pktvsn, vsns[i], (size_t)VSNLEN);
	}

	/*
	**	send pkt
	*/
	if (lSnd2RP(hp)) {
		nFree(hp);
		return(VMError);
	}

	/*
	**	get response
	*/
	for (;;) {
		rhp = lRcvFromRP(IPC_DELAY);
		if (rhp == NULL) {
			nFree(hp);
			return(VMError);
		}
		switch(rhp->vh_type) {
		    case M_ACK:
			break;
		    case M_VERROR:
			vInstallError(MSGPTR(rhp, vmpError_t));
			break;
		    case M_SHUTDOWN:
			VMError = E_UNAVAIL;
			nLogMsg(EL_CODE|EL_ERROR1, 0, "Vold shutdown");
			break;
		    case M_NOP:
			nFree(rhp);
			continue;
		    default:
			VMError = E_INTERNAL;
			nLogMsg(EL_ERROR1, 0, "Unexpected packet: %s", 
							vMsgType(rhp->vh_type));
		}
		nFree(rhp);
		return(VMError);
	}
}

/*
 *  vm_mount_sel() - mount the 1st possible of a group of vsns
 *	print result to stdout for now - just used for testing
 */
  int
vm_mount_sel(unsigned int nvsns, char **vsns, int flags, char *tag)
{
	char		*id = "vm_mount_sel";
	size_t		psize, i;
	vmpHdr_t	*hp, *rhp;
	vmpUsrMountN_t	*mp;
	char		*pktvsn;

	vClearError();			/* clear global errors */

	/*
	**	check args
	*/
	if (nvsns < 1) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR1, id, "coding error: nvsns < 1");
		return(VMError);
	}
	if (flags & (VO_ASYNC|VO_SIGNAL|VO_EXCLUSIVE)) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR1, id, 
			"coding error: ASYNC, SIGNAL, EXCLUSIVE not permitted");
		return (VMError);
	}
	if (CheckVO_Flags(flags))
		return(VMError);

	/*
	**	get pkt buf & set pointers into it
	**		since vsns are char, no need to pad
	**		vmpUsrMountN_t to word boundary
	*/
	psize = PadSize(sizeof(vmpHdr_t))
		+ sizeof(vmpUsrMountN_t)
		+ (size_t)nvsns * VSNLEN;
	hp = (vmpHdr_t *) nAlloc(psize);
	mp = MSGPTR(hp, vmpUsrMountN_t);
	pktvsn = (char *) mp + sizeof(vmpUsrMountN_t);

	/*
	**	fill in pkt
	*/
        hp->vh_magic = VMP_MAGIC;
        hp->vh_type = M_USRMOUNT_SEL;
        hp->vh_seq = vGetSeq();
	hp->vh_size = (unsigned int)sizeof(vmpUsrMountN_t) + nvsns * VSNLEN;
	hp->vh_count = nvsns;
	
	mp->vx_flags = flags;
	copytag(mp->vx_tag, tag);

	for (i=0; i<nvsns; i++, pktvsn+=VSNLEN)
		memcpy(pktvsn, vsns[i], (size_t)VSNLEN);

	/*
	**	send pkt
	*/
	if (lSnd2RP(hp)) {
		nFree(hp);
		return(VMError);
	}

	/*
	**	get response
	*/
	for (;;) {
		vmpVolStat_t	*vsp;

		rhp = lRcvFromRP(IPC_DELAY);
		if (rhp == NULL) {
			nFree(hp);
			return(VMError);
		}
		switch(rhp->vh_type) {
		    case M_ACK:
			nFree(rhp);
			continue;
		    case M_VOLSTAT:
			vsp = MSGPTR(rhp, vmpVolStat_t);
			if (vm_prvolstatarr(VolStatNULL, 0, stdout, 0,0,0,0,0))
				return (VMError);
			if (vm_prvolstatarr(vsp, 1, stdout, 0,0,0,0,0))
				return (VMError); 
			break;
		    case M_VERROR:
			vInstallError(MSGPTR(rhp, vmpError_t));
			break;
		    case M_SHUTDOWN:
			VMError = E_UNAVAIL;
			nLogMsg(EL_CODE|EL_ERROR1, 0, "Vold shutdown");
			break;
		    case M_NOP:
			nFree(rhp);
			continue;
		    default:
			VMError = E_INTERNAL;
			nLogMsg(EL_ERROR1, 0, 
					"Unexpected packet (wanted ack): %s", 
					vMsgType(rhp->vh_type));
		}
		nFree(rhp);
		return(VMError);
	}
}

/*
 *  vm_unmount() - release that drive
 */
  int
vm_unmount(vsn_t *vsnp, int flags, int waitnow, int delmsg)
{
	char		*id = "vm_unmount";
	char		pbuf[PKTSIZ(vmpVsn_t)];
	vmpHdr_t	*hp;
	vmpVsn_t	*ump;
	int		type;

	vClearError();			/* clear global errors */
	if (CheckVO_Flags(flags))
		return(VMError);
	if ((flags & VO_ASYNC) && waitnow) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR1, id, 
					"coding error: waiting on ASYNC req");
		return (VMError);
	}


	if (flags & VO_REQUEUE)
		type = M_REQUEUE;
	else
		type = M_UNMOUNT;
	hp = vBuildHdr(type, vGetSeq(), (vmpHdr_t *) pbuf);
	ump = MSGPTR(hp, vmpVsn_t);
	ump->v_vsn = *vsnp;
	ump->v_flag = flags;		/* notification flags */
		
	if (flags & VO_SIGNAL) {
		/* if VO_SIGNAL and valid vo_sig, install vo_sig in flags */
		if (vo_sig) {
			flags &= ~VO_SIGMASK;
			flags |= VO_SIGSET(vo_sig);
		} else {
			VMError = E_USER;
			nLogMsg(EL_CODE|EL_ERROR2, id, 
			  "VO_SIGNAL flag set, but no vm_signal() not called");
			return (VMError);
		}
	}

	/*
	 *  Built request, now send it
	 */

	if (lSnd2RP(hp)){
		nLogMsg(EL_DEBUG2, id, "lSnd2RP failed VMError = %d",VMError);
		return (VMError);
	}
	nLogMsg(EL_DEBUG2, id, "SENT %s", vHdrStr(hp));

	if (waitnow) {
		/* wait for status to arrive */
		nLogMsg(EL_DEBUG4, id, "Calling vm_getstat");
		(void) vm_getstat(vsnp, hp->vh_seq, 1, 0, NULL, delmsg);
	}

	nLogMsg(EL_DEBUG4, id, "Returning to caller VMError=%d",VMError);
	return (VMError);
}

  void
vPrintStatTab(FILE *fptr)
{
	int	i;

	for (i=0; vStatTab[i].cs_string; i++) {
		if (vStatTab[i].cs_code == S_DELETED)
			continue;
		fprintf(fptr, "\t%c\t%s\n",
			vStatTab[i].cs_code,
			vStatTab[i].cs_string);
	}
}

#include <sys/fcntl.h>

jmp_buf	lib_jmpbuf;

/*
 *  vm_getstat() - get status for vsn (if vsnp != NULL, else for seq)
 *	1) read in all outstanding packets and update internal
 *		status table (lProcPkts())
 *	2) if ndelay, return last received *new* status for vsn if any or
 *		NULL if none
 *	3) else if delay, wait for status for this vsn to be received
 *
 *  Return value:
 *	if status received: vmpVolStat_t *
 *	if status not received: NULL, VMError = M_NOERROR
 *	if error received: NULL, VMError = vn_code
 */
static vmpVolStat_t	dummy;
  vmpVolStat_t *
vm_getstat(vsn_t *vsnp, Seq_t seq, int delay, int ack, int *countp, int delmsg)
{
	char		*id = "vm_getstat";
	CVolReq_t	*cvrp = NULL;

	vClearError();
	if (setjmp(lib_jmpbuf)) {		/* Shutdown escape hatch */
		nLogMsg(EL_DEBUG1, id, "setjmp escape");
		return (NULL);
	}

	/*
	 *  get any waiting pkts
	 */
	lProcPkts(IPC_NDELAY);
	if (vDeadWriter)
		goto deadwrtr;

	/* 
	 *  keep reading packets until we get one for vsn or seq
	 */
	for (;;) {
		/*
		 *  check to see if we have status for the requested vsn
		 * 
		 *	cvr & pkt are deleted in lProcPkts() by a 
		 *	call to lPurgeCVReqs() if cvrp->cvr_newpkt = 0;
		 */
		if (seq != (Seq_t) NOSEQ && seq != 0)
			cvrp = lGetCVRbySeq(seq);
		else
			cvrp = lGetCVRbyVSN(vsnp);
		if (cvrp && cvrp->cvr_lastpkt) {	/* have unread status */
			vmpHdr_t	*hp = cvrp->cvr_lastpkt;
			/*
			 *  if requested, mark msg for deletion
			 */
			if (delmsg)
				cvrp->cvr_newpkt = 0;

			/*
			 *  we have received a packet for seq or VSN:
			 *	if it's a vmpError_t packet, install error
			 *	info and return NULL; if it's a vmpVolStat
			 *	packet, return it, else internal error
			 *  Stop looping and return from here.
			 */
			switch (hp->vh_type) {
			    case M_ACK:
				if (ack) {
					/*
					 *  HACK HACK HACK!!!
					 *	just want to return non-null..
					 */
					return(&dummy);
				}
				VMError = E_INTERNAL;
				nLogMsg(EL_CODE|EL_ERROR1, id, 
							"unexpected ack");
				return(NULL);
			    case M_VOLSTAT:
			    case M_VOLSTATS:
				if (ack) {
					VMError = E_INTERNAL;
					nLogMsg(EL_CODE|EL_ERROR1, id, 
						"expected ack or err..volstat");
					return(NULL);
				}
				if (hp->vh_size == 0)
					return (NULL);
				if (countp != NULL)
					*countp = hp->vh_count;
				return(MSGPTR(hp, vmpVolStat_t));
			    case M_VERROR:
				vInstallError(MSGPTR(hp, vmpError_t));
				return (NULL);
			    default:
				VMError = E_INTERNAL;
				nLogMsg(EL_CODE|EL_ERROR1, id,
				  "Unexpected message: %s", vHdrStr(hp));
				return (NULL);
			}
		}

		/* if no delay, return now and don't wait */
		if (!delay) 
			return (NULL);

		/*
		 *  wait for next packet, read it in
		 *	TODO: see that only one packet is read!
		 */
		lProcPkts(IPC_DELAY);
		if (vDeadWriter) {
  deadwrtr:		VMError = E_DOWN;
			nLogMsg(EL_TOSTDERR|EL_CODE|EL_ERROR1, id,
			  "unexpected vold termination");
				return(NULL);
			
		}
	}
}


/*
 *  vm_prvolstat() - print the last status for named VSN if there is one
 */
  int
vm_prvolstat(vsn_t *vsnp, FILE *fp)
{
	vmpVolStat_t	*vsp;
	int		count;

	vClearError();
	/*
	 *  non-null &count means buffered msg will be marked
	 *	for deletion next lProcPkts()
	 */
	if ((vsp = vm_getstat(vsnp, (Seq_t) 0, 0, 0, &count, 1)) == NULL)
		return (VMError);
	if (vm_prvolstatarr(VolStatNULL, 0, fp, 0, 0, 0, 0, 0)) /* print hdr */
		return (VMError);
	if (vm_prvolstatarr(vsp, 1, fp, 0, 0, 0, 0, 0))
		return (VMError);
	return (VMError);
}



typedef struct {
	vsn_t		vsn;
	vmpVolStat_t	*addr;
} vsnsort_t;

/*
 *  comp() - decide order of the two input vsp's
 */
static int	sorttype;

  static int
comp(REG vsnsort_t *vsnp1, REG vsnsort_t *vsnp2) 
{
	int	i;

	switch (sorttype) {
	    case 0:
		/*
		 *  sort by vsn
		 */
		i = strncmp(vsnp1->vsn.v_vsn, vsnp2->vsn.v_vsn, (size_t)VSNLEN);
		if (i)
			return(i);
		else
			return(vsnp1->addr->vv_rpord > vsnp2->addr->vv_rpord);
	    case 1:
		return(vsnp1->addr->vv_time > vsnp2->addr->vv_time);
	    case 2:
		return(vsnp1->addr->vv_rpord > vsnp2->addr->vv_rpord);
	    default:
		fprintf(stderr, "programming error\n");
		break;
	}
	exit(1);
	return(-999);
}


/*
 *  vm_prvolstatarr() - print the array of volstat struct pointed to by vsp.
 *	if vsp is null, print a heading
 */
static char	*fmt =   "%*.*s %-5s %*.*s %-8s %-12s %-8s %-8s %-12s%-4s\n";

  int
vm_prvolstatarr(REG vmpVolStat_t *vsp, int count, FILE *fp, 
		int sort, int ppid, int tag, int ptime, int dest)
{
	REG vmpVolStat_t	*tmpvsp;
	int			i;
	vsnsort_t		*vsnarray;
	extern RC_t		*RCArray;

	if (vsp == NULL) {				/* print header */
	  fprintf(fp, fmt, 
		VSNLEN, VSNLEN, "VSN   ", 
		"REQID", 
		RIDLEN, RIDLEN, (ppid ? "PID" : "RSRVD"),
		"REQUEST",
		"STATE", 
		"RC",
		(dest ? "DESTRC" : "HOST"),
		(tag ? "TAG" : "USER"),
		((ptime || sort==1) ? "TIME" : "DEVICE"));
	  return(0);
	}
	sorttype = sort;

	/*
	 *  make sure RC config is loaded
	 */
	if (RCArray == NULL)
		if (vm_configRC())
			return(1);

	/*
	 *  if printing device, need to make sure dev config
	 *	has been loaded
	 */
	if (!ptime  &&  sort !=1  &&  DriveArray == NULL) {
		if (vm_vary(NULL, VVA_DISPREQ, NULL)) 
			return(1);
	}

	/*
	 *  make an array of pointers by vsn
	 */
	vsnarray = (vsnsort_t *) nAlloc( count * sizeof(vsnsort_t));
	tmpvsp = vsp;
	for (i=0; i<count; i++, tmpvsp++) {
		memcpy(&vsnarray[i].vsn, &tmpvsp->vv_vsn, (size_t)VSNLEN);
		vsnarray[i].addr = tmpvsp;
	}

	if (sort == 1) {
		ptime = 1;
	}

	if (sort) {
		(void)qsort((void *)vsnarray, (size_t)count, 
			(size_t)sizeof(vsnsort_t),
			(int (*)(const void *, const void *))comp);
	}

	for (i=0; i<count; i++) {
		char		*device, *rc, dstrc[12];
		char		usertag[TAGLEN+1], idstr[12], reqstr[12];

		device = rc = "";
		dstrc[0] = '\0';
		vsp = vsnarray[i].addr;
		sprintf(reqstr, "%5d", vsp->vv_rpord);
		if (ppid) {
			if (vsp->vv_pid == -1)
				strcpy(idstr, "none");
			else
				sprintf(idstr, "%d", vsp->vv_pid);
		} else 
			sprintf(idstr, "%.*s", RIDLEN, vsp->vv_rid);

		if (ptime) {
			device = nas_time(vsp->vv_time);
		} else if (vsp->vv_use == U_MOUNT  ||  
			   vsp->vv_use == U_UNMOUNT) {
			if (vsp->vv_dev < 0  ||  vsp->vv_dev >= DriveTotal)
				device = "";
			else
				device = DriveArray[vsp->vv_dev].de_name;
		}

		if (tag)
			strncpy(usertag, vsp->vv_tag, (size_t)TAGLEN);
		else if (vsp->vv_flags & VV_UID) {
			strncpy(usertag, nGetUIDName((uid_t)vsp->vv_uid),
				(size_t)TAGLEN);
		}
		usertag[TAGLEN] = '\0';
		if (vsp->vv_flags & VV_RC) {
			RC_t	*rcp;

			if ((rcp = cRCInfo(vsp->vv_rc)) != NULL)
				rc = rcp->rcd_id;
			if (dest  &&  MOVE_STATES(vsp->vv_state)) {
				if ((rcp = cRCInfo(vsp->vv_dstrc)) != NULL)
					strcpy(dstrc, rcp->rcd_id);
			} else if (vsp->vv_state == VS_MOUNTED)
				strcpy(dstrc, HostInfo(vsp->vv_host)->h_name);
		}
		fprintf(fp, fmt, 
			VSNLEN, VSNLEN, &vsp->vv_vsn,
			reqstr,
			RIDLEN, RIDLEN, idstr,
		  	(vsp->vv_use == U_IDLE ?  "" : vUseStr(vsp->vv_use)),
			vReqStateStr(vsp->vv_state), 
		  	rc, dstrc, usertag, device);
	}
	return(0);
}


/*
 *  vm_detach() - make existing requests async, so this proc
 *	can exit without having the requests get cleaned up.
 */
  int
vm_detach()
{
	vmpHdr_t	h;

	vBuildHdr(M_DETACH, vGetSeq(), &h);
	(void) lSnd2RP(&h);
	return (VMError);
}



/*
 *  vm_move() - simple routine to move 1 volume to specified location; 
 *	wait for status. For multivsn move, see volmv.c
 */
  int
vm_move(vsn_t *vsnp, int rcno)
{
	char		buf[PKTSIZ(vmpMove_t) + sizeof(vsnseq_t)];
	vmpHdr_t	*hp, *hp2;
	vmpMove_t	*mp;
	vsnseq_t	*vsnseqp;

	hp = (vmpHdr_t *) buf;
	hp->vh_size = (unsigned int)(PadSize(sizeof(vmpMove_t)) 
				     + sizeof(vsnseq_t));
       	hp->vh_magic = VMP_MAGIC;
	hp->vh_type = M_MOVE;
	hp->vh_seq = vGetSeq();

	hp->vh_count = 1;

	/*
	 *  fill in move
	 */
	mp = MSGPTR(hp, vmpMove_t);
	mp->vmv_dstrc = rcno;
	memset(mp->vmv_rid, '\0', sizeof(mp->vmv_rid));
	vsnseqp = (vsnseq_t *) ((char *)mp + PadSize(sizeof(vmpMove_t)));
	vsnseqp->r_vsn = *vsnp;
	vsnseqp->r_seq = vGetSeq();

	/*
	 *  Built request, now send it 
	 */
	if (lSnd2RP(hp) != E_NOERROR)
		return(VMError);

	/*
	 *  get response
	 */

	for (;;) {
		hp2 = lRcvFromRP(IPC_DELAY);
		if (hp2 == NULL) {
			nFree(hp);
			return(VMError);
		}
		switch(hp2->vh_type) {
		    case M_ACK:
			nFree(hp2);
			return(0);
		    case M_VERROR:
			vInstallError(MSGPTR(hp2, vmpError_t));
			nFree(hp2);
			return(VMError);
		    case M_SHUTDOWN:
			VMError = E_UNAVAIL;
			nLogMsg(EL_CODE|EL_ERROR1, 0, "Vold shutdown");
			nFree(hp2);
			return(VMError);
		    case M_NOP:
			nFree(hp2);
			continue;
		    default:
			nLogMsg(EL_ERROR1, 0, "Unexpected packet: %s", 
							vMsgType(hp2->vh_type));
			nFree(hp2);
		}
	}
}


/*
 *  vm_alloc() - allocate a scratch volume, return the VDR
 */
  vsn_t *
vm_alloc(int rcno, int poolgid, int voltype, char *desc, int mode, int gid)
{
	char		*id = "vm_alloc";
	vmpHdr_t	*hp;
	vmpVAlloc_t	*ap;
	char		pbuf[PKTSIZ(vmpVAlloc_t)];
	vsn_t		*vsnp = NULL;
	CVolReq_t	*cvrp = NULL;
	
	vClearError();			/* clear global errors */
	if (desc != NULL  &&  strlen(desc) > DESCLEN) {
		VMError = E_USER;
		nLogMsg(EL_ERROR1, id, "description longer than %d", DESCLEN);
		return (NULL);
	}
	if (setjmp(lib_jmpbuf)) {		/* Shutdown escape hatch */
		VMError = E_SYS;
		nLogMsg(EL_ERROR1, id, "setjmp escape");
		return (NULL);
	}

	hp = vBuildHdr(M_VALLOC, vGetSeq(), (vmpHdr_t *) pbuf);
	ap = MSGPTR(hp, vmpVAlloc_t);
	ap->va_rc = rcno;
	ap->va_poolgid = poolgid;
	ap->va_type = voltype;
	ap->va_mode = mode;
	ap->va_gid = gid;
	if (desc)
		(void) strncpy(ap->va_desc, desc, (size_t)DESCLEN);
	else
		ap->va_desc[0] = '\0';

	/*
	 *  Built request, now send it and enqueue it...
	 */
	if (lSnd2RP(hp))
		return (NULL);

	nLogMsg(EL_DEBUG2, id, "SENT %s", vHdrStr(hp));

	do {
		lProcPkts(IPC_DELAY);
		if (vDeadWriter) {
			VMError = E_DOWN;
			nLogMsg(EL_TOSTDERR|EL_CODE|EL_ERROR1, id,
			  "unexpected vold termination");
				return(NULL);
			
		}
		cvrp = lGetCVRbySeq(hp->vh_seq);
	} while (cvrp == NULL);

	hp = cvrp->cvr_lastpkt;
	cvrp->cvr_newpkt = 0;
	if (hp == NULL) {
		VMError = E_INTERNAL;
		nLogMsg(EL_ERROR1|EL_CODE, id, "internal err - hp NULL");
		return(NULL);
	}
	switch (hp->vh_type) {
	    case M_VSN: {
		static vsn_t	vsn;
		vmpVsn_t		*vp;

		vp = MSGPTR(hp, vmpVsn_t);
		vsn = vp->v_vsn;
		vsnp = &vsn;
		break;
	    }
	    case M_VERROR:
		vInstallError(MSGPTR(hp, vmpError_t));
		break;
	    default:
		nLogMsg(EL_ERROR1, id, "Unexpected packet: %s", 
					vMsgType(hp->vh_type));
	}

	nFree(cvrp->cvr_lastpkt);		/* get rid of received packet */
	cvrp->cvr_lastpkt = NULL;
						/* TODO: get rid of cvr ent */
	return (vsnp);
}

/*
 *  vm_recycle() - recycle vsn 'good' (uerr==0) or 'bad' & set uid
 *	-1 if failure
 */
  int
vm_recycle(char *vsn, int uerr)
{
	char		*id = "vm_recycle";
	char		pbuf[PKTSIZ(vmpVsn_t)];
	vmpHdr_t	*shp, *rhp;
	vmpVsn_t	*vp;

	/*
	 *  make pkt & send it 
	 */
	shp = vBuildHdr(M_VRECYCLE, vGetSeq(), (vmpHdr_t *) pbuf);
	vp = MSGPTR(shp, vmpVsn_t);

	vp->v_flag = uerr;
	memcpy(&vp->v_vsn, vsn, (size_t)VSNLEN);
	if (lSnd2RP(shp))
		return(-1);

	/*
	 *  get response, return status
	 */
	for (;;) {
		if ((rhp = lRcvFromRP(IPC_DELAY)) == NULL)
			return(-1);

		if (rhp->vh_seq != shp->vh_seq  &&  
		    rhp->vh_type != M_SHUTDOWN  &&
		    rhp->vh_type != M_NOP) {
			nLogMsg(EL_STATUS1, id, "pkt seq wrong - skipping");
			continue;
		}
		switch (rhp->vh_type) {
		    case M_VRECYCLE:
			vp = MSGPTR(rhp, vmpVsn_t);
			if (strncmp(vp->v_vsn.v_vsn, vsn, (size_t)VSNLEN)) {
				nLogMsg(EL_ERROR1|EL_CODE, id,
					"wrong vsn returned (%.*s)",
						VSNLEN, vp->v_vsn.v_vsn);
				nFree(rhp);
				return(-1);
			}
			nFree(rhp);
			return(0);
		    case M_VERROR: 
			vInstallError(MSGPTR(rhp, vmpError_t));
			nFree(rhp);
			return(-1);
		    case M_SHUTDOWN:
			VMError = E_DOWN;
			nFree(rhp);
			return(-1);
		    case M_NOP:
			nFree(rhp);
			continue;
		    default:
			VMError = E_BADPACKET;
			nLogMsg(EL_ERROR1|EL_CODE, id, "unexpected pkt: %s", 
							vHdrStr(rhp));
			nFree(rhp);
			return(-1);
		}
	}
}

/*
 *  vm_vinfo() - look up info on volume
 */

 VDR_t *
vm_vinfo(char *vsn)
{
	vmpHdr_t	*hp;
	vmpVls_t	*ump;
	vmpVdr_t	*vdrp;
	char		pbuf[PKTSIZ2(vmpVls_t, vmpVdr_t)];
	char		*id = "vm_vinfo";

	vClearError();			/* clear global errors */
	hp = vBuildHdr(M_VLS, vGetSeq(), (vmpHdr_t *) pbuf);
	ump = MSGPTR(hp, vmpVls_t);
	strncpy(ump->l_vsn.v_vsn, vsn, (size_t)VSNLEN);

	/*
	 * Built request, now send it 
	 */
	if (lSnd2RP(hp)){
		nLogMsg(EL_DEBUG2, id, "lSnd2RP failed VMError = %d",VMError);
		return (NULL);
	}
	nLogMsg(EL_DEBUG2, id, "SENT %s", vHdrStr(hp));

	/*
	 *  get reply
	 */

	for (;;) {
		hp = lRcvFromRP(IPC_DELAY);
		if (hp == NULL) {
			nLogMsg(EL_DEBUG2, id, "failed");
			return(NULL);
		}

		nLogMsg(EL_DEBUG1, id, "RCVD %s", vHdrStr(hp));

		if (hp->vh_type == M_VDR) {
			break;
		}
		switch (hp->vh_type) {
		    case M_NOP:
			nFree(hp);
			continue;
		    case M_SHUTDOWN:
			nLogMsg(EL_STATUS1, id, "vold shutdown");
			nFree(hp);
			return(NULL);
		    case M_VERROR: {
			vmpError_t	*ep;
			
			ep = MSGPTR(hp, vmpError_t);
			nLogMsg(EL_ERROR1, id, ep->vn_msg);
			vInstallError(ep);
			nFree(hp);
			return(NULL);
		    }
		    default:
			nLogMsg(EL_ERROR1, id, "Unexpected packet: %s",
			  			vMsgType(hp->vh_type));
			nFree(hp);
			return(NULL);
		}
	}
	vdrp =  MSGPTR(hp, vmpVdr_t);
	/* if (strncmp((char *)&vdrp->vdr_vsn, vsnp, VSNLEN)) {
		nLogMsg(EL_ERROR1, id, "Got wrong vsn %.6s (?)", 
						vdrp->vdr_info.vi_vsn);
	}*/
	return(&vdrp->vv_vdr);
}

/*
 *  lProcPkts() - read all outstanding packets in;
 *	if delay==O_NDELAY, then can avoid the hang 
 *	(see vm_getstat())
 */
 void
lProcPkts(int delay)
{
	char		*id = "lProcPkts";
	int		down = 0;
	vmpHdr_t	*hp;

	vClearError();

	/*
	 *  get rid of any buffered packets which have been read
	 */
	lPurgeCVReqs();

	while ((hp = lRcvFromRP(delay)) != NULL) {
		CVolReq_t	*cvrp;
		vsn_t		*vsnp = NULL;

		nLogMsg(EL_DEBUG1, id, "RCVD %s", vHdrStr(hp));

		/*
		 *  after the 1st rcv, don't block
		 */
		delay = IPC_NDELAY;

		/*
		 * need to find the cvr of the received packet
		 */
		switch (hp->vh_type) {
		    case M_ACK:
		    case M_VSN:
		    case M_VOLSTATS:
		    case M_VDR:
		    case M_VARY:
			break;

		    /* if M_VOLSTAT, we can find out the VSN */
		    case M_VOLSTAT: {	
			vmpVsn_t	*vsp;

			vsp = MSGPTR(hp, vmpVsn_t);
			vsnp = &vsp->v_vsn;
		        break;
		      }

		    case M_SHUTDOWN:
			down = 1;
			/* fall through */
		    case M_VERROR: {
			vmpError_t	*ep;

			ep = MSGPTR(hp, vmpError_t);
			if (down) {
				ClientTermIPC(NEWERROR);
				vInstallError(ep);
				nFree(hp);		/* free packet */
				longjmp(lib_jmpbuf, 1);	/* returns to user */
			}
			break;
		      }

		    case M_NOP:			/* ignore NOPs */
			nFree(hp);
			continue;

		    default:
			nLogMsg(EL_ERROR1, id, "Unexpected packet: %s",
			  			vMsgType(hp->vh_type));
			nFree(hp);
			continue;
		}

		nLogMsg(EL_DEBUG1, id,
			  "creating a CVolReq_t for seq %x", hp->vh_seq);
		cvrp = (CVolReq_t *) nAlloc(sizeof (CVolReq_t));
		cvrp->cvr_seq = hp->vh_seq;
		cvrp->cvr_lastpkt = hp;
		if (vsnp)		/* copy vsn if available */
			cvrp->cvr_vsn = *vsnp;
		cvrp->cvr_newpkt = 1;		/* have new status */
		EnQCVR(cvrp, &CVolReqQ);
	}
	return;
}


/*
 *  lGetCVRbyVSN() - look up a Client Volume Request (on the CVolReqQ) by vsn
 */
  CVolReq_t *
lGetCVRbyVSN(REG vsn_t *vsnp)
{
	REG CVolReq_t	*cvrp;

	for (cvrp = (CVolReq_t *) CVolReqQ.q_head; cvrp;
	  cvrp = (CVolReq_t *) cvrp->cvr_q.q_next)
		if (SAMEVSN(vsnp, &cvrp->cvr_vsn))
			return (cvrp);
	return (NULL);
}


/*
 *  lGetCVRbySeq() - look up a Client Volume Request (on the CVolReqQ)
 *	by sequence number
 */
  CVolReq_t *
lGetCVRbySeq(REG Seq_t seq)
{
	REG CVolReq_t	*cvrp;

	for (cvrp = (CVolReq_t *) CVolReqQ.q_head; cvrp;
	  cvrp = (CVolReq_t *) cvrp->cvr_q.q_next)
		if (seq == cvrp->cvr_seq)
			return (cvrp);
	return (NULL);
}


/*
 *  lPurgeCVReqs() - get rid of all packets which have been read
 */
  void
lPurgeCVReqs()
{
	REG CVolReq_t	*cvrp, *nextcvrp;

	for (cvrp = (CVolReq_t *) CVolReqQ.q_head; cvrp; cvrp = nextcvrp) {
		nextcvrp = (CVolReq_t *) cvrp->cvr_q.q_next;
		if (!cvrp->cvr_newpkt) {
			DeQCVR(cvrp, &CVolReqQ);
			if (cvrp->cvr_lastpkt)
				nFree(cvrp->cvr_lastpkt);
			nFree(cvrp);
		}
	}
}


/*
 *  lPrintCVRQ() - for debugging
 */
  void
lPrintCVRQ(char *caller)
{
	CVolReq_t	*cvrp;

	nLogMsg(EL_STATUS1, caller, "Client Vol Rreq Queue:");
	for (cvrp = (CVolReq_t *) CVolReqQ.q_head; cvrp;
	  cvrp = (CVolReq_t *) cvrp->cvr_q.q_next)
		nLogMsg(EL_STATUS1, caller,
		"\tvsn=%.*s seq=%x lastpkt=%x newpkt=%d link=%x",
		  VSNLEN, &cvrp->cvr_vsn, cvrp->cvr_seq, cvrp->cvr_lastpkt,
		  cvrp->cvr_newpkt, cvrp->cvr_q.q_next);
}



/*
 *  vSetPgrp() - Disconnect from the keyboard if running in the background
 */
int
vSetPgrp()
{
	if (!isatty(fileno(stdin))  &&  !isatty(fileno(stderr))) {
#ifdef SYSV
		if( setpgrp() < 0) {
			perror("setpgrp");
			return(E_SYS);
		}
#else
		if( setpgid(0,0) < 0)  {
			perror("setpgid");
			return(E_SYS);
		}
#endif
	}
	return(E_NOERROR);
}


  int
checkvsnrange(vsn_t *vsnp, int range)
{
	int	i;

	if (!isalnum(vsnp->v_vsn[0])) {
		fprintf(stderr, "VSN %.*s: Illegal character (%c)\n",
		  VSNLEN, vsnp->v_vsn, vsnp->v_vsn[0]);
		return(1);
	}
	for (i=1; i<VSNLEN; i++) {
		/* insist on digits for ranges, alnum for non-ranges */
		if ((range && !isdigit(vsnp->v_vsn[i])) ||
		    !isalnum(vsnp->v_vsn[i])) {
			fprintf(stderr, "VSN %.*s: Illegal character (%c)\n",
			  VSNLEN, vsnp->v_vsn, vsnp->v_vsn[i]);
			return(1);
		}
	}
	for (i=0; i<VSNLEN; i++)
		if (islower(vsnp->v_vsn[i]))	/* uppercasify if lower case */
			vsnp->v_vsn[i] = toupper(vsnp->v_vsn[i]);
	return(0);
}


/*
 *  vm_writelabel() - write generic internal label 
 */
  int
vm_writelabel(int fd, char *internal, char *vsn, char *cmdname)
{
	char	vol1buf[VOL1LEN];
	int	i;

	if (ntp_taprew(fd, vsn, 0) == -1) {
		perror("rewind failed; label not written");
		return(errno);
	}
	sprintf(vol1buf, "VOL1%-75.6s4", internal);
	if ((i = (int)write(fd, vol1buf, (size_t)VOL1LEN)) != VOL1LEN) {
		fprintf(stderr,
			"%s: write of label failed. size = %d error %s\n",
			cmdname, i, strerror(errno));
		return(errno);
	}

	/* 
	 *  Add 2 tape marks for "end of data"
	 */

	if (ntp_tapwtm(fd, vsn, 0) == -1) {
		perror("writing 1st tapemark");
		return(1);
	}
	if (ntp_tapwtm(fd, vsn, 0) == -1) {
		perror("writing 2nd tapemark");
		return(1);
	}
	return(0);
}

/*
 *  vm_readlabel() - read internal label and print it
 */
  int
vm_readlabel(int fd, char *vsnp, char *cmdname)
{
	static char	vol1buf[VOL1LEN];
	int     	i;

	if ((i = (int)read(fd, vol1buf, (size_t)VOL1LEN)) != VOL1LEN) {
		fprintf(stderr,
			"%s: error reading label, error = %s, size = %d\n",
			cmdname, strerror(errno), i);
		return(1);
	}
	/*
	 *  check for VOL1
	 */

	if (strncmp(vol1buf, "VOL1", (size_t)4) == 0) {
		printf("label on tape:   (%.6s)  ANSI version: (%c)\n",
			vol1buf+4, vol1buf[79]);
		strncpy(vsnp, vol1buf+4, (size_t)VSNLEN);	
	} else {
		printf("label on tape:   none (no 'VOL1' field)\n");
		vsnp[0] = '\0';
	}
	return(0);
}

/*
 *  routines for specifying ranges of vsns
 *	note: 'prolog' arg is expected to be rounded up mod 8
 */

static vsnseq_t	*vsnarray;
static int	vsncount = -1, vsnarraysize, prebuf;
static char	*cmdname, *msgbuf = NULL;

  void
vm_initlist(char *id, int prolog)
{
	/*
	 *  check/set size of initial buffer
	 */
	if (prolog < 0) {
		fprintf(stderr, "%s: vm_initlist(): programmer error\n", id);
		exit(E_USER);
	}
	if (prolog % 8) {
		fprintf(stderr, "%s: vm_initlist(): programmer error 2\n", id);
		exit(E_USER);
	}
	prebuf = prolog;

	/*
	 *  initialize other counts
	 */
	vsncount = 0;
	vsnarraysize = 0;
	cmdname = id;

	/*
	 *  free any memory from previous use
	 */
	if (msgbuf != NULL)
		free(msgbuf);

}

/* VSN spec string tokens */
typedef enum {
	V_VSN,		/* vsn_t */
	V_DASH,		/* '-' */
	V_COMMA,	/* ',' */
	V_END,
	V_PROBLEM
} vsn_token;


  static vsn_token
getvsntoken(REG char **specpp)
{
	int		i;
	register char	*endp, *cp;

	switch (**specpp) {
	    case ',':
		(*specpp)++;			/* advance pointer past comma */
		return (V_COMMA);
	    case '-':
		(*specpp)++;			/* advance pointer past dash */
		return (V_DASH);
	    case '\0':
		return (V_END);
	    default:		/* vsn is terminated by ',', '-', or '\0' */
		/* point endp at the nearest terminator */
		endp = strchr(*specpp, '-');
		cp = strchr(*specpp, ',');
		if (cp && (endp == 0 || endp > cp))
			endp = cp;
		cp = *specpp + strlen(*specpp);
		if (endp == NULL)
			endp = cp;
		if (endp - *specpp != VSNLEN) {
			fprintf(stderr, "%s: VSN \"%.*s\" has wrong length\n",
			  cmdname, (int)(endp - *specpp), *specpp);
			*specpp = endp;		/* advance pointer past VSN */
			return(V_PROBLEM);
		}
		endp = *specpp;
		for (i = 0; i < VSNLEN; i++)
			if (!isalnum(endp[i])) {
				fprintf(stderr, "%s: VSN %.*s: bad char (%c)\n",
				  cmdname, VSNLEN, endp, endp[i]);
				*specpp = endp;	/* advance pointer past VSN */
				return(V_PROBLEM);
			}
		(*specpp) += VSNLEN;		/* advance pointer past VSN */
		break;
	}
	/* default opt put here to avoid compiler warning */
	return(V_VSN); 
}

  static void
addtovlist(vsn_t *vsnp)
{
	if (msgbuf == NULL) {

		/*
		 *  start array w/ 1K vsns
		 */
		msgbuf = malloc(prebuf + 1000 * sizeof(vsnseq_t));
		if (msgbuf == NULL) {
			fprintf(stderr, "%s: out of mem\n", cmdname);
			exit(E_SYS);
		}
		vsnarraysize = 1000;
		vsnarray = (vsnseq_t *) ((char *)msgbuf + prebuf);
	} else if (vsncount == vsnarraysize) {
		/*
		 *  grow array by factor of 2
		 */
		vsnarraysize *= 2;
		msgbuf = realloc(msgbuf, prebuf + 
				vsnarraysize * sizeof(vsnseq_t));
		if (msgbuf == NULL) {
			fprintf(stderr, "%s: out of mem\n", cmdname);
			exit(E_SYS);
		}
		vsnarray = (vsnseq_t *) ((char *)msgbuf + prebuf);
	}
	vsnarray[vsncount].r_vsn = *vsnp;
	vsnarray[vsncount].r_seq = vsncount + 1;
	vsncount++;
}

  static void
skip(FILE *fp)
{
	int	cc;

	for (;;) {
		cc = getc(fp);
		if (isspace(cc))
			break;
		if (cc == EOF) {
			ungetc(cc, fp);
			break;
		}
		if (isalnum(cc))
			fprintf(stderr, "%c", cc);
		else
			fprintf(stderr, "_");
	}
	fprintf(stderr, "\n");
}

 int
vm_getvsnfromfile(FILE *fp, char *vsnp)
{
	int		i, cc;

retry:

	/*
	 *  skip whitespace
	 */
	while (isspace(cc = getc(fp)));
	if (cc == EOF)
		return(EOF);

	/*
	 *  read vsn & check it
	 */
	for (i=0; i<VSNLEN+1; i++) {
		if (isspace(cc))
			break;
		vsnp[i] = cc;
		cc = getc(fp);
	}
	vsnp[i] = '\0';
	if (!isspace(cc)) {
		fprintf(stderr, "Vsn too long /%.*s%c",
					VSNLEN, vsnp, cc);
		skip(fp);
		fprintf(stderr, "\n");
		goto retry;
	}
	if (checkvsn(vsnp)) {
		printf("\n%s\n", nErrStr);
		goto retry;
	}

	/*
	 *  it's good
	 */
	return(0);
}

  char *
vm_getvsnfilelist(FILE *fp, int *countp)
{
	char		vsn[VSNLEN+1];

	if (vsncount == -1) {
		fprintf(stderr, "getvsnfilelist() - not set up\n");
		exit(E_USER);
	}
	for (;;) {	/* vsns */
		if (vm_getvsnfromfile(fp, vsn) == EOF)
			break;
		addtovlist((vsn_t *)vsn);
	}
	*countp = vsncount;
	vsncount = -1;
	return(msgbuf);
}

  char *
vm_getvsnarglist(char **argv, int argc, int *countp, int *errp)
{
	int		i; 

	if (vsncount == -1) {
		fprintf(stderr, "getvsnarglist() - not set up\n");
		exit(E_USER);
	}
	/*
	 *  each arg can be a compact vsnlist
	 */
	for (i=0; i<argc; i++) {
		char	*sp, *lsp;
		int	isrange = 0;	/* item is range */
		int	isend = 0;	/* end of string */
		vsn_t	lfirst;		/* first vsn of range */
		vsn_t	lvsn;		/* last vsn read */

		lvsn.v_vsn[0] = 0;
		sp = argv[i];
		while (!isend) {
			lsp = sp;
			switch (getvsntoken(&sp)) {
			    case V_PROBLEM:
				(*errp)++;
				break;
			    case V_DASH:
				isrange = 1;
				lfirst = lvsn;
				break;
			    case V_VSN:
				lvsn = *((vsn_t *) (lsp));
				break;
			    case V_END:
				isend = 1;
				/* fall through */
			    case V_COMMA:
				if (isrange) {		/* end of range */
					char	vsnch, numbuf[VSNLEN];
					int	first, last, cnt;

					if (checkvsnrange(&lfirst, isrange)) {
						(*errp)++;
						continue;
					}
					vsnch = lfirst.v_vsn[0];
					if (checkvsnrange(&lvsn, isrange)) {
						(*errp)++;
						continue;
					}
					if (lvsn.v_vsn[0] != vsnch) {
						fprintf(stderr,
	"%s: %.*s-%.*s: range VSN's must have same starting character\n",
					  		cmdname,
							VSNLEN, lfirst.v_vsn,
					  		VSNLEN, lvsn.v_vsn);
						(*errp)++;
						continue;
					}
					strncpy(numbuf, &lfirst.v_vsn[1], 
							(size_t)(VSNLEN - 1));
					numbuf[VSNLEN-1] = '\0';
					first = atoi(numbuf);
					strncpy(numbuf, &lvsn.v_vsn[1], 
							(size_t)(VSNLEN - 1));
					numbuf[VSNLEN-1] = '\0';
					last = atoi(numbuf);
					if (first >= last) {
						fprintf(stderr,
					  	  "%s: %.*s-%.*s: bad range\n",
					  	  cmdname,
						  VSNLEN, lfirst.v_vsn,
					  	  VSNLEN, lvsn.v_vsn);
						(*errp)++;
						continue;
					}

					for (cnt = first; cnt <= last; cnt++) {
						char	vsn[VSNLEN * 2];

						sprintf(vsn, "%c%.*d",
					  		vsnch, VSNLEN-1, cnt);
						addtovlist((vsn_t *)vsn);
					}
					isrange = 0;
				} else if (lvsn.v_vsn[0]) {
					if (checkvsnrange(&lvsn, isrange)) {
						(*errp)++;
						continue;
					}
					addtovlist(&lvsn);
				}
				break;
			    default:
				fprintf(stderr, 
					"%s: internal error in getvsnarglist\n",
			  		cmdname);
			}
		}
	}
	*countp = vsncount;
	vsncount = -1;
	return(msgbuf);
}


/*
 *  vm_userid() - map string to unix user id
 */
  int
vm_userid(char *str, int *uid)
{
	struct passwd *pwp;

	if (isnum(str)) {
		*uid = atoi(str);
		if (getpwuid((uid_t)*uid) != NULL)
			return(0);
	}
	pwp = getpwnam(str);
	if (pwp == NULL) {
		*uid = -1;
		return(-1);
	}
	*uid = (int)pwp->pw_uid;
	return(0);
}

/*
 *  vm_groupid() - map string to unix group
 */
  int
vm_groupid(char *str, int *gid)
{
	struct group *grp;

	/*
	 *  check if numbered group
	 */
	if (isnum(str)) {
		*gid = atoi(str);
		if (getgrgid((gid_t)*gid) != NULL)
			return(0);
	}
	/*
	 *  try for string
	 */
	grp = getgrnam(str);
	if (grp == NULL) {
		*gid = -1;
		return(-1);
	}
	*gid = (int)grp->gr_gid;
	return(0);
}

/*
 *  vm_octalperm() - interpret string as volman perm bits
 */
  int
vm_octalperm(char *str, short *perm)
{
	int     i, j;

	*perm = (short)0;
	j = strlen(str);

	/*
	 *  form is OOO or 0OOO
	 */

	if (j > 4)
		return(-1);
	
	/*
	 *  skip leading 0 if any
	 */

	if (j == 4) {
		if (str[0] != '0')
			return(-1);
		j--;
		str++;
	}
	for (i=0; i<3; i++) {
		if (str[i] == '0' ||  str[i] == '2' ||  str[i] ==  '4'
						  || str[i] == '6') {
			*perm *= 8;
			*perm += str[i] - 48;
		} else 
			return(-1);
	}
	return(0);
}
