/*
 *  vm_config.c - RC & drive configuration
 *
 *
 */

#include "vm_space.h"

extern DevEnt_t	*DriveArray;
extern int	DriveTotal;

/*
 *  vDriveName() - create drivename from host type, medium, & dev #
 *	See end of this file for naming documentation.
 */
  char *
vDriveName(int host_type, int controller, int devno, int medium)
{
	char		*id = "vDriveName";
	static char	dname[40];

	switch(host_type) {
	    case VH_NETBSD:
		/*
		 *  sprintf(dname, "/dev/nrst%d", devno);
		 *  Note: this is what we had before the 
		 *	'control node' below was adapted
		 *	to do no-rewind i/o too. The problem
		 *	was that in netBSD, one could not
		 *	open a non-control node unless the
		 *	tape was mounted, whereas volman
		 *	code was based on opening the node
		 *	& polling for mount status, then
		 *	doing i/o on the same node.
		 */
		sprintf(dname, "/dev/nrst%d", devno);
		break;
	    case VH_SGI:
		/* how does medium figure in..? */
		/* nr = no rewind, ns = no swap bytes */
		sprintf(dname, "/dev/rmt/tps%dd%dnrns", controller, devno);
		break;
	    case VH_SUN:
		switch(medium) {
		    case VM_MAGSTAR: 
			/* c = compressed, n = no rewind */
			sprintf(dname, "/dev/rmt/%dcn", devno);
			break;
		    default:
			nLogMsg(EL_STATUS1, id, "Unknown medium for sun: %d", 
									medium);
			dname[0] = '\0';
		}
		break;
	    case VH_CONVEX:
		switch(medium) {
		    case VM_D3:
			/* n = no rewind, i = compression */
			sprintf(dname, "/dev/rrdw%dni", devno);
			break;
		    case VM_34X0:
			/* n = no rewind, i = compression */
			sprintf(dname, "/dev/rtc%dni", devno);
			break;
		    case VM_MAGSTAR:
			/* dummy name */
			sprintf(dname, "/dev/mags%d", devno);
			break;
		    default:
			nLogMsg(EL_STATUS1, id, 
					"Unknown medium for convex: %d",medium);
			dname[0] = '\0';
		}
		break;
	    case VH_LINUX:
		/* so far, only dummy-drive testing, so use /dev/null */
		sprintf(dname, "/dev/null");
		break;
	    default:
		nLogMsg(EL_FATAL, id, "Unknown host type: %d", host_type);
	}
	return(dname);
}

/*
 *  vGetDev(): maps bit set in mask to table entry in the DriveArray.
 *	Returns NULL if more than one bit is set in the mask.
 *  vGetDevByNum(): same lookup with a device number instead of a map.
 *	Returns NULL if bit position is out of range.
 *
 *	The drive config must be loaded before either call is made.
 *	The RP loads from its config file, the clients load from the RC.
 */

  DevEnt_t *
vGetDev(DevMap_t *dmp)
{
	int	i;

	if (DriveArray == NULL)
		nLogMsg(EL_FATAL, 0, "calling vGetDev before configged");
	i = vDMffs(dmp);
	if (i >= DriveTotal  ||  i < 0)
		return(NULL);
	return(&DriveArray[i]); 	
}

  DevEnt_t *
vGetDevByNum(int devno)
{
	if (DriveArray == NULL)
		nLogMsg(EL_FATAL, 0, "calling vGetDevByNum before configged");
	if (devno >= DriveTotal || devno < 0)
		return(NULL);
	return(&DriveArray[devno]); 	
}

  int
vCountVaryErrors()
{
	int	devno, errs = 0;

	if (DriveArray == NULL)
		nLogMsg(EL_FATAL, 0, 
				"calling vCountVaryErrors before configged");

	for (devno=0; devno<DriveTotal; devno++) 
		if (DriveArray[devno].de_errstr)
			errs++;
	return(errs);
}

  void
PrintDrives(char *id, char *str)
{
	DevEnt_t	*dep = DriveArray;
	int	i;

	nLogMsg(EL_STATUS1, id, str);
	for (i=0; i<DriveTotal; i++, dep++)
		nLogMsg(EL_STATUS1, "   ", "%s  %d ", 
				dep->de_name, dep->de_status);
}

/*
 *  device naming documentation
 *
 * SGI (man tps)

     /dev/rmt/tps*
     /dev/rmt/jag*

     ... Since so many different types of devices are
     supported, and not all their features can be determined directly from the
     drive, a configuration table defines their capabilities.  This is usually
     found in the file /var/sysgen/master.d/scsi.

     The special files are named according to this convention:

   /dev/{r}mt/tpscontrollerdID{nr}{ns}{s}{v}{.density}{c}
   /dev/rmt/jagcontrollerdID{nr}{ns}{s}{v}{stat}{.density}

     controller is the SCSI controller number and ID is variously known as the
     SCSI ID, the SCSI address, the drive address, and the unit number.  The
     device types are:

     {nr}   no-rewind on close device

     {ns}   non-byte swapping device

     {s}   byte swapping device

     {v}   variable block size device (supported only for 9-track, DAT,
    DLT and 8 mm devices as shipped)

     {stat}   a special purpose device; it can be used even when one of
    the other names for the same physical device is already
    opened (see below)

     {.density}   for 9-track tape it is one of 800, 1600, 3200, or 6250 and
    for the Exabyte 8500 it is one of 8200 and 8500

     {c}   data compression; it is only supported on DLT and should not
    be confused with density


 * SUN (man mtio)

     /dev/rmt/<unit number><density>[<BSD behavior>][<no rewind>]

          density                  l, m,  h,  u/c  (low,  medium,
                                   high,        ultra/compressed,
                                   respectively)

          BSD behavior (optional)  b

          no rewind (optional)     n

     For example, /dev/rmt/0hbn specifies unit 0,  high  density,
     BSD behavior and no rewind.

 * CONVEX (man tc)

     Device file names for cartridge tape units usually contain
     the following: for DAT ``dat''; Fujitsu 3480 ``tc''; Metrum
     ``rsp''; Exabyte ``exb''; and ATL 2640-DLT ``dlt'' in addi-
     tion to the unit number.  Several device files may
     correspond to a single physical device; the names of the
     device files may include characters from Table 1 to indicate
     device characteristics.  Thus, the names of these devices
     have the following form:

                 /dev/r{dat|tc|rsp|exb|dlt}?[n][u][i]


     The following characters in this format have special signi-
     ficance:

                               Table 1

             ?   Unit number

             n   Do not rewind after file is closed; if n
                 is absent, tape will be rewound.

             u   Data buffering disabled at the drive; if
                 u is absent, buffering is enabled.  The
                 DAT drive does not support this option.

             i   Data compression is enabled at the
                 drive; if i is absent, compression is
                 disabled.  Note:  The Fujitsu 3480 for-
                 matter has trouble compressing records
                 that are larger than 102 KBytes.

 * NETBSD (man st)

     /dev/[n][e]rst[0-9]  general form:
     /dev/rst0            Mode 0, rewind on close
     /dev/nrst0           Mode 2, No rewind on close
     /dev/erst0           Mode 3, Eject on close (if capable)
     /dev/enrst0          Mode 4, Control Mode (elsewise like mode 1)

*/
