/*
 *  Volume Manager
 *
 *  vm_p.c: Volume Manager Protocol
 *	various tables, initializations, assignments and
 *	misc routines are found here. Directories are
 *	initialized from Makefile or Mkconfig.volman definitions.
 */

#include <ctype.h>
#include <syslog.h>
#include <signal.h>

#include "vm_space.h"

extern int	vDeadReader;

#ifdef AMDAHL_UTS
#include <malloc.h>
#else
#include <stdlib.h>
#endif

/*********************************************************************
 *  Makefile/config.csh-defined globals - all here for convenience
 */
/*  used both for installation & w/in progs  */
char	*VolManOperAcct = VOLMANACCT;
char	*vSpoolDir = VSPOOLDIR;
char	*vLibDir = VLIBDIR;
char	*vEtcDir = ETCDIR;
char	*vBinDir = BINDIR;

/*  used only w/in progs  */
char	*vNodeDir = NODEDIR;
char	*vDBDir = VDBDIR;
char	*volcd_service = VOLCD_SERVICE;
ushort	volcd_port;			/* set with getservbyname */
char	*vold_service = VOLD_SERVICE;
ushort	vold_port;			/* set with getservbyname */
char	*vVoldHost = VOLD_HOST;
char	*vVoldConfig = VCONF;
char	*vHostType = OPSYS;

/********************************************************************/

/*
 *  other globals
 */

int	VMError = 0;
char	*vClientDir = NULL;
char	*vLogFile = NULL;
int	rp_syslog = LOG_LOCAL0;

/********************************************************************/

  char *
vVsnNode(char *vsn)
{
	static char	tpnode[MAXPATHLEN];

	sprintf(tpnode, "%s/%s", vNodeDir, vsnstr(vsn));
	return(tpnode);
}

/*
 *  VmpTab[]: table of information about VMP message types.
 */
ProTab_t VmpTab[] = {
	/* request		size		*/
	{ "M_(zero)",	0,		0 },
	{ "M_ACK",	M_ACK,		0 },
	{ "M_ATTACH",	M_ATTACH,	sizeof(vmpVsn_t) },
	{ "M_DEBUG",	M_DEBUG,	sizeof(vmpDebug_t) },
	{ "M_DETACH",	M_DETACH,	0 },
	{ "M_VERROR",	M_VERROR,	sizeof(vmpError_t) },
	{ "M_GOODBYE",	M_GOODBYE,	0 },
	{ "M_HELLO",	M_HELLO,	sizeof(vmpHello_t) },
	{ "M_LBLOOK",	M_LBLOOK,	sizeof(vmpLblLook_t) },
	{ "M_RCMOUNT",	M_RCMOUNT,	sizeof(vmpRCMount_t) },
	{ "M_MOVE",	M_MOVE,		-1 },		/* variable size */
	{ "M_NOP",	M_NOP,		0 },
	{ "M_QUOTQRY",	M_QUOTQRY,	sizeof(vmpQuota_t) },
	{ "M_QUOTRPLY",	M_QUOTRPLY,	sizeof(vmpQuota_t) },
	{ "M_RCMSG",	M_RCMSG,	sizeof(vmpRcMsg_t) },
	{ "M_REQUEUE",	M_REQUEUE,	sizeof(vmpVsn_t) },
	{ "M_SHUTDOWN",	M_SHUTDOWN,	sizeof(vmpError_t) },
	{ "M_TIMER",	M_TIMER,	sizeof(vmpTimer_t) },
	{ "M_UNMOUNT",	M_UNMOUNT,	sizeof(vmpVsn_t) },
	{ "M_VALLOC",	M_VALLOC,	sizeof(vmpVAlloc_t) },
	{ "M_VARY",	M_VARY,		sizeof(vmpVary_t) },
	{ "M_DRIVEINFO",	M_DRIVEINFO,	-1 },	/* variable size */
	{ "M_IDLETIME",	M_IDLETIME,	-1 },		/* variable size */
	{ "M_VRECYCLE",	M_VRECYCLE,	sizeof(vmpVsn_t) },
	{ "M_VDR",	M_VDR,		sizeof(vmpVdr_t) },
	{ "M_VOLSTAT",	M_VOLSTAT,	sizeof(vmpVolStat_t) },
	{ "M_VOLSTATS",	M_VOLSTATS,	-1 },		/* variable size */
	{ "M_VOLQRY",	M_VOLQRY,	sizeof(vmpVsn_t) },
	{ "M_VLS",	M_VLS,		sizeof(vmpVls_t) },
	{ "M_VSN",	M_VSN,		sizeof(vmpVsn_t) },
	{ "M_ADDVSN",	M_ADDVSN,	sizeof(vmpAdd_t) },
	{ "M_DELVSN",	M_DELVSN,	sizeof(vmpVsn_t) },
	{ "M_SETQUOTA",	M_SETQUOTA,	sizeof(vmpQuota_t) },
	{ "M_LBLQRY",	M_LBLQRY,	sizeof(vmpLblQry_t) },
	{ "M_VCHMOD",	M_VCHMOD,	sizeof(vmpVsn_t) },
	{ "M_VCHGRP",	M_VCHGRP,	sizeof(vmpVsn_t) },
	{ "M_VCHOWN",	M_VCHOWN,	sizeof(vmpVsn_t) },
	{ "M_AUDIT",	M_AUDIT,	sizeof(vmpVsn_t) },
	{ "M_VATTR",	M_VATTR,	sizeof(vmpAttr_t) },
	{ "M_USRMOUNT",	M_USRMOUNT,	sizeof(vmpUsrMount_t) },
	{ "M_USRMNT_PLL",	M_USRMOUNT_PLL,	-1 },	/* variable size */
	{ "M_USRMNT_SEL",	M_USRMOUNT_SEL,	-1 },	/* variable size */
	{ "M_NEWLOG",	M_NEWLOG,	0 },
	{ "M_RESERVE",	M_RESERVE,	-1 },		/* variable size */
	{ "M_RCINFO",	M_RCINFO,	-1 },		/* variable size */
	{ "M_CREATENODE",	M_CREATENODE,	sizeof(vmpNode_t) },
	{ "M_DELETENODE",	M_DELETENODE,	sizeof(vmpNode_t) },
	{ "M_SCRATCH",	M_SCRATCH,	sizeof(vmpNode_t) },
	{ "M_NODED",	M_NODED,	sizeof(vmpHello_t) },
	{ "M_VOLCD",	M_VOLCD,	sizeof(vmpHello_t) },
	{ "M_TIMERD",	M_TIMERD,	sizeof(vmpHello_t) },
	{ "M_NEWCLIENT",	M_NEWCLIENT,	sizeof(vmpHello_t) },
	{ "M_RC",		M_RC,		sizeof(vmpHello_t) },
	{ "M_RCREADY",	M_RCREADY,	sizeof(vmpHello_t) },
	{ "M_CONSOLE",	M_CONSOLE,	sizeof(vmpHello_t) },
	{ "M_HOSTINFO",	M_HOSTINFO,	-1 },		/* variable size */
	{ "M_SIGNAL",	M_SIGNAL,	sizeof(vmpVsn_t) },
	{ "M_DEVQRY",	M_DEVQRY,	0 },
	{ "M_PRIORITY",	M_PRIORITY,	0 },
	{ NULL,		0,		0 }
};

  ssize_t
vPktSiz(int msgno)
{
	return(nPktSiz(msgno, VmpTab));
}

  char *
vMsgType(int type)
{
	return(nMsgType(type, VmpTab));
}

CodeString_t vVMErrTab[] = {
	{ E_NOERROR,		"no error" },
	{ E_NOMEM,		"out of memory" },
	{ E_SYS,		"system error" },
	{ E_DOWN,		"RP/RC down" },
	{ E_INTERNAL,		"INTERNAL ERROR" },
	{ E_BADPACKET,		"bad packet" },
	{ E_IO,			"I/O error" },
	{ E_USER,		"User error" },
	{ E_DBASE,		"Database error" },
	{ E_ADMIN,		"Database inconsistency" },
	{ E_PERM,		"Permission Denied" },
	{ E_NOTFOUND,		"Not found" },
	{ E_UNAVAIL,		"Resource unavailable/in use" },
	{ E_OPS,		"Operations problem" },
	{ E_CANCELED,		"Request canceled" },
	{ E_COMM,		"Communications problem" },
	{ E_HARDWARE,		"HARDWARE PROBLEM" },
	{ E_STUCK,		"Stuck in drive" },
	{ 0,			NULL }
};

CodeString_t vReqStateTab[] = {
	{ VS_IDLE,		"IDLE" },
	{ VS_MOUNTWAIT,		"MOUNTWAIT" },
	{ VS_DRIVEWAIT,		"DRIVEWAIT" },
	{ VS_RESERVEMOUNTWAIT,	"RESERVEWAIT" },
	{ VS_RCWAIT,		"MOUNTING" },
	{ VS_CREATENODE,	"CREATENODE" },
	{ VS_MOUNTED,		"MOUNTED" },
	{ VS_DELETENODE,	"DELETENODE" },
	{ VS_UNMOUNTWAIT,	"UNMOUNTWAIT" },
	{ VS_RCWAIT2,		"UNMOUNTING" },
	{ VS_UNMOUNTED,		"UNMOUNTED" },
	{ VS_MOVEWAIT,		"MOVEWAIT" },
	{ VS_RESERVEMOVEWAIT,	"RESERVEWAIT" },
	{ VS_MOVING,		"MOVING" },
	{ VS_EJECTED,		"EJECTED" },
	{ VS_ACCEPTED,		"ACCPTED" },
	{ VS_1MSGMOVE,		"1MSGMOVE" },
	{ 0,			NULL }
};


CodeString_t	vPortTab[] = {
	{ PT_NONE,	"(undefined)" },
	{ PT_VOLCD,	"volcd" },
	{ PT_VOLNODE,	"volnd" },
	{ PT_RC,	"RC" },
	{ PT_TIMER,	"voltimer" },
	{ PT_CONSOLE,	"console" },
	{ 0,		NULL }
};


CodeString_t	vStatTab[] = {
	{ S_VIRGIN,	"VIRGIN" },
	{ S_SCRATCH,	"SCRATCH" },
	{ S_ALLOC,	"ALLOCATED" },
	{ S_RECYCLE,	"RECYCLED" },
	{ S_RECYCLEBAD,	"RECYCLEDBAD" },
	{ S_TRASH,	"TRASH" },
	{ S_OBSOLETE,	"OBSOLETE" },
	{ S_DELETED,	"DELETED" },
	{ 0,		NULL }
};

  int
vStatusToBit(int code)
{
	char	*id = "vStatusToBit";
	/*
	 *  map code to a bit
	 */
	switch (code) {
	    case S_VIRGIN:
		return ST_VIRGIN;
	    case S_SCRATCH:	
		return ST_SCRATCH;	
	    case S_ALLOC:
		return ST_ALLOC;
	    case S_RECYCLE:	
		return ST_RECYCLE;
	    case S_RECYCLEBAD:	
		return ST_RECYCLEBAD;
	    case S_TRASH:
		return ST_TRASH;
	    case S_OBSOLETE:	
		return ST_OBSOLETE;
	    case S_DELETED:
		return ST_DELETED;
	    default:
		nLogMsg(EL_FATAL, id, "programming err on status [%c]", code);
		return 0;
	}
}

CodeString_t	StatTrans[] = {
	{ S_VIRGIN,	"VIRGIN" },
	{ S_VIRGIN,	"virgin" },
	{ S_SCRATCH,	"SCRATCH" },
	{ S_SCRATCH,	"scratch" },
	{ S_ALLOC,	"ALLOCATED" },
	{ S_ALLOC,	"allocated" },
	{ S_RECYCLE,	"RECYCLED" },
	{ S_RECYCLE,	"recycled" },
	{ S_RECYCLEBAD,	"RECYCLEDBAD" },
	{ S_RECYCLEBAD,	"recycledbad" },
	{ S_TRASH,	"TRASH" },
	{ S_TRASH,	"trash" },
	{ S_OBSOLETE,	"OBSOLETE" },
	{ S_OBSOLETE,	"obsolete" },
	{ 0,		NULL }
};

/*
 *  vStatusBit() - convert status string or char to bit
 *	0 if untranslatable
 */
 int
vStatusBit(char *str)
{
	CodeString_t	*cs;
	char		code = 0;

	/*
	 *  figure out 1-letter code per dbase
	 */
	if (strlen(str) == 1) {
		if (vVolStateStr(str[0])[0] == '[') 
			return(0);
		code = str[0];
	} else {
		for (cs=StatTrans; cs->cs_string; cs++) {
			if (strcmp(str, cs->cs_string) == 0) {
				code = cs->cs_code;
				break;
			}
		}
		if (!code)
			return(0);
	}
	return(vStatusToBit(code));
}


CodeString_t	vUseTab[] = {
	{ U_IDLE,	"IDLE" },
	{ U_MOUNT,	"MOUNT" },
	{ U_UNMOUNT,	"UNMOUNT" },
	{ U_MOVE,	"MOVE" },
	{ U_SCRATCH,	"SCRATCH" },
	{ 0,		NULL }
};


/*
 *  vClearError() - clear volman VMError & nLog err
 */
 void
vClearError()
{
	VMError = 0;
	nClearError();
}

/*
 *  vGetID() - generate somewhat unpredictable ID
 */
static unsigned idseed = 0;
 
  Id_t
vGetID()
{
	char		*id = "vGetID";
	struct timeval	now;
	struct timezone	zone;
	unsigned	cnt, seed;

	/*
	 *  use microseconds to construct the random number seed
	 */
	if (gettimeofday(&now, &zone) == -1) 
		nLogMsg(EL_FATAL|EL_SYS, id, "gettimeofday");

	/*
	 *  first time through, seed system's random # generator
	 */
	if (!idseed)
		srand((unsigned int)now.tv_usec);

	/*
	 *  use system's random # to pick a number 0..64
	 */
	cnt = ( ((unsigned)rand()) >>6) % 64;

	/*
	 *  make new seed for 2nd random # generator and seed it
	 */
	idseed += (unsigned int)now.tv_usec;
	seed = idseed;

	/*
	 *  index into 2nd random sequence by number 0..64 from
	 *	system random sequence
	 */
	while (cnt--)
        	seed = seed * 78125 + 59049;

	return((Id_t)seed);

}


/*
 *  vErrStr(): convert VM error number to string
 */
  char *
vErrStr(int err)
{
	return (nCodeToStr(err, vVMErrTab));
}


/*
 *  vReqStateStr(): return request request state string
 */
  char *
vReqStateStr(int state)
{
	return (nCodeToStr(state, vReqStateTab));
}

  char *
vPortTypeStr(int type)
{
	return (nCodeToStr(type, vPortTab));
}

  char *
vUseStr(int state)
{
	return (nCodeToStr(state, vUseTab));
}

/*
 *  vVolStateStr(): return volume request state string
 */
  char *
vVolStateStr(int state)
{
	return (nCodeToStr(state, vStatTab));
}



/*
 *  vHdrStr() - return a (static) string which describes the header (no \n)
 */
  char *
vHdrStr(vmpHdr_t *hp)
{
	static char	buf[LINELEN];

	if (hp == NULL)
		return ("[NULL header]");
	if (hp->vh_magic != VMP_MAGIC) {
		sprintf(buf, "0x%llx: bad magic number: 0x%x/0x%x",
		  (long long)(long)hp, hp->vh_magic, VMP_MAGIC);
		return (buf);
	}
	sprintf(buf, "typ=%s seq=#%d sz=%d ",
	  		nMsgType(hp->vh_type, VmpTab), hp->vh_seq, hp->vh_size);
	return(buf);
}


/*
 *  vDevMapStr() - return a (static) string which describes the device map
 */
  char *
vDevMapStr(DevMap_t *dmp)
{
	static char	buf[LINELEN * 2];
	char		tbuf[LINELEN];
	int		i, bitset;

	bitset = 0;
	buf[0] = '\0';
	for (i = 0; i < MAXDEVS; i++)
		if (DM_ISSET(i, *dmp)) {
			sprintf(tbuf, "%d ", i);
			strcat(buf, tbuf);
			bitset++;
		}

	if (bitset == 0)
		return ("[none]");
	buf[strlen(buf) - 1] = '\0';	/* get rid of extra space */
	return(buf);
}


#ifdef DEBUG
/*
 *  vDumpHdr() - print out information about the passed in header on the
 *	passed in stream
 */
vDumpHdr(vmpHdr_t *hp, FILE *str)
{
	char	*line = "--------------------------------------------------\n";

	if (hp == NULL)
		return;
	fprintf(str, line);
	if (hp->vh_magic != VMP_MAGIC) {
		fprintf(str, "0x%x: bad magic number: 0x%x/0x%x\n",
		  hp, hp->vh_magic, VMP_MAGIC);
	}
	fprintf(str, "addr	type	type	seq	size	key\n");
	fprintf(str, "0x%x	0x%x	%s	%d	%d	%d\n",
	  hp, hp->vh_type, nMsgType(hp->vh_type), hp->vh_seq, 
	  hp->vh_size, hp->vh_key);
	fprintf(str, line);
}
#endif


/*
 *  vGetSeq() - generate a sequence number
 */
  Seq_t
vGetSeq()
{
	static Seq_t	lastseq;

	if (lastseq == 0) {
		/* generalize to lPid */
		srand((int) time((time_t *) 0) + getpid());
		lastseq = (Seq_t) rand();
	}
	return (++lastseq);
}



char	*vInstErrBuf;
/*
 *  vInstallError() - install the error code and string from the passed-in
 *	M_VERROR packet to the global error storage vars
 *    Minor kludge: lower levels of the library running in the application
 *	print errors to stderr when encountered; errors installed by
 *	vInstallError aren't printed when encountered but are installed
 *	in the globals.   vm_prerr() will print nErrStr only if it
 *	is not pointing to vInstErrBuf.
 */
  void
vInstallError(vmpError_t *ep)
{
	if (vInstErrBuf == NULL)
		vInstErrBuf = nAlloc((size_t)LINELEN);
	VMError = ep->vn_error;
	memcpy(vInstErrBuf, ep->vn_msg, (size_t)LINELEN);
	vInstErrBuf[LINELEN-1] = '\0';
	nErrStr = vInstErrBuf;
}

/*
 *  vDMffs() - return the bit number of the first bit set in the DevMap_t;
 *	if no bits are set, return -1
 */
int
vDMffs(DevMap_t *dmp)
{
	REG int	i;

	for (i = 0; i < MAXDEVS; i++)
		if (DM_ISSET(i, *dmp))
			return (i);
	return (-1);
}


/*
 *  vDMand() - ANDs two device maps and returns a ptr to a static result
 */
  DevMap_t *
vDMand(DevMap_t *map1p, DevMap_t *map2p)
{
	static DevMap_t		map;
	REG unsigned int	*m1p, *m2p, *mp;
	REG int			count;

	m1p = map1p->dms_bits;
	m2p = map2p->dms_bits;
	mp = map.dms_bits;
	for (count = sizeof (DevMap_t) / sizeof (fd_mask); count > 0; count--)
		*mp++ = *m1p++ & *m2p++;
	return (&map);
}


/*
 *  vDMcnt() - return a count of the number of bits on in a DevMap_t
 */
int
vDMcnt(DevMap_t *map)
{
	REG int	count, bit;

	for (count = bit = 0;  bit < MAXDEVS; bit++)
		if (DM_ISSET(bit, *map))
			count++;
	return (count);
}

int
CheckVO_Flags(int flags)
{
	if (flags & VO_ASYNC  &&  flags & VO_WAIT) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR1, 0, 
				"VO_ASYNC and VO_WAIT incompatible");
		return(1);
	}
	if (countintbits(flags & VO_COMMAND) > 1) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR1, 0, 
				">1 VO_COMMANDs: %x", flags & VO_COMMAND);
		return(1);
	}
	return(0);
}


/*
 *  vCheckPacket() - perform sanity check on header.
 *	Return E_BADPACKET if invalid; assign VMError but don't log error;
 *	let upper level do it.
 */
int
vCheckPacket(vmpHdr_t *hp)
{
	char	*id = "vCheckPacket";
	ssize_t	size;

	if (hp->vh_magic != VMP_MAGIC) {
		char	*cp = (char *) hp;
		int	i;

		nLogMsg(EL_ERROR1, id, "Bad magic: %s", vHdrStr(hp));
		for (i=0; i<sizeof(vmpHdr_t); i++)
			if (cp[i] == VMP_MAGIC)
				nLogMsg(EL_STATUS1, id, "magic at %d", i);
		return (VMError = E_BADPACKET);
	}
	if (hp->vh_type > M_MAXTYPE) {
		nLogMsg(EL_ERROR1, id, "Bad type: %s", vHdrStr(hp));
		return (VMError = E_BADPACKET);
	}
	size = vPktSiz((int) hp->vh_type);
	if (size != -1 && size != hp->vh_size) {  /* -1 => variable length */
		nLogMsg(EL_ERROR1|EL_CODE, id, "Bad size %d/%d: %s", 
					size, hp->vh_size, vHdrStr(hp));
		return (VMError = E_BADPACKET);
	}
	return (E_NOERROR);
}


/*
 *  vBuildHdr() - build a packet header for the specified message type;
 *	if a buffer was passed in, build it there; else alloc a buffer
 *	in any event, the entire buffer is cleared.
 */
  vmpHdr_t *
vBuildHdr(int msg, Seq_t seq, vmpHdr_t *chp)
{
	vmpHdr_t	*hp;
	ssize_t		pktsiz;
	size_t		wholepkt;

	pktsiz = vPktSiz(msg);

	if (pktsiz == -1) 
		nLogMsg(EL_FATAL, 0, 
		 "programming error: using vBuildHdr on custom-size pkt");

	if (chp == NULL) {
		wholepkt = pktsiz + PadSize(sizeof(vmpHdr_t));
		hp = (vmpHdr_t *) nAlloc(wholepkt);
	} else {
		wholepkt = pktsiz + sizeof(vmpHdr_t);
		hp = chp;
	}
	memset(hp, '\0', wholepkt);
	hp->vh_magic = VMP_MAGIC;
	hp->vh_type = msg;
	hp->vh_seq = seq;
	hp->vh_size = pktsiz;
	hp->vh_count = 0;
	hp->vh_slot = -1;

	return (hp);
}


/*
 *  vSigPipe() - called on SIGPIPE, which is generated when a write is
 *	attempted on an ipc fd without a reader
 */
  void
vSigPipe(int sig)
{
	char	*id = "vSigPipe";

#if	defined(lint)
	sig = sig;
#endif
	if (signal(SIGPIPE, vSigPipe) == SIG_ERR)
		nLogMsg(EL_FATAL|EL_SYS, id, "signal(SIGPIPE)");
	vDeadReader++;				/* set global flag */
	nLogMsg(EL_DEBUG2, id, "SIGPIPE");
}

  void
vInitDirNames()
{
#ifdef DEBUG
	static int	called;
	char		*env, *getenv();

	if (!called) {
		if ((env = getenv(ALTSPOOLDIR)))
			vSpoolDir = env;
		if ((env = getenv(ALTLIBDIR)))
			vLibDir = env;
		if ((env = getenv(ALTNODEDIR)))
			vNodeDir = env;
		called = 1;
	}
#endif	/* DEBUG */
	/*
	 *  vClientDir - concat client dir on spooldir
	 */

	vClientDir = nAlloc(sizeof(VSPOOLDIR) + sizeof(CLIENTDIR) + 5);
        sprintf(vClientDir, "%s/%s", VSPOOLDIR, CLIENTDIR);
	vLogFile = nAlloc(strlen(vSpoolDir) + strlen(ERRORLOG) + 5);
        sprintf(vLogFile, "%s/%s", vSpoolDir, ERRORLOG);
}

/*
 *  vIDFile() - construct name of file to read id from - this
 *	may or may not be used, depending on the ipc
 */

char	*
vIDFile(Id_t anum)
{
	static	char	buf[sizeof(VSPOOLDIR)+sizeof(CLIENTDIR)+40];

	if (!anum)
		anum = 1;
	sprintf(buf, "%s/%d", vClientDir, anum );
	return(buf);
}

  void
copytag(char *to, char *from)
{
	int	i;

	memset(to, 0, (size_t)TAGLEN);

	if (from == NULL)
		return;
	for (i=0; i<TAGLEN; i++) {
		if (!isprint(from[i]))
			break;
		to[i] = from[i];
	}
}

static char	errbuf[256];
static int	serrno, svmerr;
static int	saved = 0;

void
SaveErrs()
{
	char	*id = "SaveErrs";

	if (saved)
		nLogMsg(EL_FATAL, id, "already saved");
	if (strlen(nErrStr) > sizeof(errbuf) - 1)
		nLogMsg(EL_FATAL, id, "string too long");
	strcpy(errbuf, nErrStr);
	serrno = errno;
	svmerr = VMError;

	saved = 1;
}

void
RestoreErrs()
{
	if (!saved)
		nLogMsg(EL_FATAL, "RestoreErrs", "nothing saved");

	strcpy(nErrStr, errbuf);
	errno = serrno;
	VMError = svmerr;

	saved = 0;
}
