/*
 *  Volume Manager
 *
 *   vm_sock.c: Volume Manager basic send/recv functions
 */

#include "vm_space.h"

#ifdef sun
# include <sys/file.h>  /* for FNDELAY */
#endif


int	vDeadWriter = 0, vDeadReader = 0;

int
SyncIPC(IPC_t *ipc)
{
	char	*id = "SyncIPC";
	int	sockflag;

	/* 
	 *  To prevent failed writes/retries to a non-blocking socket,
	 *  reset socket for blocking writes.
	 */

	if ( (sockflag = fcntl(ipc->sock_fd, F_GETFL, 0) ) == 0 ) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "fcntl F_GETFL failed id %d",
							ipc->sock_id);
		return(1);
	}
	sockflag &= ~FNDELAY;

	if ( (fcntl(ipc->sock_fd, F_SETFL, sockflag) ) < 0 ) {
		nLogMsg(EL_ERROR1|EL_SYS, id, "fcntl F_SETFL failed id %d",
							ipc->sock_id);
		return(1);
	}
	return(0);
}

/*
 *  vSndPkt() and vRcvPkt() do error checking and retries, also
 *	Snd2Client() detects a dead reader and initiates client shutdown
 *
 *  vSend() and vRecv() are the lower-level interface which send and
 *	receive packets but do no error checking or retries
 */

int
vSndPkt(vmpHdr_t *hp, IPC_t *sockp)
{
	char	*id = "vSndPkt";
	int	rcnt;			/* retry count */

	vClearError();
	if (vCheckPacket(hp) != E_NOERROR) {
		nLogMsg(EL_CODE|EL_ERROR1, id, "Tried to send bad packet");
		return (VMError);
	}
	hp->vh_key = sockp->sock_sndkey;
/*
nLogMsg(EL_STATUS1, id, "key %d", hp->vh_key);
*/
	for (rcnt = 0; rcnt < IORETRY; rcnt++) {
		vSend(hp, sockp);
		if (VMError == E_NOERROR  ||  VMError == E_CANCELED)
			break;
		sleep(1);
	}
	return (VMError);
}


 vmpHdr_t *
vRcvPkt(IPC_t *sockp, int delay)
{
	char		*id = "vRcvPkt";
	vmpHdr_t	*hp;

	vClearError();
	if ((hp = vRecv(sockp, delay)) == NULL) {
		if (vDeadWriter)
			nLogMsg(EL_CODE|EL_ERROR1, id, 
			   "Nothing read, writer died or closed connection");
		return (NULL);
	}
/*
nLogMsg(EL_STATUS1, id, "key %d expecting %d (send %d)", 
hp->vh_key, sockp->sock_rcvkey, sockp->sock_sndkey);
*/
	if (sockp->sock_state != SS_CONNECTED  &&
	    hp->vh_key != sockp->sock_rcvkey) {
		VMError = E_COMM;
		nLogMsg(EL_CODE|EL_ERROR1, id, "bad key");
		nFree(hp);
		return(NULL);
	}
	if (vCheckPacket(hp) != E_NOERROR) {
		/* VMError set already */
		nLogMsg(EL_CODE|EL_ERROR1, id, "Bad packet: %s", vHdrStr(hp));
		nFree(hp);
		return (NULL);
	}
	return (hp);
}


/*
 * vSend(): write packet to socket
 *	header buffer is provided by caller and not deallocated here
 */
int
vSend(vmpHdr_t *hp, IPC_t *sockp)
{
	char		*id = "vSend";
	int		n = 0;
	ssize_t		wcnt = 0;		/* write count */
	size_t		pktsiz;		/* packet size */

#ifdef DEBUG
	if (vCheckPacket(hp))		/* perform sanity check on packet */
		return (VMError);
#endif
	if ( hp->vh_size )
		pktsiz = PadSize(sizeof(vmpHdr_t)) + hp->vh_size;
	else
		pktsiz = sizeof(vmpHdr_t);
	errno = 0;
	wcnt = send(sockp->sock_fd, (char *)hp, pktsiz, 0);
	/*
	 *  Keep retrying the send if it would normally block.
	 *  It should go thru eventually.
	 */
	while ((wcnt == -1) && (errno == EWOULDBLOCK) && (n++ < IORETRY)) {
		sleep(1);
		nLogMsg(EL_DEBUG2, id, "send would block, retry %d ...", n);
		wcnt = send(sockp->sock_fd, (char *)hp, pktsiz, 0);
	}
	if ((wcnt == -1) && (n >= IORETRY))
		nLogMsg(EL_DEBUG1, id, "retries exhausted, giving up\n");

	if (wcnt != pktsiz) {
		VMError = E_COMM;	/* default err */
		if (wcnt == -1) {
			if (errno == EPIPE) { /* TODO - see what err we get.. */
				VMError = E_CANCELED;
				nLogMsg(EL_CODE|EL_ERROR2, id, 
					"connection terminated by %d", 
					sockp->sock_id);
			} else { 
				nLogMsg(EL_CODE|EL_ERROR1|EL_SYS, id, 
					"write (%s) failed to %d", 
					vHdrStr(hp), sockp->sock_id);
			}
		} else {
			nLogMsg(EL_CODE|EL_DEBUG1, id,
			  		"write (%s) failed %d/%d bytes to %d",
			  		vHdrStr(hp), wcnt, pktsiz,
					sockp->sock_id);
		}
	} else {
		unsigned int	lev = EL_DEBUG2;

		if (hp->vh_type == M_TIMER)
			lev = EL_DEBUG3;	/* bump level for timer ticks */
		nLogMsg(lev, id, "[%s] to %d", vHdrStr(hp), sockp->sock_id);
	}
	return (VMError);
}


/*
 *  FUNCTION:	ReadSock() 
 *		Read message from socket.
 *		Checks for fragmented messages & dead clients.
 */
  static ssize_t
ReadSock(IPC_t *sockp, char *bp, size_t bytes, char *what)
{
	char		*id = "ReadSock";
	ssize_t		bytereq = bytes, nbytes;

	for (;;) {
		if (sockp->sock_flags & IPC_NDELAY)
        		alarm(VM_SOCKTIMEOUT);
        	nbytes = recv(sockp->sock_fd, bp, bytes, 0);
                if (nbytes < 0) {
			int	saverr = errno;

                        alarm(0);
			errno = saverr;
                        if (errno == EWOULDBLOCK) {

				/*
				 *  (If it _is_ EWOULDBLOCK, there's
				 *	nothing to read & it's a non-
				 *	blocking read.)
				 */
				return(0);
			}
                        VMError = E_COMM;
			if (errno == ECONNRESET)
				nLogMsg(EL_DEBUG2, id, 
						"connection %d reset by peer",
						sockp->sock_id);
			else
                            nLogMsg(EL_CONSOLE0|EL_ERROR1|EL_CODE|EL_SYS, id,
				"Read %s failed - Connection %d died",
				what, sockp->sock_id);
			return(-1);
                } else {
                        alarm(0);
                        if (nbytes == bytes) {
				/*
				 *  got whole request or balance
				 */
                                nLogMsg(EL_DEBUG4, id, "read %s OK", what);
                                return(bytereq);
                        } else if (nbytes == 0) {
				VMError = E_COMM;
                                nLogMsg(EL_DEBUG4, id,
					"Client %d exited", sockp->sock_id);
				vDeadWriter = 1;
				return(-1);
                        } else if (nbytes < bytes) {
				/*
				 *  got too few bytes
				 */
                                nLogMsg(EL_ERROR1, id, "read %s fragment",what);
                                bp += nbytes;
                                bytes -= nbytes;
                        } else {
				/*
				 *  got too many bytes - should never happen!
				 */
                                nLogMsg(EL_ERROR1|EL_CONSOLE0, id,
                                        "BADPKT: read %d bytes instead of %d",
                                                nbytes, bytes);
                                nbytes = -1;
                                return(-1);
                        }
                }
        }
}


/*
 *  FUNCTION:	vRecv() 
 *		Receive packet from socket, return pointer to packet buffer;
 *		buffer is allocated here and deallocated by caller.
 */
 vmpHdr_t *
vRecv(IPC_t *sockp, int delay)
{
	char		*id = "vRecv";
	vmpHdr_t	*hp1, *hp2;
	char		buf[sizeof(vmpHdr_t) + 8];
	int		i, j;
	unsigned int	lev;

	vDeadWriter = 0;			/* reset global flag */

	if (sockp->sock_fd == -1)
		nLogMsg(EL_FATAL, id, "recv from unopened fd");

	if (delay != (sockp->sock_flags & IPC_NDELAY)) {
		int	delayflg = (delay == IPC_DELAY ? 0 : FNDELAY);

		/*
		 *  change socket blocking status
		 */

		if (fcntl(sockp->sock_fd, F_SETFL, delayflg) == -1) {
			nLogMsg(EL_ERROR1|EL_CONSOLE0|EL_SYS, id,
				"fcntl failed, id %d", sockp->sock_id);
			return(NULL);
		}
		sockp->sock_flags &= ~IPC_NDELAY;
		sockp->sock_flags |= delay;
	}

	/*
	 *  receive header & check it
	 */
	hp1 = (vmpHdr_t *)buf;
	for (i=0; i<10; i++) {
		/*
		 *  get 'magic' beginning-of-pkt char
		 */
		for (j=0;;j++) {
			if (ReadSock(sockp, buf, (size_t)1, "magic ") != 1)
				return(NULL);
			if (buf[0] == VMP_MAGIC)
				break;
		}
		if (j)
		  nLogMsg(EL_CONSOLE0|EL_ERROR2, id, 
		   "bad client or socket problem: %d leading bytes garbage", j);
		/*
		 *  get rest of hdr
		 */
		if (ReadSock(sockp, buf+1, sizeof(vmpHdr_t)-1, "hdr ") != 
					sizeof(vmpHdr_t)-1)
			return(NULL);
		/*
		 *  if hdr ok, go on
		 */
		if (!vCheckPacket(hp1))
			break;
	}
	if (vCheckPacket(hp1)) 
		return(NULL);

	/*
	 *  allocate a buffer for entire packet, including header
	 *	& copy header to it
	 */

	hp2 = (vmpHdr_t *) nAlloc(PadSize(sizeof(vmpHdr_t)) + hp1->vh_size);
	memcpy((char *)hp2, buf, sizeof(vmpHdr_t));

	if (hp1->vh_size > 0) {
		size_t	pktsize = 
			  hp1->vh_size + PadSize(sizeof(*hp2)) - sizeof(*hp2);
		/*
		 *  read the rest of the message
		 */

		nLogMsg(EL_DEBUG4, id, "get msg: %d bytes", hp1->vh_size);

		if (ReadSock(sockp, (char *)(hp2+1), pktsize, "msg ")
							!= pktsize) {
			return(NULL);
		}
	}
	lev = EL_DEBUG1;
	if (hp2->vh_type == M_TIMER)
		lev = EL_DEBUG4;		/* bump level for timer ticks */
	nLogMsg(lev, id, "[%s] from %d", vHdrStr(hp2), sockp->sock_id);

        return(hp2);
}



/*
 *  FUNCTION:	CloseIPC()
 *		Shut down socket depending on state and type.
 *		No return value.
 */
  void
CloseIPC(IPC_t *sp)
{
	char	*id = "CloseIPC";

	switch (sp->sock_state) {
	    case SS_NOSOCK:
		nLogMsg(EL_DEBUG3, id, "shutdown on non-socket %d",
			sp->sock_id);
		break;
	    case SS_CONNECTED:
	    case SS_HELLOD:
	    case SS_TALKING:
	    case SS_GOODBYE:
		shutdown(sp->sock_fd, 2);
		close(sp->sock_fd);
		break;
	    default:
		nLogMsg(EL_ERROR1, id, "invalid socket state %d, id %d",
			sp->sock_state, sp->sock_id);
		break;
	}

        sp->sock_fd = -1;
        sp->sock_id = -1;
	sp->sock_host = NULL;
	sp->sock_proc = NULL;
        sp->sock_state = SS_NOSOCK;
	sp->sock_ptype = PT_NONE;
}

