
/*
 *  rl_lib.c - used by vold & its servers
 */
#include "rl_space.h"
#include <pwd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <signal.h>

/*
 *  CheckUID() - make sure running as root
 */
  void
CheckUID(char *str)
{
	struct passwd	*pwp, *getpwnam();
	int		uid;

	/*
	 *  get the volume manager operator's uid (set VOLMANUID)
	 */
	if ((pwp = getpwnam(VolManOperAcct)) == NULL) {
		fprintf(stderr, "%s: there is no %s account (must be one).", 
					str, VolManOperAcct);
		exit(1);
	}
	VolManOperUID = pwp->pw_uid;
#ifndef PURIFY
	/*
	 *  must be run by RPUID;  if run by VOLMANUID, attempt a setuid()
	 */
	if ((uid = getuid()) != 0  &&  uid != VolManOperUID) {
		fprintf(stderr, "%s: must be run as %s account or root.\n", 
					str, VolManOperAcct);
		exit(1);
	}
	if (uid == VolManOperUID) {
		if (setuid(0) < 0) {
			perror("setuid failed");
			exit(1);
		}
	}
#endif
}

/*
 *  ForkExec(): path can be absolute, or it is searched for in LIBDIR
 */
  int
ForkExec(char *path, char *name, char *arg)
{
	char	*id = "ForkExec";
	char	buf[MAXPATHLEN];
	STAT_T	statb;
	int	fval;

	/* executable in libdir if not absolute path */
	if (path[0] != '/') {
		sprintf(buf, "%s/%s", vBinDir, path);
		path = buf;
	}

	/*
	 *  simple checks before fork
	 */
	if (STAT(path, &statb) == -1) {
		if (errno == ENOENT)
			nLogMsg(EL_FATAL, id, "%s: (%s): does not exist",
		  			name, path);
		nLogMsg(EL_FATAL|EL_SYS, id, "%s: (%s): Can't stat",
		  			name, path);
	}
	if (!(statb.st_mode & 0111)) {
		nLogMsg(EL_FATAL, id, "%s: (%s): not executable (0%o)",
		  			name, path, statb.st_mode);
	}
	if ((fval = fork()) < 0) {
		nLogMsg(EL_FATAL|EL_SYS, id, "%s: (%s): Can't fork",
		  name, path);
	} else if (fval == 0) {			/* child */
		int	i;

		nLogMsg(EL_STATUS1, id, "Exec: [pid %d] %s (%s)",
		  getpid(), name, path);
		for (i = 0; i < getdtablesize(); i++)	/* close all fds */
			if (i >= 3)	/* start RC off with clean slate */
				(void) close(i);	/* TODO: use v struct */
		execl(path, name, arg, 0);

		/*
		 * this message won't write to the error log
		 *	becuase the fd's have been closed; stderr's OK
		 */
		nLogMsg(EL_SYS|EL_TOSTDERR|EL_CONSOLE0, id,
					"Can't exec %s (%s)", name, path);
		exit(1);
	}
	return(fval);
}

static int		reserved_fd = -1;

  void
ReserveFD(char *str)
{
	char	*id = "ReserveFD";
	int	i;

	if (reserved_fd != -1)
		nLogMsg(EL_FATAL|EL_CONSOLE0, id, "programmer error");
	
	for (i=1; i<15; i++) {
		reserved_fd = open("/tmp/.vold", O_RDWR|O_CREAT, 0600);
		if (reserved_fd != -1)
			return;
		nLogMsg(EL_ERROR1|EL_SYS, id, "open (%s)", str);
		(void)sleep((unsigned int)(i * 2));
	}

	nLogMsg(EL_FATAL|EL_CONSOLE0, id, "CAN'T RESERVE FILE DESCRIPTOR");
}

  void
UnreserveFD(char *str)
{
	char	*id = "UnreserveFD";

	if (reserved_fd == -1)
		nLogMsg(EL_CONSOLE0|EL_FATAL, 0, "unreserve unopened fd (%s)",
									str);
	if (close(reserved_fd) == -1)
		nLogMsg(EL_SYS|EL_ERROR1, id, "close");
	reserved_fd = -1;
}

  void
vSigAlarm(int signo)
{
	nLogMsg(EL_STATUS1, 0, "vSigAlarm");
	if (SIGNAL(SIGALRM, vSigAlarm))
		nLogMsg(EL_FATAL, 0, "signal SIGALRM");
}

  void
SigHandle(int signo)
{
	nLogMsg(EL_FATAL, "SigHandle", "Caught signal %d", signo);
}

  void
SigInit()
{
	char	*id = "SigInit";
	int	sig;

	if (SIGNAL(SIGALRM, vSigAlarm))
		nLogMsg(EL_FATAL, id, "signal SIGALRM");
	if (SIGNAL(SIGPIPE, vSigPipe))
		nLogMsg(EL_FATAL, id, "signal SIGPIPE");

	for (sig = 1; sig < NSIG; sig++)
		switch (sig) {
		    case SIGABRT:	/* abort always causes core dump */
		    case SIGSEGV:
		    case SIGTRAP:
		    case SIGQUIT:
		    case SIGBUS:
		    case SIGCHLD:
		    case SIGKILL:	/* can't be caught */
		    case SIGSTOP:	/* can't be caught */
#ifdef SIGTSTP
		    case SIGTSTP:
#endif
#ifdef SIGCONT
		    case SIGCONT:
#endif
		    case SIGALRM:
		    case SIGPIPE:
			break;

		    default:
			if (SIGNAL(sig, SigHandle))
				nLogMsg(EL_FATAL, id, "signal(%d, SigHandle)", 
									sig);
		}
}
/*
 *  CreateEmptyDir(): if dirnam doesn't exist, create it; if it does, 
 *	empty it of all files.  If directory can't be created,
 *	it's a fatal error
 */
  void
CreateEmptyDir(char *dirnam)
{
	DIR		*dirp;
	struct dirent	*dep;
	char		*id = "CreateEmptyDir";

        if ((dirp = opendir(dirnam)) == NULL) {
		if (mkdir(dirnam, 0755) < 0) 			/* create it */
			nLogMsg(EL_FATAL|EL_SYS, id, "can't create %s: mkdir",
			  dirnam);
		return;
        }
	while (dep = readdir(dirp)) {
		STAT_T	sbuf;
		char	buf[MAXPATHLEN];
		
		if ((strcmp(".", dep->d_name) == MATCH) ||
		  (strcmp("..", dep->d_name) == MATCH))
			continue;
		sprintf(buf, "%s/%s", dirnam, dep->d_name);
		if (STAT(buf, &sbuf) == -1) {
			nLogMsg(EL_ERROR2|EL_SYS, id, "stat %s:", buf);
		} else {
			nLogMsg(EL_STATUS1, id, "unlinking %s", buf);
			(void) unlink(buf);
		}
	}
	closedir(dirp);
}
