int sayit = 0;
/*
 *  rp_xdrlib.c: Volume Manager vold / (volcd, volnode) XDR routines
 *
 *  Bill Ross, NASA Ames, August 1995
 */

#include "rl_space.h"
#include <rpc/xdr.h>

/*
 *	Vold/volcd communication is over sockets via XDR, which is 
 *	a protocol (part of RPC, which is used by NFS) for making 
 *	sure that data structures are the same on hosts of different 
 *	architecture.
 *
 *	On the vold side, an array of sockets indexed by file descriptor
 *	is tied to a corresponding array of XDR structures. These
 *	are used for all vold interprocess communication. The vold's
 *	servers each have a single socket/XDR connection with the vold.
 *
 *	For each connection, an XDR_t struct is initialized to point at 
 *	the IPC_t (socket connection mgt) structure and the routines
 *	that read & write raw bytes from the socket. 
 *
 *	xdr_{structname} routines generated by rpcgen from vm_pkt.x are 
 *	passed the XDR_t, and call primitive routines like xdr_int() 
 *	which invoke the socket read/write routines to get/put bytes 
 *	before/after converting data from/to the generic form XDR uses.
 *
 *	Higher-level packet-switched routines in this file read/write
 *	the header of a packet using xdr_vmpHdr_t() and then get/put
 *	the rest of the packet based on pkt->vh_type using the various
 *	other xdr_vmp{pktstruct} routines.
 *
 */


/*
 *  XDRread() - called by XDR translation routines, hooked by xdrp
 */
  static int
XDRread(const IPC_t *sockp, char *buf, const int size)
{
	char	*id = "XDRread";
	int	nb;

	if (sockp->sock_fd < 0) {
		nLogMsg(EL_STATUS1, id, "skipped read from sock fd %d id %d",
					sockp->sock_fd, sockp->sock_id);
		return(-1);
	}
        nb = recv(sockp->sock_fd, buf, size, 0);
	if (nb == 0) {
		nLogMsg(EL_STATUS2, id, 
		    "recvd 0 from slot %d - writer died or closed connection", 
		    sockp->sock_fd);
		shutdown(sockp->sock_fd, 2);
		close(sockp->sock_fd);
		errno = ENOTCONN;
		return(-1);
	}
	nLogMsg(EL_DEBUG2, id, "recvd %d of %d from %d", 
					nb, size, sockp->sock_fd);
	return(nb);
}

/*
 *  XDRwrite() - called by XDR translation routines, hooked by xdrp
 */
  static int
XDRwrite(IPC_t *sockp, char *buf, int size)
{
	char	*id = "XDRwrite";
	int	wcnt, n = 0;

	if (sockp->sock_fd < 0) {
		nLogMsg(EL_STATUS1, id, "skipped write to sock %d id %d",
					sockp->sock_fd, sockp->sock_id);
		return(-1);
	}

	wcnt = send(sockp->sock_fd, buf, size, 0);
	nLogMsg(EL_DEBUG2, id, "sent %d to %d", wcnt, sockp->sock_fd);
if (sayit)
nLogMsg(EL_STATUS1, id, "sent %d to %d", wcnt, sockp->sock_fd);
	/*
	 *  Keep retrying the send if it would normally block.
	 *  It should go thru eventually.
	 */
	while ((wcnt == -1) && (errno == EWOULDBLOCK) && (n++ < IORETRY)) {
		sleep(1);
		nLogMsg(EL_DEBUG2, id, "send would block, retry %d ...", n);
		wcnt = send(sockp->sock_fd, buf, size, 0);
	}
	if ((wcnt == -1) && (n >= IORETRY))
		nLogMsg(EL_DEBUG1, id, "retries exhausted, giving up\n");

	if (wcnt != size) {
		VMError = E_COMM;	/* default err */
		if (wcnt == -1) {
			if (errno == EPIPE) { /* TODO - see what err we get.. */
				VMError = E_CANCELED;
				nLogMsg(EL_CODE|EL_ERROR2, id, 
					"connection terminated by %d", 
					sockp->sock_id);
			} else { 
				nLogMsg(EL_CODE|EL_ERROR1|EL_SYS, id, 
					"write failed to %d", sockp->sock_id);
			}
		} else {
			nLogMsg(EL_CODE|EL_DEBUG1, id,
			  		"write failed %d/%d bytes to %d",
			  		wcnt, size, sockp->sock_id);
		}
		shutdown(sockp->sock_fd, 2);
		close(sockp->sock_fd);
	} else {
		nLogMsg(EL_DEBUG3, id, "sent to %d", sockp->sock_id);
	}
	return(wcnt);
}

/*
 *  XDRSetup() - hook xdrp to client & read/write routines
 */
  void
XDRSetup(IPC_t *sockp, XDR *xdrp)
{
	xdrrec_create(xdrp, 0, 0, (caddr_t) sockp, 
			(int (*) (char *, char *, int) )XDRread, 
			(int (*) (char *, char *, int) )XDRwrite);
	if (!xdrrec_skiprecord(xdrp))
		nLogMsg(EL_STATUS1, 0, "init xdr_skiprecord failed");
}

/*
 *  XDRSend() - basic IPC send
 */
  int
XDRSend(vmpHdr_t *hp, XDR *xdrp, int sendnow)
{
	char	*id = "XDRSend";
	char	*cp;
	int	i;

	nLogMsg(EL_DEBUG2, id, "key %d %s", hp->vh_key, vHdrStr(hp));
	if (hp->vh_type < M_MINTYPE  ||  hp->vh_type > M_MAXTYPE) {
		nLogMsg(EL_ERROR1|EL_CONSOLE0, id, "bad M_ msg type %d", 
							hp->vh_type);
		return(1);
	}

	/*
	 *  set XDR to write
	 */
	xdrp->x_op = XDR_ENCODE;

	/*
	 *  packet header
	 */
	if (!xdr_vmpHdr_t(xdrp, hp)) {
		nLogMsg(EL_ERROR1, id, "pkt hdr send failed");
		return(1);
	}

	if (hp->vh_size == 0)
		goto endrec;

	/*
	 *  packet body
	 */
	cp = MSGPTR(hp, char);
	switch (hp->vh_type) {
	    case M_TIMER:
		if (!xdr_vmpTimer_t(xdrp, (vmpTimer_t *)cp))
			return(1);
		break;
	    case M_CREATENODE:
	    case M_DELETENODE:
	    case M_SCRATCH:
		if (!xdr_vmpNode_t(xdrp, (vmpNode_t *)cp)) {
			nLogMsg(EL_ERROR1, id, "pkt body send failed");
			return(1);
		}
		break;
	    case M_NEWCLIENT:
	    case M_HELLO:
	    case M_CONSOLE:
	    case M_NODED:
	    case M_VOLCD:
	    case M_RC:
	    case M_RCREADY:
	    case M_TIMERD:
		if (!xdr_vmpHello_t(xdrp, (vmpHello_t *)cp)) 
			return(1);
		break;
	    case M_USRMOUNT:
		if (!xdr_vmpUsrMount_t(xdrp, (vmpUsrMount_t *)cp))
			return(1);
		break;
	    case M_USRMOUNT_PLL:
	    case M_USRMOUNT_SEL: {
		vmpUsrMountN_t	*mp = (vmpUsrMountN_t *) cp;

		if (!xdr_vmpUsrMountN_t(xdrp, mp))
			return(1);
		cp = (char *) mp + sizeof(*mp);
		if (!xdr_opaque(xdrp, cp, hp->vh_count * VSNLEN))
			return(1);
		break;
	    }
	    case M_RCMOUNT:
		if (!xdr_vmpRCMount_t(xdrp, (vmpRCMount_t *)cp)) 
			return(1);
		break;
	    case M_UNMOUNT:
		if (!xdr_vmpVsn_t(xdrp, (vmpVsn_t *)cp))
			return(1);
		break;
	    case M_HOSTINFO: {
		HostInfo_t	*hop = (HostInfo_t *) cp;
		for (i=0; i<hp->vh_count; i++, hop++) {
			if (!xdr_HostInfo_t(xdrp, hop))
				return(1);
		}
		break;
	    }
	    case M_MOVE: {
		vmpMove_t	*mp = (vmpMove_t *) cp;
		vsnseq_t	*vqp;

		/*  move header */
		if (!xdr_vmpMove_t(xdrp, mp)) 
			return(1);
		/* vsns */
		vqp = (vsnseq_t *) ((char *)mp + PadSize(sizeof(*mp)));
		for (i=0; i<hp->vh_count; i++, vqp++)
			if (!xdr_vsnseq_t(xdrp, vqp))
				return(1);
		break;
	    }
	    case M_IDLETIME: {
		vmpNodeTime_t	*ip = (vmpNodeTime_t *) cp;
		for (i=0; i<hp->vh_count; i++, ip++)
			if (!xdr_vmpNodeTime_t(xdrp, ip))
				return(1);
		break;
	    }
	    case M_VALLOC:
		if (!xdr_vmpVAlloc_t(xdrp, (vmpVAlloc_t *)cp))
			return(1);
		break;
	    case M_VARY:
		if (!xdr_vmpVary_t(xdrp, (vmpVary_t *)cp))
			return(1);
		break;
	    case M_DRIVEINFO: {
		int		errs = 0;
		DevEnt_t	*dp;

		/* devices */
		dp = (DevEnt_t *) cp;
		for (i=0; i<hp->vh_count; i++, dp++) {
			if (!xdr_DevEnt_t(xdrp, dp))
				return(1);
			if (dp->de_errstr) {
				errs++;
			}
		}
		if (errs) {
			cp = (char *) dp;
			/* error strings */
			if (!xdr_opaque(xdrp, cp, errs * ERRSTRSIZE))
				return(1);
		}
		break;
	    }
	    case M_VERROR:
		if (!xdr_vmpError_t(xdrp, (vmpError_t *)cp)) 
			return(1);
		break;
	    case M_QUOTQRY:
		if (!xdr_vmpQuota_t(xdrp, (vmpQuota_t *)cp))
			return(1);
		break;
	    case M_SETQUOTA:
	    case M_QUOTRPLY:
		if (!xdr_vmpQuota_t(xdrp, (vmpQuota_t *)cp))
			return(1);
		break;
	    case M_RCMSG:
		if (!xdr_vmpRcMsg_t(xdrp, (vmpRcMsg_t *)cp))
			return(1);
		break;
	    case M_ATTACH:
	    case M_VOLQRY:
	    case M_VSN:
	    case M_DELVSN:
	    case M_VCHMOD:
	    case M_VCHGRP:
	    case M_VCHOWN:
	    case M_AUDIT:
	    case M_REQUEUE:
	    case M_VRECYCLE:
		if (!xdr_vmpVsn_t(xdrp, (vmpVsn_t *)cp))
			return(1);
		break;
	    case M_SHUTDOWN:
		if (!xdr_vmpError_t(xdrp, (vmpError_t *)cp))
			return(1);
		break;
	    case M_VDR:
		if (!xdr_vmpVdr_t(xdrp, (vmpVdr_t *)cp))
			return(1);
		break;
	    case M_VOLSTAT:
		if (!xdr_vmpVolStat_t(xdrp, (vmpVolStat_t *)cp))
			return(1);
		break;
	    case M_VOLSTATS: {
		vmpVolStat_t	*vsp = (vmpVolStat_t *) cp;
		for (i=0; i<hp->vh_count; i++, vsp++) {
			if (!xdr_vmpVolStat_t(xdrp, vsp))
				return(1);
		}
		break;
	    }
	    case M_VLS:
		if (!xdr_vmpVls_t(xdrp, (vmpVls_t *)cp))
			return(1);
		break;
	    case M_RCINFO: {
		RC_t	*rcp = (RC_t *) cp;
		for (i=0; i<hp->vh_count; i++, rcp++) {
			if (!xdr_RC_t(xdrp, rcp))
				return(1);
		}
		break;
	    }
	    case M_ADDVSN:
		if (!xdr_vmpAdd_t(xdrp, (vmpAdd_t *)cp))
			return(1);
		break;
	    case M_LBLQRY:
		if (!xdr_vmpLblQry_t(xdrp, (vmpLblQry_t *)cp))
			return(1);
		break;
	    case M_VATTR:
		if (!xdr_vmpAttr_t(xdrp, (vmpAttr_t *)cp))
			return(1);
		break;
	    case M_RESERVE: {
		vmpReserveHdr_t		*vrp;

		/* header */
		vrp = (vmpReserveHdr_t *) cp;
		if (!xdr_vmpReserveHdr_t(xdrp, vrp))
			return(1);

		cp = (char *)vrp + PadSize(sizeof(*vrp));

		if (vrp->re_rcmedcount) {
			/* drives */
			vmpReserveRCMed_t	*rp = (vmpReserveRCMed_t *) cp;
			for (i=0; i<vrp->re_rcmedcount; i++, rp++) {
				if (!xdr_vmpReserveRCMed_t(xdrp, rp))
					return(1);
			}
			rp--;
			cp = (char *)rp + PadSize(sizeof(*rp));
		}
		if (vrp->re_vsncount) {
			/* vsns */
			vmpReserveVsn_t		*rvp = (vmpReserveVsn_t *) cp;
			for (i=0; i<vrp->re_vsncount; i++, rvp++) {
				if (!xdr_vmpReserveVsn_t(xdrp, rvp))
					return(1);
			}
		}
		break;
	    }
	    case M_DEBUG:
		if (!xdr_vmpDebug_t(xdrp, (vmpDebug_t *)cp)) 
			return(1);
		break;
	    case M_LBLOOK:
		if (!xdr_vmpLblLook_t(xdrp, (vmpLblLook_t *)cp)) 
			return(1);
		break;
	    case M_DEVQRY: {
		DevEnt_t	*dp;

		/* devices */
		dp = (DevEnt_t *) cp;
		for (i=0; i<hp->vh_count; i++, dp++) {
			if (!xdr_DevEnt_t(xdrp, dp))
				return(1);
		}
		break;
	    }
	    default:
		nLogMsg(EL_FATAL, id, "unexpected msg  type: %s",
						vHdrStr(hp));
	}
endrec:
	/*
	 *  end record, flush buffer if req
	 */
if (hp->vh_type == M_RCMOUNT) {
sayit = 1;
nLogMsg(EL_STATUS1, id, "%s %s",
(sendnow ? "flush" : "noflush"), vHdrStr(hp));
}
	if (!xdrrec_endofrecord(xdrp, sendnow)) {
		nLogMsg(EL_ERROR1, 0, "xdrrec_endofrecord failed");
		return(1);
	}
sayit=0;
	return(0);
}

/*
 *  XDRNextRecord() - prepare for next receive
 */
  int
XDRNextRecord(XDR *xdrp)
{
	int	more;

	nLogMsg(EL_DEBUG2, 0, "XDRNextRecord");
	/*
	 *  set XDR to read to be on the safe side..
	 */
	xdrp->x_op = XDR_DECODE;

	/*
	 *  finish this record & see if another record follows
	 */
	if (more = !xdrrec_eof(xdrp)) {
		nLogMsg(EL_DEBUG2, 0, "more input!");
	}

	/*
	 *  move to end of record boundary
	 */
	if (!xdrrec_skiprecord(xdrp))
		nLogMsg(EL_STATUS1, 0, "xdr_skiprecord failed");

	return(more);
}

/*
 *  XDRRecv() - vold / volcd,volnd,RC communication
 *
 *  	NOTE: all pointers to stuff w/in hp (e.g. where to add the
 *	next item) need to be _recalced_ after FULLPKT() since nReAlloc 
 *	occurs. This REALLY CAN BITE YOU when dealing w/ complex pkts 
 *	that require multiple FULLPKT()s.
 */
#define FULLPKT(hp, size)  \
	hp = (vmpHdr_t *) nReAlloc((void *) hp, hsize + size); \
	cp = (char *) hp + hsize;	/* beginning of 1st struct */

 vmpHdr_t *
XDRRecv(XDR *xdrp)
{
	char		*id = "XDRRecv";
	vmpHdr_t	*hp;
	int		psize, hsize = PadSize(sizeof(*hp));
	char		*cp;
	int		i, count;

	/*
	 *  set XDR to read
	 */
	xdrp->x_op = XDR_DECODE;

	/*
	 *  packet header - allow for padding to word boundary 
	 *	in case there's a pkt body
	 */
	hp = (vmpHdr_t *) nAlloc(PadSize(sizeof(vmpHdr_t)));

	if (!xdr_vmpHdr_t(xdrp, hp))
		return(NULL);

	if (hp->vh_type < M_MINTYPE  ||  hp->vh_type > M_MAXTYPE) {
		nLogMsg(EL_ERROR1|EL_CONSOLE0, id, "bad M_ msg type %d", 
							hp->vh_type);
		return(NULL);
	}
	nLogMsg(EL_DEBUG2, id, "key %d %s", hp->vh_key, vHdrStr(hp));

	if (hp->vh_size == 0)
		return(hp);

if (hp->vh_type == M_RCMOUNT)
nLogMsg(EL_STATUS1, id, "%s", vHdrStr(hp));
	/*
	 *  there is more to pkt - zero the padsize (which is
	 *	the rounded-to-wordsize gap between vmpHdr_t
	 *	and the next incoming struct. Why zero this
	 *	unused little piece of memory? So that Purify
	 *	won't complain if it gets copied later. This
	 *	just shows that we know it's there. By the
	 *	same token, need to zero all other padding 
	 *	if multiple structs come in. AND: XDR doesn't
	 *	fill the entire 'sizeof' things, so need to
	 *	zero ALL memory, which is tedious due to
	 *	realloc'ing.
	 *
	 *	Also need to set vh_size appropriate for this
	 *	machine.
	 */
#ifdef PURIFY
	cp = (char *) hp + sizeof(*hp);
	memset(cp, 0, PadSize(sizeof(*hp)) - sizeof(*hp));
#endif
	/*
	 *  packet body switch, ordered approx by frequency
	 */
	switch (hp->vh_type) {
	    case M_TIMER:
		hp->vh_size = sizeof(vmpTimer_t);
		FULLPKT(hp, sizeof(vmpTimer_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpTimer_t));
#endif
		if (!xdr_vmpTimer_t(xdrp, (vmpTimer_t *)cp)) {
			nLogMsg(EL_ERROR1, id, "xdr_vmpTimer_t");
			return(NULL);
		}
		break;
	    case M_NEWCLIENT:
	    case M_CONSOLE:
		hp->vh_size = sizeof(vmpHello_t);
		FULLPKT(hp, sizeof(vmpHello_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpHello_t));
#endif
		if (!xdr_vmpHello_t(xdrp, (vmpHello_t *)cp)) {
			nLogMsg(EL_ERROR1, id, "xdr_vmpHello_t");
			return(NULL);
		}
		break;
	    case M_RCINFO: {
		RC_t	*rcp;

		hp->vh_size = hp->vh_count * sizeof(RC_t);
		FULLPKT(hp, hp->vh_size);
#ifdef PURIFY
		memset(cp, 0, hp->vh_size);
#endif
		rcp = (RC_t *)cp;
		for (i=0; i<hp->vh_count; i++, rcp++) {
			if (!xdr_RC_t(xdrp, rcp))
				return(NULL);
		}
		break;
	    }
	    case M_VOLSTAT:
		hp->vh_size = sizeof(vmpVolStat_t);
		FULLPKT(hp, sizeof(vmpVolStat_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpVolStat_t));
#endif
		if (!xdr_vmpVolStat_t(xdrp, (vmpVolStat_t *)cp))
			return(NULL);
		break;
	    case M_VOLSTATS: {
		vmpVolStat_t	*vsp;

		hp->vh_size = hp->vh_count * sizeof(vmpVolStat_t);
		FULLPKT(hp, hp->vh_size);
#ifdef PURIFY
		memset(cp, 0, hp->vh_size);
#endif
		vsp = (vmpVolStat_t *) cp;
		for (i=0; i<hp->vh_count; i++, vsp++) {
			if (!xdr_vmpVolStat_t(xdrp, vsp))
				return(NULL);
		}
		break;
	    }
	    case M_VDR:
		hp->vh_size = sizeof(vmpVdr_t);
		FULLPKT(hp, sizeof(vmpVdr_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpVdr_t));
#endif
		if (!xdr_vmpVdr_t(xdrp, (vmpVdr_t *)cp))
			return(NULL);
		break;
	    case M_USRMOUNT:
		hp->vh_size = sizeof(vmpUsrMount_t);
		FULLPKT(hp, sizeof(vmpUsrMount_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpUsrMount_t));
#endif
		if (!xdr_vmpUsrMount_t(xdrp, (vmpUsrMount_t *)cp))
			return(NULL);
		break;
	    case M_USRMOUNT_PLL:
	    case M_USRMOUNT_SEL:
		hp->vh_size = sizeof(vmpUsrMountN_t) + hp->vh_count * VSNLEN;
		FULLPKT(hp, hp->vh_size);
#ifdef PURIFY
		memset(cp, 0, hp->vh_size);
#endif

		if (!xdr_vmpUsrMountN_t(xdrp, (vmpUsrMountN_t *)cp))
			return(NULL);

		cp += sizeof(vmpUsrMountN_t);
		if (!xdr_opaque(xdrp, cp, hp->vh_count * VSNLEN))
			return(NULL);
		break;
	    case M_RCMOUNT:
		hp->vh_size = sizeof(vmpRCMount_t);
		FULLPKT(hp, sizeof(vmpRCMount_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpRCMount_t));
#endif
		if (!xdr_vmpRCMount_t(xdrp, (vmpRCMount_t *)cp))
			return(NULL);
		break;
	    case M_VLS:
		hp->vh_size = sizeof(vmpVls_t);
		FULLPKT(hp, sizeof(vmpVls_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpVls_t));
#endif
		if (!xdr_vmpVls_t(xdrp, (vmpVls_t *)cp))
			return(NULL);
		break;
	    case M_ATTACH:
	    case M_UNMOUNT:
	    case M_VOLQRY:
	    case M_VSN:
	    case M_REQUEUE:
		hp->vh_size = sizeof(vmpVsn_t);
		FULLPKT(hp, sizeof(vmpVsn_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpVsn_t));
#endif
		if (!xdr_vmpVsn_t(xdrp, (vmpVsn_t *)cp))
			return(NULL);
		break;
	    case M_MOVE: {
		vmpMove_t	*mp;
		vsnseq_t	*vqp;


		/* fill out pkt */
		hp->vh_size = PadSize(sizeof(vmpMove_t)) 
			+ hp->vh_count * sizeof(vsnseq_t);
		FULLPKT(hp, hp->vh_size);
#ifdef PURIFY
		memset(cp, 0, hp->vh_size);
#endif

		/*  move header */
		mp = MSGPTR(hp, vmpMove_t);
		if (!xdr_vmpMove_t(xdrp, mp)) 
			return(NULL);

		/* vsns */
		vqp = (vsnseq_t *) ((char *)mp + PadSize(sizeof(*mp)));
		for (i=0; i<hp->vh_count; i++, vqp++)
			if (!xdr_vsnseq_t(xdrp, vqp))
				return(NULL);
		break;
	    }
	    case M_IDLETIME: {
		vmpNodeTime_t	*ip;

		hp->vh_size = hp->vh_count * sizeof(vmpNodeTime_t);
		FULLPKT(hp, hp->vh_size);
#ifdef PURIFY
		memset(cp, 0, hp->vh_size);
#endif
		ip = (vmpNodeTime_t *) cp;
		for (i=0; i<hp->vh_count; i++, ip++)
			if (!xdr_vmpNodeTime_t(xdrp, ip))
				return(NULL);
		break;
	    }
	    case M_VALLOC:
		hp->vh_size = sizeof(vmpVAlloc_t);
		FULLPKT(hp, sizeof(vmpVAlloc_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpVAlloc_t));
#endif
		if (!xdr_vmpVAlloc_t(xdrp, (vmpVAlloc_t *)cp))
			return(NULL);
		break;
	    case M_LBLQRY:
		hp->vh_size = sizeof(vmpLblQry_t);
		FULLPKT(hp, sizeof(vmpLblQry_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpLblQry_t));
#endif
		if (!xdr_vmpLblQry_t(xdrp, (vmpLblQry_t *)cp))
			return(NULL);
		break;
	    case M_VARY:
		hp->vh_size = sizeof(vmpVary_t);
		FULLPKT(hp, sizeof(vmpVary_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpVary_t));
#endif
		if (!xdr_vmpVary_t(xdrp, (vmpVary_t *)cp))
			return(NULL);
		break;
	    case M_DRIVEINFO: {
		int		errs = 0;
		DevEnt_t	*dp;

		hp->vh_size = hp->vh_count * sizeof(DevEnt_t);
		FULLPKT(hp, hp->vh_size);
#ifdef PURIFY
		memset(cp, 0, hp->vh_size);
#endif
		/* devices */
		dp = (DevEnt_t *) cp;
		for (i=0; i<hp->vh_count; i++, dp++) {
			if (!xdr_DevEnt_t(xdrp, dp))
				return(NULL);
			if (dp->de_errstr)
				errs++;
		}
		if (!errs)
			break;

		/* error strings */
		hp->vh_size += errs * ERRSTRSIZE;
		FULLPKT(hp, hp->vh_size);
		/* calc pointer off of possibly new hp */
		cp += hp->vh_count * sizeof(DevEnt_t);
		if (!xdr_opaque(xdrp, cp, errs * ERRSTRSIZE))
			return(NULL);
		break;
	    }
	    case M_HOSTINFO: {
		HostInfo_t	*hop;

		hp->vh_size = hp->vh_count * sizeof(HostInfo_t);
		FULLPKT(hp, hp->vh_size);
#ifdef PURIFY
		memset(cp, 0, hp->vh_size);
#endif
		hop = (HostInfo_t *) cp;
		for (i=0; i<hp->vh_count; i++, hop++) {
			if (!xdr_HostInfo_t(xdrp, hop))
				return(NULL);
		}
		break;
	    }
	    case M_RESERVE: {
		vmpReserveHdr_t		*vrp;

		/* header */
		FULLPKT(hp, sizeof(vmpReserveHdr_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpReserveHdr_t));
#endif
		vrp = MSGPTR(hp, vmpReserveHdr_t);
		if (!xdr_vmpReserveHdr_t(xdrp, vrp))
			return(NULL);

		if (vrp->re_rcmedcount == 0  &&  vrp->re_vsncount == 0) {
			hp->vh_size = sizeof(vmpReserveHdr_t);
			break;
		}

		/* expand */
		hp->vh_size = PadSize(sizeof(vmpReserveHdr_t))
		      + PadSize(vrp->re_rcmedcount * sizeof(vmpReserveRCMed_t))
		      + vrp->re_vsncount * sizeof(vmpReserveVsn_t);
		FULLPKT(hp, hp->vh_size);
#ifdef PURIFY
		vrp = MSGPTR(hp, vmpReserveHdr_t);
		cp = (char *)vrp + sizeof(*vrp);
		memset(cp, 0, hp->vh_size - sizeof(*vrp));
#endif
		/* recalc pointers off of possibly new hp */
		vrp = MSGPTR(hp, vmpReserveHdr_t);
		cp = (char *)vrp + PadSize(sizeof(*vrp));

		if (vrp->re_rcmedcount) {
			vmpReserveRCMed_t	*rp = (vmpReserveRCMed_t *)cp;
			/* drives */
			for (i=0; i<vrp->re_rcmedcount; i++, rp++) {
				if (!xdr_vmpReserveRCMed_t(xdrp, rp))
					return(NULL);
			}
			rp--;
			cp = (char *)rp + PadSize(sizeof(*rp));
		}
		if (vrp->re_vsncount) {
			vmpReserveVsn_t		*rvp = (vmpReserveVsn_t *)cp;
			/* vsns */
			for (i=0; i<vrp->re_vsncount; i++, rvp++) {
				if (!xdr_vmpReserveVsn_t(xdrp, rvp))
					return(NULL);
			}
		}
		break;
	    }
	    case M_VRECYCLE:
	    case M_DELVSN:
	    case M_VCHMOD:
	    case M_VCHGRP:
	    case M_VCHOWN:
	    case M_AUDIT:
		hp->vh_size = sizeof(vmpVsn_t);
		FULLPKT(hp, sizeof(vmpVsn_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpVsn_t));
#endif
		if (!xdr_vmpVsn_t(xdrp, (vmpVsn_t *)cp))
			return(NULL);
		break;
	    case M_ADDVSN:
		hp->vh_size = sizeof(vmpAdd_t);
		FULLPKT(hp, sizeof(vmpAdd_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpAdd_t));
#endif
		if (!xdr_vmpAdd_t(xdrp, (vmpAdd_t *)cp))
			return(NULL);
		break;
	    case M_CREATENODE:
	    case M_DELETENODE:
	    case M_SCRATCH:
		hp->vh_size = sizeof(vmpNode_t);
		FULLPKT(hp, sizeof(vmpNode_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpNode_t));
#endif
		if (!xdr_vmpNode_t(xdrp, (vmpNode_t *)cp))
			return(NULL);
		break;
	    case M_QUOTQRY:
	    case M_SETQUOTA:
	    case M_QUOTRPLY:
		hp->vh_size = sizeof(vmpQuota_t);
		FULLPKT(hp, sizeof(vmpQuota_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpQuota_t));
#endif
		if (!xdr_vmpQuota_t(xdrp, (vmpQuota_t *)cp))
			return(NULL);
		break;
	    case M_VERROR:
	    case M_SHUTDOWN:
		hp->vh_size = sizeof(vmpError_t);
		FULLPKT(hp, sizeof(vmpError_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpError_t));
#endif
		if (!xdr_vmpError_t(xdrp, (vmpError_t *)cp)) 
			return(NULL);
		break;
	    case M_DEBUG:
		hp->vh_size = sizeof(vmpDebug_t);
		FULLPKT(hp, sizeof(vmpDebug_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpDebug_t));
#endif
		if (!xdr_vmpDebug_t(xdrp, (vmpDebug_t *)cp)) 
			return(NULL);
		break;
	    case M_VATTR:
		hp->vh_size = sizeof(vmpAttr_t);
		FULLPKT(hp, sizeof(vmpAttr_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpAttr_t));
#endif
		if (!xdr_vmpAttr_t(xdrp, (vmpAttr_t *)cp))
			return(NULL);
		break;
	    case M_RCMSG:
		hp->vh_size = sizeof(vmpRcMsg_t);
		FULLPKT(hp, sizeof(vmpRcMsg_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpRcMsg_t));
#endif
		if (!xdr_vmpRcMsg_t(xdrp, (vmpRcMsg_t *)cp))
			return(NULL);
		break;
	    case M_LBLOOK:
		hp->vh_size = sizeof(vmpLblLook_t);
		FULLPKT(hp, sizeof(vmpLblLook_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpLblLook_t));
#endif
		if (!xdr_vmpLblLook_t(xdrp, (vmpLblLook_t *)cp)) 
			return(NULL);
		break;
	    case M_NODED:
	    case M_VOLCD:
	    case M_RC:
	    case M_RCREADY:
	    case M_TIMERD:
		hp->vh_size = sizeof(vmpHello_t);
		FULLPKT(hp, sizeof(vmpHello_t));
#ifdef PURIFY
		memset(cp, 0, sizeof(vmpHello_t));
#endif
		if (!xdr_vmpHello_t(xdrp, (vmpHello_t *)cp))
			return(NULL);
		break;
	    case M_DEVQRY: {
		DevEnt_t	*dp;

		hp->vh_size = hp->vh_count * sizeof(DevEnt_t);
		FULLPKT(hp, hp->vh_size);
#ifdef PURIFY
		memset(cp, 0, hp->vh_size);
#endif
		/* devices */
		dp = (DevEnt_t *) cp;
		for (i=0; i<hp->vh_count; i++, dp++) {
			if (!xdr_DevEnt_t(xdrp, dp))
				return(NULL);
		}
		break;
	    }
	    default:
		VMError = E_COMM;
		nLogMsg(EL_ERROR1, id, "unexpected type %s", vHdrStr(hp));
		return(NULL);
	}
	return(hp);
}

/*
 *  XDRFlush() - when connection goes away & socket is closed,
 *	make sure nothing remains in the buffer that would get
 *	flushed to the next client on this socket
 */
  void
XDRFlush(XDR *xdrp)
{
	(void)xdrrec_endofrecord(xdrp, 1);
}
