
/*
 *  rp_2.c: Volume Manager Request Processing of requests involving
 *	action by RCs on volumes: mounts, dismounts, moves, etc.
 *	These requests are put on the VolReqQ which maintains fair 
 *	ordering of execution. 
 *
 *	Once a request is enqueued, it is examined by routines in rpq.c, 
 *	which send it to the appropriate RC(s) when resources are available. 
 *	The RC(s) respond with volstat packets, which are processed here in 
 *	conjunction with the volreq. (Async volstats are also processed here.)
 *	
 *	Note: drive counts are tracked on a per-RC basis (rcp->rc_nvaryondev).
 *	Once a drive is allocated, it is not deallocated until a dismount 
 *	completes (or mount fails). This is because there is supposedly no 
 *	queuing in the RCs, e.g. when the ACS is completely mounted, if the
 *	voldaemon sent down a mount cmd before a dismount completed, the 
 *	mount would fail because the drive is not yet free. Drive counts
 *	are also tracked on a per-reservation basis (ridp->rid_RCMed[]).
 *	See rp_rid.c for a discussion of reservations.
 *
 *	rp_1.c	main, packet proc switch to routines, & housekeeping
 *	rp_2.c	packet proc: mounting & moving volumes
 *	rp_3.c	packet proc: reporting/changing dbase info
 *	rp_4.c	packet proc: varying drives
 *	rp_5.c	packet proc: reserving drives & volumes
 */

#include "rp_space.h"

extern int	shuttingdown;
extern int	Timeout;
extern time_t	Now;
extern int	UnmountWaitTime;


/*****************************************************************
   Volume Manager State Diagram - take with a grain of salt &
				  double-check the code.

What to do when packets are received:

 *** packets from users ***

 M_USRMOUNT: MMount():
	dbMountQuery()
	look up RC info, check RC
	set up request for client (reservation & volreq)
	fill in volreq w/ mount info
	put on VolReqQ

 M_DETACH: MDetach():
	for each reservation, volreq on client q
		if it's idle, delete it
		else set notification flags & dequeue from client
 
M_UNMOUNT: MUnmount():
	find volreq (with same client if it exists) with 'best' mount state
	connect new client if necc / msg to old
	copy new flags
	Unmount():
	  if possible
		hand off mount to waiting request
		return
	  switch state
	    case MOUNTED:
		  start unmount
	    case RCWAIT:	
		  set flag to unmount when mount completes
		  if RC type allows
			start unmount
		  put VolReq on orphan q
	    case CREATENODE:
		  set flag to unmount when mount completes
            case MOUNTWAIT:
            case DRIVEWAIT:
		  volreq status idle
		  dequeue VolReq from VolReqQ
		  Send M_ACK/Ack to client
		  if implicit reservation
			free volreq/reservation
	    case DELETENODE:
	    case RCWAIT2:
		  already unmounting - error msg to client

 M_MOVE: MMove():
	look up destination RC info
	set up request for client (reservation & volreq)
	fill in volreq w/ move info
	put on VolReqQ

 *** packets from servers (RCs, nodeservers) ***

 M_VOLSTAT: MVolStat(): (==msg from RC: 'completed vsn op'):
	find volreq
	look up RC
	if no volreq - async "volume arrived here" announcement - AsyncVolStat()
		update vsn location in db & queued requests
	else if volreq.state == RCWAIT - mount completed - RCMounted()
		RC chose a drive: set drive bit in volreq 
		if got an unmount req while mount in progress
			cleanup_open()
			return
		status = NODEWAIT & send to node setup server (checks label too)
	else if state == RCWAIT2	- unmount completed - RCUnmounted()
		update dbase
		Ack client
		if unmount was requested because of vary off of drive
			vary off drive now
		if drive &/or vsn reservation was implicit
			free reservation &/or volreq
	else if state == MOVING	- source or destination completed - RCMoved()
		if move completed (dest done or src can complete / ACS eject)
			update vsn location in db & queued requests

 M_CREATENODE:
 M_DELETENODE:
	if 'pass thru' to volnd or RC do it
	else
		find volreq
		create: tell client mount successful
		delete: send unmount to RC

 M_VERROR:
	if state == RCWAIT
		Dequeue request
		Send M_VERROR/ Ack client
	if state == DELETENODE
		unmount

cleanup_open():
	remove /dev/volume node
	state = RCWAIT2
	Send M_UNMOUNT to RC


M_REQUEUE: # not done
	find (uid,vsn)
	if state != MOUNTED
		Send M_VERROR/Ack to client
		return
	flag |= VF_REQREQ
	update drives to try
	Send M_UNMOUNT to RC

M_VERROR: MError():


*****************************************************************/

  VolReq_t *
SetupRequest(char *vsnp, Client_t *cp, int sync, Seq_t seq, char *rid)
{
	char		*id = "SetupRequest";
  	Client_t	*cp2;
	VolReq_t	*vrp = NULL;
	Rid_t		*ridp;
	int		implicitrid = 0;

	/*
	 *  any rid or vsn struct created will be tied to a
	 *	client unless the request is asynchronous
	 */
	if (sync)
		cp2 = cp;
	else
		cp2 = NULL;

	/*
	 *  get or set up reservation
	 */
	if (rid != NULL  &&  rid[0]) {

		/*
		 *  a reservation id is given - check it & get reservation
		 */
		if (BadRidName(rid)) {
			return(NULL);
		}
		ridp = GetRid(rid, cp->c_uid);
		if (ridp == NULL)
			return(NULL);
	
		/*
		 *  see if vsn is in rid in waiting-for-use form
		 */
		vrp = GetRidVsn(ridp, vsnp, VS_IDLE, cp);

	} else {
		
		/*
		 *  make an implicit reservation (no id)
		 */
		ridp = MakeRid(NULL, cp->c_uid, cp2, seq);
		if (!ridp)
			return(NULL);
		implicitrid++;
	}

	/*
	 *  add request to reservation if necc
	 *	- if not, may need to change client of reservation
	 */
	if (!vrp) {

		/*
		 *  vsn not in rid in IDLE state - 
		 *	add an instance w/ implicit flag set
		 */
		vrp = AddVsnToRid(vsnp, ridp, cp2, cp->c_uid, VF_IMPLICITVSN);
		if (!vrp) {
			if (implicitrid)
				DeleteRid(ridp, cp2);
			return(NULL);
		}
	} else {

		/*
		 *  give request to new client if sync
		 */
		vrp->vr_clientp = cp2;
		if (cp2) {
			vrp->vr_uid = cp2->c_uid;
			ENQUEUE(vrp, &vrp->vr_clientp->c_cvrq,vr_cvrq,VolReq_t)
		} else
			vrp->vr_uid = -1;
	}
	vrp->vr_clhostp = IPCHost(cp);
	vrp->vr_clhost = vrp->vr_clhostp->h_id;
	return(vrp);
}

  
/*
 *  MMount() - process M_USRMOUNT packets received by the RP
 *
 */

MMount(vmpHdr_t *hp, Client_t *cp)
{
	char		*id = "MMount";
	char		*ridname;
	vmpUsrMount_t	*mp;
	VInfo_t		*vip;
	VolReq_t	*vrp, *mvvrp;
	RPRC_t		*rcp;			
	char		*vsnp;
	int		sync, medindex, med2;
	DevMap_t	devtotry;

	mp = MSGPTR(hp, vmpUsrMount_t);
	vsnp = (char *) &mp->vu_vsn;

	if (mp->vu_rid[0])
		nLogMsg(EL_DEBUG1, id, "request %s, rid %.*s", 
			vsnstr(vsnp), RIDLEN, mp->vu_rid);
	else
		nLogMsg(EL_DEBUG1, id, "request %s, rid implicit", 
			vsnstr(vsnp));

	/*
	 *  simple pkt checks of options & reservation id
	 */

	if (CheckVO_Flags(mp->vu_flags)) {
		Err2Client(hp->vh_seq, cp);
		return(0);
	}

	/*
	 *  if exclusive req'd, make sure all ok
	 *	note: in AddVsnToRid, an err is generated if any
	 *	      other req tries to follow on a current exclusive
	 */
	if (mp->vu_flags & VO_EXCLUSIVE) {
		/*
		 *  (normally scratching) - if a req exists, vsn is 
		 *	not quiescent so there's a danger scratching 
		 *	will wipe out something
		 */
		if (FindVolReq(vsnp)) {
			VMError = E_UNAVAIL;
			nLogMsg(EL_CODE|EL_STATUS2, id, 
				"Can't get exclusive access: other request");
			Err2Client(hp->vh_seq, cp);
			return(0);
		}
	} 

	/*
	 *  check dbase (uid permission) / get fresh copy of record & RC
	 * 	note: vip is static in vdmlib
	 */
	vip = DBMountQuery(vsnp, cp->c_uid, mp->vu_flags & VO_WRITE, &rcp);
	if (vip == NULL) {
		Err2Client(hp->vh_seq, cp);
		return(0);
	}
	nLogMsg(EL_DEBUG3, id, "mount ok, flags %x", mp->vu_flags);

	/*
	 *  see what RC really will be (modulo any moves outstanding)
	 */
	mvvrp = NULL;
	for (vrp=FindVolReq(vsnp); vrp; vrp=VRNEXT(vrp->vr_cvrq)) {
		if (vrp->vr_cmd == M_MOVE)
			mvvrp = vrp;
	}
	if (mvvrp != NULL)
		rcp = mvvrp->vr_dstrcp;

	/*
	 *  see if any drives configged
	 */
	medindex = RCMedium(rcp, vVolTypeMedium(vip->vi_voltype));
	if (medindex == -1) {
		VMError = E_INTERNAL;
		nLogMsg(EL_CODE|EL_ERROR1, id, "can't get RCMedium");
		Err2Client(hp->vh_seq, cp);
		return(0);
	}
	if (rcp->rc_devs[medindex].d_nconfigdev == 0) {
		VMError = E_ADMIN;
		nLogMsg(EL_CODE|EL_ERROR2|EL_CONSOLE0, id,
				"Can't mount %.*s: RC %s %s has no %s drives",
				VSNLEN, vsnp, rcp->rc_id,
				(mvvrp ? "(move destination)" : ""),
				vMediumName(rcp->rc_devs[medindex].d_medium));
		Err2Client(hp->vh_seq, cp);
		return(0);
	}
	devtotry = rcp->rc_devs[medindex].d_devmap;
	if (mp->vu_flags & VO_MYHOST) {
		Host_t		*hop = HostIndex(cp->c_host);
		DevMap_t	*dmp;

		nLogMsg(EL_STATUS1, id, "MYHOST request %s ndrives %d", 
						hop->h_name, hop->h_ndrives);
		/*
		 *  quick check..
		 */
		if (!hop->h_ndrives) {
			VMError = E_UNAVAIL;
			nLogMsg(EL_CODE|EL_ERROR2, id, 
					"no drives on client's host (%s)",
					hop->h_name);
			Err2Client(hp->vh_seq, cp);
			return(0);
		}
		/*
		 *  exhaustive check
		 */
		dmp = vDMand(&devtotry, &hop->h_drives);
		if (vDMffs(dmp) == -1) {
			VMError = E_UNAVAIL;
			nLogMsg(EL_CODE|EL_ERROR2, id, 
				"no %s drives on client's host (%s)",
				vMediumName(rcp->rc_devs[medindex].d_medium),
				hop->h_name);
			Err2Client(hp->vh_seq, cp);
			return(0);
		}
		devtotry = *dmp;
	}

	/*
	 *  make sure RC is up
	 */
	if (rcp->rc_srv.s_sd == -1) {
		VMError = E_UNAVAIL;
		nLogMsg(EL_CONSOLE0|EL_CODE|EL_ERROR2, id, 
					"Can't mount %s: RC %s %s down",
					vsnstr(vsnp), rcp->rc_id,
					(mvvrp ? "(move destination)" : ""));
		Err2Client(hp->vh_seq, cp);
		return(0);
	}

	/*
	 *  get VolReq (hooks it up w/ rid if any and client struct
	 *	if sync)
	 */
	sync = mp->vu_flags & VO_WAIT;
	vrp = SetupRequest(vsnp, cp, sync, hp->vh_seq, mp->vu_rid);
	if (vrp == NULL) {
		Err2Client(hp->vh_seq, cp);
		return(0);
	}

	/*
	 *  fill in other request data in VolReq per this request
	 */

	vrp->vr_cmd = M_USRMOUNT;
	vrp->vr_info = *vip;			/* copy vinfo / MountQuery() */
	vrp->vr_retrycnt = 0;
	vrp->vr_clseq = hp->vh_seq;		/* client seq no from header */
	DM_ZERO(vrp->vr_devinuse);		/* no device in use yet */
	vrp->vr_rcp = rcp;			/* save pointer to RC struct */
	vrp->vr_devtotry = devtotry;
	vrp->vr_voflag = mp->vu_flags;		/* volop flags */
	copytag(vrp->vr_tag, mp->vu_tag);

	vrp->vr_state = VS_MOUNTWAIT;
	vrp->vr_time = Now;

	/*
	 *  leave request in queue for resources
	 */
	
	ENQUEUE(vrp, &VolReqQ, vr_vrq, VolReq_t)
	if (vrp->vr_ridp->rid_id[0]) 
		ridname = vrp->vr_ridp->rid_id; 
	else
		ridname = "nrsv";
	nLogMsg(EL_STATUS1, 0, 
			"MOUNTQ %.*s_%d %.*s %s %s pid %d uid %d %s",
			VSNLEN, vsnp, vrp->vr_rpseq, 
			RIDLEN, ridname,
			(vrp->vr_tag ? vrp->vr_tag : ""),
			(sync ? "sync" : "async"),
			cp->c_pid, cp->c_uid,
			(vrp->vr_voflag & VO_BYPASS ? "nlbl" : "") );

	/*
	 *  Ack so a client staging an async mount will know that 
	 *	it's basically accepted - without this part of the
	 *	protocol, clients would have to time out waiting for 
	 *	an error, which would be unreliable.
	 */

	AckClient(hp->vh_seq, cp);

	/*
	 *  See if a delayed unmount might be available
	 */

	if (UnmountWaitTime > 0  &&  Timeout  &&  !vrp->vr_ridp->rid_id[0]) {
		/*
		 *  see if a requested unmount is lying around
		 */
		if (UseWaitingUnmount(vrp))
			return(0);	/* no need to evaluate queue */
	}
	return(1);	/* evaluate queue */
}

/*
 *  MMountPll() - process M_USRMOUNT_PLL packets received by the RP 
 *
 */
static int	*RCCounts = NULL;	/* for counting resrv req'ts */
static int	VinfoCount = 0;
static VInfo_t	*Vinfo = NULL;
static RPRC_t	**RCs = NULL;

MMountPll(vmpHdr_t *hp, Client_t *cp)
{
	char		*id = "MMountPll";
	vmpUsrMountN_t	*mp;
	char		*vsnp;
	VInfo_t		*vip;
	RPRC_t		*rcp;
	Rid_t		*ridp;
	VolReq_t	*vrp, *mvvrp;
	int		i, medium, medindex, med2;

	mp = MSGPTR(hp, vmpUsrMountN_t);

	/*
	 *  input checking
	 */
	if (hp->vh_count < 1) {
		VMError = E_USER;
		Err2Client(hp->vh_seq, cp);
		return(0);
	}
	if (mp->vx_flags & (VO_ASYNC|VO_SIGNAL|VO_EXCLUSIVE)) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR1, id, 
			"coding error: ASYNC, SIGNAL, EXCLUSIVE not permitted");
		Err2Client(hp->vh_seq, cp);
		return(0);
	}
/*
TODO - specfific check 
*/
	if (CheckVO_Flags(mp->vx_flags)) {
		Err2Client(hp->vh_seq, cp);
		return(0);
	}

	/*
	 *  create/extend scratch arrays if necc
	 */
	if (Vinfo == NULL) {
		VinfoCount = hp->vh_count;
		Vinfo = (VInfo_t *) nAlloc(VinfoCount * sizeof(VInfo_t));
		RCs = (RPRC_t **) nAlloc(VinfoCount * sizeof(RPRC_t *));
	} else if (hp->vh_count > VinfoCount) {
		nFree(Vinfo);
		nFree(RCs);
		VinfoCount = hp->vh_count;
		Vinfo = (VInfo_t *) nAlloc(VinfoCount * sizeof(VInfo_t));
		RCs = (RPRC_t **) nAlloc(VinfoCount * sizeof(RPRC_t *));
	}
	if (RCCounts == NULL)
		RCCounts = (int *)
			   nAlloc(RCTotal * RC_MEDIA_TYPES * sizeof(int));

	/*
	 *  initialize counts of RC/medium req'ts
	 */
	for (i=0; i<RCTotal * RC_MEDIA_TYPES; i++)
		RCCounts[i] = 0;

	/*
	 *  check vsn access & RC drives, saving vsn records & RC pointers
	 *	in scratch arrays
	 */
	vsnp = (char *) mp + sizeof(*mp);
	for (i=0; i<hp->vh_count; i++, vsnp+=VSNLEN) {
		vip = DBMountQuery(vsnp, cp->c_uid,
				mp->vx_flags & VO_WRITE,
				&rcp);
		if (vip == NULL) {
			Err2Client(hp->vh_seq, cp);
			nLogMsg(EL_STATUS2, id, "problem w/ vsn %d", i);
			return(0);
		}
		nLogMsg(EL_DEBUG3, id, "%.*s access ok", VSNLEN, vsnp);

		/*
		 *  see what RC really will be (modulo any moves outstanding)
		 */
		mvvrp = NULL;
		for (vrp=FindVolReq(vsnp); vrp; vrp=VRNEXT(vrp->vr_cvrq)) {
			if (vrp->vr_cmd == M_MOVE)
				mvvrp = vrp;
		}
		if (mvvrp != NULL)
			rcp = mvvrp->vr_dstrcp;

		/*
		 *  see if any drives configged
		 */
		medindex = RCMedium(rcp, vVolTypeMedium(vip->vi_voltype));
		if (medindex == -1) {
			VMError = E_ADMIN;
			nLogMsg(EL_CODE|EL_ERROR2, id, 
				"%.*s %s in %s but RC not configd to hold %s",
				VSNLEN, &vip->vi_vsn, vMediumName(medium),
				rcp->rc_id, vMediumName(medium));
			Err2Client(hp->vh_seq, cp);
			return(0);
		}
		if (rcp->rc_devs[medindex].d_nconfigdev == 0) {
			VMError = E_UNAVAIL;
			nLogMsg(EL_CODE|EL_ERROR2|EL_CONSOLE0, id,
				"Can't mount %.*s: RC %s %s has no %s drives",
				VSNLEN, vsnp, rcp->rc_id,
				(mvvrp ? "(move destination)" : ""),
				vMediumName(rcp->rc_devs[medindex].d_medium));
			Err2Client(hp->vh_seq, cp);
			return(0);
		}

		/*
		 *  make sure RC is up
		 */
		if (rcp->rc_srv.s_sd == -1) {
			VMError = E_UNAVAIL;
			nLogMsg(EL_CONSOLE0|EL_CODE|EL_ERROR2, id, 
					"Can't mount %.*s: RC %s %s down",
					VSNLEN, vsnp, rcp->rc_id,
					(mvvrp ? "(move destination)" : ""));
			Err2Client(hp->vh_seq, cp);
			return(0);
		}

		/*
		 *  it could work - save info
		 */
		Vinfo[i] = *vip;
		RCs[i] = rcp;
		RCCounts[rcp->rc_index * RC_MEDIA_TYPES + medindex]++;
	}

	/*
	 *  make implicit 'parallel' reservation tied to client -
	 *	remember to delete if err below
	 */
	ridp = MakeRid(NULL, cp->c_uid, cp, hp->vh_seq);
	ridp->rid_parallel = 1;

	/*
	 *  check totals req'd and add drives to rid
	 */
	for (i=0; i<RCTotal * RC_MEDIA_TYPES; i++) {
		if (RCCounts[i]) {
			int		rcindex = i / RC_MEDIA_TYPES;
			int		medindex = i % RC_MEDIA_TYPES;
			RidRCMed_t	*ridrcmedp;

			rcp = &RCArray[rcindex];

			nLogMsg(EL_STATUS2, id, 
				"Reserve RC %s/%s: %d drive%s uid %d pid %d", 
				rcp->rc_id, 
				vMediumName(rcp->rc_devs[medindex].d_medium),
				RCCounts[i],
				(RCCounts[i] > 1 ? "s" : ""),
				cp->c_uid, cp->c_pid);
			if (RCCounts[i] > rcp->rc_devs[medindex].d_nconfigdev) {
				VMError = E_UNAVAIL;
				nLogMsg(EL_ERROR1|EL_CODE, id,
					"need %d drives in %s, %d configured",
					RCCounts[i], rcp->rc_id,
					rcp->rc_devs[medindex].d_nconfigdev);
				Err2Client(hp->vh_seq, cp);
				DeleteRid(ridp, cp);
				return(0);
			}
			ridrcmedp = &ridp->rid_RCMed[rcindex * RC_MEDIA_TYPES 
								+ medindex];

			if (RCCounts[i] > rcp->rc_devs[medindex].d_nconfigdev
					+ ridrcmedp->rr_hardalloc) {
				VMError = E_UNAVAIL;
				nLogMsg(EL_ERROR1|EL_CODE, id,
					"need %d drives in %s, %d not hardresv",
					RCCounts[i], rcp->rc_id,
					rcp->rc_devs[medindex].d_nconfigdev -
					ridrcmedp->rr_hardalloc);
				Err2Client(hp->vh_seq, cp);
				DeleteRid(ridp, cp);
				return(0);
			}
			ridrcmedp->rr_softalloc += RCCounts[i];
		}
	}

	/*
	 *  add vsns to rid
	 *	TODO - allow it to pick up waiting unmounts,
	 *	i.e. handle reservations
	 */
	vsnp = (char *) mp + sizeof(*mp);
	for (i=0; i<hp->vh_count; i++, vsnp+=VSNLEN) {
		rcp = RCs[i];
		vip = &Vinfo[i];
		medindex = RCMedium(rcp, vVolTypeMedium(vip->vi_voltype));

		vrp = AddVsnToRid(vsnp, ridp, cp, cp->c_uid, VF_IMPLICITVSN);
		if (vrp == NULL) {
			Err2Client(hp->vh_seq, cp);
			return(0);
		}
		vrp->vr_clhostp = IPCHost(cp);
		vrp->vr_clhost = vrp->vr_clhostp->h_id;

		/*
		 *  fill in other request data in VolReq per this request
		 */
		vrp->vr_cmd = M_USRMOUNT;
		vrp->vr_info = Vinfo[i];	/* copy vinfo / MountQuery() */
		vrp->vr_retrycnt = 0;
		vrp->vr_clseq = hp->vh_seq;	/* client seq no from header */
		DM_ZERO(vrp->vr_devinuse);	/* no device in use yet */
		vrp->vr_rcp = rcp;		/* save pointer to RC struct */
		vrp->vr_devtotry = rcp->rc_devs[medindex].d_devmap;
		vrp->vr_voflag = mp->vx_flags;	/* volop flags */
		copytag(vrp->vr_tag, mp->vx_tag);

		vrp->vr_state = VS_MOUNTWAIT;
		vrp->vr_time = Now;

		/*
		 *  leave request in queue for resources
		 */
	
		ENQUEUE(vrp, &VolReqQ, vr_vrq, VolReq_t)
		nLogMsg(EL_STATUS1, 0, 
			"MOUNTQ parallel %.*s_%d  %s  pid %d uid %d %s",
			VSNLEN, vsnp, vrp->vr_rpseq, 
			(vrp->vr_tag ? vrp->vr_tag : ""),
			cp->c_pid, cp->c_uid,
			(vrp->vr_voflag & VO_BYPASS ? "nlbl" : "") );

	}
	return(1);	/* evaluate queue */
}


/*
 *  MMountSel() - process M_USRMOUNT_SEL packets received by the RP 
 *
 */
MMountSel(vmpHdr_t *hp, Client_t *cp)
{
	char		*id = "MMountSel";
	vmpUsrMountN_t	*mp;
	char		*vsnp;
	VInfo_t		*vip;
	RPRC_t		*rcp;
	Rid_t		*ridp;
	VolReq_t	*vrp, *mvvrp;
	int		i, nok, medium, medindex, med2;

	mp = MSGPTR(hp, vmpUsrMountN_t);

	/*
	 *  input checking
	 */
	if (hp->vh_count < 1) {
		VMError = E_USER;
		Err2Client(hp->vh_seq, cp);
		return(0);
	}
	if (mp->vx_flags & (VO_ASYNC|VO_SIGNAL|VO_EXCLUSIVE)) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR1, id, 
			"coding error: ASYNC, SIGNAL, EXCLUSIVE not permitted");
		Err2Client(hp->vh_seq, cp);
		return(0);
	}
/*
TODO - specfific check 
*/
	if (CheckVO_Flags(mp->vx_flags)) {
		Err2Client(hp->vh_seq, cp);
		return(0);
	}

	/*
	 *  create/extend scratch arrays if necc
	 */
	if (Vinfo == NULL) {
		VinfoCount = hp->vh_count;
		Vinfo = (VInfo_t *) nAlloc(VinfoCount * sizeof(VInfo_t));
		RCs = (RPRC_t **) nAlloc(VinfoCount * sizeof(RPRC_t *));
	} else if (hp->vh_count > VinfoCount) {
		nFree(Vinfo);
		nFree(RCs);
		VinfoCount = hp->vh_count;
		Vinfo = (VInfo_t *) nAlloc(VinfoCount * sizeof(VInfo_t));
		RCs = (RPRC_t **) nAlloc(VinfoCount * sizeof(RPRC_t *));
	}
	if (RCCounts == NULL)
		RCCounts = (int *)
			   nAlloc(RCTotal * RC_MEDIA_TYPES * sizeof(int));

	/*
	 *  check vsn access & RC drives, saving vsn records & RC pointers
	 *	in scratch arrays
	 */
	vsnp = (char *) mp + sizeof(*mp);
	nok = 0;
	for (i=0; i<hp->vh_count; i++, vsnp+=VSNLEN) {
nLogMsg(EL_STATUS1, id, "vsn %.6s", vsnp);
		vip = DBMountQuery(vsnp, cp->c_uid, 
				mp->vx_flags & VO_WRITE,
				&rcp);
		if (vip == NULL) {
			nLogMsg(EL_STATUS2, id, "problem w/ vsn %d", i);
			continue;
		}
		nLogMsg(EL_DEBUG3, id, "%.*s access ok", VSNLEN, vsnp);

		/*
		 *  see what RC really will be (modulo any moves outstanding)
		 */
		mvvrp = NULL;
		for (vrp=FindVolReq(vsnp); vrp; vrp=VRNEXT(vrp->vr_cvrq)) {
			if (vrp->vr_cmd == M_MOVE)
				mvvrp = vrp;
		}
		if (mvvrp != NULL)
			rcp = mvvrp->vr_dstrcp;

		/*
		 *  see if any drives configged
		 */
		medindex = RCMedium(rcp, vVolTypeMedium(vip->vi_voltype));
		if (medindex == -1) {
			nLogMsg(EL_ERROR2, id, 
				"%.*s %s in %s but RC not configd to hold %s",
				VSNLEN, &vip->vi_vsn, vMediumName(medium),
				rcp->rc_id, vMediumName(medium));
			continue;
		}
		if (rcp->rc_devs[medindex].d_nconfigdev == 0) {
			nLogMsg(EL_ERROR2, id,
				"Can't mount %.*s: RC %s %s has no %s drives",
				VSNLEN, vsnp, rcp->rc_id,
				(mvvrp ? "(move destination)" : ""),
				vMediumName(rcp->rc_devs[medindex].d_medium));
			continue;
		}

		/*
		 *  make sure RC is up
		 */
		if (rcp->rc_srv.s_sd == -1) {
			nLogMsg(EL_CONSOLE0|EL_ERROR2, id, 
					"Can't mount %.*s: RC %s %s down",
					VSNLEN, vsnp, rcp->rc_id,
					(mvvrp ? "(move destination)" : ""));
			continue;
		}

		/*
		 *  it could work - save info
		 */
		Vinfo[nok] = *vip;
		RCs[nok] = rcp;
		nok++;
	}
	if (!nok) {
		VMError = E_UNAVAIL;
		nLogMsg(EL_ERROR1|EL_CODE, id, "problems w/ all vsns");
		Err2Client(hp->vh_seq, cp);
		return(0);
	}

	/*
	 *  make implicit 'select' reservation tied to client -
	 *	remember to delete if err below
	 */
	ridp = MakeRid(NULL, cp->c_uid, cp, hp->vh_seq);
	ridp->rid_parallel = 2;


	/*
	 *  add vsns to rid
	 *	TODO - allow it to pick up waiting unmounts,
	 *	i.e. handle reservations
	 */
	for (i=0; i<nok; i++) {
		rcp = RCs[i];
		vip = &Vinfo[i];
		vsnp = (char *)&vip->vi_vsn;
nLogMsg(EL_STATUS1, id, "round 2 vsn %.6s", vsnp);
		medindex = RCMedium(rcp, vVolTypeMedium(vip->vi_voltype));

		vrp = AddVsnToRid(vsnp, ridp, cp, cp->c_uid, VF_IMPLICITVSN);
		if (vrp == NULL) {
			Err2Client(hp->vh_seq, cp);
			return(0);
		}
		vrp->vr_clhostp = IPCHost(cp);
		vrp->vr_clhost = vrp->vr_clhostp->h_id;

		/*
		 *  fill in other request data in VolReq per this request
		 */
		vrp->vr_cmd = M_USRMOUNT;
		vrp->vr_info = Vinfo[i];	/* copy vinfo / MountQuery() */
		vrp->vr_retrycnt = 0;
		vrp->vr_clseq = hp->vh_seq;	/* client seq no from header */
nLogMsg(EL_STATUS1, id, "seq %d", hp->vh_seq);
		DM_ZERO(vrp->vr_devinuse);	/* no device in use yet */
		vrp->vr_rcp = rcp;		/* save pointer to RC struct */
		vrp->vr_devtotry = rcp->rc_devs[medindex].d_devmap;
		vrp->vr_voflag = mp->vx_flags;	/* volop flags */
		copytag(vrp->vr_tag, mp->vx_tag);

		vrp->vr_state = VS_MOUNTWAIT;
		vrp->vr_time = Now;

		/*
		 *  leave request in queue for resources
		 */
	
		ENQUEUE(vrp, &VolReqQ, vr_vrq, VolReq_t)
		nLogMsg(EL_STATUS1, 0, 
			"MOUNTQ select %.*s_%d  %s  pid %d uid %d %s",
			VSNLEN, vsnp, vrp->vr_rpseq, 
			(vrp->vr_tag ? vrp->vr_tag : ""),
			cp->c_pid, cp->c_uid,
			(vrp->vr_voflag & VO_BYPASS ? "nlbl" : "") );

	}
	return(1);	/* evaluate queue */
}

/*
 *  UseWaitingUnmount() - give a waiting unmount or unmount that
 *	is in the node delete phase to mount
 */
UseWaitingUnmount(VolReq_t *vrp)
{
	char		*id = "UseWaitingUnmount";
	VolReq_t	*vrp2, *FindWaitingUnmount();

	vrp2 = FindWaitingUnmount(vrp);
	if (vrp2 == NULL)
		return(0);

	if (HandOffMountDetailsFails(vrp2, vrp))
		return(0);

	/*
	 *  adjust rids
	 */
	vrp->vr_ridp->rid_active++;
	vrp2->vr_ridp->rid_active--;

	nLogMsg(EL_DEBUG2, id, "used %s %.*s_%d",
		(vrp->vr_state == VS_MOUNTED ? "unmountwait" : "nodedelete"),
		VSNLEN, &vrp2->vr_vsn, vrp2->vr_rpseq);

	if (vrp->vr_state == VS_MOUNTED) {
		/*
		 *  1 fewer unmountwait
		 */
		Timeout--;

		/*
		 *  inform recipient if necc
		 */
		if (vrp->vr_voflag & VO_WAIT)
			Volstat2Client(vrp);
		else
			nLogMsg(EL_DEBUG3, id, "no client waiting");
	}

	/*
	 *  inform dismount requestor
	 */

	vrp2->vr_state = VS_UNMOUNTED;
	if (vrp2->vr_voflag & VO_WAIT)
		Volstat2Client(vrp2);
	else
		nLogMsg(EL_DEBUG3, id, "unmounter was async");

	/*
	 *  delete or inactivate original request
	 */

	if (vrp2->vr_flag & VF_IMPLICITVSN) {
		FreeVolReq(vrp2, id);
	} else {
		vrp2->vr_state = VS_IDLE;
		vrp2->vr_time = Now;
		vrp2->vr_qtime = 0;
		vrp2->vr_rctime1 = 0;
		vrp2->vr_usrtime = 0;
		vrp2->vr_rctime2 = 0;
		DEQUEUE(vrp2, &VolReqQ, vr_vrq, VolReq_t);
	}
	return(1);
}


/*
 *  MDetach() - make all requests of this client VO_ASYNC.
 *	E.g. the vol command waits for status and then once
 *	the mount has occurred, wants to exit without the vol 
 *	being automatically unmounted.
 */
 void
MDetach(vmpHdr_t *hp, Client_t *cp)
{
	char		*id = "MDetach";
	Rid_t		*ridp, *nextridp;
	VolReq_t	*vrp, *nextvrp;
	int		res = 0, vol = 0;

	nLogMsg(EL_DEBUG3, id, "pid %d", cp->c_pid);
	/*
	 *  detach from rids
	 */
	res = 0;
	for (ridp=(Rid_t *)cp->c_ridq.q_head; ridp; ridp=nextridp) {
		nextridp = (Rid_t *)ridp->rid_clientq.q_next;

		if (ridp->rid_parallel == 1) {
			VMError = E_USER;
			nLogMsg(EL_ERROR1|EL_CODE, id, 
				"pid %d detaching parallel rid - not allowed");
			Err2Client(hp->vh_seq, cp);
			MClientGoodbye(cp);
			return;
		}
		if (ridp->rid_cp != cp) {
			nLogMsg(EL_STATUS1, id, "rid_cp != cp %lld %lld",
							ridp->rid_cp, cp);
			kill(getpid(),5);
		}
		DEQUEUE(ridp, &cp->c_ridq, rid_clientq, Rid_t);
		ridp->rid_cp = NULL;
		res++;
	}

	/*
	 *  detach from vsns
	 */
	vol = 0;
	for (vrp=VRHEAD(cp->c_cvrq); vrp; vrp=nextvrp) {
		nextvrp=VRNEXT(vrp->vr_cvrq);	/* since vrp might be DQ'd */

		vol++;
		if (vrp->vr_state == VS_IDLE  &&  !vrp->vr_ridp->rid_id[0]) {
			/*
			 *  implicit reservation, presumably
			 *	unmounted.
			 */
			nLogMsg(EL_DEBUG1, id, "%.*s_%d idle/implicit",
					VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
			FreeVolReq(vrp, id);
			continue;
		}

		/*
		 * got (vsn, cp)
		 *	copy new notification flags & client info
		 */

		nLogMsg(EL_DEBUG3, id, "%.*s_%d normal detach",
					VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
		vrp->vr_voflag &= ~VO_RESPONSE;
		vrp->vr_voflag |= VO_ASYNC;
		vrp->vr_clientp = NULL;
		DeQVolCVR(vrp, &cp->c_cvrq);
	}
	if (vol + res) {
		nLogMsg(EL_STATUS2, id, 
			"pid %d uid %d (%d volume%s and %d rid%s)",
			cp->c_pid, cp->c_uid,
			vol, (vol != 1 ? "s" : ""),
			res, (res != 1 ? "s" : ""));
	} else
		nLogMsg(EL_STATUS2, id, "nothing to detach");
}

/*
 *  MUmount() - process M_UNMOUNT packets recvd by the RP
 *
 *  TODO: since many req's might be queued for a vsn, consider having a flag
 *	to dismount all instances for uid, w/ root/volman additional flag
 *	to get other uids too. E.g. now a root dismount of someone else's
 *	request results in it being handed to the next queued requestor.
 */

MUnmount(vmpHdr_t *hp, Client_t *cp)
{
	char		*id = "MUnmount";
	VolReq_t	*vrp, *vrp2, *nextvrp, *FindVolReq();
	vmpVsn_t	*vp;
	char		*vsnp;
	int		uid, found = 0;

	/*
	 *  Find the volreq with the mount corresponding to this unmount.
	 *	Lookup is for (vsn, uid) except root and VOLMANUID can unmount
	 *	any request; the request furthest along in the mount
	 *	is chosen, other things being equal.
	 */

	vp = MSGPTR(hp, vmpVsn_t);
	uid = cp->c_uid;
	vsnp = (char *)&vp->v_vsn;
	for (vrp2=NULL,vrp=FindVolReq(vsnp); vrp; vrp=nextvrp) {
		nextvrp = VRNEXT(vrp->vr_vsnq);	/* because vrp might be DQed */

		/*
		 *  consider mount-related requests by appropriate uid
		 */
		if (MOUNT_STATES(vrp->vr_state)) {
		    found++;
		    if (uid == vrp->vr_uid || uid == 0 || uid == VOLMANUID) {

			/*
			 *  Got a candidate; see if it's better than any
			 *	previous one; the order of the tests is
			 *	important.
			 */
			
			if (vrp->vr_clientp == cp  &&  
			    vrp->vr_state == VS_MOUNTED) {
				/*
				 *  exactly right
				 */
				vrp2 = vrp;
				break;
			}
			if (!vrp2) {

				/*
				 *  It's the only one so far.
				 */

				nLogMsg(EL_DEBUG3, id, "1st %s", 
					vReqStateStr(vrp->vr_state));
				vrp2 = vrp;
				continue;
			}
			nLogMsg(EL_DEBUG3, id, "consider %s", 
					vReqStateStr(vrp->vr_state));
			if (vrp->vr_uid == uid  &&  vrp2->vr_uid != uid) {

				/*
				 *  The mounting uid is the same as the 
				 *	unmounter's and the uid of the 
				 *	best so far is not, i.e. authority
				 *	clients such as root will prefer to
				 *	unmount their own requests.
				 */

				nLogMsg(EL_DEBUG3, id, "same uid preferred");
				vrp2 = vrp;
				continue;
			}
			if (vrp->vr_clientp == cp  &&  vrp2->vr_clientp != cp) {

				/*
				 *  The mounting client is the same as the
				 *	unmounter, and the best so far has
				 *	a different client process (or is async)
				 */

				nLogMsg(EL_DEBUG3, id, "same client preferred");
				vrp2 = vrp;
				continue;
			}
			if (vrp2->vr_state < vrp->vr_state) {

				/*
				 *  The state is more advanced than the best
				 *	so far.
				 */

				nLogMsg(EL_DEBUG3, id, 
						"further state preferred");
				vrp2 = vrp;
				continue;
			}
			if (IPCHost(cp) == vrp->vr_clhostp  &&  
			    IPCHost(cp) != vrp2->vr_clhostp) {
				/*
				 *  prefer the client on same host
				 */
				nLogMsg(EL_DEBUG3, id, 
						"same host preferred");
				vrp2 = vrp;
				continue;
			}
		}
	    }
	}
	if (vrp2 == NULL) {

		/*
		 * no vol to unmount:  if not ASYNC, send M_VERROR pkt
		 */
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_STATUS2, id, "%.*s: %s", VSNLEN, vsnp,
		    (found ? "no permission to unmount" : "no mount found"));
		if (vp->v_flag & VO_WAIT)
			Err2Client(hp->vh_seq, cp);
		return(0);
	}
	nLogMsg(EL_DEBUG2, id, "selected %s _%d", 
				vReqStateStr(vrp2->vr_state), vrp2->vr_rpseq);
	vrp = vrp2;

	if (IPCHost(cp) != vrp->vr_clhostp)
		nLogMsg(EL_STATUS1, id, 
				"%.*s_%d: mount from %s, unmount from %s",
				VSNLEN, vsnp, vrp->vr_rpseq,
				vrp->vr_clhostp->h_name, 
				IPCHost(cp)->h_name);

	if (vrp->vr_ridp->rid_parallel) {
		/*
		 *  'parallel'/reserved or select mount - 
		 *	unmount all in reservation
		 */
		return(UnmountParallel(hp, cp, vrp));
	}

	vrp->vr_cmd = M_UNMOUNT;

	/*
	 *  Got it:
	 *	If not from the same client as that which made request, 
	 *	orphan and add to new client q if necc. 
	 *  TODO: if root/volman override, consider removing from rid too.
	 */

	if (cp != vrp->vr_clientp) {
		if (vrp->vr_clientp)
			nLogMsg(EL_DEBUG2, id, "taking from other client");
		Orphan(vrp, cp);	/* turns off vr_voflag VO_WAIT */
		if (vp->v_flag & VO_WAIT) {
			nLogMsg(EL_DEBUG2, id, "(queuing on new client)");
			EnQVolCVR(vrp, &cp->c_cvrq);
			vrp->vr_clientp = cp;
			vrp->vr_uid = cp->c_uid;
			vrp->vr_voflag |= VO_WAIT;
		}
	}

	/*
	 *  Install seq of dismounting client.
	 */
	vrp->vr_clseq = hp->vh_seq; 

	/*
	 *  Install new dismounting client's flags, except for the 
	 *	ones that determine the type of mount for purposes 
	 *	of checking handoff to another process.
	 */
	vrp->vr_voflag = vp->v_flag | (vrp->vr_voflag & VO_SETUP); 

	return(Unmount(vrp));
}

UnmountParallel(vmpHdr_t *hp, Client_t *cp, VolReq_t *vrp)
{
	char		*id = "UnmountParallel";
	vmpVsn_t	*vp = MSGPTR(hp, vmpVsn_t);
	Rid_t		*ridp = vrp->vr_ridp;
	Client_t	*cp2 = vrp->vr_clientp;
	int		ret = 0, origstate = vrp->vr_state;
	VolReq_t	*vrp2;

	/*
	 *  tell original client if it's being taken away by another
	 */
	if (cp2  &&  cp2 != cp) {
		VMError = E_CANCELED;
		nLogMsg(EL_CODE|EL_ERROR2, id, 
				"request (%.*s_%d) deleted by %s pid %d on %s", 
				VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq,
				nGetUIDName(cp->c_uid), cp->c_pid,
				IPCHost(cp)->h_name);
		Err2Client(vrp->vr_clseq, cp2);
	}

	/*
	 *  to simplify tracking, tell requesting client it's done now
	 */
	if (vp->v_flag & VO_WAIT) {
		/*
		 *  simulate 'dismount done' on this req
		 */
		vrp->vr_state = VS_UNMOUNTED;
		vrp->vr_clseq = hp->vh_seq; 
		vrp->vr_clientp = cp;
		Volstat2Client(vrp);
		vrp->vr_state = origstate;
	}

	/*
	 *  wipe all in rid, taking from client
	 */
	for (vrp=VRHEAD(ridp->rid_vsnq); vrp; vrp=vrp2) {
		vrp2 = VRNEXT(vrp->vr_ridq);	/* in case vrp is dq'd */

		if (cp2) {
			DEQUEUE(vrp, &cp2->c_cvrq, vr_cvrq, VolReq_t)
			vrp->vr_clientp = NULL;
		}
		ret += BUSY_STATES(vrp->vr_state);
		vrp->vr_clseq = hp->vh_seq; 
		vrp->vr_cmd = M_UNMOUNT;
		vrp->vr_voflag &= ~VO_WAIT;
		CancelRequest(vrp, cp, id);
	}
	DeleteRid(ridp, cp);
	return(ret);
}

/*
 *  UnmountDrive() - node host:/dev/vol/xxxxxx node removed
 *
 *	1) Update RC drive list, sentinel
 *	2) Send M_UNMOUNT to RC
 *	3) state = RCWAIT2
 */
 static void
UnmountDrive(VolReq_t *vrp)
{
	char		*id = "UnmountDrive";
	RPRC_t		*rcp = vrp->vr_rcp;
	int		bit;

	vClearError();

	if (rcp == NULL) {
		nLogMsg(EL_ERROR1, id, "INTERNAL ERROR: vr_rcp == NULL");
		return;
	}
	if (rcp->rc_srv.s_sd == -1) {
		nLogMsg(EL_STATUS1|EL_CODE|EL_CONSOLE0, id,
		  "cannot unmount %.*s: %s not running",
		  VSNLEN, &vrp->vr_vsn, rcp->rc_id);
		return;
	}

	/* TODO: check that bit is on in devmap and off in active */
	if ((bit = vDMffs(&vrp->vr_devinuse)) >= 0) {
		int	medindex;

		/*
		 *  a drive has been allocated to this request, free it
		 */
		medindex = RCMedium(rcp, DriveArray[bit].de_medium); 
		DM_SET(bit, rcp->rc_avail[medindex]);
	} 

	(void) Send2RC(vrp->vr_rcp, vrp, M_UNMOUNT, RcMsgNULL);
	vrp->vr_state = VS_RCWAIT2;
	vrp->vr_time = Now;
}

/*
 *  Unmount(): hand off or unmount this volume regardless 
 *	of its current state. If handoff, can call Volstat2Client().
 *	Return 1 if dismount sent (i.e. drive may be available to
 *	another request), 0 otherwise.
 */

Unmount(VolReq_t *vrp)
{
	char		*id = "Unmount";
	RPRC_t		*rcp = vrp->vr_rcp;
	VolReq_t	*vrp2, *HandOffUnMountOK();


	/*
	 *  see if anyone else can use it
	 */

	if (vrp2 = HandOffUnMountOK(vrp)) {

		nLogMsg(EL_STATUS1, id, 
			"UNMOUNT-handed off %.*s_%d to _%d state %s",
			VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq,
	  		vrp2->vr_rpseq, vReqStateStr(vrp->vr_state));

		/*
		 *  handed off to another request so don't free drive;
		 *	 but: decrement/increment rid_active.
		 */
		vrp->vr_ridp->rid_active--;
		vrp2->vr_ridp->rid_active++;

		/*
		 *  inform dismount requestor
		 */
		vrp->vr_state = VS_UNMOUNTED;
		vrp->vr_rctime2 = Now - vrp->vr_time;
		vrp->vr_time = Now;
		if (vrp->vr_voflag & VO_WAIT)
			Volstat2Client(vrp);
		else
			nLogMsg(EL_DEBUG3, id, "unmounter was async");

		/*
		 *  delete or inactivate original request
		 */

		if (vrp->vr_flag & VF_IMPLICITVSN) {
			FreeVolReq(vrp, id);
		} else {
			vrp->vr_state = VS_IDLE;
			vrp->vr_time = Now;
			vrp2->vr_qtime = 0;
			vrp2->vr_rctime1 = 0;
			vrp2->vr_usrtime = 0;
			vrp2->vr_rctime2 = 0;
			DEQUEUE(vrp, &VolReqQ, vr_vrq, VolReq_t);
		}
		return(0);
	}

	/*
	 *  no handoff - go ahead and unmount (if VO_HANDOFF,
	 *	mark for unmount after a delay in case another 
	 *	request comes).
	 */
	nLogMsg(EL_STATUS1, 0, "UNMOUNT %.*s_%d state %s",
			VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq,
	  		vReqStateStr(vrp->vr_state));

	switch (vrp->vr_state) {
	    case VS_MOUNTED:
		if (UnmountWaitTime > 0  &&  vrp->vr_voflag & VO_HANDOFF) {
			/*
			 *  give it a grace period
			 */
			nLogMsg(EL_DEBUG2, id, "unmntwait %.*s_%d",
					VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
			vrp->vr_state = VS_UNMOUNTWAIT;
			vrp->vr_time = time((time_t *)0);
			Timeout++;
			return(0);
		}
		return(DeleteNode(vrp, 1));
		
	    case VS_RCWAIT:
		nLogMsg(EL_STATUS1, id, "%.*s_%d: mark for unmount",
			VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
		vrp->vr_flag |= VF_UNMNTREQ;
		if (rcp->rc_flags & RCF_ASYNCUNMNT) {

			/*
			 *  ok to send unmount to RC b4 mount completes
			 */

			UnmountDrive(vrp);
			return(1);
		}
		return(0);

	    case VS_CREATENODE:
		nLogMsg(EL_STATUS1, id, "%.*s_%d: createnode; mark for unmount",
			VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
		vrp->vr_flag |= VF_UNMNTREQ;
		return(0);

	    case VS_MOUNTWAIT:
	    case VS_DRIVEWAIT:
	    case VS_RESERVEMOUNTWAIT:

		/*
		 *  Mount still queued for resources; can blow it away.
		 */
		if (vrp->vr_voflag & VO_WAIT)
			Volstat2Client(vrp);
		if (vrp->vr_flag & VF_IMPLICITVSN) {
			FreeVolReq(vrp, id);
		} else {
			DEQUEUE(vrp, &VolReqQ, vr_vrq, VolReq_t);
			vrp->vr_state = VS_IDLE;
			vrp->vr_time = Now;
			vrp2->vr_qtime = 0;
			vrp2->vr_rctime1 = 0;
			vrp2->vr_usrtime = 0;
			vrp2->vr_rctime2 = 0;
		}
		return(0);
		
	    case VS_RCWAIT2:	
	    case VS_DELETENODE:
		/*
		 *  already an unmount in progress
		 */
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_STATUS2, id, 
				"Unmount on %.*s_%d already in progress",
				VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
		if (vrp->vr_voflag & VO_WAIT && vrp->vr_clientp)
			Err2Client(vrp->vr_clseq, vrp->vr_clientp);
		return(0);

	    default:
		nLogMsg(EL_ERROR1, id, "INTERNAL ERROR: unexpected state: %s", 
				vReqStateStr(vrp->vr_state));
	}
	return(1);
}

/*
 *  HandOffMountDetailsFails() - make sure state of new request
 *	matches state known from previous request & set up node
 *	or if fails return 1.
 */
HandOffMountDetailsFails(VolReq_t *vrp1, VolReq_t *vrp2)
{
	char	*id = "HandOffMountDetailsFails";

	if (vrp1->vr_state != VS_MOUNTED  &&
	    vrp1->vr_state != VS_UNMOUNTWAIT)
		nLogMsg(EL_FATAL, id, "bad handoff attempt - state %s", 
						vReqStateStr(vrp1->vr_state));

	/*
	 *  check write ring
	 */
	if ((vrp2->vr_voflag & VO_WRITE) && !(vrp1->vr_flag & VF_RING)) {

		/*
		 *  write requested & volume not physically
		 *	write enabled: we fail the mount and perform 
		 *	the dismount (others could be waiting to just 
		 *	read this mount, but *to hell with them for now).
		 */

		VMError = E_UNAVAIL;
		nLogMsg(EL_CODE|EL_ERROR2, id, 
		  "Volume not physically write enabled - contact operator");
		ErrOutVolReq(vrp2, id);
		return(1);
	}
	if (!(vrp2->vr_voflag & VO_WRITE) && (vrp1->vr_flag & VF_RING)
				 && !(vrp2->vr_voflag & VO_IGNORE_RING)) {

		/*
		 *  read requested & write ring is set & no 'ignore_ring'
		 */

		VMError = E_UNAVAIL;
		nLogMsg(EL_CODE|EL_ERROR2, id, "%s %s",
		  "Read requested & volume is physically write enabled - ",
		  "contact operator");
		ErrOutVolReq(vrp2, id);
		return(1);
	}

	vrp2->vr_ndhostp = vrp1->vr_ndhostp;
	vrp2->vr_flag &= ~VF_RING;
	vrp2->vr_flag |= (vrp1->vr_flag & VF_RING);
	vrp2->vr_devinuse = vrp1->vr_devinuse;

	vrp2->vr_state = VS_MOUNTED;
	vrp2->vr_rctime1 += Now - vrp2->vr_time;
	vrp2->vr_time = Now;

	return(0);
}

/*
 *  HandOffUnMountOK() - give mount to waiting request under same reservation,
 *	if any. Waiting request may be async, i.e. no client connected.
 *	Since it's in the same request, no change to rid_active.
 */

 VolReq_t	*
HandOffUnMountOK(VolReq_t *vrp)
{
	char		*id = "HandOffUnMountOK";
	VolReq_t	*vrp2, *FindWaitingMount();

	if (shuttingdown)
		return(NULL);

	vrp2 = FindWaitingMount(vrp);
	if (!vrp2) {
		nLogMsg(EL_DEBUG3, id, "nothing valid waiting");
		return(NULL);
	}
	nLogMsg(EL_DEBUG3, id, "got waiting..");

	/*
	 *  vrp2 = another request for same vsn,
	 *	in single implicit reservation,
	 *	waiting for mount
	 */
	switch (vrp->vr_state) {
	    case VS_MOUNTED:
	    case VS_UNMOUNTWAIT:

		if (HandOffMountDetailsFails(vrp, vrp2))
			return(NULL);

		/*
		 *  done: inform recipient if necc
		 */
		if (!(vrp2->vr_voflag & VO_WAIT))
			nLogMsg(EL_DEBUG3, id, "no client waiting");
		else if (vrp2->vr_state == VS_MOUNTED)
			Volstat2Client(vrp2);
		return(vrp2);
		
	    case VS_RCWAIT:

		/*
		 *  to simplify things regarding mount setup
		 *	reject if requests not =
		 */

		if ((vrp->vr_voflag & VO_SETUP) != 
		    (vrp2->vr_voflag & VO_SETUP)) {
			nLogMsg(EL_DEBUG2, id, 
				"setup flags not compatible: %x %x",
				(vrp->vr_voflag & VO_SETUP), 
				(vrp2->vr_voflag & VO_SETUP));
			return(NULL);
		}
		vrp2->vr_rpseq = vrp->vr_rpseq;	/* so RC response will connect*/
		vrp2->vr_state = VS_RCWAIT;
		vrp2->vr_qtime += Now - vrp2->vr_time;
		vrp2->vr_time = Now;
		return(vrp2);
	    default:
		return(NULL);
	}
}

/*
 *  MMove() - handle M_MOVE move request
 */
MMove(vmpHdr_t *hp, Client_t *cp)
{
	char		*id = "MMove";
	vmpMove_t	*mp;
	vsnseq_t	*vsnseqp;
	VolReq_t	*vrp;
	RPRC_t		*dstrcp;	/* dst RC pointer */
	int		i;
	int		dstrc;		/* dst RC number */
	int		sync;
	char		*vReqStateStr();

	mp = MSGPTR(hp, vmpMove_t);

	/*
	 *  check the complex pkt since sizes are used to index
	 *	local memory
	 */
	if (hp->vh_count < 1) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR2, id, "bad vsn count %d", 
							hp->vh_count);
		Err2Client(hp->vh_seq, cp);
		return(0);
	}
	if (hp->vh_size != PadSize(sizeof(vmpMove_t)) + 
				hp->vh_count * sizeof(vsnseq_t)) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR2, id, 
			"bad packet length (%d vsns, %d/%d)",
					hp->vh_count,
					sizeof(vmpMove_t) + 
					hp->vh_count * sizeof(vsnseq_t),
					hp->vh_size);
		Err2Client(hp->vh_seq, cp);
		return(0);
	}

	/*
	 *  set up pointers into pkt
	 */
	vsnseqp = (vsnseq_t *) ((char *)mp + PadSize(sizeof(*mp)));
	
	/*
	 *  check destination
	 */

	dstrc = mp->vmv_dstrc;
	dstrcp = RCInfo(dstrc);
	if (dstrcp == NULL) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR2, id, 
			"destination rc (%d) for vsn %.*s %s not configured",
			dstrc, VSNLEN, &vsnseqp->r_vsn, 
			(hp->vh_count > 1 ? "and others" : ""));
		Err2Client(hp->vh_seq, cp);
		return(0);
	}

	/*
	 *  get or set up request(s) (hook up w/ rid if necc)
	 *	-postpone database & RC checks until rpq is evaluated
	 */

	sync = mp->vmv_flags & VO_WAIT;
	for (i=0; i<hp->vh_count; i++, vsnseqp++) {
		nLogMsg(EL_DEBUG3, id, "vsn %s seq %d", 
					vsnstr((char *)&vsnseqp->r_vsn),
					vsnseqp->r_seq);
		vrp = SetupRequest((char *)&vsnseqp->r_vsn, cp, sync, 
					vsnseqp->r_seq, mp->vmv_rid);
		if (vrp == NULL) {
			Err2Client(vsnseqp->r_seq, cp);
			continue;
		}

		/*
		 *  set up move request
		 */
		vrp->vr_cmd = M_MOVE;
		vrp->vr_voflag = mp->vmv_flags;
		vrp->vr_rcp = NULL;		/* don't know src RC yet */
		vrp->vr_dstrcp = dstrcp;    /* save pointer to dst RC struct */
		vrp->vr_info.vi_dstrc = dstrcp->rc_number;
		vrp->vr_clseq = vsnseqp->r_seq;
		vrp->vr_flag &= ~(VF_EJECTED|VF_ACCEPTED);

		vrp->vr_state = VS_MOVEWAIT;
		vrp->vr_time = Now;

		/*
		 *  leave request in queue for final checking & 
		 *	resources (i.e. vsn)
		 */
	
		ENQUEUE(vrp, &VolReqQ, vr_vrq, VolReq_t)
		nLogMsg(EL_STATUS1, id, "MOVE Qd %.*s_%d  %s/%d", 
			VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq,
			vReqStateStr(vrp->vr_state), vrp->vr_state);
	}
	return(1);
}

/*
 *  UpdateVolLocation() - common code for async updates.
 */
 static void
UpdateVolLocation(vmpVolStat_t *vsp, RPRC_t *destrcp)
{
	char		*id = "UpdateVolLocation";
	VInfo_t		*vip;
	VolReq_t	*vrp, *vrp2;

	if (destrcp) {
		nLogMsg(EL_STATUS1, id, "ASYNC VSN Update: VSN %.*s %s",
	  			VSNLEN, &vsp->vv_vsn, destrcp->rc_id);
	} else {
		nLogMsg(EL_STATUS1, id, "ASYNC VSN Update: VSN %.*s moved",
				VSNLEN, &vsp->vv_vsn);
	}
	vip = DBAsyncUpdateLoc((char *)&vsp->vv_vsn, destrcp);
	if (vip == NULL) {
		/* db file or index error TODO better */
		nLogMsg(EL_STATUS1, id, "vip is NULL");
		return;
	}
	nLogMsg(EL_DEBUG2, id, "vi_srcrc = %d", vip->vi_srcrc);

	if (destrcp == NULL) {

		/*
		 *  move completed & we only know the dest from the dbase
		 *	TODO better
		 */
		if ((destrcp = RCInfo(vip->vi_dstrc)) == NULL) {
			nLogMsg(EL_CONSOLE0|EL_ERROR1, id, 
						"UNKNOWN DESTINATION");
			if (vip->vi_srcrc != vip->vi_dstrc)
				nLogMsg(EL_CONSOLE0|EL_ERROR1, id, 
						" - AND CAN'T UNDO MOVE");
			return;
		}
	}

	if (vip->vi_srcrc != destrcp->rc_number) {
		vmpMove_t	*movp;
		vmpHdr_t	*hp;
		vsnseq_t	*vsnseqp;

		/*
		 * if vi_srcrc == -1 (a forbidden number for an rc), the VDM 
		 *	is saying that there's no record of this volume 
		 *	in the database.  So eject it from the system.
		 *	"handcraft" a move request to get the volume
		 *	out of the repository it was just placed in.
		 *
		 * if vi_srcrc == -2 (a forbidden number for an rc), the VDM 
		 *	is saying that the volume didn't move, so just quit
		 *	rather than wondering how it affects other requests.
		 *
		 * if vip->vi_srcrc != destrcp->rc_number, the VDM is saying 
		 *	that the user's quota was exceeded. Therefore, put it
		 *      back where it came from.
		 *	TODO: check quota in original rc, & have per-rc 
		 *	default places for orphan volumes to go.
		 */

		if (vip->vi_srcrc == -1) {
			nLogMsg(EL_ERROR1|EL_CONSOLE0, id,
			  "%.*s: entered into %s but not in database:",
			  VSNLEN, &vsp->vv_vsn, destrcp->rc_id);
			nLogMsg(EL_ERROR1|EL_CONSOLE0, id,
			  "%.*s: remove, add to database and re-enter",
			  VSNLEN, &vsp->vv_vsn);
		} else if (vip->vi_srcrc == -2) {
			nLogMsg(EL_CONSOLE0|EL_ERROR2, id,
			  "!! %.*s: entered into %s but already there:",
			  VSNLEN, &vsp->vv_vsn, destrcp->rc_id);
			return;
		} else {
			nLogMsg(EL_ERROR1|EL_CONSOLE0, id,
			  "%.*s: entered into %s but quota has been exceeded:",
			  VSNLEN, &vsp->vv_vsn, destrcp->rc_id);
			nLogMsg(EL_ERROR1|EL_CONSOLE0, id,
			  "%.*s: remove, increase quota, and re-enter",
			  VSNLEN, &vsp->vv_vsn);
			/* TODO - what if there is a request for volume? */
			if (FindVolReq((char *)&vsp->vv_vsn) != NULL)
				nLogMsg(EL_CONSOLE0|EL_STATUS1, id, 
					"!!!undoing bad move with req waiting");
		}

		/*
		 *  move it out of RC
		 */
		hp = (vmpHdr_t *) nAlloc(PKTSIZ(vmpMove_t) + sizeof(vsnseq_t));
		hp->vh_type = M_MOVE;
		hp->vh_seq = vGetSeq();
		hp->vh_size = PadSize(sizeof(vmpMove_t)) + sizeof(vsnseq_t);
		hp->vh_count = 1;
		movp = MSGPTR(hp, vmpMove_t);
		movp->vmv_srcrc = destrcp->rc_number;
		movp->vmv_dstrc = vip->vi_srcrc;	/* -1 or source rc */
		memset(movp->vmv_rid, '\0', sizeof(movp->vmv_rid));

		vsnseqp = (vsnseq_t *) ((char *)movp + PadSize(sizeof(*movp)));
		vsnseqp->r_seq = vGetSeq();
		vsnseqp->r_vsn = vsp->vv_vsn;

		nLogMsg(EL_STATUS2, id,
		  "issuing move/eject request: VSN=%.*s SRC=%d, DST=%d",
		  VSNLEN, &vsp->vv_vsn, movp->vmv_srcrc, movp->vmv_dstrc); 
		XDRSendI(hp, destrcp->rc_srv.s_sd, 1);
		nFree(hp);
		return;
	}

	/*
	 *  move ok; see how it affects any waiting requests
	 */

	vip->vi_srcrc = destrcp->rc_number;
	for (vrp=FindVolReq((char *)&vsp->vv_vsn); vrp; vrp=vrp2) {
		vrp2 = VRNEXT(vrp->vr_vsnq);	/* vrp may be closed below */
		vrp->vr_info = *vip;
		switch (vrp->vr_state) {
		    case VS_MOUNTWAIT:
		    case VS_DRIVEWAIT:
		    case VS_RESERVEMOUNTWAIT:
			ChangeRCMount(destrcp, vrp, 0);
			break;
		    case VS_IDLE:
			break;
		    case VS_RCWAIT:
			/*
			 *  tidy up request in old RC
			 */
			if (vrp->vr_rcp->rc_flags & RCF_ASYNCUNMNT) {
				/*
				 *  send a cancel and tell RC to not respond
				 *	since we want to convert this request 
				 *	now.
				 */
				nLogMsg(EL_DEBUG2,id, "cancel req, noresponse");
				vrp->vr_flag |= VF_NORESP;
				Send2RC(vrp->vr_rcp, vrp, M_UNMOUNT, RcMsgNULL);
				vrp->vr_flag &= ~VF_NORESP;
			} else {
				/*
				 *  an untidy situation that shouldn't happen;
				 *	leave the doomed request in the other RC
				 *	and hope that nothing gets done about it
				 *	in that RC.
				 */
				nLogMsg(EL_CONSOLE0|EL_ERROR1, id, 
				    "!!!mount in non-cancelable RC, vsn moved");
			}
			/*
			 *  clean up drive stuff in old RC & make sure 
			 *	mount will work in new one
			 */
			ChangeRCMount(destrcp, vrp, 1);
			break;
		    case VS_MOUNTED:
			nLogMsg(EL_STATUS1|EL_CONSOLE0, id, 
				"%.*s_%d: moved while mounted - deleting",
                                VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
			/* TODO - test; this is just a guess.. */
			FreeVolReq(vrp, id);
			break;
		    default:
			break;
		}
	}
}
/*
 *  AsyncVolStat() - volstat recvd w/ no client VolReq waiting
 */
 static void
AsyncVolStat(vmpHdr_t *hp, RPRC_t *rcp)
{
	char		*id = "AsyncVolStat";
	vmpVolStat_t	*vsp;
	RPRC_t		*rcp2;
		
	vsp = MSGPTR(hp, vmpVolStat_t);
	nLogMsg(EL_DEBUG1, id, "%s", vsnstr((char *)&vsp->vv_vsn));

	switch (vsp->vv_state) {
	    case VS_ACCEPTED:
		UpdateVolLocation(vsp, rcp);
		return;
	    case VS_1MSGMOVE:

		/*
		 *	1MSGMOVE without a vrp:  These are
		 *	from an RC which was requested to move
		 *	a vsn and the dest RC went down before
		 *	the move was complete.  This packet indicates
		 *	that we consider the move completed anyway.  
		 *	If the vv_rc field of the VolStat message is
		 *	filled in with the dest RC (!= -1, vv_flags 
		 *	& VV_RC), use that RC, otherwise assume the 
		 *	vsn reached the dest in the dbase.
		 */

		if (vsp->vv_flags & VV_RC) {
			rcp2 = RCInfo(vsp->vv_rc);
			if (rcp2)
				UpdateVolLocation(vsp, rcp2);
			else
				nLogMsg(EL_ERROR1|EL_CONSOLE0, id,
					"VS_1MSGMOVE to bogus RC %d from %s ", 
					vsp->vv_rc, rcp->rc_id);
		} else {
			/*
			 *  e.g. acs saying, "I ejected ok,
			 *	so assume it got wherever it was going"
			 */
			UpdateVolLocation(vsp, NULL);
		}
		return;
	  default:
		nLogMsg(EL_ERROR1, id, "INTERNAL ERROR: unexpected state: %s",
				vReqStateStr(vsp->vv_state));
	}
}

/*
 *  RCUnmounted() - rc sent volstat & client's request is waiting 
 *			for an unmount
 */
 static void
RCUnmounted(vmpHdr_t *hp, VolReq_t *vrp)
{
	char		*id = "RCUnmounted";
	RPRC_t		*rcp = vrp->vr_rcp;
	vmpVolStat_t	*vsp;
	int		devno, medindex;
	DevEnt_t	*dep;

	
	nLogMsg(EL_DEBUG2, 0, "UNMOUNTED %.*s_%d rid_active %d", 
		VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq, vrp->vr_ridp->rid_active);

	vsp = MSGPTR(hp, vmpVolStat_t);
	
	/*
	 *  sanity check
	 */
	if (vsp->vv_state != VS_UNMOUNTED) {
		nLogMsg(EL_ERROR1, id, "INTERNAL ERROR: volstat on unmount: %s",
				vReqStateStr(vsp->vv_state));
		return;
	}

	/*
	 *  look up drive & reset any vary err string
	 */
	devno = vDMffs(&vrp->vr_devinuse);
	if (devno == -1)
		nLogMsg(EL_ERROR1, id, "%.*s: dev not known", 
						VSNLEN, &vrp->vr_vsn);
	else if (dep = vGetDevByNum(devno))
		seterr(dep, NULL);

	/*
	 *  reset vsn in database
	 */
	DBUnmountDone((char *)&vrp->vr_vsn);		/* updated database */

	/*
	 *  decrement rid_active
	 */
	vrp->vr_ridp->rid_active--;

	/*
	 *  release drive in RC
	 */
	medindex = RCMedium(rcp, vVolTypeMedium(vrp->vr_info.vi_voltype));
	rcp->rc_devs[medindex].d_mountdev--;
	if (devno != -1)
		DM_CLR(devno, rcp->rc_inuse[medindex]);

	/*
	 *  reset user's drive reservation and update volreq state
	 */
	FreeDriveUsed(&vrp->vr_ridp->rid_RCMed[
			rcp->rc_index * RC_MEDIA_TYPES + medindex]);
	DM_ZERO(vrp->vr_devinuse);
	vrp->vr_state = VS_UNMOUNTED;
	vrp->vr_rctime2 = Now - vrp->vr_time;
	vrp->vr_time = Now;

	/*
	 *  tell client
	 */
	if (vrp->vr_voflag & VO_WAIT)
		Volstat2Client(vrp);

	/*
	 *  complete vary off of drive if requested
	 */
	if (vrp->vr_flag & VF_VARYOFF) {
		char		*devnam;

		/*
		 *  VF_VARYOFF is set when someone attempts to vary
		 *	off a device with a volume mounted on it.
		 *	First, the volume must be unmounted (which
		 *	just completed).  Now the vary off can
		 *	be completed.
		 */

		if (dep)
			devnam = dep->de_name;
		else
			devnam = "???";
		nLogMsg(EL_STATUS1, id,
			  "Completing Vary off for dev: %s", devnam);
		Vary(devno, VVA_VARYOFF, VOLMANUID, 1, NULL);
	}

	/*
	 *  delete or inactivate (idle) request
	 */

	if (vrp->vr_flag & VF_IMPLICITVSN) {
		nLogMsg(EL_DEBUG3, id, "implicit/free");
		FreeVolReq(vrp, id);
	} else {
		vrp->vr_state = VS_IDLE;
		vrp->vr_time = Now;
		vrp->vr_qtime = 0;
		vrp->vr_rctime1 = 0;
		vrp->vr_usrtime = 0;
		vrp->vr_rctime2 = 0;
		DEQUEUE(vrp, &VolReqQ, vr_vrq, VolReq_t);
	}
	nLogMsg(EL_DEBUG3, id, "done");
}

/*
 *  RCMoved() - got an RC response (volstat) on a move request
 */
 static void
RCMoved(vmpHdr_t *hp, VolReq_t *vrp)
{
	char		*id = "RCMoved";
	vmpVolStat_t	*vsp;
	int		forcemove = 0, failed = 0;
	int		flag;

	vsp = MSGPTR(hp, vmpVolStat_t);
	switch (vsp->vv_state) {
		  case VS_EJECTED: 
			vrp->vr_flag |= VF_EJECTED; 
			break;
		  case VS_1MSGMOVE: 
			/*
			 *  no need for other RC's response
			 */
			forcemove = 1; 
			nLogMsg(EL_DEBUG2, id, "%.*s_%d: 1MSGMOVE",
				 VSNLEN, &vsp->vv_vsn, vrp->vr_rpseq);
			break;
		  case VS_ACCEPTED:
			/*
			 *  no need for other RC's response
			 */
			forcemove = 1;
			vrp->vr_flag |= VF_ACCEPTED;
			/* 
			 *  forward M_VOLSTAT to src RC 
			 */
			if (vrp->vr_rcp->rc_srv.s_sd != -1) {
			    nLogMsg(EL_DEBUG1, id,
				 "%.*s_%d: forwarding VS_ACCEPTED to src (%s)",
				 VSNLEN, &vsp->vv_vsn, vrp->vr_rpseq,
				 vrp->vr_rcp->rc_id);
			    failed = XDRSendI(hp, vrp->vr_rcp->rc_srv.s_sd, 1);
			} else
				VMError = failed = E_UNAVAIL;	/* RC down! */
			if (failed)
				nLogMsg(EL_STATUS1|EL_CONSOLE0|EL_CODE, id,
				  "%.*s_%d: can't forward VOLSTAT to %s",
				  VSNLEN, &vsp->vv_vsn, vrp->vr_rpseq,
				  vrp->vr_rcp->rc_id);
			break;
		  default:
			nLogMsg(EL_ERROR1, id,
			  "INTERNAL ERROR: %s/%d: unexpected state for move",
			  vReqStateStr(vsp->vv_state), vsp->vv_state);
	}
	flag = vrp->vr_flag;
	if (((flag & VS_ACCEPTED) && (flag & VS_EJECTED)) || forcemove) {
		/*
		 *  move is complete: update dbase & tell client if any
		 */
		vrp->vr_state = VS_IDLE;
		vrp->vr_rctime1 = Now - vrp->vr_time;
		vrp->vr_time = Now;
		DBUpdateLoc((char *)&vrp->vr_vsn, vrp->vr_dstrcp);
		if (vrp->vr_clientp) {
			if (VMError)
				Err2Client(vrp->vr_clseq, vrp->vr_clientp);
			else
				AckClient(vrp->vr_clseq, vrp->vr_clientp);
		}
		/*
		 *  delete or inactivate (idle) request
		 */

		vrp->vr_ridp->rid_active--;
		if (vrp->vr_flag & VF_IMPLICITVSN) {
			FreeVolReq(vrp, id);
		} else {
			DEQUEUE(vrp, &VolReqQ, vr_vrq, VolReq_t);
		}
	}
}

extern int	obsoletetypes[];
 static void
Scratched(VolReq_t *vrp)
{
	char		*id = "Scratched";
	char		*vsnp = (char *)&vrp->vr_vsn;
	char		*label = (char *)&vrp->vr_label;
	RPRC_t		*rcp = vrp->vr_rcp;
	VInfo_t		*vip = &vrp->vr_info;
	int		medindex, i, obsolete = 0;

	/*
	 *  release RC drive 
	 */
	medindex = RCMedium(rcp, vVolTypeMedium(vip->vi_voltype));
	rcp->rc_devs[medindex].d_mountdev--;

	/*
	 *  decrement rid_active
	 */
	vrp->vr_ridp->rid_active--;

	/*
	 *  free user drive reservation
	 */
	FreeDriveUsed(&vrp->vr_ridp->rid_RCMed[
			rcp->rc_index * RC_MEDIA_TYPES + medindex]);

	/*
	 *  see if vsn is a type that's being 
	 *	pushed out of the system
	 */
	for (i=0; obsoletetypes[i]; i++) {
		if (vrp->vr_voltype == obsoletetypes[i]) {
			obsolete = 1;
			break;
		}
	}

	/*
	 *  update dbase 
	 */
	memcpy(label, vsnp, VSNLEN);
	if (DBLabelQuery(vsnp, label, vrp->vr_uid, obsolete)) {
		if (vrp->vr_clientp)
			Err2Client(vrp->vr_clseq, vrp->vr_clientp);
		return;
	}
	nLogMsg(EL_STATUS1, 0, "SCRATCHED%s %s", 
				(obsolete ? "/OBSOLETE" : ""), 
				vsnstr(vsnp));
	if (vrp->vr_clientp)
		AckClient(vrp->vr_clseq, vrp->vr_clientp);
	FreeVolReq(vrp, id);
}

/*
 *  MVolStat() - process M_VOLSTAT packets (from RCs)
 */

MVolStat(vmpHdr_t *hp, RPRC_t *rcp)
{
	char		*id = "MVolStat";
	VolReq_t	*vrp;
	vmpVolStat_t	*vsp;

	vsp = MSGPTR(hp, vmpVolStat_t);

	/*
	 *  Find the VolReq (if any) this M_VOLSTAT corresponds to
	 *	TODO: make sure ACCEPTEDs from the ACS have a sequence number,
	 *		when in response to move requests
	 */

	vrp = FindVolReqSeq((char *)&vsp->vv_vsn, hp->vh_seq);

	/*
	 *  basic information is now in hand: 
	 *	handle the packet async if it's not associated w/ a request
	 */

	if (vrp == NULL) {
		AsyncVolStat(hp, rcp);
		return(1);
	}

	/*
	 *  There's a request associated w/ packet
	 */
	switch (vrp->vr_state) {
	    case VS_RCWAIT:
		if (vrp->vr_voflag & VO_SCRATCH) {
			Scratched(vrp);
			return(0);
		} else
			return(RCMounted(hp, vrp, rcp));
	    case VS_RCWAIT2:
		RCUnmounted(hp, vrp);
		return(1);
	    case VS_MOVING:
		RCMoved(hp, vrp);
		return(1);
	    default:
		nLogMsg(EL_ERROR1, id, "INTERNAL ERROR: unexpected state: %s",
				vReqStateStr(vsp->vv_state));
		return(0);
	}
}

/*
 *  CheckParallelMount() - see if all reqs mounted
 */
 static void
CheckParallelMount(Rid_t *ridp)
{
	VolReq_t	*vrp;

	for (vrp=VRHEAD(ridp->rid_vsnq); vrp; vrp=VRNEXT(vrp->vr_ridq)) {
		if (vrp->vr_state != VS_MOUNTED)
			return;
	}
	AckClient(ridp->rid_clseq, ridp->rid_cp);
}

/*
 *  CancelOtherMounts() - for select mount, one has succeeded,
 *	so quietly delete others
 */
 static void
CancelOtherMounts(VolReq_t *vrp)
{
	char		*id = "CancelOtherMounts";
	Rid_t		*ridp = vrp->vr_ridp;
	Client_t	*cp = vrp->vr_clientp;
	VolReq_t	*vrp1, *vrp2;

	for (vrp1=VRHEAD(ridp->rid_vsnq); vrp1; vrp1=vrp2) {
		vrp2 = VRNEXT(vrp1->vr_ridq);
		if (vrp1 == vrp)
			continue;
	
		/*
		 *  remove from client queue 
		 */
		DEQUEUE(vrp1, &cp->c_cvrq, vr_cvrq, VolReq_t)

		/*
		 *  throw away
		 */
		CancelRequest(vrp1, cp, id);
	}
}


/*
 *  RCMounted() - rc sent volstat and client's request is waiting
 */

RCMounted(vmpHdr_t *hp, VolReq_t *vrp, RPRC_t *rcp)
{
	char		*id = "RCMounted";
	int		drive, medindex;
	vmpVolStat_t	*vsp;
		
	vsp = MSGPTR(hp, vmpVolStat_t);

	if (vsp->vv_state != VS_MOUNTED) {
		nLogMsg(EL_ERROR1, id, 
		    "INTERNAL ERROR: Bad protocol: volstat %s on mount: %s",
		    vReqStateStr(vsp->vv_state), 
		    vReqStateStr(vrp->vr_state));
		return(0);
	}
	nLogMsg(EL_STATUS1, 0, "MOUNTED %.*s_%d",
			VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);

	/*
	 *  mount was successful
	 */

	drive = vsp->vv_dev;	/* drive bit */
	if (drive < 0  ||  drive >= DriveTotal) {
		nLogMsg(EL_ERROR1|EL_CONSOLE0, id, "bad drive from RC");
		UnmountDrive(vrp);
		return(0);
	}
	medindex = RCMedium(rcp, DriveArray[drive].de_medium);
	DM_SET(drive, vrp->vr_devinuse);	/* set dr used bit */
	DM_CLR(drive, rcp->rc_avail[medindex]);	/* clr dr avail bit */
	DM_SET(drive, rcp->rc_inuse[medindex]);	/* set inuse bit */
	if (vsp->vv_flags & VV_RING)
		vrp->vr_flag |= VF_RING;
	else
		vrp->vr_flag &= ~VF_RING;

	/*
	 *  see if an unmount has been requested while mounting
	 */
	if (vrp->vr_flag & VF_UNMNTREQ) {
		nLogMsg(EL_STATUS2, id,
			  "%.*s_%d: unmount received during mount",
			  VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
		if (UnmountWaitTime > 0  &&  vrp->vr_voflag & VO_HANDOFF) {
			time_t	t;
			time(&t);
			/*
			 *  give it a grace period
			 */
			vrp->vr_state = VS_UNMOUNTWAIT;
			vrp->vr_rctime1 += t - vrp->vr_time;
			vrp->vr_time = t;
			Timeout++;
		} else {
			DeleteNode(vrp, 1);
		}
		return(1);
	}

	vrp->vr_rctime1 += Now - vrp->vr_time;
	vrp->vr_time = Now;

	/*
	 *  see if RC checked label / created node
	 */ 
	if (rcp->rc_flags & RCF_NONODE) {
		/*
		 *  Send label check / node setup NOT TESTED
		 *				& parallel not impl
		 */
		nLogMsg(EL_FATAL, id, "untested code");
		CreateNode(vrp);
	} else {
		/*
		 *  mount completed - RC set up node
		 */
		vrp->vr_state = VS_MOUNTED;
		if (vrp->vr_ridp->rid_parallel == 1)
			CheckParallelMount(vrp->vr_ridp);
		else {
nLogMsg(EL_STATUS1, id, "seq %d", vrp->vr_clseq);
			Volstat2Client(vrp);
			if (vrp->vr_ridp->rid_parallel == 2)
				CancelOtherMounts(vrp);
		}
	}
	return(0);
}



/*
 *  ChangeRCMount() - change mount from 1 RC to another because vol moved
 */
ChangeRCMount(RPRC_t *destrcp, VolReq_t *vrp, int undoprevsent)
{
	char		*id = "ChangeRCMount";
	RPRC_t		*rcp = vrp->vr_rcp;
	Rid_t		*ridp = vrp->vr_ridp;
	int		medium, medindex;
	RidRCMed_t	*ridrcmed;

	medium = vVolTypeMedium(vrp->vr_info.vi_voltype);
	medindex = RCMedium(rcp, medium);

	ridrcmed = &ridp->rid_RCMed[rcp->rc_index * RC_MEDIA_TYPES + medindex];

	/*
	 *  clean up drive stuff in old RC if necc
	 */

	if (undoprevsent) {
		/*
		 *  req actually was sent to prev RC, but we assume
		 *	it will never be filled now, so undo the drive
		 *	allocation
		 */
		if (vrp->vr_ridp->rid_active)
			vrp->vr_ridp->rid_active--;
		if (ridrcmed->rr_implicit > 0) {
			ridrcmed->rr_softalloc--;
			ridrcmed->rr_softused--;
			ridrcmed->rr_implicit--;
		} else if (ridrcmed->rr_hardused) {
			ridrcmed->rr_hardused--;
		}
		vrp->vr_rcp->rc_devs[medindex].d_mountdev--;
	} else {
		/*
		 *  req just waiting
		 */
		if (ridrcmed->rr_implicit > 0) {
			ridrcmed->rr_implicit--;
			ridrcmed->rr_softalloc--;
		}
	}
	/*
	 *  make sure the new RC can mount
	 */
	if (!(destrcp->rc_flags & RCF_MOUNTOK)) {
		VMError = E_UNAVAIL;
		nLogMsg(EL_CODE|EL_ERROR2, id, 
				"Volume moved to RC %s w/ no mount capability",
				destrcp->rc_id);
		ErrOutVolReq(vrp, id);
		return(1);
	}
	/*
	 *  new RC needs to be up.. 
	 */
	if (destrcp->rc_srv.s_sd == -1) {
		VMError = E_UNAVAIL;
		nLogMsg(EL_CODE|EL_ERROR2, id, 
				"Volume moved to RC %s which is down",
				destrcp->rc_id);
		ErrOutVolReq(vrp, id);
		return(1);
	}

	/*
	 *  set up for rp_q.c
	 */
	vrp->vr_devtotry = destrcp->rc_devs[medindex].d_devmap;
	vrp->vr_rcp = destrcp;
	vrp->vr_state = VS_MOUNTWAIT;
	vrp->vr_time = Now;
	return(0);
}


/*
 *  MNodeRC() - node setup/unmount req from RC - forward to volnd
 *	on drive's host if there is a request, any ? host
 *	otherwise..
 */
MNodeRC(vmpHdr_t *hp, RPRC_t *rcp)
{
	char		*id = "MNodeRC";
	vmpNode_t	*np;
	Host_t		*hop;
	VolReq_t	*vrp;

	np = MSGPTR(hp, vmpNode_t);

	if (np->vs_host == -1) {
		/*
		 *  it's an RC request to any volnd to manipulate 
		 *	the drive only, w/out reference to
		 *	creating/deleting a user node. Choose
		 *	a host & forward.
		 */
		int	i;
		for (i=0; hop=HostIndex(i); i++) {
			if (hop->h_volnd.s_sd != -1) {
				if (!XDRSendI(hp, hop->h_volnd.s_sd, 1))
					return(0);
			}
		}
		VMError = E_UNAVAIL;
		nLogMsg(EL_CODE|EL_ERROR1, id, "no volnd to do request");
		Err2Srv(hp->vh_seq, &rcp->rc_srv);
		
		return(0);
	}
	if (np->vs_drive < 0  ||  np->vs_drive >= DriveTotal) {
		VMError = E_INTERNAL;
		nLogMsg(EL_CODE|EL_ERROR1, id, "bad drive %d",
						np->vs_drive);
		Err2Srv(hp->vh_seq, &rcp->rc_srv);
		return(0);
	}
	hop = HostIndex(DriveArray[np->vs_drive].de_host);
	if (hop == NULL) {
		VMError = E_INTERNAL;
		nLogMsg(EL_CODE|EL_ERROR1, id, "bad host index %d", 
						np->vs_host);
		Err2Srv(hp->vh_seq, &rcp->rc_srv);
		return(0);
	}

	/*
	 *  forward to volnd on drive's host 
	 */
	if (hop->h_volnd.s_sd == -1) {
		VMError = E_UNAVAIL;
		nLogMsg(EL_CODE|EL_ERROR1, id, 
				"no volnd on host %s for %.*s_%d",
				hop->h_name,
				VSNLEN, &np->vs_vsn, hp->vh_seq);
		Err2Srv(hp->vh_seq, &rcp->rc_srv);
		return(0);
	}
nLogSetLevel(EL_DEBUG4);
	nLogMsg(EL_STATUS1, id, "%.*s_%d -> volnd %s", 
				VSNLEN, &np->vs_vsn, hp->vh_seq, hop->h_name);
	if (XDRSendI(hp, hop->h_volnd.s_sd, 1))
		Err2Srv(hp->vh_seq, &rcp->rc_srv);
	else
nLogMsg(EL_STATUS1, id, "(sent ok)");

	/*
	**	if there's an associated volreq, mark the
	**	volnd host in it so that an error can be
	**	sent to the RC if the host/volnd goes down
	**
	**	note - depends on RC using vold's mount seq 
	**	for hp->vh_seq
	*/
	if ((vrp = FindVolReqSeq((char *)&np->vs_vsn, hp->vh_seq)) == NULL) {
		nLogMsg(EL_STATUS1, id, "(no volreq)");
	} else {
		vrp->vr_ndhostp = hop;
	}
nLogSetLevel(EL_STATUS2);
	return(0);
}

 void
DeleteNodeDone(VolReq_t *vrp, int drive)
{
	char	*id = "DeleteNodeDone";

	DriveArray[drive].de_status &= ~DS_DELETENODE;
	vrp->vr_ndhostp = NULL;
	UnmountDrive(vrp);
}


/*
 *  MNodeV() - label check / node setup - from volnode. 
 *
 *	response to RP 
 *		createnode: not tested; volnd created user node for vold
 *		deletenode: always: volnd just removed user node for vold
 *			which will now tell RC to unmount
 *			not tested: volnd also did an unmount
 *		Return 0 if ok, 1 if e.g. cancellation could 
 *			affect other requests.
 *      response to RC
 *		pass thru to RC
 *		mount: RC has set up user node, 
 *			& will be telling rp that mount is done
 *		dismount: manual RC normally dismounts at the start
 *			of a mount, except on startup when all
 *			drives are cleared. ACS does dismounts locally.
 *		return 0
 */
MNodeV(vmpHdr_t *hp, Host_t *hop)
{
	char		*id = "MNodeV";
	vmpNode_t	*np;
	VolReq_t	*vrp;
	RPRC_t		*rcp;
	int		drive;

	np = MSGPTR(hp, vmpNode_t);

	/*
	 *  pass on if response to RC req
	 */ 
	if (np->vs_req == -1) {
		/*
		 *  it's for RC - forward
		 */
		rcp = RCInfo(np->vs_rc);
		if (rcp == NULL) {
			nLogMsg(EL_CONSOLE0|EL_ERROR2, id, 
						"rc %d unknown", np->vs_rc);
			return(0);
		}
		nLogMsg(EL_STATUS2, id, "%s -> %s", 
					vMsgType(hp->vh_type), rcp->rc_id);
		if (rcp->rc_flags & RCF_NONODE)
			nLogMsg(EL_CONSOLE0|EL_ERROR1, id, 
						"RC but RC is nonode??");
		if (rcp->rc_srv.s_sd == -1)
			nLogMsg(EL_STATUS2, id, "RC %s down, throwing away pkt",
						rcp->rc_id);
		else
			XDRSendI(hp, rcp->rc_srv.s_sd, 1);
		return(0);
	}

	/*
	 *  look up the request
	 */
	if ((vrp = FindVolReqSeq((char *)&np->vs_vsn, hp->vh_seq)) == NULL) {
		nLogMsg(EL_STATUS2, id, "no request for %.*s_%d host %s",
				VSNLEN, &np->vs_vsn, hp->vh_seq, hop->h_name);
		return(0);
	}
	rcp = vrp->vr_rcp;
	drive = vDMffs(&vrp->vr_devinuse);

	/*
	 *  vold does all user node deletes -
	 *	RC can pass a delete w/ VO_UNMOUNT to ioctl unmount
	 */
	if (vrp->vr_state == VS_DELETENODE) {
		DeleteNodeDone(vrp, drive);
		return(1);
	}

	/*
	 *  response to an RP non-delete request, deal w/ it
	 *	NOT TESTED
	 */
	if (vrp->vr_state != VS_CREATENODE) 
		nLogMsg(EL_FATAL, id, "%.*s_%d got %s but state is %s",
			VSNLEN, &np->vs_vsn, hp->vh_seq, 
			vHdrStr(hp), vReqStateStr(vrp->vr_state));

	/*
	 *  note write ring state
	 */
	if (np->vs_flag & VV_RING)
		vrp->vr_flag |= VF_RING;
	else
		vrp->vr_flag &= ~VF_RING;
	/*
	 *  see if original req canceled
	 */
	if (vrp->vr_flag & VF_UNMNTREQ) {
		nLogMsg(EL_STATUS2, id,
			  	"%.*s_%d: unmount received during node setup",
			  	VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
		if (UnmountWaitTime > 0  &&  vrp->vr_voflag & VO_HANDOFF) {
			time_t	t;
			time(&t);
			/*
			 *  give it a grace period
			 */
			vrp->vr_state = VS_UNMOUNTWAIT;
			vrp->vr_rctime1 += t - vrp->vr_time;
			vrp->vr_time = t;
			Timeout++;
			return(1);
		}
		return(DeleteNode(vrp, 1));
	} 

	/*
	 *  happily conclude the tortuous mount process
	 */
	vrp->vr_state = VS_MOUNTED;
	vrp->vr_rctime1 += Now - vrp->vr_time;
	vrp->vr_time = Now;
	Volstat2Client(vrp);
	return(0);
}

/*
 *  MErrorV() - error msg from volnd
 */
MErrorV(vmpHdr_t *hp, Host_t *hop)
{
	char		*id = "MErrorV";
	vmpError_t	*ep = MSGPTR(hp, vmpError_t);
	VolReq_t	*vrp;
	RPRC_t		*rcp;

	if (ep->vn_state == -1) {
		/*
		 *  it's for RC - forward
		 */
		rcp = RCInfo(ep->vn_flag);
		if (rcp == NULL) {
			nLogMsg(EL_CONSOLE0|EL_ERROR1, id, 
						"bad rc %d", ep->vn_flag);
			return(0);
		}
		if (rcp->rc_flags & RCF_NONODE)
			nLogMsg(EL_CONSOLE0|EL_ERROR1, id, 
						"node msg for NONODE RC!!");

		if (rcp->rc_srv.s_sd == -1)
			nLogMsg(EL_STATUS2, id, "RC %s down, throwing away pkt",
						rcp->rc_id);
		else
			XDRSendI(hp, rcp->rc_srv.s_sd, 1);
		return(0);
	}

	/*
	 *  find request
	 */
	ep = MSGPTR(hp, vmpError_t);
	if ((vrp = FindVolReqSeq((char *)&ep->vn_vsn, hp->vh_seq)) == NULL) {
		nLogMsg(EL_STATUS2, id, "no request for %.*s_%d host %s",
				VSNLEN, &ep->vn_vsn, hp->vh_seq, hop->h_name);
		return(0);
	}
	rcp = vrp->vr_rcp;
	
	/*
	 *  it's a response to an RP request,
	 *	deal w/ it NOT TESTED
	 */

	switch (vrp->vr_state) {
	    case VS_CREATENODE:
/*
TODO - retry mount?
or start unmount..
*/
		if (ep->vn_retry  && ++vrp->vr_retrycnt < 4) {
			nLogMsg(EL_ERROR2, id,
			  "resubmitting request for vsn %.*s_%d, (retry %d)",
				VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq,
				vrp->vr_retrycnt);
			nLogMsg(EL_FATAL, id, "remount not implemented");
			return(0);
		}
		if (ep->vn_retry)
			nLogMsg(EL_ERROR1, id,
				  "Retry count exceeded, vsn=%.*s_%d",
				  VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
		vInstallError(ep);
		DBFailMount((char *)&vrp->vr_vsn);
		ErrOutVolReq(vrp, id);
		return(0);
	    case VS_DELETENODE:
		/*
		 *  TODO what to do? notify ops that drive can't
		 *	be varied off? presumably leave drive
		 *	logically 'on' & let next host vary on
		 *	failure trigger off-vary.
		 *
		 *	meanwhile, tell rc to put the tape away
		 */
		UnmountDrive(vrp);
		return(1);
	    default:
		nLogMsg(EL_FATAL, id, "%.*s_%d got %s but state is %s",
			VSNLEN, &ep->vn_vsn, hp->vh_seq, 
			vHdrStr(hp), vReqStateStr(vrp->vr_state));
	} 
	return(0);
}



/*
 *  MErrorRC() - process M_VERROR pkt from RC
 */

MErrorRC(vmpHdr_t *ihp, RPRC_t *rcp)
{
	char		*id = "MErrorRC";
	VolReq_t	*vrp;
	vmpError_t	*ep;
	int		drive, medindex, unmountok = 0;
	RPRC_t		*rcp2;

	ep = MSGPTR(ihp, vmpError_t);

	/*
	 *  find the VolReq this error corresponds to
	 */
	if ((vrp = FindVolReqSeq((char *)&ep->vn_vsn, ihp->vh_seq)) == NULL) {
		nLogMsg(EL_ERROR1, id,
		 "got M_VERROR for %.*s_%d from %s but no req (%d %.*s)",
		  VSNLEN, &ep->vn_vsn, ihp->vh_seq, 
		  rcp->rc_id, ep->vn_error, LINELEN, ep->vn_msg);
		return(0);
	}

	nLogMsg(EL_STATUS1, id,			/* print error message */
	  "%s_%d M_VERROR from %s:\n err %d/%s, retry %c  msg: %.*s",
	  vsnstr((char *)&vrp->vr_vsn), vrp->vr_rpseq, rcp->rc_id, 
	  ep->vn_error, vErrStr(ep->vn_error), 
	  (ep->vn_retry ? 'y' : 'n'), 
	  LINELEN, ep->vn_msg);

	if (rcp != vrp->vr_rcp  &&  rcp != vrp->vr_dstrcp)
		nLogMsg(EL_FATAL, id, "msg from unexpected RC %s", rcp->rc_id);

	vInstallError(ep);			/* install error into globals */

	medindex = RCMedium(rcp, vVolTypeMedium(vrp->vr_info.vi_voltype));

	switch (vrp->vr_state) {

	    case VS_RCWAIT:
		/*
		 *  Mount failed..  recover RC drive 
		 */
		rcp->rc_devs[medindex].d_mountdev--;

		/*
		 *  recover user drive reservation
		 */
		FreeDriveUsed(&vrp->vr_ridp->rid_RCMed[
			rcp->rc_index * RC_MEDIA_TYPES + medindex]);

		/*
		 *  decrement rid_active
		 */
		vrp->vr_ridp->rid_active--;

		/*
		 *  Throw away req if client has canceled, but
		 *	if vn_retry bit is set (by RC)
		 *	request should be resubmitted.
		 *	Else return M_VERROR packet to user (with
		 *	code and string provied by RC) and free request
		 */
		if (!(vrp->vr_flag & VF_UNMNTREQ)  &&  ep->vn_retry) {
		    if (++vrp->vr_retrycnt > 4) {
				nLogMsg(EL_ERROR1, id,
				  "Retry count exceeded, vsn=%.*s_%d",
				  VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
		    } else {
			/*
			 *  Mount has failed and drive has been varied 
			 *	off, resubmit request by setting state
			 *	to VS_MOUNTWAIT; it will be processed by
			 *	ProcVolReqQ().
			 */

			nLogMsg(EL_ERROR2, id,
			  "resubmitting request for vsn %.*s_%d, (retry %d)",
				VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq,
				vrp->vr_retrycnt);
			vrp->vr_state = VS_MOUNTWAIT;
			vrp->vr_time = Now;
			return(1);
		    }
		}
		/* reset volume back to idle */
		DBFailMount((char *)&vrp->vr_vsn);

		ErrOutVolReq(vrp, id);		/* handles 'parallel' case */
		break;

	  case VS_RCWAIT2:

		/*
		 *  consider case where err comes on mount
		 *	after dismount has been issued
		 *	as 'dismount succeeded'.
		 */
		if (ep->vn_flag == M_RCMOUNT) {
			nLogMsg(EL_STATUS1, id, 
				"%.*s_%d: mount err on dismount %s",
				VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq, 
				" - treating as successful dismount");
			if (vrp->vr_ridp != RPRidp)
				nLogMsg(EL_ERROR1, id, "%.*s: expected RPRidp",
							VSNLEN, &vrp->vr_vsn);
			unmountok = 1;
		}

		/*
		 *  release RC drive 
		 */
		rcp->rc_devs[medindex].d_mountdev--;
		drive = vDMffs(&vrp->vr_devinuse);
		if (drive != -1)  	/* should be good.. */
			DM_CLR(drive, rcp->rc_inuse[medindex]);
        	if (vrp->vr_flag & VF_VARYOFF) {
			DevEnt_t	*dep;
                	char            *devnam;

                	/*
                	 *  VF_VARYOFF is set when someone attempts to vary
                	 *      off a device with a volume mounted on it.
                	 *      First, the volume must be unmounted (which
                	 *      just failed).  Now the vary off can
                	 *      be tried.
                	 */

			dep = vGetDevByNum(drive);
                	if (dep)
                        	devnam = dep->de_name;
                	else
                        	devnam = "???";
                	nLogMsg(EL_STATUS1, id,
                          	"Completing Vary off for dev: %s", devnam);
                	Vary(drive, VVA_VARYOFF, VOLMANUID, 1, NULL);
        	}

		/*
		 *  decrement rid_active
		 */
		vrp->vr_ridp->rid_active--;

		/*
		 *  free user drive reservation
		 */
		FreeDriveUsed(&vrp->vr_ridp->rid_RCMed[
			rcp->rc_index * RC_MEDIA_TYPES + medindex]);

		/*
		 *  update dbase
		 */
		DBUnmountDone((char *)&vrp->vr_vsn);

		if (unmountok) {
			FreeVolReq(vrp, id);
		} else {
		  	nLogMsg(EL_ERROR1, id,
		    		"error from %s unmounting %.*s_%d... %s",
		    		vrp->vr_rcp ? vrp->vr_rcp->rc_id : "RC??",
		    		VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq,
				"deleting request");
			ErrOutVolReq(vrp, id);
		}
		break;
		
	  case VS_MOVING: {
		vmpError_t	*ep2;
		vmpHdr_t	*hp;
		char		ebuf[PKTSIZ(vmpError_t)];

		/*
		 *  decrement rid_active
		 */
		vrp->vr_ridp->rid_active--;

		/*
		 *  Got an error during a move.  Possibly the database 
		 *	and physical location don't agree, or an RC
		 *	had a problem.
		 */
		if (vrp->vr_clientp)
			Err2Client(vrp->vr_clseq, vrp->vr_clientp);

		/*
		 *  build error packet for other RC & send it
		 */
		hp = vBuildHdr(M_VERROR, vrp->vr_clseq, (vmpHdr_t *) ebuf);
		ep2 = MSGPTR(hp, vmpError_t);
		ep2->vn_vsn = vrp->vr_vsn;
		ep2->vn_error = VMError;
		emsgcpy(ep2->vn_msg, nErrStr);
		if (rcp == vrp->vr_rcp)
			rcp2 = vrp->vr_dstrcp;
		else
			rcp2 = vrp->vr_rcp;
		if (rcp2->rc_srv.s_sd != -1)
			XDRSendI(hp, rcp2->rc_srv.s_sd, 1);

		DBUpdateLoc((char *)&vrp->vr_vsn, vrp->vr_rcp); /* didn't move */
		nLogMsg(EL_CONSOLE0|EL_ERROR1, id,
		  "%.*s_%d: move failed, insure that volume is in the %s",
		  VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq, 
		  vrp->vr_rcp ? vrp->vr_rcp->rc_id: "UNKNOWN RC");
		FreeVolReq(vrp, id);
		break;
	  }

	  case VS_IDLE:
	  case VS_DRIVEWAIT:
	  case VS_MOUNTED:
	  case VS_EJECTED:
	  case VS_ACCEPTED:
	  default:
	    nLogMsg(EL_ERROR1, id,
		 "unexpected M_VERROR for %.*s_%d/%s from %s ignored (%d %.*s)",
		  VSNLEN, &ep->vn_vsn, ihp->vh_seq, vReqStateStr(vrp->vr_state),
		  rcp->rc_id, ep->vn_error, LINELEN, ep->vn_msg);
		break;

	}
	return(1);
}

/*
 *  MIdleTimes() - note /dev/vol/xxxxxx idle times from volnd
 */
 void
MIdleTimes(vmpHdr_t *hp)
{
	char		*id = "MIdleTimes";
	vmpNodeTime_t	*ip = MSGPTR(hp, vmpNodeTime_t);
	int	i;

	for (i=0; i<hp->vh_count; i++, ip++) {
		VolReq_t	*vrp = FindVolReq((char *)&ip->t_vsn);
		if (vrp == NULL) {
			nLogMsg(EL_STATUS2, id, "%.*s: req not found",
					VSNLEN, &ip->t_vsn);
			continue;
		}
		if (vrp->vr_state != VS_MOUNTED) {
			nLogMsg(EL_STATUS2, id, "%.*s: req not mounted",
					VSNLEN, &ip->t_vsn);
			continue;
		}
		vrp->vr_idletime = ip->t_idle;
	}
}

/*
 *  assignvsp() - fill in vsn info for MVolQry()
 */
static void	*vsbuf = NULL;
static vmpHdr_t	*vshp = NULL;
static int	vsbufsize = 0, vsbufcount = 0;
static int	vsmsgsize, vsmsgcount;

 vmpVolStat_t *
assignvsp(vmpVolStat_t *vsp, VolReq_t *vrp, int timeflag)
{
	char	*id = "assignvsp";
 
	if (vsbufcount == vsmsgcount) {
		vsbufcount += 50;
		vsbufsize += 50 * sizeof(vmpVolStat_t);
		vsbuf = nReAlloc(vsbuf, vsbufsize);
		vshp = (vmpHdr_t *) vsbuf;
		vsp = (vmpVolStat_t *) 
			((char *)vsbuf + PadSize(sizeof(vmpHdr_t))
				+ vsmsgcount * sizeof(vmpVolStat_t));
	}
	vsp->vv_vsn = vrp->vr_vsn;
	vsp->vv_state = vrp->vr_state;
	if (timeflag) {
		if (vrp->vr_state == VS_MOUNTED)
			vsp->vv_time = vrp->vr_idletime;
		else if (vrp->vr_state == VS_IDLE)
			vsp->vv_time = 0;
		else
			vsp->vv_time = Now - vrp->vr_time;
	} else
		vsp->vv_time = Now - vrp->vr_time;
/*
TODO - maybe just send cmd? why interpret here?
*/
	if (vrp->vr_state == VS_IDLE)
		vsp->vv_use = U_IDLE;
	else if (vrp->vr_cmd == M_USRMOUNT)
		vsp->vv_use = U_MOUNT;
	else if (vrp->vr_cmd == M_UNMOUNT)
		vsp->vv_use = U_UNMOUNT;
	else if (vrp->vr_cmd == M_SCRATCH) 
		vsp->vv_use = U_SCRATCH;
	else if (vrp->vr_cmd == M_MOVE) 
		vsp->vv_use = U_MOVE;
	else 
		vsp->vv_use = 'x';
	vsp->vv_rpord = vrp->vr_rpord;
	if (vrp->vr_clientp == NULL)
		vsp->vv_pid = -1;
	else
		vsp->vv_pid = vrp->vr_clientp->c_pid;
	vsp->vv_dev = vDMffs(&vrp->vr_devinuse);
	if (vsp->vv_dev > -1)
		vsp->vv_host = DriveArray[vsp->vv_dev].de_host;
	else
		vsp->vv_host = -1;
	vsp->vv_uid = vrp->vr_uid;
	vsp->vv_rc = (vrp->vr_rcp ? vrp->vr_rcp->rc_number : -1);
	vsp->vv_dstrc = (vrp->vr_dstrcp ? vrp->vr_dstrcp->rc_number : -1);
	vsp->vv_flags = VV_UID;		/* TODO - figure out what this does */
	if (vrp->vr_state != VS_IDLE)
		vsp->vv_flags |= VV_RC;
	memcpy(vsp->vv_rid, vrp->vr_ridp->rid_id, sizeof(vsp->vv_rid));
	copytag(vsp->vv_tag, vrp->vr_tag);

	vsp++;
	vsmsgcount++;
	vsmsgsize += sizeof(vmpVolStat_t);

	return(vsp);
}

/*
 *  MVolQry(): process M_VOLQRY packets received by the RP
 *	return volume mount/move status
 */
extern QHead_t	RidQ;
 void
MVolQry(vmpHdr_t *ihp, Client_t *cp)
{
	char		*id = "MVolQry";
	vmpVsn_t	*vp;
	char		*vsnp;
	RPRC_t		*rcp;
	vmpVolStat_t	*vsp;
	int		rc;
	Rid_t		*ridp;
	VolReq_t	*vrp;

	if (VRHEAD(VolReqQ) == NULL) {
		vmpHdr_t	hdr;

		/*
		 *  no volumes in system - send 'empty' reply
		 */
		hdr.vh_magic = VMP_MAGIC;
		hdr.vh_type = M_VOLSTATS;
		hdr.vh_size = 0;
		hdr.vh_seq = ihp->vh_seq;

		Snd2Client(&hdr, cp, 1);
		return;
	}

	/*
	 *  start parsing pkt
	 */
	vp = MSGPTR(ihp, vmpVsn_t);
	vsnp = (char *)&vp->v_vsn;

	/*
	 *  make sure there's a buffer; start w/ one sized for 50
	 */
	if (vsbuf == NULL) {
		vsbufcount = 50;
		vsbufsize = PadSize(sizeof(vmpHdr_t)) + 
				vsbufcount * sizeof(vmpVolStat_t);
		vsbuf = (void *) nAlloc(vsbufsize);
		vshp = (vmpHdr_t *) vsbuf;
		vshp->vh_magic = VMP_MAGIC;
		vshp->vh_type = M_VOLSTATS;
	}

	/*
	 *  tailor pkt header for client
	 */
	vshp->vh_seq = ihp->vh_seq;
	vsp = MSGPTR(vshp, vmpVolStat_t);

	vsmsgcount = 0;
	vsmsgsize = 0;

	switch (vp->v_flag) {
	    case VQ_VSN:
		nLogMsg(EL_DEBUG2, id, "%.*s", VSNLEN, vsnp);
		for (vrp=FindVolReq(vsnp); vrp; vrp=VRNEXT(vrp->vr_vsnq)) {
			vsp = assignvsp(vsp, vrp, vp->v_flag2);
		}
		break;

	    case VQ_RC:
		rc = vp->v_rc;
		if ((rcp = RCInfo(rc)) == NULL) {
			VMError = E_USER;
			nLogMsg(EL_ERROR2|EL_CODE, id, "%d: bad RC", rc);
			Err2Client(ihp->vh_seq, cp);
			return;
		}
		for (ridp=(Rid_t *)RidQ.q_head; ridp; 
					ridp=(Rid_t *)ridp->rid_q.q_next){
			for (vrp=VRHEAD(ridp->rid_vsnq); vrp; 
					vrp=VRNEXT(vrp->vr_ridq)) {
				/* return status for all volreqs on rcp */
				if (vrp->vr_rcp == rcp) {
					vsp = assignvsp(vsp, vrp, vp->v_flag2);
				}
			}
		}
		break;

	    case VQ_UID:	/* return status for client uid's vsns */
		for (ridp=(Rid_t *)RidQ.q_head; ridp; 
					ridp=(Rid_t *)ridp->rid_q.q_next){
			if (ridp->rid_uid == cp->c_uid) {
				for (vrp=VRHEAD(ridp->rid_vsnq); vrp; 
						vrp=VRNEXT(vrp->vr_ridq)) {
					PrintVolReq(vrp, EL_STATUS1, NULL);
					vsp = assignvsp(vsp, vrp, vp->v_flag2);
				}
			}
		}
		break;

	    case VQ_ALL:	/* return status for all vsns */
		for (ridp=(Rid_t *)RidQ.q_head; ridp; 
					ridp=(Rid_t *)ridp->rid_q.q_next){
			for (vrp=VRHEAD(ridp->rid_vsnq); vrp; 
					vrp=VRNEXT(vrp->vr_ridq)) {
				vsp = assignvsp(vsp, vrp, vp->v_flag2);
			}
		}
		break;

	    default:
		VMError = E_USER;
		nLogMsg(EL_ERROR2|EL_CODE, id,
				"%d: unknown VolQry flag", vp->v_flag);
		Err2Client(ihp->vh_seq, cp);
		return;
	}
	vshp->vh_count = vsmsgcount;
	vshp->vh_size = vsmsgsize;
	Snd2Client(vshp, cp, 1);			/* send M_VOLSTAT */
}

/*
 *  MAttach() - attach to request in progress; can select mount or move
 *	or take the most advanced
 */
 void
MAttach(vmpHdr_t *ihp, Client_t *cp)
{
	char		*id = "MAttach";
	vmpVsn_t	*vp;
	char		*vsnp;
	VolReq_t	*vrp, *vrp2;
	char		*errmsg;

	vp = MSGPTR(ihp, vmpVsn_t);
	vsnp = (char *)&vp->v_vsn;

	nLogMsg(EL_DEBUG1, id, "%.*s pid %d", VSNLEN, vsnp, cp->c_pid);

	if (vp->v_flag != 0  &&  
	    vp->v_flag != M_USRMOUNT  &&  vp->v_flag != M_MOVE) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR2, id, "bad flag");
		Err2Client(ihp->vh_seq, cp);
	}

	/*
	 *  client can wait for VSN status as long as
	 *	no other client is currently doing so - choose
	 *	vrp that is furthest to execution or completed
	 */
	vrp2 = NULL;
/*
TODO - how to prioritize these err msgs & implement?
*/
	errmsg = "no request to attach";
	for (vrp=FindVolReq(vsnp); vrp; vrp=VRNEXT(vrp->vr_vsnq)) {
		nLogMsg(EL_DEBUG3, id, "prob: %s", errmsg);
		nLogMsg(EL_DEBUG3, id, "%.*s_%d ", VSNLEN, &vrp->vr_vsn, 
							vrp->vr_rpseq);
		if (vrp == vrp2)
			kill(getpid(), 5);	/* check for infinite looping */
		if (vp->v_flag  &&  vrp->vr_cmd != vp->v_flag) {
			errmsg = "request found, not requested type";
			continue;
		}
		if (vrp->vr_uid != cp->c_uid) {
			errmsg = "request belongs to other uid";
			continue;
		}
		if (vrp->vr_state == VS_IDLE) {
			errmsg = "request idle";
			continue;
		}
		if (vrp->vr_state == VS_RCWAIT2) {
			errmsg = "request unmounting";
			continue;
		}
		if (vrp->vr_clientp != NULL  &&  vrp->vr_clientp != cp) {
			errmsg = "request attached to other client";
			continue;
		}
		if (vrp->vr_clientp == cp  &&  vrp->vr_voflag & VO_WAIT  &&
					      !(vrp->vr_voflag & VO_ASYNC)) {
			errmsg = "you are already attached";
			continue;
		}
		if (vrp->vr_flag & VF_UNMNTREQ) {
			errmsg = "mount in progress canceled";
			continue;
		}
		if (vrp->vr_state == VS_UNMOUNTWAIT) {
			errmsg = "unmntwait";
			continue;
		}

		/*
		 *  got a live one; see if it's the 'best'
		 */
		if (vrp2 == NULL) {
			/*
			 *  something is better than nothing
			 */
			nLogMsg(EL_DEBUG3, id, "something");
			vrp2 = vrp;
			continue;
		}
		if (vrp->vr_state == VS_RCWAIT  ||  
		    vrp->vr_state == VS_MOUNTED ||
		    vrp->vr_state == VS_MOVING) {
			/*
			 *  active is perfect
			 */
			nLogMsg(EL_DEBUG3, id, "got perfect match");
			vrp2 = vrp;
			break;
		}
		if (vrp->vr_rpord < vrp2->vr_rpord) {
			/*
			 *  earlier waiting req somewhat arbitrarily
			 *	preferred over more recent one
			 */
			nLogMsg(EL_DEBUG3, id, "better");
			vrp2 = vrp;
		}
	}
	if (vrp2 == NULL) {
		VMError = E_USER;
		nLogMsg(EL_ERROR2|EL_CODE, id, "%.*s: %s", 
					VSNLEN, vsnp, errmsg);
		Err2Client(ihp->vh_seq, cp);
		return;
	}
	nLogMsg(EL_STATUS2, id, "%.*s_%d pid %d", VSNLEN, vsnp, 
						vrp2->vr_rpseq, cp->c_pid);
	/*
	 *  install new seq #
	 */
	vrp2->vr_voflag |= VO_WAIT;
	vrp2->vr_clseq = ihp->vh_seq;
	if (vrp2->vr_clientp == NULL) {
		/*
		 *  new client
		 */
		Host_t	*hop = IPCHost(cp);

		/* TODO - if vsn already mounted & host changes,
			have to remove node on 1st host, create on 2nd */
		if (vrp2->vr_clhostp != hop)
			nLogMsg(EL_STATUS1, id, 
				"switching clnt hosts from %s to %s, state %s",
				vrp2->vr_clhostp->h_name, hop->h_name,
				vReqStateStr(vrp2->vr_state));
		vrp2->vr_clhostp = hop;
		vrp2->vr_clhost = hop->h_id;
		EnQVolCVR(vrp2, &cp->c_cvrq);
		vrp2->vr_clientp = cp;
	}

	if (vrp2->vr_state != VS_MOUNTED) {
		/*
		 *  request is in an intermediate state - nothing
		 *	to report
		 */
		return;
	}
	/*
	 *  report mount accomplished (note that an accomplished
	 *	move would not still be sitting waiting for
	 *	client to reconnect)
	 */
	Volstat2Client(vrp2);
}

/*
 *  CancelMove(): called to deal with a move-in-progress when an RC goes
 *	down.  Sends a M_RCMSG/M_VERROR to the surviving RC.  Error code
 *	and string are pre-set by caller.
 */
 void
CancelMove(VolReq_t *vrp, RPRC_t *badrcp)
{
	char		*id = "CancelMove";
	vmpRcMsg_t	rcmsg;
	RPRC_t		*notifyrcp;
	vmpHdr_t	*hp;
	vmpError_t	*ep;

	if (vrp->vr_state != VS_MOVING) {
		nLogMsg(EL_ERROR1, id, "INTERNAL ERROR: called with state = %s",
		  vReqStateStr(vrp->vr_state));
		return;
	}
	if (vrp->vr_rcp != badrcp)
		notifyrcp = vrp->vr_rcp;
	else 
		notifyrcp = vrp->vr_dstrcp;
	if (notifyrcp->rc_srv.s_sd == -1) {
		/*
		 *  if other rc down, this move 'should' have been cancelled */
		nLogMsg(EL_ERROR1, id, 
			"INTERNAL ERROR: rc2notify = %d, rcp = %x",
		  notifyrcp->rc_number, notifyrcp);
		return;
	}

	/* set up M_RCMSG with M_VERROR inside */
	memset((char *) &rcmsg, '\0', sizeof (rcmsg));
	rcmsg.rm_rc = notifyrcp->rc_number;
	rcmsg.rm_flag = RCM_NORPFORWARD;
	rcmsg.rm_vsn = vrp->vr_vsn;
	hp = vBuildHdr(M_VERROR, NOSEQ, (vmpHdr_t *)rcmsg.rm_msg);
	ep = MSGPTR(hp, vmpError_t);
	/* M_VERROR message code and string is setup by caller */
	ep->vn_vsn = vrp->vr_vsn;
	ep->vn_state = VS_MOVING;
	ep->vn_error = VMError;
	emsgcpy(ep->vn_msg, nErrStr);
	Send2RC(notifyrcp, vrp, M_RCMSG, &rcmsg);
}

/*
 *  CreateNode() - RP does this if RC not configged to
 *	or if it started to delete a node & a new mount
 *	came in. probably NOT TESTED
 */
CreateNode(VolReq_t *vrp)
{
	char	*id = "CreateNode";
	int	drive = vDMffs(&vrp->vr_devinuse);

	/*
	 *  send msg w/ 'expectvolreq' flag=1
	 */
	if (SendNode(M_CREATENODE, vrp, drive, 1))
		return(VMError);

	/*
	 *  update state
	 */
	vrp->vr_state = VS_CREATENODE;
	DriveArray[drive].de_status |= DS_CREATENODE;

	return(0);
}

/*
 *  DeleteNode() - start unmount by deleting /dev/vol/xxxxxx
 *	user node on user's host.
 */
DeleteNode(VolReq_t *vrp, int expectvolreq)
{
	char	*id = "DeleteNode";
	int	drive = vDMffs(&vrp->vr_devinuse);

	nLogMsg(EL_STATUS1, 0, "DELETE node %.*s_%d on %s",
			VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq,
			vrp->vr_ndhostp->h_name);
	vrp->vr_usrtime = Now - vrp->vr_time;
	vrp->vr_time = Now;

	/*
	 *  send msg
	 */
	if (SendNode(M_DELETENODE, vrp, drive, expectvolreq)) {
		if (VMError == E_UNAVAIL) {
			/*
			 *  volnd down - go ahead & unmount
			 */
			nLogMsg(EL_STATUS1, id, "send failed - unmounting");
			DriveArray[drive].de_status &= ~DS_DELETENODE;
			UnmountDrive(vrp);
		} else
			nLogMsg(EL_STATUS1, id, " send failed - wait/retry");
		return(VMError);
	}
	nLogMsg(EL_DEBUG1, id, " send ok");

	/*
	 *  update state
	 */
	vrp->vr_state = VS_DELETENODE;
	DriveArray[drive].de_status &= ~DS_CREATENODE;
	DriveArray[drive].de_status |= DS_DELETENODE;

	return(0);
}

MScratch(vmpHdr_t *hp, Client_t *cp)
{
	char		*id = "MScratch";
	vmpNode_t	*np = MSGPTR(hp, vmpNode_t);
	char		*vsnp;
	VDR_t		*vdrp;
	VInfo_t		*vip;
	RPRC_t		*rcp;			
	DevMap_t	devtotry;
	int		medindex;
	VolReq_t	*vrp;

	np = MSGPTR(hp, vmpNode_t);
	vsnp = (char *)&np->vs_vsn;

	if (np->vs_flag  &&  np->vs_flag != VO_BYPASS) {
		VMError = E_USER;
		nLogMsg(EL_ERROR1|EL_CODE, id, "unexpected flag(s) 0x%x",
							np->vs_flag);
		Err2Client(hp->vh_seq, cp);
		return(0);
	}

	/*
	 *  check write permission & status in dbase for scratchability
	 */
	vdrp = DBVsnQuery(vsnp);
	if (vdrp == NULL) {
		Err2Client(hp->vh_seq, cp);
		return(0);
	}
#ifndef PURIFY
	if (!nAccess(1, cp->c_uid, vdrp->vdr_uid, vdrp->vdr_gid, 
							vdrp->vdr_mode)) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR2, id, "Permission denied");
		Err2Client(hp->vh_seq, cp);
		return(0);
        }
#endif
	if (vdrp->vdr_status != S_VIRGIN  &&
	    vdrp->vdr_status != S_RECYCLE  &&
	    vdrp->vdr_status != S_RECYCLEBAD) {
		VMError = E_USER;
		nLogMsg(EL_ERROR1|EL_CODE, id, "%s: state %s, can't scratch",
					vsnstr(vsnp), 
					vVolStateStr(vdrp->vdr_status));
		Err2Client(hp->vh_seq, cp);
		return(0);
	}

	/*
	 *  make sure access will be exclusive (safety measure)
	 */
	if (FindVolReq(vsnp)) {
		VMError = E_UNAVAIL;
		nLogMsg(EL_ERROR2|EL_CODE, id, 
				"%s: can't get exclusive access: other request",
				vsnstr(vsnp));
		Err2Client(hp->vh_seq, cp);
		return(0);
	}

	/*
	 *  get RC
	 */
	vip = &vdrp->vdr_info;
	rcp = RCInfo(vip->vi_srcrc);
	if (rcp == NULL) {
		VMError = E_INTERNAL;
		nLogMsg(EL_ERROR1, id, "%s: RC %d not defined", 
						vsnstr(vsnp), vip->vi_srcrc);
		Err2Client(hp->vh_seq, cp);
		return(0);
	}

	/*
	 *  check 'mountability' of rc
	 */
	if (!(rcp->rc_flags & RCF_MOUNTOK)) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR2, id, "can't mount in %s", rcp->rc_id);
		Err2Client(hp->vh_seq, cp);
		return(0);
	}

	/*
	 *  see if any drives configged
	 */
	medindex = RCMedium(rcp, vVolTypeMedium(vip->vi_voltype));
	if (medindex == -1) {
		VMError = E_INTERNAL;
		nLogMsg(EL_CODE|EL_ERROR1, id, "can't get RCMedium");
		Err2Client(hp->vh_seq, cp);
		return(0);
	}
	if (rcp->rc_devs[medindex].d_nconfigdev == 0) {
		VMError = E_ADMIN;
		nLogMsg(EL_CODE|EL_ERROR2|EL_CONSOLE0, id,
				"Can't mount %s: RC %s has no %s drives",
				vsnstr(vsnp), rcp->rc_id,
				vMediumName(rcp->rc_devs[medindex].d_medium));
		Err2Client(hp->vh_seq, cp);
		return(0);
	}
	devtotry = rcp->rc_devs[medindex].d_devmap;

	/*
	 *  make sure RC is up
	 */
	if (rcp->rc_srv.s_sd == -1) {
		VMError = E_UNAVAIL;
		nLogMsg(EL_CONSOLE0|EL_CODE|EL_ERROR2, id, 
					"Can't mount %s: RC %s down",
					vsnstr(vsnp), rcp->rc_id);
		Err2Client(hp->vh_seq, cp);
		return(0);
	}

	/*
	 *  get VolReq (makes default rid & hooks it up w/ client struct)
	 */
	vrp = SetupRequest(vsnp, cp, VO_WAIT, hp->vh_seq, NULL);
	if (vrp == NULL) {
		Err2Client(hp->vh_seq, cp);
		return(0);
	}

	/*
	 *  fill in other request data in VolReq per this request
	 */
	vrp->vr_cmd = M_SCRATCH;
	vrp->vr_info = *vip;			/* copy vinfo / MountQuery() */
	vrp->vr_retrycnt = 0;
	vrp->vr_clseq = hp->vh_seq;		/* client seq no from header */
	DM_ZERO(vrp->vr_devinuse);		/* no device in use yet */
	vrp->vr_rcp = rcp;			/* save pointer to RC struct */
	vrp->vr_devtotry = devtotry;
	vrp->vr_voflag = VO_WAIT|VO_WRITE|VO_SCRATCH | np->vs_flag;
	/* remember what to do in case of error */
	if (vdrp->vdr_status == S_RECYCLEBAD)
		vrp->vr_voflag |= VO_TRASH;
	else
		vrp->vr_voflag |= VO_RECYCLEBAD;
	copytag(vrp->vr_tag, "scratch");

	vrp->vr_state = VS_MOUNTWAIT;
	vrp->vr_time = Now;

	/*
	 *  leave request in queue for resources
	 */
	ENQUEUE(vrp, &VolReqQ, vr_vrq, VolReq_t)
	nLogMsg(EL_STATUS1, 0, 
			"MOUNTQ/SCRATCH %s_%d sync pid %d uid %d %s",
			vsnstr(vsnp), vrp->vr_rpseq, 
			cp->c_pid, cp->c_uid,
			(vrp->vr_voflag & VO_BYPASS ? "nlbl" : "") );
	return(1);
}

 void
ScratchErr(VolReq_t *vrp)
{
	char		*vsnp = (char *)&vrp->vr_vsn;
	RPRC_t		*rcp = vrp->vr_rcp;
	VInfo_t		*vip = &vrp->vr_info;
	int		medindex, i, obsolete = 0;

	/*
	 *  release RC drive 
	 */
	medindex = RCMedium(rcp, vVolTypeMedium(vip->vi_voltype));
	rcp->rc_devs[medindex].d_mountdev--;

	/*
	 *  decrement rid_active
	 */
	vrp->vr_ridp->rid_active--;

	/*
	 *  free user drive reservation
	 */
	FreeDriveUsed(&vrp->vr_ridp->rid_RCMed[
			rcp->rc_index * RC_MEDIA_TYPES + medindex]);

	/*
	 *  adjust status toward extinction
	 */
	if (vrp->vr_voflag & VO_RECYCLEBAD)
		DBRecyclebad(vsnp);
	else if (vrp->vr_voflag & VO_TRASH)
		DBTrash(vsnp);
}
