
/*
 *  rp_3.c: Volume Manager Request Processing of vsn requests involving
 *	only database operations. 
 *
 *	Mostly, check permission, make sure nothing is going on w/
 *	volume if dbase is to be modified, & pass on to dbase routine.
 *
 *	rp_1.c	main, packet proc switch to routines, housekeeping
 *	rp_2.c	packet proc: mounting & moving volumes
 *	rp_3.c	packet proc: reporting/changing dbase info
 *	rp_4.c	packet proc: varying drives
 *	rp_5.c	packet proc: reserving drives & volumes
 */

#include "rp_space.h"

extern int	IsChild;	/* for fork on long vls response */

/*
 *  MAddVsn() - admin put new vsn in dbase
 */
 void
MAddVsn(vmpHdr_t *hp, Client_t *cp)
{
	char		*id = "MAddVsn";
	vmpAdd_t	*ap = MSGPTR(hp, vmpAdd_t);
	char		mbuf[PKTSIZ(vmpVsn_t)];
	vmpHdr_t	*rhp;
	vmpVsn_t	*vp;

	nLogMsg(EL_DEBUG2, id, "uid %d  mode %o  status %c", 
		ap->vad_vinfo.vi_uid, ap->vad_vinfo.vi_mode, ap->vad_status);

	/*
	 *  is user special enough?
	 */
	if (CheckSpecialPerm(cp, id)) {
		Err2Client(hp->vh_seq, cp);
		return;
	}

	if (DBAddVsn(&ap->vad_vinfo, ap->vad_status)) {
		Err2Client(hp->vh_seq, cp);
		return;
	}
	rhp = vBuildHdr(M_VSN, hp->vh_seq, (vmpHdr_t *) mbuf);
	vp = MSGPTR(rhp, vmpVsn_t);
	vp->v_vsn = ap->vad_vinfo.vi_vsn;
	vp->v_flag = 0;
	Snd2Client(rhp, cp, 1);
}

/*
 *  MDeleteVsn() - admin wipe vsn from dbase
 */
 void
MDeleteVsn(vmpHdr_t *hp, Client_t *cp)
{
	char		*id = "MDeleteVsn";
	vmpVsn_t	*vp = MSGPTR(hp, vmpVsn_t);
	VolReq_t	*vrp;

	nLogMsg(EL_DEBUG2, id, "vsn %.*s", VSNLEN, &vp->v_vsn);

	if (CheckSpecialPerm(cp, id)) {
		Err2Client(hp->vh_seq, cp);
		return;
	}

	/*
	 *  can't delete a vsn in play
	 */
	if (vrp = FindVolReq((char *)&vp->v_vsn)) {
		VMError = E_UNAVAIL;
		nLogMsg(EL_CODE|EL_ERROR2, id, "Denied: %.*s_%d is in use", 
					VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
		Err2Client(hp->vh_seq, cp);
		return;
	}
	if (DBDeleteVsn((char *)&vp->v_vsn, &vp->v_rc, vp->v_flag == VQ_FORCE)) {
		Err2Client(hp->vh_seq, cp);
		return;
	}
	Snd2Client(hp, cp, 1);
}

/*
 *  MVattr() - admin set various arributes of vsn
 */
 void
MVattr(vmpHdr_t *hp, Client_t *cp)
{
	char		*id = "MVattr";
	vmpAttr_t	*ap = MSGPTR(hp, vmpAttr_t);
	VolReq_t	*vrp;

	/*
	 *  is client qualified to do plastic surgery?
	 */
	if (CheckSpecialPerm(cp, id)) {
		Err2Client(hp->vh_seq, cp);
		return;
	}
	/*
	 *  can't modify a vsn in play
	 */
	if (vrp = FindVolReq((char *)&ap->a_vsn)) {
		VMError = E_UNAVAIL;
		nLogMsg(EL_CODE|EL_ERROR2, id, "Denied: %.*s_%d is in use", 
					VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
		Err2Client(hp->vh_seq, cp);
		return;
	}
	if (DBVattr(ap)) {
		Err2Client(hp->vh_seq, cp);
		return;
	}
	Snd2Client(hp, cp, 1);
}

/*
 *  CheckSpecialPerm() - admin privileges for root or volman
 */
 int
CheckSpecialPerm(Client_t *cp, char *id)
{
	if (cp->c_uid != VolManOperUID  &&  cp->c_uid != 0) {
		VMError = E_PERM;
		nLogMsg(EL_CODE|EL_ERROR2, id,
		  "UID %d: insufficient permission", cp->c_uid);
		return(1);
	}
	return(0);
}


/*
 *  MVAlloc() - allocate a scratch volume to a user
 *	let DBAllocateScratch() do all the validation
 */
 void
MVAlloc(vmpHdr_t *hp, Client_t *cp)
{
	vmpVAlloc_t	*vap;
	vsn_t		*vsnp;
	vmpHdr_t	*rhp;
	vmpVsn_t	*vp;
	char		pbuf[PKTSIZ(vmpVsn_t)];

	vap = MSGPTR(hp, vmpVAlloc_t);
	vsnp = DBAllocateScratch(vap->va_rc, vap->va_poolgid, vap->va_type, 
					cp->c_uid, vap->va_gid,
	  				vap->va_mode, vap->va_desc);
	if (vsnp == NULL) {
		(void) Err2Client(hp->vh_seq, cp);
		return;
	} 
	nLogMsg(EL_STATUS1, 0, "ALLOCATED %.*s to uid %d pid %d", 
				VSNLEN, &vsnp->v_vsn, cp->c_uid, cp->c_pid);

	rhp = vBuildHdr(M_VSN, hp->vh_seq, (vmpHdr_t *)pbuf);
	vp = MSGPTR(rhp, vmpVsn_t);
	vp->v_vsn = *vsnp;
	Snd2Client(rhp, cp, 1);		/* send M_VSN */
}


/*
 *  MRecycle() - recycle a volume for a user
 */
 void
MRecycle(vmpHdr_t *hp, Client_t *cp)
{
	vmpVsn_t	*vsnp;
	VolReq_t	*vrp;

	vsnp = MSGPTR(hp, vmpVsn_t);
	if (vrp = FindVolReq((char *)&vsnp->v_vsn)) {
		VMError = E_UNAVAIL;
		nLogMsg(EL_CODE|EL_ERROR2, 0, "Denied: %.*s_%d is in use", 
					VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
		(void) Err2Client(hp->vh_seq, cp);
		return;
	}
	if (DBRecycle((char *)&vsnp->v_vsn, cp->c_uid, &vsnp->v_rc, 
								vsnp->v_flag)) {
		(void) Err2Client(hp->vh_seq, cp);
		return;
	}
	nLogMsg(EL_STATUS1, 0, "RECYCLED (%s) %.*s uid %d pid %d", 
			(vsnp->v_flag ? "bad" : "good"),
			VSNLEN, &vsnp->v_vsn, cp->c_uid, cp->c_pid);
	Snd2Client(hp, cp, 1);		/* send pkt as ack */
}

/*
 *  MVinfo() - look up 1 vsn for volcd client
 */
 void
MVinfo(vmpHdr_t *hp, Client_t *cp)
{
	char		*id = "MVinfo";
	VDR_t		*vdr;
	vmpVls_t	*vp;
	char		*vsn;
	vmpHdr_t	*rhp;
	vmpVdr_t	*vr;
	char		pbuf[PKTSIZ(vmpVdr_t)];
	int		i, pid, ret;

	/*
	 *  prepare reply pkt
	 */
	rhp = vBuildHdr(M_VDR, hp->vh_seq, (vmpHdr_t *)pbuf);
	vr = MSGPTR(rhp, vmpVdr_t);

	vp = MSGPTR(hp, vmpVls_t);
	vsn = (char *) &vp->l_vsn;
	if (!vsn[0]) {
		VMError = E_USER;
		nLogMsg(EL_CODE, id, "no vsn defined");
		Err2Client(hp->vh_seq, cp);
		return;
	}
	vdr = DBVsnQuery(vsn);
	if (!vdr) 
		Err2Client(hp->vh_seq, cp);
	else {
		vr->vv_vdr = *vdr;
		Snd2Client(rhp, cp, 1);
	}
}

/*
 *  MVslook() - look up vsn(s) for user (console)
 *	Note: for the case where multiple M_VDR pkts
 *	go out terminated by M_ACK or M_VERROR, the XDR
 *	record buffering is used.
 */
 void
MVslook(vmpHdr_t *hp, SrvProc_t *sp)
{
	char		*id = "MVslook";
	VDR_t		*vdr;
	vmpVls_t	*vp;
	char		*vsn;
	vmpHdr_t	*rhp;
	vmpVdr_t	*vr;
	char		pbuf[PKTSIZ(vmpVdr_t)];
	int		i, pid, ret;

	/*
	 *  prepare reply pkt
	 */
	rhp = vBuildHdr(M_VDR, hp->vh_seq, (vmpHdr_t *)pbuf);
	vr = MSGPTR(rhp, vmpVdr_t);

	vp = MSGPTR(hp, vmpVls_t);
	vsn = (char *) &vp->l_vsn;
	if (vsn[0]) {

		/*
		 *  look up a single vsn
		 */
		
		vdr = DBVsnQuery(vsn);
		if (!vdr) 
			Err2Srv(hp->vh_seq, sp);
		else {
			vr->vv_vdr = *vdr;
			Snd2Srv(rhp, sp, 1);
		}
		return;
	}

	/*
	 *  look up by uid & other criteria - 1st 20 done by this process
	 */

	for (i=0; ; i++) {
		vdr = DBVsnUidQuery(vp->l_uid, i);
		if (!vdr) {
			if (VMError == E_NOERROR)
				AckSrv(hp->vh_seq, sp);
			else
				Err2Srv(hp->vh_seq, sp);
			return;
		}

		/*
		 *  apply any other screening
		 */
		if (vp->l_media) {
			int	med = vVolTypeMedium(vdr->vdr_voltype);
			if (!(med & vp->l_media))
				continue;
		}
		if (vp->l_rc1) {
			/*
			 *  check RCs
			 */
			int	skip = 1;
			if (vp->l_rc1 == vdr->vdr_rc)
				skip = 0;
			else if (vp->l_rc2 == vdr->vdr_rc)
				skip = 0;
			else if (vp->l_rc3 == vdr->vdr_rc)
				skip = 0;
			if (skip)
				continue;
		}
		if (vp->l_status) {
			if ( !(vp->l_status & vStatusToBit(vdr->vdr_status)) )
				continue;
		}
		vr->vv_vdr = *vdr;
		if (i < 19) {
			/*
			 *  buffered send
			 */
			if (Snd2Srv(rhp, sp, 0)) 
				return;
		} else {
			/*
			 *  on 20th, flush & fork to continue
			 */
			if (Snd2Srv(rhp, sp, 1)) 
				return;
			break;
		}
	}

	/*
	 *  more than 20 vsns - copy uid index & continue as 
	 *	another process. A long vls ties up the voldaemon
	 *	unacceptably, and there is no obvious way
	 *	to share the index between two processes.
	 */

	nLogMsg(EL_DEBUG1, id, "more than 20 vsns: forking");

	if (DBCloseForUidQuery(sp->s_pid)) {
		Err2Srv(hp->vh_seq, sp);
		DBReopenForUidQuery(1);
		return;
	}
	if ( pid = fork() ) {

		/*
		 *  'real' voldaemon (pid = child's)
		 */
		DBReopenForUidQuery(1);
		nLogMsg(EL_STATUS2, id, 
			"Child %d forked to complete the look up", pid);
		return;
	}

	/*
	 *  Child: every way out of this routine from here on
	 *	should be an exit - no returns; sleep before
	 *	exit so that msgs have a chance to flush thru
	 */

	/*
	 *  set up to not shut down on error,
	 *	and not log every message
	 */
	IsChild++;
	nLogSetLevel(EL_STATUS2);

	/*
	 *  reset dbase
	 */
	ret = DBReopenForUidQuery(0);
	if (ret > 0) {
		Err2Srv(hp->vh_seq, sp);
		DBCleanupVls();
		sleep(5);
		exit(2);
	} else if (ret < 0) {
		AckSrv(hp->vh_seq, sp);
		DBCleanupVls();
		sleep(5);
		exit(0);
	}

	/*
	 *  send the rest of the vsns to client
	 */
	for (;;i++) {
		vdr = DBVsnUidQuery(vp->l_uid, i);
		if (!vdr) {
			if (VMError == E_NOERROR)
				AckSrv(hp->vh_seq, sp);
			else
				Err2Srv(hp->vh_seq, sp);
			break;
		}
		/*
		 *  apply any other screening
		 */
		if (vp->l_media) {
			int	med = vVolTypeMedium(vdr->vdr_voltype);
			if (!(med & vp->l_media))
				continue;
		}
		if (vp->l_rc1) {
			/*
			 *  check RCs
			 */
			int	skip = 1;
			if (vp->l_rc1 == vdr->vdr_rc)
				skip = 0;
			else if (vp->l_rc2 == vdr->vdr_rc)
				skip = 0;
			else if (vp->l_rc3 == vdr->vdr_rc)
				skip = 0;
			if (skip)
				continue;
		}
		if (vp->l_status) {
			if ( !(vp->l_status & vStatusToBit(vdr->vdr_status)) )
				continue;
		}
		/*
		 *  buffered send - ack will flush
		 */
		vr->vv_vdr = *vdr;
		if (Snd2Srv(rhp, sp, 0))
			break;
	}
	DBCleanupVls();
	sleep(5);
	exit(0);
}

/*
 *  MChmod() - change permissions on a volume for user
 */
 void
MChmod(vmpHdr_t *hp, Client_t *cp)
{
	vmpVsn_t	*vsnp;
	VolReq_t	*vrp;

	vsnp = MSGPTR(hp, vmpVsn_t);
	if (vrp = FindVolReq((char *)&vsnp->v_vsn)) {
		VMError = E_UNAVAIL;
		nLogMsg(EL_CODE|EL_ERROR2, 0, "Denied: volume %.*s_d in use", 
					VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
		(void) Err2Client(hp->vh_seq, cp);
		return;
	}
	if (DBChmod((char *)&vsnp->v_vsn, cp->c_uid, vsnp->v_flag, 
							vsnp->v_flag2)) {
		(void) Err2Client(hp->vh_seq, cp);
		return;
	}
	Snd2Client(hp, cp, 1);		/* send pkt as ack */
}

/*
 *  MChown() - change owner of a volume for user
 */
 void
MChown(vmpHdr_t *hp, Client_t *cp)
{
	vmpVsn_t	*vsnp;
	VolReq_t	*vrp;

	vsnp = MSGPTR(hp, vmpVsn_t);
	/*
	 *  make sure nothing going on w/ vsn
	 */
	if (vrp = FindVolReq((char *)&vsnp->v_vsn)) {
		VMError = E_UNAVAIL;
		nLogMsg(EL_CODE|EL_ERROR2, 0, "Denied: volume %.*s_d in use", 
					VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
		(void) Err2Client(hp->vh_seq, cp);
		return;
	}
	if (DBChown((char *)&vsnp->v_vsn, vsnp->v_rc, cp->c_uid)) {
		(void) Err2Client(hp->vh_seq, cp);
		return;
	}
	Snd2Client(hp, cp, 1);		/* send pkt as ack */
}

/*
 *  MChgrp() - change group of a volume for user
 */
 void
MChgrp(vmpHdr_t *hp, Client_t *cp)
{
	vmpVsn_t	*vsnp;
	VolReq_t	*vrp;

	vsnp = MSGPTR(hp, vmpVsn_t);
	/*
	 *  make sure nothing going on w/ vsn
	 */
	if (vrp = FindVolReq((char *)&vsnp->v_vsn)) {
		VMError = E_UNAVAIL;
		nLogMsg(EL_CODE|EL_ERROR2, 0, "Denied: volume %.*s_d in use", 
					VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
		(void) Err2Client(hp->vh_seq, cp);
		return;
	}
	if (DBChgrp((char *)&vsnp->v_vsn, vsnp->v_rc, cp->c_uid)) {
		(void) Err2Client(hp->vh_seq, cp);
		return;
	}
	Snd2Client(hp, cp, 1);		/* send pkt as ack */
}

/*
 *  MLabelQuery() - look up vsn by internal label & uid for user
 */
 void
MLabelQuery(vmpHdr_t *hp, Client_t *cp)
{
	vmpLblQry_t	*lp = MSGPTR(hp, vmpLblQry_t);

	if (DBLabelQuery((char *)&lp->vq_vsn, (char *)&lp->vq_label, cp->c_uid, 
							lp->vq_obsolete)) {
		(void) Err2Client(hp->vh_seq, cp);
	} else {
		Snd2Client(hp, cp, 1);		/* send back filled in msg */
	}
}

/*
 *  MGetQuota() - look up quota in RC for user
 */
 void
MGetQuota(vmpHdr_t *hp, Client_t *cp)
{
	if (DBQuotaQuery(MSGPTR(hp, vmpQuota_t)))
		(void) Err2Client(hp->vh_seq, cp);
	else {
		hp->vh_type = M_QUOTRPLY;
		Snd2Client(hp, cp, 1);		/* send back filled in msg */
	}
}

/*
 *  MSetQuota() - admin set quota in RC for user
 */
 void
MSetQuota(vmpHdr_t *hp, Client_t *cp)
{
	char	*id = "MSetQuota";

	if (CheckSpecialPerm(cp, id)) {
		(void) Err2Client(hp->vh_seq, cp);
		return;
	}
	if (DBSetQuota(MSGPTR(hp, vmpQuota_t)))
		(void) Err2Client(hp->vh_seq, cp);
	else {
		hp->vh_type = M_QUOTRPLY;
		Snd2Client(hp, cp, 1);		/* send back filled in msg */
	}
}

/*
 *  MLook() - process M_LBLOOK, M_VLS packets received by the RP
 */
 void
MLook(vmpHdr_t *hp, Client_t *cp)
{
	char		*id = "MLook";
	VDR_t		*vdrp;
	vmpHdr_t	*rhp;

	switch (hp->vh_type) {
	    case M_LBLOOK:
		vdrp = NULL;
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR2, id, "not supported");
		break;
	    case M_VLS: {
		/* dead code */
		vmpVsn_t	*vlp;

		vlp = MSGPTR(hp, vmpVsn_t);
		vdrp = DBVsnQuery((char *)&vlp->v_vsn);
		break;
	    }
	    default:
		nLogMsg(EL_ERROR1, id, "bad header: %s", vHdrStr(hp));
		return;
	}

	if (vdrp) {
		vmpVdr_t	*vvp;
		char		ebuf[sizeof (vmpHdr_t) + sizeof (vmpVdr_t)];

		rhp = vBuildHdr(M_VDR, hp->vh_seq, (vmpHdr_t *)ebuf);
		vvp = MSGPTR(rhp, vmpVdr_t);
		memcpy(&vvp->vv_vdr, vdrp, sizeof (vvp->vv_vdr));
		Snd2Client(rhp, cp, 1);
	} else {
		/*
		 *  query failed, send back error packet
		 */
		nLogMsg(EL_DEBUG2, id, "query failed");
		Err2Client(hp->vh_seq, cp);
	}

}

