/*
 *  rp_4.c: Volume Manager Request Processing of drive-related requests 
 *
 *	rp_1.c	main, housekeeping, packet proc switch to routines
 *	rp_2.c	packet proc: mounting & moving volumes
 *	rp_3.c	packet proc: reporting/changing dbase info
 *	rp_4.c	packet proc: varying drives
 *	rp_5.c	packet proc: reserving drives & volumes
 */

#include "rp_space.h"


/*
 *  MVary(): receive and process M_VARY messages.  These are messages
 *	from clients which enable and disable access to devices.
 *	Return non-0 if state change calls for re-eval of queue.
 */

MVary(vmpHdr_t *hp, Client_t *cp, SrvProc_t *sp)
{
	char		*id = "MVary";
	vmpVary_t	*vp;
	DevMap_t	*dmp;		/* device to vary		*/
	int		flags, vary, devno, vmerr, done = 0;
	char		*errstr;	/* vary off error		*/
	int		uid;

	/*
	 *  cmd can come from volcd client or server
	 */
	if (cp && sp)
		nLogMsg(EL_FATAL, id, "server & client");
	if (!cp && !sp)
		nLogMsg(EL_FATAL, id, "neither server nor client");

	if (cp != NULL)
		uid = cp->c_uid;
	else				/* from server */
		uid = 0;

	vp = MSGPTR(hp, vmpVary_t);
	flags = vp->vva_flags;
	nLogMsg(EL_DEBUG2, id, "flags 0x%x %s", flags, vHdrStr(hp));
	vary = flags & (VVA_VARYON|VVA_VARYOFF|VVA_RESET|VVA_DISPREQ);
	if (countintbits(vary) > 1) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR2, id, 
				"vary on, off, or reset only (flags %x)", vary);
		if (cp)
			Err2Client(hp->vh_seq, cp);
		else
			Err2Srv(hp->vh_seq, sp);
		return(0);
	}

	/*
	 *  if a display is requested, that's all
	 */
	if (vary == VVA_DISPREQ) {
		if (cp)
			Vary2Client(hp->vh_seq, cp);
		else
			Vary2Srv(hp->vh_seq, sp);
		return(0);
	} 
	/*
	 *  actual vary: check permission of caller;
	 *	allow either the RP operator or root
	 *	(uid = 0 above for server)
	 */
	if (uid != VolManOperUID  &&  uid != 0) {
		VMError = E_PERM;
		nLogMsg(EL_CODE|EL_ERROR2, id,
				"Permission denied [UID %d]", uid);
		if (cp)
			Err2Client(hp->vh_seq, cp);
		else
			Err2Srv(hp->vh_seq, sp);
		return(0);
	}

	nLogMsg(EL_DEBUG1, id, "DOIT!!");

	dmp = &vp->vva_dev;
	errstr = vp->vva_err;

	/*
	 *  take care of VVA_ALL here, it's kind of a special case
	 */
	if ((flags & VVA_ALL)) {
		nLogMsg(EL_DEBUG1, id, "All..!!");
		if (VaryAll(vary, uid, errstr) && (flags & VVA_RESET))
			goto badmultreset;
		if (cp)
			Vary2Client(hp->vh_seq, cp);
		else
			Vary2Srv(hp->vh_seq, sp);
		return(1);
	}

	/*
	 *  see if a bit is set in the device map
	 *	if no bit is set (got -1), then error
	 */
	if (vDMffs(dmp) == -1) {
		VMError = E_INTERNAL;
		nLogMsg(EL_CODE|EL_STATUS1, id, "no bit set in device map");
		if (cp)
			Err2Client(hp->vh_seq, cp);
		else
			Err2Srv(hp->vh_seq, sp);
		return(0);
	}

	/*
	 *  process each bit that is set
	 */
	vmerr = 0;
	for (devno = 0; devno < MAXDEVS; devno++) {
		if (DM_ISSET(devno, *dmp)) {
			nLogMsg(EL_DEBUG1, id, "%d..!!", devno);
			if (Vary(devno, vary, uid, 0, errstr)) {
				/*
				 *  error in permission, protocol..
				 *	if for 1 drive, probably all
				 */
				vmerr++;
			} else {
				/*
				 *  success or system err
				 */
				done++;
			}
		}
	}
	if (vmerr && !done) {
		/*
		 *  total washout
		 */
		if (cp)
			Err2Client(hp->vh_seq, cp);
		else
			Err2Srv(hp->vh_seq, sp);
		return(0);
	}

badmultreset:
	/*
	 *  collect errors & successes for each drive & send
	 */
	if (cp)
		Vary2Client(hp->vh_seq, cp);
	else
		Vary2Srv(hp->vh_seq, sp);
	return(done);
}

 void
seterr(DevEnt_t *dep, char *errstr)
{
	char	*cp = (char *) dep->de_errstr;

	/*
	 *  the de_errstr pointer is defined as an integer
	 *	because if it's a pointer, XDR crashes when 
	 *	receiving the DevEnt_t w/ non-NULL de_errstr, 
	 *	seemingly because of trying to access the pointer.
	 */
	if (errstr && *errstr) {
		if (cp == NULL) {
			cp = nAlloc(ERRSTRSIZE);
			dep->de_errstr = (unsigned long) cp;
			*cp = '\0';
		}
		strncpy(cp, errstr, ERRSTRSIZE-1);
	} else if (cp != NULL) {
		nFree(cp);
		dep->de_errstr = (int)NULL;
	}
}

/*
 *  Vary(): perform the actual vary operation on the devno passed-in
 */

Vary(int devno, int action, int uid, int fromUnmount, char *errstr)
{
	char		*id = "Vary", *varystr;
	DevEnt_t	*dep;
	RPRC_t		*rcp;
	int		medindex;

	if (action & VVA_VARYON  &&  action & VVA_VARYOFF)
		nLogMsg(EL_FATAL, id, "vary on & off - programming err");

	vClearError();
	switch (action) {
	    case VVA_VARYON:		varystr = "ON"; break;
	    case VVA_VARYOFF:		varystr = "OFF"; break;
	    case VVA_RESET:		varystr = "RESET"; break;
	    default:
		VMError = E_USER;
		nLogMsg(EL_ERROR1, id, "bad option %d", action);
		return(VMError);
	}

	/*
	 *  get device entry from device map; should never fail
	 */
	if ((dep = vGetDevByNum(devno)) == NULL) {
		VMError = E_INTERNAL;
		nLogMsg(EL_CODE|EL_ERROR2, id,
		  "[%s] vGetDev(%d) returned NULL", varystr, devno);
		return (VMError);
	}
	if ((rcp = RCInfo(dep->de_rc)) == NULL) {
		VMError = E_INTERNAL;
		nLogMsg(EL_CODE|EL_ERROR2, id, 
			"[%s] %s: device with bogus RC %d",
		  	varystr, dep->de_name, dep->de_rc);
		return (VMError);
	}
	medindex = RCMedium(rcp, dep->de_medium);

	/*
	 *  If resetting, see if the device is already varied off.
	 *	If not, refuse reset
	 */
	if ((action & VVA_RESET) && DM_ISSET(devno, DMVary)) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR2, id,
		  "%s: device must be offline to be reset; request refused",
		  varystr);
		return (VMError);
	}

	/*
	 *  If varying off and not called from an unmount, see if there's 
	 *	anything mounted on this device.
	 *	If so, will have to unmount before vary off.
	 *	(When called from an unmount, it's the completion
	 *	of an unmount started here or an unmount failure.)
	 */
	if (action == VVA_VARYOFF) {
	    if (!fromUnmount) {
		REG VolReq_t	*vrp, *nextvrp;

		/*
		 *  if device has a volume on it, refuse vary
		 *	unless unmount in progress
		 */
		if (DM_ISSET(devno, rcp->rc_inuse[medindex])) {
		    for (vrp = VRHEAD(VolReqQ); vrp; vrp = nextvrp) {
			nextvrp = VRNEXT(vrp->vr_vrq);

			if (DM_ISSET(devno, vrp->vr_devinuse)) {
			    switch (vrp->vr_state) {
				case VS_MOUNTED:
				case VS_DELETENODE:
				case VS_RCWAIT2:
				  	vrp->vr_flag |= VF_VARYOFF;
					seterr(dep, errstr);
				  	return(E_NOERROR);
				case VS_UNMOUNTWAIT:
				  	vrp->vr_flag |= VF_VARYOFF;
					/* start the unmount */
					DeleteNode(vrp, 1);
					seterr(dep, errstr);
					return(VMError);
				default:
					VMError = E_INTERNAL;
					nLogMsg(EL_ERROR1|EL_CODE|EL_CONSOLE0, 
						id, "unexpected state %s",
						vReqStateStr(vrp->vr_state));
					return(VMError);
			    }
			}
		    }
		    /*
		     *  no req found
		     */
		    VMError = E_INTERNAL;
		    nLogMsg(EL_ERROR1|EL_CONSOLE0|EL_CODE, id, 
				"%s in use but no vsn", dep->de_name);
		    return(VMError);
		}
	    }

	    /*
	     *  drive is not mounted (tho RC might be mounting on
	     *	it this very femtosecond).
	     *  logical vary off if varied on
	     */
	    if (DM_ISSET(devno, DMVary)) {
		dep->de_status &= ~DS_VARY;
		DM_CLR(devno, DMVary);
		rcp->rc_devs[medindex].d_nvaryondev--;
	    }
	    if (DM_ISSET(devno, rcp->rc_inuse[medindex]))
		DM_CLR(devno, rcp->rc_inuse[medindex]);

	    /*
	     *  install or remove error string
	     */
	    seterr(dep, errstr);
	}

	if (action & VVA_VARYON) {

		if (!DM_ISSET(devno, DMVary)) {

			/*
			 *  logical vary on if varied off
			 */
			dep->de_status |= DS_VARY;
			DM_SET(devno, DMVary);
			rcp->rc_devs[medindex].d_nvaryondev++;

			/*
			 *  thanks to a successful vary on,
			 *	turn off error string
			 */
			if (dep->de_errstr != (int)NULL) {
				nFree((char *)dep->de_errstr);
				dep->de_errstr = (int)NULL;
			}
		}
	} 

	nLogMsg(EL_STATUS1, id, "%s (%d) %s varied %s by UID %d",
	  		dep->de_name, devno, rcp->rc_id, 
			varystr, uid);
	nLogMsg(EL_DEBUG2, id, "on/mount= %d/%d",
			rcp->rc_devs[medindex].d_nvaryondev,
			rcp->rc_devs[medindex].d_mountdev);
	return(VMError);
}


/*
 *  VaryAll(): vary all known devices to the specified value
 *	return 0 if noerrors, 1 if any errors
 */

VaryAll(int action, int uid, char *errstr)
{
	int		devno, errs = 0;
	DevEnt_t	*dep;

	for (devno = 0; devno < MAXDEVS; devno++) {
		if (dep = vGetDevByNum(devno)) {
			int	variedon = DM_ISSET(devno, DMVary);

			if ((action == VVA_RESET)
			  || (variedon && action == VVA_VARYOFF)
			  || (!variedon && action == VVA_VARYON))
				if (Vary(devno, action, uid, 0, errstr))
					errs = 1;
		}
	}
	return (errs);
}

 void
VaryDefaults()
{
	char	*id = "VaryDefaults";
	int	devno, errs = 0;

	for (devno = 0; devno < MAXDEVS; devno++) {
		DevEnt_t	*dep = vGetDevByNum(devno);
		if (dep == NULL)
			continue;
		if (dep->de_defvary) {
			nLogMsg(EL_STATUS1, id, "default varyon %s", 
								dep->de_name);
			errs += Vary(devno, VVA_VARYON, VolManOperUID, 0, NULL);
		}
	}
	if (errs)
		nLogMsg(EL_CONSOLE0|EL_STATUS1, 0, "%d default varyon errors",
							errs);
}
