/*
 *  rp_5.c - reservation requests
 *
 *	rp_1.c	main, housekeeping, packet proc switch to routines
 *	rp_2.c	packet proc: mounting & moving volumes
 *	rp_3.c	packet proc: reporting/changing dbase info
 *	rp_4.c	packet proc: varying drives
 *	rp_5.c	packet proc: reserving drives & volumes
 */

#include "rp_space.h"
#include <ctype.h>

/*
 *  MReserve() - deal with a reservation request, which may have a variety
 *	of parts to it. Send error msgs as needed & finish with an ack.
 */

MReserve(vmpHdr_t *hp, Client_t *cp)
{
	char			*id = "MReserve";
	Rid_t			*ridp;
	vmpReserveHdr_t		*rep = MSGPTR(hp, vmpReserveHdr_t);
	RPRC_t			*rcp;
	Client_t		*cp2;
	char			*pkt_cp;
	int			i, medindex, procQ = 0;
	char			*medtype;

	pkt_cp = (char *)rep + PadSize(sizeof(*rep));

	/*
	 *  Report reservations if requested
	 */
#define UserRidReport	AllRidReport
	if (rep->re_flag == RE_LUSER) {
		UserRidReport(hp, cp);
		return(0);
	}
	if (rep->re_flag == RE_LALL) {
		AllRidReport(hp, cp);
		return(0);
	}
	if (BadRidName(rep->re_id)) {
		ErrAck2Client(hp->vh_seq, cp);
		return(0);
	}
	nLogMsg(EL_DEBUG1, id, "rid: %.*s uid %d pid %d", 
		RIDLEN, rep->re_id, cp->c_uid, cp->c_pid);
	
	/*
	 *  see if client wants to be synchronous w/ reservation
	 */
	if (rep->re_flag & RE_ASYNC)
		cp2 = NULL;
	else
		cp2 = cp;

	/*
	 *  create or find rid
	 */
	if (rep->re_flag & RE_CREATE  &&  rep->re_flag & RE_USE) {
		ridp = GetRid(rep->re_id, cp->c_uid);
		if (ridp == NULL  &&  VMError != E_PERM) {
			ridp = MakeRid(rep->re_id, cp->c_uid, cp2, hp->vh_seq);
			nLogMsg(EL_STATUS2, id, 
				"Make rid: %.*s uid %d pid %d", 
				RIDLEN, rep->re_id, cp->c_uid, cp->c_pid);
		}
	} else if (rep->re_flag & RE_CREATE) {
		nLogMsg(EL_STATUS2, id, 
				"Make rid: %.*s uid %d pid %d", 
				RIDLEN, rep->re_id, cp->c_uid, cp->c_pid);
		ridp = MakeRid(rep->re_id, cp->c_uid, cp2, hp->vh_seq);
	} else
		ridp = GetRid(rep->re_id, cp->c_uid);
	if (!ridp) {

		/*
		 *  create or find/permission failed
		 */
		ErrAck2Client(hp->vh_seq, cp);
		return(0);
	}

	/*
	 *  delete whole rid if requested
	 */
	if (rep->re_flag & RE_DELETE) {
		nLogMsg(EL_STATUS2, id, 
				"Delete rid: %.*s uid %d pid %d", 
				RIDLEN, rep->re_id, cp->c_uid, cp->c_pid);
		DeleteRid(ridp, cp);
		AckClient(hp->vh_seq, cp);
		return(1);	/* might allow another proc to move */
	}

	if (rep->re_rcmedcount) {
		vmpReserveRCMed_t	*ridrcmed = (vmpReserveRCMed_t *)pkt_cp;

		/*
		 *  add/delete drives
		 */
		for (i=0; i<rep->re_rcmedcount; i++, ridrcmed++) {

			vClearError();

			/*
			 *  check this 'drive subcmd'
			 */
			rcp = RCInfo(ridrcmed->re_rc);
			if (!rcp) {
				VMError = E_USER;
				nLogMsg(EL_CODE|EL_ERROR2, id, 
						"Nonexistent RC: %d", 
						ridrcmed->re_rc);
				Err2Client(hp->vh_seq, cp);
				continue;
			}
			medtype = vMediumName(ridrcmed->re_medium);
			medindex = RCMedium(rcp, ridrcmed->re_medium);
			if (medindex == -1) {
				VMError = E_USER;
				nLogMsg(EL_CODE|EL_ERROR2, id, 
						"%s not configured to hold %s",
						rcp->rc_id, medtype);
				Err2Client(hp->vh_seq, cp);
				continue;
			}

			nLogMsg(EL_STATUS2, id, 
			  "Reserve %.*s: RC %s/%s: %d %s drive%s uid %d pid %d",
				RIDLEN, rep->re_id,
				rcp->rc_id, medtype,
				ridrcmed->re_drives,
				(ridrcmed->re_type == RE_HARD ? "hard": "soft"),
				(ridrcmed->re_drives > 1 ? "s" : ""),
				cp->c_uid, cp->c_pid);

			if (ridrcmed->re_drives < 0)
				procQ++;	/* might let something run */

			switch (ridrcmed->re_type) {
			    case RE_HARD:
				if (HardReserve(ridp, rcp, medindex, 
							ridrcmed->re_drives))
					Err2Client(hp->vh_seq, cp);
				break;
			    case RE_SOFT:
				if (SoftReserve(ridp, rcp, medindex, 
							ridrcmed->re_drives))
					Err2Client(hp->vh_seq, cp);
				break;
			    default:
				VMError = E_USER;
				nLogMsg(EL_CODE|EL_ERROR2, id, 
						"bad reservation type %d",
						ridrcmed->re_type);
				Err2Client(hp->vh_seq, cp);
			}
		}
		ridrcmed--;
		pkt_cp = (char *) ridrcmed + PadSize(sizeof(*ridrcmed));
	}

	if (rep->re_vsncount) {
		vmpReserveVsn_t		*ridvsn = (vmpReserveVsn_t *) pkt_cp;

		/*
		 *  add/delete vsns
		 */

		for (i=0; i<rep->re_vsncount; i++, ridvsn++) {
			VolReq_t	*vrp, *AddVsnToRid();

			vClearError();
			switch (ridvsn->re_type) {
			    case RE_ADD:
				/* add to rid explicitly */
				vrp = AddVsnToRid((char *)&ridvsn->re_vsn, 
						ridp, cp2, cp->c_uid, 0);
				if (vrp == NULL)
					Err2Client(hp->vh_seq, cp);
				else
					nLogMsg(EL_STATUS2, id, 
					"added %s_#%d to %.*s uid %d pid %d", 
					vsnstr((char *)&ridvsn->re_vsn), 
					vrp->vr_rpseq, RIDLEN, rep->re_id,
					cp->c_uid, cp->c_pid);
				break;
			    case RE_DELVSN:
				if (DeleteVsnFromRid((char *)&ridvsn->re_vsn, 
								ridp, cp))
					Err2Client(hp->vh_seq, cp);
				else {
					nLogMsg(EL_STATUS2, id, 
					"deleted %.*s from %.*s  uid %d pid %d",
					VSNLEN, (char *)&ridvsn->re_vsn,
					RIDLEN, rep->re_id,
					cp->c_uid, cp->c_pid);
			    		procQ++;	/* might let smt run */
				}
				break;
			
			    default:
				VMError = E_USER;
				nLogMsg(EL_CODE|EL_ERROR2, id, 
					"unknown flag for %s", 
					vsnstr((char *)&ridvsn->re_vsn));
				Err2Client(hp->vh_seq, cp);
			}
		}
	}

	/*
	 *  finally, tell client request proc is done
	 */
	
	AckClient(hp->vh_seq, cp);
	return(procQ);
}

