#ifndef lint
static char rcsid[] = "$Header: /Volumes/TempStorage/repos/netbsd/othersrc/nastore/volman/srvr/master/rp_clnt.c,v 1.1.1.1 2000-02-28 02:18:28 wrstuden Exp $";
#endif

/*
 *  rp_clnt.c: Volume Manager Request Processor Client_t support code 
 *
 *	Volcd processes (1 per host) set up client connections,
 *	checking the uid of each, then tell vold about each client,
 *	passing msgs: client<->volcd<->vold. The vold keeps an
 *	array of Client_t structs for each host w/ a running volcd.
 *
 *	Bill Ross, NASA Ames, October 1995
 */

#include "rp_space.h"


static int	totalClients;

/*
 *  MNewClient(): null out client struct, increment count of total clients
 *
 *	Leave IPC part of client for calling IPC routine to fill in.
 */
 void
MNewClient(vmpHdr_t *hp, Host_t *hop)
{
	char		*id = "MNewClient";
	vmpHello_t	*hep = MSGPTR(hp, vmpHello_t);
	Client_t	*cp;

	nLogMsg(EL_STATUS2, id, "pid %d uid %d host %s slot %d", 
				hep->vhe_pid, hep->vhe_id, hop->h_name, 
				hep->vhe_flag);
	/*
	 *  get/check client struct
	 *	TODO: what if FD_SETSIZE varies from machine to machine?
	 *	--register volcd setsize at volcd init
	 */
	if (hep->vhe_flag < 0  ||  hep->vhe_flag >= FD_SETSIZE) {
		nLogMsg(EL_CONSOLE0|EL_ERROR1, id, 
					"host %s slot %d out of range", 
					hop->h_name, hep->vhe_flag);
		return;
	}
	if (hop->h_clients == NULL)
		nLogMsg(EL_FATAL, id, "null clients for host %s", hop->h_name);
	cp = &hop->h_clients[hep->vhe_flag];
	if (cp->c_flags != 0)
		nLogMsg(EL_CONSOLE0|EL_ERROR1, id, 
					"new client on top of old %s slot %d",
					hop->h_name, hep->vhe_flag);

	cp->c_flags = CL_EXIST;
	cp->c_id = hep->vhe_flag;	/* slot on client host */
	cp->c_uid = hep->vhe_id;
	cp->c_pid = hep->vhe_pid;
	time(&cp->c_time);

	cp->c_cvrq.q_head = NULL;
	cp->c_cvrq.q_tail = NULL;

	cp->c_ridq.q_head = NULL;
	cp->c_ridq.q_tail = NULL;

	totalClients++;
	hop->h_nclients++;
}

/*
 *  ShutdownClientVolReqs(): mark all of a client's volume requests as
 *	to be canceled and detach them from from the client's c_cvrq 
 *	(leaving them on the VolReqQ) so this client struct can be closed. 
 *	To avoid such cancellation, the client must have issued an M_DETACH.
 */
 static void
ShutdownClientVolReqs(Client_t *cp)
{
	char		*id = "ShutdownClientVolReqs";
	VolReq_t	*vrp, *nextvrp;

	for (vrp=VRHEAD(cp->c_cvrq); vrp; vrp=nextvrp) {
		nextvrp = VRNEXT(vrp->vr_cvrq);	

		nLogMsg(EL_DEBUG1, id, "%.*s_%d", VSNLEN, &vrp->vr_vsn,
				vrp->vr_rpseq);
		/*
		 * TODO: recognize volreqs which can be disposed of
		 *	here and now.  Could call FreeVolReq().
		 */
		vrp->vr_clientp = NULL;
		DeQVolCVR(vrp, &cp->c_cvrq);
		vrp->vr_uid = cp->c_uid;	/* controlling uid */

		/* 
		 *  no one cares about this request now
		 */
		vrp->vr_voflag &= ~VO_RESPONSE;
		vrp->vr_voflag |= VO_ASYNC;

		/*
		 *  handle the state
		 */
		switch (vrp->vr_state) {
		    case VS_MOVING:
		    case VS_DELETENODE:
		    case VS_RCWAIT2:
		    case VS_UNMOUNTWAIT:
			/*
			 * don't do anything to move-in-progress or
			 *	unmount-in-progress; allow it to finish
			 */
			break;
		    case VS_MOUNTWAIT:
		    case VS_DRIVEWAIT:
		    case VS_RESERVEMOUNTWAIT:
		    case VS_RCWAIT:
		    case VS_CREATENODE:
		    case VS_MOUNTED:
			Unmount(vrp);	/* unmount if one's not in progress */
			break;
		    default: 
		     /* TODO: this is too open-ended, check for coding errs */
			nLogMsg(EL_ERROR1, id, 
			    "INTERNAL ERROR: default for state %s: FreeVolReq", 
			    vReqStateStr(vrp->vr_state));
			FreeVolReq(vrp, id);	/* totally dequeue it now */
		}
	}

	cp->c_cvrq.q_head = NULL;		/* no more queue here */
}

/*
 *  ShutdownClient(): shutdown and clean up after a Client_t struct;
 *	Should be called when there are no VolReqs associated with
 *	the client struct, and after the client has been sent an
 *	M_SHUTDOWN or has said M_GOODBYE.
 *
 */
 static void
ShutdownClient(Client_t *cp, Host_t *hop)
{
	char	*id = "ShutdownClient";

	nLogMsg(EL_DEBUG4, id, "pid %d", cp->c_pid);

	if (cp->c_cvrq.q_head) {
		nLogMsg(EL_CONSOLE0|EL_ERROR1, id, 
				"internal error: called with non-empty cvrq");
		return;
	}
	cp->c_flags = 0;
	totalClients--;
	hop->h_nclients--;
}
/*
 *  MClientGoodbye() - process M_GOODBYE packets received by the RP.
 *	Take a client and all its reservations and volreqs out of commission.
 *	(Client will have issued an M_DETACH if it wanted to protect
 *	cancellations.)
 */
 void
MClientGoodbye(Client_t *cp)
{
	char	*id = "MClientGoodbye";
	Host_t	*hop;

	if (cp->c_ipcp == NULL) {
		nLogMsg(EL_STATUS1, id, "client ipc null");
		return;
	}
	hop = (Host_t *) cp->c_ipcp->sock_host;

	nLogMsg(EL_STATUS2, id, "pid %d uid %d host %s slot %d",
	  		cp->c_pid, cp->c_uid, hop->h_name, cp->c_id);

	/*
	 *  Shut down all reservations belonging to this Client_t struct
	 */
	ShutdownClientRids(cp);

	/*
	 *  Shut down all VolReqs belonging to this Client_t struct
	 *	(i.e. ones not in reservations already shut down).
	 */
	ShutdownClientVolReqs(cp);

	/*
	 *  mark client struct unused;
	 */
	ShutdownClient(cp, hop);
}

/*
 *  ShutdownClientRids() - delete reservations. To avoid this, the
 *	client must have issued an M_DETACH.
 */
ShutdownClientRids(Client_t *cp)
{
	char		*id = "ShutdownClientRids";
	Rid_t		*ridp, *nextridp;
	VolReq_t	*vrp, *nextvrp;
	int		procQ = 0;

	nLogMsg(EL_DEBUG3, id, "");

	/*
	 *  Delete any reservations (shuts down volreqs in rids as well).
	 */

	for (ridp=(Rid_t *)cp->c_ridq.q_head; ridp; ridp=nextridp) {
		nextridp = (Rid_t *)ridp->rid_clientq.q_next;

		if (ridp->rid_cp != cp) {
			nLogMsg(EL_ERROR1, id, "rid in cp q w/ different cp");
			kill(getpid(),5);
		}

		/* 
		 *  close down any vsn requests
		 */
		for (vrp=VRHEAD(ridp->rid_vsnq); vrp; vrp=nextvrp) {
			nextvrp = VRNEXT(vrp->vr_ridq);
			/*
			 *  delete req (or give it to RP if
			 *	there is residual activity)
			 */
			nLogMsg(EL_DEBUG2, id, "del req");
			CancelRequest(vrp, cp, id);
		}

		if (ridp->rid_active)
			procQ++;
		DeleteRid(ridp, cp);
	}
	return(procQ);
}




/*
 *  ShutdownAllClients(): shutdown and clean up after all clients.
 *	(Could shut down reservations to get all current requests, 
 *	but this gives the subset of all requests with a client that 
 *	can be notified.)
 */
 void
ShutdownAllClients(char *msg)
{
	int		i, j;
	Host_t		*hop;
	Client_t	*cp;
	char		pbuf[PKTSIZ(vmpError_t)];
	vmpError_t	*ep;
	vmpHdr_t	*hp;
	int		nclients = 0;

	for (i=0;; i++) {
		hop = HostIndex(i);
		if (hop == NULL)
			break;
		if (hop->h_clients == NULL)
			continue;
		 
		cp = hop->h_clients;
		for (j=0; j<FD_SETSIZE; j++, cp++) {
			if (cp->c_flags & CL_EXIST) {
				nclients++;
				nLogMsg(EL_STATUS1,0, 
					"shutdown %s clnt %d pid %d",
					hop->h_name, j, cp->c_pid);
				ShutdownClientVolReqs(cp);
				hp = vBuildHdr(M_SHUTDOWN, NOSEQ, 
							(vmpHdr_t *) pbuf);
				ep = MSGPTR(hp, vmpError_t);
				if (msg)
					emsgcpy(ep->vn_msg, msg);
				else
					ep->vn_msg[0] = '\0';
				ep->vn_error = E_DOWN;
				Snd2Client(hp, cp, 1);
			}
		}
	}
	nLogMsg(EL_STATUS1, 0, "shutdown %d clients", nclients);
}

/*
 *  vLogClients() - print summary to log
 */
 void
vLogClients()
{
	Client_t	*cp;
	Host_t		*hop;
	int		i, j;

	for (i=0;; i++) {
		hop = HostIndex(i);
		if (hop == NULL)
			break;
		if (hop->h_clients == NULL)
			continue;
		 
		nLogMsg(EL_ALWAYS|EL_NOTIME, 0, 
			"=============== Client list (%s) =================",
			hop->h_name);
		cp = hop->h_clients;
		for (j=0; j<FD_SETSIZE; j++, cp++) {
			if (cp->c_flags & CL_EXIST) {
				nLogMsg(EL_ALWAYS|EL_NOTIME, 0,
			  	"\tslot=%d, addr=%x, pid=%d, uid=%d, flags=%x",
			  	j, cp, cp->c_pid, cp->c_uid, cp->c_flags);
				PrintClientVsns(cp, NULL);
			}
		}
	}
}

