/*
 *  NAStore Volume Manager
 *
 *  rp_rc.c: Volume Manager RC & device configuration support
 *
 */

/*
 *  TODO - have densities and vary functions come from the
 *	drive spec (see SysMedium()) instead of being
 *	hardwired for the RC. For now, leaving densities
 *	to default.
 */

#include "rp_space.h"

#include <syslog.h>
#include <sys/stat.h>

static int		RCarraysize = 2;
static MoveTable_t	*MoveArray = NULL;
static int		MoveSize;
static int		MoveTotal = 0;

/*******************************************************************
			VCONF PARSING ROUTINES
 *******************************************************************/

#define TOKLEN		80

/* gettok() modes - eofbomb implies newline, sameline is for line-formatted */
#define SAMELINE	0
#define GETNEWLINE 	1
#define EOFBOMB 	2

static FILE	*fp;
static int	linenum = 0, lineindex = 0, tab = 0;
static char	tok[TOKLEN], line[TOKLEN];

char *keywd[] = {

	/* 		    main keywords 		*/
	"BEGIN_HOSTS",	"END_HOSTS",
	"BEGIN_RC",	"END_RC",	"NUMBER"	"NAME",		
	"ID", 		"TYPE",		"PATH",		"LOGIN",
	"SYSLOG",	"FLAGS",	"CAPACITY",	"MOUNTTIMEOUT",	
	"ACSID",	"DRIVES",	"BEGIN_MOVES",	"END_MOVES",	
	"VARY",		"NONODE",	"CLEARFIRST",	"REWUNLOAD",
	"REWIND",	"",
};



CodeString_t	RCFlagArr[] = {
	{ RCF_TIMER,		"TIMER_MSGS",		},
	{ RCF_ASYNCUNMNT,	"ASYNC_UNMOUNT",	},
	{ RCF_SETLED,		"SET_LED",		},
	{ RCF_MOUNTOK,		"MOUNT_OK",		},
	{ RCF_ALLOCOK,		"ALLOC_OK",		},
	{ RCF_HASQUOTA,		"HAS_QUOTA",		},
	{ RCF_RECYCLEOK,	"RECYCLE_OK",		},
	{ RCF_NODEMOUNT,	"NODEMOUNT",		},
	{ RCF_NONODE,		"NONODE",		},
	{ RCF_CACHEMOUNT,	"CACHEMOUNT",		},
	{ RCF_CLEARFIRST,	"CLEARFIRST",		},
	{ RCF_REWUNLOAD,	"REWUNLOAD",		},
	{ RCF_REWIND,		"REWIND",		},
	{ 0,			0,			},
};

 int
RCFlag(char *str)
{
	return(nStrToCode(str, RCFlagArr));
}

iskeywd()
{
	char	**p;

	for (p = keywd; strlen(*p); p++)
		if (!strcmp(tok, *p))
			return(1);
	if (vVolTypeByID(tok))
		return(1);
	if (RCFlag(tok) != -1)
		return(1);
	return(0);
}

  void
inerr(char *str1, char *str2)
{	
	char	buf[81];
	int     i;

	if (lineindex > 2) {
		for (i=0; i<lineindex+7*tab-1; i++)
			buf[i] = '.';
		buf[i] = '\0';
		nLogMsg(EL_STATUS1, 0, "\n%s\n%s", line, buf);
	}
	nLogMsg(EL_FATAL, 0, 
	 "config file %s:\n\tError on line %d: %s %s\n\tLast token was: '%s'\n",
	 vVoldConfig, linenum, str1, str2, tok);
}

readln(int eofbomb)
{
	int	i, j;

repeat:
	for (i=0; i<81; i++) {
		if ((j = fgetc(fp)) == EOF) {
			if (eofbomb)
				inerr("Unexpected EOF","");
			line[i] = '\n';
			lineindex = 0;
			tab = 0;
			return(i);
		}
		line[i] = (char) j;
		if (line[i] == '\n') {
			line[i] = '\0';
			break;
		}
	}
	if (line[i] != '\0')
		nLogMsg(EL_FATAL, 0, "%s: line too long (limit is 80):\n%.81s",
			vVoldConfig, line);
	nLogMsg(EL_STATUS1|EL_NOTIME, 0, line);
	linenum++;
	if (line[0] == '#'  ||  line[0] == '\0')
		goto repeat;
	lineindex = 0;
	tab = 0;
	return(i);
}

  void
clearline()
{
	for (; line[lineindex] != '\0'; lineindex++) {
		if (!isspace(line[lineindex]))
			inerr("Expected newline for definition of drives","");
	}
}

/*
 *  gettok() - get next token, going on to next line if requested
 *	if EOF, bomb if requested (newline==EOFBOMB)
 */
  void
gettok(int newline)
{
	static int	linelen;
	int		i;

	for (i=0; ; ) {
		switch (line[lineindex]) {
			case '\t':
				tab++;
			case ' ':
				lineindex++;
				if (i) {
					tok[i] = '\0';
					return;
				}
				break;
			case '\n':
				/* EOF */
				tok[0] = '\0';
				return;
			case '"':
				if (!i) {
					/*
					 *  beginning of string - get all
					 */
					lineindex++;
					if (line[lineindex] == '"')
						inerr("Empty string","");
					if (line[lineindex] == '\0')
						inerr("Unterminated string","");
					while(1) {
						tok[i++] = line[lineindex++];
						if (line[lineindex] == '"') {
							tok[i] = '\0';
							lineindex++;
							return;
						}
						if (line[lineindex] == '\0')
							inerr("Unterminated ",
								"string");
					}
				} else {
					/*
					 *  end of prev token
					 */
					tok[i] = '\0';
					return;
				}
			case '\0':
				if (!i) {
					if (newline == SAMELINE) {
						tok[0] = '\0';
						return;
					}
					/*
					 *  still waiting for 1st char -
					 *	get another line
					 */
					linelen = readln(
						    newline==EOFBOMB ? 1 : 0);
					continue;
				} else {
					tok[i] = '\0';
					return;
				}
			default:
				tok[i++] = line[lineindex];
				lineindex++;
				break;
		}
	}
}

  int
tokint(int *nump)
{
	int	i, j = strlen(tok);

	i = 0;
	if (tok[0] == '-') i++;
	for (; i<j; i++)
		if (!isdigit(tok[i]))
			return(1);
	*nump = atoi(tok);
	return(0);
}


 static void
ReadHosts()
{
	char	*id = "ReadHosts";

	while (1) {
		gettok(EOFBOMB);
		if (!strcmp(tok, "END_HOSTS"))
			break;
		if (strlen(tok) > MAXHOSTNAMELEN-1)
			nLogMsg(EL_FATAL, id, "%s: host name too long", tok);

		if (NewHost(tok))
			nLogMsg(EL_FATAL, id, "%s: PROBLEM w/ HOST (%s)",
					vVoldConfig, tok);
	}
}

 static void
RCGetFlags(RPRC_t *rcp)
{
	int	flag;

	while (1) {
		gettok(EOFBOMB);
		flag = RCFlag(tok);
		if (flag == -1)
			break;
		rcp->rc_flags |= flag;
	}
}

 static void
RCGetMedia(RPRC_t *rcp)
{
	char	*id = "RCGetMedia";
	int	medium;

	rcp->rc_nmedia = 0;
	while (1) {
		gettok(EOFBOMB);
		medium = vMediumCode(tok);
		if (medium == -1)
			break;
		rcp->rc_media |= medium;
		rcp->rc_nmedia++;
		if (rcp->rc_nmedia > RC_MEDIA_TYPES)
			nLogMsg(EL_FATAL|EL_CONSOLE0, id, 
					"too many media in this RC (%d)", 
					rcp->rc_nmedia);
		rcp->rc_devs[rcp->rc_nmedia-1].d_medium = medium;
	}
}

/*
 *  RCGetDrives() - get drives for this RC.
 *	for now, 1 host for each drive is assumed; later, may
 *	allow more. Since each host can have different naming,
 *	a 'human name' is given plus controller & device # info
 *	for generating the actual device name on the host in
 *	question - see lib/vm_device.c for conventions.
 */
 static void
RCGetDrives(RPRC_t *rcp)
{
	char		*id = "RCDrives";
	int		vary, controllerid, devno, medium, medindex;
	char		*dp, name[MAXDEVLEN], chan[MAXCHANLEN];
	Host_t		*hop;

	/*
	**	make sure there is a fresh line, then do 1 line = 1 drive
	*/
	clearline();
	while (1) {
		readln(1);
		gettok(SAMELINE);

		if (iskeywd())
			break;

		/*
		**	'informal' device name for humans
		*/
		if (strlen(tok) > MAXDEVLEN-1)
			nLogMsg(EL_FATAL, 0, "%s: limited to %d chars", 
							tok, MAXDEVLEN-1);
		strcpy(name, tok);

		/*
		**	default vary status
		*/
		gettok(SAMELINE);
		if (!strcmp(tok, "ON")  ||  !strcmp(tok, "on"))
			vary = 1;
		else if (!strcmp(tok, "OFF")  ||  !strcmp(tok, "off"))
			vary = 0;
		else
			nLogMsg(EL_FATAL, 0, 
			 "drive %s: expected default vary status = ON/OFF (%s)",
			 name, tok);
		
		/*
		**	host - name same as given in HOSTS section
		**		(TODO - someday maybe > 1 host)
		*/
		gettok(SAMELINE);
		hop = HostByName(tok);
		if (hop == NULL)
			nLogMsg(EL_FATAL, 0,
				"%s: expected hostname from HOSTS section [%s]",
								name, tok);

		/*
		**	controller id - sgi at least needs this; others
		**		can use a filler # like 0..
		*/
		gettok(SAMELINE);
		if (tokint(&controllerid)) 
			nLogMsg(EL_FATAL, 0, "%s: expected controller # %s", 
							name, tok);

		/*
		**	device number - every host type needs this
		*/
		gettok(SAMELINE);
		if (tokint(&devno)) 
			nLogMsg(EL_FATAL, 0, "%s: expected device # %s", 
							name, tok);
	
		/*
		**	medium
		*/
		gettok(SAMELINE);
		medium = vMediumCode(tok);
		if (medium == -1) {
			nLogMsg(EL_STATUS1, id, 
				"%s: unknown medium (%s); expected one of:", 
								name, tok);
			vListMedia(EL_STATUS1, id);
			nLogMsg(EL_FATAL, id, "--fix VCONF");
		}

		/*
		**	if medium not configured, complain
		*/
		if (!(rcp->rc_media & medium))
			nLogMsg(EL_FATAL, id, "%s: medium %s not in RC %s",
						name, tok, rcp->rc_id);
		medindex = RCMedium(rcp, medium);
		if (medindex == -1)
			nLogMsg(EL_FATAL, id, "--fix VCONF or code: medindex");

		/*
		**	get 'channel' - a string used to optimize bandwidth
		**		by showing how drives are grouped (maybe the
		**		controllerid could be used for this..)
		*/
		gettok(SAMELINE);
		if (tok[0] == '\0')
			inerr("expected channel","");
		if (strlen(tok) > MAXCHANLEN-1)
			nLogMsg(EL_FATAL, 0, "%s: longer than MAXCHANLEN %d", 
							tok, MAXCHANLEN-1);
		strcpy(chan, tok);

		/*
		**	add drive to RC config
		*/
		if (DriveTotal >= MAXDEVS)
			nLogMsg(EL_FATAL|EL_CONSOLE0, 0, 
				"Total drives exceeds MAXDEVS (%d) - %s",
				 MAXDEVS, "redimension and recompile");

		DM_SET(DriveTotal, rcp->rc_devmap);
		DM_SET(DriveTotal, rcp->rc_devs[medindex].d_devmap);
		rcp->rc_nconfigdev++;
		rcp->rc_devs[medindex].d_nconfigdev++;

		/*
		**	add drive to host array
		*/
		DM_SET(DriveTotal, hop->h_drives);
		hop->h_ndrives++;

		/*
		**	add drive to global array
		*/
		DriveArray[DriveTotal].de_rc = rcp->rc_number;
		DriveArray[DriveTotal].de_medium = medium;
		DriveArray[DriveTotal].de_rcflags = 
				rcp->rc_flags & 
				(RCF_CLEARFIRST|RCF_REWIND|RCF_REWUNLOAD);
if (DriveArray[DriveTotal].de_rcflags & RCF_REWIND)
nLogMsg(EL_STATUS1, id, "REWIND SET %s", name);
		strcpy(DriveArray[DriveTotal].de_name, name);
		DriveArray[DriveTotal].de_defvary = vary;
		DriveArray[DriveTotal].de_host = hop->h_id;
		DriveArray[DriveTotal].de_controller = controllerid;
		DriveArray[DriveTotal].de_devno = devno;
		strcpy(DriveArray[DriveTotal].de_chan, chan);

		/*
		**	see if further ['acs'] parameters defined
		**		TODO - could have RC types determining
		**		format of the optional drive spec, since
		**		stk acs is just 1 type of robot :-)
		*/
		gettok(SAMELINE);
		if (tok[0] != '\0') {
			if (tokint(&DriveArray[DriveTotal].de_acsdevloc.ad_lsm))
				inerr("expected lsm+panel+drive or nothing","");
			gettok(SAMELINE);
			if (tok[0] == '\0')
				inerr("expected lsm+panel+drive or nothing","");
			if (tokint(
				 &DriveArray[DriveTotal].de_acsdevloc.ad_panel))
				inerr("expected lsm+panel+drive or nothing","");
			gettok(SAMELINE);
			if (tok[0] == '\0')
				inerr("expected lsm+panel+drive or nothing","");
			if (tokint(&DriveArray[DriveTotal].de_acsdevloc.ad_drv))
				inerr("expected lsm+panel+drive or nothing","");
		}
		DriveTotal++;
	}
}

 static void
InitRC(RPRC_t *rcp)
{
	int	i;

	/* RCDef 'public' part */
	rcp->rc_number = -1;
	rcp->rc_index = RCTotal;
	rcp->rc_name[0] = '\0';
	rcp->rc_id[0] = '\0';
	rcp->rc_type[0] = '\0';
	rcp->rc_login[0] = '\0';
	rcp->rc_uid = -1;
	rcp->rc_syslog = 0;
	rcp->rc_flags = 0;
	rcp->rc_density = 0;
	rcp->rc_defden = 0;
	rcp->rc_media = 0;
	rcp->rc_nmedia = 0;
	rcp->rc_mounttimeout = 0;
	rcp->rc_capacity = 0;
        rcp->rc_used = -1;
        rcp->rc_freeslots = 0;
	rcp->rc_nconfigdev = 0;
	DM_ZERO(rcp->rc_devmap);
	rcp->rc_acsid = 0;

	/* RP 'private' part */
	rcp->rc_path = NULL;
	rcp->rc_vary = NULL;
	for (i=0; i<RC_MEDIA_TYPES; i++) {
        	DM_ZERO(rcp->rc_avail[i]);
        	DM_ZERO(rcp->rc_inuse[i]);
        	rcp->rc_devs[i].d_medium = 0;
        	rcp->rc_devs[i].d_nconfigdev = 0;
        	rcp->rc_devs[i].d_nvaryondev = 0;
        	rcp->rc_devs[i].d_mountdev = 0;
        	DM_ZERO(rcp->rc_devs[i].d_devmap);
	}
        rcp->rc_srv.s_sd = -1;
}

 static void
ReadRC()
{
	char	*id = "ReadRC";
	RPRC_t	*rcp;
	int	itmp, err = 0;

	/*
	 *  expand RCArray if necc
	 */
	if (RCTotal >= RCarraysize) {
		RCarraysize += 5;
		RCArray = (RPRC_t *) nReAlloc(RCArray, 
				RCarraysize * sizeof(RPRC_t));
	}
	rcp = &RCArray[RCTotal];

	/*
	 *  initialize this RC to allow checking fields
	 */
	InitRC(rcp);

	/*
	 *  get attributes
	 */
	gettok(EOFBOMB);
	while (1) {

		if (!strcmp(tok, "END_RC"))
			break;

		/*
		 *  each case needs to do a final gettok() 
		 *	for the next iteration
		 */
		if (!strcmp(tok, "NUMBER")) {
			gettok(EOFBOMB);
			if (iskeywd())
				inerr("expected RC number, got keyword","");
			if (tokint(&rcp->rc_number))
				inerr("expected RC number","");
			if (rcp->rc_number < 1)
				inerr("RC NUMBER: must be > 0","");
			gettok(EOFBOMB);
		} else if (!strcmp(tok, "NAME")) {
			gettok(EOFBOMB);
			if (iskeywd())
				inerr("expected RC name, got keyword","");
			if (strlen(tok) > RC_STRLEN-1)
				inerr("RC NAME: too long","");
			strcpy(rcp->rc_name, tok);
			gettok(EOFBOMB);
		} else if (!strcmp(tok, "TYPE")) {
			gettok(EOFBOMB);
			if (iskeywd())
				inerr("expected RC type, got keyword","");
			if (strlen(tok) > RC_STRLEN-1)
				inerr("RC TYPE: too long","");
			strcpy(rcp->rc_type, tok);
			gettok(EOFBOMB);
		} else if (!strcmp(tok, "ID")) {
			gettok(EOFBOMB);
			if (iskeywd())
				inerr("expected RC id, got keyword","");
			if (strlen(tok) > RC_STRLEN-1)
				inerr("RC ID: too long","");
			strcpy(rcp->rc_id, tok);
			gettok(EOFBOMB);
		} else if (!strcmp(tok, "LOGIN")) {
			gettok(EOFBOMB);
			if (iskeywd())
				inerr("expected RC login, got keyword","");
			if (strlen(tok) > RC_STRLEN-1)
				inerr("RC LOGIN: too long","");
			strcpy(rcp->rc_login, tok);
			gettok(EOFBOMB);
		} else if (!strcmp(tok, "SYSLOG")) {
			gettok(EOFBOMB);
			if (iskeywd())
				inerr("expected RC login, got keyword","");
			rcp->rc_syslog = nSysLogCode(tok);
			if (rcp->rc_syslog == 0)
				inerr("Syslog not defined","");
			gettok(EOFBOMB);
		} else if (!strcmp(tok, "PATH")) {
			gettok(EOFBOMB);
			if (iskeywd())
				inerr("expected RC path, got keyword","");
			rcp->rc_path = nAlloc(strlen(tok)+1);
			strcpy(rcp->rc_path, tok);
			gettok(EOFBOMB);
		} else if (!strcmp(tok, "CAPACITY")) {
			gettok(EOFBOMB);
			if (iskeywd())
				inerr("expected RC capacity, got keyword","");
			if (tokint(&rcp->rc_capacity))
				inerr("expected RC capacity","");
			gettok(EOFBOMB);
		} else if (!strcmp(tok, "MOUNTTIMEOUT")) {
			gettok(EOFBOMB);
			if (iskeywd())
				inerr("expected RC timeout, got keyword","");
			if (tokint(&rcp->rc_mounttimeout))
				inerr("expected RC mount timeout","");
			if (rcp->rc_mounttimeout < 1)
				inerr("RC mount timeout must be > 0",
							"(or don't specify)");
			gettok(EOFBOMB);
		} else if (!strcmp(tok, "ACSID")) {
			gettok(EOFBOMB);
			if (iskeywd())
				inerr("expected RC acsid, got keyword","");
			if (tokint(&itmp))
				inerr("expected RC acsid","");
			if (itmp < 0  ||  itmp > 255)
				inerr("RC acsid must be in 0..255","");
			rcp->rc_acsid = itmp;
			gettok(EOFBOMB);
		} else if (!strcmp(tok, "VARY")) {
			extern int	(*VaryName())();

			gettok(EOFBOMB);
			rcp->rc_vary = VaryName(tok);
			if (rcp->rc_vary == NULL)
				inerr("Vary function not defined ",
					"in rp_device.c");
			gettok(EOFBOMB);
		} else if (!strcmp(tok, "FLAGS")) {
			RCGetFlags(rcp);	/* quits on 1st non-flag tok */
		} else if (!strcmp(tok, "MEDIA")) {
			RCGetMedia(rcp); /* quits on 1st non-media tok */
		} else if (!strcmp(tok, "DRIVES")) {
			RCGetDrives(rcp);	/* quits on 1st non-flag tok */
		} else
			inerr("Unexpected RC token","");
	}
	/*
	 *  check flags
	 */
	if (rcp->rc_flags  &  (RCF_CLEARFIRST|RCF_CACHEMOUNT) ==
				(RCF_CLEARFIRST|RCF_CACHEMOUNT)) {
		nLogMsg(EL_CONSOLE0|EL_ERROR1, id,
			"CLEARFIRST & CACHEMOUNT incompatible");
		err++;
	}
if (rcp->rc_flags & RCF_REWIND)
nLogMsg(EL_STATUS1, id, "REWIND SET");
	/*
	 *  check RC req'd fields
	 */
	if (rcp->rc_number == -1) {
		nLogMsg(EL_CONSOLE0|EL_ERROR1, 0, "No RC NUMBER given");
		err++;
	}
	if (rcp->rc_name[0] == '\0') {
		nLogMsg(EL_CONSOLE0|EL_ERROR1, 0, "No RC NAME given");
		err++;
	}
	if (rcp->rc_type[0] == '\0') {
		nLogMsg(EL_CONSOLE0|EL_ERROR1, 0, "No RC TYPE given");
		err++;
	}
	if (rcp->rc_id[0] == '\0') {
		nLogMsg(EL_CONSOLE0|EL_ERROR1, 0, "No RC ID given");
		err++;
	}
	if (rcp->rc_login[0] == '\0') {
		nLogMsg(EL_CONSOLE0|EL_ERROR1, 0, "No RC LOGIN given");
		err++;
	}
	if (rcp->rc_media == 0) {
		nLogMsg(EL_CONSOLE0|EL_ERROR1, 0, 
					"No RC MEDIA (volume types) given");
		err++;
	}
	if (rcp->rc_capacity == 0) {
		nLogMsg(EL_CONSOLE0|EL_ERROR1, 0, 
				"No RC CAPACITY given (-1 for unlimited)");
		err++;
	}

	/*
	 *  make sure the things that should be unique are
	 */
	for (itmp=0; itmp<RCTotal; itmp++) {
		if (rcp->rc_number != -1) {
			if (RCArray[itmp].rc_number == rcp->rc_number) {
				nLogMsg(EL_CONSOLE0|EL_ERROR1, 0, 
					"Duplicate RC NUMBER %d", 
					rcp->rc_number);
				err++;
			}
		}
		if (rcp->rc_name[0] != '\0') {
			if (!strcmp(RCArray[itmp].rc_name, rcp->rc_name)) {
				nLogMsg(EL_CONSOLE0|EL_ERROR1, 0, 
					"Duplicate RC NAME %s", 
					rcp->rc_name);
				err++;
			}
		}
		if (rcp->rc_id[0] != '\0') {
			if (!strcmp(RCArray[itmp].rc_id, rcp->rc_id)) {
				nLogMsg(EL_CONSOLE0|EL_ERROR1, 0, 
					"Duplicate RC ID %s", 
					rcp->rc_id);
				err++;
			}
		}
	}
	if (err)
		nLogMsg(EL_FATAL|EL_CONSOLE0, id, "RC configuration problem%s",
					(err > 1 ? "s" : ""));

	for (itmp=0; itmp<rcp->rc_nmedia; itmp++) {
		nLogMsg(EL_STATUS1|EL_NOTIME, 0, 
			"Configured `%s' (%s), number %d, with %d %s drive%s",
			rcp->rc_name, rcp->rc_id, rcp->rc_number,
			rcp->rc_devs[itmp].d_nconfigdev,
			vMediumName(rcp->rc_devs[itmp].d_medium),
			(rcp->rc_devs[itmp].d_nconfigdev == 1  ? "" : "s") );
	}
	if (rcp->rc_path == NULL)
		nLogMsg(EL_STATUS1|EL_NOTIME, 0, 
				"(%s: No RC PATH given - will not be forked)",
				rcp->rc_id);

	RCTotal++;
}


  static RPRC_t *
GetRcp(char *rcname)		/* get the rcp based on RC name or number */
{
	RPRC_t	*rcp = RCArray;
	char	*cp;
	int	i;

	if (rcp == NULL)
		nLogMsg(EL_FATAL, 0, "RCArray not initted");

	/*
	 *  see if rcname is all digits, if so, consider it the rc number
	 *	else it is an RC ID string
	 */

	for (cp = rcname; *cp; cp++)
		if (!isdigit(*cp))
			break;
	if (*cp == '\0')
		return(RCInfo(atoi(rcname)));

	/*
	 *  look up RC ID
	 */

	for (i=0; i<RCTotal; i++, rcp++)
		if (strcmp(rcp->rc_id, rcname) == MATCH)
			return(rcp);
	return (NULL);
}

 static void
ReadMoves()
{
	RPRC_t	*rcp1, *rcp2;

	if (MoveArray != NULL) {
		inerr("Only 1 MOVES section allowed","");
	} else {
		if (RCTotal < 2)
			inerr("Must be > 1 RC to define moves","");

		/*
		 *  total possible moves in both directions
		 */
		MoveSize = RCTotal * RCTotal;
		MoveArray = (MoveTable_t *) nAlloc(MoveSize * 
							sizeof(MoveTable_t));
	}
	while (1) {
		/*
		 *  get pairs of tokens for RCs
		 */
		gettok(EOFBOMB);
		if (!strcmp(tok, "END_MOVES"))
			break;
		if (MoveTotal == MoveSize)
			inerr("Too many moves defined - max is #RCs * #RCs","");
		if (iskeywd())
			inerr("Got keyword - ",
				"expected RC number or id defined above");
		if ((rcp1 = GetRcp(tok)) == NULL)
			inerr("Expected RC number or id defined above","");
		gettok(EOFBOMB);
		if (iskeywd())
			inerr("Got keyword - ",
				"expected RC number or id defined above");
		if ((rcp2 = GetRcp(tok)) == NULL)
			inerr("Expected RC number or id defined above","");
		MoveArray[MoveTotal].source = rcp1->rc_number;
		MoveArray[MoveTotal].dest = rcp2->rc_number;
		MoveTotal++;
	}
}

/*******************************************************************
			CONFIG ACCESS ROUTINES
 *******************************************************************/

  RPRC_t *
RCInfo(int number)
{
	int	rcindex;
	RPRC_t	*rcp = RCArray;

	for (rcindex=0; rcindex<RCTotal; rcindex++,rcp++) {
		if (rcp->rc_number == number)
			return(rcp);
	}
	return(NULL);
}

 void
PrintRCArray(int lev, char *id)
{
	RPRC_t	*rcp;
	int	rcindex;

	if (!(lev & nLogLevel()))
		return;

	nLogMsg(EL_ALWAYS|EL_NOTIME, id,
			"Repository Controller list -------------------------");
	nLogMsg(EL_ALWAYS|EL_NOTIME, NULL, 
			"#    RC     quota medium |var|all| avail * devmap");
	for (rcp=RCArray, rcindex=0; rcindex<RCTotal; rcindex++,rcp++) {
		char	dmstr[LINELEN * 3];
		int	i;

		for (i=0; i<rcp->rc_nmedia; i++) {
			Media_t	*mp = &rcp->rc_devs[i];

			strcpy(dmstr, vDevMapStr(&rcp->rc_avail[i]));
			strcat(dmstr, " * ");
			strcat(dmstr, vDevMapStr(&mp->d_devmap));

			nLogMsg(EL_ALWAYS|EL_NOTIME, NULL, 
				"%-2d %s%-6.6s %s   %c   |%3d|%3d| %s",
		  		rcp->rc_number, 
				(rcp->rc_srv.s_sd==-1 ? "  " : "X "),
		  		rcp->rc_id, 
				(rcp->rc_flags & RCF_HASQUOTA ? 'y' : 'n'),
				vMediumName(mp->d_medium),
		  		mp->d_nvaryondev, mp->d_mountdev, dmstr);
		}
	}
	nLogMsg(EL_ALWAYS|EL_NOTIME, id, 
			"----------------------------------------------------");
}

RCVolTypeProblem(RPRC_t *rcp, int voltype)
{
	int	medium;

	medium = vVolTypeMedium(voltype);
	if (medium == -1) {
		VMError = E_ADMIN;
		nLogMsg(EL_CODE|EL_CONSOLE0|EL_ERROR1, 0, 
				"volume type %d not defined", voltype);
		return(1);
	}
	if (rcp->rc_media & medium)
		return(0);

	VMError = E_USER;
	nLogMsg(EL_CODE, 0, "RC %s is not configured to hold volume type %s",
			rcp->rc_id, vVolTypeName(voltype));
	return(1);
}

MoveProblem(RPRC_t *srcrcp, RPRC_t *destrcp, int voltype)
{
	int	i;

	if (RCVolTypeProblem(destrcp, voltype))
		return(1);
	for (i=0; i<MoveTotal; i++) {
		if (MoveArray[i].source == srcrcp->rc_number  &&
		    MoveArray[i].dest == destrcp->rc_number)
			return(0);
	}
	return(1);
}

RCMedium(RPRC_t *rcp, int medium)
{
	int	i;

	if (medium == -1)
		return(-1);

	for (i=0; i<rcp->rc_nmedia; i++)
		if (rcp->rc_devs[i].d_medium == medium)
			return(i);
	return(-1);
}

/* 
 *  RPConfig() - read in config file 
 */
 void
RPConfig()
{
	char		*id = "RPConfig";

	/*
	 *  open config file, read 1st line, and make initial RC array
	 */
	if ((fp = fopen(vVoldConfig, "r")) == NULL) {
		nLogMsg(EL_FATAL|EL_SYS, id,
		  "can't open RC configuration file (%s)", vVoldConfig);
	}
	nLogMsg(EL_STATUS1, id, "Config file: %s", vVoldConfig);
	readln(1);
	RCArray = (RPRC_t *) nAlloc(RCarraysize * sizeof(RPRC_t));

	/*
	 *  parse config file
	 */
	while (1) {
		gettok(GETNEWLINE);
		if (tok[0] == '\0')
			break;
		if (!strcmp(tok, "BEGIN_HOSTS"))
			ReadHosts();
		else if (!strcmp(tok, "BEGIN_RC"))
			ReadRC();
		else if (!strcmp(tok, "BEGIN_MOVES"))
			ReadMoves();
		else
		    inerr("expected BEGIN_HOSTS, BEGIN_RC or BEGIN_MOVES","");
	}
	fclose(fp);
}

/*
**	SetVolndMapON() - update drive bitmap for volnd
*/
 void
SetVolndMapON(int host)
{
	int	i;

	for (i=0; i<DriveTotal; i++) {
		if (DriveArray[i].de_host == host)
			DM_SET(i, DMVolnd);
	}
}
/*
**	SetVolndMapOFF() - update drive bitmap for volnd
*/
 void
SetVolndMapOFF(int host)
{
	int	i;

	for (i=0; i<DriveTotal; i++) {
		if (DriveArray[i].de_host == host)
			DM_CLR(i, DMVolnd);
	}
}

