
/*
 *  rp_q.c: Volume Manager Volume Request Queue Processing.
 *
 *	After volume requests have been made active (volume
 *	reservations w/out specific requests for action are 
 *	inactive), they are appended to the VolReqQ (in addition 
 *	to various other queues they may be in), which maintains 
 *	their priority order. When a state-changing event occurs, 
 *	this queue is examined by ProcVolReqQ() to see if any 
 *	waiting requests can be dispatched.
 *
 * 	..maybe have per-RC VolReq queue? -no, since vols can be 
 *	waiting on groups of vols in a reservation, not assoc w/ 
 *	just 1 rc.
 *
 *	Bill Ross, NASA Ames, October 1995
 */

#include "rp_space.h"

extern int	UnmountWaitTime;
extern int	Timeout;
extern time_t	Now;

int		activerequests;

static int	drivesneeded;


/*
 *  RCTimeoutMount() - check how long the request is outstanding
 *	if the RC has a positive mount timeout
 */
 static void
RCTimeoutMount(VolReq_t *vrp)
{
	char		*id = "RCTimeoutMount";
	RPRC_t		*rcp = vrp->vr_rcp;
	time_t		wait;

	if (rcp->rc_mounttimeout < 1)
		return;
	wait = Now - vrp->vr_time;
	if (wait > rcp->rc_mounttimeout)
		nLogMsg(EL_CONSOLE0|EL_ERROR2, id, 
			"TIMEOUT %s: %.*s waiting %s",
			rcp->rc_id, VSNLEN, (char *)&vrp->vr_vsn, 
			nas_time(wait));
}

/*
 *  VNTimeoutMount() - check how long req is outstanding in volnd
 */
 static void
VNTimeoutMount(VolReq_t *vrp)
{
	char		*id = "VNTimeoutMount";
	time_t		wait;

	wait = Now - vrp->vr_time;
	/* TODO - an arbitrary amount for now.. tune this */
	if (wait > 300)
		nLogMsg(EL_CONSOLE0|EL_ERROR2,
			id, "TIMEOUT host %s: %.*s waiting %s",
		  	vrp->vr_ndhostp->h_name, VSNLEN, (char *)&vrp->vr_vsn, 
		  	nas_time(wait));
}

/*
 *  ZeroVaryCounts() - initialize array keeping track of need for 'vary on'
 *	msgs to ops
 */
static int	*rccount, parmounts;
 static void
ZeroVaryCounts()
{
	int 	i;

	for (i=0; i<RCTotal*RC_MEDIA_TYPES; i++)
		rccount[i] = 0;
	parmounts = 0;
}

 void
InitVaryCounts()
{
	rccount = (int *) nAlloc(RCTotal*RC_MEDIA_TYPES*sizeof(int));
	ZeroVaryCounts();
}


/*
 *  TellOperToVary() - called when a volreq can't be satisfied because
 *	there are no devices available.  See if any applicable devs
 *	aren't varied on.  If so, tell RP operator.
 *
 *  	Maintain list of RCs needing messages and print them out all 
 *	at once when vrp is null.
 */
 static void
TellOperToVary(VolReq_t *vrp)
{
	char		*id = "TellOperToVary";
	int		rcno, prmsg, medindex;
	RPRC_t		*rcp;

	if (vrp) {
		DevMap_t	map, *mp2;

		rcp = vrp->vr_rcp;

		medindex = RCMedium(rcp, 
				      vVolTypeMedium(vrp->vr_info.vi_voltype));

		/*
		 *  TODO - if rc_avail is updated to reflect vary,
		 *	2-stage AND not necc
		 */
		map = *vDMand(&vrp->vr_devtotry, &rcp->rc_avail[medindex]);
		mp2 = vDMand(&map, &DMVary);
		if (nLogLevel() > EL_DEBUG2) {
			char	buf1[LINELEN * 2], buf2[LINELEN * 2];
			char	buf3[LINELEN * 2];

			strcpy(buf1, vDevMapStr(&DMVary));
			strcpy(buf2, vDevMapStr(&rcp->rc_avail[medindex]));
			strcpy(buf3, vDevMapStr(&vrp->vr_devtotry));
			nLogMsg(EL_DEBUG3, id, 
			    "%.*s_%d\nvary  %s\navail %s\ntotry %s\n      %s", 
				VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq,
				buf1, buf2, buf3,
				vDevMapStr(mp2));
		}
		if (!vDMcnt(mp2)) {
			nLogMsg(EL_DEBUG3, id, "count it");
			if (!drivesneeded)
				ZeroVaryCounts();
			drivesneeded++;
			rccount[rcp->rc_index*RC_MEDIA_TYPES + medindex]++;
			if (vrp->vr_ridp->rid_parallel)
				parmounts++;
		}
		return;
	}
	nLogMsg(EL_DEBUG3, id, "print");

	/*
	 *  vrp == NULL.  this is the signal to go ahead and print
	 *	messages about all RCs needing attention
	 */
	prmsg = 0;
	for (rcp=RCArray,rcno=0; rcno<RCTotal; rcno++,rcp++) {
		int	medindex;

		for (medindex=0; medindex<rcp->rc_nmedia; medindex++) {
		    int	index = rcno * RC_MEDIA_TYPES + medindex;

		    if (rccount[index]  &&
					rcp->rc_devs[medindex].d_nvaryondev < 
					rcp->rc_devs[medindex].d_nconfigdev) {
			char	*sfmt;

			prmsg = 1;
			sfmt = (rccount[index] == 1) ? "" : "s";
			nLogMsg(EL_ERROR2, NULL,
			    "%d %s %s device%s needed for pending mount%s.",
			    rccount[index], rcp->rc_id, 
			    vMediumName(rcp->rc_devs[medindex].d_medium), 
			    sfmt, sfmt);
		    }
		}
	}
	if (prmsg) {
	  nLogMsg(EL_ERROR2, NULL,
	  "Please fix drives if necessary and issue a VOLVARY ON command");
	  if (parmounts)
	  	nLogMsg(EL_ERROR2, NULL,
			"Parallel mount(s) blocked - could be deadlock");
	}

	/*
	 *  reset the counts to 0 for the next round
	 */
	ZeroVaryCounts();
}

 static void
SelectMountWait(VolReq_t *vrp)
{
	char		*id = "SelectMountWait";
	Rid_t		*ridp = vrp->vr_ridp;
	RPRC_t		*rcp = vrp->vr_rcp;
	int		medindex, rw = vrp->vr_voflag & VO_WRITE ? 1 : 0;
	RidRCMed_t	*ridrcmed;
	int		unsafe1, unsafe2;

	if (ridp->rid_active) {
		nLogMsg(EL_STATUS1, id, "%.*s: skipping for busy select",
						VSNLEN, &vrp->vr_vsn);
		return;
	}

	medindex = RCMedium(rcp, vVolTypeMedium(vrp->vr_info.vi_voltype));
	ridrcmed = &ridp->rid_RCMed[rcp->rc_index * RC_MEDIA_TYPES + medindex];

	nLogMsg(EL_DEBUG3, id, "type %s medium %s, medindex %d", 
		vVolTypeName(vrp->vr_info.vi_voltype),
		vMediumName(vVolTypeMedium(vrp->vr_info.vi_voltype)), medindex);

	/*
	 *  make sure this vsn wouldn't hang any reservations
	 *	(just this vsn for select mount, so don't
	 *	volintersect all vsns in this rid)
	 */
	if (InAnotherActiveRid(vrp, ridp))
		return;

	/*
	 *  quick drive availability check
	 */
	if (rcp->rc_devs[medindex].d_mountdev >= 
				rcp->rc_devs[medindex].d_nvaryondev) {
		if (rcp->rc_devs[medindex].d_nvaryondev < 
					rcp->rc_devs[medindex].d_nconfigdev) {
			TellOperToVary(vrp);
		}
nLogMsg(EL_STATUS1, id, "mountdev %d  varyon %d",
rcp->rc_devs[medindex].d_mountdev,
rcp->rc_devs[medindex].d_nvaryondev);
		vrp->vr_state = VS_DRIVEWAIT;
		nLogMsg(EL_DEBUG2, id, "%.*s_%d no drive in %s",
				VSNLEN, (char *)&vrp->vr_vsn, 
				vrp->vr_rpseq, rcp->rc_id);
		return;
	}

	/*
	 *  Make sure an acceptable drive is available
	 */
	if (DriveMapCheck(vrp, rcp, medindex))
		return;

	/*
	 *  need an implicit reservation
	 */
	nLogMsg(EL_DEBUG3, id, "  try for implicit soft drive allocation");


	unsafe1 = Unsafe();
	ridrcmed->rr_softalloc = 1;
	ridrcmed->rr_softused = 1;
	ridrcmed->rr_implicit = 1;
	ridp->rid_active = 1;

	/*
	 *  compare prospective state to previous one to see
	 *	if it's any more unsafe
	 */
	unsafe2 = Unsafe();
	nLogMsg(EL_DEBUG2, id, "--postcheck %d", unsafe2);

	if (unsafe2 > unsafe1) {
		nLogMsg(EL_DEBUG2, id, 
				"%.*s_%d rid waiting for soft drive - unsafe",
				VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq);
		ridrcmed->rr_implicit = 0;
		ridrcmed->rr_softalloc = 0;
		ridrcmed->rr_softused = 0;
		ridp->rid_active = 0;
		vrp->vr_state = VS_RESERVEMOUNTWAIT;
		return;
	}

	nLogMsg(EL_STATUS2, 0, "MOUNT %s %.*s_%d", 
		rcp->rc_id, VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq);
	if (RCMount(vrp)) {
		nLogMsg(EL_DEBUG2, id, "rcmount failed %.*s_%d ", 
			VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq);
		ridrcmed->rr_implicit = 0;
		ridrcmed->rr_softalloc = 0;
		ridrcmed->rr_softused = 0;
		ridp->rid_active = 0;
		return;		/* failed */
	}
	vrp->vr_state = VS_RCWAIT;
	vrp->vr_qtime += Now - vrp->vr_time;
	vrp->vr_time = Now;

	if (DBMountInit((char *)&vrp->vr_vsn, vrp->vr_cmd, rw)) { 
		/*
		 *  TODO should err to client but
		 *	keep vrp around to cancel mount
		 *	when it completes, or maybe shut
		 *	down altogether..
		 */
		nLogMsg(EL_FATAL, id, "dbase err, shutting down");

		ridrcmed->rr_implicit = 0;
		ridrcmed->rr_softalloc = 0;
		ridrcmed->rr_softused = 0;
		ridp->rid_active = 0;
		ErrOutVolReq(vrp, id);
		return;
	}
	rcp->rc_devs[medindex].d_mountdev++;
	nLogMsg(EL_DEBUG2, id, "all ok " );
}

/*
 *  MountWait() - request is waiting to be sent to RC
 */
 static void
MountWait(VolReq_t *vrp)
{
	char		*id = "MountWait";
	Rid_t		*ridp = vrp->vr_ridp;
	RPRC_t		*rcp = vrp->vr_rcp;
	int		medindex, rw = vrp->vr_voflag & VO_WRITE ? 1 : 0;
	RidRCMed_t	*ridrcmed;
	int		unsafe1, unsafe2;

	nLogMsg(EL_DEBUG3, id, "%s_%d ", 
			vsnstr((char *)&vrp->vr_vsn), vrp->vr_rpseq);

	if (ridp->rid_parallel == 2) {
		SelectMountWait(vrp);
		return;
	}

	/* TODO - see if RC is up */

	if (ridp->rid_RCMed == NULL) {
		nLogMsg(EL_ERROR1, id, "INTERNAL ERROR: rid_RCMed is null");
		return;
	}
	medindex = RCMedium(rcp, vVolTypeMedium(vrp->vr_info.vi_voltype));
	ridrcmed = &ridp->rid_RCMed[rcp->rc_index * RC_MEDIA_TYPES + medindex];

	nLogMsg(EL_DEBUG3, id, "type %s medium %s, medindex %d", 
		vVolTypeName(vrp->vr_info.vi_voltype),
		vMediumName(vVolTypeMedium(vrp->vr_info.vi_voltype)), medindex);

	/*
	 *  see if vol can be used now
	 */
	if (ridp->rid_active) {
		/*
		 *  Reservation is active, so just need to see if
		 *	vol is already active within it.
		 */
		VolReq_t	*vrp2;

		nLogMsg(EL_DEBUG3, id, "  in active rid");
		vrp2 = VolActiveInRid(vrp);

		if (vrp2 != NULL) {
			nLogMsg(EL_DEBUG2, id, "%.*s_%d active in rid", 
					VSNLEN, (char *)&vrp->vr_vsn, 
					vrp->vr_rpseq);
			vrp->vr_state = VS_RESERVEMOUNTWAIT;

			/*
			 *  Test other client
			 */
			if (vrp2->vr_clientp != NULL) {

				if (vrp2->vr_clientp == vrp->vr_clientp) {

					/*
					 *  Client is own enemy.
					 */

					VMError = E_USER;
					nLogMsg(EL_CODE|EL_ERROR2, id,
					  "Client mounting %.*s_%d twice",
					  VSNLEN, (char *)&vrp->vr_vsn,
					  vrp->vr_rpseq);
					ErrOutVolReq(vrp, id);
				} 
			}
			return;
		}
		nLogMsg(EL_DEBUG3, id, "  vsn not active");
	} else {

		/*
		 *  Need to see if whole reservation can become active.
		 */
		if (VolIntersect(ridp)) {
			vrp->vr_state = VS_RESERVEMOUNTWAIT;
			nLogMsg(EL_DEBUG2, id, "intersect failed");
			nLogMsg(EL_DEBUG2, id, "%.*s_%d rid vol use blocked",
				VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq);
			return;
		}
	}

	/*
	 *  Quick pre-checks on possibility of drive availability.
	 */
	if (rcp->rc_devs[medindex].d_nconfigdev == 0) {
		VMError = E_ADMIN;
		nLogMsg(EL_CODE|EL_ERROR2|EL_CONSOLE0, id,
				"Can't mount %.*s: RC %s has no %s drives",
				VSNLEN, (char *)&vrp->vr_vsn, rcp->rc_id, 
				vMediumName(rcp->rc_devs[medindex].d_medium));
		ErrOutVolReq(vrp, id);
		return;
	}
	if (rcp->rc_devs[medindex].d_mountdev >= 
				rcp->rc_devs[medindex].d_nvaryondev) {
		/*
		 *  all varied on drives are mounted
		 */
		nLogMsg(EL_DEBUG2, id, "all drives already mounted");
		if (ridp->rid_active) {
			if (ridp->rid_id[0])
			    nLogMsg(EL_DEBUG3, id, "rid %.*s active", 
						RIDLEN, ridp->rid_id);
			else
			    nLogMsg(EL_DEBUG3, id, "rid (implicit) active");
			if ((rcp->rc_devs[medindex].d_nvaryondev < 
			     rcp->rc_devs[medindex].d_nconfigdev)   &&
			    (ridrcmed->rr_hardalloc - 
			     ridrcmed->rr_hardused > 0  ||
			     ridrcmed->rr_softalloc - 
			     ridrcmed->rr_softused > 0)) {
			    nLogMsg(EL_ALWAYS, id, 
			    "RESERVATION BLOCKED, %.*s NEEDS %s DRIVE in %s",
				VSNLEN, (char *)&vrp->vr_vsn,
				vMediumName(rcp->rc_devs[medindex].d_medium),
				rcp->rc_id);
			    vrp->vr_state = VS_DRIVEWAIT;
nLogMsg(EL_STATUS1, id, "all drives already mounted 1");
			    TellOperToVary(vrp);
			} else if (ridrcmed->rr_hardalloc == 0  &&
			    ridrcmed->rr_softalloc == ridrcmed->rr_implicit) {
				/*
				 *  waiting for an implicit alloc
				 */
				vrp->vr_state = VS_DRIVEWAIT;
nLogMsg(EL_STATUS1, id, "all drives already mounted 2");
			} else {
				/*
				 *  waiting on an explicitly alloc'd drive
				 */
				nLogMsg(EL_DEBUG2, id, "drive alloc'd/wait");
				vrp->vr_state = VS_RESERVEMOUNTWAIT;
			}
		} else {
			if (rcp->rc_devs[medindex].d_nvaryondev < 
					rcp->rc_devs[medindex].d_nconfigdev) {
				TellOperToVary(vrp);
			}
			vrp->vr_state = VS_DRIVEWAIT;
nLogMsg(EL_STATUS1, id, "all drives already mounted 3");
		}
		nLogMsg(EL_DEBUG2, id, "%.*s_%d no drive in %s",
				VSNLEN, (char *)&vrp->vr_vsn, 
				vrp->vr_rpseq, rcp->rc_id);
		return;
	}
	nLogMsg(EL_DEBUG3, id, "  drive not unavail %.*s_%d ", 
			VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq);

	/*
	 *  Make sure an acceptable drive is available
	 */
	if (DriveMapCheck(vrp, rcp, medindex))
		return;

	/*
	 *  Get a drive for client - use hard allocation if available.
	 *	Safety is tested for hard allocations and also
	 *	soft allocations in active reservations even though
	 *	on the face of it this should not be necessary, because
	 *	drives may have been varied off when the voldaemon has
	 *	no reserve drive to bring the pool back up. If this
	 *	condition were tracked explicitly, Unsafe() would
	 *	only need to be called every time when the condition
	 *	obtained, otherwise not for hard allocs and soft
	 *	allocs in active rids.
	 */

	if (ridrcmed->rr_hardalloc > ridrcmed->rr_hardused) {

		/*
		 *  a hard allocation is available; check safety 
		 *	in case drives have been varied off
		 */
		nLogMsg(EL_DEBUG3, id, "checking hard reservation");

		if (Unsafe()) {

			/*
			 *  can't do any more - softalloc would fail too
			 */

			nLogMsg(EL_DEBUG2, id, "%.*s_%d no hard alloc drive",
				VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq);
			vrp->vr_state = VS_DRIVEWAIT;
nLogMsg(EL_STATUS1, id, "unsafe");
			TellOperToVary(vrp);
			return;
		}

		/*
		 *  hard allocation available & safe - finish up
		 */
		
		if (RCMount(vrp)) {
			/*
			 *  failed - retry later
			 */
			vrp->vr_state = VS_MOUNTWAIT;
			return;
		}
		vrp->vr_state = VS_RCWAIT;
		vrp->vr_qtime += Now - vrp->vr_time;
		vrp->vr_time = Now;
		if (DBMountInit((char *)&vrp->vr_vsn, vrp->vr_cmd, rw)) {
			/*
			 *  dbase problem but mount was sent -
			 *	TODO should err to client but
			 *	keep vrp around to cancel mount
			 *	when it completes, or maybe shut
			 *	down altogether..
			 */
			nLogMsg(EL_FATAL, id, "dbase err, shutting down");
			ErrOutVolReq(vrp, id);
			return;
		}

		ridp->rid_active++;
		ridrcmed->rr_hardused++;
		/* 
		 *  (don't increase rcp->rc_mountdev vars for hard alloc
		 *	since they are incremented at alloc time)
		 *	TODO - consider calling it rcp->rc_allocdev
		 */
		return;
	}

	/*
	 *  soft drive allocation; count degree of unsafeness,
	 *	then set prospective state
	 */

	unsafe1 = Unsafe();
	nLogMsg(EL_DEBUG2, id, "--precheck %d", unsafe1);
	if (ridrcmed->rr_softalloc > ridrcmed->rr_softused) {
		/*
		 *  soft reservation already made & available
		 */
		nLogMsg(EL_DEBUG3, id, "trying using soft drive allocation");
		ridrcmed->rr_softused++;
	} else {
		nLogMsg(EL_DEBUG2, id, "  try implicit");

		/*
		 *  need an implicit reservation
		 */
		nLogMsg(EL_DEBUG3, id, 
			"  try for implicit soft drive allocation");

		if (SoftReserve(ridp, rcp, medindex, 1)) {
			nLogMsg(EL_STATUS2, id, 
				"%.*s_%d no implicit reservation possible",
				VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq);
			if (ridp->rid_active) {
				/*
				 *  trying to add a vsn to an active
				 *	resevation - this could be a 
				 *	deadlock situation
				 */
				ErrOutVolReq(vrp, id);
			} else {
				vrp->vr_state = VS_DRIVEWAIT;
nLogMsg(EL_STATUS1, id, "SoftReserve");
			}
			return;
		}
		ridrcmed->rr_softused++;
		ridrcmed->rr_implicit++;

	}
	ridp->rid_active++;

	/*
	 *  compare prospective state to previous one to see
	 *	if it's any more unsafe
	 */
	unsafe2 = Unsafe();
	nLogMsg(EL_DEBUG2, id, "--postcheck %d", unsafe2);

	if (unsafe2 > unsafe1) {
		nLogMsg(EL_DEBUG2, id, 
				"%.*s_%d rid waiting for soft drive - unsafe",
				VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq);
		ridrcmed->rr_softused--;
		ridp->rid_active--;
		vrp->vr_state = VS_RESERVEMOUNTWAIT;
		return;
	}

	nLogMsg(EL_STATUS2, 0, "MOUNT %s %.*s_%d", 
					rcp->rc_id, 
					VSNLEN, (char *)&vrp->vr_vsn, 
					vrp->vr_rpseq);
	if (RCMount(vrp)) {
		nLogMsg(EL_DEBUG2, id, "rcmount failed %.*s_%d ", 
			VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq);
		ridrcmed->rr_softused--;
		ridp->rid_active--;
		if (ridrcmed->rr_implicit > 0) {
			/*
			 *  roll back the implicit reservation and let
			 *	it be recreated if necessary.
			 */
			ridrcmed->rr_softalloc--; /* TODO - make sure ok */
			ridrcmed->rr_implicit--;
		}
		return;		/* failed */
	}
	vrp->vr_state = VS_RCWAIT;
	vrp->vr_qtime += Now - vrp->vr_time;
	vrp->vr_time = Now;

	if (DBMountInit((char *)&vrp->vr_vsn, vrp->vr_cmd, rw)) {
		/*
		 *  TODO should err to client but
		 *	keep vrp around to cancel mount
		 *	when it completes, or maybe shut
		 *	down altogether..
		 */
		nLogMsg(EL_FATAL, id, "dbase err, shutting down");

		ridp->rid_active--;
		ridrcmed->rr_softused--;
		if (ridrcmed->rr_implicit > 0) {
			/*
			 *  undo the implicit reservation 
			 */
			ridrcmed->rr_softalloc--;  /* TODO - make sure ok */
			ridrcmed->rr_implicit--;
		}
		ErrOutVolReq(vrp, id);
		return;
	}
	rcp->rc_devs[medindex].d_mountdev++;
	nLogMsg(EL_DEBUG2, id, "all ok " );
}

/*
 *  UnmountWait() - see if vol has waited the prescribed grace
 *	period before unmounting
 */
 static void
UnmountWait(VolReq_t *vrp)
{
	char		*id = "UnmountWait";

	if (vrp->vr_time + UnmountWaitTime < Now) {
		Timeout--;
		DeleteNode(vrp, 1);
		nLogMsg(EL_DEBUG2, id, "discarded unmntwait %.*s_%d",
			VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq);
	}
}

/*
 *  SaveMove() - write packet to file
 */
 static void
SaveMove(vmpHdr_t *hp)
{
	/* maybe this'll be worth doing.. */
}

/*
 *  MoveWait() - see if vol can be moved now
 */
 static void
MoveWait(VolReq_t *vrp)
{
	char		*id = "MoveWait";
	Rid_t		*ridp = vrp->vr_ridp;
	RPRC_t		*dstrcp = vrp->vr_dstrcp;
	RPRC_t		*srcrcp, *CheckRC();
	VInfo_t		*vip;
	int		i, nbatch;
	vmpHdr_t	*hp;
	vmpMove_t	*mp;
	vsnseq_t	*vsnseqp;
	VolReq_t	*vrp2, *vrnext;

        /*
         *  Get VDR from dbase (checking permissions)
         *      if NULL, then error info is set in globals
         */

	nLogMsg(EL_DEBUG2, id, "%.*s_%d",
		VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq);

        vip = DBMoveQuery((char *)&vrp->vr_vsn, vrp->vr_uid, dstrcp);
        if (vip == NULL) {
                ErrOutVolReq(vrp, id);
		return;
        }

	/*
	 *  see if both RCs up 
	 *	(TODO: in some cases (like ACS Eject) only 1 
	 *	needs to be up)
	 */

	if ((srcrcp = CheckRC(vip->vi_srcrc, (char *)&vrp->vr_vsn, id))==NULL) {
		ErrOutVolReq(vrp, id);
		return;
	}
	vrp->vr_rcp = srcrcp;
	if (dstrcp->rc_srv.s_sd == -1) {
		VMError = E_UNAVAIL;
               	nLogMsg(EL_STATUS1|EL_CODE|EL_CONSOLE0, id,
                 "cannot perform operation (vsn=%.*s_%s): destination %s down",
                 VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq, dstrcp->rc_id);
		ErrOutVolReq(vrp, id);
		return;
	}

	/*
	 *  see if vol can be used now
	 */

	if (ridp->rid_active) {
		/*
		 *  Reservation is active, so just need to see if
		 *	vol is already active within it.
		 */
		if (VolActiveInRid(vrp)) {
			vrp->vr_state = VS_RESERVEMOVEWAIT;
			return;
		}
	} else {

		/*
		 *  Need to see if whole reservation can become active
		 */
		if (VolIntersect(ridp)) {
			vrp->vr_state = VS_RESERVEMOVEWAIT;
			return;
		}
	}

	/* 
	 *  copy vinfo from MoveQuery() - needs to be done b4 MoveBatchWait
	 */
	vrp->vr_info = *vip;            

	if (DBMoveInit((char *)&vrp->vr_vsn, vrp->vr_uid, vrp->vr_dstrcp)) {
		ErrOutVolReq(vrp, id);
		return;
	}

	/*
	 *  find out how many other requests can be batched w/ this
	 */
	nbatch = 1 + MoveWaitBatch(vrp);
	nLogMsg(EL_DEBUG3, id, "batch total %d", nbatch);

	/*
	 *  build batch request
	 */
	hp = (vmpHdr_t *) nAlloc(PKTSIZ(vmpMove_t) + nbatch * sizeof(vsnseq_t));
	hp->vh_magic = VMP_MAGIC;
	hp->vh_type = M_MOVE;
	hp->vh_seq = vGetSeq();
	hp->vh_size = PadSize(sizeof(vmpMove_t)) + nbatch * sizeof(vsnseq_t);
	hp->vh_count = nbatch;
	mp = MSGPTR(hp, vmpMove_t);
	mp->vmv_srcrc = srcrcp->rc_number;
	mp->vmv_dstrc = dstrcp->rc_number;
	vsnseqp = (vsnseq_t *) ((char *)mp + PadSize(sizeof(vmpMove_t)));
	for (i=0, vrp2=vrp; i<nbatch; i++, 
				      vsnseqp++, 
				      vrp2=VRNEXT(vrp2->vr_scrq)) {
		vsnseqp->r_seq = vrp2->vr_rpseq;
		vsnseqp->r_vsn = vrp2->vr_vsn;
	}

	/*
	 *  send request to both RCs 
	 */
	if (Snd2Srv(hp, &srcrcp->rc_srv, 1) == E_NOERROR) {
		if (Snd2Srv(hp, &dstrcp->rc_srv, 1) != E_NOERROR) {
			nLogMsg(EL_CONSOLE0, id, 
			    "PROTOCOL PROBLEM!!! - move sent to source %s",
			    "but destination send failed - notify programmer");
			/*
			 *  HACK - at the NAS site, the main (99.99%) move
			 *	is ACS->vault, & the vault ignores the
			 *	msg anyway.
			 */
			VMError = E_NOERROR;
		}
	}
	/*
	 *  if send error, fail batch
	 */
	if (VMError != E_NOERROR) {
		for (i=0, vrp2=vrp; i<nbatch; i++, vrp2=vrnext) {
			vrnext = VRNEXT(vrp2->vr_scrq); /* since DQing */
			DBMoveFail((char *)&vrp2->vr_vsn, vrp2->vr_uid,
							vrp2->vr_dstrcp);
			ErrOutVolReq(vrp2, id);
		}
		return;
	}

	/*
	 *  success: update reqs and dissolve scratch q
	 */
	for (i=0, vrp2=vrp; i<nbatch; i++, vrp2=vrnext) {
		vrnext = VRNEXT(vrp2->vr_scrq);		/* since DQing */

		/*
		 *  make sure rid active and set request state
		 */
		vrp2->vr_ridp->rid_active++;
		vrp2->vr_state = VS_MOVING;
		vrp->vr_qtime += Now - vrp->vr_time;
		vrp->vr_time = Now;

		/*
		 *  undo scratch queue
		 */
		vrp2->vr_headvrp = NULL;
		vrp2->vr_scrq.q_next = NULL;
		vrp2->vr_scrq.q_prev = NULL;
	}
	/*
	 *  save the packet in case we need to restart after
	 *	a shutdown
	 */
	SaveMove(hp);

	/*
	 *  free that custom-built pkt
	 */
	nFree(hp);
}
/*
 *  ProcVolReq() - see if anything can be done for this request
 */

ProcVolReq(VolReq_t *vrp)
{
	char	*id = "ProcVolReq";
	int	ret = 1;

	/*
	 *  handle this request 
	 */
	switch(vrp->vr_state) {
	    case VS_MOUNTWAIT:
	    case VS_DRIVEWAIT:
	    case VS_RESERVEMOUNTWAIT:
		MountWait(vrp);
		break;
	    case VS_RCWAIT:
	    case VS_RCWAIT2:
		/*
		 *  Requests are taken out of these states by "mounted"
		 *	or "unmounted" msgs from an RC 
		 */
		RCTimeoutMount(vrp);
		break;
	    case VS_CREATENODE:
	    case VS_DELETENODE:
		/*
		 *  Requests are taken out of these states by "node 
		 *	created/deleted" from a volnode daemon.
		 */
		VNTimeoutMount(vrp);
		break;
	    case VS_UNMOUNTWAIT:
		UnmountWait(vrp);
		break;
	    case VS_MOVEWAIT:
	    case VS_RESERVEMOVEWAIT:
		MoveWait(vrp);
		break;
	    case VS_IDLE:
		ret = 0;
		break;
	    default:
		nLogMsg(EL_DEBUG3, id, "skipping state %s\n",
				vReqStateStr(vrp->vr_state));
		break;
	}
	return(ret);
}


  int
DriveMapCheck(VolReq_t *vrp, RPRC_t *rcp, int medindex)
{
	char		*id = "DriveMapCheck";
	DevMap_t	mp2, *mp3, map;
	int		variedon;

	/*
	 *  Make sure an acceptable drive is available (e.g.
	 *	one not used in a prev mount if this is a requeue).
	 *	Need to AND the devices-to-try map with the RC's
	 *	devices-available map and AND that with the
	 *	VARY-ON map.  The result is a bitmap of all
	 *	drives which can be used for this request now.
	 */
	map = *vDMand(&vrp->vr_devtotry, &rcp->rc_avail[medindex]);
	mp2 = *vDMand(&map, &DMVary);
	variedon = vDMffs(&mp2);
	mp3 = vDMand(&mp2, &DMVolnd);

	if (vDMffs(mp3) == -1) {

		/*
		 *  None of the drives the request is willing to
		 *	accept are currently available. Leave it
		 *	alone till a drive changes state.
		 */
		if (variedon != -1)
			nLogMsg(EL_CONSOLE0, id, 
				"%.*s_%d request blocked because volnd down",
				VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq);
else
nLogMsg(EL_STATUS1, id, "%.*s_%d no acceptable drive",
				VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq);
		nLogMsg(EL_DEBUG2, id, "%.*s_%d no acceptable drive",
				VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq);
		vrp->vr_state = VS_DRIVEWAIT;
		if (rcp->rc_devs[medindex].d_nvaryondev < 
					rcp->rc_devs[medindex].d_nconfigdev) {
			TellOperToVary(vrp);
		}
		return(1);
	}
	return(0);
}
 

/*
 *  RCMount() - send mount to RC
 */
 int
RCMount(VolReq_t *vrp)
{
	char		*id = "RCMount";

	/*
	 *  send mount request off to RC
	 */
	nLogMsg(EL_DEBUG2, id, "Sending mount to RC  %.*s_%d label (%.*s)",
				VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq,
				VSNLEN, (char *)&vrp->vr_label);

	if (Send2RC(vrp->vr_rcp, vrp, M_RCMOUNT, RcMsgNULL) != E_NOERROR) {
		nLogMsg(EL_ERROR2, id, "%.*s_%d: Send2RC failed",
				VSNLEN, (char *)&vrp->vr_vsn, vrp->vr_rpseq);
		ErrOutVolReq(vrp, id);
		return(1);
	}
	return (0);
}



/*
 *  MoveWaitBatch() - see if vols can be added to a batch move
 */
MoveWaitBatch(VolReq_t *headvrp)
{
	char		*id = "MoveWaitBatch";
	RPRC_t		*dstrcp = headvrp->vr_dstrcp;
	VolReq_t	*vrp, *vrnext, *vrlast = NULL;
	int		count = 0;

	/*
	 *  if there are any other moves = headvrp ready to go, batch
	 *	them on tmp q
	 */
	for (vrp=VRNEXT(headvrp->vr_vrq); vrp; vrp=vrnext) {
		vrnext = VRNEXT(vrp->vr_vrq);	/* OkBatchMove may DQ vrp */

		if ((vrp->vr_state == VS_MOVEWAIT  ||
		     vrp->vr_state == VS_RESERVEMOVEWAIT) &&
					vrp->vr_dstrcp == dstrcp) {
			if (OkBatchMove(vrp, headvrp)) {
				/*
				 *  add to batch in scratch q
				 */
				count++;
				if (vrlast) {
					vrlast->vr_scrq.q_next = (char *) vrp;
					vrp->vr_scrq.q_prev = (char *) vrlast;
				} else {
					headvrp->vr_scrq.q_prev = NULL;
					headvrp->vr_scrq.q_next = (char *) vrp;
					vrp->vr_scrq.q_prev = (char *) headvrp;
				}
				vrp->vr_scrq.q_next = NULL;
				vrp->vr_headvrp = headvrp;
				vrlast = vrp;
			}
		}
	}
	return(count);
}

/*
 *  OkBatchMove() - see if req can be included in batch move.
 *	(Quite similar to 1st part of MoveWait.)
 */
OkBatchMove(VolReq_t *vrp, VolReq_t *headvrp)
{
	char		*id = "OkBatchMove";
	VInfo_t		*vip;
	VolReq_t	*vrp2;

        /*
         *  Get VDR from dbase (checking permissions)
         *      if NULL, then error info is set in globals
         */

        vip = DBMoveQuery((char *)&vrp->vr_vsn, vrp->vr_uid, vrp->vr_dstrcp);
        if (vip == NULL) {
        	ErrOutVolReq(vrp, id);
		return(0);
	}

	/*
	 *  see if src RC == headvrp
	 *	(TODO: in some cases (like ACS Eject) only 1 
	 *	needs to be up)
	 */

	if (vip->vi_srcrc != headvrp->vr_srcrc)
		return(0);
	vrp->vr_rcp = headvrp->vr_rcp;

	/*
	 *  Request is a candidate for batching; make sure it
	 *	doesn't duplicate a vsn already in the batch by 
	 *	checking samevsn queue back & forward. If another
	 *	one duplicates it, then this is the 'bad' one,
	 *	since the reqs are put in the temp Q in priority
	 *	order.
	 */
	for (vrp2=VRPREV(vrp->vr_vsnq); vrp2; vrp2=VRPREV(vrp2->vr_vsnq)) {
		if (vrp2->vr_headvrp == headvrp) {
			VMError = E_USER;
			nLogMsg(EL_CODE|EL_ERROR2, id, 
						"%.*s: duplicate move request",
						VSNLEN, (char *)&vrp->vr_vsn);
        		ErrOutVolReq(vrp, id);
			return(0);
		}
	}
	for (vrp2=VRNEXT(vrp->vr_vsnq); vrp2; vrp2=VRNEXT(vrp2->vr_vsnq)) {
		if (vrp2->vr_headvrp == headvrp) {
			VMError = E_USER;
			nLogMsg(EL_CODE|EL_ERROR2, id, 
						"%.*s: duplicate move request",
						VSNLEN, (char *)&vrp->vr_vsn);
        		ErrOutVolReq(vrp, id);
			return(0);
		}
	}

	/*
	 *  see if vol can be used now
	 */

	if (vrp->vr_ridp->rid_active) {
		/*
		 *  Reservation is active, so just need to see if
		 *	vol is already active within it.
		 */
		if (VolActiveInRid(vrp)) {
			vrp->vr_state = VS_RESERVEMOVEWAIT;
			return(0);
		}
	} else if (vrp->vr_ridp != headvrp->vr_ridp) {

		/*
		 *  Need to see if whole reservation can become active
		 *	(no need if = headvrp's rid, since it was checked
		 */
		if (VolIntersect(vrp->vr_ridp)) {
			vrp->vr_state = VS_RESERVEMOVEWAIT;
			return(0);
		}
	}

	vrp->vr_info = *vip;            /* copy vinfo from MoveQuery() */

	if (DBMoveInit((char *)&vrp->vr_vsn, vrp->vr_uid, vrp->vr_dstrcp)) {
		ErrOutVolReq(vrp, id);
		return(0);
	}

	return(1);
}

/*
 *  CheckRC(): 
 *	checks that rcnum is a valid RC number and if so, that the RC
 *	is up and running.  If not, sets up errors 
 */

  RPRC_t *
CheckRC(int rcnum, char *vsnp, char *id)
{
	RPRC_t		*rcp;

	if ((rcp = RCInfo(rcnum)) == NULL) {
		VMError = E_INTERNAL;
		nLogMsg(EL_CODE|EL_ERROR1, id, "RC %d not configured", rcnum);
		return(NULL);
	} 
	if (rcp->rc_srv.s_sd == -1) {
		VMError = E_UNAVAIL;
                nLogMsg(EL_STATUS1|EL_CODE|EL_CONSOLE0, id,
                  "cannot perform operation (vsn=%.*s): destination %s down",
                  VSNLEN, vsnp, rcp->rc_id);
		return(NULL);
	}
	return (rcp);
}

/*
 *  DoVaryMsg(): scan the volume request list and send messages to the
 *	(volman) operator if drives need to be varied on (for M_TIMER)
 */
 void
DoVaryMsg()
{
	char		*id = "DoVaryMsg";
	REG VolReq_t	*vrp;

	nLogMsg(EL_DEBUG3, id, "");
	for (vrp = VRHEAD(VolReqQ); vrp; vrp = VRNEXT(vrp->vr_vrq)) {
		/*
		 * check for VolReqs waiting for a drive
		 */
		if (vrp->vr_state == VS_DRIVEWAIT)
			TellOperToVary(vrp);
	}
	if (drivesneeded) {
		nLogMsg(EL_DEBUG2, id, "putting out any vary msgs");
		TellOperToVary(NULL);	/* causes msg(s) to be printed */
	}
}

/*
 *  ProcVolReqQ() - loop over the grand line of waiting requests
 */
 void
ProcVolReqQ()
{
	char		*id = "ProcVolReqQ";
	VolReq_t	*vrp, *vrp2;

	drivesneeded = 0;
	activerequests = 0;
	/*
	 *  go over the queue
	 */ 
	for (vrp=VRHEAD(VolReqQ); vrp; vrp=vrp2) {
		vrp2 = VRNEXT(vrp->vr_vrq);  /* in case vrp gets cleaned up */
		activerequests += ProcVolReq(vrp);
	}
/*
	PrintVolReqQ(EL_DEBUG3, id);
	PrintReservations();
	nLogClients();
	vPrintRCArray(EL_DEBUG3, id);
*/
	if (drivesneeded)
		TellOperToVary(NULL);
}
