/*
 *  rp_rid.c - reservation of drives & vsns using banker's algorithm.
 *
 *	Banker's algorithm prototyped by Tom Proett.
 *	Volume Manager implementation and comments by Bill Ross.
 *
 *	The request-processing front end to this library of reservation 
 *	routines is in rp_5.c
 *
 *  Reservations & deadlock
 *
 *	What is deadlock? For example, if there are 2 drives and 2 
 *	users each need both at once, it would be a mistake to give
 *	each user one, since each would then wait infinitely long
 *	for the other to give up a drive. In order to be able to
 *	guarantee to a user that all resources can be provided
 *	without deadlock, one must know in advance what is the extent 
 *	of the resources that the user will request: the user
 *	communicates this by making a reservation.
 *
 *	The user makes a reservation for a set of resources, then 
 *	when ready, makes a request to activate the reservation by
 *	requesting one or more of the resources - then that request
 *	waits in line until all the resources in the reservation can 
 *	be provided without deadlock (though further waiting for other
 *	resources in the reservation can occur). The banker's algorithm 
 *	detects when a reservation can be activated by simulating what 
 *	would happen if the reservation became active to see if an 
 *	'unsafe' situation would result. See Unsafe() below for details.
 *	This algorithm guarantees optimal use of resources and dates
 *	back to bankers lending money in ancient times (Babylonia
 *	or thereabouts).
 *	
 *  Operation
 *
 *	A reservation can be made in advance with a user-supplied
 *	id string - an "explicit" reservation (the user must supply
 *	a unique id string). Such a reservation can have drives in 
 *	various RCs and vsns added and deleted. When a reserved
 *	resource is requested, e.g. moving or mounting a reserved vsn,
 *	the banker's algorithm is used to ensure that the request
 *	does not execute until _all_ the reserved resources are
 *	available. At the point where one request in a reservation
 *	begins execution, that request becomes `active'. As long as 
 *	a resource in the reservation is being used, the request is
 *	active and all the resources are reserved. 
 *
 *	A request to use a reserved volume must reference the reservation 
 *	id to invoke the reservation and the associated guarantee that all 
 *	reserved resources will be available once execution begins. Without
 *	the reference to the reservation, the request generates an `implicit'
 *	volume reservation, which executes on its own.  Drives and volumes 
 *	that have not been explicitly reserved may be requested under a
 *	reservation: then they are implicitly added to the reservation to 
 *	satisfy the request and freed when the individual request is satisfied.
 *
 *	Moreover, a user can request a resource (e.g. vsn move or mount) 
 *	without making any reservation. In this case, an implicit
 *	_reservation_ (with no id string) is automatically made, and deleted
 *	when the move or unmount completes.
 *
 *	The banker's algorithm is applied on a per-RC/medium basis for 
 *	drives.  Drive reservations can be either "hard" or "soft." A hard
 *	reservation guarantees the holder immediate access to a drive
 *	(unless the drive pool has shrunk since the reservation was
 *	granted). A soft reservation guarantees immediate access once
 *	the reservation is active - the reservation becoming active
 *	will wait until resources are available.
 *
 *	For vsns, if two reservations have a single volume in common,
 *	only one reservation can be active.
 */


#include "rp_space.h"
#include "avltree.h"
#include <ctype.h>

/*
 *  The RP can keep a drive or more per RC reserved in order to
 *	release it when a drive fails, in order to prevent deadlock.
 *	RPRID is the reservation id used for orphaned requests and 
 *	appears when one does a 'vreserve -a'. 
RPRID will also be used for a fallback hard-reserved drive pool
that can be freed when a drive failure/varyoff would cause deadlock
 */
#define RPRID	"volman"

#ifndef TRUE
#define TRUE	1
#define FALSE	0
#endif

/*
 *  index stuff used for rids and the vsn hash queues
 */
AVL_IX_DESC	rid_ix;
AVL_IX_REC	*rid_entry;

/*
 *  scratch stuff used to test a hypothetical situation
 *	to see if it would create deadlock; used by Unsafe()
 */
#define RID_SCRATCH	5		/* initial rid size of ridRCMeds */

static int		rid_scratch = 0;
static RidRCMed_t	*ridRCMeds = NULL;
static int		*rid_active = NULL;

static int		all_drive_groups = 0;
static int		*soft_used = NULL;
static int		*hard_alloc = NULL;
static int		*safe_drives = NULL;

/*
 *  queue for all reservations
 */
QHead_t		RidQ;
int		num_rids = 0;

/*
 *  InitRidSpace() - initialize index for rids and vsn hash queue;
 *		     make hard drive reservations for replacing
 *		     downed drives; initialize scratch space
 */
 void
InitRidSpace()
{
	char	*id = "InitRidSpace";
	Rid_t	*MakeRid();

	if (BadRidName(RPRID))
		nLogMsg(EL_FATAL, id, "RPRID (%s) violates convention", RPRID);

	RidQ.q_head = NULL;
	RidQ.q_tail = NULL;

	/*
	**	create index for looking up things by name: rids,
	**		and vsns (for rid-rid vsn intersections: hash queue w/ vr_vsnq)
	**	make index record w/ 'plenty of space' for string keys (vsns & rids).
	*/
	avl_create_index(&rid_ix, AVL_NO_DUP_KEYS, 0);
	rid_entry = (AVL_IX_REC *) nAlloc(sizeof(AVL_IX_REC) + 3 * RIDLEN);

	/*
	 *  TODO: add default hard drive reservations to request queue
	 *	for a reserve pool to get out of deadlock situations
	 *	resulting from varying off when activity is high.
	 */

	/*
	 *  for convenience, count total of drive groups that need
	 *	to be checked for each reservation
	 */
	all_drive_groups = RCTotal * RC_MEDIA_TYPES;
nLogMsg(EL_STATUS1, id, "all_drive_groups = %d", all_drive_groups);
	/*
	 *  initialize scratch space for Unsafe()
	 *
	 *	rids * all_drive_groups
	 */
	rid_scratch = RID_SCRATCH;
	ridRCMeds = (RidRCMed_t *) 
		nAlloc(rid_scratch * all_drive_groups * sizeof(RidRCMed_t));
	rid_active = (int *) nAlloc(rid_scratch * sizeof(int));

	/*
	 *  overall counts over all_drive_groups
	 */
	soft_used = (int *) nAlloc(all_drive_groups * sizeof(int));
	hard_alloc = (int *) nAlloc(all_drive_groups * sizeof(int));
	safe_drives = (int *) nAlloc(all_drive_groups * sizeof(int));

	/*
	 *  make special reservation for default drive reservations
	 *	and rid-deleted user requests
	 */
	RPRidp = MakeRid(RPRID, RPUID, NULL, 0);
	if (RPRidp == NULL)
		nLogMsg(EL_FATAL, id, "can't make RP reservation");

}

/*
 *  BadRidName() - decide if a reservation id follows the rules
 */
  int
BadRidName(char *name)
{
	char	*id = "BadRidName";
	int	i;

	for (i=0; i<RIDLEN; i++) {
		if (name[i] == '\0')
			break;
		if (!isalnum(name[i]) && name[i] != ' ') {
			VMError = E_USER;
			nLogMsg(EL_CODE|EL_ERROR2, id, 
				"non-alphanumeric char in id: %d (%c)",
				name[i], name[i]);
			return(1);
		}
	}
	if (!i) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_ERROR2, id, "0-length id");
		return(1);
	}
	return(0);
}

/*
 *  findrid() - look up rid in index
 */
  static Rid_t *
findrid(char *name)
{
	char	*id = "findrid";

	sprintf(rid_entry->key, "_%.*s", RIDLEN, name);
	if (avl_find_key(rid_entry, &rid_ix) == AVL_IX_FAIL) 
		return(NULL);
	return((Rid_t *) rid_entry->recptr);
}

/*
 *  GetRid() - get rid for user & check permission
 */
  Rid_t *
GetRid(char *name, int uid)
{
	char	*id = "GetRid";
	Rid_t	*ridp;

	if (BadRidName(name))
		return(NULL);
	if ((ridp = findrid(name)) == NULL) {
		VMError = E_USER;
		nLogMsg(EL_CODE|EL_STATUS2, id, 
					"reservation id not found (%.*s)",
					RIDLEN, name);
		return(NULL);
	}
	if (ridp->rid_uid != uid) {
		VMError = E_PERM;
		nLogMsg(EL_CODE|EL_ERROR2, id, "reservation not owned by user");
		return(NULL);
	}
	return(ridp);
}

/*
 *  MakeRid() - make rid for client
 */
  Rid_t *
MakeRid(char *name, int uid, Client_t *cp, Seq_t seq)
{
	char		*id = "MakeRid";
	Rid_t		*ridp;
	RidRCMed_t	*rc;
	int		i;

	if (name != NULL)
		nLogMsg(EL_DEBUG2, id, "%.*s", RIDLEN, name);
	else
		nLogMsg(EL_DEBUG3, id, "implicit rid");

	/*
	 *  check legality
	 */
	if (name != NULL) {
		/*
		 *  client has supplied a name - see if it's unique
		 */
		if (findrid(name) != NULL) {
			VMError = E_USER;
			nLogMsg(EL_CODE|EL_ERROR2, id, 
				"Reservation name %.*s already exists",
							RIDLEN, name);
			return(NULL);
		}
	}

	/*
	 *  allocate & initialize
	 */

	ridp = (Rid_t *) nAlloc(sizeof(Rid_t));
	rc = (RidRCMed_t *) nAlloc(all_drive_groups * sizeof(RidRCMed_t));
	ridp->rid_RCMed = rc;
	memset(ridp->rid_id, '\0', sizeof(ridp->rid_id));

	nLogMsg(EL_DEBUG3, id, "ridp %d  cp %d", ridp, cp);
	if (name != NULL) {

		/*
		 *  explicit rid w/ name
		 */
		memcpy(ridp->rid_id, name, sizeof(ridp->rid_id));
	
		/*
		 *  add to index
		 */
		sprintf(rid_entry->key, "_%.*s", RIDLEN, ridp->rid_id);
		rid_entry->recptr = ridp;
		if (avl_add_key(rid_entry, &rid_ix) == AVL_IX_FAIL) 
			nLogMsg(EL_FATAL|EL_SYS, id, "avl_add_key(%s) failure", rid_entry->key);
	} 

	ridp->rid_vsns = 0;
	ridp->rid_active = 0;
	ridp->rid_vsnq.q_head = NULL;
	ridp->rid_vsnq.q_tail = NULL;
	rc = ridp->rid_RCMed;
	for (i=0; i<all_drive_groups; i++, rc++) {
		rc->rr_hardalloc = 0;
		rc->rr_hardused  = 0;
		rc->rr_softalloc = 0;
		rc->rr_softused  = 0;
		rc->rr_implicit  = 0;
	}
	ridp->rid_parallel = 0;
	ridp->rid_uid = uid;
	ridp->rid_cp = cp;
	ridp->rid_clseq = seq;
	if (cp) {
		
		/*
		 *  add to client's ridq
		 */

		ENQUEUE(ridp, &cp->c_ridq, rid_clientq, Rid_t);
	}

	num_rids++;

	/*
	 *  add to overall rid queue
	 */

	ENQUEUE(ridp, &RidQ, rid_q, Rid_t)

	return(ridp);
}

/*
 *  DeleteRid() - return any resources owned by rid.
 */
  void
DeleteRid(Rid_t *ridp, Client_t *cp)
{
	char		*id = "DeleteRid";
	VolReq_t	*vrp, *vrp2;
	int		i;

	if (ridp == RPRidp)
		nLogMsg(EL_FATAL, id, "attempt to delete RPRid");

	if (ridp->rid_id[0])
		nLogMsg(EL_DEBUG3, id, "%.*s", RIDLEN, ridp->rid_id);
	else
		nLogMsg(EL_DEBUG3, id, "implicit");

	if (ridp->rid_active != 0) 
		nLogMsg(EL_FATAL, id, "rid has %d active, %d total vsns",
			ridp->rid_active, ridp->rid_vsns);

	/* 
	 *  close down any vsn requests (returns drives in use to pools)
	 */
	for (vrp=VRHEAD(ridp->rid_vsnq); vrp; vrp=vrp2) {
		vrp2 = VRNEXT(vrp->vr_ridq);
		CancelRequest(vrp, cp, id);
	}

	num_rids--;

	/*
	 *  dequeue from / notify client if necc
	 */

	if (ridp->rid_cp) {
		DEQUEUE(ridp, &ridp->rid_cp->c_ridq, rid_clientq, Rid_t);
		if (ridp->rid_cp != cp) {
			VMError = E_CANCELED;
			nLogMsg(EL_CODE|EL_ERROR2, id, 
				"Reservation canceled by other client");
			Err2Client(ridp->rid_clseq, ridp->rid_cp);
		}
	}

	/*
	 *  dequeue from rid q
	 */
	DEQUEUE(ridp, &RidQ, rid_q, Rid_t)

	/*
	 *  free memory
	 */
	nFree(ridp->rid_RCMed);
	if (ridp->rid_id[0]) {

		/*
		 *  delete from index
		 */

		sprintf(rid_entry->key, "_%.*s", RIDLEN, ridp->rid_id);
		if (avl_delete_key(rid_entry, &rid_ix) == AVL_IX_FAIL) 
			nLogMsg(EL_FATAL, id, "avl_delete_key(%s)", rid_entry->key);
	}
	nFree(ridp);
	nLogMsg(EL_DEBUG2, id, "done");
}

/*
 *  AddVsnToRid() - make/add volreq to rid if vsn exists & would not
 *	be in 2 active rids at once.
 */

  VolReq_t *
AddVsnToRid(char *vsn, Rid_t *ridp, Client_t *cp, int uid, int implicit)
{
	char		*id = "AddVsnToRid";
	VolReq_t	*vrp, *vrp2;

	nLogMsg(EL_DEBUG3, id, "%.*s", VSNLEN, vsn);

	/*
	 *  make sure vsn is in dbase
	 */
	if (DBFetchVsnKey(vsn, 1))	/* 1 == generate error msg */
		return(NULL);

	/*
	 *  don't add if exclusive op (normally a scratch)
	 * 	is in progress
	 */
	if (vrp2 = FindVolReq(vsn)) {		
		/* 
		 *  (FindVolReq finds 1st in Q on this vsn -
		 *	if it's exclusive, then it's the only 1)
		 */
		if (vrp2->vr_voflag & VO_EXCLUSIVE) {
			VMError = E_UNAVAIL;
			nLogMsg(EL_CODE|EL_ERROR2, id, 
						"volume is in exclusive use");
			return(NULL);
		}
	}

	/*
	 *  if this rid is active, cannot add vsn if it's in another
	 *	active rid, else the 2 rids could deadlock
	 */
	if (ridp->rid_active) {
		for (; vrp2; vrp2=VRNEXT(vrp2->vr_vsnq)) {

			/*
			 *  allow > 1 instance of a vsn per rid
			 */
			if (vrp2->vr_ridp == ridp)
				continue;

			/*
			 *  is other rid active?
			 */
			if (vrp2->vr_ridp->rid_active) {
				VMError = E_UNAVAIL;
				nLogMsg(EL_CODE|EL_ERROR2, id, 
				    "vsn part of another active request #%d",
				    vrp2->vr_rpseq);
				return(NULL);
			}
		}
	}

	/*
	 *  make a partial VolReq and enqueue it
	 */
	vrp = NewVolReq();
	memcpy(&vrp->vr_vsn, vsn, VSNLEN);
	vrp->vr_flag |= implicit;
	vrp->vr_ridp = ridp;
	vrp->vr_clientp = cp;
	vrp->vr_uid = uid;
	vrp->vr_state = VS_IDLE;

	EnQVolReq(vrp);

	return(vrp);
}

/*
 *  DeleteVsnFromRid() - delete all instances of vsn from rid
 *		TODO: for now only delete client-owned vsns -
 *		might want to sweep all.. 
 */
  int
DeleteVsnFromRid(char *vsn, Rid_t *ridp, Client_t *cp)
{
	char		*id = "DeleteVsnFromRid";
	VolReq_t	*vrp, *vrpnext;
	int		err = 1;

	nLogMsg(EL_DEBUG2, id, "%.*s", VSNLEN, vsn);

	/*
	 *  look at all instances of this vsn
	 */
	for (vrp=FindVolReq(vsn); vrp; vrp=vrpnext) {
		/*
		 *   save next for loop, in case this one is deleted
		 *  	(FindVolReq gets head of Q on vsn)
		 */
		vrpnext = VRNEXT(vrp->vr_vsnq);
		if (vrp->vr_ridp == ridp  &&  vrp->vr_clientp == cp) {
			
			/*
			 *  VolReq in rid: remove from client, put in
			 *	vold rid, clean up
			 */
			CancelRequest(vrp, cp, id);
			err = 0;
		}
	}
	return(err);
}

/*
 *  MoveVolReq2RPRid() - called by CancelRequest() on active requests.
 *	Take volreq from client, put in RP rid, make implicit. 
 *	rid_active counts are handled by caller.
 */
  void
MoveVolReq2RPRid(VolReq_t *vrp, Client_t *cp)
{
	char	*id = "MoveVolReq2RPRid";
	Rid_t	*ridp = vrp->vr_ridp;

	nLogMsg(EL_DEBUG2, id, "%.*s_%d", VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);

	/*
	 *  take away from client, if any
	 */
	Orphan(vrp, cp);

	/*
	 *  remove from current rid and assign to RP's rid
	 */
	DEQUEUE(vrp, &ridp->rid_vsnq, vr_ridq, VolReq_t);
	ridp->rid_vsns--;
	vrp->vr_ridp = RPRidp;
	ENQUEUE(vrp, &RPRidp->rid_vsnq, vr_ridq, VolReq_t);
	RPRidp->rid_vsns++;

	/*
	 *  set implicit flag so that req will cancel when finished
	 */
	vrp->vr_flag |= VF_IMPLICITVSN;
}

/*
 *  VolActiveInRid() - find another, active instance of vsn in rid
 *	Since starting from a particular vrp, must search Q on vsn
 *	both forward & back from current.
 */
  VolReq_t *
VolActiveInRid(VolReq_t *vrp)
{
	char		*id = "VolActiveInRid";
	VolReq_t	*vrp2;

	/* 
	 *  check samevsn queue back
	 */
	for (vrp2=VRPREV(vrp->vr_vsnq); vrp2; vrp2=VRPREV(vrp2->vr_vsnq)) {
		nLogMsg(EL_DEBUG3, id, "back: %.*s_%d", VSNLEN, &vrp2->vr_vsn);
/*
		if (!SAMEVSN(&vrp->vr_vsn, &vrp2->vr_vsn)) {
			nLogMsg(EL_ERROR1, id, "%.6s != %.6s",
				&vrp->vr_vsn, &vrp2->vr_vsn);
			kill(getpid(), 5);
		}
*/
		if (BUSY_STATES(vrp2->vr_state)) {
			nLogMsg(EL_DEBUG3, id, "  busy");
/*
			if (vrp2->vr_ridp != vrp->vr_ridp) {
				nLogMsg(EL_ERROR1, id, 
						"other rid active/%d: %s",
						vrp2->vr_ridp->rid_active,
						vReqStateStr(vrp2->vr_state));
				kill(getpid(), 5);
			}
*/
			return(vrp2);
		}
	}
	/* 
	 *  check samevsn queue forward
	 */
	for (vrp2=VRNEXT(vrp->vr_vsnq); vrp2; vrp2=VRNEXT(vrp2->vr_vsnq)) {
		nLogMsg(EL_DEBUG3, id, "forward: %.*s_%d", VSNLEN, &vrp2->vr_vsn);
/*
		if (!SAMEVSN(&vrp->vr_vsn, &vrp2->vr_vsn)) {
			nLogMsg(EL_ERROR1, id, "%.6s != %.6s",
				&vrp->vr_vsn, &vrp2->vr_vsn);
			kill(getpid(), 5);
		}
*/
		if (BUSY_STATES(vrp2->vr_state)) {
			nLogMsg(EL_DEBUG3, id, "  busy");
/*
			if (vrp2->vr_ridp != vrp->vr_ridp) {
				nLogMsg(EL_ERROR1, id, 
						"other rid active/%d: %s",
						vrp2->vr_ridp->rid_active,
						vReqStateStr(vrp2->vr_state));
				kill(getpid(), 5);
			}
*/
			return(vrp2);
		}
	}
	nLogMsg(EL_DEBUG3, id, "  not active");
	return(NULL);
}

InAnotherActiveRid(VolReq_t *vrp, Rid_t *ridp)
{
	char		*id = "InAnotherActiveRid";
	VolReq_t	*vrp2;

	/* 
	 *  check samevsn queue back
	 */
	for (vrp2=VRPREV(vrp->vr_vsnq); vrp2; vrp2=VRPREV(vrp2->vr_vsnq)) {
		nLogMsg(EL_DEBUG3, id, "back ... %.*s_%d %s rid", 
				VSNLEN, &vrp2->vr_vsn, vrp2->vr_rpseq,
				(vrp2->vr_ridp == ridp ? "same" : "different"));
/*
		if (vrp2 == VRPREV(vrp2->vr_vsnq)) {
			nLogMsg(EL_ERROR1, id, "loop");
			kill(getpid(), 5);
		}
*/
		if (vrp2->vr_ridp == ridp)
			continue;
		if (vrp2->vr_ridp->rid_active) {
			return(1);
		}
	}
	/* 
	 *  check samevsn queue forward
	 */
	for (vrp2=VRNEXT(vrp->vr_vsnq); vrp2; vrp2=VRNEXT(vrp2->vr_vsnq)){
		nLogMsg(EL_DEBUG3, id, "fwd ... %.*s_%d %s rid", 
				VSNLEN, &vrp2->vr_vsn, vrp2->vr_rpseq,
				(vrp2->vr_ridp == ridp ? "same" : "different"));
/*
		if (vrp2 == VRNEXT(vrp2->vr_vsnq)) {
			nLogMsg(EL_ERROR1, id, "loop");
			kill(getpid(), 5);
		}
*/
		if (vrp2->vr_ridp == ridp)
			continue;
		if (vrp2->vr_ridp->rid_active)
			return(1);
	}
	return(0);
}

/*
 *  VolIntersect() - make sure that no vsn in rid is in another, active rid.
 *	This is a vital condition before activating a rid for avoiding deadlock.
 */
  int
VolIntersect(Rid_t *ridp)
{
	char		*id = "VolIntersect";
	VolReq_t	*vrp, *vrp2;

	if (ridp->rid_id[0])
		nLogMsg(EL_DEBUG3, id, "rid %.*s", RIDLEN, ridp->rid_id);
	else
		nLogMsg(EL_DEBUG3, id, "(implicit rid, addr %lld)", ridp);

	/*
	 *  look at all vsn-work in this rid
	 */
	for (vrp=VRHEAD(ridp->rid_vsnq); vrp; vrp=VRNEXT(vrp->vr_ridq)) {
		nLogMsg(EL_DEBUG3, id, "checking %.*s_%d", 
				VSNLEN, &vrp->vr_vsn, vrp->vr_rpseq);
		/*
		 *  look at all rids (other than requestor) w/ this vsn
		 *	to see if one is active: if so, intersection=true
		 */
		if (InAnotherActiveRid(vrp, ridp))
			return(1);
	}
	nLogMsg(EL_DEBUG3, id, "no conflicts");
	return(0);
}

/*
 *  Unsafe() - check the safety of the drive allocations (over all RCs/media).
 *	The caller has provisionally made an allocation, so now see if
 *	the resulting situation (i.e. activating the reservation) could 
 *	result in deadlock. This is done by simulation: iteratively 
 *	making sure that some reservation can finish based on the 
 *	currently unused resources, then 'freeing' the resources, until 
 *	either all reservations have been satisfied or some remain 
 *	unable to complete (deadlock).
 */
  int
Unsafe()
{
	char		*id = "Unsafe";
	int		progress_made, active_rids, rid_num;
	int		rcindex, medindex, count, i;
	Rid_t		*ridp;
	RidRCMed_t	*ridrcmedp;
	RPRC_t		*rcp;

	/*
	 *  gather all info into convenient arrays
	 */

	/* initialize basic totals for each RC/medium */

	rcp = RCArray;
	for (rcindex=0; rcindex<RCTotal; rcindex++, rcp++) {
	    for (medindex=0; medindex<rcp->rc_nmedia; medindex++) {
		int	index = rcindex * RC_MEDIA_TYPES + medindex;
		soft_used[index] = 0;
		hard_alloc[index] = 0;
	    }
	}

	/* try to reuse scratch for rid checks to avoid malloc */

	if (num_rids > rid_scratch) {
		rid_scratch = num_rids + 5;;
		ridRCMeds = (RidRCMed_t *) nReAlloc(ridRCMeds, 
			rid_scratch * all_drive_groups * sizeof(RidRCMed_t));
		rid_active = (int *) nReAlloc(rid_active, 
					rid_scratch * sizeof(int));
	}

	/*
	 *  copy/accumulate info for active rids for deadlock checking
	 */
	active_rids = 0;
	for (ridp = (Rid_t *)RidQ.q_head, i = 0;
			ridp;
			ridp = (Rid_t *)ridp->rid_q.q_next, i++) {
		if (!ridp->rid_active) {
			nLogMsg(EL_DEBUG3, id, "(skipping inactive %d)", i);
			continue;
		}
		nLogMsg(EL_DEBUG3, id, "rid %d active", i);
		rid_active[active_rids] = 1;

		/*
		 *  copy array of RCMedia for this rid
		 */

		ridrcmedp = ridp->rid_RCMed;
		memcpy(&ridRCMeds[active_rids * all_drive_groups], ridrcmedp,
				all_drive_groups * sizeof(RidRCMed_t));

		/*
		 *  accumulate totals per rc/medium of currently used drives
		 */
		rcp = RCArray;
		for (rcindex=0; rcindex<RCTotal; rcindex++, rcp++) {
		    int	rc_offset = rcindex * RC_MEDIA_TYPES;
		    for (medindex=0; medindex<rcp->rc_nmedia; medindex++) {
			int	index = rc_offset + medindex;

			ridrcmedp = &ridp->rid_RCMed[index];
			if (ridrcmedp->rr_softused || ridrcmedp->rr_hardalloc) {
			  if (ridp->rid_id[0])
			    nLogMsg(EL_DEBUG3, id, "%s/%s %.*s soft %d hard %d",
				rcp->rc_id, 
				vMediumName(rcp->rc_devs[medindex].d_medium),
				RIDLEN, ridp->rid_id,
				ridrcmedp->rr_softused,
				ridrcmedp->rr_hardalloc);
			  else
			    nLogMsg(EL_DEBUG3, id, "%s/%s %s soft %d hard %d",
				rcp->rc_id, 
				vMediumName(rcp->rc_devs[medindex].d_medium),
				"implicit",
				ridrcmedp->rr_softused,
				ridrcmedp->rr_hardalloc);
			}
			soft_used[index] += ridrcmedp->rr_softused;
			hard_alloc[index] += ridrcmedp->rr_hardalloc;
		    }
		}
		active_rids++;
	}
	nLogMsg(EL_DEBUG3, id, "%d rid%s %d active", 
				num_rids, (num_rids > 0 ? "s" : ""),
				active_rids);

	/*
	 *  The initial number of drives to work with per RC/medium
	 *	is the number varied on that haven't been allocated 
	 *	for mounts.
	 */
	for (rcp=RCArray, rcindex=0; rcindex<RCTotal; rcindex++, rcp++) {
	    int	rc_offset = rcindex * RC_MEDIA_TYPES;
	    for (medindex=0; medindex<rcp->rc_nmedia; medindex++) {
		int	index = rc_offset + medindex;

		safe_drives[index] = rcp->rc_devs[medindex].d_nvaryondev - 
				soft_used[index] - hard_alloc[index];
		nLogMsg(EL_DEBUG3, id, 
			"%s/%s safe %d = vary %d - softused %d - hardalloc %d", 
			rcp->rc_id, 
			vMediumName(rcp->rc_devs[medindex].d_medium),
			safe_drives[index], 
			rcp->rc_devs[medindex].d_nvaryondev, 
			soft_used[index], hard_alloc[index]);
	    }
	}

	/*
	 *  Keep going over all active reservations, checking if 
	 *	they can finish with the current total of drives 
	 *	that can be worked with (across all RC/medium groups). 
	 *	If a reservation can finish, add its drives to the 
	 *	pool that the others can work with.
	 */
	nLogMsg(EL_DEBUG3, id, "---- do while");
	do {
	    RidRCMed_t	*ridRCMedI = ridRCMeds;	/* start at 1st rid */

	    nLogMsg(EL_DEBUG3, id, "-------- check all");
	    progress_made = FALSE;

	    for (rid_num=0; rid_num<active_rids; 
			    rid_num++, ridRCMedI+=all_drive_groups) {
		int	free = 0;

		/*
		 *  skip rid if it's been done (if only for 
		 *	easier debug logging)
		 */
		if (!rid_active[rid_num])
			continue;

		/*
		 *  loop over each rc/medium in this rid
		 */

		nLogMsg(EL_DEBUG3, id, "test rid %d", rid_num);
		rcp = RCArray;
		for (rcindex=0; rcindex<RCTotal; rcindex++, rcp++) {
		    int	rc_offset = rcindex * RC_MEDIA_TYPES;

		    for (medindex=0; medindex<rcp->rc_nmedia; medindex++) {
			int	index = rc_offset + medindex;

			nLogMsg(EL_DEBUG3, id, 
			  "   %s/%s safe %d ?>= softalloc %d - softused %d", 
			  rcp->rc_id, 
			  vMediumName(rcp->rc_devs[medindex].d_medium),
			  safe_drives[index], ridRCMedI[index].rr_softalloc,
			  ridRCMedI[index].rr_softused);

			if (ridRCMedI[index].rr_softused == 0) {
				/* no drives used => nothing to free */
				continue;
			}
			if (safe_drives[index] >=
						ridRCMedI[index].rr_softalloc -
						ridRCMedI[index].rr_softused) {
				free = 1;
			} else {
				nLogMsg(EL_DEBUG3, id, "   ->free = 0!");
				free = 0;
				break;
			}
		    }
		}
		if (free) {
		    nLogMsg(EL_DEBUG3, id, "   -- free!");
		    rid_active[rid_num] = 0;

		    /*
		     *  rid can be deactivated, so return its drives
		     *		to the pool
		     */
		    progress_made++;
		    rcp = RCArray;
		    for (rcindex=0; rcindex<RCTotal; rcindex++, rcp++) {
		    	for (medindex=0; medindex<rcp->rc_nmedia; medindex++) {
			    int	index = rcindex * RC_MEDIA_TYPES + medindex;

			    safe_drives[index] += ridRCMedI[index].rr_softused;
			    ridRCMedI[index].rr_softused = 0;
			}
		    }
	        }
	    }
	} while (progress_made);
	nLogMsg(EL_DEBUG2, id, "no more progress..");

	/*
	 *  check resulting state
	 */
	count = 0;
	rcp = RCArray;
	for (rcindex=0; rcindex<RCTotal; rcindex++, rcp++) {
	    int	rc_offset = rcindex * RC_MEDIA_TYPES;
	    for (medindex=0; medindex<rcp->rc_nmedia; medindex++) {
		int	index = rc_offset + medindex;
		if (safe_drives[index] != 
		    rcp->rc_devs[medindex].d_nvaryondev - hard_alloc[index]) {
			nLogMsg(EL_DEBUG2, id, "--unsafe:");
	  		nLogMsg(EL_DEBUG2, id, 
		  	   "%s/%s: safe_drives %d  varied on %d  hard alloc %d",
		  	   rcp->rc_id, 
			   vMediumName(rcp->rc_devs[medindex].d_medium), 
			   safe_drives[index], 
			   rcp->rc_devs[medindex].d_nvaryondev, 
			   hard_alloc[index]);
	  		nLogMsg(EL_DEBUG2, 0, 
				"      soft in use %d  total in use %d",
				soft_used[index], 
				rcp->rc_devs[medindex].d_mountdev);
			count += rcp->rc_devs[medindex].d_nvaryondev - 
				 hard_alloc[index] -
				 safe_drives[index];
		}
	    }
	}
	nLogMsg(EL_DEBUG2, id, "count.. %d", count);
	return(count);
}

/*
 *  TestTotalRCDrives() - simple check to see if an additional reservation
 *	by a given rid would exceed the total drives in the RC
 */
TestTotalRCDrives(RPRC_t *rcp, int medindex, RidRCMed_t *ridrcmedp, int drives)
{
	char	*id = "TestTotalRCDrives";

	if (drives + ridrcmedp->rr_hardalloc + ridrcmedp->rr_softalloc 
					> rcp->rc_devs[medindex].d_nconfigdev) {
		nLogMsg(EL_STATUS2, id, "soft alloc %d hard alloc %d",
			ridrcmedp->rr_softalloc, ridrcmedp->rr_hardalloc);
		VMError = E_UNAVAIL;
		nLogMsg(EL_CODE|EL_ERROR2, id, 
		    "%s %d+%d=%d %s drives > configured (%d)",
		    "Total used+request",
		    ridrcmedp->rr_softalloc + ridrcmedp->rr_hardalloc, drives,
		    ridrcmedp->rr_softalloc + ridrcmedp->rr_hardalloc +drives,
		    vMediumName(rcp->rc_devs[medindex].d_medium),
		    rcp->rc_devs[medindex].d_nconfigdev);
		return(1);
	}
	return(0);
}

/*
 *  HardReserve() - make 'hard' reservation if possible. A hard
 *	reservation is dedicated & not usable by anyone else.
 */
  int
HardReserve(Rid_t *ridp, RPRC_t *rcp, int medindex, int drives)
{
	char	*id = "HardReserve";
	RidRCMed_t	*ridrcmedp;

	ridrcmedp = &ridp->rid_RCMed[rcp->rc_index * RC_MEDIA_TYPES + medindex];

	/* 
	 *  quick tests to see if enough drives available now
	 */
	if (TestTotalRCDrives(rcp, medindex, ridrcmedp, drives)) {
		nLogMsg(EL_STATUS2, id, "failed");
		return(1);
	}

	/* TODO - queue request if impossible now but ultimately possible */

	if (rcp->rc_devs[medindex].d_nvaryondev - 
	    rcp->rc_devs[medindex].d_mountdev < drives) {
		VMError = E_UNAVAIL;
		nLogMsg(EL_CODE|EL_ERROR2, id, 
			"Hard reservation %d %s drives not safe - %d %s, %d %s",
			drives, vMediumName(rcp->rc_devs[medindex].d_medium),
			rcp->rc_devs[medindex].d_nvaryondev, "varied on",
			rcp->rc_devs[medindex].d_mountdev, "mounted");
		return(1);
	}
	ridrcmedp->rr_hardalloc += drives;
	rcp->rc_devs[medindex].d_mountdev += drives;
	if (Unsafe()) {
		ridrcmedp->rr_hardalloc -= drives;
		rcp->rc_devs[medindex].d_mountdev -= drives;
		VMError = E_UNAVAIL;
		nLogMsg(EL_CODE|EL_ERROR2, id, 
			"Hard reservation of %d %s drives not safe", 
			drives, vMediumName(rcp->rc_devs[medindex].d_medium));
		return(1);
	}
	return(0);
}

/*
 *  SoftReserve() - make 'soft' reservation. Client may have to wait
 *	to use it because it shares drives w/ other reservations.
 */
  int
SoftReserve(Rid_t *ridp, RPRC_t *rcp, int medindex, int drives)
{
	char	*id = "SoftReserve";
	RidRCMed_t	*ridrcmedp;

	ridrcmedp = &ridp->rid_RCMed[rcp->rc_index * RC_MEDIA_TYPES + medindex];
	
	/* 
	 *  quick test to see if enough drives available in RC
	 */

	if (TestTotalRCDrives(rcp, medindex, ridrcmedp, drives)) {
		nLogMsg(EL_STATUS2, id, "failed");
		return(1);
	}

	ridrcmedp->rr_softalloc += drives;
	if (ridp->rid_active) {
		if (Unsafe()) {
			ridrcmedp->rr_softalloc -= drives;
			VMError = E_UNAVAIL;
			nLogMsg(EL_CODE|EL_ERROR2, id, 
						"Soft reservation not safe");
			return(1);
		}
	}
	return(0);
}

/*
 *  GetRidVsn() - find 1st instance of vsn w/ given state/client in rid; 
 *	if none, the last clientless one w/ given state
 */
  VolReq_t *
GetRidVsn(Rid_t *ridp, char *vsn, int state, Client_t *cp)
{
	char		*id = "GetRidVsn";
	VolReq_t	*vrp, *vrp2 = NULL;

	for (vrp=VRHEAD(ridp->rid_vsnq); vrp; vrp=VRNEXT(vrp->vr_ridq)) {
		if (vrp->vr_state == state  &&  SAMEVSN(vsn, &vrp->vr_vsn)){
			if (vrp->vr_clientp == cp) {
				vrp2 = vrp;
				break;
			}
			if (vrp->vr_clientp == NULL)
				vrp2 = vrp;
		}
	}
	return(vrp2);
}

/*
 *  FreeDriveUsed() - free a drive in ridrcmedp (don't dealloc unless implicit)
 */
  void
FreeDriveUsed(RidRCMed_t *ridrcmedp)
{
	if (ridrcmedp->rr_implicit > 0) {
		/*
		 *  implicit takes precedence because it's
		 *	an expedient
		 */
		ridrcmedp->rr_implicit--;
		ridrcmedp->rr_softalloc--;
		ridrcmedp->rr_softused--;
	} else if (ridrcmedp->rr_hardused) {
		/*
		 *  hard alloc is next because it's
		 *	a priority
		 */
		ridrcmedp->rr_hardused--;
	} else {
		ridrcmedp->rr_softused--;
	}
}

 void
PrintReservations()
{
	Rid_t		*ridp;

	nLogMsg(EL_ALWAYS|EL_NOTIME, 0, 
			"================ reservations ================");
	for (ridp=(Rid_t *)RidQ.q_head; ridp
					; ridp=(Rid_t *)ridp->rid_q.q_next) {
		if (!ridp->rid_id[0])
			nLogMsg(EL_ALWAYS|EL_NOTIME, 0, 
					"implicit addr %lld", ridp);
		else
			nLogMsg(EL_ALWAYS|EL_NOTIME, 0, "%.*s", 
					RIDLEN, ridp->rid_id);
		PrintReservationReqs(EL_DEBUG1, ridp, NULL);
	}
	nLogMsg(EL_ALWAYS|EL_NOTIME, 0, "----------------------------------");

}

 void
UserRidReport(vmpHdr_t *ihp, Client_t *cp)
{
	char	*id = "UserRidReport";
	Rid_t	*ridp;
	int	pktsize = 0;

	/*
	 *  figure size of pkt
	 */
	for (ridp=(Rid_t *)RidQ.q_head; ridp; ridp=(Rid_t *)ridp->rid_q.q_next){
	    if (ridp->rid_uid == cp->c_uid) {
		/* TODO.. */
	    }
	}
}

static vmpHdr_t	*reportpkt = NULL;
static int	reportsize = 0;

/*
 *  AllRidReport() - multiple M_RESERVE pkts - use XDR buffering,
 *	allowing final ack to flush
 */
 void
AllRidReport(vmpHdr_t *ihp, Client_t *cp)
{
	char			*id = "AllRidReport";
	Rid_t			*ridp;
	RPRC_t			*rcp;
	VolReq_t		*vrp;
	vmpHdr_t		*rhp;
	vmpReserveHdr_t		*rp;
	vmpReserveRCMed_t	*rrcp;
	vmpReserveVsn_t		*rvsnp;

	for (ridp=(Rid_t *)RidQ.q_head; ridp; ridp=(Rid_t *)ridp->rid_q.q_next){
		int	i, j, vsns = 0, rcmeds = 0;
		int	pktsize = PadSize(sizeof(vmpHdr_t))
				 + PadSize(sizeof(vmpReserveHdr_t));

		/*
		 *  figure size of msg
		 */
		rcp = RCArray;
		for (i=0; i<RCTotal; i++, rcp++) {
			for (j=0; j<rcp->rc_nmedia; j++) {
				int	index = RC_MEDIA_TYPES * i + j;
				if (ridp->rid_RCMed[index].rr_hardalloc)
					rcmeds++;
				if (ridp->rid_RCMed[index].rr_softalloc)
					rcmeds++;
			}
		}
		for (vrp=VRHEAD(ridp->rid_vsnq); vrp; vrp=VRNEXT(vrp->vr_ridq))
			vsns++;
		pktsize += PadSize(vsns * sizeof(vmpReserveVsn_t)) +
			   rcmeds * sizeof(vmpReserveRCMed_t);

		if (reportpkt == NULL) {
			/*
			 *  alloc space & fill in constant part of header 
			 */
			reportpkt = (vmpHdr_t *) nAlloc(pktsize);
			reportpkt->vh_magic = VMP_MAGIC;
			reportpkt->vh_type = M_RESERVE;
		} else if (pktsize > reportsize) {
			/*
			 *  nReAlloc the extra space
			 */
			reportpkt = (vmpHdr_t *) nReAlloc(reportpkt, pktsize);
			reportsize = pktsize;
		}

		/*
		 *  fill in pkt-specific hdr & update pointer in case
		 *	of nReAlloc
		 */
		reportpkt->vh_size = pktsize - PadSize(sizeof(vmpHdr_t));
		reportpkt->vh_seq = ihp->vh_seq;
		rp = MSGPTR(reportpkt, vmpReserveHdr_t);

		/*
		 *  stuff info in msg: header then substructures
		 */
		memcpy(rp->re_id, ridp->rid_id, sizeof(rp->re_id));
		rp->re_flag = ridp->rid_active;
		rp->re_uid = ridp->rid_uid;
		rp->re_rcmedcount = rcmeds;
		rp->re_vsncount = vsns;
		rrcp = (vmpReserveRCMed_t *) 
					((char *)rp + PadSize(sizeof(*rp)));
		rcp = RCArray;
		for (i=0; i<RCTotal; i++, rcp++) {
		    for (j=0; j<rcp->rc_nmedia; j++) {
			int	index = RC_MEDIA_TYPES * i + j;

			if (ridp->rid_RCMed[index].rr_hardalloc) {
				rrcp->re_rc = rcp->rc_number;
				rrcp->re_medium = rcp->rc_devs[j].d_medium;
				rrcp->re_type = RE_HARD;
				rrcp->re_drives = 
					ridp->rid_RCMed[index].rr_hardalloc;
				rrcp->re_used = 
					ridp->rid_RCMed[index].rr_hardused;
				rrcp++;
			}
			if (ridp->rid_RCMed[index].rr_softalloc) {
				rrcp->re_rc = rcp->rc_number;
				rrcp->re_medium = rcp->rc_devs[j].d_medium;
				nLogMsg(EL_DEBUG2, id, 
						"--soft rc %d medium %d",
						rrcp->re_rc, rrcp->re_medium);
				rrcp->re_type = RE_SOFT;
				rrcp->re_drives = 
					ridp->rid_RCMed[index].rr_softalloc;
				rrcp->re_used = 
					ridp->rid_RCMed[index].rr_softused;
				rrcp++;
			}
		    }
		}
		rrcp--;
		rvsnp = (vmpReserveVsn_t *) 
					((char *)rrcp + PadSize(sizeof(*rrcp)));
		for (vrp=VRHEAD(ridp->rid_vsnq); vrp; vrp=VRNEXT(vrp->vr_ridq)){
			rvsnp->re_type = 0;  /* not used for reporting */
			rvsnp->re_vsn = vrp->vr_vsn;
			rvsnp++;
		}

		/*
		 *  send msg
		 */

		Snd2Client(reportpkt, cp, 0);
	}
	/*
	 *  send final ack (flushes XDR buffer too)
	 */
	AckClient(ihp->vh_seq, cp);
}

