/*
**  IBM 3494 Repository Controller
**
**	ibm_dismount.c:  issue & handle volume dismounts
*/
#include "vm_space.h"		/* volman */
#include "ibmrc.h"		

#include <sys/stat.h>
#include <sys/errno.h>
#include <fcntl.h>


extern DevEnt_t	*DriveArray;
extern DevMap_t	HungDrives, AvailDrives;
extern int	*DriveChannels, *ChannelUse;

/*
**  FUNCTION:	ibm_dismount()
**		Construct and execute a dismount request to the robot.
**		Let caller clean up if there is a problem.
*/
int
ibm_dismount(const char *vsnp, int ibmdrv, uint *req_id)
{
	static char		*id = "ibm_dismount";
	struct mtldarg		dmtreq;

	dmtreq.versn = 0;
	dmtreq.device = ibmdrv;
	dmtreq.wait_flg = 1;	/* sync since only 1 hand in devel */
	dmtreq.target_cat = 0;
	memset(dmtreq.volser, 0, sizeof(dmtreq.volser));
	memcpy(dmtreq.volser, vsnp, VSNLEN);

	if (ioctl_ibmatl(ibm_fd, MTIOCLDM, &dmtreq) < 0) {
		VMError = E_IO;
		nLogMsg(EL_CODE|EL_ERROR1|EL_SYS|EL_CONSOLE0|EL_CONSOLE1, id,
			"ioctl_ibmatl error (%s)", IBM_CC(dmtreq.mtldret.cc));
		return(-1);
	}
	if (dmtreq.mtldret.cc != MTCC_COMPLETE) {
		VMError = E_HARDWARE;
		nLogMsg(EL_CODE|EL_ERROR1|EL_CONSOLE0|EL_CONSOLE1, id,
			"dismount problem: %s", IBM_CC(dmtreq.mtldret.cc));
		return(-1);
	}
	*req_id = dmtreq.mtldret.req_id;
	return(0);
}


#ifdef WOW
/*
**  FUNCTION:	ibm_dismount_rsp()
**		Construct and send a dismount response to the RP.
**		- not used for synchronous ops; fill in when this
**		  is implemented
*/
ibm_dismount_rsp(buf)
char	*buf;
{
	char			*id = "ibm_dismount_rsp";
	RCReq_t			*reqp;
	unsigned		seq;
	int			bitdrv;
	int			err = 0;
	char			*vsnp;

	/*
	**  Match response to request
	*/

	reqp = FindReqSeq(seq);
	if (reqp == NULL) {
		/*
		**  no active request .. not a common situation,
		**	so error handling is somewhat approximate
		*/
		nLogMsg(EL_CONSOLE1|EL_CONSOLE0|EL_STATUS1, id, 
		    "response seq %d (%s) (%s from %s)",
			seq, (DM_ISSET(bitdrv, HungDrives) ? 
				"response to hung drive clear" :
				"possibly prev acsrc"),
				vsnp, DriveArray[bitdrv].de_devnam);
		/*
		**  success
		*/
		if (DM_ISSET(bitdrv, HungDrives)) {
			/*
			**  an eventual success on a startup or pre-startup
			**	 dismount that failed initially, probably
			**	 because the drive was busy
			*/
			nLogMsg(EL_CONSOLE0|EL_STATUS1, id, 
				"DRIVE %s DISMOUNTED, MAY WANT TO VARY ON",
				DriveArray[bitdrv].de_devnam);
			DM_SET(bitdrv, AvailDrives);
			DM_CLR(bitdrv, HungDrives);
		}

		return(0);
	}

	if (reqp->rr_drive != bitdrv) {
		nLogMsg(EL_ERROR2, id, "msg drive %d != request drive %d", 
						bitdrv, reqp->rr_drive);
	}

	if (reqp->rr_state != S_UNMOUNTING) {
		VMError = E_INTERNAL;
		nLogMsg(EL_CODE|EL_ERROR1|EL_CONSOLE0|EL_CONSOLE1, id, 
			"%s unmounted by mistake (??) bad state %d",
			vsnp, reqp->rr_state);
		RCErr2RP(&reqp->rr_vsn, reqp->rr_seq, M_UNMOUNT, 0);
/*
		if ((drp->message_status.status == STATUS_SUCCESS) ||
		    (drp->message_status.status == STATUS_DRIVE_AVAILABLE)) {
			DM_SET(bitdrv, AvailDrives);
		}
*/
		DelReq(reqp);
		return(-1);
	}

/*
	switch () {
	  case STATUS_SUCCESS:
		FinishUnmount(reqp);
		return(0);
	  case STATUS_DRIVE_IN_USE:
		nLogMsg(EL_ERROR1|EL_CONSOLE1|EL_CONSOLE0, id, 
		    	"%.*s drive %s in use - will keep RETRYING",
			VSNLEN, &reqp->rr_vsn, DriveArray[bitdrv].de_devnam);
		reqp->rr_state = S_RETRY;
		reqp->rr_retries = 0;
		break;
	  case STATUS_LIBRARY_FAILURE:
		nLogMsg(EL_ERROR1|EL_CONSOLE1|EL_CONSOLE0, id, 
		    "LIBRARY FAILURE dismounting %.*s from %s - RETRYING",
			VSNLEN, &reqp->rr_vsn,
			DriveArray[bitdrv].de_devnam);
		reqp->rr_state = S_RETRY;
		break;
	  case STATUS_PROCESS_FAILURE:
		nLogMsg(EL_ERROR1|EL_CONSOLE1|EL_CONSOLE0, id, 
		    "PROCESS FAILURE dismounting %.*s from %s - (%s) RETRYING",
			VSNLEN, &reqp->rr_vsn,
			DriveArray[bitdrv].de_devnam,
			"LSM may be full");
		reqp->rr_state = S_RETRY;
		break;
	  case STATUS_ACS_FULL:
		nLogMsg(EL_CONSOLE0|EL_STATUS1, id, 
		    "ACS is FULL - no free cells - will keep RETRYING");
		reqp->rr_state = S_RETRY;
		reqp->rr_retries = 0;
		break;
	  case STATUS_AUDIT_IN_PROGRESS:
		VMError = E_INTERNAL;
		nLogMsg(EL_CONSOLE0|EL_CONSOLE1|EL_CODE|EL_ERROR1, id, 
		    "audit in progress - will keep RETRYING");
		reqp->rr_state = S_RETRY;
		reqp->rr_retries = 0;
		break;
	  case STATUS_MISPLACED_TAPE:
		VMError = E_INTERNAL;
		nLogMsg(EL_CONSOLE0|EL_CODE|EL_ERROR1, id,
			"misplaced tape, found vsn %s in %s instead of %.*s",
			vsnp, DriveArray[bitdrv].de_devnam, 
			VSNLEN, &reqp->rr_vsn);
		err = 1;
		break;
	  case STATUS_NOT_IN_SAME_ACS:
		VMError = E_INTERNAL;
		nLogMsg(EL_CODE|EL_ERROR1, id, "vol not in same ACS as drive");
		err = 1;
		break;
	  case STATUS_UNREADABLE_LABEL:
		VMError = E_INTERNAL;
		nLogMsg(EL_CONSOLE0|EL_CODE|EL_ERROR1, id, 
			"volume in %s has unreadable label", 
			DriveArray[bitdrv].de_devnam);
		err = 1;
		break;
	  case STATUS_DRIVE_AVAILABLE:
		VMError = E_INTERNAL;
		nLogMsg(EL_CODE|EL_ERROR1, id, "drive %s empty", 
					DriveArray[bitdrv].de_devnam);
		DM_SET(bitdrv, AvailDrives);
		err = 1;
		break;
	  case STATUS_VOLUME_NOT_IN_DRIVE:
		VMError = E_INTERNAL;
		nLogMsg(EL_CODE|EL_ERROR1|EL_CONSOLE0, id, 
		   "%s: vol %s in drive instead of %.*s (or not in drive if =)", 
		   DriveArray[bitdrv].de_devnam, vsnp, VSNLEN, &reqp->rr_vsn);
		err = 1;
		break;
	  default:
		VMError = E_INTERNAL;
		nLogMsg(EL_ERROR1|EL_CONSOLE0|EL_CONSOLE1, id, 
		    "UNEXPECTED ERR dismounting %.*s from %s: %s (RETRYING)",
			VSNLEN, &reqp->rr_vsn,
			DriveArray[bitdrv].de_devnam,
			cl_status(drp->message_status));
		reqp->rr_state = S_RETRY;
		break;
	}
*/

	if (err) {
		/*
		**  Error - notify vold & clean up
		*/
		VMError = E_HARDWARE;
		if (reqp->rr_type == T_UNMOUNTFIRST) {
			nLogMsg(EL_ERROR1|EL_CODE|EL_CONSOLE0, id,
			  "%.*s %s stuck in drive %d (%s), %s",
			  VSNLEN, &reqp->rr_vsn, "was already", drive,
			  DriveArray[bitdrv].de_devnam,
			  "& recovery failed");
		} else {
			nLogMsg(EL_ERROR1|EL_CODE|EL_CONSOLE0, id,
				"unmount failed for %.*s, drive %d (%s)",
				VSNLEN, &reqp->rr_vsn, drive,
				DriveArray[bitdrv].de_devnam);
		}
		DM_SET(reqp->rr_drive, AvailDrives);
		DM_SET(reqp->rr_drive, HungDrives);
		if (reqp->rr_errstr[0]) {
			/*
			**  unmount was recovering from failed mount
			*/
			SendRecoverErr(reqp, "--unmount failed too");
		} else
			RCErr2RP(&reqp->rr_vsn, reqp->rr_seq, M_UNMOUNT, 0);
		DelReq(reqp);
		return(-1);
	}
	return(0);
}
#endif

