/*
 *  vltlib.c - database routines for the Manual Vault Server
 *
 *	vltlib.c uses Hunter & Associates' B-PLUS btree indexing shareware
 *		library. See the bplus.doc file for further information.
 *
 *	Tom Proett and Bill Ross, Computer Sciences Corporation
 *	NASA Ames Research Center, March, 1993
 */
/*
 *	Btree errors should cause an immediate shutdown. Here is how they
 *	are handled/reported. If the error is reported, the caller should
 *	perform the shutdown, including calling CloseDb().
 *
 *	shuts down within routine (EL_FATAL) if error:
 *	
 *	        InitDb()
 */

/****************************************************************************/

#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <signal.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "vmerror.h"
#include "vconfig.h"
#include "sysfile.h"
#include "bplus.h"

#include "vltlib.h"


/*****************************************************************
 			vars used to return values
 *****************************************************************/

vsnrec_t	vsnbuf;		/* ALL info on vsn */

/*****************************************************************
		vars used for accessing dbase
 *****************************************************************/

static		STAT_T		sbuf;
static		int		dbopen = 0;
extern		int		VMError;

OFF_T	vsnaddr;

/***************************************
	table file descriptors 
 ***************************************/

int		db_vsn_fd = -1;

/***************************************
	bplus index structures	
 ***************************************/

IX_DESC		db_vsn_idx, db_loc_idx;		/* unique key index */
ENTRY		db_vsn_entry, db_loc_entry;

/***********************************************************************
   			basic dbase table routines
 ***********************************************************************/

/***********************************************************
						TOINT()
 ***********************************************************/

/*
 *  toInt() - check & translate char -> int
 */

 static int
toInt(num, str)
 int	*num;
 char	*str;
{
	char	*id = "db to integer";
	int     i;

	if (str[INTLEN] != ' ') {
		vLogMsg(EL_ERROR1, id,
			"int field discrepancy (%.*s)",
			INTLEN+1, str);
		return(1);
	}
	for (i=0; i<INTLEN; i++)
		if (str[i] != ' '  &&  !isdigit(str[i])  &&  str[i] != '-') {
			vLogMsg(EL_ERROR1, id, 
				"int field (%.*s) has non-numeral",
				INTLEN, str);
			return(1);
		}
	*num = atoi(str);
	return(0);
}


/***********************************************************
						SETINT()
 ***********************************************************/

/*
 *  SetInt() - translate int into char field
 *	return: 0 if ok, otherwise 1 & set VMerror & EL_CODE
 */

 static	int
SetInt(str, num)
 char	*str;
 int	num;
{
	char	*id = "SetInt";
	char	arr[32];
	int	len;

	sprintf(arr, "%d", num);
	len = strlen(arr);

	if (len > INTLEN) {
		VMError = E_DBASE;
		vLogMsg(EL_CODE, id, "%d: int len %d > reserved %d", 
						num, len, INTLEN);
		return(1);
	}
	sprintf(str, "%*d", INTLEN, num);
	str[INTLEN] = ' ';
	return(0);
}


/***********************************************************
						FIELDFILL()
 ***********************************************************/
/*
 *  FieldFill() - fill an area in the database record with
 *	blanks, then copy from the source string up until a
 *	zero is seen or size chars have been copied.
 */
 void
FieldFill(dest, src, num)
 char	*dest, *src;
 int	num;
{
	int	i;

	memset(dest, ' ', num+1);
	for (i=0; i<num; i++) {
		if (src[i] == '\0')
			break;
		dest[i] = src[i];
	}
}

/***********************************************************
						TBL_OP()
 ***********************************************************/

/*
 *  tbl_op() - seek and read or write an indexed table file
 *		if error, logs error & returns 1 
 */

#define OP_READ		111
#define OP_WRITE	222
#define OP_APPEND	333
#define TBL_EOF		444
#define TBL_ERR		999

 static int
tbl_op(op, buf)
 int		op;
 vsnrec_t	*buf;
{
	char	*id = "dbop";
	int	err;
	int	ret;


	/*
	 *  it is assumed that the file offset has been set by a 
	 *	table-specific means (e.g index operation) and that
	 *	(for write) the buffer has been set up.
	 */

	if (op == OP_APPEND) {
		err = ((vsnaddr = LSEEK(db_vsn_fd, 0, SEEK_END)) == -1) ?
			 1 : 0;
	}
	else {
		err = (LSEEK(db_vsn_fd, vsnaddr, SEEK_SET) != vsnaddr) ?
			1 : 0;
	}
	if (err) {
		VMError = E_DBASE;
		vLogMsg(EL_LCLCON|EL_RPCON|EL_SYS|EL_CODE, id,
			"table LSEEK error");
		return(TBL_ERR);
	}

	if (vsnaddr % sizeof(vsnrec_t)) {	/* consistency check */
		VMError = E_DBASE;
		vLogMsg(EL_CODE|EL_RPCON, id,
			"dbase vsn table bad size (%d)", vsnaddr);
		return(TBL_ERR);
	}
	switch (op) {
	case OP_READ:
		ret = read(db_vsn_fd, buf, sizeof(vsnrec_t));
		break;
	case OP_APPEND:
	case OP_WRITE:
		ret = write(db_vsn_fd, buf, sizeof(vsnrec_t));
		break;
	default:
		vLogMsg(EL_FATAL, id, "programming error op %d", op);
	}
	if ((ret == 0) && (op == OP_READ)) {

		/* 
		 *  this should only happen when reading a whole table 
		 *	into memory
		 */

		vLogMsg(EL_STATUS2, id, "end of file");
		return(TBL_EOF);
	}
	if (ret != sizeof(vsnrec_t)) {
		if (ret == -1) {
			vLogMsg(EL_SYS|EL_ERROR1, id, "db table %s errno %d", 
				(op == OP_READ) ? "read" : "write", errno);
		}
		else {
			vLogMsg(EL_ERROR1, id, 
				"db table %s: %d bytes, attempted %d", 
				(op == OP_READ) ? "read" : "write",
				ret, sizeof(vsnrec_t));
		}
		VMError = E_DBASE;
		return(TBL_ERR);
	}
	vLogMsg(EL_DEBUG4, id, "did %d on %.6s", op, buf->v_vsn);
	return(0);
}


/***********************************************************
					  VSNRECCHECK()
 ***********************************************************/
/*
 *  VsnRecCheck() - using global vsnaddr, read record &
 *	fill in vsnloc structure with current location.
 *	check to see that the record is not corrupt.
 *	return 0 if no error, otherwise set VMError & EL_CODE 
 *	and return 1.
 */
 static	int
VsnRecCheck(rec)
 vsnrec_t	*rec;
{
	char	*id = "VsnRecCheck";
	int	num;

	/*
	 *  read record from table
	 */

	vsnaddr = db_vsn_entry.recptr;
	if (tbl_op(OP_READ, rec)) {
		VMError = E_DBASE;
		vLogMsg(EL_RPCON|EL_CODE, id, 
				"dbase table error offset %d", vsnaddr);
		return(1);
	}

	/*
	 *  basic record sanity check
	 */
	if (rec->end != '\n') {
		VMError = E_DBASE;
		vLogMsg(EL_RPCON|EL_CODE, id, 
			"dbase index error on (%.6s): bad record terminator", 
			rec->v_vsn);
		return(1);
	}

	/*
	 *  see if record is really there
	 */
	if (rec->v_status == S_DELETED) {
		VMError = E_DBASE;
		vLogMsg(EL_RPCON|EL_CODE, id, 
			"dbase index error on (%.6s): deleted record", 
			rec->v_vsn);
		return(1);
	}

	/*
	 *  check to make sure last known location is ok
	 */

	num = rec->v_lastloc - '0';
	if (num < 0 || HISTORY <= num) {
		VMError = E_DBASE;
		vLogMsg(EL_RPCON|EL_CODE, id, 
			"%.6s: bad location '%c' in history array",
			rec->v_vsn, rec->v_lastloc);
		return(1);
	}

	return(0);
}

/***********************************************************
						ADDLOCTOINDEX()
 ***********************************************************/

/*
 *  AddLocToIndex() - if location is in vault, add it to index
 *  if it works (or isn't needed) return 0 otherwise return 1.
 */

 int
AddLocToIndex(lrp)
 locrec_t	*lrp;
{
	char		*id = "AddLocToIndex";
	OFF_T		tmprecptr = -1;
	OFF_T		holdaddr = vsnaddr;
	vsnrec_t	tmprec;
	locrec_t	*tlrp;
	int		lastloc, nextloc;

	if (lrp->l_type != VAULT)
		return(0);

	vLogMsg(EL_DEBUG4, id, "add %.*s to index",
		sizeof(vaultrec_t),
		&lrp->l_field.vu_vault);
	clearkey(&db_loc_entry);
	db_loc_entry.recptr = vsnaddr;
	sprintf(db_loc_entry.key, "%.*s",
		sizeof(vaultrec_t),
		&lrp->l_field.vu_vault);


	if (add_key(&db_loc_entry, &db_loc_idx) == IX_FAIL) {
		vLogMsg(EL_ERROR1, id, 
		"%.*s record %d: probable duplicate",
			sizeof(vaultrec_t),
			&lrp->l_field.vu_vault,
			vsnaddr/sizeof(vsnbuf));
		/*
		 *  Look for the location to see if there is
		 *  a vsn already there.  If there is, update it
		 *  to be location "unknown".
		 */
		if (find_key(&db_loc_entry, &db_loc_idx) == IX_FAIL ||
		    db_loc_entry.recptr == vsnaddr)
			return(1);

		tmprecptr = db_loc_entry.recptr;
		if (delete_key(&db_loc_entry, &db_loc_idx) == IX_FAIL) {
			vLogMsg(EL_ERROR1, id,
				"delete duplicate failed");
			return(1);
		}
		db_loc_entry.recptr = vsnaddr;
		if (add_key(&db_loc_entry, &db_loc_idx) == IX_FAIL) {
			vLogMsg(EL_ERROR1, id, 
				"%.*s record %d: second add",
				sizeof(vaultrec_t),
				&lrp->l_field.vu_vault,
				vsnaddr/sizeof(vsnbuf));
			return(1);
		}
		/*
		 *  Here we check to see if a record needs to be updated
		 *  with the position "unknown".  We don't want to return
		 *  an error if one occurs... the "real" update has been
		 *  done.
		 */

		/* read record to be updated */
		db_vsn_entry.recptr = tmprecptr;
		if (VsnRecCheck(&tmprec) == 0) {
			lastloc = tmprec.v_lastloc - '0';
			nextloc = (lastloc + 1) % HISTORY;
			tmprec.v_lastloc = nextloc + '0';
			tlrp = &tmprec.v_loc[nextloc];
			
			FieldFill(tlrp->l_date, lrp->l_date, DATELEN);
			tlrp->a01 = ' ';
			tlrp->l_type = UNKNOWN;
			memset(tlrp->l_field.vu_general, ' ', LOCSIZE);
			tbl_op(OP_WRITE, &tmprec);
			
			vLogMsg(EL_STATUS1, id,
				"reset %.6s location to unknown",
				tmprec.v_vsn);
		}
		db_vsn_entry.recptr = vsnaddr = holdaddr;
	}
	return(0);
}


/***********************************************************
						DELETELOCFROMINDEX()
 ***********************************************************/

/*
 *  DeleteLocFromIndex() - if location is in vault, delete it from index
 *  if it works (or isn't needed) return 0 otherwise return 1.
 */
 int
DeleteLocFromIndex(lrp)
 locrec_t	*lrp;
{
	char	*id = "DeleteLocFromIndex";

	/*
	 *  see if it is a vault location
	 */

	if (lrp->l_type != VAULT)
		return(0);

	vLogMsg(EL_DEBUG4, id, "delete %.*s from index",
		sizeof(vaultrec_t),
		lrp->l_field.vu_vault);

	clearkey(&db_loc_entry);
	db_loc_entry.recptr = vsnaddr;
	sprintf(db_loc_entry.key, "%.*s",
		sizeof(vaultrec_t),
		lrp->l_field.vu_vault);

	if (delete_key(&db_loc_entry, &db_loc_idx) == IX_FAIL) {
		VMError = E_DBASE;
		vLogMsg(EL_RPCON, id, 
			"dbase loc delete_key() error on (%.*s)",
			sizeof(vaultrec_t),
			lrp->l_field.vu_vault);
		return(1);
	}
	return(0);
}

/***********************************************************
						STARTINDEX()
 ***********************************************************/

/*
 *  StartIndex() - open it if it exists and return 0,
 *  otherwise make it and return 1.
 */

 static int
StartIndex(name, ix, dup)
 char		*name;
 IX_DESC	*ix;
{
	char	*id = "StartIndex";

	if (stat(name, &sbuf) != -1) {
		if (open_index(name, ix, dup) != IX_OK)
			vLogMsg(EL_FATAL, id, "open_index(%s) failed", name);
		return(0);
	} 
	if (errno != ENOENT)
		vLogMsg(EL_FATAL|EL_SYS, id, "index (%s) stat failed:", name);

	vLogMsg(EL_RPCON, id,
		"creating %s index (%s)", 
		dup ? "" : "exclusive", name);
	if (make_index(name, ix, dup) != IX_OK)
		vLogMsg(EL_FATAL, id, "make_index(%s) failed", name);
	return(1);
}


/***********************************************************
						CHECKVSN()
 ***********************************************************/

/*
 *  CheckVsn() - make sure vsnbuf.v_vsn ok
 */

 int
CheckVsn(vsn)
 char	*vsn;
{
	char	*id = "CheckVsn";
	int	i;

	if (vsn == vsnbuf.v_vsn && vsn[VSNLEN] != ' ') {
		vLogMsg(EL_CODE, id, "%.7s: no space seperator after vsn", 
			vsn);
		return(1);
	}
	for (i=0; i<VSNLEN; i++) {
		if (vsn[i] == ' ') {
			vLogMsg(EL_CODE, id, "vsn %.6s: too short", 
				vsn);
			return(1);
		}
		if (!isalnum(vsn[i])) {
			vLogMsg(EL_CODE, id,
				"vsn %.6s: Illegal character (%c)",
				vsn, vsn[i]);
			return(1);
		}
	}
	return(0);
}

/***********************************************************
 						PRINTVSN()
 ***********************************************************/

/*
 *  printvsn() - print contents of vsnbuf
 */

 static void
printvsn()
{
	char	*buf = (char *) &vsnbuf;

	vLogMsg(EL_STATUS2, 0, 
			"addr %d\n\t|%.58s|\n\t%.72s", 
			vsnaddr, buf, buf+58);
}

/***********************************************************
 					REBUILDVSNINDEX()
 ***********************************************************/

/*
 *  RebuildVsnIndex() - read vsn table & build vsn index 
 */

 static void
RebuildVsnIndex()
{
	char		*id = "RebuildVsnIndex";
	int		i, ret;

	vLogMsg(EL_DEBUG1, id, "entered %d", sizeof(vsnbuf));

	/*
	 *  read the whole table
	 */

	vsnaddr = 0;
	for (i=1;;i++) {

		ret = tbl_op(OP_READ, &vsnbuf);
		if (ret) {
			if (ret == TBL_EOF)
				break;
			else
				exit(2);
		}

		/*
		 *  see if record is really there
		 */

		if (vsnbuf.v_status == S_DELETED) {
			vsnaddr += sizeof(vsnbuf);
			continue;
		}

		/*
		 *  check vsn part of record & end of record
		 */

		if (CheckVsn(vsnbuf.v_vsn)) {
			printvsn();
			vLogMsg(EL_RPCON|EL_ERROR1, id, 
					"%s: bad record %d", DB_VSN_TBL, i);
			exit(2);
		}
		if (vsnbuf.end != '\n') {
			printvsn();
			vLogMsg(EL_RPCON|EL_ERROR1, id, 
					"%s: bad record %d terminator", 
					DB_VSN_TBL, i);
			exit(2);
		}

		/*
		 *  add to vsn index
		 */

		vLogMsg(EL_DEBUG4, id, "add %.6s to index", vsnbuf.v_vsn);
		clearkey(&db_vsn_entry);
		db_vsn_entry.recptr = vsnaddr;
		sprintf(db_vsn_entry.key, "%.6s", vsnbuf.v_vsn);
		if (add_key(&db_vsn_entry, &db_vsn_idx) == IX_FAIL) {
			vLogMsg(EL_RPCON|EL_ERROR1, id, 
		    "%.6s record %d: add_key() failure - probable duplicate",
				vsnbuf.v_vsn, vsnaddr/sizeof(vsnbuf));
		}
		vsnaddr += sizeof(vsnbuf);
		if ((i%25000) == 0)
			vLogMsg(EL_DEBUG1, id, "record %d..", i);
	}

	vLogMsg(EL_STATUS1, id, "vsn index build complete");
}

/***********************************************************
 					REBUILDLOCINDEX()
 ***********************************************************/

/*
 *  RebuildLocIndex() - read vsn table & build loc index 
 */

 static void
RebuildLocIndex()
{
	char		*id = "RebuildLocIndex";
	int		i, num, ret;

	vLogMsg(EL_DEBUG1, id, "entered");

	/*
	 *  read the whole table
	 */
	vsnaddr = 0;
	for (i=1;;i++) {

		ret = tbl_op(OP_READ, &vsnbuf);
		if (ret) {
			if (ret == TBL_EOF)
				break;
			else
				exit(2);
		}

		/*
		 *  end of record
		 */

		if (vsnbuf.end != '\n') {
			printvsn();
			vLogMsg(EL_RPCON|EL_ERROR1, id, 
					"%s: bad record %d terminator", 
					DB_VSN_TBL, i);
			exit(2);
		}

		/*
		 *  see if record is really there
		 */

		if (vsnbuf.v_status == S_DELETED) {
			vsnaddr += sizeof(vsnbuf);
			continue;
		}

		/*
		 *  get the vault location
		 */

		num = vsnbuf.v_lastloc - '0';
		if (num < 0 || HISTORY <= num) {
			vLogMsg(EL_RPCON|EL_ERROR1, id, 
				"%.6s: bad location '%c' in history array",
				vsnbuf.v_vsn, vsnbuf.v_lastloc);
			exit(2);
		}

		/*
		 *  add to location index
		 */
		AddLocToIndex(&vsnbuf.v_loc[num]);

		vsnaddr += sizeof(vsnbuf);
		if ((i%10000) == 0)
			vLogMsg(EL_DEBUG1, id, "record %d..", i);
	}

	vLogMsg(EL_STATUS1, id, "location index build complete");
}

/***********************************************************
  					    	UNLINK()
 ***********************************************************/

/*
 *  Unlink() - ignore ENOENT
 */

 static void
Unlink(name)
 char	*name;
{
	char	*id = "Unlink";

	if (unlink(name) == -1) {
		if (errno != ENOENT) {
			vLogMsg(EL_SYS|EL_FATAL, id,
				"%s: unlink failed", name);
		}
	}
	else
		vLogMsg(EL_DEBUG1, id, "unlinked %s", name);
}

/***********************************************************
  						INITDB()
 ***********************************************************/

/*
 *  InitDb() - open database & initialize indexes
 */

 void
InitDb()
{	
	char		*id = "initdb";
	int		ret;
	OFF_T		addr;
	int		cflag = O_RDWR|O_SYNC|O_CREAT;

	/*
	 *  set up buffer
	 */

	memset(&vsnbuf, ' ', sizeof(vsnbuf));
	vsnbuf.end = '\n';

	/*
	 *  db about to be set up, so lock
	 */

	if ((ret=open(DB_LOCK, O_CREAT|O_EXCL, 0600)) == -1) {
		if (errno == EEXIST) {
			/*
			 *  dbase wasn't closed properly
			 *  prepare to rebuild indexes
			 */

			vLogMsg(EL_RPCON, id,
				"DBASE NOT CLOSED - REBUILDING INDEX");
			Unlink(DB_VSN_IDX);
			Unlink(DB_LOC_IDX);
		}
		else
			vLogMsg(EL_FATAL|EL_SYS, id, "create lock failed");
	}
	else
		close(ret);

	vLogMsg(EL_DEBUG1, id, "open tables & indexes:");

	/* vsn table and associated indexes */

	if (stat(DB_VSN_TBL, &sbuf) != -1) {

		/*
		 *  table exists - open & check size
		 */

		if ((db_vsn_fd = open(DB_VSN_TBL, O_RDWR|O_SYNC, 0644))== -1)
			vLogMsg(EL_FATAL|EL_SYS, id, 
				"vsn table (%s) open failed: %s", DB_VSN_TBL);
		if ((addr = LSEEK(db_vsn_fd, 0, SEEK_END)) == -1) 
			vLogMsg(EL_FATAL|EL_SYS, id, 
					"table LSEEK (%s) failed: %s", 
					DB_VSN_TBL);
		if (addr % sizeof(vsnbuf)) {	/* consistency check */
			vLogMsg(EL_FATAL, id, 
				"table (%s) size (%d) bad - records %d: %d",
				DB_VSN_TBL, addr, sizeof(vsnbuf),
				addr % sizeof(vsnbuf));
		}
	}
	else {
		/*
		 *  table doesn't exist
		 */

		if (errno != ENOENT) {
			vLogMsg(EL_FATAL|EL_SYS, id,
				"table stat problem: %s:",
				DB_VSN_TBL);
		}
		vLogMsg(EL_RPCON, id,
			"CREATING VSN TABLE %s AND INDEXES", 
			DB_VSN_TBL);
		if ((db_vsn_fd = open(DB_VSN_TBL, cflag, 0644))== -1) {
			vLogMsg(EL_FATAL|EL_SYS, id, 
				"vsn table (%s) open/create failed:", 
				DB_VSN_TBL);
		}
		Unlink(DB_VSN_IDX);
		Unlink(DB_LOC_IDX);
	}

	if (StartIndex(DB_VSN_IDX, &db_vsn_idx, 0))
		RebuildVsnIndex();
	if (StartIndex(DB_LOC_IDX, &db_loc_idx, 0))
		RebuildLocIndex();

	vLogMsg(EL_STATUS2, id, "table opened / size checked");

	dbopen = 1;
	vLogMsg(EL_DEBUG2, id, "DBASE: database started");
}


/***********************************************************
   						CLOSEDB()
 ***********************************************************/

/*
 *  CloseDb() - close indexes & tables
 */

 void
CloseDb()
{
	char	*id = "closedb";

	if (!dbopen)
		return;

	if (close_index(&db_vsn_idx) != IX_OK) {
		vLogMsg(EL_RPCON, id, "%s: btree close_index() error",
							DB_VSN_IDX);
		Unlink(DB_VSN_IDX);
	}

	if (close_index(&db_loc_idx) != IX_OK) {
		vLogMsg(EL_RPCON, id, "%s: btree close_index() error",
							DB_LOC_IDX);
		Unlink(DB_LOC_IDX);
	}

	/*
 	 *  close table
 	 */

	if (db_vsn_fd > -1) {
		fsync(db_vsn_fd);
		if (close(db_vsn_fd) == -1)
			vLogMsg(EL_RPCON, id,
				"%s: close() error", DB_VSN_TBL);
	}

	/*
	 *  release lock
	 */

	if (unlink(DB_LOCK) == -1)
		vLogMsg(EL_RPCON|EL_SYS, id,
			"unlink lock file (%s)", DB_LOCK);

	vLogMsg(EL_STATUS1, id, "Database shut down");
}

 void
TruncStr(str, len)
 char	*str;
 int	len;
{
	int	i;

	for (i=len-1; i>=0; i--)
		if (str[i] != ' ')
			break;
	i++;
	if ((0 <= i) && (i < len) &&		/* in range */
	    str[i] == ' ')			/* and is last blank */
		str[i] = '\0';
}

/***********************************************************
					  LOCRECTOSPACETIME()
 ***********************************************************/
/*
 *  LocRecToSpaceTime() - convert an ascii locrec structure
 *	into a spacetime struct.
 *	return 0 if no error, otherwise set VMError & EL_CODE 
 *	and return 1.
 */
 static	int
LocRecToSpaceTime(lrp, stp)
 locrec_t	*lrp;
 spacetime_t	*stp;
{
	char	*id = "LocRecToSpaceTime";

	vLogMsg(EL_DEBUG4, id, "entered");

	stp->when = TimeStr(lrp->l_date);
	
	/*
	 *  do the right thing for each type of location
	 */
	switch (lrp->l_type) {
	case UNKNOWN:
		vLogMsg(EL_DEBUG4, id, "unknown");
		stp->where.tag = unknown;
		memset(stp->where.general, ' ', LOCSIZE);
		break;
	case VAULT: {
		vaultrec_t *vrp = &lrp->l_field.vu_vault;

		vLogMsg(EL_DEBUG4, id, "vault");
		
		stp->where.tag = vault;
		memcpy(stp->where.section, vrp->vr_section, SECSIZE);
		if (vrp->a00 != ' ') {
			VMError = E_DBASE;
			vLogMsg(EL_RPCON|EL_CODE, id,
				"%.6s: no seporator after section",
				vsnbuf.v_vsn);
			printvsn();
			return(1);
		}
		TruncStr(stp->where.section, SECSIZE);
		if (toInt(&stp->where.row, vrp->vr_row) ||
		    toInt(&stp->where.slot, vrp->vr_slot)) {
			VMError = E_DBASE;
			vLogMsg(EL_RPCON|EL_CODE, id,
				"%.6s: dbase translation error",
				vsnbuf.v_vsn);
			printvsn();
			return(1);
		}
		break;
	}
	case DISPLAY:
		vLogMsg(EL_DEBUG4, id, "display");
		stp->where.tag = display;
		memcpy(stp->where.general,
		       lrp->l_field.vu_general, LOCSIZE);
		TruncStr(stp->where.general, LOCSIZE);
		break;
	default:
		VMError = E_DBASE;
		vLogMsg(EL_RPCON|EL_CODE, id,
			"%.6s: bad location type tag %c",
			vsnbuf.v_vsn, lrp->l_type);
		printvsn();
		return(1);
	}

	return(0);
}

/***********************************************************
					  VSNADDRTOLOCREC()
 ***********************************************************/
/*
 *  VsnAddrToLocRec() - using global vsnaddr, read record &
 *	fill in vsnloc structure with current location.
 *	return pointer if no error, otherwise set VMError & EL_CODE 
 *	and return NULL.
 */

 static	vsnloc_t *
VsnAddrToLocRec(vsn)
 vsn_t	*vsn;
{
	static		vsnloc_t	vloc;
	int		lastloc;

	if (VsnRecCheck(&vsnbuf))
		return( NULL );

	if (memcmp(vsnbuf.v_vsn, vsn, VSNLEN) != 0) {
		VMError = E_DBASE;
		vLogMsg(EL_RPCON|EL_CODE, NULL, 
			"%.6s: index found wrong record: %.6s",
			vsn, vsnbuf.v_vsn);
		return(NULL);
	}

	memcpy((char *)&vloc.vsn, vsnbuf.v_vsn, VSNLEN);

	lastloc = vsnbuf.v_lastloc - '0';
	if (LocRecToSpaceTime(&vsnbuf.v_loc[lastloc], &vloc.location))
		return( NULL );

	return( &vloc );
}

/***********************************************************
					  VSNADDRTOHISTORY()
 ***********************************************************/
/*
 *  VsnAddrToHistory() - using global vsnaddr, read record &
 *	fill in full vsnhistory structure.
 *	return pointer if no error, otherwise set VMError & EL_CODE 
 *	and return NULL.
 */

 static	vsnhistory_t *
VsnAddrToHistory(vsn)
 vsn_t	*vsn;
{
	static		vsnhistory_t	vinfo;

	int		i, lastloc;

	if (VsnRecCheck(&vsnbuf))
		return( NULL );

	if (memcmp(vsnbuf.v_vsn, vsn, VSNLEN) != 0) {
		VMError = E_DBASE;
		vLogMsg(EL_RPCON|EL_CODE, NULL, 
			"%.6s: index found wrong record: %.6s",
			vsn, vsnbuf.v_vsn);
		return(NULL);
	}

	memcpy(&vinfo.vsn, vsnbuf.v_vsn, VSNLEN);
	lastloc = vsnbuf.v_lastloc - '0';

	for (i=0; i<HISTORY; i++) {
		if ( LocRecToSpaceTime(&vsnbuf.v_loc[(lastloc+i)%HISTORY],
				       &vinfo.history[i]) )
			return( NULL );
	}
	return( &vinfo );
}

/***********************************************************
						FINDVSN()
 ***********************************************************/
/*
 *  FindVsn() - see if record exists in index
 *	return 0 if no error, otherwise set VMError & EL_CODE 
 *	and return 1
 */
 int
FindVsn(vsn)
 vsn_t	*vsn;
{
	char	*id = "FindVsn";


	vLogMsg(EL_DEBUG4, id, "entered %.6s", vsn);
	clearkey(&db_vsn_entry);
	sprintf(db_vsn_entry.key, "%.6s", vsn);
	if (find_key(&db_vsn_entry, &db_vsn_idx) == IX_FAIL) {
		VMError = E_NOTFOUND;
		vLogMsg(EL_STATUS1|EL_CODE, id,
			"volser (%.6s) not found", vsn);
		return(1);
	}
	return(0);
}


/***********************************************************
						GETVSN()
 ***********************************************************/
/*
 *  GetVsn() - get vsn record into vsnbuf and translate it
 *	return pointer if no error, otherwise set VMError & EL_CODE 
 *	and return NULL
 */

 spacetime_t *
GetVsn(vsn)
 vsn_t	*vsn;
{
	char		*id = "GetVsn";
	vsnloc_t	*vlp;

	vLogMsg(EL_DEBUG3, id, "entered %.6s", vsn);
	/*
	 *  look up vsn in index
	 */
	if (FindVsn(vsn))
	    return(NULL);

	if ((vlp = VsnAddrToLocRec(vsn)) == NULL)
		return(NULL);
	return(&vlp->location);
}


/***********************************************************
 						SpaceTimeToLocRec()
 ***********************************************************/

/*
 *  SpaceTimeToLocRec() - Convert the spacetime struct into
 *  a locrec structure.
 *  return a 1 if an error occurs, else return 0
 */
 int
SpaceTimeToLocRec(stp, lrp)
 spacetime_t	*stp;
 locrec_t	*lrp;
{
	char	*id = "SpaceTimeToLocRec";

	/*
	 *  check time stamp and provide one if needed.
	 */

	if (stp->when == 0) {
		time_t	now = time((long *) 0);

		if (now != -1)
			stp->when = now;
		else {
			vLogMsg(EL_SYS|EL_ERROR2, id,
				"time() error %d", errno);
			return(1);
		}
	}
	FieldFill(lrp->l_date, ctime(&stp->when), DATELEN);
	lrp->a01 = ' ';

	switch (stp->where.tag) {
	case	unknown:
		lrp->l_type = UNKNOWN;
		memset(lrp->l_field.vu_general, ' ', LOCSIZE);
		break;
	case	vault:
		lrp->l_type = VAULT;
		FieldFill(lrp->l_field.vu_vault.vr_section,
			stp->where.section, SECSIZE);
		if (SetInt(lrp->l_field.vu_vault.vr_row, stp->where.row) ||
		    SetInt(lrp->l_field.vu_vault.vr_slot, stp->where.slot))
			return(1);
		break;
	case	display:
		lrp->l_type = DISPLAY;
		FieldFill(lrp->l_field.vu_general,
			  stp->where.general, LOCSIZE);
		break;
	}

	vsnbuf.end = '\n';
	return(0);
}

 int
LocEqual(a, b)
 location_t	*a, *b;
{
	
	int	equal = 0;
	
	if (a->tag == b->tag) {
		switch (a->tag) {
		case unknown:
			equal = 1;
			break;
		case vault:
			if (strncmp(a->section, b->section, SECSIZE) == 0 &&
			    a->row == b->row &&
			    a->slot == b->slot)
				equal = 1;
			break;
		case display:
			if (strncmp(a->general, b->general, LOCSIZE) == 0)
				equal = 1;
			break;
		}
		if (equal) {
			vLogMsg(EL_STATUS1, NULL,
				"repeat of location");
		}
	}
	return(equal);
}

/***********************************************************
 						UpdateVsn()
 ***********************************************************/

/*
 *  UpdateVsn() - Find the record for this vsn if it exists
 *  and update the information.  Add a new record of needed.
 *  return 0 if all is well, on error return 1
 */

 int
UpdateVsn(info)
 vsnloc_t	*info;
{
	char	*id = "UpdateVsn";
	int	ret, lastloc, nextloc;
	extern	spacetime_t	*PrevLoc;
	static	spacetime_t	oldloc;


	vLogMsg(EL_STATUS1, id, "%.6s", &info->vsn);
	PrevLoc = NULL;

	/*
	 *  see if vsn exists already
	 */

	if (FindVsn(&info->vsn) == 0) {
		/*
		 *  it does, fill in complete history from database
		 */
		if (VsnRecCheck(&vsnbuf))
			return(1);

		if (memcmp(vsnbuf.v_vsn, &info->vsn, VSNLEN) != 0) {
			VMError = E_DBASE;
			vLogMsg(EL_RPCON|EL_CODE, NULL, 
				"%.6s: index found wrong record: %.6s",
				&info->vsn, vsnbuf.v_vsn);
			return(1);
		}

		lastloc = vsnbuf.v_lastloc - '0';
		nextloc = (lastloc + 1) % HISTORY;
		vsnbuf.v_lastloc = nextloc + '0';

		if (LocRecToSpaceTime(&vsnbuf.v_loc[lastloc], &oldloc))
			return(1);

		/*
		 *  set global to inform anyone interested in the
		 *  previous location
		 */
		PrevLoc = &oldloc;

		/*
		 *  Check to see if the location is stale
		 */
		if (info->location.when < oldloc.when) {
			vLogMsg(EL_ERROR2, id,
				"stale date %d, lastloc %d",
				info->location.when, oldloc.when);
			return(1);
		}
		if (LocEqual(&info->location.where, &oldloc.where)) {
			PrevLoc = &info->location;
			return(0);
		}

		if (SpaceTimeToLocRec(&info->location,
				      &vsnbuf.v_loc[nextloc]))
			return(1);

		/*
		 *  add new location to index
		 */
		if (AddLocToIndex(&vsnbuf.v_loc[nextloc]))
			return(1);
		/*
		 *  delete old location from index
		 */
		if (DeleteLocFromIndex(&vsnbuf.v_loc[lastloc]))
			return(1);

		if (tbl_op(OP_WRITE, &vsnbuf) == TBL_ERR)
			return(1);

		vLogMsg(EL_DEBUG1, id,
			"%.6s: updated key %s addr %d", vsnbuf.v_vsn,
			db_vsn_entry.key, vsnaddr);
	
		vLogMsg(EL_STATUS1, id, "updated");
	}
	else {
		/*
		 *  doesn't exist, set up record & add vsn to table
		 */
		memset(&vsnbuf, ' ', sizeof(vsnbuf));
		memcpy(vsnbuf.v_vsn, &info->vsn, VSNLEN);
		vsnbuf.v_lastloc = '0';

		if (SpaceTimeToLocRec(&info->location, &vsnbuf.v_loc[0]))
			return(1);
		if (tbl_op(OP_APPEND, &vsnbuf) == TBL_ERR)
			return(1);

		/*
		 *  add vsn to index
		 */

		clearkey(&db_vsn_entry);
		db_vsn_entry.recptr = vsnaddr;
		sprintf(db_vsn_entry.key, "%.*s", VSNLEN, &info->vsn);
		if (add_key(&db_vsn_entry, &db_vsn_idx) == IX_FAIL) {
			VMError = E_DBASE;
			vLogMsg(EL_CODE|EL_RPCON, id,
				"%.6s: add_key() failure",
				&info->vsn);
			return(1);
		}

		/*
		 *  add location to index
		 */

		if (AddLocToIndex(&vsnbuf.v_loc[0]))
			return(1);

		vLogMsg(EL_DEBUG1, id,
			"%.6s: added key %s addr %d", vsnbuf.v_vsn,
			db_vsn_entry.key, vsnaddr);
	
		vLogMsg(EL_STATUS1, id, "added");
	}
	return(0);
}

/***********************************************************
 						DELETEVSN()
 ***********************************************************/

/*
 *  DeleteVsn() - change status & delete from indexes
 *	(other indexes will require checking if vsn deleted)
 *	return: pointer to the vsn's last location if ok,
 *	otherwise NULL & set EL_CODE
 */

 spacetime_t	*
DeleteVsn(vsn)
 vsn_t	vsn;
{
	char		*id = "DeleteVsn";
	int		lastloc;
	spacetime_t	*stp;

	VMError = E_NOERROR;

	/*
	 *  get record
	 */

	if ((stp = GetVsn(vsn)) == NULL)
		return(NULL);	/* not there */

	/*
	 *  update status & write back record
	 */
	vsnbuf.v_status = S_DELETED;

	if (tbl_op(OP_WRITE, &vsnbuf)) {
		VMError = E_DBASE;
		vLogMsg(EL_ERROR1|EL_RPCON|EL_CODE, id, 
			"dbase table error on %s", db_vsn_entry.key);
		return(NULL);
	}

	/*
	 *  delete from vsn index 
	 */

	clearkey(&db_vsn_entry);
	db_vsn_entry.recptr = vsnaddr;
	sprintf(db_vsn_entry.key, "%.6s", vsn); 
	if (delete_key(&db_vsn_entry, &db_vsn_idx) == IX_FAIL) {
		VMError = E_DBASE;
		vLogMsg(EL_RPCON, id, 
			"dbase vsn delete_key() error on (%.6s)",
			db_vsn_entry.key);
		return(NULL);
	}

	lastloc = vsnbuf.v_lastloc - '0';
	(void)DeleteLocFromIndex(&vsnbuf.v_loc[lastloc]);

	return(stp);
}
