/*
 *  cl_lib.c - NAStore Virtual Volume Manager client interface
 *
 *	Bill Ross, NASA Ames, March 1998
 */

#include <ctype.h>
#include "vv_cl.h"

static int	connected = 0;

static vvSeq_t	seq = 1;

static VVInfo_t	VVInfo;
static Vsn_t	VsnInfo;

static int 	HandleMagic = 11223344;	/* for checking what users pass in */

char	vv_errstr[MAXVVERRLEN];

  char *
vv_handle2name(vv_handle_t *vhp)
{
	VVError = VVE_NOERROR;

	if (vhp->v_magic != HandleMagic) {
		strcpy(vv_errstr, "bad vv_handle_t");
		VVError = VVE_USER;
		return(NULL);
	}
	return(vhp->v_name);
}


/*
 *  vv_init() - connect to vvmd; if hostname NULL, use localhost
 */
  int
vv_init(char *hostname, int retry)
{
	VVError = VVE_NOERROR;

	if (connected)
		return(0);

	if (VVCLInit(hostname, VVM_CLIENT, retry)) {
		VVError = VVE_UNAVAIL;
		strcpy(vv_errstr, nErrStr);
		return(1);
	}
	connected = 1;
	return(0);
}

/*
 *  vv_term() - close conn to vvmd
 */
  void
vv_term()
{
	VVError = VVE_NOERROR;

	if (connected) {
		VVCLGoodbye(1);
		connected = 0;
	}
}

static int
seqcheck(vvSeq_t got, vvSeq_t expected)
{
	if (got == expected)
		return(0);

	VVError = VVE_INTERNAL;
	sprintf(vv_errstr, "got bad pkt seq %d expected %d\n",
			got, expected);
	return(1);
}

/*
**	CLMsg2Handle() - send a client request to vvmd & get back
**		a handle; used internally.
**		NOTE: '*id' is for debugging.
*/
  static vv_handle_t *
CLMsg2Handle(vvpHdr_t *hp, char *id, int write)
{
	vv_handle_t	*vhp;
	VVInfo_t	*vip;
	vvpError_t	*vep;
	vvpHdr_t	*rhp;
	char		*fname;

	VVError = VVE_NOERROR;

	if (VVCLSend(hp, 1)) {
		vv_term();
		VVError = VVE_COMM;
		strcpy(vv_errstr, nErrStr);
		return(NULL);
	}

	/*
	**	since only VVCLRecv is used to drive the
	**		main loop - i.e. no select())  - no need to use
	**		the result of VVCLNextRecord() to see if any pkt
	**		is buffered.
	*/
	rhp = VVCLRecv("response");

	if (rhp == NULL) {
		vv_term();
		VVError = VVE_COMM;
		strcpy(vv_errstr, nErrStr);
		return(NULL);
	}
	(void)VVCLNextRecord();
	switch (rhp->vh_type) {

	    case VVM_VOLINFO:
		if (seqcheck(rhp->vh_seq, hp->vh_seq)) {
			nFree(rhp);
			return(NULL);
		}

		/*
		**	success - set name field
		*/
		vip = VV_MSGPTR(rhp, VVInfo_t);
		vhp = nAlloc(sizeof(vv_handle_t));
		vhp->v_magic = HandleMagic;
		vhp->v_fd = -1;
		vhp->v_hdroffset = -1;
		vhp->v_writing = 0;
		vhp->v_info = *vip;
		strcpy(vhp->v_name, VV2Name(&vhp->v_info.v_vv));
		nFree(rhp);
		break;

	    case VVM_VVERROR:
		vep = VV_MSGPTR(rhp, vvpError_t);
		VVError = vep->vn_error;
		sprintf(vv_errstr, "error from vvmd: %.*s", 
						LINELEN, vep->vn_msg);
		nFree(rhp);
		return(NULL);

	    case VVM_SHUTDOWN:
		connected = 0;
		sprintf(vv_errstr, "vvmd shutdown received");
		nFree(rhp);
		VVError = VVE_DOWN;
		return(NULL);

	    default:
		sprintf(vv_errstr, "unexpected vvmd response %s",
						vvHdrStr(rhp));
		nFree(rhp);
		VVError = VVE_INTERNAL;
		return(NULL);
	}

	/*
	**	success
	*/

	if (write == -1)
		return(vhp);

	/*
	**	open the vv file 
	*/
	fname = VV2Path(&vhp->v_info.v_vv);
	if (write) {
		vhp->v_fd = open(fname, O_RDWR);
		vhp->v_writing = 1;
	} else
		vhp->v_fd = open(fname, O_RDONLY);

	if (vhp->v_fd == -1) {
		int	serrno = errno;
		char	buf[100];

		sprintf(buf, "opening vv file %s: %s", fname, strerror(errno));
		(void)vv_close(vhp, 0);
		nFree(vhp);
		errno = serrno;
		VVError = VVE_SYS;
		strcpy(vv_errstr, buf);
		return(NULL);
	}

	/*
	**	note hdr offset
	*/
	vhp->v_hdroffset = CheckDiskHdr(vhp->v_fd, fname,
				vhp->v_info.v_vv.class, 
				vhp->v_info.v_vv.serial, 
				vhp->v_info.v_id);
	if (vhp->v_hdroffset == -1) {
		int	serrno = errno;
		int	vverr = VVError;
		char	buf[100];

		(void)vv_close(vhp, 0);
		nFree(vhp);
		errno = serrno;
		VVError = vverr;
		strcpy(vv_errstr, buf);
		return(NULL);
	}
	return(vhp);
}

  int
vv_close(vv_handle_t *vhp, int finish)
{
	char		buf[VV_PKTSIZ(vvpVol_t)];
	int		cmd;
	vvpHdr_t	*hp, *rhp;
	vvpVol_t	*vp;
	vvpError_t	*vep;

	VVError = VVE_NOERROR;

	if (vhp->v_magic != HandleMagic) {
		strcpy(vv_errstr, "bad vv_handle_t");
		VVError = VVE_USER;
		return(-1);
	}
	if (vhp->v_fd != -1) {
		close(vhp->v_fd);
		vhp->v_fd = -1;
	}

	/*
	**	build request
	*/
	if (!vhp->v_writing) {
		if (finish) {
			strcpy(vv_errstr, "'finish' used on rdonly vv");
			VVError = VVE_USER;
			return(-1);
		}
		return(0);
	}

	if (!connected) {
		strcpy(vv_errstr, "not connected to vvmd");
		VVError = VVE_USER;
		return(-1);
	}

	cmd = VVM_CWRITE;
	hp = vvBuildHdr(cmd, seq++, (vvpHdr_t *)buf);
	vp = VV_MSGPTR(hp, vvpVol_t);

	vp->vo_vv = vhp->v_info.v_vv;
	vp->vo_flag = finish;

	/*
	**	send request
	*/
	if (VVCLSend(hp, 1)) {
		vv_term();
		strcpy(vv_errstr, nErrStr);
		VVError = VVE_COMM;
		return(1);
	}

	/*
	**	get response
	**
	**	since only VVCLRecv is used to drive the
	**		main loop - i.e. no select())  - no need to use
	**		the result of VVCLNextRecord() to see if any pkt
	**		is buffered.
	*/
	rhp = VVCLRecv("response");

	if (rhp == NULL) {
		vv_term();
		strcpy(vv_errstr, nErrStr);
		VVError = VVE_COMM;
		return(1);
	}
	(void)VVCLNextRecord();
	switch (rhp->vh_type) {

	    case VVM_ACK:
		if (seqcheck(rhp->vh_seq, hp->vh_seq)) {
			nFree(rhp);
			return(1);
		}
		nFree(rhp);
		return(0);

	    case VVM_VVERROR:
		vep = VV_MSGPTR(rhp, vvpError_t);
		VVError = vep->vn_error;
		sprintf(vv_errstr, "error from vvmd: %s", vep->vn_msg);
		nFree(rhp);
		return(1);

	    case VVM_SHUTDOWN:
		connected = 0;
		sprintf(vv_errstr, "vvmd shutdown received");
		nFree(rhp);
		VVError = VVE_DOWN;
		return(1);

	    default:
		sprintf(vv_errstr, "unexpected vvmd response %s",
						vvHdrStr(rhp));
		nFree(rhp);
		VVError = VVE_INTERNAL;
		return(1);
	}
}

  vv_handle_t *
vv_openwrite(char *xid, char *class, OFF_T want2write, OFF_T reservesize)
{
	char		*id = "vv_openwrite";
	char		buf[VV_PKTSIZ(vvpOpenwrite_t)];
	int		i;
	size_t		len;
	vvpHdr_t	*hp;
	vvpOpenwrite_t	*op;
	vv_handle_t	*vhp;
	OFF_T		data, maxdata, fsize;

	VVError = VVE_NOERROR;

	if (!connected) {
		strcpy(vv_errstr, "not connected to vvmd");
		VVError = VVE_USER;
		return(NULL);
	}

	/*
	**	build request hdr
	*/
	hp = vvBuildHdr(VVM_OWRITE, seq++, (vvpHdr_t *)buf);
	op = VV_MSGPTR(hp, vvpOpenwrite_t);

	/*
	**	check args & fill in request
	*/
	if (want2write < 1  ||  reservesize < 0) {
		VVError = VVE_USER;
		sprintf(vv_errstr, "bad size param(s)");
		return(NULL);
	}
	SetHexOFF(op->ow_size, sizeof(op->ow_size), want2write + reservesize);

	len = strlen(xid);
	if (len == 0) {
		VVError = VVE_USER;
		sprintf(vv_errstr, "id 0 length");
		return(NULL);
	}
	if (len > sizeof(op->ow_id)) {
		VVError = VVE_USER;
		sprintf(vv_errstr, "id too long, limit %d [%s]", 
						(int)sizeof(op->ow_id), xid);
		return(NULL);
	}
	for (i=0; i<len; i++) {
		if (!isprint(xid[i])  ||  isspace(xid[i])) {
			VVError = VVE_USER;
			sprintf(vv_errstr, "id has illegal char %d [%c]",
						xid[i], xid[i]);
			return(NULL);
		}
	}
    
			
	memset(op->ow_id, ' ', sizeof(op->ow_id));
	memcpy(op->ow_id, xid, len);

	len = strlen(class);
	if (len == 0) {
		VVError = VVE_USER;
		sprintf(vv_errstr, "class name 0 length");
		return(NULL);
	}
	if (len > sizeof(op->ow_class)) {
		VVError = VVE_USER;
		sprintf(vv_errstr, "class name too long, limit %d [%s]", 
					(int)sizeof(op->ow_class), class);
		return(NULL);
	}
	memset(op->ow_class, ' ', sizeof(op->ow_class));
	memcpy(op->ow_class, class, len);

	/*
	**	get handle on vv
	*/
	vhp = CLMsg2Handle(hp, id, 1);
	if (vhp == NULL)
		return(NULL);

	/*
	**	remember how much to save at the end
	*/
	vhp->v_reserve = reservesize;

	/*
	**	position to eof
	*/
	if (C2BIN(HexOFF, &data, vhp->v_info.v_data)  ||
	    C2BIN(HexOFF, &maxdata, vhp->v_info.v_maxdata)) { 
		sprintf(vv_errstr, "%s: bad size info field(s)", vhp->v_name);
		VVError = VVE_INTERNAL;
		return(NULL);
	}
	fsize = LSEEK(vhp->v_fd, 0, SEEK_END);
	if (fsize == -1) {
		int	serrno = errno;
		char	*fname = VV2Path(&vhp->v_info.v_vv);

		sprintf(vv_errstr, "%s: lseek vv %s: %s", 
						fname, vhp->v_name,
						strerror(errno));
		(void)vv_close(vhp, 0);
		errno = serrno;
		VVError = VVE_SYS;
		return(NULL);
	}
	if (data == 0)
		data = DiskHdrSize();
	if (fsize != data) {
		VVError = VVE_INTERNAL;
		sprintf(vv_errstr, "%s: file sizes inconsistent (%lld/%lld)",
				vhp->v_name, (long long)data, (long long)fsize);
		return(NULL);
	}
	vhp->v_writable = maxdata - fsize - reservesize;
	return(vhp);
}

  static int
writecheck(vv_handle_t *vhp, char *use)
{
	static	char	id[] = "writecheck";

	if (vhp->v_magic != HandleMagic) {
		sprintf(vv_errstr, "(%s) bad vv_handle_t", use);
		VVError = VVE_USER;
		return(1);
	}
	if (vhp->v_info.v_use != VVU_WRITING) {
		VVError = VVE_USER;
		sprintf(vv_errstr, "%s (%s): read-opened vv", id, vhp->v_name);
		return(1);
	}
	if (vhp->v_writing == 0) {
		VVError = VVE_USER;
		sprintf(vv_errstr, "%s: %s not opened for write",
			id, vhp->v_name);
		return(1);
	}
	if (vhp->v_fd == -1) {
		VVError = VVE_USER;
		sprintf(vv_errstr, "%s (%s): VV was closed", id, vhp->v_name);
		return(1);
	}
	return(0);
}

/*
 *  readcheck() - make sure vv_handle_t was opened for read
 */
  static int
readcheck(vv_handle_t *vhp, char *use)
{
	if (vhp->v_magic != HandleMagic) {
		sprintf(vv_errstr, "(%s) bad vv_handle_t", use);
		VVError = VVE_USER;
		return(-1);
	}
	if (vhp->v_writing) {
		VVError = VVE_USER;
		sprintf(vv_errstr, "%s: can't %s write-opened vv", 
						vhp->v_name, use);
		return(1);
	}
	if (vhp->v_fd == -1) {
		VVError = VVE_USER;
		sprintf(vv_errstr, "%s (%s): VV was closed", vhp->v_name, use);
		return(1);
	}
	return(0);
}

/*
 *  vv_available() - report available space in current mode
 *	(main body or reserved portion)
 */
  OFF_T
vv_available(vv_handle_t *vhp)
{
	VVError = VVE_NOERROR;

	if (writecheck(vhp, "vv_available"))
		return((OFF_T)-1);
	return(vhp->v_writable);
}

  OFF_T
vv_tell(vv_handle_t *vhp)
{
	OFF_T	pos;

	VVError = VVE_NOERROR;

	if (writecheck(vhp, "vv_tell"))
		return(-1);

	pos = LSEEK(vhp->v_fd, (OFF_T)0, SEEK_CUR);
	if (pos == -1) {
		VVError = VVE_SYS;
		sprintf(vv_errstr, "%s: lseek: %s", 
						VV2Path(&vhp->v_info.v_vv),
						strerror(errno));
		return(-1);
	}
	pos -= (OFF_T)vhp->v_hdroffset;
	return(pos);
}

/*
 *  vv_write() - write to handle opened w/ vv_openwrite()
 */
  ssize_t
vv_write(vv_handle_t *vhp, void *bp, size_t len)
{
	OFF_T	written;

	VVError = VVE_NOERROR;

	if (writecheck(vhp, "vv_write"))
		return(-1);

	if (vhp->v_writable == 0) {
		VVError = VVE_NOSPC;
		sprintf(vv_errstr, "%s: VV has no space", vhp->v_name);
		return(-1);
	}

	/*
	 *  cut req down if necc
	 */
	if (len > vhp->v_writable)
		len = (size_t)vhp->v_writable;

	written = write(vhp->v_fd, bp, len);
	if (written == -1) {
		VVError = VVE_SYS;
		sprintf(vv_errstr, "%s: write: %s",
					VV2Path(&vhp->v_info.v_vv),
					strerror(errno));
		return(-1);
	}
	if (written > 0)
		vhp->v_writable -= written;
	return(written);
}

/*
 *  vv_startreserve() - start tracking reserved bytes from
 *	vv_openwrite()
 */
  int
vv_startreserve(vv_handle_t *vhp)
{
	char	*id = "vv_startreserve";

	VVError = VVE_NOERROR;

	if (writecheck(vhp, id))
		return(1);
	if (vhp->v_writing == -1) {
		VVError = VVE_USER;
		sprintf(vv_errstr, "%s: already called vv_startreserve()",
							vhp->v_name);
		return(1);
	}
	vhp->v_writing = -1;
	vhp->v_writable = vhp->v_reserve;
	return(0);
}

/*
 *  vv_finish() - mark VV as no longer appendable
 *	(therefore eligible to be copied to tape
 *	and removed from disk)
 */
  vv_handle_t *
vv_finish(char *name)
{
	char		*id = "vv_finish";
	char		buf[VV_PKTSIZ(vvpVol_t)];
	vvpHdr_t	*hp;
	vvpVol_t	*vp;
	VV_t		*vvp;

	VVError = VVE_NOERROR;

	if (!connected) {
		strcpy(vv_errstr, "not connected to vvmd");
		VVError = VVE_USER;
		return(NULL);
	}

	vvp = Name2VV(name, 1);
	if (vvp == NULL)
		return(NULL);

	/*
	 *  build request
	 */
	hp = vvBuildHdr(VVM_FINISH, seq++, (vvpHdr_t *)buf);
	vp = VV_MSGPTR(hp, vvpVol_t);

	vp->vo_vv = *vvp;
	
	/*
	 *  get handle on vv, which means finish is ok
	 */
	return(CLMsg2Handle(hp, id, -1));
}


  vv_handle_t *
vv_openread(char *name)
{
	char		*id = "vv_openread";
	char		buf[VV_PKTSIZ(vvpVol_t)];
	vvpHdr_t	*hp;
	vvpVol_t	*vp;
	vv_handle_t	*vhp;
	int		i;
	VV_t		*vvp;

	VVError = VVE_NOERROR;

	if (!connected) {
		strcpy(vv_errstr, "not connected to vvmd");
		VVError = VVE_USER;
		return(NULL);
	}

	vvp = Name2VV(name, 1);
	if (vvp == NULL)
		return(NULL);

	/*
	 *  build request
	 */
	hp = vvBuildHdr(VVM_OREAD, seq++, (vvpHdr_t *)buf);
	vp = VV_MSGPTR(hp, vvpVol_t);

	vp->vo_vv = *vvp;
	
	/*
	 *  get handle on vv, which means reading is ok
	 */
	for (i=1; i<20; i++) {
		vhp = CLMsg2Handle(hp, id, 0);
		if (vhp != NULL)
			break;
		if (VVError == VVE_UNAVAIL) {
			sleep((unsigned int)(i*i*i));
			continue;
		}
		return(NULL);
	}
	if (vhp == NULL) {
		if (VVError == VVE_UNAVAIL)
			sprintf(vv_errstr, "timed out");
		return(NULL);
	}

	/*
	 *  position after the VVM hdr
	 */
	if (LSEEK(vhp->v_fd, (OFF_T)vhp->v_hdroffset, SEEK_SET) == -1) {
		int	serrno = errno;
		char	hbuf[100];

		sprintf(hbuf, "%s: lseek to end of vvm hdr: %s", 
					VV2Path(&vhp->v_info.v_vv), 
					strerror(errno));
		(void)vv_close(vhp, 0);
		nFree(vhp);
		errno = serrno;
		VVError = VVE_SYS;
		strcpy(vv_errstr, hbuf);
		return(NULL);
	}
	return(vhp);
}

  OFF_T
vv_lseek(vv_handle_t *vhp, OFF_T offset, int whence)
{
	OFF_T	ret;

	VVError = VVE_NOERROR;

	if (readcheck(vhp, "lseek"))
		return(-1);

	/*
	 *  if seeking from beginning, allow for VVM hdr
	 */
	if (whence == SEEK_SET)
		offset += (OFF_T) vhp->v_hdroffset;

	ret = LSEEK(vhp->v_fd, offset, whence);
	if (ret == -1) {
		VVError = VVE_SYS;
		sprintf(vv_errstr, "%s: lseek: %s", 
						VV2Path(&vhp->v_info.v_vv),
						strerror(errno));
		return(ret);
	}
	/*
	 *  allow for VVM hdr in return value too
	 */
	return(ret - (OFF_T)vhp->v_hdroffset);
}

  ssize_t
vv_read(vv_handle_t *vhp, void *bp, size_t len)
{
	char	*id = "vv_read";
	ssize_t	ret;

	VVError = VVE_NOERROR;

	if (readcheck(vhp, id))
		return(-1);

	ret = read(vhp->v_fd, bp, len);
	if (ret == -1) {
		VVError = VVE_SYS;
		sprintf(vv_errstr, "%s: read: %s", 
						VV2Path(&vhp->v_info.v_vv),
						strerror(errno));
		return(-1);
	}
	return(ret);
}


  VVInfo_t *
vv_info(char *name)
{
	char		*id = "vv_info";
	char		buf[VV_PKTSIZ(vvpVolqry_t)];
	vvpHdr_t	*hp;
	vvpVolqry_t	*vqp;
	VV_t		*vvp;
	vv_handle_t	*vhp;

	VVError = VVE_NOERROR;

	if (!connected) {
		strcpy(vv_errstr, "not connected to vvmd");
		VVError = VVE_USER;
		return(NULL);
	}

	vvp = Name2VV(name, 1);
	if (vvp == NULL)
		return(NULL);

	hp = vvBuildHdr(VVM_VOLQRY, seq++, (vvpHdr_t *)buf);

	hp->vh_count = 1; /* no ACK expected per CLMsg2Handle() */

	vqp = VV_MSGPTR(hp, vvpVolqry_t);

	vqp->vq_vv = *vvp;
	vqp->vq_flags = VQF_VOL;

	vhp = CLMsg2Handle(hp, id, -1);
	if (vhp == NULL)
		return(NULL);

	memcpy(&VVInfo, &vhp->v_info, sizeof(VVInfo));
	return(&VVInfo);
}

  VVInfo_t *
vv_handle2info(vv_handle_t *vhp)
{
	VVError = VVE_NOERROR;

	if (vhp->v_magic != HandleMagic) {
		strcpy(vv_errstr, "bad vv_handle_t");
		VVError = VVE_USER;
		return(NULL);
	}
	return(&vhp->v_info);
}


/*
 *  vv_recycle() - throw away VV
 */
  int
vv_recycle(char *name)
{
	VV_t		*vvp;
	char		buf[VV_PKTSIZ(vvpVol_t)];
	vvpHdr_t	*hp, *rhp;
	vvpVol_t        *vop;
	vvpError_t	*vep;
	
	VVError = VVE_NOERROR;

	if (!connected) {
		strcpy(vv_errstr, "not connected to vvmd");
		VVError = VVE_USER;
		return(-1);
	}

	vvp = Name2VV(name, 1);
	if (vvp == NULL)
		return(1);

	hp = vvBuildHdr(VVM_VRECYCLE, seq++, (vvpHdr_t *)buf);
	vop = VV_MSGPTR(hp, vvpVol_t);
	vop->vo_vv = *vvp;

	/*
	**	send request
	*/
	if (VVCLSend(hp, 1)) {
		vv_term();
		strcpy(vv_errstr, nErrStr);
		VVError = VVE_COMM;
		return(1);
	}

	/*
	**	get response
	**
	**	since only VVCLRecv is used to drive the
	**		main loop - i.e. no select())  - no need to use
	**		the result of VVCLNextRecord() to see if any pkt
	**		is buffered.
	*/
	rhp = VVCLRecv("response");

	if (rhp == NULL) {
		vv_term();
		strcpy(vv_errstr, nErrStr);
		VVError = VVE_COMM;
		return(1);
	}
	(void)VVCLNextRecord();
	switch (rhp->vh_type) {

	    case VVM_ACK:
		if (seqcheck(rhp->vh_seq, hp->vh_seq)) {
			nFree(rhp);
			return(1);
		}
		nFree(rhp);
		return(0);

	    case VVM_VVERROR:
		vep = VV_MSGPTR(rhp, vvpError_t);
		VVError = vep->vn_error;
		sprintf(vv_errstr, "error from vvmd: %s", vep->vn_msg);
		nFree(rhp);
		return(1);

	    case VVM_SHUTDOWN:
		connected = 0;
		sprintf(vv_errstr, "vvmd shutdown received");
		nFree(rhp);
		VVError = VVE_DOWN;
		return(1);

	    default:
		sprintf(vv_errstr, "unexpected vvmd response %s",
						vvHdrStr(rhp));
		nFree(rhp);
		VVError = VVE_INTERNAL;
		return(1);
	}
}

/*
 *  vv_destroyhandle() - free memory used by handle & set pointer to NULL
 */
  int
vv_destroyhandle(vv_handle_t **vhpP)
{
	if ((*vhpP)->v_magic != HandleMagic) {
		VVError = VVE_USER;
		sprintf(vv_errstr, "bad handle");
		return(1);
	}
	if ((*vhpP)->v_fd != -1) {
		VVError = VVE_USER;
		sprintf(vv_errstr, "bad handle");
		return(1);
	}
	nFree(*vhpP);
	*vhpP = NULL;
	return(0);
}

#if 0
/* not needed w/ vvmd's sorted LRU cache mgt  */
  int
vv_spacefree(int free)
{
	char		buf[VV_PKTSIZ(vvpDebug_t)];
	vvpHdr_t	*hp, *rhp;
	vvpDebug_t	*dp;
	vvpError_t	*vep;

	VVError = VVE_NOERROR;

	if (!connected) {
		strcpy(vv_errstr, "not connected to vvmd");
		VVError = VVE_USER;
		return(-1);
	}
	hp = vvBuildHdr(VVM_DEBUG, seq++, (vvpHdr_t *)buf);
	dp = VV_MSGPTR(hp, vvpDebug_t);
        dp->vd_cmd = 3;
        dp->vd_level = free;

	/*
	 *  msg is built - send it
	 */
	(void)VVCLSend(hp, 1);		/* dies if problem */
}
#endif /* CROLF */

  Vsn_t *
vv_vsninfo(char *vsnp)
{
	char		buf[VV_PKTSIZ(vvpVsn_t)];
	vvpHdr_t	*hp;
	vvpVsn_t	*vp;

	VVError = VVE_NOERROR;

	if (checkvsn(vsnp)) {
		VVError = VVE_USER;
		strcpy(vv_errstr, nErrStr);
		return(NULL);
	}
	if (!connected) {
		strcpy(vv_errstr, "not connected to vvmd");
		VVError = VVE_USER;
		return(NULL);
	}

	hp = vvBuildHdr(VVM_VSNQRY, seq++, (vvpHdr_t *)buf);

	hp->vh_count = 1;	/* no ACK expected */

	vp = VV_MSGPTR(hp, vvpVsn_t);

	vp->vq_flags = VQF_VSN;
	memcpy(vp->vq_vsn, vsnp, sizeof(vp->vq_vsn));

	if (VVCLSend(hp, 1)) {
		strcpy(vv_errstr, nErrStr);
		return(NULL);
	}

	/*
	**	get response(s)
	*/
	for (;;) {
		vvpHdr_t	*rhp;
		vvpError_t	*vep;

		/*
		**	since only VVCLRecv is used to drive the
		**	main loop - i.e. no select())  - no need to use
		**	the result of VVCLNextRecord() to see if any pkt
		**	is buffered.
		*/
		rhp = VVCLRecv("response");	/* dies if problem */
		(void)VVCLNextRecord();

		switch (rhp->vh_type) {
		    case VVM_VSNINFO:
			if (seqcheck(hp->vh_seq, rhp->vh_seq)) {
				nFree(rhp);
				return(NULL);
			}
			memcpy(&VsnInfo, VV_MSGPTR(rhp, Vsn_t), 
							sizeof(VsnInfo));
			nFree(rhp);
			return(&VsnInfo);

		    case VVM_VVERROR:
			vep = VV_MSGPTR(rhp, vvpError_t);
			VVError = vep->vn_error;
			strcpy(vv_errstr, vep->vn_msg);
			return(NULL);
		    case VVM_SHUTDOWN:
			connected = 0;
			VVError = VVE_DOWN;
			sprintf(vv_errstr, "vvmd shutdown received");
			return(NULL);
		    default:
			VVError = VVE_INTERNAL;
			sprintf(vv_errstr, "unexpected pkt: %s", vvHdrStr(rhp));
			return(NULL);
		}
	}
}
