/*
 *  vv_class.c - Virtual Volume Manager volume classes.
 *
 *	Changing existing classes can invalidate an existing
 *	database and lead to unpredictable confusion which
 *	can further exacerbate the crisis. Adding classes is
 *	ok.
 */

#include <ctype.h>
#include "vv_space.h"
#include "cl_space.h"

/*
 *  nvsns <= MAXVSN; must have dummy values of rc/med to fill out to MAXVSN
 */
static Class_t	ClassInfoTab[] = {
	{
		/* name		prefix		flags */
		"basic",	"X",		CS_FREEDISK,
			/* size		finishsize	nvsns */
		      100*1024*1024,    80*1024*1024,     2,
					/* rc1		type1 */
				{
#ifdef FILEDRIVES
					   { "VAULTRC",	"SQR800" },
#else
					   { "ACS_1",	"D3" },
#endif
					/* rc2		type2 */
#ifdef FILEDRIVES
					   { "VAULTRC",	"D3" }
#else
					   { "ACS_1",	"D3" }
				}
#endif
	},
	{
		NULL,		NULL,		0,
			0,		0,		0
	}
};

static char	*serchar = NULL;

  static void
InitSerChar()
{
	int	i, order = 0;

	/*
	**	set up serial char comparison array
	*/
	serchar = (char *) nAlloc((size_t)128);

	for (i='0'; i<='9'; i++)
		serchar[i] = order++;
	for (i='a'; i<='z'; i++)
		serchar[i] = order++;
	for (i='A'; i<='Z'; i++)
		serchar[i] = order++;
}

  static int
notupper(char *str)
{
	int	i;
	size_t	len = strlen(str);

	for (i=0; i<len; i++)
		if (!isupper(str[i]))
			return(1);
	return(0);
}

  int
CheckClasses()
{
	char		*id = "CheckClasses";
	Class_t		*cip;
	int		i, j;

	for (i=0,cip=ClassInfoTab; cip->cs_name; cip++,i++) {
		int	nvsns = cip->cs_nvsns;
		char	*cp;

		nLogMsg(EL_STATUS2|EL_NOTIME, id, "%s / %s",
					cip->cs_name, cip->cs_prefix);
		for (j=0; j<nvsns; j++)
			nLogMsg(EL_STATUS2|EL_NOTIME, 0, 
					"  slot %d: rc %s type %s", j,
					cip->cs_vsns[j].vs_rc,
					cip->cs_vsns[j].vs_voltype);
		if (strlen(cip->cs_prefix) > VV_CLASSLEN)
			nLogMsg(EL_FATAL, id, 
				"%s: class prefix [%s] len must be <= %d",
				cip->cs_name, cip->cs_prefix, VV_CLASSLEN);
		if (vvm_class_policy_class_len) {
			if (strlen(cip->cs_prefix) != 
						vvm_class_policy_class_len)
				nLogMsg(EL_FATAL, id, 
				    "%s: class prefix [%s] len must be %d %s",
				    cip->cs_name, cip->cs_prefix, 
				    vvm_class_policy_class_len,
				"-- see Mkconfig.vvm VVM_CLASSPOLICYCLASSLEN");
			if (notupper(cip->cs_prefix))
				nLogMsg(EL_FATAL, id, 
				    "%s: prefix [%s] must be uppercase %s %s",
				    cip->cs_name, cip->cs_prefix, 
				    "if length is limited",
				"-- see Mkconfig.vvm VVM_CLASSPOLICYCLASSLEN");
		}
		for (cp=cip->cs_prefix; *cp; cp++)
			if (!isalnum(*cp))
				nLogMsg(EL_FATAL, id, 
				  "%s: class prefix [%s] must be alphanum",
				  cip->cs_name, cip->cs_prefix);
	}
	return(i);
}

/*
 *  IncSer() - 0..9,a..z,A..Z = base-62 number system
 */
  static int
IncSer(char *buf, int size)
{
	size--;
	if (!size)
		return(1);

	if (buf[size] == '9') {
		buf[size] = 'a';
		return(0);
	}
	if (buf[size] == 'z') {
		buf[size] = 'A';
		return(0);
	}
	if (buf[size] == 'Z') {
		buf[size] = '0';
		/* carry the 1 */
		return(IncSer(buf, size));
	}
	if (buf[size] == ' ') {
		buf[size] = '1';
		return(0);
	}
	buf[size]++;
	return(0);
}

  int
CmpClassSerial(const char *s1, const char *s2)
{
	char	*id = "CmpClassSerial";
	int	check = VV_SERIALLEN - vvm_class_policy_serial_len - 1;

	if (serchar == NULL)
		InitSerChar();

	/*
	 *  figure high-order position
	 */
	if (vvm_class_policy_serial_len  &&  
				vvm_class_policy_serial_len != VV_SERIALLEN) {
		check = VV_SERIALLEN - vvm_class_policy_serial_len - 1;

		/*
		 *  check policy
		 */
		if (s1[check] != ' ')
			nLogMsg(EL_FATAL, id, "s1 bigger than policy");
		if (s2[check] != ' ')
			nLogMsg(EL_FATAL, id, "s2 bigger than policy");
		check++;
	} else
		check = 0;

	while (s1[check] == ' '  &&  s2[check] == ' '  &&  check < VV_SERIALLEN)
		check++;

	if (check == VV_SERIALLEN)
		return(0);	/* both equal 0 */

	if (s1[check] == ' ')	/* s1 < s2 */
		return(-1);
	if (s2[check] == ' ')	/* s2 < s1 */
		return(1);

	/*
	 *  same # of digits - compare
	 */
	while (check < VV_SERIALLEN) {
		if (serchar[(int)s1[check]] < serchar[(int)s2[check]])
			return(-1);	/* s1 < s2 */
		if (serchar[(int)s1[check]] > serchar[(int)s2[check]])
			return(1);	/* s2 < s1 */
		check--;
	}
	return(0);	/* equal non-0 */
}

  int
IncClassSerial(char *serial)
{
	char	*id = "IncClassSerial";
	int	check = VV_SERIALLEN - vvm_class_policy_serial_len - 1;

	if (serial[check] != ' ') {
		VVError = VVE_UNAVAIL;
		nLogMsg(EL_ERROR1|EL_CODE, id, 
			"serial [%.*s] already exceeds policy len of %d %s",
			VV_SERIALLEN, serial, 
			vvm_class_policy_serial_len,
			"-- see Mkconfig.vvm VVM_CLASSPOLICYSERIALLEN");
		return(1);
	}
	if (IncSer(serial, VV_SERIALLEN)) {
		VVError = VVE_UNAVAIL;
		nLogMsg(EL_ERROR1|EL_CODE, id, 
			"serial [%.*s] would exceed max len of %d",
			VV_SERIALLEN, serial, VV_SERIALLEN);
		return(1);
	}
	if (serial[check] != ' ') {
		VVError = VVE_UNAVAIL;
		nLogMsg(EL_ERROR1|EL_CODE, id, 
			"serial [%.*s] would exceed policy len of %d %s",
			VV_SERIALLEN, serial, 
			vvm_class_policy_serial_len,
			"-- see Mkconfig.vvm VVM_CLASSPOLICYSERIALLEN");
		return(1);
	}
	return(0);
}

  void
ListClasses()
{
	Class_t		*cip;

	fprintf(stderr, "Classes: name\tprefix\n");
	for (cip=ClassInfoTab; cip->cs_name; cip++)
		fprintf(stderr, "\t%s\t%s\n", cip->cs_name, cip->cs_prefix);
}

  void
PrintClass(Class_t *cip)
{
	int	i;

	if (cip == NULL) {
		printf("NAME\tPREFIX\tFLAGS\tMAX\tFINISH");
		for (i=0; i<MAXVSN; i++)
			printf("\tRC%d\tTYPE%d", i, i);
		printf("\n");
		return;
	}
	printf("%s\t%s\t0x%x\t%lld\t%lld",
		cip->cs_name, cip->cs_prefix, cip->cs_flags,
		(long long)cip->cs_vvsize, (long long)cip->cs_finishsize);
	for (i=0; i<cip->cs_nvsns; i++)
		printf("\t%s\t%s",
			cip->cs_vsns[i].vs_rc, cip->cs_vsns[i].vs_voltype);
	printf("\n");
}

  Class_t *
ClassInfo(char *prefix)
{
	char		*id = "ClassInfo";
	Class_t		*cip;

	for (cip=ClassInfoTab; cip->cs_name; cip++) {
		if (!strcmp(prefix, cip->cs_prefix))
			return(cip);
	}
	VVError = VVE_INTERNAL;
	nLogMsg(EL_ERROR1|EL_CODE, id, "unknown class %s", prefix);
	return(NULL);
}

  Class_t *
ClassIndex(int i)
{
	if (i < 0  ||  i > sizeof(ClassInfoTab) / sizeof(Class_t) - 1)
		return(NULL);
	if (ClassInfoTab[i].cs_name)
		return(&ClassInfoTab[i]);
	return(NULL);
}

static int	counter = 0;
  void
ResetClasses()
{
	counter = 0;
}

  Class_t *
NextClass()
{
	return(ClassIndex(counter++));
}

  char *
ClassString(char *class)
{
	static char	buf[VV_CLASSLEN+1];
	int		i;

	/*
	 *  class is a left-justified alphanum thing
	 */
	for (i=0; i<VV_CLASSLEN; i++, class++) {
		if (*class == '\0'  ||  *class == ' ')
			break;
		buf[i] = *class;
	}
	if (i == 0)
		nLogMsg(EL_FATAL, "ClassString", "no class defined");
	buf[i] = '\0';
	return(buf);
}
