/*
 *  vv_p.c - NAStore Virtual Volume Manager (VVM) communications
 *
 *	Bill Ross, NASA Ames, October 1997
 */

#include <stdlib.h>
#include <syslog.h>

#include "vv_space.h"

/*********************************************************************
 *  Mkconfig.vvm-defined globals - all here for convenience
 */
#include "vv_p.h"

/*  used both for installation & w/in progs  */

char	*VVOperAcct = VVM_ACCT;
char	*vvLibDir = VVM_LIBDIR;

/*  used only w/in progs  */

#ifdef VVMD_HOST
char	*vvHost = VVMD_HOST;
#else
char	*vvHost = NULL;
#endif

int	vvm_class_policy_class_len = VVM_CLASSPOLICYCLASSLEN;
int	vvm_class_policy_serial_len = VVM_CLASSPOLICYSERIALLEN;

char	*vvm_service = VVMD_SERVICE;
short	vvm_port;
char	*vvm_config = VVM_CONF;
char	*vvDBDir = VVM_DBDIR;
char	*vvSpoolDir = VVM_SPOOLDIR;
char	*vvVVDir = VVM_VVDIR;
int	vvVsnGroupId = VVM_GID;

/********************************************************************/

int	vv_syslog = LOG_LOCAL0;
int	VVError = VVE_NOERROR;

static char	logbuf[MAXPATHLEN];

  char *
vvLogFile()
{

	sprintf(logbuf, "%s/ErrorLog", vvSpoolDir);
	return(logbuf);
}

static char	vsnbuf[MAXPATHLEN];

  char *
vvVsnFile(const char *vsn, const char *extra)
{
	if (extra != NULL)
		sprintf(vsnbuf, "%s/VSNCHECK/%.*s_%s", 
				vvVVDir, VSNLEN, vsn, extra);
	else
		sprintf(vsnbuf, "%s/VSNCHECK/%.*s", vvVVDir, VSNLEN, vsn);
	return(vsnbuf);
}


/*
 *  VVmpTab[]: table of information about VVP message types.
 */
static ProTab_t VVmpTab[] = {
	/* request		size		*/
	{ "VVM_(zero)",	0,		0 },
	{ "VVM_MOVER",	VVM_MOVER,	sizeof(vvpHello_t) },
	{ "VVM_CLIENT",	VVM_CLIENT,	sizeof(vvpHello_t) },
	{ "VVM_TIMERD",	VVM_TIMERD,	sizeof(vvpHello_t) },
	{ "VVM_GOODBYE",	VVM_GOODBYE,	0 },
	{ "VVM_ACK",	VVM_ACK,	0 },
	{ "VVM_DEBUG",	VVM_DEBUG,	sizeof(vvpDebug_t) },
	{ "VVM_VVERROR",	VVM_VVERROR,	sizeof(vvpError_t) },
	{ "VVM_SHUTDOWN",	VVM_SHUTDOWN,	sizeof(vvpError_t) },
	{ "VVM_TIMER",	VVM_TIMER,	sizeof(vvpTimer_t) },
	{ "VVM_NEWLOG",	VVM_NEWLOG,	0 },
	{ "VVM_OWRITE",	VVM_OWRITE,	sizeof(vvpOpenwrite_t) },
	{ "VVM_CWRITE",	VVM_CWRITE,	sizeof(vvpVol_t) },
	{ "VVM_FINISH",	VVM_FINISH,	sizeof(vvpVol_t) },
	{ "VVM_VOLQRY",	VVM_VOLQRY,	sizeof(vvpVol_t) },
	{ "VVM_VOLINFO",	VVM_VOLINFO,	sizeof(VVInfo_t) },
	{ "VVM_OREAD",	VVM_OREAD,	sizeof(vvpVol_t) },
	{ "VVM_WTAPE",	VVM_WTAPE,	sizeof(vvpMvreq_t) },
	{ "VVM_RTAPE",	VVM_RTAPE,	sizeof(vvpMvreq_t) },
	{ "VVM_VRECYCLE",	VVM_VRECYCLE,	sizeof(vvpVol_t) },
	{ "VVM_VSNQRY",	VVM_VSNQRY,	sizeof(vvpVsn_t) },
	{ "VVM_VSNINFO",	VVM_VSNINFO,	sizeof(Vsn_t) },
	{ "VVM_CHECKVSN",	VVM_CHECKVSN,	sizeof(vvpVsn_t) },
	{ "VVM_BADVSN",	VVM_BADVSN,	sizeof(vvpVsn_t) },
	{ "VVM_RECYCVSN",	VVM_RECYCVSN,	sizeof(vvpVsn_t) },
	{ "VVM_COPYIX",	VVM_COPYIX,	sizeof(vvpIx_t) },
	{ "VVM_STATUS",	VVM_STATUS,	0 },
	{ "VVM_BYTES",	VVM_BYTES,	-1 },		/* variable size */
	{ "VVM_FINCLASS",	VVM_FINCLASS,	sizeof(vvpFinish_t) },
	{ "VVM_FINISHVSN",VVM_FINISHVSN,	sizeof(vvpVsn_t) },
	{ "VVM_DISKCK",	VVM_DISKCK,	0 },
	{ "VVM_NOP",	VVM_NOP,	0 },
	{ NULL,		0,		0 }
};

  char *
vvMsgType(int msgno)
{
	return(nMsgType(msgno, VVmpTab));
}

  ProTab_t *
vvMsgInfo(int msgno)
{
	return(nMsgInfo(msgno, VVmpTab));
}

  static ssize_t
vvPktSiz(int msgno)
{
	return(nPktSiz(msgno, VVmpTab));
}


static CodeString_t vvErrTab[] = {
	{ VVE_NOERROR,		"no error" },
	{ VVE_NOMEM,		"out of memory" },
	{ VVE_SYS,		"system error" },
	{ VVE_DOWN,		"RP/RC down" },
	{ VVE_INTERNAL,		"INTERNAL ERROR" },
	{ VVE_BADPACKET,	"bad packet" },
	{ VVE_IO,		"I/O error" },
	{ VVE_USER,		"User error" },
	{ VVE_DBASE,		"Database error" },
	{ VVE_ADMIN,		"Database inconsistency" },
	{ VVE_PERM,		"Permission Denied" },
	{ VVE_NOTFOUND,		"Not found" },
	{ VVE_UNAVAIL,		"Resource unavailable/in use" },
	{ VVE_OPS,		"Operations problem" },
	{ VVE_CANCELED,		"Request canceled" },
	{ VVE_COMM,		"Communications problem" },
	{ VVE_HARDWARE,		"HARDWARE PROBLEM" },
	{ VVE_FILE,		"problem with VV file" },
	{ VVE_NOSPC,		"write would exceed file size" },
	{ 0,			NULL }
};

/*
 *  vvErrStr(): convert VV error number to string
 */
  char *
vvErrStr(int err)
{
	return(nCodeToStr(err, vvErrTab));
}

CodeString_t	vvPortTab[] = {
	{ S_NOSOCK,	"(no connection)" },
	{ S_CONNECTED,	"(connection only)" },
	{ S_TIMER,	"timer" },
	{ S_MOVER,	"mover" },
	{ S_CLIENT,	"client" },
	{ 0,		NULL }
};

  char *
vvPortTypeStr(int port)
{
	return(nCodeToStr(port, vvPortTab));
}

CodeString_t	vvStatusTab[] = {
	{ VVS_HOT,	"hot/append" },
	{ VVS_FINISHED,	"finished" },
	{ VVS_MIGRATED,	"ontape" },
	{ VVS_RECOVER,	"recover" },
	{ VVS_RECYCLING,"recycling" },
	{ VVS_RECYCLED,	"recycled" },
	{ VVS_ADDED,	"added" },
	{ VVS_LOST,	"lost" },
	{ VVS_DELETED,	"deleted" },
	{ 0,		NULL}
};

  char *
vvStatusStr(char status)
{
	return(nCodeToStr((int)status, vvStatusTab));
}

CodeString_t	vvUseTab[] = {
	{ VVU_IDLE,	"idle" },
	{ VVU_WRITING,	"writing" },
	{ VVU_WTAPE,	"totape" },
	{ VVU_RTAPE,	"fromtape" },
	{ 0,		NULL }
};

  char *
vvUseStr(char use)
{
	return(nCodeToStr((int)use, vvUseTab));
}
 
/*
 *  vvHdrStr() - return a (static) string which describes the header (no \n)
 */
  char *
vvHdrStr(vvpHdr_t *hp)
{
	static char	buf[LINELEN];

	if (hp == NULL)
		return ("[NULL header]");
	if (hp->vh_magic != VVP_MAGIC) {
		sprintf(buf, "0x%llx: bad magic number: 0x%x/0x%x",
		  (long long)(long)hp, hp->vh_magic, VVP_MAGIC);
		return (buf);
	}
	sprintf(buf, "typ=%s seq=#%d sz=%d ",
	  vvMsgType(hp->vh_type), hp->vh_seq, hp->vh_size);
	return(buf);
}


/*
 *  vvCheckPacket() - perform sanity check on header.
 *	Return VVE_BADPACKET if invalid; assign VVError but don't log error;
 *	let upper level do it.
 */
  int
vvCheckPacket(vvpHdr_t *hp)
{
	char	*id = "vvCheckPacket";
	ssize_t	size;

	if (hp->vh_magic != VVP_MAGIC) {
		char	*cp = (char *) hp;
		int	i;

		nLogMsg(EL_ERROR1, id, "Bad magic: %s", vvHdrStr(hp));
		for (i=0; i<sizeof(vvpHdr_t); i++)
			if (cp[i] == VVP_MAGIC)
				nLogMsg(EL_STATUS1, id, "magic at %d", i);
		return (VVError = VVE_BADPACKET);
	}
	if (hp->vh_type > VVM_MAXTYPE) {
		nLogMsg(EL_ERROR1, id, "Bad type: %s", vvHdrStr(hp));
		return (VVError = VVE_BADPACKET);
	}
	size = vvPktSiz((int) hp->vh_type);
	if (size != -1 && size != hp->vh_size) {  /* -1 => variable length */
		nLogMsg(EL_ERROR1, id, "Bad size: %s", vvHdrStr(hp));
		return (VVError = VVE_BADPACKET);
	}
	return (VVE_NOERROR);
}


/*
 *  vvBuildHdr() - build a packet header for the specified message type;
 *	if a buffer was passed in, build it there; else alloc a buffer
 *	in any event, the entire buffer is cleared.
 */
  vvpHdr_t *
vvBuildHdr(int msg, vvSeq_t seq, vvpHdr_t *chp)
{
	vvpHdr_t	*hp;
	ssize_t		pktsiz; 
	size_t		wholepkt;

	pktsiz = vvPktSiz(msg);

	if (pktsiz == -1) 
		nLogMsg(EL_FATAL, 0, 
		 "programming error: using vvBuildHdr on custom-size pkt");
	if (pktsiz)
		wholepkt = (size_t)pktsiz + PadSize(sizeof(vvpHdr_t));
	else
		wholepkt = sizeof(vvpHdr_t);

	if (chp == NULL) {
		hp = (vvpHdr_t *) nAlloc(wholepkt);
	} else {
		hp = chp;
	}
	memset(hp, '\0', wholepkt);
	hp->vh_magic = VVP_MAGIC;
	hp->vh_type = msg;
	hp->vh_seq = seq;
	hp->vh_size = (unsigned int)pktsiz;
	hp->vh_count = 0;

	return (hp);
}

/*
 *  vvGetID() - generate somewhat unpredictable ID
 */
static unsigned idseed = 0;
 
  vvId_t
vvGetID()
{
	char		*id = "vvGetID";
	struct timeval	now;
	struct timezone	zone;
	unsigned	cnt, seed;

	/*
	 *  use microseconds to construct the random number seed
	 */
	if (gettimeofday(&now, &zone) == -1) 
		nLogMsg(EL_FATAL|EL_SYS, id, "gettimeofday");

	/*
	 *  first time through, seed system's random # generator
	 */
	if (!idseed)
		srand((unsigned int)now.tv_usec);

	/*
	 *  use system's random # to pick a number 0..64
	 */
	seed = (unsigned)rand();
	cnt = (seed>>6) % 64;

	/*
	 *  make new seed for 2nd random # generator and seed it
	 */
	idseed += (unsigned)now.tv_usec;
	seed = idseed;

	/*
	 *  index into 2nd random sequence by number 0..64 from
	 *	system random sequence
	 */
	while (cnt--)
        	seed = seed * 78125 + 59049;

	return((vvId_t)seed);

}

/*
 *  ForkExec(): path can be absolute, or it is searched for in LIBDIR
 */
  int
ForkExec(char *path, char *name, char *arg)
{
	char	*id = "ForkExec";
	char	buf[MAXPATHLEN];
	STAT_T	statb;
	int	fval;

	/* executable in libdir if not absolute path */
	if (path[0] != '/') {
		sprintf(buf, "%s/%s", vvLibDir, path);
		path = buf;
	}

	/*
	 *  simple checks before fork
	 */
	if (STAT(path, &statb) == -1) {
		if (errno == ENOENT)
			nLogMsg(EL_FATAL, id, "%s: (%s): does not exist",
		  			name, path);
		nLogMsg(EL_FATAL|EL_SYS, id, "%s: (%s): Can't stat",
		  			name, path);
	}
	if (!(statb.st_mode & 0111)) {
		nLogMsg(EL_FATAL, id, "%s: (%s): not executable (0%o)",
		  			name, path, statb.st_mode);
	}
	if ((fval = fork()) < 0) {
		nLogMsg(EL_FATAL|EL_SYS, id, "%s: (%s): Can't fork",name, path);
	} else if (fval == 0) {			/* child */
		int	i;

		nLogMsg(EL_STATUS1, id, "Exec: [pid %d] %s (%s) %s",
						getpid(), name, path, 
						(arg ? arg : ""));

		for (i = 0; i < getdtablesize(); i++)	/* close all fds */
			if (i >= 3)	/* start child off with clean slate */
				(void) close(i);	/* TODO: use v struct */

		execl(path, name, arg, 0);

		/*
		 * this message won't write to the error log
		 *	becuase the fd's have been closed; stderr's OK
		 */
		nLogMsg(EL_SYS|EL_TOSTDERR|EL_CONSOLE0, id,
					"Can't exec %s (%s)", name, path);
		exit(1);
	}
	return(fval);
}

