
/*
 *  vv_xdr.c: NAStore Virtual Volume Manager XDR routines
 *
 *  Bill Ross, NASA Ames, October 1997
 */

#include "vv_space.h"
#include <rpc/xdr.h>

/*
 *	VVM communication is over sockets via XDR, which is 
 *	a protocol (part of RPC, which is used by NFS) for making 
 *	sure that data structures are the same on hosts of different 
 *	architecture.
 *
 *	On the vvmd side, an array of sockets indexed by file descriptor
 *	is tied to a corresponding array of XDR structures. These
 *	are used for all vvmd interprocess communication. The vvmd's
 *	clients and servers ('movers') each have a single socket/XDR 
 *	connection with the vvmd.
 *
 *	For each connection, an XDR_t struct is initialized to point at 
 *	the Sock_t (socket connection mgt) structure and the routines
 *	that read & write raw bytes from the socket. 
 *
 *	xdr_{structname} routines generated by rpcgen from vv_pkt.x are 
 *	passed the XDR_t, and call primitive routines like xdr_int() 
 *	which invoke the socket read/write routines to get/put bytes 
 *	before/after converting data from/to the generic form XDR uses.
 *
 *	Higher-level packet-switched routines in this file read/write
 *	the header of a packet using xdr_vvpHdr_t() and then get/put
 *	the rest of the packet based on pkt->vh_type using the various
 *	other xdr_vvp{pktstruct} routines.
 *
 */

  void
XDRCloseSock(Sock_t *sockp)
{
	char	*id = "XDRCloseSock";

	/* see if this flushes any partly sent/rcvd pkts.. */
	xdr_destroy(sockp->sock_xdrp);
	XDRSetup(sockp, sockp->sock_xdrp);

	switch (sockp->sock_ptype) {
	    case S_NOSOCK:
		nLogMsg(EL_ERROR1, id, "attempted shutdown on non-socket");
		break;
	    default:
		shutdown(sockp->sock_fd, 2);
		if (close(sockp->sock_fd) == -1)
		    nLogMsg(EL_ERROR2|EL_SYS, id, "socket close failed");
		break;
	}
	sockp->sock_fd = -1;
	sockp->sock_id = -1;
	sockp->sock_host = NULL;
	sockp->sock_proc = NULL;
        sockp->sock_ptype = S_NOSOCK;
}

/*
**	XDRread() - called by XDR translation routines, hooked by xdrp
*/
  static int
XDRread(const Sock_t *sockp, char *buf, int size)
{
	char	*id = "XDRread";
	int	nb;

	if (sockp->sock_fd < 0) {
		nLogMsg(EL_STATUS1, id, "skipped read from sock fd %d id %d",
					sockp->sock_fd, sockp->sock_id);
		return(-1);
	}
        nb = recv(sockp->sock_fd, buf, (size_t)size, 0);
	if (nb == 0) {
		nLogMsg(EL_STATUS2, id, 
		    "recvd 0 from slot %d - writer died or closed connection", 
		    sockp->sock_fd);
		errno = ENOTCONN;
		return(-1);
	}
	nLogMsg(EL_DEBUG2, id, "recvd %d of %d from %d", 
					nb, size, sockp->sock_fd);
	return(nb);
}

/*
**	XDRwrite() - called by XDR translation routines, hooked by xdrp
*/
  static int
XDRwrite(Sock_t *sockp, char *buf, int size)
{
	char	*id = "XDRwrite";
	int	wcnt, n = 0;

	if (sockp->sock_fd < 0) {
		nLogMsg(EL_DEBUG1, id, "skipped write to sock %d id %d",
					sockp->sock_fd, sockp->sock_id);
		return(-1);
	}
		
	wcnt = send(sockp->sock_fd, buf, (size_t)size, 0);
	nLogMsg(EL_DEBUG2, id, "sent %d to %d", wcnt, sockp->sock_fd);
	/*
	**	Keep retrying the send if it would normally block.
	**		It should go thru eventually.
	*/
	while ((wcnt == -1) && (errno == EWOULDBLOCK) && (n++ < VVM_IORETRY)) {
		sleep(1);
		nLogMsg(EL_DEBUG2, id, "send would block, retry %d ...", n);
		wcnt = send(sockp->sock_fd, buf, (size_t)size, 0);
	}
	if ((wcnt == -1) && (n >= VVM_IORETRY))
		nLogMsg(EL_DEBUG1, id, "retries exhausted, giving up\n");

	if (wcnt == size) {
		nLogMsg(EL_DEBUG3, id, " all sent to %d", sockp->sock_id);
	} else {
		VVError = VVE_COMM;	/* default err */
		if (wcnt == -1) {
			if (errno == EPIPE) { /* TODO - see what err we get.. */
				VVError = VVE_CANCELED;
				nLogMsg(EL_CODE|EL_ERROR2, id, 
					"connection terminated by %d", 
					sockp->sock_id);
			} else { 
				nLogMsg(EL_CODE|EL_ERROR1|EL_SYS, id, 
					"write failed to %d", sockp->sock_id);
			}
		} else {
			nLogMsg(EL_CODE|EL_DEBUG1, id,
			  		"write failed %d/%d bytes to %d",
			  		wcnt, size, sockp->sock_id);
		}
	}
	return(wcnt);
}

/*
**	XDRSetup() - hook xdrp to client & read/write routines
*/
  void
XDRSetup(Sock_t *sockp, XDR *xdrp)
{
	sockp->sock_xdrp = xdrp;
	xdrrec_create(xdrp, 0, 0, (caddr_t) sockp, 
				(int (*) (char *, char *, int) )XDRread, 
				(int (*) (char *, char *, int) )XDRwrite);
	if (!xdrrec_skiprecord(xdrp))
		nLogMsg(EL_STATUS1, 0, "init xdr_skiprecord failed");
}


  int
XDRSend(vvpHdr_t *hp, XDR *xdrp, int now)
{
	char	*id = "XDRSend";
	char	*cp;

/*
if (hp->vh_type != VVM_TIMER  &&  hp->vh_type != VVM_BYTES
	&&  hp->vh_type != VVM_VSNQRY  &&  hp->vh_type != VVM_VSNINFO
	&&  hp->vh_type != VVM_VOLQRY) 
nLogMsg(EL_STATUS1, id, "key 0x%x %s", hp->vh_key, vvHdrStr(hp));
*/
	nLogMsg(EL_DEBUG2, id, "key 0x%x %s", hp->vh_key, vvHdrStr(hp));
	if (hp->vh_type < VVM_MINTYPE  ||  hp->vh_type > VVM_MAXTYPE) {
		nLogMsg(EL_ERROR1|EL_CONSOLE0, id, "bad VVM_ msg type %d", 
							hp->vh_type);
		return(1);
	}

	/*
	**	set XDR to write
	*/
	xdrp->x_op = XDR_ENCODE;

	/*
	**	packet header
	*/
	if (!xdr_vvpHdr_t(xdrp, hp)) {
		static vvId_t	key;

		if (key != hp->vh_key) {
			nLogMsg(EL_ERROR1, id, "pkt hdr send failed %s key %d",
				vvHdrStr(hp), hp->vh_key);
			key = hp->vh_key;
		}
		return(1);
	}

	if (hp->vh_size == 0)
		goto endrec;

	/*
	**	packet body
	*/
	cp = VV_MSGPTR(hp, char);
	switch (hp->vh_type) {
	    case VVM_TIMER:
		if (!xdr_vvpTimer_t(xdrp, (vvpTimer_t *)cp))
			return(1);
		break;
	    case VVM_CLIENT:
	    case VVM_MOVER:
	    case VVM_TIMERD:
		if (!xdr_vvpHello_t(xdrp, (vvpHello_t *)cp)) 
			return(1);
		break;
	    case VVM_OWRITE:
		if (!xdr_vvpOpenwrite_t(xdrp, (vvpOpenwrite_t *)cp))
			return(1);
		if (!xdr_VVInfo_t(xdrp, (VVInfo_t *)cp))
			return(1);
		break;
	    case VVM_VOLINFO:
		if (!xdr_VVInfo_t(xdrp, (VVInfo_t *)cp))
			return(1);
		break;
	    case VVM_WTAPE:
	    case VVM_RTAPE:
		if (!xdr_vvpMvreq_t(xdrp, (vvpMvreq_t *)cp))
			return(1);
		break;
	    case VVM_BYTES:
		if (!xdr_bytes(xdrp, &cp, &hp->vh_size, hp->vh_size))
			return(1);
		break;
	    case VVM_VVERROR:
	    case VVM_SHUTDOWN:
		if (!xdr_vvpError_t(xdrp, (vvpError_t *)cp)) 
			return(1);
		break;
	    case VVM_VRECYCLE:
	    case VVM_OREAD:
	    case VVM_CWRITE:
	    case VVM_FINISH:
		if (!xdr_vvpVol_t(xdrp, (vvpVol_t *)cp))
			return(1);
		break;
	    case VVM_VOLQRY:
		if (!xdr_vvpVolqry_t(xdrp, (vvpVolqry_t *)cp))
			return(1);
		break;
	    case VVM_VSNQRY:
	    case VVM_RECYCVSN:
	    case VVM_CHECKVSN:
	    case VVM_FINISHVSN:
	    case VVM_BADVSN:
		if (!xdr_vvpVsn_t(xdrp, (vvpVsn_t *)cp)) 
			return(1);
		break;
	    case VVM_VSNINFO:
		if (!xdr_Vsn_t(xdrp, (Vsn_t *)cp))
			return(1);
		break;
	    case VVM_COPYIX:
		if (!xdr_vvpIx_t(xdrp, (vvpIx_t *)cp))
			return(1);
		break;
	    case VVM_FINCLASS:
		if (!xdr_vvpFinish_t(xdrp, (vvpFinish_t *)cp))
			return(1);
		break;
	    case VVM_DEBUG:
		if (!xdr_vvpDebug_t(xdrp, (vvpDebug_t *)cp)) 
			return(1);
		break;
	    default:
		nLogMsg(EL_FATAL, id, "unexpected msg type: %s",
						vvHdrStr(hp));
	}
endrec:
	/*
	 *  end record, flush buffer if req
	 */
	if (!xdrrec_endofrecord(xdrp, now)) {
		nLogMsg(EL_ERROR1, 0, "xdrrec_endofrecord failed");
		return(1);
	}
	return(0);
}

  int
XDRNextRecord(XDR *xdrp)
{
	int	more;

	nLogMsg(EL_DEBUG2, 0, "XDRNextRecord");
	/*
	 *  set XDR to read to be on the safe side..
	 */
	xdrp->x_op = XDR_DECODE;

	/*
	 *  finish this record & see if another record follows
	 */
	if ((more = !xdrrec_eof(xdrp)) != 0) {
		nLogMsg(EL_DEBUG2, 0, "more input!");
	}

	/*
	 *  move to end of record boundary
	 */
	if (!xdrrec_skiprecord(xdrp))
		nLogMsg(EL_STATUS1, 0, "xdr_skiprecord failed");

	return(more);
}

/*
 *  XDRRecv() - vvmd / client, mover communication
 *
 *  	NOTE: all pointers w/in hp need to be recalced after FULLPKT()
 *	since nReAlloc occurs. This REALLY CAN BITE YOU when dealing
 *	w/ complex pkts that require multiple FULLPKT()s.
 */
#define FULLPKT(hp, size)  \
	hp = (vvpHdr_t *) nReAlloc((void *) hp, hsize + size); \
	cp = VV_MSGPTR(hp, char);

 vvpHdr_t *
XDRRecv(XDR *xdrp)
{
	char		*id = "XDRRecv";
	vvpHdr_t	*hp;
	char		*cp;
	size_t		hsize = PadSize(sizeof(vvpHdr_t));

	/*
	**	set XDR to read
	*/
	xdrp->x_op = XDR_DECODE;

	/*
	**	packet header
	*/
	hp = (vvpHdr_t *) nAlloc(hsize);

	if (!xdr_vvpHdr_t(xdrp, hp)) {
		nFree(hp);
		return(NULL);
	}

	if (hp->vh_type < VVM_MINTYPE  ||  hp->vh_type > VVM_MAXTYPE) {
		nLogMsg(EL_ERROR1|EL_CONSOLE0, id, "bad VVM_ msg type %d", 
							hp->vh_type);
		nFree(hp);
		return(NULL);
	}
/*
if (hp->vh_type != VVM_TIMER  &&  hp->vh_type != VVM_BYTES
	&&  hp->vh_type != VVM_VSNQRY  &&  hp->vh_type != VVM_VSNINFO
	&&  hp->vh_type != VVM_VOLQRY) 
nLogMsg(EL_STATUS1, id, "key 0x%x %s", hp->vh_key, vvHdrStr(hp));
*/

	nLogMsg(EL_DEBUG2, id, "key 0x%x %s", hp->vh_key, vvHdrStr(hp));

	if (hp->vh_size == 0)
		return(hp);

	/*
	**	packet body, ordered approx by frequency
	*/
	switch (hp->vh_type) {
	    case VVM_TIMER:
		FULLPKT(hp, sizeof(vvpTimer_t));
		if (!xdr_vvpTimer_t(xdrp, (vvpTimer_t *)cp)) {
			nLogMsg(EL_ERROR1, id, "xdr_vvpTimer_t");
			return(NULL);
		}
		break;
	    case VVM_CLIENT:
	    case VVM_MOVER:
	    case VVM_TIMERD:
		hp->vh_size = sizeof(vvpHello_t);
		FULLPKT(hp, sizeof(vvpHello_t));
		if (!xdr_vvpHello_t(xdrp, (vvpHello_t *)cp)) {
			nLogMsg(EL_ERROR1, id, "xdr_vvpHello_t");
			return(NULL);
		}
		break;
	    case VVM_OWRITE:
		hp->vh_size = sizeof(vvpOpenwrite_t);
		FULLPKT(hp, sizeof(vvpOpenwrite_t));
		if (!xdr_vvpOpenwrite_t(xdrp, (vvpOpenwrite_t *)cp)) {
			nLogMsg(EL_ERROR1, id, "xdr_vvpOpenwrite_t");
			return(NULL);
		}
		break;
	    case VVM_VRECYCLE:
	    case VVM_OREAD:
	    case VVM_CWRITE:
	    case VVM_FINISH:
		hp->vh_size = sizeof(vvpVol_t);
		FULLPKT(hp, sizeof(vvpVol_t));
		if (!xdr_vvpVol_t(xdrp, (vvpVol_t *)cp)) {
			nLogMsg(EL_ERROR1, id, "xdr_vvpVol_t");
			return(NULL);
		}
		break;
	    case VVM_WTAPE:
	    case VVM_RTAPE:
		hp->vh_size = sizeof(vvpMvreq_t);
		FULLPKT(hp, sizeof(vvpMvreq_t));
		if (!xdr_vvpMvreq_t(xdrp, (vvpMvreq_t *)cp)) {
			nLogMsg(EL_ERROR1, id, "vvpMvreq_t");
			return(NULL);
		}
		break;
	    case VVM_VOLQRY:
		hp->vh_size = sizeof(vvpVolqry_t);
		FULLPKT(hp, sizeof(vvpVolqry_t));
		if (!xdr_vvpVolqry_t(xdrp, (vvpVolqry_t *)cp)) {
			nLogMsg(EL_ERROR1, id, "vvpVolqry_t");
			return(NULL);
		}
		break;
	    case VVM_VSNQRY:
	    case VVM_RECYCVSN:
	    case VVM_CHECKVSN:
	    case VVM_FINISHVSN:
	    case VVM_BADVSN:
		hp->vh_size = sizeof(vvpVsn_t);
		FULLPKT(hp, sizeof(vvpVsn_t));
		if (!xdr_vvpVsn_t(xdrp, (vvpVsn_t *)cp)) 
			return(NULL);
		break;
	    case VVM_VSNINFO:
		hp->vh_size = sizeof(Vsn_t);
		FULLPKT(hp, sizeof(Vsn_t));
		if (!xdr_Vsn_t(xdrp, (Vsn_t *)cp)) 
			return(NULL);
		break;
	    case VVM_COPYIX:
		hp->vh_size = sizeof(vvpIx_t);
		FULLPKT(hp, sizeof(vvpIx_t));
		if (!xdr_vvpIx_t(xdrp, (vvpIx_t *)cp))
			return(NULL);
		break;
	    case VVM_FINCLASS:
		hp->vh_size = sizeof(vvpFinish_t);
		FULLPKT(hp, sizeof(vvpFinish_t));
		if (!xdr_vvpFinish_t(xdrp, (vvpFinish_t *)cp))
			return(NULL);
		break;
	    case VVM_VOLINFO:
		hp->vh_size = sizeof(VVInfo_t);
		FULLPKT(hp, sizeof(VVInfo_t));
		if (!xdr_VVInfo_t(xdrp, (VVInfo_t *)cp)) {
			nLogMsg(EL_ERROR1, id, "xdr_VVInfo_t");
			return(NULL);
		}
		break;
	    case VVM_BYTES:
		/* bytes are raw */
		FULLPKT(hp, hp->vh_size);
		if (!xdr_bytes(xdrp, &cp, &hp->vh_size, hp->vh_size)) {
			nLogMsg(EL_ERROR1, id, "xdr_bytes");
			return(NULL);
		}
		break;
	    case VVM_VVERROR:
	    case VVM_SHUTDOWN:
		hp->vh_size = sizeof(vvpError_t);
		FULLPKT(hp, sizeof(vvpError_t));
		if (!xdr_vvpError_t(xdrp, (vvpError_t *)cp)) {
			nLogMsg(EL_ERROR1, id, "xdr_vvpError_t");
			return(NULL);
		}
		break;
	    case VVM_DEBUG:
		hp->vh_size = sizeof(vvpDebug_t);
		FULLPKT(hp, sizeof(vvpDebug_t));
		if (!xdr_vvpDebug_t(xdrp, (vvpDebug_t *)cp)) {
			nLogMsg(EL_ERROR1, id, "xdr_vvpDebug_t");
			return(NULL);
		}
		break;
	    default:
		VVError = VVE_COMM;
		nLogMsg(EL_ERROR1, id, "unexpected type %s", vvHdrStr(hp));
		return(NULL);
	}
	return(hp);
}

/*
 *  XDRFlush() - when connection goes away & socket is closed,
 *	make sure nothing remains in the buffer that would get
 *	flushed to the next client on this socket
 */
  void
XDRFlush(XDR *xdrp)
{
	(void)xdrrec_endofrecord(xdrp, 1);
}
