/*
 *  vvm_class.c - NAStore Virtual Volume Manager mgt of storage
 *	classes, i.e. mapping vv's onto different categories of
 *	tape. This is where Volume Manager requests to mount tapes
 *	for writing and Mover requests to write the tapes are generated.
 */

#include "vvmd.h"
#include "cl_space.h"

extern int	volman_sock;

int		ClassMax = 0;
VVMClass_t	*ClassArray = NULL;

#define VV_PRINT	24	/* for formatting output */

/*
 *  InitClasses() - set up vvmd's vv class mgt (basically
 *	the save-to-tape aspect of vvmd)
 */
  void
InitClasses()
{
	char		*id = "InitClasses";
	int		class, slot;
	Class_t		*libcip;

	/*
	**	make sure the config in lib/vv_classes.c is ok
	*/
	ClassMax = CheckClasses();

	/*
	**	allocate/fill-in vvmd's super-version
	**	of the class array
	*/
	ClassArray = (VVMClass_t *) nAlloc(ClassMax * sizeof(VVMClass_t));
	memset(ClassArray, 0, ClassMax * sizeof(VVMClass_t));

	for (class=0; libcip=ClassIndex(class); class++) {
		/*
		**	copy lib/vv_classes.c config
		*/
		ClassArray[class].vc_class = *libcip;

		/*
		**	initialize/null the queues of vv<->tape requests
		*/
		ClassArray[class].vc_writeq.q_head = NULL;
		ClassArray[class].vc_writeq.q_tail = NULL;
		ClassArray[class].vc_readq.q_head = NULL;
		ClassArray[class].vc_readq.q_tail = NULL;
		ClassArray[class].vc_recycq.q_head = NULL;
		ClassArray[class].vc_recycq.q_tail = NULL;

		/*
		**	per-slot initialization
		*/
		for (slot=0; slot<libcip->cs_nvsns; slot++) {
			VolType_t	*vtp;

			/*
			**	check volman types, init vsn properties
			*/
			vtp = vVolTypeByID(libcip->cs_vsns[slot].vs_voltype);
			if (vtp == NULL)
				nLogMsg(EL_FATAL, id, "unknown voltype: %s",
					libcip->cs_vsns[slot].vs_voltype);

			ClassArray[class].vc_idlemount[slot] = vtp->vt_weartime;
		}
		for (slot=0; slot<MAXVSN; slot++) {
			/*
			**	0 various slot counts/queues
			*/
			ClassArray[class].vc_vsloterrs[slot] = 0; 
			ClassArray[class].vc_msloterrs[slot] = 0; 
			ClassArray[class].vc_allocwait[slot] = 0;
			ClassArray[class].vc_waitct[slot] = 0;
			ClassArray[class].vc_activect[slot] = 0;
			ClassArray[class].vc_slotq[slot].q_head = NULL;
			ClassArray[class].vc_slotq[slot].q_tail = NULL;
		}

		/* NOTE: vc_lastnum is filled in on dbase init */

		/* NOTE: vc_hotpanel is filled in after VVCONF is read */
		ClassArray[class].vc_hotpanel = NULL;
		ClassArray[class].vc_npanels = 1;	/* default */
	}
}


/*
 *  VVMClassInfo() - look up the class by its prefix;
 *	can use null-terminated or unterminated string 
 *	for class
 */
  VVMClass_t *
VVMClassInfo(char *prefix)
{
	char		*id = "VVMClassInfo";
	VVMClass_t	*cip = ClassArray;
	char		*cp;
	int		i;

	for (i=0; i<VV_CLASSLEN; i++)
		if (prefix[i] == '\0')
			break;
		
	for (i=0; i<ClassMax; i++, cip++) {
		if (!strncmp(prefix, cip->vc_class.cs_prefix, i))
			return(cip);
	}
	VVError = VVE_INTERNAL;
	nLogMsg(EL_ERROR1|EL_CODE, id, "unknown class %s", prefix);
	return(NULL);
}

/*
 *  VVMClassInfoByName() - use name to look up; if 1 char, try prefix too
 */
  VVMClass_t *
VVMClassInfoByName(char *name)
{
	char		*id = "VVMClassInfoByName";
	VVMClass_t	*cip = ClassArray;
	int		i;

	for (i=0; i<ClassMax; i++, cip++) {
		if (!strcmp(cip->vc_class.cs_name, name))
			return(cip);
	}
	VVError = VVE_INTERNAL;
	nLogMsg(EL_ERROR1|EL_CODE, id, "unknown class %s", name);
	return(NULL);
}


/*
 *  VVMClassIndex() - a way of indexing the ClassArray
 *	w/out knowing about ClassMax
 */
  VVMClass_t *
VVMClassIndex(int i)
{
	if (i < 0  ||  i >= ClassMax)
		return(NULL);
	return(&ClassArray[i]);
}


/*
**	LogClasses() - log info on classes: hot panels, write/read reqs
*/
  void
LogClasses()
{
	char		*id = "LogClasses";
	VVMClass_t	*cip = ClassArray;
	int		i, panel, slot;
	int		alloc = 0, mount = 0, len;
	char		*cp, hbuf[VV_PRINT];
	char		vsnbuf1[MAXVSN*VSNLEN*2+6];
	char		vsnbuf2[MAXVSN*VSNLEN*2+6];
	char		vsnbuf3[MAXVSN*VSNLEN*2+6];

	nLogMsg(EL_STATUS1, id, "----------------------------------------");
	for (i=0; i<ClassMax; i++, cip++) {
		int		npanels = cip->vc_npanels;
		int		nvsns = cip->vc_class.cs_nvsns;
		HotPanel_t	*hp = cip->vc_hotpanel;
		WriteReq_t	*wrp;
		ReadReq_t	*rp;

		memset(vsnbuf1, ' ', sizeof(vsnbuf1));
		memset(vsnbuf2, ' ', sizeof(vsnbuf2));
		for (slot=0; slot<nvsns; slot++) {
			char	buf[INTLEN+1];
			size_t	len;

			(void)sprintf(buf, "slot%d", slot);
			len = strlen(buf);
			memcpy(vsnbuf1+2*slot*VSNLEN, buf, len);
			(void)sprintf(buf, "%d/%d", cip->vc_vsloterrs[slot], 
						      cip->vc_msloterrs[slot]);
			len = strlen(buf);
			memcpy(vsnbuf2+2*slot*VSNLEN, buf, len);
		}
		vsnbuf1[2*VSNLEN*nvsns+3] = '|';
		vsnbuf1[2*VSNLEN*nvsns+4] = '\0';
		vsnbuf2[2*VSNLEN*nvsns+3] = '|';
		vsnbuf2[2*VSNLEN*nvsns+4] = '\0';
		nLogMsg(EL_STATUS1|EL_NOTIME, 0, 
		    "=============> class %s: panels %d slots %d <=============",
				cip->vc_class.cs_prefix, npanels, nvsns);

		memset(hbuf, ' ', sizeof(hbuf)-1);
		hbuf[sizeof(hbuf)-1] = '\0';
		nLogMsg(EL_STATUS1|EL_NOTIME, 0, "%s%s", hbuf, vsnbuf1);
		(void)strcpy(hbuf, "errs (volman/mover):");
		len = strlen(hbuf);
		cp = hbuf + len;
		memset(cp, ' ', sizeof(hbuf)-1-len);
		nLogMsg(EL_STATUS1|EL_NOTIME, 0, "%s%s", hbuf, vsnbuf2);
		nLogMsg(EL_STATUS1|EL_NOTIME, 0, "");
		
		for (panel=0; panel<npanels; panel++, hp++) {
			HotVsn_t	*hvp = hp->hp_vsns;

			memset(vsnbuf1, ' ', sizeof(vsnbuf1));
			memset(vsnbuf2, ' ', sizeof(vsnbuf2));
			memset(vsnbuf3, ' ', sizeof(vsnbuf3));
			if (volman_sock == -1) {
			    memcpy(vsnbuf1, "--- volman down ---", (size_t)19);
			} else {
			    for (slot=0; slot<nvsns; slot++, hvp++) {
				Vsn_t		*vp = &hvp->hv_vsnrec;

				wrp = hvp->hv_writerp;

				if (isalnum(vp->vs_vsn[0]))
					memcpy(vsnbuf1+(2*slot*VSNLEN), 
						vp->vs_vsn, (size_t)VSNLEN);
				else if (hvp->hv_error)
					memcpy(vsnbuf1+(2*slot*VSNLEN),
							"BAD CLASS", (size_t)9);
				if (hvp->hv_vrp != NULL) {
					char	buf[INTLEN+1];
					int	len;
					(void)strcpy(buf, 
					  vvVsnStatus(hvp->hv_vrp->vr_status));
					len = strlen(buf);
					memcpy(vsnbuf2+2*slot*VSNLEN, buf, len);
				} else if (hvp->hv_status == HV_VOLALLOC)
					memcpy(vsnbuf2+2*slot*VSNLEN,
								"alloc", 
								(size_t)5);
				else if (hvp->hv_status == HV_NOVSN)
					memcpy(vsnbuf2+2*slot*VSNLEN,
								"NO_SCRATCH", 
								(size_t)10);
				else
					memcpy(vsnbuf2+2*slot*VSNLEN,
							"no_mount", (size_t)8);
				if (wrp != NULL) {
					char	*name;
					size_t	len2;

					name = VV2Name(&wrp->wr_vv.v_vv);
					len2 = strlen(name);
					memcpy(vsnbuf3+(2*slot*VSNLEN), 
						name, len2);
				}
			    }
			}
			vsnbuf1[2*VSNLEN*nvsns+3] = '|';
			vsnbuf1[2*VSNLEN*nvsns+4] = '\0';
			vsnbuf2[2*VSNLEN*nvsns+3] = '|';
			vsnbuf2[2*VSNLEN*nvsns+4] = '\0';
			vsnbuf3[2*VSNLEN*nvsns+3] = '|';
			vsnbuf3[2*VSNLEN*nvsns+4] = '\0';

			nLogMsg(EL_STATUS1|EL_NOTIME, 0, "-- hot panel %d", 
								panel);
			(void)strcpy(hbuf, "vsns:");
			len = strlen(hbuf);
			cp = hbuf + len;
			memset(cp, ' ', sizeof(hbuf)-1-len);
			nLogMsg(EL_STATUS1|EL_NOTIME, 0, "%s%s", hbuf, vsnbuf1);

			(void)strcpy(hbuf, "stat:");
			len = strlen(hbuf);
			cp = hbuf + len;
			memset(cp, ' ', sizeof(hbuf)-1-len);
			nLogMsg(EL_STATUS1|EL_NOTIME, 0, "%s%s", hbuf, vsnbuf2);

			(void)strcpy(hbuf, "active vv:");
			len = strlen(hbuf);
			cp = hbuf + len;
			memset(cp, ' ', sizeof(hbuf)-1-len);
			nLogMsg(EL_STATUS1|EL_NOTIME, 0, "%s%s", hbuf, vsnbuf3);
		}
		nLogMsg(EL_STATUS1|EL_NOTIME, 0, "");
		nLogMsg(EL_STATUS1|EL_NOTIME, 0, "--------- write requests");
		if (CWRITEHEAD(cip) == NULL)
			nLogMsg(EL_STATUS1|EL_NOTIME, 0, 
						 "              (none)");
		for (wrp=CWRITEHEAD(cip); wrp; wrp=CWRITENEXT(wrp)) {
			char	statbuf[MAXVSN*2*VSNLEN+6];

			(void)sprintf(hbuf, "%s/%s", VV2Name(&wrp->wr_vv.v_vv),
						IdString(wrp->wr_vv.v_id));
			len = strlen(hbuf);
			cp = hbuf + len;
			memset(cp, ' ', sizeof(hbuf)-1-len);

			cp = statbuf;
			for (slot=0; slot<nvsns; slot++) {
				switch (wrp->wr_statpanel[slot]) {
				    case WR_STWAIT:
					(void)strcpy(cp, "wait"); break;
				    case WR_STDONE:
					(void)strcpy(cp, "done"); break;
				    case WR_STFAILED:
					(void)strcpy(cp, "fail"); break;
				    default:
					(void)sprintf(cp, "panel %d", 
						wrp->wr_statpanel[slot]);
				}
				len = strlen(cp);
				cp += len;
				len = 2 * VSNLEN - len;
				memset(cp, ' ', len);
				cp += len;
			}
			(void)strcpy(cp, "   |");
			nLogMsg(EL_STATUS1|EL_NOTIME, 0, "%s%s", hbuf, statbuf);
break;
		}
		nLogMsg(EL_STATUS1|EL_NOTIME, 0, "");
		nLogMsg(EL_STATUS1|EL_NOTIME, 0, "--------- read requests");
		if (CREADHEAD(cip) == NULL)
			nLogMsg(EL_STATUS1|EL_NOTIME, 0, "             (none)");
		for (rp=CREADHEAD(cip); rp; rp=CREADNEXT(rp)) {
			char	statbuf[MAXVSN*2*VSNLEN+6];

			(void)sprintf(hbuf, "%s/%s", VV2Name(&rp->r_vv.v_vv),
						IdString(rp->r_vv.v_id));
			len = strlen(hbuf);
			cp = hbuf + len;
			memset(cp, ' ', sizeof(hbuf)-1-len);

			cp = statbuf;
			(void)sprintf(cp, "%.*s", nvsns * VSNLEN, rp->r_panel);
			nLogMsg(EL_STATUS1|EL_NOTIME, 0, "%s%s", hbuf, statbuf);
			if (rp->r_vrp == NULL) {
				nLogMsg(EL_STATUS1|EL_NOTIME, 0, 
						"           (no mount)");
			} else
				nLogMsg(EL_STATUS1|EL_NOTIME, 0, 
					"           %s",
					vvVsnStatus(rp->r_vrp->vr_status));
		}
		nLogMsg(EL_STATUS1|EL_NOTIME, 0, "");
	}
	nLogMsg(EL_STATUS1, id, "----------------------------------------");
}		

/*
**	MLogClasses() - send info on classes: hot panels, write/read reqs
*/
  void
MLogClasses(int cl_slot, unsigned int writes)
{
	char		*id = "MLogClasses";
	VVMClass_t	*cip = ClassArray;
	int		i, panel, slot;
	int		alloc = 0, mount = 0;
	size_t		len;
	char		*cp, msgbuf[200];
	char		vsnbuf1[MAXVSN*VSNLEN*2+6];
	char		vsnbuf2[MAXVSN*VSNLEN*2+6];
	char		vsnbuf3[MAXVSN*VSNLEN*2+6];
	char		vsnbuf4[MAXVSN*VSNLEN*2+6];

	if (SendBuf(cl_slot, "----------------------------------------"))
		return;
	for (i=0; i<ClassMax; i++, cip++) {
		int		npanels = cip->vc_npanels;
		int		nvsns = cip->vc_class.cs_nvsns;
		HotPanel_t	*hp = cip->vc_hotpanel;
		WriteReq_t	*wrp;
		ReadReq_t	*rp;

		memset(vsnbuf1, ' ', sizeof(vsnbuf1));
		memset(vsnbuf2, ' ', sizeof(vsnbuf2));
		memset(vsnbuf3, ' ', sizeof(vsnbuf3));
		memset(vsnbuf4, ' ', sizeof(vsnbuf4));
		for (slot=0; slot<nvsns; slot++) {
			char	buf[INTLEN+1];

			/* hdrs */
			(void)sprintf(buf, "slot%d", slot);
			len = strlen(buf);
			memcpy(vsnbuf1+2*slot*VSNLEN, buf, len);

			/*  defns */
			len = strlen(cip->vc_class.cs_vsns[slot].vs_rc);
			memcpy(vsnbuf2+2*slot*VSNLEN, 
					cip->vc_class.cs_vsns[slot].vs_rc, len);

			len = strlen(cip->vc_class.cs_vsns[slot].vs_voltype);
			memcpy(vsnbuf3+2*slot*VSNLEN, 
					cip->vc_class.cs_vsns[slot].vs_voltype,
					len);

			(void)sprintf(buf, "%d/%d", cip->vc_vsloterrs[slot], 
					      cip->vc_msloterrs[slot]);

			/* errs */
			len = strlen(buf);
			memcpy(vsnbuf4+2*slot*VSNLEN, buf, len);
		}
		vsnbuf1[2*VSNLEN*nvsns+3] = '|';
		vsnbuf1[2*VSNLEN*nvsns+4] = '\n';
		vsnbuf1[2*VSNLEN*nvsns+5] = '\0';
		vsnbuf2[2*VSNLEN*nvsns+3] = '|';
		vsnbuf2[2*VSNLEN*nvsns+4] = '\0';
		vsnbuf3[2*VSNLEN*nvsns+3] = '|';
		vsnbuf3[2*VSNLEN*nvsns+4] = '\n';
		vsnbuf3[2*VSNLEN*nvsns+5] = '\0';
		vsnbuf4[2*VSNLEN*nvsns+3] = '|';
		vsnbuf4[2*VSNLEN*nvsns+4] = '\0';

		(void)sprintf(msgbuf, 
		   "=============> class %s: panels %d slots %d <=============",
				cip->vc_class.cs_prefix, npanels, nvsns);
		if (SendBuf(cl_slot, msgbuf))
			return;

		/* hdrs */

		memset(msgbuf, ' ', (size_t)VV_PRINT);
		cp = msgbuf + VV_PRINT;

		(void)strcpy(cp, vsnbuf1);
		if (SendBuf(cl_slot, msgbuf))
			return;

		/* defns */

		(void)strcpy(cp, vsnbuf2);
		if (SendBuf(cl_slot, msgbuf))
			return;

		(void)strcpy(cp, vsnbuf3);
		if (SendBuf(cl_slot, msgbuf))
			return;

		/* errs */

		(void)strcpy(msgbuf, "errs (volman/mover):");
		len = strlen(msgbuf);
		cp = msgbuf + len;
		memset(cp, ' ', (size_t)VV_PRINT-len);
		cp = msgbuf + VV_PRINT;
		(void)strcpy(cp, vsnbuf4);
		if (SendBuf(cl_slot, msgbuf))
			return;

		if (SendBuf(cl_slot, ""))
			return;
		
		for (panel=0; panel<npanels; panel++, hp++) {
			HotVsn_t	*hvp = hp->hp_vsns;
			int		mvstatus = 0;

			memset(vsnbuf1, ' ', sizeof(vsnbuf1));
			memset(vsnbuf2, ' ', sizeof(vsnbuf2));
			memset(vsnbuf3, ' ', sizeof(vsnbuf3));
			memset(vsnbuf4, ' ', sizeof(vsnbuf4));

			if (volman_sock == -1) {
			    memcpy(vsnbuf1, "--- volman down ---", (size_t)19);
			} else {
			    for (slot=0; slot<nvsns; slot++, hvp++) {
				Vsn_t		*vp = &hvp->hv_vsnrec;

				wrp = hvp->hv_writerp;

nLogMsg(EL_STATUS1, id, "hvp 0x%x panel %d slot %d wrp 0x%x",
hvp, panel, slot, wrp); 
				if (isalnum(vp->vs_vsn[0]))
					memcpy(vsnbuf1+(2*slot*VSNLEN), 
						vp->vs_vsn, (size_t)VSNLEN);
				else if (hvp->hv_error)
					memcpy(vsnbuf1+(2*slot*VSNLEN),
							"BAD CLASS", (size_t)9);
				if (hvp->hv_vrp != NULL) {
					char	buf[INTLEN+1];
					int	len;
					(void)strcpy(buf, 
					  vvVsnStatus(hvp->hv_vrp->vr_status));
					len = strlen(buf);
					memcpy(vsnbuf2+2*slot*VSNLEN, buf, len);
				} else if (hvp->hv_status == HV_VOLALLOC)
					memcpy(vsnbuf2+2*slot*VSNLEN,
							"alloc", (size_t)5);
				else if (hvp->hv_status == HV_NOVSN)
					memcpy(vsnbuf2+2*slot*VSNLEN,
						"NO_SCRATCH", (size_t)10);
				else
					memcpy(vsnbuf2+2*slot*VSNLEN,
							"no_mount", (size_t)8);
				if (wrp != NULL) {
					char	*name;
					int		len;
					name = VV2Name(&wrp->wr_vv.v_vv);
					len = strlen(name);
					memcpy(vsnbuf3+(2*slot*VSNLEN), 
								name, len);
				}

				if (hvp->hv_vrp != NULL  &&
				    hvp->hv_vrp->vr_status == VSQ_MOUNTED  &&
				    HostMover(hvp->hv_vrp->vr_hostp) == -1) {
					mvstatus++;
					memcpy(vsnbuf4+(2*slot*VSNLEN), 
							"NO MVR", (size_t)6);
				}
			    }
			}
			vsnbuf1[2*VSNLEN*nvsns+3] = '|';
			vsnbuf1[2*VSNLEN*nvsns+4] = '\0';
			vsnbuf2[2*VSNLEN*nvsns+3] = '|';
			vsnbuf2[2*VSNLEN*nvsns+4] = '\0';
			vsnbuf3[2*VSNLEN*nvsns+3] = '|';
			vsnbuf3[2*VSNLEN*nvsns+4] = '\0';
			vsnbuf4[2*VSNLEN*nvsns+3] = '|';
			vsnbuf4[2*VSNLEN*nvsns+4] = '\0';

			(void)sprintf(msgbuf, "-- hot panel %d\n", panel);
			if (SendBuf(cl_slot, msgbuf))
				return;
			
			(void)strcpy(msgbuf, "vsns:");
			len = strlen(msgbuf);
			cp = msgbuf + len;
			memset(cp, ' ', (size_t)VV_PRINT-len);
			cp = msgbuf + VV_PRINT;
			(void)strcpy(cp, vsnbuf1);
			if (SendBuf(cl_slot, msgbuf))
				return;

			(void)strcpy(msgbuf, "stat:");
			len = strlen(msgbuf);
			cp = msgbuf + len;
			memset(cp, ' ', (size_t)VV_PRINT-len);
			cp = msgbuf + VV_PRINT;
			(void)strcpy(cp, vsnbuf2);
			if (SendBuf(cl_slot, msgbuf))
				return;

			(void)strcpy(msgbuf, "active vv:");
			len = strlen(msgbuf);
			cp = msgbuf + len;
			memset(cp, ' ', (size_t)VV_PRINT-len);
			cp = msgbuf + VV_PRINT;
			(void)strcpy(cp, vsnbuf3);
			if (SendBuf(cl_slot, msgbuf))
				return;

			if (mvstatus) {
				memset(msgbuf, ' ', (size_t)VV_PRINT);
				cp = msgbuf + VV_PRINT;
				(void)strcpy(cp, vsnbuf4);
				if (SendBuf(cl_slot, msgbuf))
					return;
			}
			if (SendBuf(cl_slot, ""))
				return;
		}
		if (SendBuf(cl_slot, "--------- write requests\n"))
			return;
		if (CWRITEHEAD(cip) == NULL)
			if (SendBuf(cl_slot, "              (none)"))
				return;
		if (writes) {
			/*
			**	report all writes
			*/
			for (wrp=CWRITEHEAD(cip); wrp; wrp=CWRITENEXT(wrp)) {

				(void)sprintf(msgbuf, "%s/%s", 
						VV2Name(&wrp->wr_vv.v_vv),
						IdString(wrp->wr_vv.v_id));
				len = strlen(msgbuf);
				cp = msgbuf + len;
				memset(cp, ' ', (size_t)VV_PRINT-len);
				cp = msgbuf + VV_PRINT;

				for (slot=0; slot<nvsns; slot++) {
					switch (wrp->wr_statpanel[slot]) {
					    case WR_STWAIT:
						(void)strcpy(cp, "wait"); break;
					    case WR_STDONE:
						(void)strcpy(cp, "done"); break;
					    case WR_STFAILED:
						(void)strcpy(cp, "fail"); break;
					    default:
						(void)sprintf(cp, "panel %d", 
						    wrp->wr_statpanel[slot]);
					}
					len = strlen(cp);
					cp += len;
					len = 2 * VSNLEN - len;
					memset(cp, ' ', len);
					cp += len;
				}
				(void)strcpy(cp, "   |");
				if (SendBuf(cl_slot, msgbuf))
					return;
			}
		} else {
			int	wait[MAXVSN], done[MAXVSN]; 
			int	fail[MAXVSN], writing[MAXVSN];
			int	total = 0;

			for (slot=0; slot<nvsns; slot++) {
				wait[slot] = done[slot] = 0;
				fail[slot] = writing[slot] = 0;
			}
			for (wrp=CWRITEHEAD(cip); wrp; wrp=CWRITENEXT(wrp)) {
				total++;
				for (slot=0; slot<nvsns; slot++) {
					switch (wrp->wr_statpanel[slot]) {
					    case WR_STWAIT:
						wait[slot]++; break;
					    case WR_STDONE:
						done[slot]++; break;
					    case WR_STFAILED:
						fail[slot]++; break;
					    default:
						writing[slot]++; break;
					}
				}
			}
			if (total) {
				(void)sprintf(msgbuf, "total VV's:");
				len = strlen(msgbuf);
				cp = msgbuf + len;
				memset(cp, ' ', (size_t)VV_PRINT-len);
				cp = msgbuf + VV_PRINT;
				for (slot=0; slot<nvsns; slot++) {
					(void)sprintf(cp, "%d", total);
					len = strlen(cp);
					cp += len;
					len = 2 * VSNLEN - len;
					memset(cp, ' ', len);
					cp += len;
				}
				if (SendBuf(cl_slot, msgbuf))
					return;

				(void)sprintf(msgbuf, "wait:");
				len = strlen(msgbuf);
				cp = msgbuf + len;
				memset(cp, ' ', (size_t)VV_PRINT-len);
				cp = msgbuf + VV_PRINT;
				for (slot=0; slot<nvsns; slot++) {
					(void)sprintf(cp, "%d", wait[slot]);
					len = strlen(cp);
					cp += len;
					len = 2 * VSNLEN - len;
					memset(cp, ' ', len);
					cp += len;
				}
				if (SendBuf(cl_slot, msgbuf))
					return;

				(void)sprintf(msgbuf, "done:");
				len = strlen(msgbuf);
				cp = msgbuf + len;
				memset(cp, ' ', (size_t)VV_PRINT-len);
				cp = msgbuf + VV_PRINT;
				for (slot=0; slot<nvsns; slot++) {
					(void)sprintf(cp, "%d", done[slot]);
					len = strlen(cp);
					cp += len;
					len = 2 * VSNLEN - len;
					memset(cp, ' ', len);
					cp += len;
				}
				if (SendBuf(cl_slot, msgbuf))
					return;

				(void)sprintf(msgbuf, "failed:");
				len = strlen(msgbuf);
				cp = msgbuf + len;
				memset(cp, ' ', (size_t)VV_PRINT-len);
				cp = msgbuf + VV_PRINT;
				for (slot=0; slot<nvsns; slot++) {
					(void)sprintf(cp, "%d", fail[slot]);
					len = strlen(cp);
					cp += len;
					len = 2 * VSNLEN - len;
					memset(cp, ' ', len);
					cp += len;
				}
				if (SendBuf(cl_slot, msgbuf))
					return;

				(void)sprintf(msgbuf, "active:");
				len = strlen(msgbuf);
				cp = msgbuf + len;
				memset(cp, ' ', (size_t)VV_PRINT-len);
				cp = msgbuf + VV_PRINT;
				for (slot=0; slot<nvsns; slot++) {
					(void)sprintf(cp, "%d", writing[slot]);
					len = strlen(cp);
					cp += len;
					len = 2 * VSNLEN - len;
					memset(cp, ' ', len);
					cp += len;
				}
				if (SendBuf(cl_slot, msgbuf))
					return;
			}
		}
		if (SendBuf(cl_slot, ""))
			return;
		if (SendBuf(cl_slot, "--------- read requests\n"))
			return;
		if (CREADHEAD(cip) == NULL)
			if (SendBuf(cl_slot, "              (none)"))
				return;
		for (rp=CREADHEAD(cip); rp; rp=CREADNEXT(rp)) {

			(void)sprintf(msgbuf, "%s/%s", VV2Name(&rp->r_vv.v_vv),
						IdString(rp->r_vv.v_id));
			len = strlen(msgbuf);
			cp = msgbuf + len;
			memset(cp, ' ', (size_t)VV_PRINT-len);

			cp = msgbuf + VV_PRINT;
			(void)sprintf(cp, "%.*s", nvsns * VSNLEN, rp->r_panel);
			if (SendBuf(cl_slot, msgbuf))
				return;
			if (rp->r_vrp == NULL) {
				if (SendBuf(cl_slot, "           (no mount)"))
					return;
			} else {
				(void)sprintf(msgbuf, "           %s",
					  vvVsnStatus(rp->r_vrp->vr_status));
				if (SendBuf(cl_slot, msgbuf))
					return;
			}
		}
		if (SendBuf(cl_slot, ""))
			return;
	}
	if (SendBuf(cl_slot, "----------------------------------------"))
		return;
}		


