
/*
 *  vvm_clread.c - routines for restoring a VV from tape
 *
 *	the requests are queued on the VVMClass_t
 *	vsns are mounted via the 'select' cmd
 */

#include "vvmd.h"

int	read_idle = 0;		/* vsn mounts waiting to expire */

  static void
DelReadReq(ReadReq_t *rp)
{
	VVMClass_t	*cip = (VVMClass_t *) rp->r_cp;

	CREADDEQ(rp, cip);
	if (rp->r_vrp != NULL)
		DelVsnReq(rp->r_vrp);
	
	nFree(rp);
}

/*
**	StartVsnReadReq() - push a read request into the vsn stage
**		(i.e. set r_vrp)
*/
  static void
StartVsnReadReq(ReadReq_t *rp)
{
	char		*id = "StartVsnReadReq";
	VVMClass_t	*cip = (VVMClass_t *) rp->r_cp;
	int		nvsns = rp->r_vv.v_nvsns;

	/*
	**	see if an (idle or busy) mounted vsn is available
	*/
	if (rp->r_vrp = UseMountedVsn(nvsns, rp->r_panel2, rp, cip)) {
		if (rp->r_vrp->vr_status == VSQ_MOUNTED  ||
		    rp->r_vrp->vr_status == VSQ_IDLE)
			CRMoverRead(rp->r_vrp);
		return;
	}
	/*
	**	see if a vsn is a hot one waiting for an excuse
	**	to mount
	*/
	if (rp->r_vrp = VolmanMountHotVsn(nvsns, rp->r_panel2, rp, cip)) {
		return;
	}

	/*
	**	nothing mounted - send mount request
	*/
	rp->r_vrp = VolmanMount(nvsns, rp->r_panel2, VQT_READ,(void *)rp, cip);
}


/*
 *  HandoffMountedVsn() - give vsn to read or write, if waiting.
 *	If something waiting, hand off & dequeue/delete vrp;
 *	if nothing waiting, mark idle & leave for a bit.
 */
  static void
HandoffMountedVsn(VsnReq_t *vrp)
{
	char		*id = "HandoffMountedVsn";
	int		slot = vrp->vr_readslot;
	VsnReq_t	*vrp2 = (VsnReq_t *)vrp->vr_waitq.q_next;
	int		i;

	if (vrp2) {
		/*
		**	another read req is queued on vsn
		**	dequeue this req from this waitq
		*/
		char	*vsnp = vrp->vr_vsns + VSNLEN*slot;
		/* q_prev should be NULL */
		vrp2->vr_waitq.q_prev = vrp->vr_waitq.q_prev; 

		/*
		**	pass on mount characteristics to
		**	waiting vrp
		*/
		vrp2->vr_status = VSQ_MOUNTED;
		vrp2->vr_hostp = vrp->vr_hostp;
		vrp2->vr_bufsize = vrp->vr_bufsize;

		/*
		**	figure out which slot in q'd one
		**	(since slot defn's may have changed
		**	from 1 VV to the next)
		*/
if (vrp2->vr_nvsns < 1)
nLogMsg(EL_FATAL, id, "vrp2->vr_nvsns %d", vrp2->vr_nvsns);

		for (i=0; i<vrp2->vr_nvsns; i++) {
nLogMsg(EL_STATUS1, id, "compare %.6s", vrp2->vr_vsns + i*VSNLEN);
			if (!memcmp(vsnp, vrp2->vr_vsns + i*VSNLEN, 
							(size_t)VSNLEN))
				break;
		}
		if (i == vrp2->vr_nvsns)
			nLogMsg(EL_FATAL|EL_CONSOLE0, id,
					"vsn %.*s not found in q'd req",
					VSNLEN, vsnp);
		vrp2->vr_readslot = i;

		/*
		**	start mover req for the waiting read
		*/
		CRMoverRead(vrp2);

	} else if (vrp->vr_waitq.q_prev) {

		/*
		**	write / hot
		*/
		vrp2 = (VsnReq_t *) vrp->vr_waitq.q_prev;

		/*
		**	dequeue this req from prev's waitq
		*/
		vrp2->vr_waitq.q_next = NULL;

		/*
		**	mark it ready and give to any waiting write..
		*/
		vrp2->vr_status = VSQ_MOUNTED;
		CWHandoffHotMount(vrp2);

	} else {
		/*
		**	nothing waiting - mark idle & leave a bit
		**	just in case..
		*/
		read_idle++;
		vrp->vr_status = VSQ_IDLE;
		vrp->vr_time = Now;
		vrp->vr_caller = NULL;
		return;
	}

	/*
	 *  remove this req's refs for safety's sake & delete
	 */
	vrp->vr_waitq.q_prev = NULL;
	vrp->vr_waitq.q_next = NULL;
	DelVsnReq(vrp);
}

static void
SendErrs(ReadReq_t *rp)
{
	ReadReq_t	*rp2, *rp3;

	for (rp2=rp; rp2!=NULL; rp2=rp3) {
		rp3 = CWAITNEXT(rp2);

		if (rp2->r_clnt != -1)
			SendErr(rp2->r_seq, rp2->r_clnt);

		if (rp2 != rp)
			DelReadReq(rp2);
	}
}

  static void
RetryRestore(ReadReq_t *rp, int badslot)
{
	char		*id = "RetryRestore";
	VVMClass_t	*cip = (VVMClass_t *) rp->r_cp;
	VsnReq_t	*vrp2;

nLogMsg(EL_STATUS1, id, "nvsns %d", rp->r_nvsns);

	rp->r_nvsns--;
	if (!rp->r_nvsns) {
		/*
		**	nothing left to try; free reserved space
		*/
		DiskUnreserve(cip);
		VVError = VVE_UNAVAIL;
		nLogMsg(EL_CONSOLE0|EL_ERROR1|EL_CODE, 0, 
					"%s: RESTORE %sFAILED",
					VV2Name(&rp->r_vv.v_vv),
					(rp->r_clnt != -1 ? 
						"for recovery " : ""));
		SendErrs(rp);
		nLogMsg(EL_CONSOLE0|EL_STATUS1, 0, "%s: failed vsns: %.*s",
					VV2Name(&rp->r_vv.v_vv),
					rp->r_vv.v_nvsns*VSNLEN,
					rp->r_panel);
		DelReadReq(rp);
		return;
	}

	/*
	 *  wipe the offending vsn from the r_panel2 and retry
	 */
	memset(rp->r_panel2+badslot*VSNLEN, ' ', (size_t)VSNLEN);
	StartVsnReadReq(rp);
}

static ReadReq_t *
FindLastVVRestore(VVMClass_t *cip, VVInfo_t *vip)
{
	char 		*id = "FindLastVVRestore";
	ReadReq_t	*rp, *rplast = NULL;

	for (rp=CREADHEAD(cip); rp!=NULL; rp=CREADNEXT(rp)) {
		if (!memcmp(&vip->v_vv, &rp->r_vv.v_vv, sizeof(vip->v_vv))) {
			rplast = rp;
		}
	}
if (rplast && CWAITNEXT(rplast) != NULL)
nLogMsg(EL_FATAL, id, "next != null");
	return(rplast);
}

/*
 *  CRestore() - enqueue request, send mount if possible
 *	if clnt_sock==-1, it's a pre-restore to recover
 *	a lost vsn.
 */
  void
CRestore(vvSeq_t seq, int clnt_sock, VVInfo_t *vip)
{
	char		*id = "CRestore";
	VVMClass_t	*cip;
	ReadReq_t	*rp, *rp2;
	int		i, nvsns;

nLogMsg(EL_STATUS1, id, "%s", VV2Name(&vip->v_vv));
	cip = VVMClassInfo(vip->v_vv.class);
	if (cip == NULL)
		nLogMsg(EL_FATAL, id, "unknown class %s", 
					ClassString(vip->v_vv.class));

	/*
	**	build ReadReq_t, stuffing vsns VV has been 
	**		saved to into array
	*/
	rp = (ReadReq_t *) nAlloc(sizeof(ReadReq_t));
	rp->r_cp = (void *)cip;
	rp->r_clnt = clnt_sock;
	rp->r_seq = seq;
	rp->r_vv = *vip;
	rp->r_waitq.q_prev = NULL;
	rp->r_waitq.q_next = NULL;
	rp->r_vrp = NULL;

	/*
	**	see if restore already queued, if so wait on that
	*/
	rp2 = FindLastVVRestore(cip, vip);
	if (rp2 != NULL) {
nLogMsg(EL_STATUS1, id, "adding to prev VV");
		rp->r_waitq.q_prev = (void *)rp2;
		rp2->r_waitq.q_next = (void *)rp;
		CREADENQ(rp, cip);
		return;
	}

	/*
	**	r_panel is for reference; r_panel2 has failed vsns deleted
	*/
	rp->r_nvsns = 0;
	memset(rp->r_panel, ' ', sizeof(rp->r_panel));
	nvsns = vip->v_nvsns;
	for (i=0; i<nvsns; i++) {
		if (isalnum(vip->v_vsns[i].v_vsn[0])) {
			memcpy(rp->r_panel+i*VSNLEN, 
				vip->v_vsns[i].v_vsn, (size_t)VSNLEN);
			rp->r_nvsns++;
		}
	}
	memcpy(rp->r_panel2, rp->r_panel, sizeof(rp->r_panel2));

	/*
	**	check that there are valid vsn(s) to restore from
	*/
	if (rp->r_nvsns == 0) {
		VVError = VVE_INTERNAL;
		nLogMsg(EL_ERROR1|EL_CONSOLE0|EL_CODE, id, 
					"VV %s not on disk or on any vsns",
					VV2Name(&vip->v_vv));
		SendErr(seq, clnt_sock);
		nFree(rp);
		return;
	}

	/*
	**	make sure there's room for the VV on disk
	*/
	if (DiskReserve(cip)) {
		SendErr(seq, clnt_sock);
		nFree(rp);
		return;
	}

	CREADENQ(rp, cip);

	/*
	**	start vsn action with full r_panel2
	*/
	StartVsnReadReq(rp);
}

  void
CRChooseAlternateMount(VsnReq_t *vrp)
{
	ReadReq_t	*rp = (ReadReq_t *) vrp->vr_caller;
	int		badslot = vrp->vr_readslot;

nLogMsg(EL_STATUS1, "CRChooseAlternateMount",
"%s: badslot %d", VV2Name(&rp->r_vv.v_vv), badslot);

	/*
	 *  now that vital info is extracted from vrp, delete it
	 */
	DelVsnReq(vrp);

	/*
	 *  push the restore req down the road
	 */
	RetryRestore(rp, badslot);
}


/*
 *  CRFailQdReads() - requests waiting on another vsn req
 *	readslot is known or the multi-vsn readreq wouldn't
 *	be q'd.. 
 */
  void
CRFailQdReads(VsnReq_t *vrp)
{
	VsnReq_t	*vrp1, *vrp2;

	for (vrp1=vrp; vrp1; vrp1=vrp2) {
		vrp2 = (VsnReq_t *) vrp1->vr_waitq.q_next;

		CRChooseAlternateMount(vrp1);
	}
}

/*
**	CRMountFailed() - since it's a 'mount any vsn' request after
**		trying/failing any currently hot-mounted vsns, fail restore. 
*/
  void
CRMountFailed(VsnReq_t *vrp)
{
	char		*id = "CRMountFailed";
	ReadReq_t	*rp, *rp2;
	VVMClass_t	*cip;

	rp = (ReadReq_t *) vrp->vr_caller;
	if (rp == NULL)
		nLogMsg(EL_FATAL, id, "no caller");

	cip = (VVMClass_t *) rp->r_cp;

	VVError = VVE_UNAVAIL;
	nLogMsg(EL_CONSOLE0|EL_ERROR1|EL_CODE, 0, 
					"%s: RESTORE %sFAILED",
					VV2Name(&rp->r_vv.v_vv),
					(rp->r_clnt == -1 ? 
						"for recovery " : ""));
	/*
	**	send out error on this and any q'd on the VV
	*/
	SendErrs(rp);

	/*
	**	deal w/ the vsn aspect
	*/
	nLogMsg(EL_CONSOLE0|EL_STATUS1, 0, "%s: failed vsns: %.*s/%.*s",
					VV2Name(&rp->r_vv.v_vv),
					rp->r_vv.v_nvsns*VSNLEN,
					vrp->vr_vsns,
					rp->r_vv.v_nvsns*VSNLEN,
					rp->r_panel);
	/*
	**	in case reads get queued on mount/read
	*/
	CRFailQdReads((VsnReq_t *)vrp->vr_waitq.q_next);

	DelReadReq(rp);
}

/*
**	CRMountCanceled() - mount was taken away by outside agency
*/
  void
CRMountCanceled(VsnReq_t *vrp)
{
	char		*id = "CRMountCanceled";
	ReadReq_t	*rp = (ReadReq_t *) vrp->vr_caller;

	/*
	**	in case reads get queued on mount/read
	*/
	CRFailQdReads((VsnReq_t *)vrp->vr_waitq.q_next);

	if (rp != NULL) {
		/*
		**	let reader clean up when vvmvd responds -
		**		it may be able to finish the VV
		*/
		return;
	}
	DelVsnReq(vrp);
}

 void
CRMoverRead(VsnReq_t *vrp)
{
	char		*id = "CRMoverRead";
	ReadReq_t	*rp = (ReadReq_t *) vrp->vr_caller;
	int		slot = vrp->vr_readslot;
	char		*vsnp = vrp->vr_vsns + VSNLEN*slot;
	char		buf[VV_PKTSIZ(vvpMvreq_t)];
	vvSeq_t		seq = vvGetSeq();
	vvpHdr_t	*hp;
	vvpMvreq_t	*mvp;
	int		comslot;

nLogMsg(EL_STATUS1, id, "%.*s_%d", VSNLEN, vsnp, seq);
	/*
	**	make sure mover exists on host
	*/
	comslot = HostMover(vrp->vr_hostp);
	if (comslot == -1) {
nLogMsg(EL_STATUS1, id, "%.*s_%d - mover down (remount?)", VSNLEN, vsnp, seq);
		return;
	}

	/*
	**	build pkt for mover
	*/
	hp = vvBuildHdr(VVM_RTAPE, seq, (vvpHdr_t *)buf);
	mvp = VV_MSGPTR(hp, vvpMvreq_t);

	memcpy(mvp->m_vsn, vsnp, sizeof(mvp->m_vsn));
	mvp->m_checkvol1 = 0;	/* TODO - think! */

	/* vsn record info */
	mvp->m_block = rp->r_vv.v_vsns[slot].v_block;
	mvp->m_rec = rp->r_vv.v_vsns[slot].v_nrec;

	/* hdr info for this vv */
	mvp->m_vv = rp->r_vv.v_vv;
	memcpy(mvp->m_id, rp->r_vv.v_id, sizeof(mvp->m_id));
	memcpy(mvp->m_fsize, rp->r_vv.v_data, sizeof(mvp->m_fsize));

	/* useful tape info */
	mvp->m_bufsize = vrp->vr_bufsize;

	/*
	**	send pkt
	*/
	if (XDRSendI(hp, comslot, 1))
		return;

	/*
	**	mark status
	*/
	vrp->vr_mvseq = seq;
	vrp->vr_status = VSQ_MOVER;
	return;
}


/*
 *  CRMoverErr() - failed to get VV off a tape -
 *	try other tapes if possible.
 */
  void
CRMoverErr(VsnReq_t *vrp)
{
	char		*id = "CRMoverErr";
	ReadReq_t	*rp = (ReadReq_t *) vrp->vr_caller;
	int		badslot = vrp->vr_readslot;

nLogMsg(EL_STATUS1, id, "%s: badslot %d", VV2Name(&rp->r_vv.v_vv), badslot);

	/*
	**	take care of (un)mounted vsn - deletes vrp
	*/
	HandoffMountedVsn(vrp);
	rp->r_vrp = NULL;

	/*
	**	try again if possible, or err to client
	*/
	RetryRestore(rp, badslot);
}


  void
CRMoverOk(VsnReq_t *vrp)
{
	char		*id = "CRMoverOk";
	ReadReq_t	*rp = (ReadReq_t *) vrp->vr_caller;
	ReadReq_t	*rp2;
	VVMClass_t	*cip = (VVMClass_t *) rp->r_cp;
	int		slot = vrp->vr_readslot;
	char		*vsnp = vrp->vr_vsns + VSNLEN*slot;
	char		buf[VV_PKTSIZ(VVInfo_t)];
	vvpHdr_t	*rhp;

	nLogMsg(EL_STATUS1, id, "%s from %.*s",
				VV2Name(&rp->r_vv.v_vv), VSNLEN, vsnp);

	/*
	**	let go of space reservation
	*/
	DiskUnreserve(cip);

	/*
	**	take care of mounted vsn - deletes vrp
	*/
	HandoffMountedVsn(vrp);
	rp->r_vrp = NULL;

	/*
	**	update dbase
	*/
	DBRestored(&rp->r_vv.v_vv);

	if (rp->r_clnt == -1) {
		/*
		**	continue restore of lost vsn copy
		*/
		nLogMsg(EL_STATUS1, id, "%s back on disk - continue recovery",
						VV2Name(&rp->r_vv.v_vv));
		CWriteVV2Vsns(&rp->r_vv);
	}

	/*
	**	tell client(s)
	*/
	rhp = vvBuildHdr(VVM_VOLINFO, rp->r_seq, (vvpHdr_t *)buf);
	*VV_MSGPTR(rhp, VVInfo_t) = rp->r_vv;

	for (; rp!=NULL; rp=rp2) {
		rp2 = CWAITNEXT(rp);

		if (rp->r_clnt != -1)
			(void)XDRSendI(rhp, rp->r_clnt, 1);

		DelReadReq(rp);
        }
}


/*
 *  CRVolmanDown() - throw away mount request & wait for
 *	volman to restart
 */
  void
CRVolmanDown(VsnReq_t *vrp)
{
	char		*id = "CRVolmanDown";
	ReadReq_t	*rp = (ReadReq_t *) vrp->vr_caller;
	VVMClass_t	*cip = (VVMClass_t *) rp->r_cp;

	if (rp->r_vrp == NULL)
		nLogMsg(EL_FATAL, id, "reqp null");
	DelVsnReq(rp->r_vrp);
	rp->r_vrp = NULL;
}
