
/*
**	vvm_recycle.c - routines for restoring a bunch of VV's 
**		from a tape and recycling the tape.
**
**		The requests (1 per vsn) are queued on the VVMClass_t.
*/

#include "vvmd.h"


  static void
DelRecycleReq(RecycReq_t *ryp)
{
	char		*id = "DelRecycleReq";
	VVMClass_t	*cip = (VVMClass_t *) ryp->ry_cp;

	CRECYCDEQ(ryp, cip);
	if (ryp->ry_vrp != NULL) {
		if (ryp->ry_vrp->vr_status == VSQ_MOUNTED)
			VolmanUnmount(ryp->ry_vrp, id);
		DelVsnReq(ryp->ry_vrp);
	}
	
	nFree(ryp);
}


  static int
RecycRestoreVV(VVMClass_t *cip, RecycReq_t *ryp, int checkvol1)
{
	char		*id = "RecycRestoreVV";
	VsnReq_t	*vrp = ryp->ry_vrp;
	int		slot = vrp->vr_readslot;
	char		*vsnp = vrp->vr_vsns + VSNLEN*slot;
	char		buf[VV_PKTSIZ(vvpMvreq_t)];
	vvSeq_t		seq = vvGetSeq();
	vvpHdr_t	*hp;
	vvpMvreq_t	*mvp;
	int		comslot;

	comslot = HostMover(vrp->vr_hostp);
	if (comslot == -1) {
		nLogMsg(EL_ERROR1, id, "%.*s_%d - mover down (remount?)", 
							VSNLEN, vsnp, seq);
		return(1);
	}

	/*
	**	make sure the space is there
	*/
	if (DiskReserve(cip)) 
		return(1);

	/*
	**	build pkt for mover
	*/
	hp = vvBuildHdr(VVM_RTAPE, seq, (vvpHdr_t *)buf);
	mvp = VV_MSGPTR(hp, vvpMvreq_t);

	memcpy(mvp->m_vsn, vsnp, sizeof(mvp->m_vsn));

	mvp->m_checkvol1 = checkvol1;

	/* vsn record info */
	mvp->m_block = ryp->ry_vv.v_vsns[slot].v_block;
	mvp->m_rec = ryp->ry_vv.v_vsns[slot].v_nrec;

	/* hdr info for this vv */
	mvp->m_vv = ryp->ry_vv.v_vv;
	memcpy(mvp->m_id, ryp->ry_vv.v_id, sizeof(mvp->m_id));
	memcpy(mvp->m_fsize, ryp->ry_vv.v_data, sizeof(mvp->m_fsize));

	/* useful tape info */
	mvp->m_bufsize = vrp->vr_bufsize;

	/*
	**	send pkt to mover
	*/
	if (XDRSendI(hp, comslot, 1))
		return(1);

	/*
	**	mark status
	*/
	vrp->vr_mvseq = seq;
	vrp->vr_status = VSQ_MOVER;

	return(0);
}

/*
**	MountVsnRecycleReq() - push a recycle request into the vsn stage
**		(i.e. set ry_vrp)
*/
  static void
MountVsnRecycleReq(RecycReq_t *ryp)
{
	char		*id = "MountVsnRecycleReq";

	/*
	**	send mount request
	*/
	ryp->ry_vrp = VolmanMount(ryp->ry_slot, ryp->ry_panel, VQT_RECYCLE,
				(void *)ryp, (VVMClass_t *) ryp->ry_cp);
}


/*
**	MRecycleVsn() - setup/start process of:
**		- mount tape 
**		- pull off any nonresident VV's that are live,
**			scheduling them for save of new copy
**		- recycle vsn to volman
*/
  void
MRecycleVsn(vvpHdr_t *hp, int slot)
{
	char		*id = "MRecycleVsn";
	vvpVsn_t	*vp = VV_MSGPTR(hp, vvpVsn_t);
	int		recovervv;
	VVMClass_t	*cip;
	Vsn_t		*Vsnp;
	RecycReq_t	*ryp;

	/*
	**	get vsn record
	*/
	Vsnp = DBLookupVsn(vp->vq_vsn);
	if (Vsnp == NULL) {
		SendErr(hp->vh_seq, slot);
		return;
	}

	/*
	**	make sure class is still defined
	*/
	cip = VVMClassInfo(Vsnp->vs_class);
	if (cip == NULL) {
		VVError = VVE_ADMIN;
		nLogMsg(EL_ERROR1|EL_CODE, id, 
					"%.*s status %s: unknown class %s",
					VSNLEN, vp->vq_vsn,
					vvStatusStr(Vsnp->vs_status),
					ClassString(Vsnp->vs_class));
		SendErr(hp->vh_seq, slot);
		return;
	}

	/*
	**	vsn needs to be finished for writing to recycle
	*/
	if (Vsnp->vs_status != VVS_FINISHED) {
		VVError = VVE_UNAVAIL;
		nLogMsg(EL_ERROR1|EL_CODE, id, "%.*s: not finished, status %s",
						VSNLEN, vp->vq_vsn,
						vvStatusStr(Vsnp->vs_status));
		SendErr(hp->vh_seq, slot);
		return;
	}

	/*
	**	vsn can't be in use for restores either
	**	TODO - maybe allow this.
	*/
	if (VolmanBusy(vp->vq_vsn)) {
		VVError = VVE_UNAVAIL;
		nLogMsg(EL_ERROR1|EL_CODE, id, "%.*s: in use", 
						VSNLEN, vp->vq_vsn);
		SendErr(hp->vh_seq, slot);
		return;
	}

	/*
	**	DBInitRecycleVsn marks 'recycling' status on 
	**	the vsn record as well as in the vsn's part of
	**	any VV that needs to be written to another vsn
	**	(even if it's on disk). List indexes are updated
	**	appropriately to carry on the activity across
	**	shutdowns.
	*/
	recovervv = DBInitRecycleVsn(vp->vq_vsn);
	if (recovervv == -1) {
		SendErr(hp->vh_seq, slot);
		return;
	}
	if (recovervv == 0) {
		/*
		**	no mount needed, VV's (if any) all on disk
		*/
		SendAck(hp->vh_seq, slot);
		return;
	}

	/*
	**	need to restore VV's from tape - set up
	**	RecycReq_t in class q 
	*/
	ryp = (RecycReq_t *)nAlloc(sizeof(RecycReq_t));
	ryp->ry_cp = cip;
	ryp->ry_slot = Vsnp->vs_classI;
	memset(ryp->ry_panel, ' ', sizeof(ryp->ry_panel));
	memcpy(ryp->ry_panel + Vsnp->vs_classI * VSNLEN, vp->vq_vsn, 
							(size_t)VSNLEN);
	ryp->ry_totvv = recovervv;
	ryp->ry_donevv = 0;

	/* set nrec to indicate no vv done yet */
	ryp->ry_vv.v_vsns[ryp->ry_slot].v_nrec = -1;
	CRECYCENQ(ryp, cip);

	/*
	**	tell user all set up ok
	*/
	SendAck(hp->vh_seq, slot);

	/*
	**	try the mount
	*/
	MountVsnRecycleReq(ryp);
}

/*
**	CRRecycleRemount() - need to start new mount; delete old too
*/
 void
CRRecycleRemount(VsnReq_t *vrp)
{
	char		*id = "RecycleMounted";
	RecycReq_t	*ryp = (RecycReq_t *) vrp->vr_caller;
	VVMClass_t	*cip = (VVMClass_t *) ryp->ry_cp;
	char		*vsnp = ryp->ry_panel + VSNLEN * ryp->ry_slot;

	DelVsnReq(ryp->ry_vrp);
	MountVsnRecycleReq(ryp);
}

/*
**	RecycleMounted() - vsn has been mounted, so start/resume
**	restoring VV's
**	TODO - think about error handling
*/
 void
CRecycleMounted(VsnReq_t *vrp)
{
	char		*id = "RecycleMounted";
	RecycReq_t	*ryp = (RecycReq_t *) vrp->vr_caller;
	VVMClass_t	*cip = (VVMClass_t *) ryp->ry_cp;
	char		*vsnp = ryp->ry_panel + VSNLEN * ryp->ry_slot;

	/*
	**	make sure mover exists on host
	*/
	if (HostMover(vrp->vr_hostp) == -1) {
nLogMsg(EL_STATUS1, id, "%.*s_%d - mover down (remount?)", 
VSNLEN, vsnp, vrp->vr_mvseq);
		return;
	}

	/*
	**	get db info on 1st vv to restore
	*/
	if (DBRecycleGetNextRestoreVV(vsnp, ryp->ry_slot, &ryp->ry_vv)) {
		DelRecycleReq(ryp);
		return;
	}

	/*
	**	see if anything to restore; -1 is the db flag for nothing
	*/
	if (ryp->ry_vv.v_vsns[ryp->ry_slot].v_nrec == -1) {
		nLogMsg(EL_STATUS1, id, "%.*s: mounted but no vv's", 
							VSNLEN, vsnp);
		/* TODO - upgrade to recyc->vold? */
		/* DBRecycleVsnRestoresDone(vsnp); */
		DelRecycleReq(ryp);
		return;
	}

	/*
	**	send req to mover, flagged to check vsn's VOL1
	**	label, since vsn was just mounted
	*/
	if (RecycRestoreVV(cip, ryp, 1)) {
		DelRecycleReq(ryp);
		return;
	}
}	




/*
**	CRecycleMountFailed() - since there's only 1 vsn, fail req.
**		TODO - go into 'mark vsn bad' 
**		(i.e. get vv's from other tapes) mode?
*/
  void
CRecycleMountFailed(VsnReq_t *vrp)
{
	char		*id = "CRecycleMountFailed";
	RecycReq_t	*ryp = (RecycReq_t *) vrp->vr_caller;
	VVMClass_t	*cip = (VVMClass_t *) ryp->ry_cp;
	char		*vsnp = ryp->ry_panel + VSNLEN * ryp->ry_slot;

	nLogMsg(EL_CONSOLE0|EL_ERROR1, 0, "%.*s: RECYCLE MOUNT FAILED",
							VSNLEN, vsnp);

	DelRecycleReq(ryp);
}



  void
CRecycleMoverOk(VsnReq_t *vrp)
{
	char		*id = "CRecycleMoverOk";
	RecycReq_t	*ryp = (RecycReq_t *) vrp->vr_caller;
	VVMClass_t	*cip = (VVMClass_t *) ryp->ry_cp;
	int		slot = vrp->vr_readslot;
	char		*vsnp = vrp->vr_vsns + VSNLEN*slot;

	nLogMsg(EL_STATUS1, id, "%s from %.*s",
				VV2Name(&ryp->ry_vv.v_vv), VSNLEN, vsnp);

	/*
	**	release reserved space
	*/
	DiskUnreserve(cip);

	/*
	**	update dbase
	*/
	DBRecycRestoredVV(&ryp->ry_vv);

	/*
	**	start write of fresh copy of VV to vsn
	*/
	nLogMsg(EL_STATUS1, id, "%s back on disk - continue recovery",
						VV2Name(&ryp->ry_vv.v_vv));
	CWriteVV2Vsns(&ryp->ry_vv);

	/*
	**	on to next vv
	*/
	if (DBRecycleGetNextRestoreVV(vsnp, ryp->ry_slot, &ryp->ry_vv)) {
		DelRecycleReq(ryp);
		return;
	}

	/*
	**	see if anything to restore; -1 is the db flag for nothing
	*/
	if (ryp->ry_vv.v_vsns[ryp->ry_slot].v_nrec == -1) {
		nLogMsg(EL_STATUS1, id, 
				"%.*s: nothing left to restore", VSNLEN, vsnp);
		/*
		**	recyc->vold
		*/
		DBRecycleVsnRestoresDone(vsnp);
		DelRecycleReq(ryp);
		return;
	}

	/*
	**	send req to mover, flagged to not check vsn's VOL1
	**	label, since it was done on 1st VV
	*/
	if (RecycRestoreVV(cip, ryp, 0)) {
		DelRecycleReq(ryp);
		return;
	}
}

/*
**	CRecycleMoverErr() - failed to get VV off a tape -
**		put it in line to try other tapes if possible.
*/
  void
CRecycleMoverErr(VsnReq_t *vrp)
{
	char		*id = "CRecycleMoverErr";
	RecycReq_t	*ryp = (RecycReq_t *) vrp->vr_caller;
	VVMClass_t	*cip = (VVMClass_t *) ryp->ry_cp;

	/*
	**	release reserved space
	*/
	DiskUnreserve(cip);

nLogMsg(EL_FATAL, id, "unimplemented");
	/*
	**	make a single-VV version of DBBadVsn()
	**	to put this VV in line for restore
	*/
	/*
	**	get next VV & continue? count VV errs &
	**	bail w/ general DBBadVsn() if too many?
	*/
}


/*
**	CRecycleVolmanDown() - throw away mount request & wait for
**		volman to restart
*/
  void
CRecycleVolmanDown(VsnReq_t *vrp)
{
	char		*id = "CRecycleVolmanDown";
	RecycReq_t	*ryp = (RecycReq_t *) vrp->vr_caller;
	VVMClass_t	*cip = (VVMClass_t *) ryp->ry_cp;

	if (ryp->ry_vrp == NULL)
		nLogMsg(EL_FATAL, id, "reqp null");
	DelVsnReq(ryp->ry_vrp);
	ryp->ry_vrp = NULL;
}

 void
CRecycleCancel(VsnReq_t *vrp)
{
	char	*id = "CRecycleCancel";

	nLogMsg(EL_FATAL, id, "unexpected cancel");
}


 int
CRecycleRestarts()
{
	char		*id = "CRecycleRestarts";
	int		i = 0, recycle = 0, err = 0, recovervv;
	Vsn_t		*Vsnp;
	VVMClass_t	*cip;
	RecycReq_t	*ryp;

        while (Vsnp = DBGetIncompleteRecycle(i++, &recovervv)) {
		cip = VVMClassInfo(Vsnp->vs_class);
		if (cip == NULL) {
			nLogMsg(EL_ERROR1|EL_CONSOLE0, id, 
					"%.*s status %s: unknown class %s",
					VSNLEN, Vsnp->vs_vsn,
					vvStatusStr(Vsnp->vs_status),
					ClassString(Vsnp->vs_class));
			err++;
			continue;
		}
		if (Vsnp->vs_status != VVS_RECYCLING) {
			nLogMsg(EL_ERROR1, id, 
					"%.*s: unexpected status %s",
					VSNLEN, Vsnp->vs_vsn,
					vvStatusStr(Vsnp->vs_status));
			err++;
			continue;
		}
		if (recovervv < 1) {
			/* TODO - add any checks needed & start recycle/vold */
			nLogMsg(EL_ERROR1, id, 
					"%.*s: %d to recover",
					VSNLEN, Vsnp->vs_vsn, recovervv);
			err++;
			continue;
		}
		nLogMsg(EL_STATUS1, id, "  restarting %.*s, %d VV", 
					VSNLEN, Vsnp->vs_vsn, recovervv);
                recycle++;
		/*
		**	need to restore VV's from tape - set up
		**	RecycReq_t in class q 
		*/
		ryp = (RecycReq_t *)nAlloc(sizeof(RecycReq_t));
		ryp->ry_cp = cip;
		ryp->ry_slot = Vsnp->vs_classI;
		memset(ryp->ry_panel, ' ', sizeof(ryp->ry_panel));
		memcpy(ryp->ry_panel + Vsnp->vs_classI * VSNLEN, 
					Vsnp->vs_vsn, (size_t)VSNLEN);
		ryp->ry_totvv = recovervv;
		ryp->ry_donevv = 0;

		/* set nrec to indicate no vv done yet */
		ryp->ry_vv.v_vsns[ryp->ry_slot].v_nrec = -1;

		/*
		**	try the mount
		*/
		MountVsnRecycleReq(ryp);
		CRECYCENQ(ryp, cip);
	}
	return(recycle);
}
