/*
 *  NAStore Virtual Volume Manager
 *
 *  vvm_config.c: load host info and other config
 *
 */

#include "vvmd.h"

#include <syslog.h>
#include <sys/stat.h>


/*******************************************************************
			VVCONF PARSING ROUTINES
 *******************************************************************/

#define TOKLEN		80

/* gettok() modes - eofbomb implies newline, sameline is for line-formatted */
#define SAMELINE	0
#define GETNEWLINE 	1
#define EOFBOMB 	2

static FILE	*fp;
static int	linenum = 0, lineindex = 0, tab = 0;
static char	tok[TOKLEN], line[TOKLEN];

static char	vvConfig[MAXPATHLEN];

char *keywd[] = {

	/* 		    main keywords 		*/
	"BEGIN_HOSTS",		"END_HOSTS",
	"BEGIN_PANELS",		"END_PANELS",
	"START_FREE_SPACE",	"END_FREE_SPACE",
	"END",		"",
};


  static int
iskeywd()
{
	char	**p;

	for (p = keywd; strlen(*p); p++)
		if (!strcmp(tok, *p))
			return(1);
	return(0);
}

  void
inerr(char *str1, char *str2)
{	
	char	buf[81];
	int     i;

	if (lineindex > 2) {
		for (i=0; i<lineindex+7*tab-1; i++)
			buf[i] = '.';
		buf[i] = '\0';
		nLogMsg(EL_STATUS1, 0, "\n%s\n%s", line, buf);
	}
	nLogMsg(EL_FATAL|EL_TOSTDERR, 0, 
	 "config file %s:\n\tError on line %d: %s %s\n\tLast token was: '%s'\n",
	 vvConfig, linenum, str1, str2, tok);
}

 static int
readln(int eofbomb)
{
	int	i, j;

repeat:
	for (i=0; i<81; i++) {
		if ((j = fgetc(fp)) == EOF) {
			if (eofbomb)
				inerr("Unexpected EOF","");
			line[i] = '\n';
			lineindex = 0;
			tab = 0;
			return(i);
		}
		line[i] = (char) j;
		if (line[i] == '\n') {
			line[i] = '\0';
			break;
		}
	}
	if (line[i] != '\0')
		nLogMsg(EL_FATAL|EL_TOSTDERR, 0, 
			"%s: line too long (limit is 80):\n%.81s",
			vvConfig, line);
	nLogMsg(EL_STATUS1|EL_NOTIME, 0, line);
	linenum++;
	if (line[0] == '#'  ||  line[0] == '\0')
		goto repeat;
	lineindex = 0;
	tab = 0;
	return(i);
}


/*
 *  gettok() - get next token, going on to next line if requested
 *	if EOF, bomb if requested (newline==EOFBOMB)
 */
  void
gettok(int newline)
{
	static int	linelen;
	int		i;

	for (i=0; ; ) {
		switch (line[lineindex]) {
		    case '\t':
			tab++;
		    case ' ':
			lineindex++;
			if (i) {
				tok[i] = '\0';
				return;
			}
			break;
		    case '\n':
			/* EOF */
			tok[0] = '\0';
			return;
		    case '"':
			if (!i) {
				/*
				 *  beginning of string - get all
				 */
				lineindex++;
				if (line[lineindex] == '"')
					inerr("Empty string","");
				if (line[lineindex] == '\0')
					inerr("Unterminated string","");
				while(1) {
					tok[i++] = line[lineindex++];
					if (line[lineindex] == '"') {
						tok[i] = '\0';
						lineindex++;
						return;
					}
					if (line[lineindex] == '\0')
						inerr("Unterminated ",
								"string");
				}
			} else {
				/*
				 *  end of prev token
				 */
				tok[i] = '\0';
				return;
			}
		    case '\0':
			if (!i) {
				if (newline == SAMELINE) {
					tok[0] = '\0';
					return;
				}
				/*
				 *  still waiting for 1st char -
				 *	get another line
				 */
				linelen = readln(newline==EOFBOMB ? 1 : 0);
				continue;
			} else {
				tok[i] = '\0';
				return;
			}
		    default:
			tok[i++] = line[lineindex];
			lineindex++;
			break;
		}
	}
}

  int
tokint(int *nump)
{
	int	i, j = strlen(tok);

	i = 0;
	if (tok[0] == '-') i++;
	for (; i<j; i++)
		if (!isdigit(tok[i]))
			return(1);
	*nump = atoi(tok);
	return(0);
}

  static int
tokOFF(OFF_T *nump)
{
	int		i, j = strlen(tok);
	int		max = sizeof(OFF_T);

	*nump = 0;
	for (i=0; i<j; i++) {
		if (!isdigit(tok[i]))
			return(1);
		*nump *= 10;
		*nump += tok[i] - '0';
	}
	return(0);
}

 static void
ReadHosts()
{
	char	*id = "ReadHosts";

	while (1) {
		gettok(EOFBOMB);
		if (!strcmp(tok, "END_HOSTS"))
			break;
		if (strlen(tok) > MAXHOSTNAMELEN-1)
			nLogMsg(EL_FATAL|EL_TOSTDERR, id, 
					"%s: host name too long", tok);

		if (NewHost(tok))
			nLogMsg(EL_FATAL|EL_TOSTDERR, id, 
					"%s: PROBLEM w/ HOST (%s)",
					vvConfig, tok);
	}
}

 static void
ReadPanels()
{
	char		*id = "ReadPanels";
	VVMClass_t	*cip;
	int		num;

	while (1) {
		gettok(EOFBOMB);
		if (!strcmp(tok, "END_PANELS"))
			break;
		cip = VVMClassInfoByName(tok);
		if (cip == NULL)
			cip = VVMClassInfo(tok);
		if (cip == NULL)
			inerr("Unknown class", tok);

		gettok(EOFBOMB);
		if (tokint(&num))
			inerr("expected # of panels", tok);
		if (num < 1)
			inerr("expected # of panels > 0", tok);
		cip->vc_npanels = num;
	}
}
/* 
 *  VVConfig() - read in config file 
 */
  void
VVConfig()
{
	char	*id = "VVConfig";

	/*
	 *  open config file, read 1st line
	 */
	(void)sprintf(vvConfig, "%s/rc.vvm", ETCDIR);
	if ((fp = fopen(vvConfig, "r")) == NULL) {
		nLogMsg(EL_FATAL|EL_TOSTDERR, id,
		  "can't open configuration file (%s)", vvConfig);
	}
	nLogMsg(EL_STATUS1, id, "Config file: %s", vvConfig);

	/*
	**	parse config file
	*/
	(void)readln(1);
	while (1) {
		gettok(GETNEWLINE);
		if (tok[0] == '\0')
			break;
		if (!strcmp(tok, "BEGIN_HOSTS")) {
			ReadHosts();
		} else if (!strcmp(tok, "BEGIN_PANELS")) {
			ReadPanels();
		} else if (!strcmp(tok, "END")) {
			break;
		} else
		    inerr("expected BEGIN_HOSTS, BEGIN_PANELS, or END","");
	}
	(void)fclose(fp);
}



