
/*
 *  Virtual Volume Manager
 *
 *  vvm_init.c: Virtual Volume Manager Daemon initialization/termination
 */

/* 
 *  make globals non-extern in this file 
 */
#undef  EXTERN
#define EXTERN

#include "vvmd.h"

#include <stdlib.h>
#include <signal.h>

#include <dirent.h>
#include <sys/time.h>		/* for gettimeofday() */
#include <pwd.h>

int		shuttingdown = 0;
int		timerpid;		/* tracked for log screening purposes */


extern int	IsChild;
extern char	**argvIn;
extern char	StartDir[];



/*
 *  lock for critical section of BecomeVvmd()
 *	(also used in VVShutdown())
 */

static char	*startup_lock = "/tmp/newvvmd_lock";

/*
 *  SigChild(): called on SIGCHLD which is generated when one of the
 *	children dies.
 */
static int	timerforks = 0;
#define MAXTIMERFORK	3

 static void
SigChild(int sig)
{
	char		*id = "SigChild";
	int		status, pid, csig;

	if ((pid = wait(&status)) < 0) {
		nLogMsg(EL_SYS, id, "wait");
		if (SIGNAL(sig, SigChild))
			nLogMsg(EL_FATAL|EL_SYS, id, "signal SIGCHLD");
		return;
	}
	status &= 0177777;
	csig = status & 0177;
	status >>= 8;
	if (csig) 
		nLogMsg(EL_STATUS1, id, "child %d terminated on signal %d",
								pid, csig);
	
	if (!shuttingdown) {
		if (pid == timerpid) {
			if (status == 7)
				nLogMsg(EL_FATAL, id, 
					"TIMER (PID %d) DIED ON CONFIG PROBLEM",
					pid);
			if (++timerforks > MAXTIMERFORK)
				nLogMsg(EL_FATAL, id, 
					"TIMER (PID %d) DIED TOO MANY TIMES",
					pid);
			nLogMsg(EL_CONSOLE0|EL_ERROR1, id, 
				"TIMER (PID %d) DIED STATUS %x - restarting",
				pid, status);
			timerpid = ForkExec(VVTIMER, VVTIMER, VVM_TIMERSECS);
			if (SIGNAL(sig, SigChild))
				nLogMsg(EL_FATAL|EL_SYS, id, "signal SIGCHLD");
			return;
		}
		nLogMsg(EL_STATUS2, id, "pid %d exited with status %x", 
					pid, status);
	}
	/* reinstate handler (must be after wait()) */
	if (SIGNAL(sig, SigChild))
		nLogMsg(EL_FATAL|EL_SYS, id, "signal SIGCHLD");
}

  static void
SigAlarm(int signo)
{
	nLogMsg(EL_STATUS1, 0, "SigAlarm");
	if (SIGNAL(SIGALRM, SigAlarm))
		nLogMsg(EL_FATAL, 0, "signal SIGALRM");
}

  static void
SigPipe(int signo)
{
	nLogMsg(EL_STATUS1, 0, "SigPipe");
	if (SIGNAL(SIGPIPE, SigPipe))
		nLogMsg(EL_FATAL, 0, "signal SIGPIPE");
}

  static void
SigHandle(int signo)
{
	nLogMsg(EL_FATAL, "SigHandle", "Caught signal %d", signo);
}

  static void
SigInit()
{
	char	*id = "SigInit";
	int	sig;

	if (SIGNAL(SIGALRM, SigAlarm))
		nLogMsg(EL_FATAL, id, "signal SIGALRM");
	if (SIGNAL(SIGPIPE, SigPipe))
		nLogMsg(EL_FATAL, id, "signal SIGPIPE");

	for (sig = 1; sig < NSIG; sig++)
		switch (sig) {
		    case SIGABRT:	/* abort always causes core dump */
		    case SIGSEGV:
		    case SIGTRAP:
		    case SIGQUIT:
		    case SIGBUS:
		    case SIGCHLD:
		    case SIGKILL:	/* can't be caught */
		    case SIGSTOP:	/* can't be caught */
#ifdef SIGTSTP
		    case SIGTSTP:
#endif
#ifdef SIGCONT
		    case SIGCONT:
#endif
		    case SIGALRM:
		    case SIGPIPE:
			break;

		    default:
			if (SIGNAL(sig, SigHandle))
				nLogMsg(EL_FATAL, id, "signal(%d, SigHandle)", 
									sig);
		}
}

/*
**	InitCatchups() - get closed vv's that weren't completely
**	written and vsn's in the process of recycling from dbase 
**	and resume work on them. Only use this at startup time - 
**	existing reqs should recover themselves if volman restarts 
**	after this.
*/
  static void
InitCatchups()
{
	char		*id = "InitCatchups";
	VVInfo_t	*vip;
	int		i = 0, write = 0, read1st = 0, recycle = 0;

	/*
	**	VV's to be written out
	*/
	while (vip = DBGetIncompleteSave(i++)) {
		switch (vip->v_status) {
		    case VVS_FINISHED:
			CWriteVV2Vsns(vip);
			write++;
			break;
		    case VVS_RECOVER:
			CRestore(NOSEQ, -1, vip);
			read1st++;
			break;
		    default:
			nLogMsg(EL_FATAL, id, "%s: unexpected status %c/%s",
						VV2Name(&vip->v_vv), 
						vip->v_status,
						vvStatusStr(vip->v_status));
		}
	}

	/*
	**	vsn recycling
	*/
	recycle = CRecycleRestarts();

	nLogMsg(EL_STATUS2|EL_NOTIME, id, 
			"started %d incomplete save-to-vsn, %d pre-restore",
			write, read1st);
	if (recycle)
		nLogMsg(EL_STATUS2|EL_NOTIME, id, "restarted %d vsn recycles",
							recycle);
	LogClasses();
}

/*
 *  BecomeVvmd(): Check for another vvmd running and get a lock on
 *	the critical "becoming a vvmd" section.
 */
  static void
BecomeVvmd()
{
	char		*id = "BecomeVvmd";

	/*
	 *  get exclusive lock
	 */
	nGetLock(startup_lock, "vvmd");

	/*
	 *  start log file as vvmd
	 */
	nLogInit("VV", VVShutdown);
	nLogSetLevel(EL_STATUS2);
	nLogInitFile(vvLogFile(), EL_LEVEL, 1, NULL);
	nLogInitConsole(vv_syslog, EL_CONSOLE0, EL_LEVEL);
	nLogMsg(EL_STATUS1, id, "--------------------------------------------");
#ifdef DEBUG
	{	char	*lognam;

		if (lognam = getlogin()) {
			if (strlen(VolManOperAcct) >= strlen(lognam))
				(void)strcpy(VolManOperAcct, lognam);
		}
	}
#endif
}

/*
 *  VVInit(): initialize VVMD
 *
 *	initialize data structures, get lock on database, and start IPC
 */
  void
VVInit()
{
	char		*id = "VVInit";
	DIR		*dirp;
	char		fbuf[100];

#ifdef MALLOC_DEBUG
	mallopt(MALLOC_DEBUG, 1);
#endif
	/*
	**	initialize EXTERN vars from vvmd.h
	*/
	IsChild = 0;
	VsnQ.q_head = NULL;
	VsnQ.q_tail = NULL;
	(void)time(&Now);

	/*
	**	make sure vSpoolDir exists and go to it in case of coredump
	*/
        if ((dirp = opendir(vvSpoolDir)) == NULL) {
		if (mkdir(vvSpoolDir, 0755) < 0) {		/* create it */
			(void)fprintf(stderr, "%s: mkdir(%s): %s", 
					id, vvSpoolDir, strerror(errno));
			exit(1);
		}
        } else
		(void)closedir(dirp);
	if (chdir(vvSpoolDir) < 0) {
		(void)fprintf(stderr, "%s: chdir(%s): %s", 
					id, vvSpoolDir, strerror(errno));
		exit(1);
	}

	/*
	**	negotiate any locks and start nLogMsg() logging
	*/
	BecomeVvmd();
	nLogMsg(EL_CONSOLE0|EL_STATUS1, id, "vvmd starting, version %d",
								VV_VERSION);

	/*
	**	make sure disk cache dir exists
	*/
	DiskStat();

	/*
	**	init socket structs
	*/
	IPCInit();

	/*
	**	map hard-coded part of classes to VVM struct
	*/
	InitClasses();

	/*
	**	read in the config file
	*/
	VVConfig();

	/*
	**	Start dbase - must be done before setting SIGCHLD
	**		since InitDb() needs to wait() if rebuilding indexes,
	**					[in parallel - to be added]
	**	and before starting other children since they might
	**	also interfere w/ index process wait()s.
	*/
	DBInit();

	/*
	**	get atime's of VV's on disk for cache mgt
	*/
	DiskInitList();

	/*
	**	catch all signals; SIGCHILD has a specific handler
	*/
	SigInit();
	if (SIGNAL(SIGCHLD, SigChild))
		nLogMsg(EL_FATAL|EL_SYS, id, "signal SIGCHLD");

	/*
	**	init socket/XDR links, start listening for connections
	*/
	IPCStartup();

	/*
	**	start the timer
	*/
	timerpid = ForkExec(VVTIMER, VVTIMER, TIMERSECS);

	/*
	**	now that children are off, close stdin, stdout
	*/
	(void)fclose(stdin);	
	(void)fclose(stdout);

	/*
	**	start volman connection
	*/
	(void)VolmanInit();

	/*
	**	mount or allocate current hot panels and resume
	**		work on partially-taped vv's
	*/
	CWInitHotPanels();
	InitCatchups();

	nLogMsg(EL_CONSOLE0|EL_STATUS1, id, 
				"vvmd initialized (pid %d)", getpid());
}


/*
 *  VVShutdown() - terminate vvmd
 */
  int
VVShutdown(char *caller, char *msg)
{
	char		*id = "VVShutdown";

	/*
	 *  if this is a child process (forked to write a long msg)
	 *	don't clean up - this would mess up the indexes etc.
	 */
	if (IsChild) {
		DBCleanupChild();	/* in case index was copied */
		exit(0);
	}

	if (caller == NULL)
		caller = "";
	if (shuttingdown) {
		nLogMsg(EL_STATUS2, id, "recursive call by %s", caller);
		return(0);
	}

	shuttingdown = 1;
	nLogMsg(EL_STATUS1, id, "called by %s", caller);
	if (msg && *msg)
		nLogMsg(EL_STATUS1|EL_TOSTDERR, id, msg);

	/*
	 *  close dbase 1st in case machine is going down
	 *	and vvmd might die soon - this preserves
	 *	indexes so that they don't need to be
	 *	rebuilt next time. NOTE: if any other
	 *	shutdown steps come to need dbase, this
	 *	would need to be rethought.
	 */
	nLogMsg(EL_STATUS2, id, "Closing dbase");
	DBClose();

	nLogMsg(EL_STATUS2, id, "No longer accepting connections");
	IPCShutdown();

	nLogMsg(EL_STATUS2, id, "Shutting down volman conn");
	(void)vm_term();

	nLogMsg(EL_STATUS2, id, "Shutting down clients & server procs");
	ShutdownServers();

	(void)kill(timerpid, SIGINT);

	/*
	 *  release startup lock in case death at startup critical section
	 */
	if (unlink(startup_lock)) {
		if (errno != ENOENT)
			nLogMsg(EL_SYS, id, 
				"unable to delete lock %s", startup_lock);
	}
	nLogMsg(EL_STATUS1|EL_CONSOLE0, id, "exiting");
	exit(0);
	return(0);	/* for compiler warnings */
}

