
/*
 *  NAStore Virtual Volume Manager: Mover
 *	(Adapted from volman volnd)
 *
 *  mv_1.c: vvmvd - mover daemon - server prog to check labels & 
 *	read /dev/vol/xxxxxx nodes into virtual volume files. Rcvs 
 *	commands from vvd, forks process to do request, times out / 
 *	waits for process, informs vvd of result.
 *
 *	TODO: consider forking 1 proc per drive at startup to do
 *	all work on that drive.. unless it hangs...
 *
 *  mv_2.c: child routines
 *
 *  Bill Ross, NASA Ames, September 1997
 */


#include "mv.h"

#include <stdio.h>
#include <dirent.h>
#include <signal.h>
#include <sys/wait.h>
#include <setjmp.h>
#include <sys/param.h>
#include <sys/mount.h>

#define VVMD_TIMEOUT	600	/* timeout on vvmd rcv */

extern int		immortal;
extern jmp_buf		jmpbuf_i;

time_t			Now, LastTime = 0;
char			**args;		/* argv */
char			*arg0;
char			myhost[MAXHOSTNAMELEN], logstr[MAXHOSTNAMELEN];

Req_t			*ReqArray = NULL;
int			MaxReq = 0;

 void
usage()
{
	fprintf(stderr, "usage: %s [-h vvmd_host] [-i]\n", args[0]);
	exit(1);
}

int
bomb()
{
	VVCLGoodbye(1);
	sleep(3);		/* try to let vvmd rcv goodbye */
	kill(0, 5);	/* try to take out children, too */
	return(1);	/* for compiler warning */
}

int
SigTerm()
{
	nLogMsg(EL_STATUS1, 0, "caught SIGTERM");
	(void)bomb();
}


/*
 *  SigChild() - handle status & reset signal-catcher
 */
 void
SigChild(int sig)
{
	char		*id = "SigChild";
	int		status, pid, csig, req;

	/*
	 *  handle all dead children
	 */
	while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
		Req_t		*rp;

		if (status == 0)
			nLogMsg(EL_DEBUG1, id, "pid%d - status 0x%x", 
							pid, status);
		else
			nLogMsg(EL_STATUS1, id, "pid%d - status 0x%x", 
							pid, status);
		status &= 0177777;
		csig = status & 0177;
		status >>= 8;
		if (csig) 
			nLogMsg(EL_STATUS1, id, 
					"child %d terminated on signal %d",
					pid, csig);
		/*
		 *  figure out which request was involved
		 */
		for (req=0; req<MaxReq; req++) {
			if (ReqArray[req].r_seq == VVNOSEQ)
				continue;
			if (pid == ReqArray[req].r_pid)
				break;
		}
		if (req == MaxReq) {
			nLogMsg(EL_STATUS2, id, 
					"pid%d forgotten - status 0x%x", 
					pid, status);
			continue;
		}
		rp = &ReqArray[req];

		/*
		 *  handle req, reset it
		 */
		if (status != EXIT_OK) {
			if (csig)
				nLogMsg(EL_ERROR1|EL_CODE, id,
					"request exited %d on signal %d", 
					status, csig);
			else
				nLogMsg(EL_ERROR1|EL_CODE, id,
					"request failed, status %d", status);
		}
		ReqArray[req].r_seq = VVNOSEQ;
		ReqArray[req].r_pid = 0;
	}

	if (pid == -1  &&  errno != ECHILD)	/*  ECHILD == no child */
		nLogMsg(EL_SYS, id, "waitpid (status 0x%x)", status);

	/*
	 *  reset catcher
	 */
	if (SIGNAL(sig, SigChild))
		nLogMsg(EL_FATAL|EL_SYS, id, "signal SIGCHLD");
}

/*
 *  CheckChild() - to be sure a sigcld didn't get lost (may
 *	be overkill) - see if info can be gotten
 *	return=1 means child status handled.
 */
CheckChild(int req)
{
	char	*id = "CheckChild";
	Req_t	*rp = &ReqArray[req];
	int	ret, status, csig;

	/*
	**	waitpid() w/ 'instant return'
	*/
	ret = waitpid(rp->r_pid, &status, WNOHANG);
	if (ret == -1) {
		nLogMsg(EL_SYS, id, "waitpid");
		if (errno == ECHILD) {
			/*
			 *  child death wasn't detected (?)
			 *	assume all went well
			 */
			ReqArray[req].r_seq = VVNOSEQ;
			ReqArray[req].r_pid = 0;
			return(1);
		}
		return(0);
	} 
	if (ret == rp->r_pid) {
		/*
		 *  child's SIGCLD was lost, handle
		 *	as if receiving now -translate
		 */
		status &= 0177777;
		csig = status & 0177;
		status >>= 8;
		if (csig) 
			nLogMsg(EL_STATUS1, id, 
					 "child pid%d terminated on signal %d",
					 rp->r_pid, csig);
		ReqArray[req].r_seq = VVNOSEQ;
		ReqArray[req].r_pid = 0;
		return(1);
	}
	if (ret == 0)
		nLogMsg(EL_CONSOLE0|EL_ERROR1, id,
			"pid%d hung on open or read - problem w/ drive",
			rp->r_pid);
	return(0);
}

void
ResetReqs()
{
	int	i;

	for (i=0; i<MaxReq; i++) {
		Req_t	*rp = &ReqArray[i];
		if (rp->r_pid  &&  rp->r_pid != -1)
			kill(rp->r_pid, SIGTERM);
		rp->r_pid = 0;
		rp->r_cmd = 0;
	}
}

/*
 *  MReq() - tape read/write request from vvmd
 */
 static void
MReq(vvpHdr_t *hp)
{
	char		*id = "MReq";
	vvpMvreq_t	*mp = VV_MSGPTR(hp, vvpMvreq_t);
	int		i, slot = -1;
	static sigset_t	block_sig;

	/*
	 *  make sure cache is mounted
	 */
	if (nCheckMount(vvVVDir)) {
		MVErr2VV(&mp->m_vv, mp->m_vsn, hp->vh_seq, hp->vh_type);
		return;
	}

	/*
	 *  allocate slot / check for duplicate seq
	 */
	for (i=0; i<MaxReq; i++) {
		if (ReqArray[i].r_seq == hp->vh_seq) {
			VVError = VVE_INTERNAL;
			nLogMsg(EL_ERROR1|EL_CODE, id, 
					"duplicate seq %d", hp->vh_seq);
			MVErr2VV(&mp->m_vv, mp->m_vsn, 
					hp->vh_seq, hp->vh_type);
			return;
		}
		if (ReqArray[i].r_seq == VVNOSEQ  &&  slot == -1)
			slot = i;
	}
	if (slot == -1) {
		/*
		 *  expand request array
		 */
		slot = MaxReq;
		MaxReq += 10;
		ReqArray = nReAlloc(ReqArray, MaxReq * sizeof(Req_t));
		for (i=slot; i<MaxReq; i++)
			ReqArray[i].r_seq = VVNOSEQ;
	}

	/*
	 *  track request
	 */
	ReqArray[slot].r_cmd = hp->vh_type;
	ReqArray[slot].r_seq = hp->vh_seq;
	ReqArray[slot].r_req = *mp;
	ReqArray[slot].r_time = Now;

	/*
	 *  child does the work, single parent waits on all results
	 */

	for (i=1; i<5; i++) {
		/*
		**	Block notice of any children dying while forking, 
		**		since we have observed the child to 
		**		complete and the caught signal to cause 
		**		reset of r_pid before the test of it happens.
		*/
		sigemptyset(&block_sig);
		sigaddset(&block_sig, SIGCHLD);
		if (sigprocmask(SIG_BLOCK, &block_sig, NULL) == -1)
			nLogMsg(EL_FATAL|EL_SYS, id, "sig_block");

		/*
		**	try the fork
		*/
		if ((ReqArray[slot].r_pid = fork()) != -1)
			break;

		/*
		**	fork failed; reinstate SIGCHLD, checking for
		**		any dead children, before waiting to retry
		*/
		nLogMsg(EL_ERROR1|EL_SYS, id, "fork");

		sigemptyset(&block_sig);
		sigaddset(&block_sig, SIGCHLD);
		if (sigprocmask(SIG_UNBLOCK, &block_sig, NULL) == -1)
			nLogMsg(EL_FATAL|EL_SYS, id, "sig_unblock");
		sleep(i*i*i);
	}
	if (ReqArray[slot].r_pid == -1)
		nLogMsg(EL_FATAL|EL_SYS|EL_CONSOLE0, id, "can't fork");

	/*
	**	child does its thing
	*/
	if (ReqArray[slot].r_pid == 0)
		exit(MoverChild(slot));

	/*
	**	parent notes the event & turns on SIGCHLD again
	*/
	nLogMsg(EL_STATUS1, 
			(hp->vh_type == VVM_WTAPE ? "WTAPE" : "RTAPE"),
			"%s_%d on %.*s: forked pid%d req %d", 
			VV2Name(&mp->m_vv), hp->vh_seq, VSNLEN, mp->m_vsn, 
			ReqArray[slot].r_pid, slot);
	sigemptyset(&block_sig);
	sigaddset(&block_sig, SIGCHLD);
	if (sigprocmask(SIG_UNBLOCK, &block_sig, NULL) == -1)
		nLogMsg(EL_FATAL|EL_SYS, id, "sig_unblock");
}

/*
 *  MTimer() - heartbeat from vvmd 
 */
 static void
MTimer(vvpHdr_t	*hp)
{
	char		*id = "MTimer";
	vvpTimer_t	*tp;
	int		i;

	/*
	 *  since we are pkt-driven, make sure they have not
	 *	piled up somehow - no point in thrashing the
	 *	log file
	 */
	if (Now - LastTime < 5) {
		return;
	}
	LastTime = Now;

	/*
	 *  consider msg time-of-flight from vvmd
	 */
	tp = VV_MSGPTR(hp, vvpTimer_t);
/*
	if (Now - tp->vt_time > 5) 
		nLogMsg(EL_STATUS1, id, "(msg time from vvmd %d)",
					Now - tp->vt_time);
*/
	/*
	 *  check requests
	 */
	for (i=0; i<MaxReq; i++) {
		Req_t	*rp = &ReqArray[i];

		if (rp->r_seq == VVNOSEQ)
			continue;

		/*
		 *  check node for activity
		 */
		if (Now - rp->r_time > TIMEOUT) {
			/*
			 *  see if a SIGCLD was missed
			 */
			if (CheckChild(i))
				continue;

			VVError = VVE_HARDWARE;
			nLogMsg(EL_CONSOLE0|EL_CODE|EL_ERROR1, id, 
				"%s_%d: [%.*s] pid%d hung on %s",
				VV2Name(&rp->r_req.m_vv), rp->r_seq, 
				VSNLEN, rp->r_req.m_vsn, 
				rp->r_pid,
                        	(rp->r_cmd == VVM_WTAPE ? "WTAPE" : "RTAPE"));
			MVErr2VV(&rp->r_req.m_vv, rp->r_req.m_vsn, 
						rp->r_seq, rp->r_cmd);

			/*
			 *  plan to ignore child death (seq=VVNOSEQ)
			 *	and try to kill it
			 */
	    		rp->r_seq = VVNOSEQ;
			if (!kill(rp->r_pid, SIGTERM))
				nLogMsg(EL_SYS, id, "kill pid%d", rp->r_pid);
			rp->r_pid = -1;
		}
	}
}



main(int argc, char *argv[])
{
	char		*id = argv[0];
	int		cc, i, vvmd_sd, err, dflg=0;
	extern int	optind;
	extern char	*optarg;
	char		*cp, *host = NULL;
	char		buf[MAXPATHLEN];
	DIR		*dirp;
	fd_set		fds;

	args = argv;
	arg0 = argv[0];
#ifndef PURIFY
	if (getuid() != 0) {
		fprintf(stderr, "%s: must be run as root\n", argv[0]);
		exit(1);
	}
#endif

	/*
	 *  get host info & set default
	 */
	if (gethostname(myhost, MAXHOSTNAMELEN) == -1) {
		perror("gethostname");
		exit(1);
	}
	host = myhost;

	/*
	**	parse args
	*/
	while ((cc = getopt(argc, argv, "h:id")) != EOF) {
		switch (cc) {
		    case 'h':
			host = optarg;
			break;
		    case 'i':
			immortal = 1;
			break;
		    case 'd':
			dflg = 1;
			break;
		    default:
			fprintf(stderr, "unexpected option (%c)\n", cc);
			usage();
		}
	}
#ifdef PURIFY
	dflg = 1;
#endif
	if (optind < argc) {
		fprintf(stderr, "argc %d, last [%s]\n", argc, argv[argc-1]);
		usage();
	}

	/*
	**	get to an appropriate place in case core is dumped
	*/
	if (strstr(vvSpoolDir, vvVVDir) != NULL) {
		/*
		**	spooldir is under vvdir
		*/
		if (nCheckMount(vvVVDir)) {
			fprintf(stderr, "%s MOUNT PROBLEM\n", vvVVDir);
			exit(1);
		}

	}
	sprintf(buf, "%s/%s_%s", vvSpoolDir, "mover", myhost);
	if ((dirp = opendir(buf)) == NULL) {
		if (mkdir(buf, 0755) == -1) {
			fprintf(stderr, "%s: mkdir(%s): %s",
				arg0, buf, strerror(errno));
			exit(1);
		}
	} else
		closedir(dirp);

	if (chdir(buf) == -1) {
		fprintf(stderr, "%s: chdir(%s): %s",
				arg0, buf, strerror(errno));
		exit(1);
	}

	/*
	**	go into the background
	*/
	if (!dflg) {
		switch (fork()) {
		    case -1:				/* fork failed */
			perror("fork");
			exit(1);
		    case 0:				/* child */
			if (setpgid(0,0))
				perror("setpgid");
			break;
		    default:				/* parent */
			exit(0);	/* detach */
		}
	}

	/*
	**	build log tag string
	*/
	cp = strchr(myhost, '.');
	if (cp != NULL)
		*cp = '\0';
	sprintf(logstr, "MV.%s", myhost);

	/*
	**	start logging
	*/
	nLogInit(logstr, bomb);
	nLogInitFile(vvLogFile(), EL_LEVEL, 0, NULL);
	nLogInitConsole(vv_syslog, 0x00040000, EL_LEVEL);
	nLogMsg(EL_STATUS1, 0, "-------------------- start on %s", myhost);

	/*
	**	build initial ReqArray
	*/
	MaxReq = 20;
	ReqArray = (Req_t *) nAlloc(MaxReq * sizeof(Req_t));
	memset(ReqArray, 0, MaxReq * sizeof(Req_t));
	for (i=0; i<MaxReq; i++)
		ReqArray[i].r_seq = VVNOSEQ;

	/*
	**	prepare for immortality in case req'd
	*/
	if (setjmp(jmpbuf_i)) {
		nLogMsg(EL_STATUS1, 0, "setjmp to beginning ok");
		ResetReqs();
		sleep(30);
	}
	if (SIGNAL(SIGTERM, SigTerm))
		nLogMsg(EL_FATAL|EL_SYS, id, "signal SIGCHLD");

	/*
	**	connect/register w/ vvmd
	*/
	nLogMsg(EL_STATUS1, id, "vvmd host is %s", vvHost);
	(void)VVCLInit(vvHost, VVM_MOVER, RETRY_FOREVER);

	/*
	**	make sure disk cache is mounted on this machine
	*/
	if (nCheckMount(vvVVDir))
		nLogMsg(EL_CONSOLE0|EL_FATAL, id, "MOUNT %s PROBLEM", vvVVDir);

	/*
	**	set up for select/timeout
	*/
	vvmd_sd = VVCLSockId();
	FD_ZERO(&fds);
	FD_SET(vvmd_sd, &fds);

	if (SIGNAL(SIGINT, bomb))
		nLogMsg(EL_FATAL|EL_SYS, id, "signal SIGINT");
	if (SIGNAL(SIGCHLD, SigChild))
		nLogMsg(EL_FATAL|EL_SYS, id, "signal SIGCHLD");

	nLogMsg(EL_CONSOLE0|EL_STATUS1, 0, "vvmvd initialized (pid%d) %s", 
				getpid(), (immortal ? "(immortal)" : ""));

	/*
	**	service node setup/delete requests from vvmd
	*/
	while (1) {
		int		nsoc, err = 0;
		struct timeval	tv;

		/*
		**	wait for pkt or timeout (timer pkts should
		**	arrive regularly - every 30 sec - and if not,
		**	assume a transport problem since vvmvd's have 
		**	sat on a rcv for weeks when vvmd host went down
		*/
		tv.tv_sec = VVMD_TIMEOUT;
		while ((nsoc = select(vvmd_sd+1, &fds, (fd_set *) NULL,
                                        (fd_set *) NULL, &tv)) == -1
                                        && errno == EINTR);
                if (nsoc == 0) {
                        nLogMsg(EL_STATUS1|EL_CONSOLE0, id, 
						"timeout - vvmd presumed down");
			if (immortal) {
				VVCLGoodbye(0);
				longjmp(jmpbuf_i, 1);
				nLogMsg(EL_SYS, 0, "longjmp on timeout");
			}
			continue;	/* to see what happens */
                }

		/*
		**	got at least 1 packet
		*/
		do {
			vvpHdr_t	*hp;
			hp = VVCLRecv("main loop");

			nLogMsg(EL_DEBUG3, id, vvHdrStr(hp));

			time(&Now);	/* grab a unique second of eternity */

			switch (hp->vh_type) {
			    case VVM_WTAPE:
			    case VVM_RTAPE:
				MReq(hp); break;
			    case VVM_TIMER:
				MTimer(hp); break;
			    case VVM_SHUTDOWN:
				if (immortal) {
					nLogMsg(EL_CONSOLE0|EL_STATUS1, 0, 
								"shutdown");
					longjmp(jmpbuf_i, 1);
					nLogMsg(EL_SYS, 0, 
							"longjmp on shutdown");
				}
				nLogMsg(EL_CONSOLE0|EL_STATUS1, 0, 
						"shutdown rcvd - exiting");
				VVCLGoodbye(0);
				/*
				**	try to take out children, too
				*/
				kill(0, SIGTERM);
			    case VVM_NEWLOG:
				sleep(1);
				nLogMsg(EL_STATUS1, "MNewlog", 
				"============ Restarting log ================");
				nLogInitFile(vvLogFile(), EL_LEVEL, 0, NULL);
				nLogMsg(EL_STATUS1, "MNewLog", 
				"--------------------------------------------");
				nLogMsg(EL_STATUS1, "MNewLog", 
				"---------- RESTARTED LOG -------------------");
				break;
			    case VVM_DEBUG:
				MMDebug(hp); break;
			    default:
				nLogMsg(EL_ERROR1|EL_CONSOLE0, 0, 
						"unexpected msg type %s",
						vvHdrStr(hp));
				break;
			}
			nFree(hp);
		} while (VVCLNextRecord());
	}
}




 void
PrintRequests(char *str)
{
	int	i;

	nLogMsg(EL_STATUS1, str, "-------------- ");

	nLogMsg(EL_STATUS1, 0, "seq     vv      vol      cmd      time");
	for (i=0; i<MaxReq; i++) {
		Req_t	*rp = &ReqArray[i];

		if (rp->r_seq != VVNOSEQ)
			nLogMsg(EL_STATUS1, 0, "%d  %s  %.*s  %s  %s",
				rp->r_seq, 
				VV2Name(&rp->r_req.m_vv),
				VSNLEN, &rp->r_req.m_vsn, 
				nas_time(Now - rp->r_time));
	}
	nLogMsg(EL_STATUS1, 0, "-------------- %s", str);
}

  void
MVErr2VV(VV_t *vvp, char *vsnp, vvSeq_t seq, int req)
{
	char		*id = "MVErr2VV";
	vvpHdr_t	*hp;
	vvpError_t	*ep;
	char		ebuf[VV_PKTSIZ(vvpError_t)];

	hp = vvBuildHdr(VVM_VVERROR, seq, (vvpHdr_t *)ebuf);
	if (hp->vh_seq == -1) {
		nLogMsg(EL_ERROR1, id, "seq -> -1 from %d", seq);
		hp->vh_seq = seq;
	}
	ep = VV_MSGPTR(hp, vvpError_t);
	ep->vn_req = req;
	ep->vn_vv = *vvp;
	memcpy(ep->vn_vsn, vsnp, VSNLEN);
	ep->vn_error = VVError;
	emsgcpy(ep->vn_msg, nErrStr);
	if (hp->vh_seq == -1) {
		nLogMsg(EL_ERROR1, id, "#2: seq -> -1 from %d", seq);
		hp->vh_seq = seq;
	}
	VVCLSend(hp, 1);
	nLogMsg(EL_DEBUG3, 0, "srv_ sent err");
}

