/*	$NetBSD$	*/

/*-
 * Copyright (c) 1998-1999 Brett Lymn
 *                         (blymn@baea.com.au, brett_lymn@yahoo.com.au)
 * All rights reserved.
 *
 * This code has been donated to The NetBSD Foundation by the Author.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software withough specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 */

#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include "edit.h"
#include "buffer.h"
#include "screen.h"
#include "utils.h"

/* declared in edit.c */
extern int mark_count, cursor_line, cursor_pos, screen_pos, marked;
extern EDIT_LINE *current_line, *select_line, *top_line, *first_line;

EDIT_LINE *alloc_buffer = NULL;

/* function prototypes */
void
copy_to_buffer(EDIT_LINE *line, EDIT_LINE **pos);
void
free_cut_buffer(void);

/*
 * Release the storage allocated to the cut buffer.
 */
void
free_cut_buffer(void)
{
	EDIT_LINE *this;
	
	if (alloc_buffer == NULL)
		return;

	for (this = alloc_buffer; this != NULL; this = this->next)
		free(this->line);

	free(alloc_buffer);
	alloc_buffer = NULL;
}


/*
 * Copy the specified line into the copy buffer.  An appropriately sized
 * copy buffer is assumed to already have been allocated.  The parameter
 * pos points at the location in the cut buffer at which the line should
 * be placed, this parameter will be updated to point at the newly inserted
 * line.
 */
void
copy_to_buffer(EDIT_LINE *line, EDIT_LINE **pos)
{
	EDIT_LINE *old_pos;
	
	if (((*pos)->line = (char *) malloc(sizeof(char) * (line->length + 1)))
	    == NULL) {
		fprintf(stderr, "Could not malloc copy buffer string\n");
		return;
	}
	strcpy((*pos)->line, line->line);
	(*pos)->length = line->length;
	(*pos)->alloced = line->alloced;
	
	if (*pos == alloc_buffer) {
		(*pos)->prev = NULL;
	}

	old_pos = *pos;
	*pos = *(pos) + 1;
	old_pos->next = *(pos);
	(*pos)->prev = old_pos;
}

	
	
/*
 * Copy the region from the select line to the current line.  The variable
 * mark_count tells us which way to copy, negative is towards the top of
 * the edit buffer.
 */
void
copy_region(void)
{
	EDIT_LINE *this, *buf_current = NULL;

	if (marked == 0)
		return;
	
	if (alloc_buffer != NULL)
		free_cut_buffer();

	if ((alloc_buffer = (EDIT_LINE *)
	     malloc((abs(mark_count)) * sizeof(EDIT_LINE))) == NULL) {
		fprintf(stderr, "Cannot allocate copy buffer\n");
		return;
	}

	buf_current = alloc_buffer;
	
	if (mark_count >= 0) {
		for (this = select_line; this != current_line;
		     this = this->next) {
			copy_to_buffer(this, &buf_current);
		}
		copy_to_buffer(current_line, &buf_current);
	} else {
		for (this = current_line; this != select_line;
		     this = this->next) {
			copy_to_buffer(this, &buf_current);
		}
		copy_to_buffer(select_line, &buf_current);
	}

	buf_current->prev->next = NULL;
}

/*
 * Delete the region between the select line and the current line from the
 * edit buffer.
 */
void
delete_region(void)
{
	EDIT_LINE *this, *new_current, *new_this;

	if ((select_line == NULL) || (marked == 0))
		return;

	if (mark_count >= 0) {
		new_current = select_line->prev;
		if (new_current != NULL) {
			select_line->prev->next = current_line->next;
			if (current_line->next != NULL) {
				current_line->next->prev = select_line->prev;
			}
		} else {
			if (current_line->next != NULL) {
				new_current = current_line->next;
				current_line->next->prev = NULL;
			} else {
				/* the cut will empty the buffer */
				top_line = NULL;
				new_current = NULL;
				first_line = NULL;
				cursor_line = 0;
				cursor_pos = 0;
				screen_pos = 0;
			}
		}

		for (this = select_line; this != current_line;) {
			new_this = this->next;
			free_line(this);
			this = new_this;
		}
		free_line(current_line);
	} else {
		new_current = select_line->next;
		if (new_current != NULL) {
			select_line->next->prev = current_line->prev;
			if (current_line->prev != NULL) {
				current_line->prev->next = select_line->next;
			}
		} else {
			if (current_line->prev != NULL) {
				new_current = current_line->prev;
				current_line->prev->next = NULL;
			} else {
				/* the cut will empty the buffer */
				top_line = NULL;
				new_current = NULL;
				first_line = NULL;
				cursor_line = 0;
				cursor_pos = 0;
				screen_pos = 0;
			}
		}
		
		for (this = current_line; this != select_line;) {
			new_this = this->next;
			free_line(this);
			this = new_this;
		}
		free_line(select_line);
	}

	current_line = new_current;
	redraw_screen();
}

/*
 * Insert the cut buffer into the edit buffer at the current line.
 */
void
insert_buffer(void)
{
	EDIT_LINE *this, *old_next;

	if (alloc_buffer == NULL)
		return;

	old_next = current_line->next;
	for (this = alloc_buffer; this != NULL; this = this->next) {
		insert_line(this->line);
	}

	if (old_next != NULL)
		old_next->prev = current_line;
	
	redraw_screen();
}

	
