/*	$NetBSD: driver.c,v 1.3 2000/05/07 13:16:48 blymn Exp $	*/

/*-
 * Copyright (c) 1998-1999 Brett Lymn
 *                         (blymn@baea.com.au, brett_lymn@yahoo.com.au)
 * All rights reserved.
 *
 * This code has been donated to The NetBSD Foundation by the Author.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software withough specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 */

#include <ctype.h>
#include <malloc.h>
#include <menu.h>
#include <stdlib.h>
#include <strings.h>
#include "edit.h"
#include "driver.h"
#include "file.h"
#include "help.h"
#include "screen.h"
#include "search.h"
#include "utils.h"

/* these are defined in edit.c */
extern WINDOW *edit_win, *top_win, *filler;
extern MENU *top_menu;
extern int cursor_line, cursor_pos, screen_pos, edit_maxy, mark_count;
extern int pushback;
extern EDIT_LINE *current_line, *top_line;


int menu_mode = 0;

/* function prototypes. */
void
handle_menus(int c);
void
handle_edit(int c);
void
next_line(void);
void
prev_line(void);
void
next_char(void);
void
prev_char(void);
void
insert_char(char c);
void
split_line(void);
void
join_line(int direction);
void
delete_character(void);
void
backspace_character(void);

/*
 * Defines for join_line
 */
#define LINE_NEXT 0
#define LINE_PREV 1

/*
 * Editor event loop, grab characters from the keyboard and either
 * perform menu functions if we are in the top menu (we have seen a
 * single ESC) or perform edit functions otherwise.
 */
void
edit_driver(void)
{
	int c;
	
	for (;;) {
		if (pushback != 0) {
			c = pushback;
			pushback = 0;
			ungetch('\n'); /* this will pop the menu down */
		} else
			c = wgetch((menu_mode)? top_win : edit_win);
		
		if (menu_mode) {
			pos_menu_cursor(top_menu);
			handle_menus(c);
		} else {
			handle_edit(c);
		}
		wrefresh(top_win);
		wrefresh(filler);
		wrefresh(edit_win);
	}
}

/*
 * Handle the menu mode keystrokes.
 */
void
handle_menus(int c)
{
	Menu_Hook cb;
	ITEM *current;

	if (!handle_shortcut(c)) {
		switch (c) {
			case '\e':
				menu_mode = 0;
				break;
		  
			case 'k':
			case KEY_UP:
				beep();
				break;

			case 'j':
			case '\r':
			case '\n':
			case KEY_DOWN:
				if ((current = current_item(top_menu)) != NULL) {
					cb = (Menu_Hook) item_userptr(current);
					if (cb != NULL) {
						cb(top_menu);
						touchwin(edit_win); /* clean up damage */
						refresh_screen();
					}
				}
				break;

			case 'h':
			case KEY_LEFT:
				menu_driver(top_menu, REQ_LEFT_ITEM);
				break;
				
			case 'l':
			case '\t':
			case KEY_RIGHT:
				menu_driver(top_menu, REQ_RIGHT_ITEM);
				break;

			default:
				menu_driver(top_menu, c); /* this makes item search work */
				break;
		}
	}
	
	wmove(edit_win, cursor_line, screen_pos);
}

/*
 * Handle the editing window events.
 */
void
handle_edit(int c)
{
	if (!handle_shortcut(c)) {
		switch (c) {
			case '\e':
				menu_mode = 1;
				break;

			case KEY_DOWN:
				next_line();
				break;

			case KEY_UP:
				prev_line();
				break;

			case KEY_LEFT:
				prev_char();
				break;

			case KEY_RIGHT:
				next_char();
				break;

			case KEY_F(1):
				help_keys_item_callback();
				break;
		  
			default:
				if ((c == '\r') || (c == '\n') || (c == '\t') ||
				    (isprint((char)c))) {
					insert_char((char)c);
				}
				break;
		}
	}
	
	wrefresh(edit_win);
}

/*
 * Check the input character c to see if it is a short cut character,
 * if it is then process the shortcut.  Returns 1 if a short cut was
 * processed, 0 otherwise.
 */
int
handle_shortcut(int c)
{
	int result = 1;

	switch (c) {
	  case '\f':
		  touchwin(edit_win);
		  wrefresh(edit_win);
		  break;

	  case '\000': /* control-space */
		  edit_select_item_callback(NULL);
		  break;

	  case '\002': /* control-b */
	  case KEY_PPAGE:
		  scroll_down(edit_maxy);
		  break;
		  
	  case '\003': /* control-c */
		  edit_copy_item_callback(NULL);
		  break;

	  case '\004': /* control-d */
	  case KEY_DC:
		  delete_character();
		  break;

	  case '\006': /* control-f */
	  case KEY_NPAGE:
		  scroll_up(edit_maxy);
		  break;
		  
	  case '\010': /* control-h */
	  case KEY_BACKSPACE:
		  backspace_character();
		  break;

	  case '\017': /* control-o */
		  file_open_item_callback(NULL);
		  break;
		  
	  case '\020': /* control-p */
		  edit_paste_item_callback(NULL);
		  break;
		  
	  case '\022': /* control-r */
		  search_backward_item_callback(NULL);
		  touchwin(edit_win);
		  break;
		  
	  case '\023': /* control-s */
		  search_forward_item_callback(NULL);
		  touchwin(edit_win);
		  break;

	  case '\026': /* control-v */
		  edit_cut_item_callback(NULL);
		  break;

	  case '\027': /* control-w */
		  file_save_item_callback(NULL);
		  break;
		  
	  case '\030': /* control-x */
		  endwin();
		  exit(0);
		  break;

	  default:
		  result = 0;
		  break;
	}

	return result;
}

/*
 * Move to the next line in the buffer, updating screen as appropriate.
 */
void
next_line(void)
{
	if ((current_line == NULL) || (current_line->next == NULL))
		return;

	mark_count++;
	cursor_line++;
	if (cursor_line > edit_maxy - 1) {
		scroll_up(1); /* this will reset current_line */
		cursor_line = edit_maxy - 1;
	} else
		current_line = current_line->next;

	if (cursor_pos > current_line->length) {
		cursor_pos = current_line->length;
		screen_pos = tab_expanded_length(current_line->length);
	}

	wmove(edit_win, cursor_line, screen_pos);
	wrefresh(edit_win);
}

/*
 * Move to the previous line in the buffer, updating the screen as appropriate.
 */
void
prev_line(void)
{
	if ((current_line == NULL) || (current_line->prev == NULL))
		return;

	mark_count--;
	cursor_line--;
	if (cursor_line < 0) {
		scroll_down(1); /* this will reset current_line */
		cursor_line = 0;
	} else
		current_line = current_line->prev;

	if (cursor_pos > current_line->length) {
		cursor_pos = current_line->length;
		screen_pos = tab_expanded_length(current_line->length);
	}

	wmove(edit_win, cursor_line, screen_pos);
	wrefresh(edit_win);
}

/*
 * Move to the next character position in the buffer.
 */
void
next_char(void)
{
	if (current_line == NULL)
		return;
	
	if (current_line->line[cursor_pos] == '\t') {
		screen_pos = screen_pos - (screen_pos % 8) + 8;
	} else {
		screen_pos++;
	}

	cursor_pos++;
	
	if (cursor_pos > current_line->length) {
		cursor_pos = 0;
		screen_pos = 0;
		next_line();
	}
	wmove(edit_win, cursor_line, screen_pos);
}

/*
 * Move to the previous character position in the buffer.
 */
void
prev_char(void)
{
	if (current_line == NULL)
		return;
	
	cursor_pos--;
	if (cursor_pos < 0) {
		prev_line();
		cursor_pos = current_line->length;
		screen_pos = tab_expanded_length(cursor_pos);
	} else {
		screen_pos--;
		if (current_line->line[cursor_pos] == '\t')
			screen_pos = tab_expanded_length(cursor_pos);
	}
	
	wmove(edit_win, cursor_line, screen_pos);
}

/*
 * Insert the given character into the buffer at the current cursor position.
 */
void
insert_char(char c)
{
	char *new_str;
	
	if ((c == '\n') || (c == '\r')) {
		split_line();
	} else {
		if (current_line == NULL) /* if buffer is empty then add a line */
			insert_line("\n");
		
		if (current_line->alloced < (current_line->length + 1)) {
			if ((new_str = (char *) realloc(current_line->line,
					       (current_line->length + 32)
					       * sizeof(char))) == NULL) {
				fprintf(stderr,
					"Cannot realloc insert space\n");
				return;
			}
			current_line->alloced += 32;
			current_line->line = new_str;
		}
		bcopy(current_line->line + cursor_pos,
		      current_line->line + cursor_pos + 1,
		      current_line->length - cursor_pos + 1);
		current_line->line[cursor_pos] = c;
		current_line->length++;
		cursor_pos++;
		screen_pos = tab_expanded_length(cursor_pos);
		redraw_current_line();
	}

	wrefresh(edit_win);
}

/*
 * Split a line at cursor_pos making two lines.
 */
void
split_line()
{
	EDIT_LINE *old_line;

	if (current_line == NULL) {
		insert_line("");
	} else {
		old_line = current_line;
		insert_line(&current_line->line[cursor_pos]);
		old_line->line[cursor_pos] = '\0';
		old_line->length = strlen(old_line->line);
		if ((current_line->next != NULL) &&
		    (current_line->next->prev != NULL))
			  /* fix up pointer to inserted line */
			current_line->next->prev = current_line;
								    
		cursor_pos = 0;
		screen_pos= 0;
		if ((cursor_line >= edit_maxy - 1) && (top_line->next != NULL))
			top_line = top_line->next;
	}
	
	redraw_screen();
	wrefresh(edit_win);
}

/*
 * Delete the character at cursor_pos.
 */
void
delete_character(void)
{
	if ((current_line->length == 0) ||
	    (current_line->line[cursor_pos] == '\0')) {
		join_line(LINE_NEXT);
	} else {
		bcopy(current_line->line + cursor_pos + 1,
		      current_line->line + cursor_pos,
		      current_line->length - cursor_pos + 1);
		current_line->length--;
		redraw_current_line();
		wrefresh(edit_win);
	}
}

/*
 * Delete the character to the left of cursor_pos.  If we are at the
 * beginning of a line then join the current line to the previous one, if
 * any.
 */
void
backspace_character(void)
{
	if (cursor_pos == 0) {
		join_line(LINE_PREV);
	} else {
		bcopy(current_line->line + cursor_pos,
		      current_line->line + cursor_pos - 1,
		      current_line->length - cursor_pos + 1);
		cursor_pos--;
		screen_pos = tab_expanded_length(cursor_pos);
		current_line->length--;
		redraw_current_line();
		wrefresh(edit_win);
	}
}

/*
 * Join a the current line to either the next or previous line depending
 * on the direction indicated by the direction parameter.
 */
void
join_line(int direction)
{
	char *new_line;
	EDIT_LINE *del_line;
	int new_len;
	
	if (direction == LINE_NEXT) {
		if (current_line->next == NULL)
			return;

		new_len = current_line->length + current_line->next->length;
		
		if (new_len > current_line->alloced) {
			if ((new_line = (char *)
			     realloc(current_line->line,
				     (new_len + 1) * sizeof(char))) == NULL) {
				fprintf(stderr,
					"Cannot realloc joined line\n");
				return;
			}
			current_line->line = new_line;
			current_line->alloced = new_len;
		}
		bcopy(current_line->next->line,
		      current_line->line + current_line->length,
		      current_line->next->length + 1);
		del_line = current_line->next;
		current_line->next = current_line->next->next;
		current_line->length = new_len;
		if (current_line->next != NULL)
			current_line->next->prev = current_line;
		free_line(del_line);
	} else {
		if (current_line->prev == NULL)
			return;

		cursor_pos = current_line->prev->length;
		screen_pos = tab_expanded_length(cursor_pos);
		new_len = current_line->length + current_line->prev->length;
		if (new_len > current_line->prev->alloced) {
			if ((new_line = (char *)
			     realloc(current_line->prev->line,
				     (new_len + 1) * sizeof(char))) == NULL) {
				fprintf(stderr,
					"Could not realloc joined line 2\n");
				return;
			}

			current_line->prev->line = new_line;
			current_line->prev->alloced = new_len;
		}
		bcopy(current_line->line,
		      current_line->prev->line + current_line->prev->length,
		      current_line->length + 1);
		del_line = current_line;
		current_line->prev->next = current_line->next;
		if (current_line->next != NULL)
			current_line->next->prev = current_line->prev;
		current_line = current_line->prev;
		current_line->length = new_len;
		free_line(del_line);
		if (cursor_line <= 0)
			top_line = current_line;
	}
	redraw_screen();
}


	
	
	
