/*	$NetBSD: edit.c,v 1.2 2000/05/06 11:44:34 blymn Exp $	*/

/*-
 * Copyright (c) 1998-1999 Brett Lymn
 *                         (blymn@baea.com.au, brett_lymn@yahoo.com.au)
 * All rights reserved.
 *
 * This code has been donated to The NetBSD Foundation by the Author.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software withough specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 */

#include <stdlib.h>
#include <menu.h>
#include "edit.h"
#include "driver.h"
#include "file.h"
#include "help.h"
#include "search.h"
#include "utils.h"
#include "buffer.h"

MENU *top_menu, *edit_menu;
ITEM *top_items[7], *edit_items[5];

int col_width, mark_count, cursor_pos, screen_pos, cursor_line, edit_maxy;
int marked = 0, pushback = 0;
WINDOW *screen, *top_win, *edit_win, *pop_up, *filler;

EDIT_LINE *first_line = NULL, *current_line = NULL, *select_line = NULL,
	*buffer = NULL, *top_line = NULL;

/* callbacks for menu items */
void
dummy_item_callback(MENU *parent);
void
edit_copy_item_callback(MENU *parent);
void
edit_cut_item_callback(MENU *parent);
void
edit_paste_item_callback(MENU *parent);
void
edit_select_item_callback(MENU *parent);

/*
 * Initialise the items for the edit menu.
 */
void
edit_item_init(void)
{
	if ((edit_items[0] = new_item("Copy", "Ctrl-C")) == NULL) {
                error_exit("Edit copy item failed");
        }
	if ((edit_items[1] = new_item("Cut", "Ctrl-V")) == NULL) {
                error_exit("Edit cut item failed");
        }
	if ((edit_items[2] = new_item("Paste", "Ctrl-P")) == NULL) {
                error_exit("Edit paste item failed");
        }
	if ((edit_items[3] = new_item("Select", "Ctrl-Space")) == NULL) {
                error_exit("Edit select item failed");
        }

	set_item_userptr(edit_items[0], (char *) &edit_copy_item_callback);
	set_item_userptr(edit_items[1], (char *) &edit_cut_item_callback);
	set_item_userptr(edit_items[2], (char *) &edit_paste_item_callback);
	set_item_userptr(edit_items[3], (char *) &edit_select_item_callback);
	
	edit_items[4] = NULL;

	if ((edit_menu = new_menu(edit_items)) == NULL)
		error_exit("Failed to create edit menu\n");

	set_menu_format(edit_menu, 4, 1);
	menu_opts_on(edit_menu, O_SHOWDESC);
}

/*
 * Copy the lines from the edit buffer into the cut buffer.
 */
void
edit_copy_item_callback(MENU *parent)
{
	copy_region();
	marked = 0;
}

/*
 * Remove the lines from the edit buffer and place them in the cut buffer.
 */
void
edit_cut_item_callback(MENU *parent)
{
	copy_region();
	delete_region();
	marked = 0;
}

/*
 * Copy the lines from the cut buffer into the edit buffer.
 */
void
edit_paste_item_callback(MENU *parent)
{
	insert_buffer();
}

/*
 * Mark the start of a region that will/may be cut or copied later.
 */
void
edit_select_item_callback(MENU *parent)
{
	select_line = current_line;
	marked = 1;
	mark_count = 0;
}

/*
 * Do the edit menu thang
 */
void
edit_item_callback(MENU *parent)
{
	WINDOW *edit_popup;

	if ((edit_popup = create_menu_popup(parent, edit_menu, 1, -1, -1, -1))
	    == NULL) {
		fprintf(stderr, "Cannot create edit menu popup\n");
		return;
	}

	handle_menu_popup(edit_menu, edit_popup, "edit");
	
}


int
main(int argc, char **argv)
{
	int top_rows, top_cols, status;
	char message[80];
	
	  /* firstly, set the terminal environment if not done already */
	setenv("TERM", "vt100", 0);

	screen = initscr();
	if ((top_items[0] = new_item("File", "File Menu")) == NULL) {
                error_exit("File item failed");
        }

	if ((top_items[1] = new_item("Edit", "Edit Menu")) == NULL) {
                error_exit("Edit item failed");
        }

	if ((top_items[2] = new_item("Search", "Search Menu")) == NULL) {
                error_exit("Search item failed");
        }

	if ((top_items[3] = new_item("Help", "")) == NULL) {
                error_exit("Help item failed");
        }

	top_items[4] = NULL;

	set_item_userptr(top_items[0], (char *) &file_item_callback);
	set_item_userptr(top_items[1], (char *) &edit_item_callback);
	set_item_userptr(top_items[2], (char *) &search_item_callback);
	set_item_userptr(top_items[3], (char *) &help_item_callback);

	  /* set default menu attributes - all menus will inherit these */
	set_menu_back(NULL, A_REVERSE);
	set_menu_fore(NULL, A_NORMAL);
	/* ignore case and show matches */
	menu_opts_on(NULL, O_IGNORECASE | O_SHOWMATCH); 
	  /* done */
	
	file_item_init();
	edit_item_init();
	search_item_init();
	help_item_init();

	top_menu = new_menu(top_items);
	if (top_menu == NULL) {
		error_exit("Failed to create top menu");
	}
	
	menu_opts_off(top_menu, O_ROWMAJOR);
	set_menu_format(top_menu, 1, 4);

	scale_menu(top_menu, &top_rows, &top_cols);
	
	col_width = top_cols / 4;
	
	if ((top_win = newwin(1, top_cols, 0,0)) == NULL) {
		error_exit("Failed to create menu win");
	}

	if ((filler = newwin(1, COLS - top_cols, 0, top_cols - 1)) == NULL) {
		error_exit("Failed to create filler window");
	}
	
	if ((edit_win = newwin(LINES - 1, COLS, 1, 0)) == NULL) {
		error_exit("Failed to create edit win");
	}
	
  	set_menu_sub(top_menu, top_win);
	if ((status = post_menu(top_menu)) < 0) {
		sprintf(message, "Post of top_menu failed with status %d",
			status);
		error_exit(message);
	}

	wattron(filler, A_REVERSE);
	whline(filler, ' ', COLS - top_cols);
	wrefresh(filler);
	wrefresh(top_win);
	wclear(edit_win);
	
        keypad(top_win, TRUE);
        nodelay(top_win, FALSE);
        keypad(edit_win, TRUE);
        nodelay(edit_win, FALSE);
	scrollok(edit_win, TRUE);
	edit_maxy = getmaxy(edit_win);
        raw();
        noecho();

	if (argv[1] != NULL)
		read_file(argv[1]);
	wrefresh(edit_win);

	edit_driver();

	clear();
	refresh();
	endwin();
	exit(0);
}
