/*	$NetBSD: file.c,v 1.1 2000/05/04 13:58:52 blymn Exp $	*/

/*-
 * Copyright (c) 1998-1999 Brett Lymn
 *                         (blymn@baea.com.au, brett_lymn@yahoo.com.au)
 * All rights reserved.
 *
 * This code has been donated to The NetBSD Foundation by the Author.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software withough specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <malloc.h>
#include <menu.h>
#include <string.h>
#include <unistd.h>
#include "edit.h"
#include "file.h"
#include "screen.h"
#include "utils.h"

static ITEM *file_items[5];
static MENU *file_menu;
static char file_open[MAXPATHLEN] = ""; /* name of file that is open, if any */

/* defined in edit.c */
extern int col_width, cursor_pos, screen_pos;
extern EDIT_LINE *first_line, *current_line, *top_line;

/* menu callbacks */
void
file_open_item_callback(MENU *parent);
void
file_save_item_callback(MENU *parent);
void
file_save_as_item_callback(MENU *parent);
void
file_quit_item_callback(MENU *parent);

/* prototypes. */
void
file_item_init(void);
int
item_sort(const void *a, const void *b);
ITEM **
load_file_items(int *item_count);
void
destroy_items(ITEM **items, int count);
WINDOW *
create_file_popup(MENU *menu);

/*
 * Initialise the file menu structures.
 */
void
file_item_init(void)
{
	if ((file_items[0] = new_item("Open", "Ctrl-O")) == NULL) {
                error_exit("File Open item failed");
        }
	if ((file_items[1] = new_item("Save", "Ctrl-W")) == NULL) {
                error_exit("File Save item failed");
        }
	if ((file_items[2] = new_item("Save As...", "      ")) == NULL) {
                error_exit("File Save As item failed");
        }
	if ((file_items[3] = new_item("Quit", "Ctrl-X")) == NULL) {
                error_exit("File Quit item failed");
        }

	file_items[4] = NULL;

	set_item_userptr(file_items[0], (char *) &file_open_item_callback);
	set_item_userptr(file_items[1], (char *) &file_save_item_callback);
	set_item_userptr(file_items[2], (char *) &file_save_as_item_callback);
	set_item_userptr(file_items[3], (char *) &file_quit_item_callback);
	
	if ((file_menu = new_menu(file_items)) == NULL) {
		error_exit("Failed to create file menu");
	}

	set_menu_format(file_menu, 4, 1);
	menu_opts_on(file_menu, O_SHOWDESC);
}

/*
 * Callback for the file menu, pops up the file menu and handles the
 * selections.
 */
void
file_item_callback(MENU *parent)
{
	WINDOW *file_popup;

	if ((file_popup = create_menu_popup(parent, file_menu, 1, -1, -1, -1))
	    == NULL) {
		fprintf(stderr, "File menu popup failed.\n");
		return;
	}
	
	handle_menu_popup(file_menu, file_popup, "file");
}

void
read_file(filename)
	char *filename;
{
	FILE *handle;
	char line[256];

	if ((handle = fopen(filename, "r")) == NULL) {
		fprintf(stderr, "Open of file %s failed\n", filename);
		return;
	}

	strcpy(file_open, filename); /* save name of file we have opened */
	
	if (first_line != NULL) /* already a file in the buffer - clear it */
		destroy_all_lines();
	
	while (!feof(handle)) {
		if (fgets(line, 256, handle) != NULL) {
			insert_line(line);
		}
	}

	fclose(handle);
	current_line = first_line;
	top_line = first_line;
	cursor_pos = 0;
	screen_pos = 0;
	redraw_screen();
}

/*
 * Do the work necessary to create a popup window to hold the given
 * menu.  Size the window to fit the menu and allow 10 rows of items to
 * be displayed in a centred popup window.  Return NULL if anything failed.
 */
WINDOW *
create_file_popup(MENU *menu)
{
	int rows, cols, start_row, start_col;
	WINDOW *new;
	
	scale_menu(menu, &rows, &cols);

	if (cols > COLS) {
		fprintf(stderr, "file open menu too wide for screen\n");
		return NULL;
	}
	
	start_row = (LINES / 2) - 5;
	if (start_row < 0)
		start_row = 0;

	start_col = (COLS - cols) / 2;
	
	if ((new = newwin(10, cols, start_row, start_col)) == NULL) {
		fprintf(stderr, "failed to create open window\n");
		return NULL;
	}

	keypad(new, TRUE);
	nodelay(new, FALSE);
	
	set_menu_format(menu, 10, 1);
	set_menu_sub(menu, new);

	return new;
}

/*
 * Destroy the given array of items and free the allocated array space.
 */
void
destroy_items(ITEM **items, int count)
{
	int i;
	
	for (i = 0; i < count; i++)
		free_item(items[i]);
	free(items);
}

/*
 * Comparison function for item sorting.
 */
int
item_sort(const void *a, const void *b)
{
	return strcmp(item_name(*(ITEM **)a), item_name(*(ITEM **)b));
}

/*
 * Allocate and load up an array of items for the open menu popup by
 * scanning the current directory and converting them to menu items.
 */
ITEM **
load_file_items(int *item_count)
{
	int i = 0;
	DIR *dirp;
	struct dirent *dp;
	ITEM **items;

	(*item_count) = 0;
	if ((dirp = opendir(".")) == NULL) {
		fprintf(stderr, "Could not open directory for read\n");
		return NULL;
	}

	  /* count the directory items */
	while ((dp = readdir(dirp)) != NULL) {
		if (!((dp->d_name[0] == '.') && (dp->d_name[1] == '\0')))
			(*item_count)++;
	}

	  /* allocate memory for the items array based on the number of
	     directory entries */
	if ((items =
	     (ITEM **) malloc((*item_count + 1) * sizeof (ITEM *))) == NULL) {
		fprintf(stderr,
			"Could not malloc open file popup items array\n");
		closedir(dirp);
		return NULL;
	}

	rewinddir(dirp);

	  /* create items from directory entries */
	while (((dp = readdir(dirp)) != NULL) && (i <= *item_count)) {
		if (!((dp->d_name[0] == '.') && (dp->d_name[1] == '\0'))) {
			if ((items[i++] = new_item(dp->d_name, "")) == NULL)
				fprintf(stderr,
					"Failed to add %s to file items\n",
					dp->d_name);
		}
	}

	items[*item_count] = NULL;

	(void)closedir(dirp);

	  /* nicely sort the items */
	qsort(items, *item_count, sizeof(ITEM *), &item_sort);
	return items;
}

/*
 * Call back to handle the open menu function.
 */
void
file_open_item_callback(MENU *parent)
{
	ITEM **open_items, **new_items;
	MENU *open_menu;
	WINDOW *open_window;
	struct stat stat_buffer;
	char *name;
	int item_count, done, c, status;
	int new_item_count;

	if ((open_items = load_file_items(&item_count)) == NULL) {
		fprintf(stderr, "load of directory items failed\n");
		return;
	}
	
	if ((open_menu = new_menu(open_items)) == NULL) {
		fprintf(stderr, "failed to create file open menu\n");
		goto the_exit;
	}

	if ((open_window = create_file_popup(open_menu)) == NULL) {
		fprintf(stderr, "failed to create popup window\n");
		goto the_exit;
	}
	
	menu_opts_on(open_menu, O_ROWMAJOR);
	menu_opts_off(open_menu, O_IGNORECASE); /* want filename search case
						   sensitive */
	
	if ((status = post_menu(open_menu)) < 0) {
		fprintf(stderr, "failed to post open menu (%d)\n", status);
		goto the_exit;
	}
	
	wrefresh(open_window);

	done = 0;
	do {
		c = wgetch(open_window);
		switch (c) {
		  case '\e':
			  done = 1;
			  break;

		  case '\t':
		  case KEY_DOWN:
			  menu_driver(open_menu, REQ_DOWN_ITEM);
			  break;

		  case KEY_UP:
			  menu_driver(open_menu, REQ_UP_ITEM);
			  break;

		  case '\n':
		  case '\r':
			  name = item_name(current_item(open_menu));
			  if (stat(name, &stat_buffer) < 0) {
				  fprintf(stderr, "Cannot stat %s\n",
					  name);
				  break;
			  }
			  
			  if ((stat_buffer.st_mode & S_IFDIR) == S_IFDIR) {
				  if (chdir(name) < 0) {
					  fprintf(stderr,
						  "Could not chdir to %s\n",
						  name);
					  break;
				  }

				  if ((new_items =
				       load_file_items(&new_item_count))
				      == NULL) {
					  fprintf(stderr,
						  "Cannot create new items list\n");
					  break;
				  }
					  
				  unpost_menu(open_menu);
				  if (set_menu_items(open_menu,
						     new_items) < 0) {
					  fprintf(stderr,
						  "attach of new items failed\n");
					  destroy_items(new_items,
							new_item_count);
					  post_menu(open_menu);
					  break;
				  }
				  
				  destroy_items(open_items, item_count);
				  open_items = new_items;
				  item_count = new_item_count;
				  if ((open_window =
				       create_file_popup(open_menu)) == NULL) {
					  fprintf(stderr,
						  "Could not recreate window\n");
					  goto the_exit;
				  }
				  
				  if (post_menu(open_menu) < 0) {
					  fprintf(stderr,
						  "Repost of menu failed\n");
					  goto the_exit;
				  }
			  } else if ((stat_buffer.st_mode
				      & (S_IFREG | S_IFLNK)) != 0) {
				  read_file(name);
				  done = 1;
			  } else {
				  fprintf(stderr,
					  "Cannot open that sort of file\n");
			  }
			  break;

		  default:
			  menu_driver(open_menu, c);
			  break;
		}
		wrefresh(open_window);
	} while (done == 0);

	unpost_menu(open_menu);
	
  the_exit:
	destroy_items(open_items, item_count);
	return;
}

/*
 * Call back to handle the saving of files, save the edit buffer in a file
 * named by file_open.
 */
void
file_save_item_callback(MENU *parent)
{
	FILE *handle;
	EDIT_LINE *line;

	if (file_open[0] == '\0') { /* no file open */
		fprintf(stderr, "No file has been opened, use \"save as\"\n");
		return;
	}

	if ((handle = fopen(file_open, "w")) == NULL) {
		fprintf(stderr, "Cannot open %s for writing\n", file_open);
		return;
	}

	for (line = first_line; line != NULL; line = line->next)
		fprintf(handle, "%s\n", line->line);

	fclose(handle);
	return;
}

/*
 * Call back to implement the save as menu item.
 */
void
file_save_as_item_callback(MENU *parent)
{
	WINDOW *prompt_window;
	char name[MAXPATHLEN];

	if ((prompt_window = newwin(1, 30, LINES / 2, (COLS - 30) / 2))
	    == NULL) {
		fprintf(stderr, "Cannot create file prompt window\n");
		return;
	}

	wattron(prompt_window, A_REVERSE);
	wmove(prompt_window, 0, 0);
	wprintw(prompt_window, "Filename:                     ");
	wmove(prompt_window, 0, 11);
	wrefresh(prompt_window);
	noraw();
	echo();
	wscanw(prompt_window, "%s", name);
	raw();
	noecho();
	delwin(prompt_window);
	
	if (name[0] != '\0') {
		strcpy(file_open, name);
		file_save_item_callback(parent);
	}

	return;
}

/*
 * Call back to handle quitting the editor.
 */
void
file_quit_item_callback(MENU *parent)
{
	clear();
	refresh();
	endwin();
	exit(0);
}
