/*	$NetBSD$	*/

/*-
 * Copyright (c) 1998-1999 Brett Lymn
 *                         (blymn@baea.com.au, brett_lymn@yahoo.com.au)
 * All rights reserved.
 *
 * This code has been donated to The NetBSD Foundation by the Author.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software withough specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 */
#include <malloc.h>
#include <menu.h>
#include <stdio.h>
#include <string.h>
#include "edit.h"
#include "screen.h"
#include "search.h"
#include "utils.h"

ITEM *search_items[5], *search_popup_items[3], *replace_popup_items[5];
MENU *search_menu, *search_popup_menu, *replace_popup_menu;
WINDOW *find_popup, *replace_popup;
int find_active;

/* externs defined in edit.c */
extern EDIT_LINE *current_line;
extern int cursor_pos, mark_count;
extern WINDOW *edit_win;


/* menu callbacks */
void
replace_forward_item_callback(MENU *parent);
void
replace_backward_item_callback(MENU *parent);
void
popup_find_next_callback(MENU *parent);
void
destroy_search_popup(MENU *parent);
void
replace_item_callback(MENU *parent);
void
replace_all_item_callback(MENU *parent);
void
replace_skip_item_callback(MENU *parent);
void
replace_destroy_popup(MENU *parent);

/* The following are to control the direction match_string works in. */
#define E_SEARCH_FORWARD  0
#define E_SEARCH_BACKWARD 1

/* prototypes */
char *
match_char(char *origin, char *start, char c);
char *
match_string(EDIT_LINE **this, const char *string, unsigned char direction);
void
handle_search_popup(void);
void
handle_replace_popup(void);
EDIT_LINE *
perform_find(void);
void
replace(void);
void
search_item_init(void);

/* globals used to implement the search and replace functions. */
char search_string[251];
char replace_string[251];
unsigned char search_direction, match_next;

/*
 * Replace the string starting at cursor_pos with the replacement string.
 * This function assumes that a search has been performed and the current
 * cursor position and line point at the search string, search string
 * length characters will be removed from the buffer and the new string
 * inserted.
 */
void
replace(void)
{
	int s_len, r_len, new_len;
	char *new_line;

	s_len = strlen(search_string);
	r_len = strlen(replace_string);
	new_len = current_line->length + r_len - s_len;
	
	if ((r_len > s_len) && (current_line->alloced < new_len)) {
		if ((new_line = (char *) realloc(current_line->line,
				   1 + sizeof(char) * new_len)) == NULL) {
			fprintf(stderr,
				"Cannot malloc new line for replace\n");
			return;
		}
		current_line->line = new_line;
		current_line->alloced = new_len;
	}

	bcopy(current_line->line + cursor_pos + s_len,
	      current_line->line + cursor_pos + r_len,
	      current_line->length - cursor_pos - s_len + 1);
	bcopy(replace_string, current_line->line + cursor_pos, r_len);
	current_line->length = new_len;
}

/*
 * Attempt to match the given char in the string working from the
 * starting pointer backwards towards the origin of the string.
 * Return a pointer to the matching character in the string, if there
 * is no match then return NULL.
 */
char *
match_char(char *origin, char *start, char c)
{
	char *tp;

	for (tp = start; tp != origin; tp--) {
		if (*tp == c)
			return tp;
	}

	if (*origin == c) /* lucky last, check if first char is the one */
		return origin;
	
	return NULL;
}


/*
 * Match the given string starting at current_line and traversing the
 * lines in the given direction.  If a match is found set the this pointer
 * to the line that matched and return the pointer to the position in the
 * string where the match occurred.  If the match fails then return NULL.
 * The contents of the this pointer when a match fails is undefined.
 */
char *
match_string(EDIT_LINE **this, const char *string, unsigned char direction)
{
	int match_len, line_count;
	char *match_loc, *char_loc, *match_start;
	EDIT_LINE *line;

	match_len = strlen(string);
	if (match_len == 0) /* don't try to match null string. */
		return NULL;

	line_count = 0;
	for (line = current_line; line != NULL;
	     line = (direction == E_SEARCH_FORWARD)? line->next : line->prev,
		     line_count++) {
		if (line->length < match_len) /* don't bother if line short */
			continue;

		if ((line == current_line) && (match_next == 1)) {
			if (direction == E_SEARCH_FORWARD) {
				if ((cursor_pos + match_len) > line->length)
					continue;
				match_start = line->line + cursor_pos
					+ match_len;
			} else {
				if ((cursor_pos - match_len) < 0)
					continue;
				match_start = line->line + cursor_pos
					- match_len;
			}
		} else {
			match_next = 0;
			match_start = line->line +
				((direction == E_SEARCH_FORWARD)? 0 :
				 line->length);
		}

		    
		if (direction == E_SEARCH_FORWARD) {
			if ((match_loc = strstr(match_start, string))
			    != NULL) {
				*this = line;
				match_next = 1;
				mark_count += line_count;
				return match_loc;
			}
		} else {
		  rematch:
			if ((char_loc = match_char(line->line, match_start,
						    (int) string[0])) != NULL) {
				if ((match_loc = strstr(char_loc, string))
				    != NULL) {
					*this = line;
					match_next = 1;
					mark_count -= line_count;
					return match_loc;
				} else {
					if (char_loc == line->line)
						continue;
					match_start = char_loc - 1;
					goto rematch;
				}
			}
			
		}
	}
	
	return NULL;
}

/*
 * Initialise the search item menu ready for the search_item_callback
 * function to popup.
 */

void
search_item_init(void)
{
	if ((search_items[0] = new_item("Search Forward", "Ctrl-S")) == NULL) {
                error_exit("Search Forward item failed");
        }
	if ((search_items[1] =
	     new_item("Search Backward", "Ctrl-R")) == NULL) {
                error_exit("Search Backward item failed");
        }
	if ((search_items[2] = new_item("Replace Forward", "")) == NULL) {
                error_exit("Replace Forward item failed");
        }
	if ((search_items[3] = new_item("Replace Backward", "")) == NULL) {
                error_exit("Replace Backward item failed");
        }

	search_items[4] = NULL;

	set_item_userptr(search_items[0],
			 (char *) search_forward_item_callback);
	set_item_userptr(search_items[1],
			 (char *) search_backward_item_callback);
	set_item_userptr(search_items[2],
			 (char *) replace_forward_item_callback);
	set_item_userptr(search_items[3],
			 (char *) replace_backward_item_callback);

	if ((search_menu = new_menu(search_items)) == NULL) {
		error_exit("Failed to create search menu");
	}

	set_menu_format(search_menu, 4, 1);
	menu_opts_on(search_menu, O_SHOWDESC);

	  /* init the search popup menu */
	if ((search_popup_items[0] = new_item("Find Next", "")) == NULL) {
                error_exit("Find Next item failed");
        }
	if ((search_popup_items[1] = new_item("Done", "")) == NULL) {
                error_exit("Find Done item failed");
        }

	search_popup_items[2] = NULL;

	set_item_userptr(search_popup_items[0],
			 (char *) popup_find_next_callback);
	set_item_userptr(search_popup_items[1],
			 (char *) destroy_search_popup);
	
	if ((search_popup_menu = new_menu(search_popup_items)) == NULL) {
		error_exit("Failed to create search popup menu");
	}

	set_menu_format(search_popup_menu, 1, 2);
	menu_opts_off(search_popup_menu, O_SHOWDESC);

	  /* init the replace popup menu */
	if ((replace_popup_items[0] = new_item("Replace", "")) == NULL) {
                error_exit("Replace item failed");
        }
	if ((replace_popup_items[1] = new_item("Replace All", "")) == NULL) {
                error_exit("Replace All item failed");
        }
	if ((replace_popup_items[2] = new_item("Skip", "")) == NULL) {
                error_exit("Replace item failed");
        }
	if ((replace_popup_items[3] = new_item("Done", "")) == NULL) {
                error_exit("Replace item failed");
        }

	replace_popup_items[4] = NULL;

	set_item_userptr(replace_popup_items[0],
			 (char *) replace_item_callback);
	set_item_userptr(replace_popup_items[1],
			 (char *) replace_all_item_callback);
	set_item_userptr(replace_popup_items[2],
			 (char *) replace_skip_item_callback);
	set_item_userptr(replace_popup_items[3],
			 (char *) replace_destroy_popup);
	
	if ((replace_popup_menu = new_menu(replace_popup_items)) == NULL) {
		error_exit("Failed to create replace popup menu");
	}

	set_menu_format(replace_popup_menu, 1, 4);
	menu_opts_off(replace_popup_menu, O_SHOWDESC);

	
}

/*
 * Popup the search menu and handle the items selected therein.
 */
void
search_item_callback(MENU *parent)
{
	WINDOW *search_popup;

	if ((search_popup = create_menu_popup(parent, search_menu, 1, -1,
					      -1, -1)) == NULL) {
		fprintf(stderr, "Search menu popup failed.\n");
		return;
	}

	handle_menu_popup(search_menu, search_popup, "search");
}

/*
 * Search forward for a string.
 */
void
search_forward_item_callback(MENU *parent)
{
	search_direction = E_SEARCH_FORWARD;
	handle_search_popup();
}

/*
 * Search backward for a string
 */
void
search_backward_item_callback(MENU *parent)
{
	search_direction = E_SEARCH_BACKWARD;
	handle_search_popup();
}

/*
 * Replace strings forward.
 */
void
replace_forward_item_callback(MENU *parent)
{
	search_direction = E_SEARCH_FORWARD;
	handle_replace_popup();
}

/*
 * Replace strings backward.
 */
void
replace_backward_item_callback(MENU *parent)
{
	search_direction = E_SEARCH_BACKWARD;
	handle_replace_popup();
}

/*
 * Popup the search window which contains a prompt for the search string
 * and a small menu of functions to do with the search.
 */
void
handle_search_popup(void)
{
	WINDOW *local_win;
	ITEM *current;
	EDIT_LINE *this;
	Menu_Hook cb;
	char *mp;
	int status, c, done, sl, sc;

	sl = LINES / 2 - 3;
	sc = COLS /2 - 10;
	if (sl < 0)
		sl = 0;
	if (sc < 0)
		sc = 0;
	
	find_popup = create_menu_popup(NULL, search_popup_menu, sl,
				       sc, 7, 24);

	if (find_popup == NULL) {
		fprintf(stderr, "Cannot create find popup\n");
		return;
	}

	keypad(find_popup, FALSE);
	nodelay(find_popup, FALSE);
	
	box(find_popup, '*', '*');
	wmove(find_popup, 2,2);
	wprintw(find_popup, "Search for: ");
	if ((local_win = derwin(find_popup, 1, 22, 4, 1)) == NULL) {
		fprintf(stderr, "Cannot create local menu\n");
		return;
	}

	keypad(local_win, TRUE);
	nodelay(local_win, FALSE);
	set_menu_sub(search_popup_menu, local_win);
	
	if ((status = post_menu(search_popup_menu)) < 0) {
		fprintf(stderr, "Post of local find menu failed (%d)\n",
			status);
		return;
	}

	wrefresh(find_popup);
	wrefresh(local_win);

	find_active = 1;
	
	do {
		wmove(find_popup, 2, 14);
		wrefresh(find_popup);
		noraw();
		echo();
		wgetstr(find_popup, search_string);
		noecho();
		raw();
		mp = match_string(&this, search_string, search_direction);
		if (mp == NULL)
			break;
		current_line = this;
		show_match(this, mp);
		touchwin(find_popup);
		touchwin(local_win);
		wrefresh(find_popup);
		wrefresh(local_win);
		
		done = 0;
		do {
			c = wgetch(local_win);
			switch (c) {
			  case '\t':
			  case KEY_RIGHT:
			  case 'l':
				  menu_driver(search_popup_menu,
					      REQ_RIGHT_ITEM);
				  break;

			  case KEY_LEFT:
			  case 'h':
				  menu_driver(search_popup_menu,
					      REQ_LEFT_ITEM);
				  break;
				  
			  case '\r':
			  case '\n':
				  if ((current =
				       current_item(search_popup_menu))
				      != NULL) {
					  cb = (Menu_Hook) item_userptr(current);
					  cb(search_popup_menu);
				  }
				  break;
				  
			  case '\e':
				  done = 1;
				  break;
			}

			if (find_active == 1) {
				touchwin(find_popup);
				touchwin(local_win);
				wrefresh(find_popup);
				wrefresh(local_win);
			}
		} while ((done == 0) && (find_active == 1));
	} while (find_active == 1);
	unpost_menu(search_popup_menu);
	delwin(find_popup);
}

/*
 * Popup the replace window which has the search and replace string
 * prompts and a small menu to perform the search and replace functions.
 */
void
handle_replace_popup(void)
{
	WINDOW *local_win;
	ITEM *current;
	EDIT_LINE *this;
	Menu_Hook cb;
	int status, c, done, sl, sc;

	sl = LINES / 2 - 4;
	sc = COLS /2 - 25;
	if (sl < 0)
		sl = 0;
	if (sc < 0)
		sc = 0;
	
	replace_popup = create_menu_popup(NULL, replace_popup_menu, sl,
				       sc, 9, 50);

	if (replace_popup == NULL) {
		fprintf(stderr, "Cannot create replace popup\n");
		return;
	}

	keypad(replace_popup, FALSE);
	nodelay(replace_popup, FALSE);
	
	box(replace_popup, '*', '*');
	wmove(replace_popup, 2,2);
	wprintw(replace_popup, "Search for: ");
	wmove(replace_popup, 4, 2);
	wprintw(replace_popup, "Replace with: ");
	
	if ((local_win = derwin(replace_popup, 1, 48, 6, 1))
	    == NULL) {
		fprintf(stderr, "Cannot create replace local win\n");
		return;
	}

	keypad(local_win, TRUE);
	nodelay(local_win, FALSE);
	set_menu_sub(replace_popup_menu, local_win);
	
	if ((status = post_menu(replace_popup_menu)) < 0) {
		fprintf(stderr, "Post of local find menu failed (%d)\n",
			status);
		return;
	}

	wrefresh(replace_popup);
	wrefresh(local_win);

	find_active = 1;
	
	do {
		wmove(replace_popup, 2, 14);
		wrefresh(replace_popup);
		noraw();
		echo();
		wgetstr(replace_popup, search_string);
		wmove(replace_popup, 4, 15);

		wrefresh(replace_popup);
		wgetstr(replace_popup, replace_string);
		noecho();
		raw();
		if ((this = perform_find()) == NULL)
			break;
		touchwin(replace_popup);
		touchwin(local_win);
		wrefresh(replace_popup);
		wrefresh(local_win);
		
		done = 0;
		do {
			c = wgetch(local_win);
			switch (c) {
			  case '\t':
			  case KEY_RIGHT:
			  case 'l':
				  menu_driver(replace_popup_menu,
					      REQ_RIGHT_ITEM);
				  break;

			  case KEY_LEFT:
			  case 'h':
				  menu_driver(replace_popup_menu,
					      REQ_LEFT_ITEM);
				  break;
				  
			  case '\r':
			  case '\n':
				  if ((current =
				       current_item(replace_popup_menu))
				      != NULL) {
					  cb = (Menu_Hook) item_userptr(current);
					  cb(replace_popup_menu);
				  }
				  break;
				  
			  case '\e':
				  done = 1;
				  break;
			}

			if (find_active == 1) {
				touchwin(replace_popup);
				touchwin(local_win);
				wrefresh(replace_popup);
				wrefresh(local_win);
			}
		} while ((done == 0) && (find_active == 1));
	} while (find_active == 1);
	unpost_menu(replace_popup_menu);
	delwin(replace_popup);
}

/*
 * Find the next instance of the string, set the current line to it
 * and refresh the screen to show the match.  Returns a pointer to the
 * line matched or NULL if the match fails.
 */
EDIT_LINE *
perform_find(void)
{
	EDIT_LINE *this;
	char *mp;
	
	mp = match_string(&this, search_string, search_direction);
	if (mp == NULL)
		return NULL;
	current_line = this;
	show_match(this, mp);
	return this;
}

/*
 * Search for the next occurrence of the find string in the buffer.
 */
void
popup_find_next_callback(MENU *parent)
{
	perform_find();
}

/*
 * Destroy the find popup, this is called when the done button is pressed.
 */
void
destroy_search_popup(MENU *parent)
{
	find_active = 0;
}

/*
 * Replace the currently matched string with the replacement string in
 * the buffer.
 */
void
replace_item_callback(MENU *parent)
{
	EDIT_LINE *this;
	
	replace();
	redraw_current_line();
	wrefresh(edit_win);
	if ((this = perform_find()) == NULL)
		return;
}

/*
 * Replace all occurrences of the search string with the replacement string
 * in the buffer.
 */
void
replace_all_item_callback(MENU *parent)
{
	EDIT_LINE *this;

	replace();
	while ((this = perform_find()) != NULL) {
		replace();
	}
	redraw_current_line();
	wrefresh(edit_win);
}

/*
 * Skip to the next matching search string without replacing.
 */
void
replace_skip_item_callback(MENU *parent)
{
	perform_find();
}

/*
 * Close the replace window.
 */
void
replace_destroy_popup(MENU *parent)
{
	find_active = 0;
}
