/*	$NetBSD: utils.c,v 1.2 2000/05/04 14:11:55 simonb Exp $	*/

/*-
 * Copyright (c) 1998-1999 Brett Lymn
 *                         (blymn@baea.com.au, brett_lymn@yahoo.com.au)
 * All rights reserved.
 *
 * This code has been donated to The NetBSD Foundation by the Author.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software withough specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 */

#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <curses.h>
#include "edit.h"
#include "driver.h"
#include "utils.h"

/* these are defined in edit.c */
extern WINDOW *edit_win, *filler, *top_win;
extern EDIT_LINE *first_line, *current_line, *top_line;
extern int col_width, cursor_line, cursor_pos, screen_pos, pushback;

/* these are defined in driver.c */
extern int menu_mode;

/*
 * End curses mode, print an error message to stderr and then exit.
 */
void
error_exit(char *message)
{
	clear();
	endwin();
	fprintf(stderr, "%s\n", message);
	exit(1);
}

/*
 * Insert a line after the current one and set the contents of the line
 * to that passed.
 */
void
insert_line(char *line)
{
	EDIT_LINE *new_line;
	int len;
	
	if ((new_line = (EDIT_LINE *) malloc(sizeof(EDIT_LINE))) == NULL) {
		fprintf(stderr, "Malloc of new line failed\n");
		return;
	}

	new_line->prev = NULL;
	new_line->next = NULL;
	len = strlen(line);

	if (line[len - 1] == '\n')
		new_line->length = len - 1; /* line is terminated by a \n
					       which we don't store */
	else
		new_line->length = len; /* no \n on line */
	
	new_line->alloced = new_line->length;

	if ((new_line->line =
	     (char *) malloc((new_line->length + 1) * sizeof(char))) == NULL) {
		fprintf(stderr, "Malloc of new line string failed\n");
		free(new_line);
		return;
	}

	strcpy(new_line->line, line);
	new_line->line[new_line->length] = '\0';

	if (current_line == NULL) {
		  /* buffer was empty, this must be the first line in */
		first_line = new_line;
		current_line = first_line;
		top_line = first_line; /* and it must be at top of screen */
		cursor_line = 0;
		cursor_pos = 0;
		screen_pos = 0;
	} else {
		new_line->prev = current_line;
		new_line->next = current_line->next;
		current_line->next = new_line;
		current_line = new_line;
	}
	return;
}

/*
 * Free all the allocated storage associated with the line given.  The
 * line pointer will be invalid after this function call.
 */
void
free_line(EDIT_LINE *line)
{
	free(line->line);
	line->line = NULL;
	line->prev = NULL;
	line->next = NULL;
	free(line);
}


/*
 * Destroy all the lines in the buffer, freeing all allocated storage.
 */
void
destroy_all_lines(void)
{
	EDIT_LINE *this_line, *np;

	this_line = first_line;

	while (this_line != NULL) {
		np = this_line->next;
		free_line(this_line);
		this_line = np;
	}

	current_line = NULL;
	first_line = NULL;
}

/*
 * Create and return an appropriately sized and positioned window for
 * the given menu.  If x is less than 0 then the function uses the
 * parent menu position to calculate the correct location of the
 * popup.  The parameter y must always be a valid location.  If either
 * drows or dcols are less than 0 then the child menu dimensions are
 * used to set the window size. The parent parameter may be NULL iff x
 * is greater than or equal to 0.  Returns NULL on error.
 */
WINDOW *
create_menu_popup(MENU *parent, MENU *child, int y, int x,
		  int drows, int dcols)
{
	int lhs, index, rows, cols;
	WINDOW *popup;
	
	scale_menu(child, &rows, &cols);
	if (drows > 0)
		rows = drows;
	if (dcols > 0)
		cols = dcols;

	if (x >= 0) {
		lhs = x;
	} else {
		if (parent == NULL) {
			fprintf(stderr, "Bad parent in create_menu_popup\n");
			return NULL;
		}
		
		index = item_index(current_item(parent));
		lhs = index * col_width;
		if (lhs > COLS)
			lhs = COLS - cols;
	}
	
	if ((popup = newwin(rows, cols, y, lhs)) == NULL) {
		fprintf(stderr, "Could not create popup win\n");
		return NULL;
	}

	keypad(popup, TRUE);
	nodelay(popup, FALSE);
	set_menu_sub(child, popup);
	return popup;
}

/*
 * Do a generic menu handling for the given menu in the given popup window.
 * This includes posting the menu, handling the key event loop and then
 * unposting the menu when done.  The name parameter is used when reporting
 * errors.
 */
void
handle_menu_popup(MENU *menu, WINDOW *popup, char *name)
{
	int c, done = 0, status;
	Menu_Hook callback;
	ITEM *current;

	
	if ((status = post_menu(menu)) < 0) {
		fprintf(stderr, "Post of %s menu failed (%d)\n", name, status);
		return;
	}

	wrefresh(popup);
	do {
		c = wgetch(popup);
		if (!handle_shortcut(c)) {
			switch (c) {
				case '\e':
					done = 1;
					break;

				case 'j':
				case '\t':
				case KEY_DOWN:
					menu_driver(menu, REQ_DOWN_ITEM);
					break;

				case 'k':
				case KEY_UP:
					menu_driver(menu, REQ_UP_ITEM);
					break;

				case '\n':
				case '\r':
					if ((current = current_item(menu)) != NULL) {
						callback = (Menu_Hook) item_userptr(current);
						if (callback != NULL) {
							callback(menu);
							menu_mode = 0;
							done = 1;
						}
					}

					break;

				case 'h':
				case 'l':
				case KEY_LEFT:
				case KEY_RIGHT:
					pushback = c;
					done = 1;
					break;
				
				default:
					menu_driver(menu, c);
					break;
			}
		}

		if (refresh_screen())
			touchwin(popup); /* redraw popup if screen updated */
		wrefresh(popup);
	} while (done == 0);
	
	unpost_menu(menu);
	return;
}

/*
 * Calculate the length of the displayed line allowing for any tab
 * characters that need to be expanded.  We assume that the tab stops
 * are 8 characters apart.  The parameter pos is the character position in
 * current_line we want to get the length of, the function returns the number
 * of characters from start of the line that the character at pos should
 * be displayed at after any intervening tabs have been expanded.
 */
int
tab_expanded_length(int pos)
{
	int len, i;

	len = 0;
	for (i = 0; i < pos; i++) {
		if (current_line->line[i] == '\t')
			len = len - (len % 8) + 8;
		else
			len++;
	}

	return len;
}

/*
 * Refresh the screen - if edit_win has been touched then force a redraw
 * of the top menu bar and filler.  Returns 1 if screen was redrawn.
 */
int
refresh_screen(void)
{
	if (is_wintouched(edit_win) == TRUE) {
		wrefresh(edit_win);
		touchwin(filler);
		touchwin(top_win);
		wrefresh(filler);
		wrefresh(top_win);
		return 1;
	}

	return 0;
}
