/*	$NetBSD: ungencat.c,v 1.1 2008/11/20 21:29:24 ginsbach Exp $	*/

/*
 * Copyright (c) 1996, 2007 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by J.T. Conklin and Brian Ginsbach.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD 
 *	  Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its 
 *    contributors may be used to endorse or promote products derived 
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
#if defined(__RCSID) && !defined(lint)
__RCSID("$NetBSD: ungencat.c,v 1.1 2008/11/20 21:29:24 ginsbach Exp $");
#endif

/***********************************************************
Copyright 1990, by Alfalfa Software Incorporated, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that Alfalfa's name not be used in
advertising or publicity pertaining to distribution of the software
without specific, written prior permission.

ALPHALPHA DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
ALPHALPHA BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

If you make any modifications, bugfixes or other changes to this software
we'd appreciate it if you could send a copy to us so we can keep things
up-to-date.  Many thanks.
				Kee Hinckley
				Alfalfa Software, Inc.
				267 Allston St., #3
				Cambridge, MA 02139  USA
				nazgul@alfalfa.com

******************************************************************/
#define _NLS_PRIVATE

#include <sys/types.h>
#include <sys/queue.h>

#include <netinet/in.h>	/* Needed by arpa/inet.h on NetBSD */
#include <arpa/inet.h>	/* Needed for htonl() on POSIX systems */

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <nl_types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

struct _msgT {
	long    msgId;
	char   *str;
        LIST_ENTRY(_msgT) entries;
};

struct _setT {
	long    setId;
        LIST_HEAD(msghead, _msgT) msghead;
        LIST_ENTRY(_setT) entries;
};

LIST_HEAD(sethead, _setT) sethead;
static struct _setT *curSet;

static	char   *xstrdup(const char *);
static	void   *xmalloc(size_t);

void	MCReadCat(int fd);
void	MCDumpCat(int fd);
void	MCAddMsg(int msgId, const char *msg);
void	MCAddSet(int setId);
void	usage(void);

#define CORRUPT			"corrupt message catalog"
#define NOMEMORY		"out of memory"

int convert;
int quote;
int Xflag;
char *sep;

static void *
xmalloc(size_t len)
{
	void   *p;

	if ((p = malloc(len)) == NULL)
		errx(1, NOMEMORY);
	return (p);
}

static char *
xstrdup(const char *str)
{
	char *nstr;

	if ((nstr = strdup(str)) == NULL)
		errx(1, NOMEMORY);
	return (nstr);
}

void
MCAddSet(int setId)
{
	struct _setT *p, *q;

	if (setId <= 0) {
		errx(1, "setId's must be greater than zero");
		/* NOTREACHED */
	}
	if (setId > NL_SETMAX) {
		errx(1, "setId exceeds limit");
		/* NOTREACHED */
	}

	p = sethead.lh_first;
	q = NULL;
	for (; p != NULL && p->setId < setId; q = p, p = p->entries.le_next);

	if (p && p->setId == setId) {
		;
	} else {
		p = xmalloc(sizeof(struct _setT));
		memset(p, '\0', sizeof(struct _setT));
		LIST_INIT(&p->msghead);

		p->setId = setId;

		if (q == NULL) {
			LIST_INSERT_HEAD(&sethead, p, entries);
		} else {
			LIST_INSERT_AFTER(q, p, entries);
		}
	}

	curSet = p;
}

void
MCAddMsg(int msgId, const char *str)
{
	struct _msgT *p, *q;

	if (!curSet)
		errx(1, "can't specify a message when no set exists");

	if (msgId <= 0) {
		errx(1, "msgId's must be greater than zero");
		/* NOTREACHED */
	}
	if (msgId > NL_MSGMAX) {
		errx(1, "msgID exceeds limit");
		/* NOTREACHED */
	}

	p = curSet->msghead.lh_first;
	q = NULL;
	for (; p != NULL && p->msgId < msgId; q = p, p = p->entries.le_next);

	if (p && p->msgId == msgId) {
		free(p->str);
	} else {
		p = xmalloc(sizeof(struct _msgT));
		memset(p, '\0', sizeof(struct _msgT));

		if (q == NULL) {
			LIST_INSERT_HEAD(&curSet->msghead, p, entries);
		} else {
			LIST_INSERT_AFTER(q, p, entries);
		}
	}

	p->msgId = msgId;
	p->str = xstrdup(str);
}

void
MCReadCat(int fd)
{
	void   *msgcat;		/* message catalog data */
	struct _nls_cat_hdr cat_hdr;
	struct _nls_set_hdr *set_hdr;
	struct _nls_msg_hdr *msg_hdr;
	char   *strings;
	int	m, n, s;
	int	msgno, setno;

	n = read(fd, &cat_hdr, sizeof(cat_hdr));
	if (n < sizeof(cat_hdr)) {
		if (n == 0)
			return;		/* empty file */
		else if (n == -1)
			err(1, "header read");
		else
			errx(1, CORRUPT);
	}
	if (ntohl(cat_hdr.__magic) != _NLS_MAGIC)
		errx(1, "%s: bad magic number (%#x)", CORRUPT, cat_hdr.__magic);

	cat_hdr.__mem = ntohl(cat_hdr.__mem);
	msgcat = xmalloc(cat_hdr.__mem);

	cat_hdr.__nsets = ntohl(cat_hdr.__nsets);
	cat_hdr.__msg_hdr_offset = ntohl(cat_hdr.__msg_hdr_offset);
	cat_hdr.__msg_txt_offset = ntohl(cat_hdr.__msg_txt_offset);
	if ((cat_hdr.__mem < 0) ||
	    (cat_hdr.__msg_hdr_offset < 0) ||
	    (cat_hdr.__msg_txt_offset < 0) ||
	    (cat_hdr.__mem < (cat_hdr.__nsets * sizeof(struct _nls_set_hdr))) ||
	    (cat_hdr.__mem < cat_hdr.__msg_hdr_offset) ||
	    (cat_hdr.__mem < cat_hdr.__msg_txt_offset))
		errx(1, "%s: catalog header", CORRUPT);

	n = read(fd, msgcat, cat_hdr.__mem);
	if (n < cat_hdr.__mem) {
		if (n == -1)
			err(1, "data read");
		else
			errx(1, CORRUPT);
	}

	set_hdr = (struct _nls_set_hdr *)msgcat;
	msg_hdr = (struct _nls_msg_hdr *)((char *)msgcat +
	    cat_hdr.__msg_hdr_offset);
	strings = (char *)msgcat + cat_hdr.__msg_txt_offset;

	setno = 0;
	for (s = 0; s < cat_hdr.__nsets; s++, set_hdr++) {
		set_hdr->__setno = ntohl(set_hdr->__setno);
		if (set_hdr->__setno < setno)
			errx(1, "%s: bad set number (%d)",
			     CORRUPT, set_hdr->__setno);
		setno = set_hdr->__setno;

		MCAddSet(setno);

		set_hdr->__nmsgs = ntohl(set_hdr->__nmsgs);
		set_hdr->__index = ntohl(set_hdr->__index);
		if (set_hdr->__nmsgs < 0 || set_hdr->__index < 0)
			errx(1, "%s: set header", CORRUPT);

		/* Get the data */
		msgno = 0;
		for (m = 0; m < set_hdr->__nmsgs; m++, msg_hdr++) {
			msg_hdr->__msgno = ntohl(msg_hdr->__msgno);
			msg_hdr->__offset = ntohl(msg_hdr->__offset);
			if (msg_hdr->__msgno < msgno)
				errx(1, "%s: bad message number (%d)",
				     CORRUPT, msg_hdr->__msgno);
		        if ((msg_hdr->__offset < 0) ||
			    ((strings + msg_hdr->__offset) >
			     ((char *)msgcat + cat_hdr.__mem)))
				errx(1, "%s: message header", CORRUPT);

			msgno = msg_hdr->__msgno;
			MCAddMsg(msg_hdr->__msgno, strings + msg_hdr->__offset);
		}
	}
	free(msgcat);
}

void
MCDumpCat(int fd)
{
	char *cp;
	FILE *ofp;
	struct _setT *set;
	struct _msgT *msg;

	if ((ofp = fdopen(fd, "w")) == NULL)
		err(1, "No message file open");

	LIST_FOREACH(set, &sethead, entries) {
		fprintf(ofp, "$set %ld\n", set->setId);
		LIST_FOREACH(msg, &set->msghead, entries) {
			fprintf(ofp, "%ld%s", msg->msgId, (sep)? sep : " ");
			if (quote)
				fputc(quote, ofp);
			for (cp = msg->str; *cp; cp++) {
				if ((!Xflag && iscntrl(((unsigned char)*cp))) ||
			            (convert && !isascii(((unsigned char)*cp))))
					switch (*cp) {
					case '\n':	/* newline */
						fprintf(ofp, "\\n");
						break;
					case '\t':	/* horizontal tab */
						fprintf(ofp, "\\t");
						break;
					case '\v':	/* vertical tab */
						fprintf(ofp, "\\v");
						break;
					case '\b':	/* backspace */
						fprintf(ofp, "\\b");
						break;
					case '\r':	/* carriage-return */
						fprintf(ofp, "\\r");
						break;
					case '\f':	/* form-feed */
						fprintf(ofp, "\\f");
						break;
					default:	/* bit pattern */
						fprintf(ofp, "\\%o", (int)*cp);
						break;
					}
				else
					fputc((int)*cp, ofp);
			}
			if (quote)
				fputc(quote, ofp);
			fputc('\n', ofp);
		}
	}
}

void
usage(void)
{

	fprintf(stderr, "Usage: %s [-CX] [-q c] [-s c] catfile [msgfile]\n",
		getprogname());
	exit(1);
}

int
main(int argc, char *argv[])
{
	int ifd, ofd;
	int c;
	char *catfile, *msgfile;

	while ((c = getopt(argc, argv, "Cq:s:X")) != -1) {
		switch (c) {
		case 'C':
			/* Convert non-ASCII to escape sequences (SCO) */
			convert = 1;
			break;
		case 'q':	/* quote */
			if (strlen(optarg) != 1)
				errx(1, "Bad quote character: %s", optarg);
			quote = optarg[0];
			break;
		case 's':	/* separator */
			if ((strlen(optarg) != 1) || !isblank(optarg[0]))
				errx(1, "Bad separator character: %s", optarg);
			sep = optarg;
			break;
		case 'X':
			/* Inhibit production of only X/Open-compatible
			 * sources  (SCO) */
			/* XXX what should this do? */
			Xflag = 1;
			break;
		default:
			usage();
			/* NOTREACHED */
		}
	}
	argc -= optind;
	argv += optind;

	if ((argc < 1) || (argc > 2))
		usage();

	catfile = *argv++;
	argc--;

	if ((catfile[0] == '-') && (catfile[1] == '\0')) {
		ifd = STDIN_FILENO;
	} else {
		if ((ifd = open(catfile, O_RDONLY)) < 0) {
			err(1, "Unable to open %s", catfile);
			/* NOTREACHED */
		}
	}

	if (argc == 1) {
		msgfile = *argv;
		ofd = open(msgfile, O_WRONLY | O_CREAT | O_EXCL, 0666);
		if (ofd < 0) {
			err(1, "Unable to create new %s", msgfile);
			/* NOTREACHED */
		}
	} else {
		ofd = STDOUT_FILENO;
	}

	MCReadCat(ifd);
	MCDumpCat(ofd);

	return 0;
}
