/* $Id: main.c,v 1.2 2009-01-19 03:42:49 christos Exp $ */

/*
 * Copyright (c) 2005 Manuel Freire.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>

#ifndef lint
__COPYRIGHT("@(#) Copyright (c) 2005 Manuel Freire");
__RCSID("$Id: main.c,v 1.2 2009-01-19 03:42:49 christos Exp $");
#endif

#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <paths.h>

#include "bpg.h"

#define MAX_CMDLEN 20
#define MAX_INPUTFILES 20
#define MAX_PATHLEN 200		/* revise this */

typedef struct bpginfo_t {
	int detached;
	int verbose;
	char *pub_file;
	char *sec_file;
	char *input_files[MAX_INPUTFILES];
	char *output_file;
	char *det_sig_file;
	BPG_CTX *opt;
} bpginfo_t;

typedef struct cmd_t {
	const char *name;
	int (*f)(void);
} cmd_t;

static void usage(void);
void clean(void);
int bpg_sign(void);
int bpg_verify(void);
int bpg_encrypt(void);
int bpg_symmenc(void);
int bpg_decrypt(void);
int bpg_encsig(void);

static cmd_t cmds[] = {
	{	"sign"		,	bpg_sign	},
	{	"verify"	,	bpg_verify	},
	{	"encrypt"	,	bpg_encrypt	},
	{	"symmenc"	,	bpg_symmenc	},
	{	"decrypt"	,	bpg_decrypt	},
	{	"encsig"	,	bpg_encsig	},
	{	""		,	0		},
};

bpginfo_t *info;

/*
 * The bpg(1) utility is a command line interface for the bpg(3) library. It
 * performs OpenPGP operations, such as sign, verify, encrypt or decrypt data
 * in a public key system.
 */
int
main(int argc, char *argv[])
{
	int ch, i;
	cmd_t *cmd;

	setprogname(argv[0]);

	/* parse command */
	if (argc < 2)
		usage();
	for (cmd = cmds; strncmp(cmd->name, argv[1], MAX_CMDLEN); cmd++)
		if (*cmd->name == 0)
			usage();
	argc--;
	argv++;

	/* parse options */
	if ((info = malloc(sizeof(bpginfo_t))) == NULL)
		err(EXIT_FAILURE, NULL);
	atexit(clean);
	info->detached = 0;
	info->verbose = 0;
	info->det_sig_file = 0;
	info->pub_file = 0;
	info->sec_file = 0;
	info->output_file = 0;		/* stdout */
	info->input_files[0] = NULL;	/* stdin */
	info->opt = BPG_CTX_new();
	while ((ch = getopt(argc, argv, "abdvc:h:p:o:r:u:z:D:U:R:P:Z:"))
	    != -1) {
		switch (ch) {
		case 'a':
		case 'c':
		case 'h':
		case 'p':
		case 'z':
		case 'Z':
			errx(EXIT_FAILURE, "-%c not yet supported", optopt);
			/* NOTREACHED */
		case 'D':
			info->detached = 1;
			info->det_sig_file = optarg;	/* dangerous */
			break;
		case 'P':
			/*
			 * VERY DANGEROUS OPTION: ps reveals passphrase;
			 * use a file with the password instead.
			 */
			info->opt->pass = optarg;	/* dangerous */
			break;
		case 'R':
			info->pub_file = optarg;	/* dangerous */
			break;
		case 'U':
			info->sec_file = optarg;	/* dangerous */
			break;
		case 'b':
			info->opt->ascii = 0;
			break;
		case 'd':
			info->detached = 1;
			break;
		case 'o':
			info->output_file = optarg;	/* dangerous */
			break;
		case 'r':
			info->opt->pub_uid = optarg;	/* dangerous */
			break;
		case 'u':
			info->opt->sec_uid = optarg;	/* dangerous */
			break;
		case 'v':
			info->verbose = 1;
			break;
		case '?':
		default:
			usage();
			/* NOTREACHED */
		}
	}
	argc -= optind;
	argv += optind;

	/* parse input files */
	for (i = 0; argc > 0 && i < MAX_INPUTFILES; i++, argc--, argv++)
		info->input_files[i] = *argv;
	info->input_files[i] = NULL;

	return (*cmd->f)();
}

static void
usage(void)
{

#if 0
	/* not quite yet */
	(void)fprintf(stderr,
"usage: %s ( sign | verify | encrypt | symmenc | decrypt | encsig ) ...\n",
		      getprogname());
#endif
	(void)fprintf(stderr, "usage: %s ( sign | verify ) ...\n",
		      getprogname()); 

	exit(EXIT_FAILURE);
}

void
clean(void)
{

	if (info != NULL) {
		BPG_CTX_free(info->opt);
		free(info);
	}
}

int
bpg_sign(void)
{
	int ret;

	if (info->sec_file == 0 && strcmp(info->opt->sec_uid, "00000000") == 0)
		errx(EXIT_FAILURE,
			"-u <identity> or -U <priv key file> is required");

	if (info->input_files[0] != NULL && info->input_files[1] != NULL)
		warnx("Multiple files not yet supported, signing `%s' only",
		      info->input_files[0]);

	if (info->sec_file != NULL) {		/* -U option set */
		if (info->detached) {
			if (info->verbose)
				warnx("Signing `%s' (detached).",
				      info->input_files[0]);
			if (info->verbose)
				warnx("Using key `%s'.", info->sec_file);
			ret = BPG_sign_detached_k(info->input_files[0],
						  info->sec_file,
						  info->output_file, info->opt);
		} else {
			if (info->verbose)
				warnx("Signing `%s'.", info->input_files[0]);
			if (info->verbose)
				warnx("Using key `%s'.", info->sec_file);
			ret = BPG_sign_k(info->input_files[0], info->sec_file,
					 info->output_file, info->opt);
		}
	} else {				/* -U option unset */
		if (info->detached) {
			if (info->verbose)
				warnx("Signing `%s' (detached).",
				      info->input_files[0]);
			if (info->verbose)
				warnx("Using key `%s'.", info->opt->sec_uid);
			ret = BPG_sign_detached(info->input_files[0],
						info->output_file, info->opt);
		} else {
			if (info->verbose)
				warnx("Signing `%s'.", info->input_files[0]);
			if (info->verbose)
				warnx("Using key `%s'.", info->opt->sec_uid);
			ret = BPG_sign(info->input_files[0],
				       info->output_file, info->opt);
		}
	}

	if (ret)
		errx(EXIT_FAILURE, BPG_get_error_string());

	return EXIT_SUCCESS;
}

int
bpg_verify(void)
{
	int ret;

	if (info->pub_file == 0 && strcmp(info->opt->pub_uid, "00000000") == 0)
		errx(EXIT_FAILURE, "-r or -R is required");

	if (info->input_files[0] != NULL && info->input_files[1] != NULL)
		warnx("Multiple files not yet supported, verifying `%s' only",
		      info->input_files[0]);

	if (info->pub_file != NULL) {			/* -R option set */
		if (info->detached) {
			if (info->det_sig_file == 0)
				errx(EXIT_FAILURE, "-D option required.");
			if (info->verbose)
				warnx("Verifying `%s' (detached) with `%s'.",
				      info->input_files[0], info->det_sig_file);
			ret = BPG_verify_detached_k(info->input_files[0],
						    info->det_sig_file,
						    info->pub_file, info->opt);
		} else {
			if (info->verbose)
				warnx("Verifying `%s'.", info->input_files[0]);
			ret = BPG_verify_k(info->input_files[0], info->pub_file,
					   info->opt);
		}
	} else {					/* -R option unset */
		if (info->detached) {
			if (info->det_sig_file == 0)
				errx(EXIT_FAILURE, "-D option required.");
			if (info->verbose)
				warnx("Verifying `%s' (detached) with `%s'.",
				      info->input_files[0], info->det_sig_file);
			ret = BPG_verify_detached(info->input_files[0],
						  info->det_sig_file,
						  info->opt);
		} else {
			if (info->verbose)
				warnx("Verifying `%s'.", info->input_files[0]);
			ret = BPG_verify(info->input_files[0], info->opt);
		}
	}

	if (ret < 0)
		errx(EXIT_FAILURE, BPG_get_error_string());
	if (ret == 0)
		warnx("***** Invalid signature of file `%s'! *****",
		      info->input_files[0]);
	if (info->verbose)
		warnx("Valid signature of file `%s'.", info->input_files[0]);

	return ret;
}

int
bpg_encrypt(void)
{

	errx(EXIT_FAILURE, "Encryption not yet supported");
}

int
bpg_symmenc(void)
{

	errx(EXIT_FAILURE, "Symmetric encryption not yet supported");
}

int
bpg_decrypt(void)
{

	errx(EXIT_FAILURE, "Decryption not yet supported");
}

int
bpg_encsig(void)
{

	errx(EXIT_FAILURE, "Encryption and signing not yet supported");
}

