/* $NetBSD: bpg.c,v 1.1.1.1 2008/05/18 18:31:41 agc Exp $ */

/*
 * Copyright (c) 2005 Manuel Freire.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>

#ifndef lint
__COPYRIGHT("@(#) Copyright (c) 2005 Manuel Freire. All rights reserved.");
__RCSID("$Id: bpg.c,v 1.2 2009-01-19 03:42:17 christos Exp $");
#endif

#include <sys/types.h>
#include <sys/wait.h>

#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <openssl/dsa.h>
#include <openssl/engine.h>
#include <openssl/err.h>
#include <openssl/evp.h>
#include <openssl/rand.h>
#include <openssl/rsa.h>

#include <paths.h>

#include "constants.h"
#include "data.h"
#include "error.h"
#include "packet.h"
#include "bpgutil.h"

#include "bpg.h"

const char *BPG_version = "Summer-of-Code";

/* Internal functions prototypes. */
int BPG_do_sign_DSA(BPG_BUF *, BPG_BUF *, DSA *);

/* Sign file `msg_p' into `dst_p'. */
int
BPG_sign(char *msg_p, char *dst_p, BPG_CTX *opt)
{
	/*
	 * FIXME: this function is a patch for using gpg in key extraction.
	 * When the bpg-key library is available it will be used instead.
 	 */

	int fd, status, ret;
	BPG_CTX *ctx;
	
	warnx("GnuPG is used for key extraction.");
	if (opt)
		ctx = BPG_CTX_new();
	else
		ctx = opt;

	fd = fork();
	switch (fd) {
	case 0:
		execlp("gpg", "gpg", "-o", "/tmp/bpg-seckey.tmp",
		       "--export-secret-key", opt->sec_uid, NULL);
		exit(errno);
		/*FALLTHROUGH*/
	case -1:
		return -1;
	default:
		wait(&status);
		if (status)
			return -1;
	}

	ret = BPG_sign_k(msg_p, "/tmp/bpg-seckey.tmp", dst_p, opt);

	if (opt == NULL)
		BPG_CTX_free(ctx);
	remove("/tmp/bpg-seckey.tmp");

	return ret;
}

/* Create a detached signature of file `msg_p' in `dst_p'. */
int
BPG_sign_detached(char *msg_p, char *dst_p, BPG_CTX *opt)
{
	/*
	 * FIXME: this function is a patch for using gpg in key extraction.
	 * When the bpg-key library is available it will be used instead.
 	 */

	int fd, status, ret;
	BPG_CTX *ctx;
	
	warnx("GnuPG is used for key extraction.");
	if (opt)
		ctx = BPG_CTX_new();
	else
		ctx = opt;

	fd = fork();
	switch (fd) {
	case 0:
		execlp("gpg", "gpg", "-o", "/tmp/bpg-seckey.tmp",
		       "--export-secret-key", opt->sec_uid, NULL);
		exit(errno);
		/*FALLTHROUGH*/
	case -1:
		return -1;
	default:
		wait(&status);
		if (status)
			return -1;
	}

	ret = BPG_sign_detached_k(msg_p, "/tmp/bpg-seckey.tmp", dst_p, opt);

	if (opt == NULL)
		BPG_CTX_free(ctx);
	remove("/tmp/bpg-seckey.tmp");

	return ret;
}

/* Sign `msg_p' into `dst_p' using secret key in file `skey_p'. */
int
BPG_sign_k(char *msg_p, const char *skey_p, char *dst_p, BPG_CTX *opt)
{
	FILE *fm, *fs, *fd;
	BPG_BUF *msg, *skey, *dst, *onepass, *literal;
	BPG_PKT *pkt_onepass, *pkt_literal;

	if (msg_p == NULL)
		fm = stdin;
	else
		if ((fm = fopen(msg_p, "r")) == NULL)
			return -1;
	if ((fs = fopen(skey_p, "r")) == NULL)
			return -1;
	if ((msg = BPG_BUF_new(0)) == NULL)
		return -1;
	if ((skey = BPG_BUF_new(0)) == NULL)
		return -1;
	if ((dst = BPG_BUF_new(0)) == NULL)
		return -1;
	if ((onepass = BPG_BUF_new(0)) == NULL)
		return -1;
	if ((literal = BPG_BUF_new(0)) == NULL)
		return -1;
	if ((pkt_onepass = BPG_PKT_new()) == NULL)
		return -1;
	if ((pkt_literal = BPG_PKT_new()) == NULL)
		return -1;
	if (BPG_BUF_read_f(msg, fm) == -1)
		return -1;
	if (BPG_BUF_read_f(skey, fs) == -1)
		return -1;
	if (msg_p != NULL)
		fclose(fm);
	fclose(fs);

	if (BPG_do_sign(msg, skey, dst, opt) == -1)
		return -1;

	if (dst_p == NULL)
		fd = stdout;
	else
		if ((fd = fopen(dst_p, "w")) == NULL)
			return -1;

	/* Write signature: One-Pass Signature + Literal Packet + Signature. */
	/* FIXME: next block is a patch I cannot fix now because of time. */
	if (BPG_ONEPASS_build(onepass, opt) == -1)
		return -1;
	if (BPG_BUF_read(pkt_onepass->body, onepass->body, onepass->len) == -1)
		return -1;
	BPG_PKT_build(pkt_onepass, PKTTYPE_ONEPASS, opt);
	BPG_PKT_write(pkt_onepass, onepass);
	if (BPG_LITERAL_build(msg, literal, msg_p) == -1)
		return -1;
	if (BPG_BUF_read(pkt_literal->body, literal->body, literal->len) == -1)
		return -1;
	BPG_PKT_build(pkt_literal, PKTTYPE_LITERAL, opt);
	BPG_PKT_write(pkt_literal, literal);

	fwrite(onepass->body, 1, onepass->len, fd);
	fwrite(literal->body, 1, literal->len, fd);
	fwrite(dst->body, 1, dst->len, fd);

	if (dst_p != NULL)
		fclose(fd);

	BPG_BUF_free(msg);
	BPG_BUF_free(skey);
	BPG_BUF_free(dst);
	BPG_BUF_free(onepass);
	BPG_BUF_free(literal);
	BPG_PKT_free(pkt_onepass);
	BPG_PKT_free(pkt_literal);

	return 0;
}

/* Create a detached signature of `msg_p' in `dst_p' using key in `skey_p'. */
int
BPG_sign_detached_k(char *msg_p, const char *skey_p, char *dst_p, BPG_CTX *opt)
{
	FILE *fm, *fs, *fd;
	BPG_BUF *msg, *skey, *dst;

	if (msg_p == NULL)
		fm = stdin;
	else
		if ((fm = fopen(msg_p, "r")) == NULL)
			return -1;
	if ((fs = fopen(skey_p, "r")) == NULL)
			return -1;
	if ((msg = BPG_BUF_new(0)) == NULL)
		return -1;
	if ((skey = BPG_BUF_new(0)) == NULL)
		return -1;
	if ((dst = BPG_BUF_new(0)) == NULL)
		return -1;
	if (BPG_BUF_read_f(msg, fm) == -1)
		return -1;
	if (BPG_BUF_read_f(skey, fs) == -1)
		return -1;
	if (msg_p != NULL)
		fclose(fm);
	fclose(fs);

	if (BPG_do_sign(msg, skey, dst, opt) == -1)
		return -1;

	if (dst_p == NULL)
		fd = stdout;
	else
		if ((fd = fopen(dst_p, "w")) == NULL)
			return -1;
	fwrite(dst->body, 1, dst->len, fd);
	if (dst_p != NULL)
		fclose(fd);

	BPG_BUF_free(msg);
	BPG_BUF_free(skey);
	BPG_BUF_free(dst);

	return 0;
}

/* Write in `dst' a signature of `msg' using the secret key in `skey'. */
int
BPG_do_sign(BPG_BUF *msg, BPG_BUF *skey, BPG_BUF *dst, BPG_CTX *opt)
{
	int stype;
	bpg_time_t t;
	BPG_CTX *ctx;
	BPG_SECKEY *seckey;
	BPG_SIG *sig;
	BPG_BUF *to_hash, *hash;
	EVP_MD_CTX hctx;
	const EVP_MD *htype;

	if (opt == NULL) {	/* use defaults */
		if ((ctx = BPG_CTX_new()) == NULL)
			return -1;
	} else
		ctx = opt;
	if ((seckey = BPG_SECKEY_new()) == NULL)
		return -1;
	if ((sig = BPG_SIG_new()) == NULL)
		return -1;
	if ((to_hash = BPG_BUF_new(0)) == NULL)
		return -1;
	if ((hash = BPG_BUF_new(0)) == NULL)
		return -1;

	if (BPG_SECKEY_read_packet(seckey, skey->body, skey->len, ctx) == -1) {
		BPG_err(BPGERR_COULDNT_PARSE_SECKEY);
		return -1;
	}
	stype = SIGTYPE_BINARY_DOC;	/* FIXME: other types? */
	t = bpg_time_get();

	/* Hash */
	switch (ctx->sigv) {
	case SIGVERSION_3:
		BPG_BUF_resize(to_hash, 5);
		to_hash->body[0] = stype;
		util_uint32_to_str(t, &(to_hash->body[1]));
		break;
	case SIGVERSION_4:
		BPG_err(BPGERR_SIGVERSION_4_NOT_SUPPORTED);
		return -1;
	default:
		BPG_err(BPGERR_SIGVERSION_UNKNOWN);
		return -1;
	}
	switch (ctx->hash_algo) {
	case HA_SHA1:
		BPG_BUF_resize(hash, SHA1_SIZE);
		if (ctx->pubkey_algo == PKA_DSA)
			htype = EVP_dss1();	/* dsa requires dss1 */
		else
			htype = EVP_sha1();
		break;
	default:
		BPG_err(BPGERR_ALGORITHM_NOT_SUPPORTED);
		return -1;
	}
	EVP_MD_CTX_init(&hctx);	
	if (EVP_DigestInit_ex(&hctx, htype, NULL) == 0) {
		BPG_err(BPGERR_OPENSSL_DIGEST);
		return -1;
	}
	if (EVP_DigestUpdate(&hctx, msg->body, msg->len) == 0) {
		BPG_err(BPGERR_OPENSSL_DIGEST);
		return -1;
	}
	if (EVP_DigestUpdate(&hctx, to_hash->body, to_hash->len) == 0) {
		BPG_err(BPGERR_OPENSSL_DIGEST);
		return -1;
	}
	if (EVP_DigestFinal_ex(&hctx, hash->body, &(hash->len)) == 0) {
		BPG_err(BPGERR_OPENSSL_DIGEST);
		return -1;
	}
	EVP_MD_CTX_cleanup(&hctx);

	/* Sign */
	RAND_seed("SDJA0O103UJ0WAE98RUA093UR02Q3", 20); /* FIXME */
	switch (ctx->pubkey_algo) {
	case PKA_DSA:
		sig->dsa_sig = DSA_do_sign(hash->body, (int)hash->len,
			seckey->dsa);
		if (sig->dsa_sig == NULL)
			return -1;
		break;
	default:
		BPG_err(BPGERR_ALGORITHM_NOT_SUPPORTED);
		return -1;
	}
	
	/* Build packet */
	sig->version = ctx->sigv;
	switch (sig->version) {
	case SIGVERSION_3:
		sig->type = stype;
		sig->creation_time = t;
		(void)memcpy(sig->key_id, ctx->sec_kid, 8);
		sig->pubkey_algo = ctx->pubkey_algo;
		sig->hash_algo = ctx->hash_algo;
		(void)memcpy(sig->hash_start, hash->body, 2);
		break;
	case SIGVERSION_4:
		BPG_err(BPGERR_SIGVERSION_4_NOT_SUPPORTED);
		return -1;
	default:
		BPG_err(BPGERR_SIGVERSION_UNKNOWN);
		return -1;
	}
	BPG_SIG_build_packet(sig, dst, ctx);

	BPG_SECKEY_free(seckey);
	BPG_SIG_free(sig);
	if (opt == NULL)
		BPG_CTX_free(ctx);
	BPG_BUF_free(to_hash);
	BPG_BUF_free(hash);

	return 0;
}

/* Verify signed file `smsg_p'. */
int
BPG_verify(char *smsg_p, BPG_CTX *opt)
{
	/*
	 * FIXME: this function is a patch for using gpg in key extraction.
	 * When the bpg-key library is available it will be used instead.
 	 */

	int fd, status, ret;
	BPG_CTX *ctx;
	
	warnx("GnuPG is used for key extraction.");
	if (opt)
		ctx = BPG_CTX_new();
	else
		ctx = opt;

	fd = fork();
	switch (fd) {
	case 0:
		execlp("gpg", "gpg", "-o", "/tmp/bpg-pubkey.tmp", "--export",
		       opt->pub_uid, NULL);
		exit(errno);
		/*FALLTHROUGH*/
	case -1:
		return -1;
	default:
		wait(&status);
		if (status)
			return -1;
	}

	ret = BPG_verify_k(smsg_p, "/tmp/bpg-pubkey.tmp", opt);

	if (opt == NULL)
		BPG_CTX_free(ctx);
	remove("/tmp/bpg-pubkey.tmp");

	return ret;
}

/* Verify detached signature `sig_p' of file `msg_p'. */
int
BPG_verify_detached(char *msg_p, char *sig_p, BPG_CTX *opt)
{
	/*
	 * FIXME: this function is a patch for using gpg in key extraction.
	 * When the bpg-key library is available it will be used instead.
 	 */

	int fd, status, ret;
	BPG_CTX *ctx;
	
	warnx("GnuPG is used for key extraction.");
	if (opt)
		ctx = BPG_CTX_new();
	else
		ctx = opt;

	fd = fork();
	switch (fd) {
	case 0:
		execlp("gpg", "gpg", "-o", "/tmp/bpg-pubkey.tmp", "--export",
		       opt->pub_uid, NULL);
		exit(errno);
		/*FALLTHROUGH*/
	case -1:
		return -1;
	default:
		wait(&status);
		if (status)
			return -1;
	}

	ret = BPG_verify_detached_k(msg_p, sig_p, "/tmp/bpg-pubkey.tmp", opt);

	if (opt == NULL)
		BPG_CTX_free(ctx);
	remove("/tmp/bpg-pubkey.tmp");

	return ret;
}

/* Verify signed file `smsg_p' using the public key in `pkey_p'. */
int
BPG_verify_k(char *smsg_p, const char *pkey_p, BPG_CTX *opt)
{
	int ret, nread;
	FILE *fm, *fp;
	BPG_BUF *smsg, *msg, *sig, *pkey;
	BPG_PKT *pkt;

	if (smsg_p == NULL)
		fm = stdin;
	else
		if ((fm = fopen(smsg_p, "r")) == NULL)
			return -1;
	if ((fp = fopen(pkey_p, "r")) == NULL)
		return -1;
	if ((smsg = BPG_BUF_new(0)) == NULL)
		return -1;
	if ((msg = BPG_BUF_new(0)) == NULL)
		return -1;
	if ((sig = BPG_BUF_new(0)) == NULL)
		return -1;
	if ((pkey = BPG_BUF_new(0)) == NULL)
		return -1;
	if ((pkt = BPG_PKT_new()) == NULL)
		return -1;
	if (BPG_BUF_read_f(smsg, fm) == -1)
		return -1;
	if (BPG_BUF_read_f(pkey, fp) == -1)
		return -1;
	if (smsg_p != NULL)
		fclose(fm);
	fclose(fp);

	/* Read one-pass signature (TODO: don't ignore this packet). */
	if (BPG_PKT_read(pkt, smsg->body, smsg->len) == -1)
		return -1;
	nread = pkt->full_length;

	/* Read literal data. */
	if (BPG_PKT_read(pkt, &smsg->body[nread], smsg->len - nread) == -1)
		return -1;
	nread += pkt->full_length;
	if (BPG_LITERAL_read(msg, pkt->body) == -1)
		return -1;

	/* Read signature. */
	if (BPG_BUF_read(sig, &smsg->body[nread], smsg->len - nread) == -1)
		return -1;

	ret = BPG_do_verify(msg, sig, pkey, opt);

	BPG_BUF_free(smsg);
	BPG_BUF_free(msg);
	BPG_BUF_free(sig);
	BPG_BUF_free(pkey);

	return ret;
}

/* Verify detached signature `sig_p' of file `msg_p' using key in `pkey_p'. */
int
BPG_verify_detached_k(char *msg_p, char *sig_p, const char *pkey_p, BPG_CTX *opt)
{
	int ret;
	FILE *fm, *fs, *fp;
	BPG_BUF *msg, *sig, *pkey;

	if (msg_p == NULL)
		fm = stdin;
	else
		if ((fm = fopen(msg_p, "r")) == NULL)
			return -1;
	if ((fs = fopen(sig_p, "r")) == NULL)
		return -1;
	if ((fp = fopen(pkey_p, "r")) == NULL)
		return -1;
	if ((msg = BPG_BUF_new(0)) == NULL)
		return -1;
	if ((sig = BPG_BUF_new(0)) == NULL)
		return -1;
	if ((pkey = BPG_BUF_new(0)) == NULL)
		return -1;
	if (BPG_BUF_read_f(msg, fm) == -1)
		return -1;
	if (BPG_BUF_read_f(sig, fs) == -1)
		return -1;
	if (BPG_BUF_read_f(pkey, fp) == -1)
		return -1;
	if (msg_p != NULL)
		fclose(fm);
	fclose(fs);
	fclose(fp);

	ret = BPG_do_verify(msg, sig, pkey, opt);

	BPG_BUF_free(msg);
	BPG_BUF_free(sig);
	BPG_BUF_free(pkey);

	return ret;
}

/* Verify signature of buffer `msg' using key in `pkey'.  `opt' is ignored. */
/*ARGSUSED3*/
int
BPG_do_verify(BPG_BUF *msg, BPG_BUF *sig, BPG_BUF *pkey, BPG_CTX *opt)
{
	int ret;
	BPG_SIG *signature;
	BPG_PUBKEY *pubkey;
	BPG_BUF *to_hash, *hash;
	EVP_MD_CTX hctx;
	const EVP_MD *htype;

	if ((signature = BPG_SIG_new()) == NULL)
		return -1;
	if ((pubkey = BPG_PUBKEY_new()) == NULL)
		return -1;
	if ((to_hash = BPG_BUF_new(0)) == NULL)
		return -1;
	if ((hash = BPG_BUF_new(0)) == NULL)
		return -1;

	if (BPG_SIG_read_packet(signature, sig->body, sig->len) == -1)
		return -1;
	if (BPG_PUBKEY_read_packet(pubkey, pkey->body, pkey->len) == -1)
		return -1;

	/* Calculate hash. */
	switch (signature->version) {
	case SIGVERSION_3:
		BPG_BUF_resize(to_hash, 5);
		to_hash->body[0] = signature->type;
		util_uint32_to_str(signature->creation_time,
		    &(to_hash->body[1]));
		break;
	case SIGVERSION_4:
		BPG_err(BPGERR_SIGVERSION_4_NOT_SUPPORTED);
		return -1;
	default:
		BPG_err(BPGERR_SIGVERSION_UNKNOWN);
		return -1;
	}
	switch (signature->hash_algo) {
	case HA_SHA1:
		BPG_BUF_resize(hash, SHA1_SIZE);
		if (signature->pubkey_algo == PKA_DSA)
			htype = EVP_dss1();	/* dsa requires dss1 */
		else
			htype = EVP_sha1();
		break;
	default:
		BPG_err(BPGERR_ALGORITHM_NOT_SUPPORTED);
		return -1;
	}
	EVP_MD_CTX_init(&hctx);	
	if (EVP_DigestInit_ex(&hctx, htype, NULL) == 0) {
		BPG_err(BPGERR_OPENSSL_DIGEST);
		return -1;
	}
	if (EVP_DigestUpdate(&hctx, msg->body, msg->len) == 0) {
		BPG_err(BPGERR_OPENSSL_DIGEST);
		return -1;
	}
	if (EVP_DigestUpdate(&hctx, to_hash->body, to_hash->len) == 0) {
		BPG_err(BPGERR_OPENSSL_DIGEST);
		return -1;
	}
	if (EVP_DigestFinal_ex(&hctx, hash->body, &(hash->len)) == 0) {
		BPG_err(BPGERR_OPENSSL_DIGEST);
		return -1;
	}
	EVP_MD_CTX_cleanup(&hctx);

	/* Verification. */
	if (memcmp(hash->body, signature->hash_start, 2) != 0)
		return 0;
	switch (signature->pubkey_algo) {
	case PKA_DSA:
		ret = DSA_do_verify(hash->body, (int)hash->len,
			signature->dsa_sig, pubkey->dsa);
			/* TODO: use openssl_evp(3) */
		break;
	default:
		BPG_err(BPGERR_ALGORITHM_NOT_SUPPORTED);
		return -1;
	}

	BPG_SIG_free(signature);
	BPG_PUBKEY_free(pubkey);
	BPG_BUF_free(to_hash);
	BPG_BUF_free(hash);

	return ret;
}

/* Encrypt file `msg_p' into `dst_p'. */
/* ARGSUSED0 */
int
BPG_encrypt(char *msg_p, char *dst_p, BPG_CTX *opt)
{

	BPG_err(BPGERR_NOT_IMPLEMENTED);
	return -1;
}

/* Encrypt `msg_p' into `dst_p' using the key in `pkey_p'. */
/* ARGSUSED0 */
int
BPG_encrypt_k(char *msg_p, char *pkey_p, char *dst_p, BPG_CTX *opt)
{

	BPG_err(BPGERR_NOT_IMPLEMENTED);
	return -1;
}

/* Encrypt buffer `msg' into `dst' using the key in `pkey'. */
/* ARGSUSED0 */
int
BPG_do_encrypt(BPG_BUF *msg, BPG_BUF *pkey, BPG_BUF *dst, BPG_CTX *opt)
{

	BPG_err(BPGERR_NOT_IMPLEMENTED);
	return -1;
}

/* Symmetrically encrypt file `msg_p' into `dst_p'. */
/* ARGSUSED0 */
int
BPG_symmenc(char *msg_p, char *dst_p, BPG_CTX *opt)
{

	BPG_err(BPGERR_NOT_IMPLEMENTED);
	return -1;
}

/* Symmetrically encrypt buffer `msg' into `dst'. */
/* ARGSUSED0 */
int
BPG_do_symmenc(BPG_BUF *msg, BPG_BUF *dst, BPG_CTX *opt)
{

	BPG_err(BPGERR_NOT_IMPLEMENTED);
	return -1;
}

/* Decrypt file `msg_p' into `dst_p'. */
/* ARGSUSED0 */
int
BPG_decrypt(char *msg_p, char *dst_p, BPG_CTX *opt)
{

	BPG_err(BPGERR_NOT_IMPLEMENTED);
	return -1;
}

/* Decrypt file `msg_p' into `dst_p' using the key in `skey_p'. */
/* ARGSUSED0 */
int
BPG_decrypt_k(char *msg_p, char *skey_p, char *dst_p, BPG_CTX *opt)
{

	BPG_err(BPGERR_NOT_IMPLEMENTED);
	return -1;
}

/* Decrypt buffer `msg' into `dst' using the key in `skey'. */
/* ARGSUSED0 */
int
BPG_do_decrypt(BPG_BUF *msg, BPG_BUF *skey, BPG_BUF *dst, BPG_CTX *opt)
{

	BPG_err(BPGERR_NOT_IMPLEMENTED);
	return -1;
}

/* Encrypt and sign file `msg_p' into `dst_p'. */
/* ARGSUSED0 */
int
BPG_encrypt_sign(char *msg_p, char *dst_p, BPG_CTX *opt)
{

	BPG_err(BPGERR_NOT_IMPLEMENTED);
	return -1;
}

/* Encrypt and sign `msg_p' into `dst_p' using keys in `pkey_p' and `skey_p'. */
/* ARGSUSED0 */
int
BPG_encrypt_sign_k(char *msg_p, char *pkey_p, char *skey_p, char *dst_p,
		   BPG_CTX *opt)
{

	BPG_err(BPGERR_NOT_IMPLEMENTED);
	return -1;
}

/* Encrypt and sign `msg' into `dst' using keys in `pkey' and `skey'. */
/* ARGSUSED0 */
int
BPG_do_encrypt_sign(BPG_BUF *msg, BPG_BUF *pkey, BPG_BUF *skey, BPG_BUF *dst,
		    BPG_CTX *opt)
{

	BPG_err(BPGERR_NOT_IMPLEMENTED);
	return -1;
}

/* Allocate a new bpg context with defaults set. */
BPG_CTX *
BPG_CTX_new(void)
{
	BPG_CTX *ctx;

	if ((ctx = malloc(sizeof(BPG_CTX))) == NULL)
		return NULL;

	ctx->cipher = BPG_DEFAULT_CIPHER;
	ctx->pubkey_algo = BPG_DEFAULT_PKA;
	ctx->hash_algo = BPG_DEFAULT_HASH;
	ctx->pktv = BPG_DEFAULT_PKTV;
	ctx->sigv = BPG_DEFAULT_SIGV;
	ctx->keyv = BPG_DEFAULT_KEYV;
	ctx->pass = NULL;
	ctx->pub_uid = strdup("00000000");	/* FIXME: set the default key ID. */
	ctx->sec_uid = strdup("00000000");	/* FIXME: set the default key ID. */
	/*ctx->pub_kid = ;*/	/* FIXME: set the default key ID. */
	/*ctx->sec_kid = ;*/	/* FIXME: set the default key ID. */
	ctx->pub_path = strdup("");
	ctx->sec_path = strdup("");
	
	return ctx;
}

/* Free a bpg context. */
void
BPG_CTX_free(BPG_CTX *ctx)
{

	if (ctx != NULL)
		free(ctx);
}

/* Allocate a new buffer. */
BPG_BUF *
BPG_BUF_new(size_t len)
{
	BPG_BUF *buf;

	if ((buf = malloc(sizeof(BPG_BUF))) == NULL)
		return NULL;

	buf->body = NULL;
	BPG_BUF_resize(buf, len);
	buf->pos = 0;

	return buf;
}

/* Free a buffer. */ 
void
BPG_BUF_free(BPG_BUF *buf)
{

	if (buf != NULL) {
		if (buf->body != NULL)
			free(buf->body);
		free(buf);
	}
}

/* Reallocate buffer.  Buffer length is set to `len'. */
int
BPG_BUF_resize(BPG_BUF *buf, size_t len)
{

	if (buf == NULL) {
		BPG_err(BPGERR_NULL_POINTER);
		return -1;
	}

	if (len) {
		if ((buf->body = realloc(buf->body, len)) == NULL) {
			BPG_err(BPGERR_MALLOC);
			return -1;
		}
	}

	buf->len = (unsigned int)len;

	return 0;
}

/* Print a buffer, formatted. */
void
BPG_BUF_print(BPG_BUF *buf, FILE *f)
{
	int i;

	if (buf != NULL && f != NULL) {
		(void)fprintf(f, "Buffer (%u bytes):", buf->len);
		for (i = 0; i < buf->len; i++) {
			if (i % 16 == 0)
				(void)fprintf(f, "\n  ");
			(void)fprintf(f, "%02x ", buf->body[i]);
		}
		(void)fprintf(f, "\n");
	}
}

/* Read `flen' bytes from `from'. */
int
BPG_BUF_read(BPG_BUF *buf, unsigned char *from, size_t flen)
{

	if (buf == NULL || from == NULL) {
		BPG_err(BPGERR_NULL_POINTER);
		return -1;
	}

	if (BPG_BUF_resize(buf, flen) == -1)
		return -1;

	memcpy(buf->body, from, flen);

	return 0;
}

/* Dumps stream `f' into a buffer. */
int
BPG_BUF_read_f(BPG_BUF *buf, FILE *f)
{
	size_t nread;

	if (f == NULL || buf == NULL) {
		BPG_err(BPGERR_NULL_POINTER);
		return -1;
	}

	BPG_BUF_resize(buf, 100);
	while ((nread = fread(&buf->body[buf->len - 100], 1, 100, f)) == 100)
		BPG_BUF_resize(buf, buf->len + 100);
	BPG_BUF_resize(buf, buf->len - 100 + nread);

	return 0;
}

